
/*
 * @copyright
 *
 *  Copyright 2014 Neo Natura
 *
 *  This file is part of the Share Library.
 *  (https://github.com/neonatura/share)
 *        
 *  The Share Library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version. 
 *
 *  The Share Library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with The Share Library.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  @endcopyright
 */  

#include <math.h>
#include "shcoind.h"
#include "fractal.h"
#include "spring.h"

/* see end of file */
extern uint64_t _spring_matrix[256][256];

typedef struct spring_bits
{
  shnum_t lat;
  shnum_t lon;
  unsigned int x, y;
  unsigned int lat1, lat2;
  unsigned int lon1, lon2;
} spring_bits;

/*not needed */
extern uint8_t* shsha1_result(shsha1_t *s);

static unsigned int spring_matrix_bit(uint64_t seed)
{
  shsha1_t s;
  uint8_t *result;
  uint32_t val;
  unsigned int *i_ptr;
  int i;

  seed = htonll(seed);

  memset(&s, 0, sizeof(s));
  shsha1_init(&s);
  shsha1_write(&s, (const char *)&seed, sizeof(seed));
  result = shsha1_result(&s);

  val = 0;
  i_ptr = (uint32_t *)result;
  for (i = 0; i < 5; i++) {
    val += i_ptr[i];
  }

  return (val % 32);
}

static void spring_loc_bits(spring_bits *bits)
{
  uint32_t abs_lat, abs_lon;
  uint64_t b_lat, b_lon;
  shnum_t lat;
  shnum_t lon;

  lat = bits->lat;
  lon = bits->lon;

  /* collapse hemisphere into 'uni-sphere' */
  if (lat < 0) lat = lat * -1;
  if (lon < 0) lon = lon * -1;

  abs_lat = (uint32_t)lat;
  abs_lon = (uint32_t)lon;

  /* transition offset and ratio used to match matrix size */
  bits->y = (int)((lat - SPRING_OFFSET_LATITUDE) * SPRING_Y_FACTOR) % 256;
  bits->x = (int)((lon - SPRING_OFFSET_LONGITUDE) * SPRING_X_FACTOR) % 256;

  b_lat = (uint64_t)((lat - (shnum_t)abs_lat) * 1000.0);
  bits->lat1 = spring_matrix_bit(b_lat) % 32;
  bits->lat2 = spring_matrix_bit(b_lat/8) % 32;

  b_lon = (uint64_t)((lon - (shnum_t)abs_lon) * 1000.0);
  bits->lon1 = spring_matrix_bit(b_lon) % 32;
  bits->lon2 = spring_matrix_bit(b_lon/8) % 32;
}



void spring_loc_set(double lat, double lon)
{
  spring_bits bits;
  uint32_t *i_ar;

  memset(&bits, 0, sizeof(bits));
  bits.lat = lat;
  bits.lon = lon;
  spring_loc_bits(&bits);

  i_ar = (uint32_t *)&_spring_matrix[bits.y][bits.x];
  i_ar[0] |= (1 << bits.lat1);
  i_ar[0] |= (1 << bits.lat2);
  i_ar[1] |= (1 << bits.lon1);
  i_ar[1] |= (1 << bits.lon2);
}


int is_spring_loc(double lat, double lon)
{
  spring_bits bits;
  uint32_t *i_ar;

  memset(&bits, 0, sizeof(bits));
  bits.lat = lat;
  bits.lon = lon;
  spring_loc_bits(&bits);

  if (bits.lat1 == bits.lat2 || bits.lon1 == bits.lon2)
    return (FALSE);

  i_ar = (uint32_t *)&_spring_matrix[bits.y][bits.x];
  if (
      (i_ar[0] & (1 << bits.lat1)) &&
      (i_ar[0] & (1 << bits.lat2)) &&
      (i_ar[1] & (1 << bits.lon1)) &&
      (i_ar[1] & (1 << bits.lon2))
     ) {
    return (TRUE);
  }

  return (FALSE);
}

void spring_loc_claim(double lat, double lon)
{
  spring_bits bits;
  uint32_t *i_ar;

  memset(&bits, 0, sizeof(bits));
  bits.lat = lat;
  bits.lon = lon;
  spring_loc_bits(&bits);

  if (bits.lat1 == bits.lat2 || bits.lon1 == bits.lon2)
    return;

  /* remove first bits */
  i_ar = (uint32_t *)&_spring_matrix[bits.y][bits.x];
  if (
      (i_ar[0] & (1 << bits.lat1)) &&
      (i_ar[0] & (1 << bits.lat2)) &&
      (i_ar[1] & (1 << bits.lon1)) &&
      (i_ar[1] & (1 << bits.lon2))
     ) {
    i_ar[0] &= ~(1 << bits.lat1);
    i_ar[1] &= ~(1 << bits.lon1);
  }

}

int spring_loc_search(double cur_lat, double cur_lon, double *lat_p, double *lon_p)
{
  double s_lat, e_lat;
  double s_lon, e_lon;
  double lat, lon;
  int total;
  int found;

  s_lat = cur_lat - 0.25;
  s_lon = cur_lon - 0.25;
  e_lat = cur_lat + 0.25;
  e_lon = cur_lon + 0.25;

  for (lat = s_lat; lat <= e_lat; lat += 0.001) {
    for (lon = s_lon; lon <= e_lon; lon += 0.001) {
      if (is_spring_loc(lat, lon)) {
        *lat_p = lat;
        *lon_p = lon;
        return (0);
      }
    }
  }

  return (SHERR_NOENT);
}

void spring_matrix_compress(uint32_t matrix[3][3])
{
  int row, col;
  int y, x;

  for (row = 0; row < 3; row++) {
    for (col = 0; col < 3; col++) {
      matrix[row][col] = 0;
    }
  }

  for (y = 0; y < 256; y++) {
    for (x = 0; x < 256; x++) {
      row = (int)(y / 85) % 3;
      col = (int)(x / 85) % 3;
      matrix[row][col] += (uint32_t)_spring_matrix[y][x];
    }
  }

}

#if 0
int spring_render_fractal(char *img_path, double zoom, double span, double x_of, double y_of)
{
  static const unsigned int width = 256, height = 256;
  BMP *bmp;
  uint32_t m_seed;
  uint32_t val;
  shnum_t seed;
  shnum_t rate;
  shnum_t K;
  shnum_t Z;
  shnum_t C;
  shnum_t min_cord;
  shnum_t max_cord;
  shnum_t dx, dy;
  unsigned int x, y;
  int cval;
  int idx;
  unsigned char r, g, b;
  int px_width, px_height;

  px_width = (int)((shnum_t)256 / span);
  px_height = (int)((shnum_t)256 / span);
  bmp = BMP_Create(px_width, px_height, 32);

  m_seed = 0;
  for (y = 0; y < height; y++) {
    for (x = 0; x < width; x++) {
      m_seed += _spring_matrix[y][x];
    }
  }
  //seed = (shnum_t)m_seed / 16;
  seed = (shnum_t)m_seed;
  seed = seed / 65536;

  zoom = MAX(0.001, MIN(100.0, zoom));
  rate = span * zoom;
  min_cord = (-128 * zoom);
  max_cord = (128 * zoom);
  for (dy = min_cord; dy <= max_cord; dy += rate) {
    for (dx = min_cord; dx <= max_cord; dx += rate) {
      C = (dx * dy) + seed;

      Z = 0;
      for (idx = 0; idx < 16; idx++) {
        Z = Z * Z + C;
        if (fabs(Z) >= 2)
          break;
      }
      if (idx == 16) {
        r = g = b = 0;
      } else {
        K = fabs(Z*C);
        val = 4294967296 % (uint32_t)(K+1);
        r = (val >> 16) & 0xff;
        g = (val >> 8) & 0xff;
        b = val & 0xff;

        r = MIN(255, r + idx);
        g = MIN(255, g + idx);
        b = MIN(255, b + idx);
      }
      x = (int)((dx + max_cord) / rate) % px_width;
      y = (int)((dy + max_cord) / rate) % px_height;
      BMP_SetPixelRGB( bmp, x, y, r, g, b);
    }
  }
  BMP_WriteFile(bmp, img_path);
  BMP_CHECK_ERROR( stdout, -2 );
  BMP_Free( bmp );
  return 0;
}
#endif
int spring_render_fractal(char *img_path, double zoom, double span, double x_of, double y_of)
{
  static const int width = 256, height = 256;
  uint32_t m_seed;
  double seed;
  int y, x;

  m_seed = 0;
  for (y = 0; y < height; y++) {
    for (x = 0; x < width; x++) {
      m_seed += _spring_matrix[y][x];
    }
  }
  seed = (double)m_seed;

  return (fractal_render(img_path, seed, zoom, span, x_of, y_of));
}




/**
 * A 256x256 matrix with 'dual' bitvectors initialized to common US locations.
 */
uint64_t _spring_matrix[256][256] = {
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200200000000500,0x0,0x4008080000,0x2004008008000,0x9100020000404000,0xc00000080400,0x0,0x0,0x100000880080000,0x1210000008088100,0x0,0x0,0x0,0x0,0x0,0x8000000102000100,0x480000000005000,0x0,0x0,0x0,0x0,0x101005000000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc702420a0886600,0x4122784201022700,0x4000000210000400,0x40504001002600,0xa21801b10d482200,0x10154302262400,0x800000280000400,0x0,0x0,0x200040020004000,0x38002010000500,0x2002000005000,0x4002000008080000,0x0,0x2100080000400,0x0,0x0,0x0,0x0,0x0,0x10400000300000,0x9000100c49007000,0x4484018020244000,0x50108000000500,0x0,0x101000002200,0x1004000000000500,0x600e001880400,0x1020c20401882600,0x427e10812a2c00,0x210040809100020,0xe000400005060,0x0,0x0,0x41000000002400,0x0,0x120084000840c000,0x2010000000120,0x410000000300000,0x0,0x0,0x0,0x8040000002200,0x2040002000400,0x0,0x0,0x0,0x0,0x20040008884000,0x80000400000060,0x8002000000000400,0x0,0x0,0x0,0x4000200000300000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100002000880000,0x0,0x0,0x0,0x204000004002,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1200204000222800,0x10400000022000,0x0,0x281c00400022400,0x448610000822a800,0x438008100a80800,0x48000880000,0x20080000022000,0x0,0x10100008004000,0x80010008008000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa00026002,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x50200082002000,0x4010141a00920120,0xd0100800024520,0x0,0x0,0x0,0x18200120020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8004000000000060,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x44000000140,0x0,0x0,0x6060000100460,0x0,0x0,0x8000880000080920,0x0,0x0,0x0,0x804000080020,0x204701020021c0,0x0,0x20800080800,0x8040000000440,0x0,0x4008102000020140,0x4000001012000000,0x0,0x0,0x0,0x0,0xba00244116820560,0x2804010004100400,0x0,0x4200042004060840,0x0,0x0,0x80000412000000,0x8005a0006080d00,0x0,0x0,0x3200002000004440,0x0,0x0,0x40200000420,0x0,0x0,0x100400000000500,0x800000400000140,0x0,0x10402192000440,0x88000000020020,0x0,0x200200000000500,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1800002002000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6000800040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x204000800040,0x0,0x0,0x200000400200400,0x0,0x8102000c00,0x0,0x0,0x0,0x0,0x4000010000800200,0x0,0x0,0x80000200800040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc00000000800,0x208000000800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x81c454832182404,0x480000002000100,0x0,0x0,0x0,0x800880201082000,0x4000002044000000,0x8000060022140000,0x0,0x0,0x100080000002800,0x4000010002000004,0x0,0x8000400001002000,0x5000200254000004,0x0,0x0,0x0,0x0,0x0,0x101002002000,0xa00005210802100,0x2020080100000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x405008c0140040,0x20004002008000,0x4002000000080400,0x0,0x0,0x0,0x1420f440c2005446,0x200240280040040,0x40803042000044,0x200108882800040,0x0,0x8000040002000100,0x210000010040000,0x45000082100000,0x108080001000,0x24048000000,0x100d0084140000,0x400004002040000,0x0,0x0,0x200040020100000,0x2800000108,0x0,0x1000080040000400,0x940020412040044,0x8804052040082440,0x0,0x0,0x8000200044000000,0x2000840000080,0x10040040000080,0x0,0x100002048000000,0x8200000000800040,0x0,0x8400480046040000,0xc12404a2820500c4,0x5001004012000104,0x0,0x0,0x8000080000400004,0x0,0x0,0x0,0x0,0x2200000040004,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2002082000000,0x0,0x0,0x0,0x0,0x0,0x210000010000040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20020090000000,0x0,0x11000002800,0x0,0x48090000000,0x0,0x0,0x8000004004100000,0x0,0x0,0x0,0x18000002800,0x8100000090000000,0x0,0x0,0x0,0x100002044000000,0x0,0x100080002000004,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x204004100000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8010100101104208,0x641808000110c210,0x13206a000b02210,0x8200001000001,0x8000090001100200,0x8080000000c000,0x4000400800004,0x0,0x8010000000004080,0x0,0x0,0x0,0x0,0x1080000000400020,0x0,0x0,0x0,0x0,0x4008004000,0x0,0x8000010008000020,0x0,0x80c10088002000,0xc08020608800030a,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2004408002800208,0x400100000800001,0x0,0x6000040000002,0x0,0x0,0x0,0x0,0x0,0x800020000000c000,0x40e0008c0004282,0x98040001020020,0x2020001004000,0x0,0x0,0x0,0x80100800001,0x4040824080044280,0x20240088000002,0x4000000200c00030,0x100020008000100,0x0,0x400006400006480,0x41200d408880260a,0x12000002000,0x0,0x840014040800012,0x10002088000a00,0x84000088000000,0x4480004408082100,0x6000000008000020,0x0,0x2020000008004000,0x0,0x8302000008020300,0x0,0x20080000002400,0x0,0x0,0x0,0x1100d2208000e440,0x2000821080002440,0x0,0x2800400020,0x0,0x600000002001,0x0,0x0,0x420080000010,0x0,0x8000020000400020,0x0,0x110c40208c06081,0x841f4e7a0ac0e2a3,0x4000100000000300,0x12000002020,0x4020000004080,0x10000200002020,0x0,0x1004000000020200,0x0,0x10020000020200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x800000808000020,0x0,0x41001004000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x800010000200200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200280000040,0x6000000000200200,0x8100800002,0xc10002080810040,0x0,0x8100000000200200,0x0,0x0,0x1004040808404040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80100000200200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4092302202402122,0xa88a2d40cb04b10e,0x808000680000c00a,0x0,0x20000201000200,0x0,0x0,0x2000004000004008,0x0,0x0,0x0,0x0,0x24000004002,0x0,0x0,0x8102040080a100,0x0,0x8010000000440000,0x0,0x1000004001004000,0x0,0x8000140000040,0x80010000008100,0x804080200000,0x0,0xa0000400008,0x0,0x0,0x0,0x0,0x100002101004102,0x0,0x4080001004000,0x10002000008100,0x0,0x0,0x0,0x300000400008,0x0,0x0,0x200101000200,0x802020040000,0x0,0x0,0x100000840004000,0xa00008400000,0x0,0x0,0x0,0x26082004108,0x44000500000,0x800010001004000,0x0,0x0,0x2800018000,0xa00440050004000,0x0,0x80000208000002,0x2100000000060,0x0,0x4000001000400008,0x1000000440004000,0x0,0x0,0x200020080000002,0x0,0x202000604000c100,0x4000000000040004,0x108000004040,0x0,0x0,0x20040000008008,0x0,0x48030004000c108,0x40201000200,0x80040000804000,0x100080001008000,0x240000002100,0x50000000004002,0x5001000200,0x400100001008000,0x200200002100,0x0,0x2045000040019000,0x44e54408045c00a,0xb08e56b3eb50c112,0x9001004000848004,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x900400882500000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80000808000002,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10400008400000,0x4000200000000102,0x0,0x0,0x0,0x8040000000060,0x0,0x0,0x0,0x0,0x8080a48408000,0x410402040008000,0x200040000000102,0x0,0x0,0x2000240000040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400024020100c000,0x8000050080018000,0x80400000020040,0x201000001020,0x0,0x0,0x0,0x0,0x0,0x0,0x120000018000,0x0,0x0,0x0,0x4800001040,0xa88862c8040ca01,0x0,0x0,0x20800005000,0x100020040800000,0x0,0x0,0x0,0x60040000080,0x0,0x0,0x81080280208000,0x4d08100405200,0x8000048000007080,0xa24040280208301,0x830065202203f020,0x801003202084a100,0x20044200038800,0x18005000000d200,0x0,0xc000400280003080,0x0,0x248805010418a00,0x404010400000,0xb18c20804002a149,0x41428020d040,0x120000008200,0x800000400020040,0x80040000018000,0x10001000000a00,0x4102080000005201,0x0,0x840000010400000,0x2100000000001040,0x0,0x2186d50c2223281,0xc006286a81c2e020,0x4602000090400000,0x0,0x0,0x82400488001d000,0x240000020800,0x0,0x0,0x0,0x0,0x0,0x200200005000,0x0,0x0,0x9000000000802000,0x0,0x0,0x0,0x820804004000c004,0x8000002040000020,0x0,0x3004000200808100,0x48000ca0040c820,0x800010010400000,0x0,0x0,0x0,0xa00002080,0x0,0x0,0x300000001020,0x0,0x0,0x8214440000400a00,0xc1845820802da08,0x5080200080041002,0x0,0x4000040000404000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8400000008200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x50c00040040006,0x0,0x800080000008200,0x0,0x0,0x0,0x0,0x8040000101000,0x0,0x80000008200,0x81000040000004,0x2100000008200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1080100000008200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2200020800808220,0x0,0x8002000400200,0x0,0x0,0x0,0x0,0x100000208000800,0x0,0x8002020100422200,0xc00000808000,0x0,0x48000000001,0x1480244002000a09,0x5880612a88e20201,0x20800022000,0x0,0x0,0x0,0x0,0x800020000010200,0x0,0x0,0x0,0x1004000000000600,0x802200210001201,0x4000400a00000,0x0,0x0,0x808004c81c20000,0x88490244013200,0x0,0x30960108432200,0x4000800000010200,0x8000020200022010,0x80c00408000206,0x0,0x0,0x8200000081000,0x14004020000,0x108000020010,0x0,0x0,0x0,0x288050000022210,0x80800000101000,0xa781048801201,0x18420002000221,0x1000000406000000,0x9fa487047e23801,0x3028044a00a00601,0x8100004002000,0x0,0x0,0x2204100810201,0x4000001000080001,0x820040000001,0x140008000004,0x108904020600,0x8002008000080,0x8000004000004800,0x0,0x0,0x0,0x0,0x0,0x0,0x40020008000001,0xa004020000,0x8000002000400200,0x20004000060000,0x0,0x0,0x0,0x0,0x0,0x200011440a00001,0x2200000004800,0x410000000800001,0x0,0x0,0x0,0x842000000e20000,0x480002000001,0x1a0036800905a48,0x8000200000a00000,0x0,0x0,0x0,0x401004800000,0x0,0x4002004020000,0x0,0x0,0x120000000010200,0x0,0x2004000004800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa00008000100,0x200000408000100,0x0,0x0,0x400002000a00000,0x0,0x0,0x0,0x0,0x0,0x8100202001080801,0x0,0x0,0x0,0x0,0x30420000a60000,0x88000000a00000,0x8100000004002000,0x0,0x8088101000800,0x0,0x0,0x4000001000060000,0x0,0xa00080001,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x840000000060000,0x0,0x0,0x1a01220010e0801,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000120000080,0x2200020000080,0x200040000000050,0x0,0x0,0x0,0x0,0x0,0x84021000000,0x0,0x0,0x0,0x0,0x20040000000006,0x0,0x240800200010c0,0x8000002000400010,0x0,0xc000000001040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4080000000082,0x0,0x500024000000,0xb00411560741040,0x0,0x8001000400008,0x0,0x1080000000400008,0x7000000220108400,0x0,0x61021021000,0x800010021000000,0x0,0x10001000100200,0x0,0x14000600000,0x80010000000006,0x8010000000003000,0x6000000400010,0x100400000200040,0x8042000020420080,0x401000040040,0x841002020600080,0x284a44028101442,0x8084640020805060,0x802109023400088,0x220c000020808000,0x48000000000000a,0x1100040021001020,0x0,0x10800001400,0x0,0x0,0x840000000400010,0x20000222000000,0x0,0x80080000021000,0x10400000001020,0x0,0x4000000200001004,0x0,0x0,0x0,0x0,0x0,0x0,0x80100000003000,0x0,0x5000000022000000,0x50020008000,0x6000000000001400,0x0,0x100000200101000,0x0,0x0,0x800400000000050,0x4080000020000400,0x140000001400000,0x500020000200,0x0,0x8000010000001400,0x200400000000000a,0x0,0x0,0x0,0x0,0x0,0x8000000200200040,0x0,0x0,0x0,0x0,0x0,0x4040000040040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6000000000021000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200080460000080,0x0,0x480000000001020,0x0,0x0,0x0,0x0,0x0,0x0,0x4000200060000000,0x2090004020420080,0x4020820020000080,0x6000000420000,0x0,0x0,0x18020000080,0x0,0x4000800000021000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200a8020008080,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x9044040144800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8400000060000,0x0,0x0,0x0,0x0,0x202000200800,0x14008000800,0x480000200800,0x0,0x0,0xc00000008002,0x0,0x4400140010c4000,0x0,0x8200000002040000,0x0,0x0,0x10100002000100,0x0,0x804a40213100400,0x4500000101804800,0x4000000000800800,0x0,0x480000004100000,0x1100060002108002,0x72822414130ec580,0x0,0x0,0x0,0x0,0x0,0x0,0x2100000004100000,0x200001004004000,0x402240c2000800,0x400004000800800,0x0,0x2302002084800,0x0,0x1000000400200800,0xa95088300804,0x400201080000,0x4480a010021c0804,0x4400012000000,0x401000082000000,0x0,0x0,0x500008000040,0x0,0x400010001000800,0x0,0x0,0x0,0x0,0x0,0x1000200012000000,0x200101080000,0x200000100004,0x0,0x0,0x40040000040800,0x2100002040000,0x0,0x0,0x2200000000808000,0x840000000040800,0x0,0x8088c8001a042000,0x42200801084800,0x0,0x0,0x100040012000000,0x4000008000040800,0x0,0x6480042080800,0x20004000008002,0x100000282000000,0x0,0x40001000014000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4280000800,0x0,0x0,0x2200002100000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x42004004000,0x6000000008000040,0x0,0x811020200c084840,0x0,0x81000008000040,0x0,0x0,0x100002004004000,0x8010a00000060000,0x0,0x10100004104000,0x0,0x0,0xa00004004000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80004004000,0x0,0x0,0x0,0x140002040000,0x8010200004064000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4100000004000400,0x0,0x0,0x220000000000000a,0x0,0x20040008000100,0x4100000000120,0x0,0x0,0x0,0x0,0x0,0x100810010200402,0x0,0x0,0x500202000100408,0x2400201000000a,0x20402000010,0x2000800040000100,0x0,0x50210080000300,0x40200000104,0x0,0x0,0x828000200100020,0x0,0x0,0x0,0x8020000000000120,0x0,0x20102000400,0x850680228080501,0x441405e010288613,0x408008000400,0x101000002,0x0,0x0,0x240000008001,0x0,0x0,0x0,0x0,0x0,0x400820000100,0x110020000002,0x8400000000005,0x0,0x404040000020,0x0,0x842000808000601,0x10020000020040,0x804008000100,0x8020240020100400,0x110000020000100,0x60002000010,0x0,0x0,0x8010804001000442,0x0,0x0,0x40000800200010,0x8400004000400,0x0,0x0,0x100800002000010,0x0,0x0,0x1000200000000408,0x0,0x8020000001000002,0x2000008001000002,0x0,0x44021000a000110,0x2004002000400,0x0,0x4000400000120,0x0,0x10001000000440,0x0,0x0,0x41000000040020,0x3084c002100030,0x440000020000002,0x300000080020,0x0,0x0,0x480010000002,0x0,0x0,0x240000100020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x800002000000000a,0x0,0x0,0x100802000100,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x280000000a,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8080000000101,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa0000000000a,0x10100000000408,0x0,0x0,0x0,0x200c0002000501,0x4000400100020,0x0,0x0,0x4100202002000501,0x0,0x280002000400,0x10100000000101,0xa02000400,0x0,0x200540000000509,0x80000000408,0x8082280202080501,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x208440002000400,0x0,0x0,0x0,0x0,0x1200000002000010,0x10000200000000a,0x801000000000000a,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x860010810b00180,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400004000000280,0x802004400000,0x4800004040,0x849010880400,0x0,0x0,0x0,0xa042a000203020c4,0x0,0x0,0x0,0x8010200000124200,0x8020000000120000,0x0,0x0,0x0,0x2000400000804000,0x6000020000040,0x5000804000,0x0,0x0,0x0,0x0,0x0,0x6020200000,0x0,0x5485202104e802e8,0x4000800000022000,0x0,0x0,0x0,0x42000084000,0x0,0x0,0x0,0x40100000800400,0x0,0x0,0x600200008a0008,0x0,0x0,0x6000000020008,0x1000002000020020,0x0,0x140000000200040,0x0,0x0,0x0,0x200200002040,0x1400000000060,0x0,0x8020000000280,0x0,0x0,0x0,0x0,0x0,0x40400000804000,0x9418400040010e4,0x0,0x0,0x100800000028,0x0,0xc000200004881000,0x0,0x10100000004040,0x0,0x0,0x8000100000000028,0x0,0xa102000000006060,0x0,0x8000004000002040,0x12000000140,0x2000e00001002c0,0x10040000020008,0x14000101000,0x0,0x800002020000040,0x2020004400000,0x0,0x0,0x201004010000,0xa00000084000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x140000000000060,0x0,0x0,0x88000000084000,0xc00000120000,0x0,0x0,0x40d0740000b82060,0x0,0x8100000000000280,0x0,0x0,0x24000880000,0x0,0x0,0x0,0x0,0x2800120000,0x40000002040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x280000000060,0x0,0x0,0x0,0x0,0x40000280000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10040000020,0x0,0x0,0x0,0x0,0x0,0x4000000204080000,0xc204000004a8,0x50000000001400,0x0,0x8034260220905030,0x1000010c04040480,0x4000000000100200,0x141080084141400,0x0,0x0,0x4080000000801000,0x0,0x0,0x88810040000460,0x4000001000040080,0x0,0x8010000000000028,0x0,0x21000104000,0x0,0x0,0x0,0x80100000001008,0x0,0x80400000401000,0x300000000021,0x20004000104000,0x0,0x800004000801000,0x0,0x0,0x80000001001,0x0,0x60000000000021,0x0,0x0,0x8c00200000d000,0x0,0x160104040600,0x2004000401000,0x100002004000800,0x0,0x0,0x108000001008,0x1000031004040004,0x0,0x800000204000004,0x10001000401000,0x2000800005000,0x0,0x14000010020,0x20080004080000,0x80000404000004,0x0,0x0,0x2800000000c000,0x820000801000,0x0,0x0,0x0,0x0,0x1002000001400,0x2004020000020,0x0,0x8000100000000040,0x10020000000040,0x0,0x0,0x0,0x0,0x24000801000,0x0,0x8400000020000020,0x1200000080100000,0x8020000010040000,0x0,0x0,0x0,0x0,0x408000000001001,0x200000114000000,0xc00080001000,0x24000010000400,0x0,0x0,0x8004008100000,0x0,0x0,0x800040004000004,0x0,0x0,0x20004000801000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1004000010000400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2800801000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2000800008100000,0x0,0x0,0x0,0x0,0x0,0x0,0x2808104000,0x800040000000440,0x0,0x80080080100000,0x24002000040,0x204000004000,0x8001000004000800,0x0,0x0,0x800080040000020,0x204000104000,0x20080040000020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x500040000020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x21000000140,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x204220010000680,0x100400000010040,0x10000220040000,0x8002001000040,0x88000002400000,0x8000010002400000,0x11000004040,0x1114001400a044,0x80100002040,0x0,0x280210080880200,0xc00002400000,0x80000100000140,0x0,0x0,0x0,0x0,0x2080002400000,0x0,0x0,0x0,0x80010000080,0x4000400000002002,0x0,0x0,0x20080010800140,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x142810048000,0x0,0x100002000400100,0x0,0x0,0x0,0x100500002400600,0x0,0x700800041002000,0x10000680802000,0x4002001800000,0x1080500044000044,0x6010000080,0x0,0x0,0x0,0x40010001000002,0x0,0x100000100840000,0x180000800,0x0,0x0,0x0,0x8000000110800000,0x200004000400200,0x440000010000080,0x2000002040002000,0x0,0x8400000010040,0x0,0x0,0x20004002400000,0x20004001000400,0x0,0x800100000008040,0x0,0x0,0x1000001080000040,0x200010000400100,0x50000080000040,0x2100000000220,0x0,0x200201000002,0x0,0x0,0x0,0x0,0x0,0x200200040040000,0x0,0x0,0x1000080000008040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000100040000004,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80000800000210,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000800080002000,0x10100001000002,0x0,0x0,0x0,0x0,0x4000001040040000,0x0,0x60080000040,0x0,0x0,0x940800090002080,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2130598089402222,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1100008004000,0x6000008010,0x0,0x0,0x10100000300000,0x0,0x1002000000400,0x2010040200000,0x8000014080282000,0x8000000100080100,0x40001001000400,0x0,0x4008012000000,0x0,0x0,0x0,0x800004a080008810,0x120a0090020100,0x0,0x0,0x0,0x20040000020100,0x8010000008008000,0x0,0x8000480800202810,0x0,0x4400000240000,0x0,0x0,0x81000000080100,0x800008002000100,0x400400000080100,0x40100000020100,0x0,0x20090298004000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40020000240000,0x0,0x0,0x0,0x0,0x0,0x0,0x1400100484080000,0x100080012000000,0x408000200400,0x0,0x108180090000400,0x0,0x0,0x0,0x0,0x0,0x211040008008500,0x0,0x0,0x0,0x0,0x0,0x22000008008000,0x0,0x0,0x0,0x800010090000000,0x0,0x0,0x0,0x8000100000400,0x480080000800,0x0,0x8000004000002001,0x0,0x21000084000,0x80004c0080082400,0x24001000400,0x0,0x0,0x0,0x0,0x200020090000000,0x0,0x1004000000008010,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x208000008004000,0x0,0x0,0x0,0x0,0x0,0x0,0x400100000080100,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2880100000,0x0,0x0,0x0,0x0,0x0,0x280000880000,0x0,0x6000080100000,0x0,0x80100000000400,0x0,0x8100000000880000,0x0,0x40004090000000,0x0,0x0,0x100000100080100,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4480302000880000,0x0,0x80000200200800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200000100020,0x0,0x0,0x0,0x0,0x0,0x80940000160020,0x420000022000,0x0,0x200000402000400,0x0,0x800004000008400,0x40800000400,0x0,0x800040009000000,0x0,0x0,0x0,0x1204000000022140,0x1200000000100100,0x10800022000,0x0,0x6002000200800500,0x0,0x500000020040,0x0,0x0,0x0,0x0,0x0,0x800002000000440,0x42000001800000,0x0,0x0,0x108000000000140,0x0,0x0,0x0,0x28000000500,0x0,0x8002001000200,0x0,0x80100000002200,0x0,0x6008008000,0x0,0x0,0x2001040800000,0x0,0x0,0x80010000000140,0x0,0x0,0x0,0x424008000000580,0xc000280009800100,0x8000062000062200,0x8040020000020820,0x800080000840000,0x404000000020400,0x0,0x9100004000120420,0x50620000860120,0x0,0x0,0x8200000000440,0x0,0x0,0x0,0x0,0x0,0x80050000424020,0x0,0x2000800001800000,0x0,0x0,0xa00000000005000,0x0,0x820001000200,0x0,0x0,0x20200000000440,0x0,0x80c49f0a00027642,0x10441002800440,0x2010000840040,0x50000000140,0x4000001020004000,0x0,0x8002000000020400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x408000008008000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x42002000400,0x0,0x0,0x0,0x18002000400,0x0,0x0,0x0,0x1080000000004002,0x0,0x0,0x2d00000100100,0x0,0x0,0x4100000008400000,0x20020000022000,0x0,0x0,0x0,0x6000022000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80100002000400,0x0,0x0,0x140022002004402,0x208040002004402,0x0,0x500000022000,0x81000002000400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1000020040000040,0x0,0x0,0x201044000000,0x8200000002000400,0xc00030000000,0x0,0x0,0x0,0x0,0x84010104a000000,0x20004048000000,0x80408448040004,0x0,0x800040002800000,0x0,0x0,0x44000842004,0x0,0x808052042040460,0x164040210180450,0x14082000000,0x0,0x4002820044100580,0x2010000030000000,0x0,0x0,0x0,0x0,0x80010000080800,0x140000020800,0x0,0x0,0x0,0x0,0x0,0x6000000040000040,0x44000012000000,0x8000248000000,0x100804000240404,0x0,0x408000040004,0x400000804100000,0x0,0x0,0x80004000100000,0x0,0x60648420420c0,0x200000500,0xc000000040040,0x0,0x8010000000100000,0x0,0x0,0x80040040000040,0x200040000000420,0x0,0x1800000000420,0x2100002002000,0x4000002a00102800,0x0,0x6104082800000,0x0,0x0,0x0,0x0,0x0,0x8000400000000402,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x24000000200400,0x1200000040000400,0x0,0x0,0x100400000080400,0x48002000040,0x100080030000000,0x120002040000,0x0,0x0,0x120000000020800,0x0,0x0,0x0,0x0,0x44001002000,0x0,0x0,0x80400020100000,0x8000000240000040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20004000082000,0x0,0x0,0x0,0x0,0x10400000800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000400030000000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x140000000002800,0x0,0x8000200000002800,0x0,0x0,0x0,0x800088104100800,0x0,0x0,0x8100000000002800,0x0,0x0,0x400010000000420,0x0,0x0,0x0,0x10020000080400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x810000044000000,0x0,0x0,0x0,0x0,0x0,0x80000000800,0x8040000200400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x802090000000,0x630024200a300,0xa000000082000000,0x11000000000a00,0x0,0x0,0x100020000002400,0x80800040000002,0x0,0x0,0x200180004000,0x40001002800000,0x0,0x820000000000a00,0x4200282040004,0x480060008a000100,0x8000000208000100,0x0,0x0,0x0,0x0,0x24002282801000,0x7800082810000,0x0,0x0,0x0,0x0,0x41000008000100,0x0,0x0,0x0,0x0,0x400000802000004,0x0,0x440000000002400,0x200088000000,0x0,0x8000400008000008,0x0,0x0,0x0,0x0,0x0,0x0,0x48008400000,0x100000800040200,0x20040000000300,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80010000002000,0x400004008000020,0x800030802800110,0x50021088081000,0x1004042090004020,0x20080008000008,0x0,0xc00002000004,0x0,0x0,0x200100008200,0x0,0x0,0x0,0x80040010000100,0x8000000000002020,0x0,0x4020010000100,0x0,0x0,0x0,0x0,0x8000001000002001,0x0,0x0,0x0,0x2004000200002,0x0,0x40200080000002,0x0,0x0,0x100000880001000,0x0,0x2080002000004,0x0,0x0,0x0,0x0,0x80000202000004,0x0,0x0,0x24082000000,0x0,0x0,0x180002000,0x5000800002,0x0,0x0,0x0,0x0,0x0,0x0,0x2000020000002020,0x0,0x0,0x0,0x4000010088000000,0x4000200040002000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x28000002000004,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000200080000040,0x0,0x200000402000004,0x0,0x6010040080810040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80004000810000,0x0,0x0,0x28000090000000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x104c0080008240,0x80100008000008,0x2010000000800002,0xc800400a8210c200,0x840000000810000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2100000000208,0x0,0x0,0x40040050000000,0x800020001004000,0x0,0x0,0x1000540080202020,0x0,0x0,0x0,0x10020000008008,0x0,0xc000000000108,0x0,0x80400000400100,0x840000000008100,0x2000200400020,0x500000004080,0x10000000018000,0x0,0x140000082000000,0x10044000008218,0x1000044400404022,0x0,0x800000800000208,0x120140404020,0x0,0xa050000000,0x0,0x200000482000000,0x0,0x110000000108,0x20400040004000,0x0,0x0,0x0,0x0,0x0,0x20004004004000,0x0,0x8000902080408102,0x0,0x0,0x0,0x0,0x0,0x80080000400020,0x0,0x0,0x0,0x8000400000c000,0x0,0x0,0x8000002000040020,0x400100800001,0x0,0x400000808004000,0x2101403c0004004,0x0,0x0,0x0,0x840000040001000,0x0,0x1001000000400100,0x820000004008,0x10400004080,0x0,0x2542000001300220,0xc00000800004,0x0,0x240000004008,0x0,0x1800040004000,0x0,0x2004000000800001,0x810000000108,0x0,0x0,0x210000000400100,0x60000002020,0x2004080000002,0x81000060c0000042,0x0,0x2000008000004080,0x0,0x0,0x0,0x0,0x0,0x0,0x4000001000000060,0x800040000000060,0xc00000000102,0x0,0x0,0x500000808000,0x80000008100,0x0,0x0,0x0,0x8000002000200200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8010000000020200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000002000400020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x81000000000102,0x6000000000000102,0x0,0x840026800a1c200,0x0,0x20020000200200,0x0,0x0,0x10400000000060,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x21000000060,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x810400001100102,0x500000000102,0x29101164020c260,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x202000000c000,0x0,0x0,0x0,0x284000080803000,0x0,0x0,0x0,0x840000000800002,0x20400018000,0x100040020040000,0x0,0x0,0x0,0x1080000080000000,0x0,0x0,0x0,0x4100000001448000,0x0,0x802001040000,0x80000108400000,0x206080002002082,0x14000040200,0x4002000002020,0x0,0x0,0x0,0x0,0x820000040004,0x0,0x4000002000500000,0x0,0x0,0x40410020042004,0x4000008000020001,0xc00000002002,0x8000180000001,0x0,0x0,0x210000002000002,0xc00001004000,0x2100000001000200,0x0,0x0,0x4800020008,0x20401008000,0x0,0x100001008400000,0x0,0x0,0x0,0x0,0x8002000002000002,0x1000000400020020,0x0,0x200010001000200,0x8000d00000042220,0x204000000008200,0x10001080200000,0x200100020040,0x0,0x0,0x0,0x8000002020040000,0x0,0x0,0x2020001000200,0x0,0x420000500000,0x10400000002002,0x10002000000c000,0x0,0xc0000020020,0x0,0x0,0x1000004002002000,0x20004020040000,0x802000000002080,0x0,0x10002000020040,0x50001000200,0x4000200000002080,0x0,0x2080800000000102,0x48080001000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa00101000,0x0,0x44000500000,0x8000010001000200,0x1000020020042002,0x8010000008100,0x0,0x0,0x0,0x0,0x600101000,0x0,0x0,0x0,0x104000000008100,0x0,0x140000800002,0x0,0x0,0x0,0x480000c000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20104a080100a122,0x0,0x4000001000002080,0x0,0x8004000000020008,0x0,0x0,0x200002000000c000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x102000000500000,0x0,0x440008b00040102,0x0,0x0,0x8290000601109102,0x840000000008200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4020006000000,0x0,0x0,0x0,0x200010400460200,0x0,0x6000000a00000,0x1002004800000,0x0,0x0,0x0,0x4800115004c12a00,0x0,0x0,0x0,0x0,0x80000100022000,0x8002004000400a00,0x0,0x0,0xc00010400000,0x0,0x80800040800000,0x0,0x0,0x0,0x0,0x480010400000,0x0,0x844000805002800,0x4002000000a00,0xc00808081200,0x40800000a00,0x1004000000060000,0x8200000000201,0x0,0x80000400020010,0x0,0x0,0x0,0x4800000000000220,0x5000012001822000,0x0,0x0,0x8410200080420800,0x0,0x0,0x60001800000,0x0,0x10005c4006000201,0x0,0x0,0x2000800000c00,0x0,0x0,0x0,0x0,0x0,0x0,0x420000000005000,0x6001000800,0x0,0x0,0xa10000100a00201,0x0,0x0,0x0,0x0,0x0,0x0,0x400100000000201,0x0,0x0,0x0,0x0,0x0,0x0,0x8000020000420000,0x0,0x0,0x0,0x0,0x1000080004002000,0x8000800000404000,0x2000800000404000,0xa030440040101c80,0x5000204040822000,0x0,0x0,0x4140000400208,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x84000800040,0x0,0xc002304040401e28,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400004006000000,0x1800004002000,0x2100000000400200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x82010400000,0x2000040000010200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10400000600,0x0,0x0,0x0,0x0,0x0,0x8000100a00000,0x0,0x0,0x0,0x0,0x100800000400200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200040000a00000,0x0,0x0,0x280000600a60000,0x840200240a60004,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd08010420200,0x0,0x8010000000001040,0x4800000000080001,0x0,0x400000821000000,0x0,0x44008000080,0x82000000000a,0x0,0x0,0x2000008020008000,0x900208000000008a,0x100100000082,0xc3060220200000de,0x8010000000081000,0x1004000000000a,0x4430042000440050,0x484100000040041,0x4080000008020000,0x0,0x0,0x0,0x0,0x201000400010,0x0,0x40000800001200,0x0,0x8000002020000008,0x0,0x0,0x808668202501201,0x0,0x80100000800001,0x300000021000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8002000000003000,0x80140001000,0x0,0x2080000600000,0x0,0x20100021000,0x0,0x0,0x100100040001000,0x0,0x44000020020000,0x0,0x0,0x0,0x0,0x2280008000101,0x200000800001200,0x0,0x202402000000004e,0x410000002000001,0x408000000000202,0x0,0x0,0x4000208000800,0x80002000001,0x0,0x942030040441001,0x108000000021000,0x8200008000100,0x0,0x0,0x41000008000080,0x0,0x0,0x0,0x0,0x4000100000000a,0x0,0x20400020000008,0x20808000100,0xf218257069083043,0x8000100000000050,0x0,0x44000000050,0x0,0x0,0x0,0x0,0x0,0x0,0x2080008000001,0x4000240001000,0x80000120000080,0x0,0x4000008040001000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc000000000006,0x0,0x0,0x0,0x0,0x0,0x280000000000082,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc000010220020080,0x0,0x0,0x0,0x100002000420000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x24040000001,0x0,0x208140000000005a,0x0,0x8000100003000,0x20004000001040,0x4002000001200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200040000600000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x101000101000,0x628440000421000,0x4000468604200c0,0x885400084a0101,0x4000001000080001,0x482c80020420080,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1014080022840000,0x0,0x0,0x0,0x0,0x8003002000801c00,0x0,0x882100a02101400,0x0,0x0,0x0,0x0,0x70002a0422101054,0x0,0x0,0x0,0x0,0x40200000044000,0x100001024000000,0x200000200001020,0x800010060000000,0x0,0x0,0x110000000001400,0x0,0x0,0x200040000044000,0x0,0x800010000004040,0x0,0x200401802108002,0x24000200040,0x0,0x2010000000100004,0x100040000044000,0x2002000060000000,0x30000101000,0x0,0x200010022000000,0x0,0x0,0x0,0x404000000240000,0x20080024000000,0x0,0xc0801000004840,0x0,0x0,0x0,0x0,0x480020800000,0x0,0x16000300a00,0x0,0x0,0x0,0x140024000100404,0x0,0x200100200800,0x14000004800,0x0,0x400200000000040,0x1400000800800,0x0,0x0,0x0,0x0,0x1400008040000,0x0,0x4002000200800,0x0,0x0,0x0,0x1000800000060000,0x0,0x100001000080800,0x0,0x0,0x0,0x880050000008052,0x5600000020800000,0x0,0x0,0x0,0x0,0x0,0x0,0x300000008002,0x1400000001020,0x0,0x0,0x0,0x0,0x8080000200040,0x0,0x4400000000004800,0x0,0x0,0x480000044000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1180040020001404,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x840000000200800,0x0,0x0,0x700001101800,0x11000200040,0x0,0x0,0x0,0x0,0x0,0x10100000080800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x108004e006160000,0x4042060004800,0x0,0x4050061024960000,0x400400060000000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x810000000101,0x1000080008000040,0x0,0x0,0x0,0x2400000c000,0x0,0x40200000000104,0x804000008200000,0x200010000000104,0x0,0x2000040008000400,0xc00000000120,0x0,0x0,0x1004048042200010,0x8010000008080,0x0,0x8000000100808000,0x1004020000400,0x6000000201,0x0,0x804002808014100,0x44000000000104,0x0,0x200040000008001,0x0,0x0,0x0,0x0,0x80400002000010,0x0,0x80000008080,0x0,0x0,0x21002000010,0x0,0x4804092009000702,0x0,0x8000000201004000,0x0,0x0,0x4000000200808000,0x0,0x200000000120,0x0,0x400100040000100,0x100080080000100,0x10400000012,0x800010000000a,0x2000800003000000,0x80100000000440,0x4080000002000010,0x1000002008002000,0x0,0x0,0x24004004000,0x1000040002000100,0x0,0x0,0x1000020040000100,0xa02000100,0x200201000002,0x0,0x0,0x24000000000101,0x2000010002000010,0x500012000000,0x0,0x0,0x20000400000c000,0x0,0x0,0x0,0x0,0x1000020000000012,0x0,0x2008012000000,0x5080001002200110,0x42004004000,0x20203000000,0x0,0x0,0x0,0x1080000000000012,0x100040020000400,0x0,0x6001004000,0x0,0x0,0x8000080000808000,0x0,0x0,0x0,0x20040000008001,0x0,0x0,0x1200808000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x42082000000,0x110c42000808180,0x100080000000101,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10020001000002,0x0,0x0,0x4000820801014000,0x0,0x10020000200010,0x0,0x18000200010,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x800010000000101,0x0,0x0,0x0,0x0,0x0,0x20040080000100,0x110402000014101,0x0,0x42000008040,0x100800020000400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100080080000020,0x210020000980000,0x20040000880000,0x0,0x42020000002,0x200200002040,0x0,0x400010000820000,0x0,0x80000200020040,0x0,0x0,0x0,0x404201000004840,0x0,0x200000200080002,0x4066080000804228,0x1410088200988020,0x0,0x0,0x8000020010000400,0x0,0x0,0x0,0x50000004200,0x4000100000200400,0x0,0x0,0x0,0x0,0x4002000020100,0x840000080020000,0x20000080000000a,0x0,0x2000002000002040,0x0,0x8010000000084000,0x0,0x0,0x0,0x4804002010300400,0x8010000004001000,0x0,0x480080000400,0x0,0x200010000000440,0x0,0x0,0x0,0x0,0x200000040020,0x0,0x0,0x0,0x2020020000040,0x100800000004800,0x0,0x8000800000200400,0x0,0x48000800400,0x24000000280000,0x4046020204800,0xd000800000880400,0x0,0x8010080000400,0x0,0x4810000400,0x0,0x0,0x410000000080002,0x0,0x200200000000140,0x0,0x0,0x828008004011000,0x0,0x0,0x800040080000020,0x500010000400,0x12000006002800a0,0x40804010000,0x0,0x28002000090000a,0x0,0x0,0x22000020200000,0x0,0x0,0x2080000804000,0x0,0x420000011000000,0x184804210904400,0x800044000b80000,0x5000000000820000,0x840000000800400,0x0,0x40040001200000,0x0,0x20400040020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x904000200006060,0x140014003200400,0x0,0x0,0x800040004000100,0x0,0x2000800040000020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x832c008880a0,0x4856d2181499446a,0x6000000000002040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100080000820000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x448400000806040,0x40040000084000,0x30300000024300,0xc800674040084362,0x4004001204012040,0x654040000008400a,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc008140100503040,0x24000040000020,0x0,0x0,0x0,0x0,0x1004000001001,0x0,0x200100000040,0x0,0x0,0x60000000000180,0x2010040000020,0x0,0x0,0x144800044000024,0x0,0x0,0x4000800005000,0x200114000000,0x0,0x14000001080,0x0,0x40058000304100,0x8002080000000280,0x0,0x0,0x10400000020020,0x20400040000020,0x0,0x0,0x11000020020,0x42000040000020,0x8000200000084,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x802022002004040,0x140000000040,0x28000040000020,0x0,0x12004080000,0x20400010020,0x2002000001400,0x0,0x400200040000020,0x80010000004000,0x0,0x8000400000101000,0x8000040040000020,0x0,0x0,0x100400000401000,0x2004000000001400,0x0,0x0,0x20080080020000,0x0,0x0,0x80000100020020,0x0,0x0,0x0,0x0,0x50000000020004,0x0,0x6000802004000980,0x80010000022000,0x42000120000,0x0,0x200020000001080,0x1080000000000084,0x0,0x4000400020000020,0x0,0x0,0x48002000040,0x1004200010020,0x240000001008,0x0,0x100000084,0x0,0x0,0x0,0x2004000000040,0x0,0x0,0x0,0x0,0x0,0x60004000004,0x100001000300000,0x0,0x0,0x0,0x0,0x0,0x0,0x805200400c800,0x0,0x50000000000028,0x0,0x0,0x0,0x0,0x100a00200005000,0x0,0x8002000001008,0x0,0x2004000010020,0x0,0x0,0x0,0x100000800401000,0x0,0x0,0x0,0x0,0x0,0x0,0x210000004000800,0x0,0x0,0x0,0x8000000000110000,0x0,0xce22c20b241b42a4,0x1014880000502d,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4042024080020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000100000404000,0x0,0x0,0x0,0xd1a00c40044814e4,0x8100030000104440,0x2080000104000,0x65540c0060120060,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8002000040008,0x200400050000000,0x0,0x9084000005480,0x0,0x0,0x0,0x0,0x0,0x100010020040000,0x0,0x4000001004000040,0x0,0x0,0x1000080040002000,0x0,0x202080100000,0x0,0x181000000084c0,0x0,0x0,0x8042644180100250,0x0,0x480000000100800,0x0,0x0,0x0,0x6001000400,0x1800000040002,0x0,0x0,0x8040000040008,0x0,0x1800000000600,0x0,0x420000004040,0x100040041040040,0x2010000040000004,0x6000000000801000,0x0,0x0,0x820008100000,0x0,0x0,0x0,0x14050000000,0x0,0x0,0x0,0x1800001000400,0x0,0x0,0x4020000000040008,0x100001020040000,0x0,0x0,0x4008000000240,0x0,0x210000004040000,0x0,0x8001040040000,0x0,0x0,0x1020084054040000,0x0,0x840000050000000,0x0,0x0,0x0,0x1800000100200,0x0,0x10400040000001,0x0,0x800050080100000,0x0,0x0,0x2200000000040008,0x88001000002,0x0,0x42000000600,0x4000040010008000,0x800010050000000,0x40040000840000,0x500000400400,0x0,0x0,0x80804051000400,0x0,0x0,0x4000001000000408,0x0,0x4080000008040,0x1280800001802040,0x0,0x0,0x0,0x0,0x0,0x0,0x80010040000,0x42010040000,0x8100000000408,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2050208101008,0x0,0xc302840204040640,0x8000000200001008,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc00840000,0x0,0x0,0x0,0x0,0x0,0x2220210008140,0x80101482050004d0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x802010000400,0x0,0x0,0x0,0x0,0x0,0x480000000600,0x540010010040400,0x0,0x0,0x10450000000,0x0,0x0,0x0,0x0,0x0,0x4000410800000,0x0,0x0,0x0,0x0,0x0,0x0,0x1000080004040000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1302000041082,0x0,0x44010008000,0x400220040000,0x8181004050803080,0x1068000100040c0,0x8040a19100082,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x5ac0326900006540,0xffb777f318a06570,0xd8000f2a0080a530,0x42021800000540,0xa40008a00004160,0x442610e008804560,0xa440036800002720,0x193400f808000560,0x1e83c24000a02560,0x940401000a00520,0x843201e000002560,0x1000a0080000340,0xd080021000004140,0xc141957cd8202760,0xc14000000540,0x1008098108400500,0xabc5842080006720,0xc006212000400100,0x8840110800002520,0x8002101080000140,0x28a4446a00006500,0x1180126708006520,0x4020258000000160,0x8010962000006520,0x41080800002520,0xab10ca6008a04520,0x607214a608002520,0x8200000000120,0x2908441010802540,0x780f00000004520,0x841585800008530,0x14b0540000204500,0x950eb53408006720,0x8000518008000400,0x4885d22400802520,0xe5383c4008a06501,0x8100000400,0x514b852800a04520,0xa1223e4408202400,0xc00000200400,0x80064a1808002520,0x5986d4a88802540,0x2211000080320,0x4984196008000520,0x8390f24200c04760,0x8004b4100002500,0x588001000000120,0xd00aa20c12200620,0xb02444980006d40,0xabecd2080900e520,0x4402000800200520,0x20e048100000520,0x77020acc0a400560,0x9700ed5300a06520,0x8214100008002500,0x28cc181100206500,0x961cb5090802500,0x4474267008800760,0x1400e1000a04500,0x70000510c0000500,0xc08c00400000520,0x940814008080720,0x132a04a688800560,0x80c0040008000400,0x2032145008802500,0x7903c24008800500,0xc2881008200400,0x2003492000200540,0x938180400020a500,0x8530a64008006540,0x4204830201002560,0x2150176000006520,0x1001848300004520,0x8950964000082720,0x16007008002400,0x8080006480002700,0x28401e8000202520,0x202040200000520,0x313a064a08206500,0x10083000000520,0x0,0x100d00000008500,0x804100000330,0x1421d42000804500,0x9092359392202560,0x9aba0d5008a06560,0x44d3b70a08202530,0x1484000000400,0x142404e408002504,0x8124062080802520,0x0,0x202680000208510,0x150824a00000520,0x2580013200006500,0x8000000008400000,0x0,0xc8a228020020a740,0x0,0x8820050888200d00,0x450012080004300,0x18000004100,0x46000000400,0x802000000000120,0x2024010000804100,0x0,0x0,0x0,0x0,0x4400000002400,0x0,0x799e557198a06560,0xa3383f6000a06564,0x180800000200520,0x14040400400500,0x0,0x0,0xc00010800000500,0x0,0x0,0x12050008000400,0x4004000000000400,0x4000200000000120,0x0,0x2400101000000400,0x104410000002500,0x400026000802500,0x0,0x3000000000000400,0x100100000400,0x0,0x0,0x80008000400,0x50284c0200202400,0x0,0x6a96ce9c90a02560,0x80040000000400,0x0,0x0,0x0,0x200004000800100,0x1014526800004520,0x20100f0000002520,0x280200100002400,0x0,0x20100000120,0x11000000000120,0x6000220000002520,0x8240400000002500,0xe316cf2e08a06560,0x0,0x48000004500,0x1080104008000400,0x50008000400,0x8300d00010000130,0x0,0x2000038008000400,0x0,0x40000400200,0x0,0x2000800800100,0x40004000200400,0x444040000002402,0x0,0x0,0x104800400000520,0x0,0x3008550208000500,0x200240100000520,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x800220788412440,0x0,0x10c0064090808010,0x4800000000008200,0x8441308000a014,0x0,0x800480000004720,0x2800000400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xff7df77b8f3eff67,0xfffefff7fffeff73,0xdebc6ffe9ffc6f67,0xc1643402c0086462,0xa31bd5508d7c2d23,0xf4b6f6fc89acff73,0x9d6ffe638aace573,0xfb9e7ffaef6eff71,0xcfd6fcafcf6eec72,0x3bf7369bdb3c6d23,0xc0def3b1673eaf60,0x51aebe78b864f673,0xaaa24d64cf366f61,0x8bc4f3e54b686760,0xe1e59ae80aac5513,0xbefb5fbc8f7eff62,0xd4e1f74a2ffe7f73,0xf6767ffcef6e7573,0xcbf1f3efefec7f73,0x6af39fdadd6c7d67,0x9e9fffffef46f723,0xf7feb6ee75627e63,0x697c4dc13bbc7766,0xeb00e76e1a3c3566,0xe4fab35901ec7762,0x2ffedf7afffcf772,0xf59e5efb4baa7767,0xdfb8ff736dfc7f67,0xff947e76dfee7722,0xedf7fe1e89ec7f63,0xab5ecfcb2f7a6763,0x9cf7fee9d9eefe60,0xf59bff7afb6c7f72,0xcbdaffbecffcfc60,0xcdd5ff1e8ffc6d35,0xeefefddbdf6eff63,0xbcb67ffbff6ef763,0xd1f5ae680b6e7f63,0xdfc62d7dfb6e7f77,0xc7eaf2ab3b6e4767,0xb5d5fdf93aefd53,0xb7dfddffdbbc7f77,0xfd9cd860492c4f62,0xbf3c5d2c9238f521,0x5bc5f32fdf6e7d62,0xedfb7f5bdbec7f60,0x91bacd4325eef771,0xd5f7feffcbaef565,0x9ea61de14bc87762,0xf3c6eb368bac6e63,0x9ecb3f51cf2e7661,0x1f9beffb7fec7e70,0xd7b226680b0a4765,0xb771ff70db3e7766,0xfb9cff3b3f487d73,0x63ce763f933e7722,0x665ecedf87fc6563,0xf4befe7dfb7e7732,0x95ffbf7e9feefd65,0xcfae777fabf87763,0xc9d6fe74cf6e6e70,0x7a30061aca404c20,0x8491c4cb287c6e20,0x51b4fee8cf3e2740,0xf7922f539f647562,0xeffaff3febec7766,0xeffe9efe8f4e7573,0xbf3edffbbbfcf777,0xb67d37788a746503,0x8af6755e01ac3e20,0xcbd4f39babceef73,0xe7fffefbdbee7766,0x97ac4de217acef62,0xfff67ebefbae7763,0x4aac7da31bae6e62,0xf974e33ecf4a7f27,0xe8cffe937ffc7d63,0x2b3e4f7ab36e7d61,0x77f776ff7f7c7d76,0x81c8f5674a2cfe63,0xd2994a62812e6430,0x5fdffebd4b7eff73,0x2ffedfd9df7afe32,0x76b4f6e0cf7cec66,0xc9ba7522fb2a2563,0xff9ebe49db7e6572,0xcad7be3359ece663,0xa5c4f1c132e7665,0x75b377e7df6cff63,0x814885201a304443,0xcbbabfe2ba6c6767,0xf2cfda625bfc7555,0x2bfdcfd18b6ea560,0xf77a67bf8a6ced73,0x31c8702080042565,0x70a6387a036a7742,0x4a84e3aa0bea6722,0xe49b3cd74b28e721,0xf50e75282643d60,0x8141106a35082560,0x40a8055012004742,0xf993ff9fcb6eef63,0x830014a000400,0xc0ba4568c122fe71,0x700206acc8006e20,0x2034000008700000,0x12002000100,0x0,0xd6b545e11085e21,0x0,0xefc67ff1db7c6766,0xfab02f758a7c6761,0x100400010000400,0x0,0x205000000,0x0,0x1800220580006822,0xd34409260a7c0d62,0x8b5e5ac02b7c7760,0x8990c02915003462,0x53a8054013083460,0x8a207d62083c7022,0x8138d19a82686660,0xce77e558b4e6766,0xb1bedfeb6ffc7f64,0xf535f7e24bbc7f62,0xe84118010302402,0x8152802c48244042,0x214e120f106ced22,0x308c400018a0000,0x100800000401,0x950000884062,0xc900d5850a4c4c20,0xc42120818884500,0x4f1ddf5a93eef763,0x80188020404020,0x21419b0048487100,0x4000202018400400,0x41a30020b0a4520,0xe936df43daacf572,0xb4bffff77ffe7f63,0xd114f5cfca486760,0x8ba22d4250ec6602,0x9c4020402080501,0x620cd0814f402920,0x9c3a696228c84432,0xd484f24000383160,0xc866354019247420,0xc350ffbb0b3e7f62,0x24931a78b2646624,0x414000002000,0x404004e058402400,0xee8e65600fec7620,0x990e122c9ec7001,0x826adb5e0b4c7f66,0x693eddfb4ffee543,0x71c05a0869487431,0x40200002000400,0x80c77c0b664523,0x8542f21a4b6e4c00,0x108000048000000,0x2024481082000,0x16876010942f402,0x2100c5409420e00,0x2004010208084002,0x2081dfc00b425f21,0x882601ca08384506,0x1148054811280e20,0x8380040080602440,0x800100020000020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100080000080100,0x8212010c488c0020,0x0,0x81c000880402,0x5426004080882400,0x220244080646000,0x0,0x0,0x8264b460a405d70,0x81b0014882002541,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x14001800000,0x4bf53684e983d40,0xf5f7ffffafbefd68,0xffbebf71c7bb3f60,0xffdeffef5efeff60,0xbfd6febecf9a3f68,0x675dcf47d7b23d68,0xf6feffff7fbf7d60,0xddf7ff7b5eb27f60,0xbb2235785e9eac60,0xfd57ef38df9af760,0xab7d5d5b9f9ebf40,0xbcfb5de9dabb3f68,0x51eeff735fbaff60,0xdaa42f431a966d60,0xe54ef12a1f9e2460,0x6f7fbe5b4faa2f60,0x9ef75ffe17b67f40,0xfc46fceb871f6d68,0x7fbaaf7956fe6740,0xc177fb0a5eb20d60,0xaadc195a67962f40,0x2accffb0e3f7d48,0x9662f6f413bf4f20,0x5bee7dfadfbe6f60,0x8b18e2678e9e7c40,0x8eebbadd2bba7e60,0xe7becfdadfbf7d60,0xd1b7feff87fb5f68,0xcb2bffe01ef65f60,0xffbe7f7d7fff7f60,0xefeeff3bb79b7f60,0xaddacddf4f9e3d68,0x74bdd6eaff966d68,0xb18e5f7258821d40,0xcf1c677edab67c68,0xcdf7ff9e7fbe6d40,0x237f5a504af76f48,0xb57b7fe90fb66760,0xbbe1ee7452baad60,0xfe9fafff0ff23d68,0x61d2f38b07da3d40,0x6b7eddc34af23568,0xa4bacffbd7be3d60,0xddffdffae6beff60,0xdf2e3f74e73aad60,0xab466d6bd6ba1560,0x5edf3eff9e96ed60,0x9fbfdfdb4edadd60,0xdebfe6ec3c8b2f40,0xfffe7ffbdebe7d68,0xfb57df6ed7be7d60,0xeefb9b7fffbe7d60,0x9fafcf5bcffe7d68,0x74dff6eefbfefd60,0xbfbe7d637fbefd40,0xeb26ffffdff63f60,0x7fdeba9bce877f68,0x7bdf9ddb5f9eff60,0xd69cff6b1fbf7d60,0xffee7f625fb77d60,0xcb8621728e329540,0xddcfff8ed6b63f60,0x826ddd535f982540,0x84b3422a92927f20,0x71cc373d82926748,0x7ab6a7eebfba6f60,0x8dd7fa7f5e9eff60,0x6ffedfdfdffe7d68,0x1ebb0fcf8fb2ff40,0x552687f813862e40,0x88be6c7196b23560,0xebcff77f7fbf6d08,0xe7f6967f57f63f60,0xffffffdf5ffeff40,0xd5f737e79fbe7d48,0xc69efdd117965540,0xeb93db8b17d6ad60,0xfffb36becfbe6d60,0xdb3edff285fa7d40,0x34df7febd7aefc68,0xffeed737ff62560,0xc69beb30cada2f40,0xffdffebfdeff7d68,0xab1ecb5606b63f20,0xb4f97e6816fe7d48,0xdbacff72b6beff68,0xec38367a76be7560,0x8ee6a22cd29a7d40,0x63d7528f07da6d20,0x2d3607eb07964d60,0x19581e2a329e2420,0xae9f3d7696b7ed40,0xfdbfff3b17ae3d48,0x6ff6bf4aff863d68,0xf6114fba56125560,0x1200060a08030c00,0x57047579439a7d00,0xcb81eb28869e2d60,0x6000000001100000,0xbbaf5d3996be7d08,0x8432086002ae2400,0x32b625f02ef67d60,0x5200e62414920720,0x840000000900000,0x1602c30a80964c40,0x434244a92822900,0xa90d29200f136c20,0x8004021004800140,0xc42000040120400,0x4018010040200400,0x100305c40b822c60,0x8001246002020560,0x4602000800140,0x0,0x0,0xa000020020,0x0,0xbba0044493820940,0xe4a4600483906600,0x401014804160100,0x207e054901400,0x8850000100440,0x9b9e5c721fba0f60,0xca7ff2468a280d20,0x9c41fd1c2d61d00,0x5fbecd9edbba7760,0xc44cbe216ee63d60,0x1028ad000a362d40,0x4a10131824103c40,0x8114104321040,0x10002000020040,0x4000001000104000,0x20040000200040,0xf802860050a62c08,0x4902940014924040,0xef5fcf5fffbf7f68,0x80040ca002b83400,0x2100000100100,0x2828a000a001420,0xcfcfae3fcebe7d68,0xefffdf557ffbfc60,0xf7bfffefffff3f48,0x15b2f7241b27540,0x54260d600fb26540,0x5788d10920c71d40,0x2966060118923420,0x8fbacff4bfb26f60,0x100cf43b1683bd40,0x400404200e100000,0x89d6fb3d56be3d60,0x76d95a864716cd40,0x1080418110860800,0x610c246050802768,0x620204010900100,0x8294046640944408,0xedf37e9f7f96bf40,0xb364fb246b62f60,0xfdff77ffd7be7d68,0xffffff77ffbf7f20,0x18c523401b04d00,0x40c2122244280440,0x800240800802200,0x11049002902d20,0x8002040000400,0x490024001800240,0x10200800040,0xa000080041100400,0x106018202848340,0x4064030000863000,0xb200080202852c40,0x104c20414000540,0x3088860110901d00,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1400044010800900,0x8800020000020900,0x0,0x880090002082000,0x9111024001960020,0x0,0x0,0x8922100801620,0x2008188002200440,0x1000000200004040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe7bfdffbd7967bdf,0xb6ff77ecf7b63fff,0xdfb2fd73f32c23ff,0xfb8cef2cf6162f7c,0xcbffdf1bf7be7bff,0xb7f0cdd0d5142ce3,0xf4f7f6fef6ae6fdf,0xfcff3773d79c63fc,0x8a8b7d2a92a42954,0xffdfffbef63e6fc7,0x6ece5bd1f7ac625b,0x1e23effef694236e,0xf15ecfe2d224775e,0xfebe3feef7b660f5,0xedd6f7bbd7263fd6,0x2fdfbfc5d61e6d81,0xb79a5ff3d68c6de4,0xd53dfee9d62e2bf6,0x9fee7d43d63e2dfe,0xc9d6f53af6166bdd,0xcea0edb52042f84,0x3ff91d81d6a452de,0xfcf7be77d62e2ffb,0xf3bead70d6262fe7,0x8b08df6af006230c,0xcf47bf75d70e2766,0x9bbedd71d68c697d,0xd536fefbd79e6bec,0xbfbaafe1d6a673c7,0xed36ffffd6ae6def,0xdff7f787d6b43fec,0xaf7fdddfd6062bfe,0xf63a76e2d71c6bed,0x97fe3778d6166a6f,0xff326e5ff79e69f4,0xf7f7bf3bd6be6bff,0x4c04090004200,0xbdbd5bf3d79c7fe7,0xdfeea7b8f69e37ce,0xdabf3dffd73c27dd,0xedfcf1b8d6863bd4,0x236e3b59d79421d7,0xb6bbfffff6ae6fee,0xc93abd7ff6967f7c,0xa4e70e8b6002944,0x8322d77ad7342848,0x9edf3fd6d6be3b6d,0x23bccd73d62623ce,0xeb7ff6fbd30e65fc,0x7ffe7dfbd63e2fcf,0xdbf5ff3ed78e6fff,0xefdbfffbd6ae2f5f,0xfffedffbd6a47ff7,0xf7f4f6fed63e7fd5,0x9bbeffe3d69c6375,0x93347b7cd7ac63e8,0x59f5e21d70e2105,0xabdfdfdbd6bc6bcd,0xb6f9d6e5f4042ff5,0xfd966773d7342bef,0xbea0577af71650e6,0xc1e3ffb6d62621dd,0xa1481a82920c2945,0xaeba75efd61c67d4,0xff69ee30d6943be6,0xdc9a2ff5d28663fe,0xedf5f235c6867bc5,0xcbde5f79d7162bec,0x17346ff3b2a02e4d,0x70aca64af62421c5,0x1e9c6f6cd62467fd,0xa9f7af77d22c2ffd,0xeed63bdbd22e2ff4,0xbfbcefdff6a66f7d,0xd776f6ffd68e2be7,0xfefe7df3d62e27df,0xf3f6ff3ef72e37ef,0xce5f97e2d6947ff4,0xf6bfcff7d6162eed,0x30b7feebf79e63fc,0xdf0aff72d69c73fa,0x5e83e16ed3022c86,0xc6dffe27f71e6ded,0xa7d6cfdb96243d7d,0xfcf37e6cd62623ff,0x9126afe0d63630bd,0xcb9012d9d2a027f4,0x8727ef3ed60c2fad,0xdd7d9c4d71225c4,0xb49b5de5d6a02e55,0xbfedfe70d22461ec,0xdbbf6fffd7366757,0xedb6f73ed20c3ffb,0x6f5f1cd9939e295d,0x7287f72928439ec,0x102307769210273c,0x7694355ed2067e34,0x8bc7a20d95946bc4,0x69a73d80d712414c,0xb4d2c38ad21421d7,0x34762666d20420db,0x4e845dc2d51441c2,0xa0c4c10250022f00,0x2870838ac0046140,0x300050054000004,0x3107066e560c2865,0xe5388320d60c2d00,0x99839b76d69065d4,0xc5cfee39b206235d,0xb1f475ac2022980,0x8020042040002880,0x8805000000200048,0x4020220010000804,0x0,0x0,0x0,0x0,0x800230090002014,0x4180921050002384,0x2952965190242175,0xa4300560c2002481,0xd2292e70d63021cf,0x2e12f562d22c68eb,0xebf1f32ed7b673f6,0xe8f35f5dd69473e6,0x7ebfcfdbd72c63fd,0x40140280102000,0xc6a82f40560609e4,0x8a102112c2042104,0x90308354002041,0x8080000041000,0x904400c0000102,0x2000882000002801,0x800002002000000,0x8882722982022850,0xcfbe4f53d61e7be5,0x26a897e0d60c2dd4,0x58110000a00,0xf496062cd2102f01,0xcce7af3ed39c2dc6,0xeaffffdff62c7f7f,0xfffffffff6be7fff,0xfbffff7bf6b62f5f,0x9226356092347505,0x88245ea0d4902b41,0xab745ecbd30c25d6,0xb3becff3d7067397,0x91008570d7022a04,0x24042012020200,0x80003000020c0004,0x2fff7ed3d7be6fd7,0xa1ccbd4b2062264,0x100880000010,0x24040000802440,0x804012890002304,0xecf73eabd6346bf6,0x85ff5f74d28c7ff5,0xfffff7fff7be6fff,0x97fffff7f7be7fff,0x2a10670bc20400ba,0x29568608d20403c4,0x209480056002020,0x410006f46222002,0x80410500b2140084,0x4010271282142404,0x1068704c4000058,0x802121882302004,0x430142590100058,0x8d0242820c6101,0x0,0x100401830000020,0x2400000000040080,0x22200cd082802c40,0x880080000008,0x0,0x0,0x0,0x0,0x801081000000,0x452c055074042c80,0xc202a56ac6a0654c,0x1000020000040080,0x0,0x0,0x94100100c10400a0,0x100008000000000c,0x42021012000104,0x0,0xf0142020c6042008,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa9421c08813053a8,0xafb7cffb8bbe7bb3,0x70b6fefb8bbc72e6,0x5afefd728bd47bfe,0xf626af678ba85bea,0xcfc3fe36a9f0767e,0xefffddd6abbd53fe,0x5dbff6e9abbe5bfe,0xbf7cff78abba7bfe,0x9b6ceb2f8bb463be,0xd063ef3aa9f472cf,0x7ffedfdfa9fc57bc,0x3e9615b588b877a6,0xdde3fdf08bb24b6b,0xdebe2f7babde5b25,0xefffff6f8bf55fbf,0x3dfefffb8bf77fbf,0xbab5cff88bd84726,0xddec37968b3c57ae,0xcb5e6f778bbc57ae,0xcb92f3eeaabc5364,0x64935ff98aa43306,0xb7b9def28b785bee,0xf5fefefe8bbc5fec,0x5f46bd3189b45fbf,0xe991fb2d89fc73bf,0xcfd3ff86899c4ffe,0xffbf5fd98bf047ab,0xd77737ef89bc73a6,0xdb7eeff1abae7f3f,0xfbbeff7f8bff5fe3,0xcfcfff9babfd7fff,0xabfecfc98b9c535e,0xfcbf56728bdd5327,0xbffefff3abbb5b76,0xcfbfbf7fa9b47bbe,0xcfeffee8abf47f7a,0x699fdf478bf657fe,0xb2b65fee8bbc7ff5,0x3960ce7c89b8473e,0x8fb22f738bb5722e,0xc7f7ee35abe42b63,0xaa77ffcb8bf03f7e,0xbfbedfff8bbf5bbf,0xf9dffff28bd47bf6,0xffd6bf7789fe1777,0x8933fb8a89b8177e,0x7dd71fd88bda7b78,0x5bbd6dce8bb63bbe,0xf77dbbfd8bfd4bff,0xfffe7ffb8bff7fff,0xfbffffffabff7fff,0x6febfede8bbd7bf2,0x6fbfdffb8bfd7bbf,0xf1f6def9abbc73ee,0xdabef562abdc73ef,0xf700af078a945a2a,0xccd2b019a9856a6a,0xea7f0fdbabbe43f6,0x74efdffd89ec1f6b,0xb4c9af708974122c,0xf21677b4893c536e,0xe7c7f73a89f57727,0x3b765e5fa39e1367,0xfc5eddfaa99c63bf,0x1082661a98c0062,0x8af63e53a39a7ad2,0xefcfdf58abf65b6f,0x2bffdfe189dc63a6,0x8ebadff98b9c5f26,0xb7cd176ba9b25f34,0x2b1e3c4589846332,0xe1b7ff3dabbb473e,0xeffbbfdfabfe6f7f,0xb6b545e88bd06336,0xd573fb7babbe6233,0x7bbe7d7389b57bbf,0xdbf6df7e89f45fff,0xeaf3bfdf89755b6e,0xda3eed7889bf1b66,0x5efddfeaa9bd4b63,0x92922d5088500be2,0xce06e52d89b66223,0xfd473b1b89fe4be3,0xbb0ecddb89be53b3,0xf4b51ce20b9843aa,0xf7e3ff7b8bd957e7,0xffb267fa8bbe53ae,0x49d2b682888c5226,0xa35dce5c0b9c43a1,0xb7f7df6fabdd4b8e,0x5967ae7a8bbc6f66,0xfe8e3f2aa991736e,0x6fcaf23f89bc7fe6,0x57dedf8881b47ea2,0xb6d84ee109bf432e,0x50400080110020c,0x18c0010408440120,0x2004000008004000,0x4cd6de5c81935bf6,0x88aca42aa8b856b0,0xd0cfa6f58adc53a0,0xca0204428ba04200,0xd0c1714e0b844308,0x940020000000328,0x97cc0000880a27,0x4486f2d08929436a,0x8000040000000120,0xc290732a8b505a02,0xfcd7bebfabdf0baf,0x4812094202842300,0x24008080000,0x200010000040200,0x0,0x0,0x100041880005080,0x0,0x8000080001000020,0xfabcbfff8bac63eb,0x4000240200140220,0x2220004281340020,0x4a85950a3844a18,0x453e8d6389104be6,0x1a96256009d02a60,0xfbc6eb478bad53be,0xaff6dadf89d45bf5,0x9f9ec5cb8bcb23a7,0xd674026881b4032a,0xa8003c00a001130a,0x0,0x840060000080320,0x0,0x404010800a804000,0x42030000003a0,0x8000000400800020,0x494bbc0b8bc0632e,0x69df9d53a9b07727,0xf5fffffb8bf75bff,0xd4e8242088c40227,0x9024356489344b00,0x4f8bfbbea99c5fa5,0xfbffdfdfabfe7bff,0xbfffffffabff7f7f,0xbbefaf7ea95d47e6,0xc012617081b00a13,0xc022081001b00210,0x16d4ce0a8194130a,0x7904772893a4ba0,0x2040b37aa100536e,0x28221600a8840b0,0x0,0xaf7f13d3abf463ae,0xb3dcdf38bec6b3d,0x9474c7668bcd332f,0x24006009400020,0x314cec088b304626,0xdde3bf18ab8c7732,0xbffecffaabff43ee,0xfdfffffbabfd7fff,0xd7bfbf77abff7bff,0xdc2e6ffe8bb14bab,0x886098108b200060,0x22e00400b004060,0x4002001100000,0x40000800020,0x8a00000800902320,0xc0000800010,0x40c01810080c0320,0x18002180200002,0x8a64088500300,0x2000240002005200,0x580800080801002,0xc80092030b920202,0x800204800a804000,0x80101a0202840280,0x300080000800,0x0,0x0,0x0,0x1100004803004002,0x82a0c6280100f28,0xc0a0e37209c042a6,0x0,0x80040c8102806020,0x28000100200,0x8081416080a00062,0xc000110208500a00,0x502080000041a0,0x4000000300,0x48000000300,0x120022000300200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2fef1a9bd1d5db6a,0xffbecf73c9ddebba,0x75bffefce955f32a,0xfffe6ff3fbc5f36a,0xfe05ffae7bdcc372,0xce7bbf33fbcde32a,0xbbfe4f5fe2d4d16a,0xf7f7fee7ebd5f1ba,0xd9feef7b4bc4d33a,0xf2b67fe6cb5cf1da,0xffd7ea7ee9d5f37a,0xefffffdbf9ddf3aa,0xf4fe5fbb89dde3ba,0xfbe7ef6aebcce16a,0xc3ba7472cbd5d30a,0xe1f7f7aacbc5e32a,0x46fedfdff354e31a,0xbf9beff1c1dde33a,0x24769f4051c5c302,0x5dbe7e736bc483ba,0xabec5e167354d32a,0xcbc23b5b48c4c13a,0x97925fe7d8dcf13a,0xdffffffeebddf3fa,0xfebe3df2e3dddbba,0xbbd4ef7febc5e9ba,0x6fda860fabc4c12a,0xb73fcff3f9c5f13a,0x757576ebc3d4e93a,0xdffebfe26bddc3fa,0xfffffffffbddfbfa,0xefdeffbff3d4d33a,0x8dd64f5fcbd4f3ba,0xf4f6feaafb5cf19a,0xdffbff73fbdddb3a,0xefb77f37e9dce3ba,0xcdf5ff0ef3c4fbba,0xe19edd5b5bc4c1aa,0x74fa5bfb7bd4d31a,0x5948b5a9c1c5d9aa,0xdabbefe2c3dcfbea,0xc7c7fb73fb54d36a,0xed7edffffbddd33a,0xffbffffffbddebfa,0xd57dff7afb54f3da,0xfffe7f7ffbd5f3ba,0xf30effeeabd5ebfa,0x4ff77eeacbd5d3aa,0xfffffffef9dcf3fa,0xd7bff2fdfbc4f9da,0xfffebdfbfbddf3ba,0xfbffff3ffbddf3ba,0xefdfbc3be3d4e3ba,0xff7fdffff3dcf3fa,0xfef7fffafbddfbba,0xf7fafff27bdcd3ba,0xf283e577f144db3a,0xc856fe3169ccd00a,0xae4c4fdad984e31a,0x7cf77fedf3dce17a,0xa1fea76ac0ccd30a,0xee9e2f7ed354eb02,0xc863ee2ed1d4e3ba,0x97de5d4b8bd5893a,0xfcdb7fa3c1c4d32a,0xa54727985150d10a,0xeab62f7e4bd4937a,0xefa7f76eebc5c32a,0xbf7efadefbdcf32a,0xb7b7effbebd4fb2a,0x94aca760d1c4431a,0xdbfe6d54f1c5cbb2,0xd3f3ef2a79c4d33a,0x5ed6ba6df354a32a,0xbf7e5ffef2c4fb2a,0xf7f7f6ffe9d5eb3a,0x5fd63dd3e3ccdb3a,0xfbfdfffffbddf3ba,0xcfe3f6e7dbd4e32a,0xe7bf4ffaf1c5e3f2,0x7db43ceb6bd5d23a,0xf7beeff0ebddf31a,0xfe14f25e5154d3b8,0xcedef24260c4c332,0x5f9fdfcff9c5f32a,0xf4b5f66ff884eb32,0xcd6e67f1f1ddf32a,0xfa97373f49d4d17a,0xef77ffbdebddf3da,0x1bac9dd88154629a,0x347e15eff9dce12a,0xd974ad9bfb44c1a2,0xdc063ff452d5813a,0x8b563a02c854e82a,0x4e76ecc74bddd10a,0xb9b34d7d5bdcc23a,0x9790ac018000e12a,0x5602314648c48122,0x6084e30e80c4c32a,0x280222098200a922,0x1fb0ffe8e2d4e07a,0xb0c45e4200d4c15a,0xfaa40de2d1d4f11a,0x49bf532243d4d320,0x40011610e3c04008,0x40180c01e084e000,0xf18277e8e944e12a,0x936045209540202,0x5c60020041005108,0x102900040008012,0x8100040010,0x421000080004002,0x8080000001004000,0x0,0x0,0xab7a914a7280d12a,0xe148008002,0x8400000040008000,0xf2b6249170018128,0x9c2e4220384c2a0,0x10100000000102,0x24a0040809020,0xd0dcb6c009d4c36a,0x17960f648055e92a,0x8becc57be9c4eb4a,0xfca2e5feb45a36a,0xb69c7d400188c31a,0x1412e6e4c0c0e10a,0x273006010005c30a,0x5210410861004000,0xc0001182000802,0x401000008000002,0x140400800080c002,0x2c1008440000,0x0,0x8fee3b3bebc5eb3a,0xab7fdfdb69c5d2ba,0xfeff7fff7bc5c1ba,0x4000a00040048010,0x2a90400888006012,0x894bfe0b4145813a,0x8bdedfdfebc4cb5a,0x15bf37e2e344c15a,0x91cc87e8e2c4f03a,0xc0b401f8f104c322,0x85f6e11a89c5ca0a,0x588800900c0c028,0x7feeed7ae3c4f33a,0x500000000108,0x5a5e25e361c4d14a,0xc5c11001ccc042,0xc31f706145c00a,0x861a02100b00802a,0x66a31408045c818,0x28ae3c63d3d5ca8a,0xc9f4e7aef1c499ba,0xe6c1065a4184c30a,0xefbfdffbfbd5cb6a,0x74fffeece1c5f32a,0xffffff77fbcdf37a,0x5212291500c0c13a,0xcc0b9608e101a322,0x2308090000c0400a,0x348400408201c002,0xd1a855004045c01a,0xa81082008300480a,0x10430a004100d020,0x4010000000400002,0xdcd477edc1d4f33a,0xc170bf405ac5c13a,0x3000012080002000,0x208000008020,0x8000080000008020,0x2080143040004102,0x9000040840008100,0x2004001000200,0x4020080004000,0x0,0x0,0x340424080800812a,0x40086d42824cc062,0x1000ebb041808192,0x840000080002000,0x102004000004000a,0x0,0xa00010201040280,0x0,0x0,0x418048100008102,0x48000008100,0x88500040008100,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xccfb17ffcedbf7f6,0xffbfdffbcedbfffe,0xfcfdfff7cecbfafc,0x57be3772c6d2327a,0xff9ee7becedbf6e0,0xccc22ab946cbfaa0,0xef9edfdfced3fea6,0xf7ffffffcedbbffe,0xffffffe3cedbf3f6,0xca2a6f44ced3f324,0xe787ef2accdbf3b8,0x27f6dfdbccdbf6e8,0x75bdfdebc6cbbf2c,0xfb4ffef6cecbbef4,0xfa963df8ce92b3fc,0xc9f6f623cadbf766,0xc812f6c7c682732e,0xfefb7fe9c6daba5c,0xa1771e6bcedaba7c,0x1976f5714493b300,0x89de5b6bccc3faac,0x4ec91a9fc4db9a60,0xdef9fffbcec2b7f2,0xf7ffbfffcedbfff2,0xefbeff73cedbfffc,0xfb97eb3fcedbbbfc,0xa8c7bfd9cecbbacc,0xd1bfcffb4edafff6,0xf4fdf7ebc6cbf972,0xdbbfff72ced2feb2,0xffbdffffced3fffe,0xefeffeffcecbfffe,0x2f9ffedfce8bf6fc,0xf2d2fff9cedab37a,0xdbfffffbcedbfff6,0xffbebf5e4edbf7e6,0xfffef6bec8c2bf74,0xc6dfde53cedabe52,0xb5b36de3cad3bf60,0xfbd9f7aacedbba20,0xdfb62f7386d3be32,0xcd8cb3e4ec2ff64,0xdffffedfcedbfffe,0xf7bfdffbcedbfffe,0xb3f9ee7ace9bb2a6,0xfeb6effec6cabf74,0xe9efdb9fcec2fef4,0xc8dabf698edabac2,0x5fda6ddb4cdbb676,0x9cfe77c38edab3fe,0xfffe7d72cedbfffe,0x8bfdd6bfcec2bf6e,0xfddaf75eccd1f7ec,0xefffcff7cedbbeee,0x54f7d6fbcecbbfe4,0x7f9efd734ecbbfee,0xf6b6bb764e4bf7aa,0x5ade3e9fc6cbf68e,0xa34e8f4a4cd2f230,0xbcfe5e654cc3beee,0xdfc66d624ecaf2bc,0xdebfbf76cec3b77c,0xeac6f71a4ec2b7e8,0xa7ccadec8cab276,0x8838d5cb4c43a220,0xd27c3d180ec2a322,0xe58d3562448ad5a8,0xedc6f2af4ed3bf54,0xafdfbdd9cedbfb3e,0xfffbffffcedbfffe,0xd5fdbf7dcecafa64,0x1042e56488407024,0x83c7cf03c8d2b924,0x6a733f5bcc43bae4,0x9d9acdfacec2babc,0xf477f2f1cadbf3c2,0xebeeff6a4edbfbfe,0xfbdfff3fcedbbff6,0xec63bf79ced39f1c,0x8b3e4f56c48ab22c,0x5a94d7eccec3b33a,0xdf03f5514ed2b28c,0xc98c742b48cad670,0x4dcfbe494ecabfbc,0x2f6adfda4ecab6e6,0xf6fdf6fdced3bbee,0x916eb75ac6c2b896,0x9fba6776ce83f3e2,0xf7e7f33a8edbf3e8,0xa91efddac6d2f676,0xb5f37feacedbfaae,0x9928877286cab2a4,0x1a300f12488b9320,0xa46fb2a448cbbe80,0xd7fde994ecab728,0x96ba7c6d0e5afb6a,0x95b49706c0c273e4,0x2023d4004418b08,0x8a2eba3f86839302,0x802021100020140,0xffff7ffbcedbfefe,0x81a2d62084c970a0,0xab614200600a30a,0x89f6cb36c6cb126c,0xc757be8b4ecafe2c,0x44880800004,0xf5f756fd8ed3ffe4,0xdb6bbc75cecbba62,0x8280200000420300,0x0,0x800200002002000,0x44048004002080,0x940420844028120,0x8200000002000200,0x8400000400200,0x201100804808040,0x0,0x109c06080c28b00,0xa8282760cc827606,0x5ceca2286c0f282,0x2152c611c8c1b404,0x8064000103020,0x810000080800000,0xd79a2ffc8ecbfbb2,0xe9ffff7bcec3ffe2,0xefdabeffcedbb2fe,0x169c05e84cc0af08,0x94356270ce80b6e4,0x500002022040,0xdc4526c6c01aa0,0x10022e0002129002,0x8402048001004,0x30108a0004903080,0x303a107004438068,0x0,0xded7be1bc4d3fb2a,0x2bfb1b5a46cb934c,0x78f636f9c6d3b764,0x8c06242044d13020,0x8800041080002004,0x59c6ea02cccaaf64,0x2f1e9556c6ca3aca,0xdb2647e14cdafa30,0xf3ecefee4ed3f7b4,0xca061d6c0e82a700,0x8f439b224ec2a660,0x15ef282448ab078,0x4028674100c2a200,0x10268780c0989104,0xdb3647f6ce82b3f2,0xe9ddd602cc82b234,0x214011c188818200,0xb9320502c4821260,0x9401a07048c18204,0x8b14b521ca82a214,0x4b166ba846c3bfa2,0x10061e0104503600,0xbfffdffbcedbf6e6,0x71ff7fffced37ffe,0x9bf7ef53cedbf3fe,0xeb18776cc69af6e0,0x890e3e1f46dab7e6,0x414a018bc6880600,0x80500084021000,0x10c103704800a632,0x4006202c04822080,0x8000800040020000,0x8040040000004,0x140856eecc81b260,0x2023484002a164,0x1210010000422100,0x8010000000020100,0x80800000018000,0x20204400000a042,0x8034000000121020,0x6000000020100,0x0,0x0,0x0,0x24002180403000,0x5e80244008c9b290,0x302800c00107020,0x0,0x8400080001000,0x402690c820220,0x0,0x0,0x8000000104002000,0x400400040000004,0x1080048004800000,0x120a82044909204,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xebd0e92aaa72b39b,0xecea3ebb99737b9b,0xafbedf77fdf77a9b,0x5dd77eefdfd73ebb,0x9ffeb5919973da9b,0xff94ff7f5ff37fbb,0x68cfff225cd2beab,0xabe8f5b4b137e8b,0xfcf6f6fdd9737fbb,0xdfeff7f96f53fb39,0xf3bf1f6f4cd3bf8b,0x7f5ef7bfad73be8b,0xaff7fa4f88f27e82,0xbeb7dfeb8fe33a0b,0xdbfdbfeeaf573abb,0xcbaefff32f73fab3,0xe867fe2059523e1b,0x9886fe0bafe23f13,0xffbeeffb6dd33f93,0xd1e5566e2ad31b93,0x97ac2bc08b521a01,0xdad77ea609d31fbb,0xab775e1218737a11,0x9e1bcdd349567b2b,0xfffffffffff7ffbb,0xfffefdf3cff77fbb,0xbbd5eb2b2ef33b83,0xe8c7ffbdddf3babb,0xefbfcffa6ff37eb3,0xfcf676ff8ad3b23b,0xdb92bfd17e737eab,0xdffdeb7f0e57ff9b,0xefdfff3fedf7ffab,0xeadddfdf1ef3ff3b,0xb09ac7e803645a0b,0xdff2ff6fabf3ffbb,0xf29a2ff4cb635ea9,0x8bf2a2bc08537b90,0xead77ddd4b76bb9b,0xbe77d6ef4a73bb9b,0xb3edf77a1af39ab1,0xdeae2ffe0963fe13,0xcfdbfff92dd7ff3b,0xbffe9fff6f773e9b,0xbeffdfffc9f3ffbb,0x927eff7a18735fb1,0xffae7def6ed37b9b,0xebcefbddcb73ff8b,0xff5b5fdf4bc37b33,0x92bccf6bcc473a83,0xf5efb67bfb73dbb1,0xfffa5d7b6fd3fb89,0xeb94f7369bc7fb9b,0xa6ce2ef208d3be93,0x4e96454e0dc73a81,0xb4f5f6c4a9f21a91,0xdbbaff628873b3b3,0x7106a7210844bb11,0x5dce36930c627e2b,0x2bdecddb46d37a03,0xf5fff6feedd3ba39,0xfffefdf3fff7ffbb,0xffbf3fffeff7ff9b,0xffd7ef3e2b733eb3,0x6bdfffdb5bf3bfa9,0xbd3e55fa9a733ea1,0xf97c7f461c77329b,0xfe9f28e7cad37b3b,0xc1deffef4e53bb9b,0x6efede5b6b73df8b,0xffffffffeff7ffbb,0x75ecf67b09e33ebb,0x7f123d67aac37b81,0xe977ffaf3bf3dabb,0x4cf71ed32a727e9b,0xffbfffffeff7ffbb,0xfdf6f7ff5bf3ffab,0xdf9e7d710f767bb1,0xf3e6ff6ec9535f9b,0xcfdbbbdf4a739e9b,0x7bbbcfdb2df7bbb3,0x84e637e8c8931a8b,0xeffafd616a77bf9b,0xe9b4ef618af3dba9,0xef5fbebfaff33f83,0xab7fff4abb531eab,0xd4b7766a0a72bfb1,0xf9a5ef7aefd33fb1,0xbbb437745cf23e01,0xe9fef8a2b1f3ba9b,0x2b76194803439a03,0x9cf557ea8bf37a39,0xf9dd76730e723b13,0xaa0016208431b91,0xc95bfeffcbf3bfbb,0xc0d33a1908d21a1a,0xa9a84d780d037a8b,0x90587a2024033a01,0xc10e3ca003921a03,0x121a6040521001,0x8008480108601001,0xffffdfeb9fc77eb1,0x404922805023801,0xcb362c0001177e01,0x83d0b6115852b303,0x9c4bbeebadf36eb1,0x2730140080021602,0x502ca35008035a81,0x80082e5444d03a03,0x820000000202,0x40c00c1048004801,0x4842091208521900,0x0,0x0,0x40000400000000a,0x0,0x0,0x0,0x0,0x8000004004002000,0x80032601831e08,0xa8c29a5208633f21,0x402006000008281,0x10502e80ac022210,0xdef66d7c48765a83,0xbfaff6f4dc73ea9,0x69e64e1a05137e9a,0x90908242021811,0x9000808400050602,0x4f0264400a521301,0x8982610a41521801,0x110000000201,0x86500010120b,0x0,0x0,0x4000800000001800,0x88d7be139ad23a93,0xb25a95cb49521b81,0x61b516b849563e81,0x9dfca7730b535b81,0x72962f7ce87313ab,0xef77fed669531fbb,0xac4ddd88d733a8b,0x983e5ffa89137a8b,0xfd2eff6aabe75bb3,0xbeb617614b731a8b,0x8136d3b809905a8b,0xaeda3a9b4d433f8b,0x80400000000201,0x1388f0044910181b,0x8b32ef7201535a01,0x81a044120800161b,0x40023810411a0a,0x8cf267e4ac56b289,0x9242124613701801,0xcf523c4a08721f39,0xfb54d33dcb537f31,0xb242a25148521a19,0x8e3f4f734f757e03,0x75ffdff81df3fe9b,0xdbaeaf710a77129b,0x5200e10b1b021901,0x242220002401001,0x940001008000aa0,0x380aa809021a80,0x28240600084102a0,0x140000000600,0x0,0x90d2c0c0021009,0x444000001000a01,0x84000000500000,0x80000240000001,0xc1000040020a,0x0,0x2000010000400001,0x8010240040020011,0x0,0x8100025400031600,0x110000500000,0x34dbefed2e773b91,0xb4a52420085010a1,0x43a83d3009d21a91,0x9292730e0b863a01,0x0,0x0,0x0,0xa00500001101800,0x2c2009000900,0x840000001000800,0x150040000029001,0x28400001008801,0x813258200002900b,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6812247864760954,0xbbfcf7fee67f17da,0xffdfbbdf6a7f3cda,0xeffbcff3a36f1cca,0xddf7deece4763c4e,0x21b6fd636876acc4,0xffbfef7b6f7f1fde,0xefd7b74fe77735de,0xa33a9fcf667d99de,0xf4fffefa677fd9ce,0xd9cfffd66d7eb5de,0xdebf2ffc6777fdd8,0xf7fbf7bae2773dde,0x8f7f5b0ee27e1dd6,0x8dfe77eaaa771dce,0xfbfcfefeaf779dde,0xaaaeef4c6a76b4de,0xcbf6ffa3677f96de,0xeffffedfef7fffde,0xbebfffffe67f9fde,0x9dfcf7caef7fd9da,0xd6baff66ea7e3d5c,0xc823e352227d154c,0xefd79e5a60765fd6,0xbabfcd6b236e15ce,0xfdffff7deb7eddce,0xf6feffe1ef769fde,0xa39fefdf277ef8de,0xdfdffeffef7ffdde,0xbebe5ffee36f5cd2,0xd5f676786a7eb5ce,0x5fbeffe3277f96de,0xffbfff7fe7779dde,0xd9cffe9fe47d1dce,0xa8164ed6215fdc44,0xf4d27eec636f11de,0xdfeebe62a7761ddc,0xff977577e67f5aca,0xfdffff3f677f7fde,0xa2dfffdee6771dce,0xb6fffeffe67f5dde,0xf76aef496a763358,0xf7b6effee67f7dda,0x67eaff26227eb4cc,0xabbfff79a27e14c6,0xf7bb7ff5e05fb4de,0xff3d8ff32b7fb8cc,0xfef26f74ef7e97ce,0xcbfefbbfe27e9fde,0xafff7ebb2f7f7dce,0xbdb25f73697d1dde,0xd7fef65eab7f9dc6,0x4f7e6d7a2f7468ce,0xc9f7df7fe37f15de,0xc9cbff90e24717c2,0x9a9a5570a74413c6,0x1482b6e8267684ca,0x3bbe67f3ae7611d6,0xf9a47f8f2676900c,0xcccab7bb62775ec2,0x3b8cdf572a767dc6,0x36957fe8e37cb5de,0xf9eff773ef5ffbce,0xfebe7fffa37f13ce,0xf554fe04267d79c2,0xb5cdfdae77ecbde,0x6cd25df1a2741dd6,0xf12ebb50227e3c4e,0xda8ebb6fe67631da,0x89eef7ffef7f9fde,0xd3feff592e7e15ca,0xffbb7fffe77e95de,0xd7fe3fede27619cc,0x5f0e2d64623e1656,0xfbf6ffffee7f99de,0xae5237d626771958,0x3eff75fe6e7f11ce,0xf1f7feffa37f9fde,0x75be7f72e67697de,0xdbd7e337e77e1cde,0xceefffff6b7737de,0xafb6cfe1e17ea1ce,0xcdff7efbeb7695da,0x5ba8ef31e9771ed6,0xdbb4dfbe6c7717de,0xddd7fe23277493dc,0xcbfe5f52637635ce,0xf6fffffee7773fdc,0xfffdff7f677f59dc,0xdbbbef776f7f7fda,0xe5d5f73ee774a1ce,0x2a51845a60245850,0xc4f75fe92e769fde,0xb1e9afd823724e4e,0xdeae2f4ee37617de,0xecf7ff7eef7f9fce,0xac363b54a67210c6,0x7005150a2640850,0x98e53e28666259d2,0x68163c63604e145e,0x8000e1302234000e,0x140021000000,0x829005aa60435910,0x8442046100611440,0x90102140202400c0,0x0,0x2842000e22021016,0x80c00160200800,0x4441628a4700042,0x24402860602010d0,0x8400004000060050,0x490812140a441448,0x0,0x0,0x4224000000,0x8050400520002,0x0,0x0,0x0,0x40020008000,0x10021000200840,0x80000400000140,0x96134322723004,0x28004100001450,0x9201034882161002,0x74b0255a227615de,0xa8e430836270154c,0x6d901ec26a75204a,0x8223444a00471146,0x420000000102000,0x99ea950225485c6,0x50c00c00640048,0xa413289207e084c,0x25044000345800,0x900f128000630848,0x4022010020220800,0x0,0x84a3235025e11de,0x2330105e027c1c50,0x60ae005022671248,0x70ca8f30a21c114a,0x16900350236600c2,0x80d7eb92a57c94de,0x8901d4f227619c4,0x1c1f6f77223f11c2,0x73ecb6f5eb7f15d2,0x96962e49227f48ce,0x69dfdf7ae2561456,0x3b5a36c7635e95d8,0x80808640000110d0,0x134c97226a74148e,0x4a9e29f22b5e14ce,0x8202e360201a81cc,0x2d0060124720006,0xf29147e0e6571450,0x86fff2eb227f994c,0x4082fc10816a0a42,0xe996f336a6779dca,0xbc921e07624681ce,0x9bec7ca207658c6,0xf5ffd7ff23775dce,0xb600e173227488ce,0xcc20022101400,0x4144881422644808,0x8480221620880,0x60201630a270100a,0x4100040000301800,0xc220214228620048,0x0,0x800000800102000,0x401012000001448,0x0,0x800050000420240,0x2026000100006,0x8410c00222101050,0x808000900300044,0x0,0x2040020008000,0x8120007021001010,0x4080000004800,0xfffffffbea7ffdde,0xa55c3268a256104a,0x50626c4060060000,0xc380b215e07400ca,0x4847b20423509248,0x14026623000240,0x0,0x18000420000,0x80800014a0200800,0x40070a3060248840,0x8a548f1b62348844,0xf0a8c66822761554,0xdffeff7fef7f7fde,0x928c2b40a774034e,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x91fdaf7a9e6cef1a,0xdffebd739f7cfef6,0xf9e7f7368f3cddf1,0xedc332b18a3ddf98,0x819dd7f34e75e639,0x35431765001dc791,0x73eaef615318dcf0,0xfbbcfbfddf1cdedc,0x8dd3fcbb9a1dfefe,0xafdecfdf8b3cfefd,0xf4f87e7e4b3dce59,0xfffde7f98f3df79d,0xf7920b658b20ce92,0xccfef6be8d19fe96,0x6e5e5ade8b7dffd1,0x5c9f7fe70e7cc6fb,0xdd5ef7fd9f1ccfff,0xdeae2f42881ccfde,0xefd6cdbe0b5ce7df,0xebfedfffdb3dfff6,0xbeba7ffa9e3dc7fd,0x93bbf77e8b1cbe9a,0x9efe7ffd1e3dffd8,0xe8c4f26a9b088ff8,0xeabe9ec69f6cdc58,0xb2b3fef2cb3dcede,0xfd7dff65df3ddfff,0xcb3e7ff24e7dd7df,0xf9f5fa36ce3dfefb,0xdf7fffffcf7ddfff,0x7eb84f8a091cfefd,0x70f1dee04b5df7f2,0xdbbfff628b7dfefc,0xffbdff6b9f3ceffb,0xb8c7fe129b197ff0,0xfbfffd3de3ccff8,0x76d586e59f5cef98,0xd97eb77a0e34f7df,0x9bbe7f679f3cdff5,0xd9f7ff3ecf3ccff7,0xff5fdfdacb7dfff9,0xbffe5ffbdf7cfff9,0xffefffffcd7ccfff,0xdefe3d7f8f3ceef2,0x83d5efb7c91c8cfd,0x39ef3f4d4f1dced4,0x9cdb5fe85718cefb,0xedffbf768b3deffd,0xfabe47e5db1cefb5,0xe33cfb129e3cdcfe,0xefffbbdf9f3decfe,0xbfffcffbce7dfffd,0x95f6f26f1f7ccbf4,0x7ffefff34f1dcff8,0xfbffff2f9f2dff91,0xde7fbf8f8f3ccf9b,0xa69eca028b1cce9e,0x30405224060c4600,0x4bbe2df08e3dcae2,0xc190357a891cde98,0xc782aef3097cdad5,0xb7de5f9b8b1cef9d,0x74bdde6a4e7ccedd,0xbdff75638b2cefde,0xbbbb3f7d9b3dffdd,0xf90ffb8e9b6c3efc,0x2fb65f5e1b7ce47f,0xb09f5beb855cce97,0x99faefe7da388cf8,0xae9927ad4f1cfff0,0xebdff37fdf1ddffc,0x2e7798d91f3dead2,0xb7bbfff89f64ff98,0xf16cfebe8b19ffd0,0xffb625e2421c9297,0xa9c5ffeb9e3d7e7c,0xe9f33ed79b3c9eb8,0x1eb5dff98a1cee3d,0x55fefee55f7cdf7b,0xf2fe7df9cf7defdf,0xeb56f77b8f7ddfdd,0xfa6bfff95f3ddffc,0xff9e0ffb873dccfe,0xbdb1774e1b5dfed0,0xddf6ff731f3de799,0x4105fe5f9b0ce4d1,0xcfebfebada7ddfbd,0xffdbdfdb8f3cffdf,0xfeffffffdf7dffff,0xfbefffffdf7dffde,0xffffffffdf7cfffd,0xefffffbfdf3dffdf,0x29bddddc0f01f7da,0xb4567ce6cf1cfcdc,0xe9fdbf71df1deebd,0xff920df78a3dcffd,0x89edd33a8b2cef3e,0x80000108002000,0x85919250021c80d2,0x9294f70f9b1dc997,0xcf6e0d208b18ebe9,0x8104830013108c80,0x494008201010,0x6984d08811948d0,0x251c906243108a11,0x51244211a002108,0x202020480100480,0x6c8a92010808cad2,0x4610440000c0014,0x40040008002000,0xc2a00840080ca041,0xfb906710cb3cccd0,0x28879e04081c2080,0x261000180800c881,0x10000200000408,0x0,0x8000000400080400,0x6000000002000200,0x2100c00008201010,0x0,0x0,0x8aa60e6082040c98,0x884099b0004c0484,0x8100b4000109032,0x100003882000a00,0x1124042002080430,0x62860ce08a1c4ad0,0xc44310010c4460,0x200895c28a080d18,0x4098442802088098,0x500081c009214000,0x40490000108022,0x880410841108802,0x94c8f243c554c696,0x7a38cfdb5f5cf6d8,0x36d256635b3cdcdc,0x0,0x4100008002000,0x54a0e0a4a3090d8,0x234641d088195e90,0x680040009004808,0xbfcfd7b8b2dcebf,0x7e982f549f29df5b,0x4d298904218c0b0,0xbf1d45b4e3cefc3,0x35ddfdf7cb1d5cdc,0x400083a0080c840e,0xeaae5de2097cc4e3,0x4cd7f37e1f3ddff8,0x2fda1edb0e1ccddd,0x35386a72cb39cc52,0xf0e573f51a2dd4d8,0x4a447524801c9694,0x8800d2000b204900,0x2408b00002089492,0xb49f4fea1f20dcdb,0xd577f67edf39ffdb,0xcaa23e02823dcd3e,0xfbdcf73e5a3cbedf,0xe843a2934b1059d8,0xb574df48da389df3,0x5491a66c9b38aa99,0xfbacfc620a0d7fd8,0x6840804808140084,0xc8b980208284c88,0x40054280140a04,0x36340460081860d0,0x1094010808140640,0x0,0x110902080208a10,0x44000000008010,0x10002000080080,0x1080006001080810,0x4400000000008010,0xc00240108080,0x9100100054080,0x0,0x2086000100c001,0xba323d720b28fed8,0x8043a22a0b204a18,0x7eff7e58cf38dffa,0xffbfdff3df3cffff,0xc00fa0621a2c8514,0x6000440001000a0,0x4940f4228600d450,0xfad73eb7cf11ff58,0x2b19de490201cd18,0x4000002002000010,0x14ebd604b344558,0xff8b6b7bcb7dcfd6,0x60573e210b3cc5cc,0xa75edfde5b7deedf,0x74bcf5e84b7cafdd,0xfffefffbdf7dffff,0xd6bebffb8b2cdf59,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xdffff7ffddbadfff,0xdfbe7dfbcdba0fff,0xfbbff736dc9adf6b,0x8e03b74915ba05a7,0x8e9c4f7a752a0f6b,0xfd507eeb6d2a5fe1,0xdaff2f72a4ba57ef,0xef26ff7f4dba97eb,0x8fffffbbfdba8ffb,0xef9cffda7dba07ea,0xf43efee74d9a5f63,0xd3ab6f5935920fcb,0x8bb26a597c980722,0x8cf2b7371c884db2,0x6f5edb1b7d9acd4b,0xf7ff5febfcba9feb,0xf9f3af68dd8a5da6,0xaebc2d7115b80523,0xcfafffbd753a0576,0x4dffbffecd3a0f73,0xffbadfff5dba9feb,0xf1feff314dba4fef,0xffdc7d301dba9747,0x9afd3af2cba87e3,0xbee7bec6649a0fe1,0xdd3e6fef659a09eb,0xe0fff3eb3d2a977f,0xdbee7df26dba8beb,0xfbc4f31f7daaddcb,0xccfbb7c9dd3a5ff7,0xaf30df83d11a075f,0xf4d1d563c5289fa3,0x99fe7ff3c5badf6b,0xfd8ee77ffc2ad7ef,0xc9effe3b4dba8feb,0xeedddddafcba1dea,0xf4b8ffe8bdb84feb,0x98ca77374c3a9f6b,0xdfbebffa75bad7ff,0xeddffc8aad3a0f63,0xef7f5fdffd3a9fef,0xfebfffe6fdba4ff3,0xf8efb7fafdba1fe3,0xdef7af537dbacfe3,0x811dd7332c38c7a3,0x297dfccd3caacde7,0xb339cde3dd9a075b,0x9d7dffeeddba577a,0x7fbe3d7c5cba1f6b,0x8976ffaa64bad761,0xabeb3ffbfdaadf63,0x84ceccd6481a077b,0xffddf378fdba4f6f,0x7afefd737d3acfeb,0xabd5f732fdbacfa3,0x6c5f9c1af13a852b,0xf69855705c8a5762,0x38d2476c243acf63,0xfb7ab9702d1a0f37,0xfbb2e77fddba0ff3,0xcec3feb35d88c5b3,0x950118954820542,0xf6f77ff03dba57af,0xd34ef57fedbadfe3,0xdb927bf57d3847eb,0xccebd71a74a291eb,0xbb7efc5b1cbad7a3,0x347b57eed4929fe3,0xfdc4f74e74aa5feb,0xfeae3f7eecaa07e7,0xf3f4f37f6cba57fb,0xac89a5b05381753,0x359bdffac5badfb3,0xfbfdfef7fdb20ffb,0xdfbefffffdba8fef,0xcba3f36e65ba57e3,0xedde7e55fda80fab,0x43bc4ef1a49a5f62,0xfaf3377a9dba8fe7,0x7da65d6b55ba1ffb,0xfbffffff7dbadf7f,0xffe7bef9cdba4f7b,0xbbbfef787dba4f73,0xdcd6b6eed83acf83,0x93b2ad412dbac7c3,0xfbb5fbffdcba4df3,0x4ddffeadfdaa0f6b,0xffffffdffdbadfff,0xfffffffffdbadfff,0xfffefffbfdbadfff,0xfffffffffdbadfff,0xffffffbffdbadfff,0xcbb6ff5b7daa03eb,0xbdb77ff7edbadfe7,0xcd4d676395b857a3,0xfe8e2f70bdba1f63,0x408847121c201702,0xf81a11c4900520,0x930247d2358a15f2,0x53bfaffe5dba17e3,0x74823da18cba4d02,0x8280e56e901a4ce3,0x104c0000280020,0x1c469044b80700,0x9494be4c0cba4b42,0x2004002000020140,0xb890001804085701,0x3880020820100,0x880289100c8001e0,0x71f4b6fe4dbacf7b,0x0,0xc28003044c200140,0x848a38008c8a4404,0x2a51d4502cb219e5,0x4000106004300200,0xa84b2f5a5daa4de2,0x0,0x801001020000012a,0xb28ecd8008a08cb,0x8002000020004,0x2004060818100500,0xfebe3d579dbacfef,0x0,0xed563a008cba0504,0x8000442000a80100,0x20261080020452,0x100020000280000,0xc862f7fa15b85563,0x84606000c101100,0x64a05120c2a0906,0x1024008040020060,0x92a445410d1a0921,0x364c62ea12a47e2,0x5c1794840920da3,0xa73e4fea4d3acf2b,0xe0b7ffeb6dba9573,0x93b00c5005384702,0x4204020601000502,0x4140001000020840,0x697509427cb80532,0x3420462449800183,0x102514030a00300,0x9baf633f6cba0f3f,0x8c81ca080ca85140,0xbeedfd83d3a1feb,0x14d7fe6bfdbacf2b,0x911aa5602da803e3,0xd2b67f768dbadfeb,0xcdffff264d3ad7ab,0x23d9dad70db80760,0xfe2a4d7f4dba8fef,0xd3229742148a9d23,0xdbde4dd37dba0f63,0x84001080c820000,0xe48306cf041a47b2,0x35f96fd3d5ba4fe3,0xffd7f77a7c3adfbb,0x5bbe57f274185daa,0xfbdc677ee53897cb,0x8cffbf35fdbadf67,0x8bbccf4244ba5f67,0x4dd45668cd08553a,0xb0122d050cba0d26,0x8080001410020900,0xca8218060018450a,0xeb6ec9cb0cb89b6b,0xc4941021481a0d01,0x800481080c020304,0x5a24252044081833,0x141e8200d220246,0xa000841040280120,0x904400010000a2,0x1000040000020040,0x4002008000100,0x0,0x4480001000822,0x20680040001a0,0xd07d2c1231280f0b,0x888425202c208141,0xcb847a6e0dba5fe1,0xf3d76ed5fc3a4deb,0xdfbecfff7dbadfff,0x8001004040000022,0x18848400000030b,0xf9d6ffbf7dba5ffb,0xffffffbfddbadffb,0xbfbfdd7fcdaadf6b,0x5424c3606420572a,0x6918b531593a0d27,0xd8a47be1248a47e3,0x8dd3be1dbd289733,0x6f5fefdfcdba1fe2,0xe6b776f31dba4feb,0xfffffffffdbadfff,0xfffefffffdbadfff,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x94fb5fe0b0136ea6,0xffffffffbcffffee,0xdffeef73b4d3f6ee,0xdb96fbf6bc9cefea,0xd947bf5b94bbfbe6,0x8f9ece57b4bdf7ee,0x14c777e934135bee,0xcbbaff63b4bafce6,0xffaff797b4b6f3ee,0xc95ef6b31cfb6ce6,0xaf76ffdb3cf7fbee,0xe4f3f6c634d16de4,0xdd6ff77bbcd7ffe6,0xdfbe3f77b4bfebee,0xcdf7f33e34baf6ec,0xafdfdfcb34d37aea,0xfdd71ffba4f2f6e4,0xfb4ff7f334bbffec,0xfa9e3fff9cb6fee6,0xe9c4f13630bf7ec4,0x2ffffefeb4b7ffee,0xbfbffffbb433ffee,0xdd7cb77f3cf26eee,0xdfde3f77b4bfceea,0xfbfffb7e3413ffee,0xefffde5f34977f6e,0x39bdcf6f94d7f6e6,0xf7f7de6db4baf6ea,0xffbe3ffbb4bb7fee,0xfb70ff2f3c92feee,0xe7da795904d07ce0,0x7ffbeffbb4b3ffe6,0x7cf55febb4d7f6a6,0x9fbfbf60b493f8ee,0xd880db3e10b2feec,0xa77ffebf94da6ee6,0x779edfdeb493fee6,0xb43f5ef23495fce6,0xc9cf357bb4137cee,0x1994aff92493e8ec,0xeccfea9e34fa6de6,0xe3ffd7dbbcb6fee4,0xfdf65ffbb49bfeee,0xfbedf7ffbc93fae6,0xfe3e3f76b493fce6,0xc7debe2eb4b27eec,0xb56fede3cdee2e4,0x9e0a6ffb1492fdec,0xf5edaf3a349eefee,0xff9e7f633cffffe6,0xcbfbff6f14f8ffc6,0x7ee6fe1eb49b7fe4,0x8a94cd6bb812e9ee,0xd4fef6cb34bf7de6,0xebfe7df234b2f6ee,0xcbfee39e3cbafee2,0x8cc2784a249a5aa4,0x1a188d13149a30ec,0x556631d0041246ac,0xef2cad600492eac4,0xffbef3ffb4f3feee,0x5ac3ff133cb666e4,0x2a0f9dcb1898e8ec,0xf8f6bffebcdeeeea,0x5bad676b34976eee,0x8b04e26da49412ac,0xfa5ef7bbb4ba7ce6,0x5e1d7558b0126cae,0xc4fed76b149b76e6,0x597c7f0e14b16eee,0xeabf3f7e149efeee,0xafedff3fb4b2fce6,0xec52f24234b2c2e4,0xb7b2ff7b34b3fee6,0x95f5dfebbc9bfce4,0xb9967572bcbadfe4,0xebd6f66ebc9bdee6,0xa0f49fdf3c336ee6,0x6fe26bc394dee7c4,0xd5e5f750a49c6fee,0x9ba8bd631cd0fcc6,0xebf4f72f349076e6,0xcfffffff3cdfffee,0x67bdde6814daf2ee,0xf4fffee7b0f3f6e2,0x969c2e700412eeac,0xdb9dcaab34bb6eec,0xdff7b3b1bcb3feae,0xffffffdbbcfffeee,0xf7ffffffbcffffee,0xffffbf7bbcffffee,0xfffe7fffbcffffee,0xffffffffbcffffee,0x2fd67fdb9cbbfee6,0xbdfb7fefbcfffeee,0xd97d7ffa9cb3faee,0xdebe3ffdb4f77eee,0x9c6ef2aac127cee,0x695a3a4994d2f3e2,0x8b20473114123864,0xd189f6df3cb3fbee,0x1d067d2210926a84,0x88001b0010126a40,0x8844a40008a2084,0x210443208024284,0x44406691093a886,0x42000411100220aa,0xa314c006b490cc44,0x4244b642a0d04160,0x1880414224904ca0,0x41120220022c000,0x1080240000808280,0x200000400880000,0x940860100901068,0x8102c10204101484,0x3000000000101000,0xc000000020004,0x110000004000020,0x0,0x8080000000c000,0x0,0x2000002000000084,0xe6b2afde3c9a7fec,0x4000001010000800,0xef7f7bdf34bb7fe4,0x0,0xc510106b00123aa0,0x8000400000000880,0xc9f6fbabbcffffee,0xeed33ad0b4bffbec,0x128b04621482c2e4,0x100807224003160,0x1082040004003040,0xc0d4d82e30125b62,0xeedfd919a4bffde6,0x7ffed6fbb4f37fee,0x3df5fffab4bffeee,0xb5192c5534bff6e6,0x4084803028004040,0x8400a0004106040,0x4013c820002060,0x500094c6080062e0,0xc81502280402c484,0x8995777d3c9b6fe4,0xc13ef33834937aea,0x35fc61914d57c86,0xde3e7febb4b26bee,0x4f6efffbb49bfeee,0xffbe3fffb4fffeee,0xabb4fbfd3cfe7eee,0x7ffedfff3cfffee6,0x9dbacb31109a7ee4,0xe07da798b49854c4,0xda0425b29492f6a0,0x82328a0218d15042,0x64b05a7814996060,0xfffefffb34b37eee,0xdb70fe7d94b2ffee,0x4fbe3f63b4b67ae4,0xdb9ce33ebc3356e6,0x7ad3f7bbb491fee4,0x25300756a41232c2,0x50d176e224d1a2a4,0x8041400004005040,0x800001480402002c,0xa40000100004a80,0xb5e4f5bb412fae2,0x2004c0004304000,0x254080980000,0x1204156000880280,0x8940c80080106160,0xa802000030000280,0x100004e200114a80,0x20080010010000,0xa80000000121004,0x80900000906040,0x50101600111000,0x101001a014200802,0x0,0x800680000004a80,0x418186200088e2c4,0x79d21edf309b75e6,0x18cc69014105c28,0x18105040200048c0,0x0,0xebddf77a1cfafeee,0xefffffffbcfbffee,0xffbfddfabcdfffee,0x982060008a0064,0x22ba6d31340238c6,0x7a0c009a4484,0xdfc33e37b49afcea,0xa7d6dfd734bb4ee6,0xf4f5bfedb4fbefe4,0xffffffffbcffffee,0xffbfffffbcfffeee,0x0,0x0,0x0,0x0,0x0,0xc7e6260224102068,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x850840100022c0,0xfe9efffbf0fd57fa,0xfffffffffafdffff,0xfbfefd63f8fcfffe,0x2819334100017ea,0x8b621eddd095dffa,0xbbbfeff9d095d7fa,0xf0ff7ef79a1dbeeb,0xfbfeff72f2fddfea,0xc3b7bf7c329d77ee,0xdfc7bfbbf0fcd7eb,0xaf9fdfd3d035f7ed,0x9cd4e76852d4b4da,0x9bff77e3b25cfffe,0xdbb12579f2fcdb70,0xedffeffef8dcfefa,0x2ff9dfdff27cf7e8,0x36bb5ff67a9cd6fb,0xf7fcbff8f01d5efb,0xdabab77ff27cdfe9,0xefdff3bff27dffff,0xeffedfdff2f5fff9,0xbfffdffffafcffef,0xddffff78929dfeea,0xbfbe3dfed0dcdfdb,0xfbffffebf8fcdfef,0xefdffedff27cfeff,0xdfbffffffaddfeff,0xf5fef7eff2fcfffd,0xfffe7ffaf2fddffb,0xfbfdeb7fd2dcd7fb,0x87e7fff9f23cdffa,0x2fbb4f5a927dd3e8,0xf9f3fefc90fcf7ca,0xbefee533f8557eee,0xeb93efb9d09cfef3,0x8dd1380a783198e8,0xeffedf9ff27dfffa,0x74fdfefab29c7eed,0xc0ea2741c0c09469,0xdbdebaf61874b7d8,0xcec7ff127254be70,0x4ffedd1fd29d9ffc,0xcff27ffbf274ffea,0xdbfeef7ff25cd7ef,0xfebebf7372d5f7ea,0xf73affba78d4fee4,0x23563e583030de62,0xf6aee36ab01c37f2,0x536c9773f214b7ea,0xcf367f7e925dffea,0x8b82d7af32d4be7a,0xeed70b77f0b1deee,0xbffc6fff927cffca,0xd5f7f7ffd2fcfee7,0xffee7de2d0bddefa,0xfb86f7367a1c1cfa,0xcfd7bbeffadd5ef8,0x23becd31921c367a,0xbeb7fedc10dcb6ca,0xab33af52901cde86,0xd3bee3f2b0fcf5eb,0x8bd6fb1638941ef8,0xa94649c2b214d4e8,0x7664fee5b27cbef2,0xd973ee71d27436ce,0xffb6eb76529cd6ea,0x68d3e73a588096e9,0xeb6e7d5bda945fca,0xb42ed7c3d08c14a8,0xd92237e6d2b4de6b,0xfeaa2fdff8fc5ee8,0xe7d2fe725834ffc2,0x6b579ed9d21c1e5a,0x96bfcfeffab1f668,0xbb7d9fdb50d5fef5,0x5e82b642d00417ea,0x8be6f2cd50b456ab,0xaef6bf5210b5bff6,0x3efcc5db3014ce7e,0xf4ffee79921c17da,0x838e7d6150d4f9ea,0xd9afeb2e78bc7eea,0x4ffbbe4150d516ec,0xe9e5ffb5274f7e2,0xf9f73f7ef21cd6fd,0xdffeff73fafddffa,0xfafd7fa0f8143673,0xcfdbbabb505c7fde,0xffffffdbfafdffff,0xfefffffffafdffff,0xffffbffffaf4fffb,0xffbf7ffff2fddfff,0xefdfffbff8dcd6fe,0xab7e7ddfd0fcbfd8,0xf4ff77ffbafcfeea,0xb3e22f4b50207466,0x3a960f2292d0dc48,0x4dcadf29fa14d7ea,0xffdf7f4ef87c7cfa,0x8693757a7034d47a,0x11a497f21214dcc2,0x4c800d4230049662,0x7080411200801640,0x61540009104092c2,0xb831457ad01c3442,0x90098640408414a2,0x4aaa2cc0d21c9cd6,0x72806300d01034e0,0x9c22e04200017e8,0x87128f137044d4e6,0x3440526040011668,0x200560200404e0,0x840010080001002,0x600c00812100840,0x800010050000000,0x2220000100800,0x41000040008,0x80000102000040,0x400000800000028,0x0,0x0,0xa080002090000240,0xd2b62d7c7055177b,0x110000000001400,0x297d1fcbf2bdd6f8,0x78e45bbd01c16c2,0x0,0x4802240200501340,0x8b3ffbeb72e15768,0x4f4a0e3020215668,0x8880298101c3040,0x1400806010008240,0x44060d22401c0e30,0x2ba4ab0fd004d5c1,0xefffbffff25cfefa,0xfbbe5f73da84d7e9,0xf5ff7ffbf2d4ffff,0xdbaeed2070b45aea,0x21003040004048a,0x840021110d01410,0x2b4034810346640,0x2004024010204400,0x8824200020801424,0xa2144f647060b068,0xd000993850115768,0xca0ddb5b1214c7d0,0xce3bdfe9f26dd6ea,0xfffdeff7fafcfefa,0xfe9fbffefae4f6e8,0x71d6d7edf295ffe0,0x2bbfdfdff27cfeea,0xc60226e7fa909f61,0x97fdbf9a32347ffa,0xdf044d7fb2545c7e,0xe924f32f3210de68,0xae4316eaf8101ffb,0xbeffefff6afcffeb,0xff75f67d52b55efa,0x6c9631615a1cd542,0xd9d1ff265aa4d7e2,0xfd4bbf9ad014fef6,0x6b1e1f7fb0955761,0x2052fd40141cf8,0x3082002000250062,0x8000010010100140,0x400260140801020,0x8475d529014149e,0xf7ed5dea783cfffb,0x98e10720921440d2,0x8820090210100480,0x8081020000011022,0x830040020100220,0x8010040020100820,0x24000000000401,0x1020060c10101218,0x4102c30810841008,0x540022818001408,0xa00004000,0x44410000000c240,0x3000000000005000,0x1800080001000,0x44c1ae2830b99ae0,0xa0444000811028,0x8420000000105200,0x10010000000240,0xc9dccbbf787cf6fa,0xffdfbffbfafdfeee,0xef3fddd3f291f6eb,0x0,0x91c023008141ac0,0xfb966d66b8fd17ec,0x4fcffeab7af5f6fb,0xeffedfdaf0f5d6e9,0xfdfffffef2fcf7fa,0xdbffeffbf2f5f7fb,0xfff7ff7ffabdfffe,0x0,0x0,0x0,0x0,0x0,0x2080000001008,0x28021108121012c0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa8f61bd0a9e4a36c,0xdfbafffbf9fd23da,0xf5fff76df1ed2b7e,0xfffefdf3f9fdaffe,0xfb7ce3ae81ed2e5e,0x9fdbbe9ff1eda36e,0xbf9edfdbf1efaff6,0xd1d0ffecf9eea3c6,0xd7bffd62d1c7af5c,0xdfae6d6fd9efa76e,0xdfffffbff9eeafde,0x9f776fdfd9eca34e,0xf4b5de7cd9e52b74,0xed6cff73596d365e,0xbbbcfb7fd1f4ab7a,0xfdcffec2f9ed2b6e,0x2abedf9be9dfae4c,0xb7ffffffd9ff2e76,0xf76f3f6af1652a7e,0xfebe7ffef9fc237e,0xc58ef63fd1ed261a,0xafdbffdbd1ff22d6,0xb7bffffbf1a7bbfa,0xdf74dd19d9a4388c,0xfbb47775d9fdbbfa,0xf9fffbfff1ffb3fe,0xfef7dfdfd1efb3fe,0xfabbcdfbf9ffbafa,0xd7f4ffe8d9efb34a,0x7fbe7f72d9ffa75e,0xbbb5f75ed1bda34a,0xcfeb7feaf9cfb37a,0xebbfdffbe9eda3ce,0xfdff57ffd9ffb35e,0xfdaeff77d9f7abfe,0xeb38677ff11f237e,0x4bfcfc27f9e5ba50,0xebfecfdff9ffabfe,0xf4faffecd9f5abfa,0xd4eebf71d071b34e,0xf3bc2f77c9edb77a,0xe8f5f902d189aa40,0x2bfedf92d1eda35a,0xe5ff5fecf9ecaf7a,0xdfbafe5af9cd3bd2,0xdab67fffd9f73bfe,0xe1fdff79f1edabfa,0x3dfbd759d8cd2a5a,0x8382cef3d8cc275c,0xd57eff7cd9c4abda,0xef367fe0d1e5a85a,0x83eeff76c1f5bbce,0xefffbed9d9eda3c6,0xff7f4f62d9ffa35e,0x75f6f6f9f9eda766,0x9fbe75f9d1ef3f56,0xf8e4e33ed9dc234a,0xee533ffbf9ef235e,0x86250970d965a846,0x35bd5fe8f9ec2ade,0xc93e4d43f9ed2b86,0xd39e35afd9c53362,0x6ece320ad9c53652,0xa35acfd9f9fdaa4c,0x94f97265f13c2016,0x8decffe3f9efa2c2,0xc700eb148145035a,0x2587e72ac9cf2352,0x2b37ff5bd1412bee,0xbcf977eaf187a3f0,0x510cf170d0ef2848,0x4abd2f32e94ba3f6,0xcdfff374d1e9a266,0xaf5e5e17c8f42b68,0xb9bacfeff9bdabe8,0xd6f5ff7ff9df237e,0xfa046534f109a35a,0xebf6ebbad9f1bb6c,0xc8c71f6fd9f523fa,0xd8bfdfebf1cfaf7e,0xddfb767bd9e5a3de,0x7b7efd7af9edb76a,0xabf77b7ed9ef3bd2,0x64cb32cac124a24e,0xfebecdb8f9fca740,0x9cd5ff62d9f3a37a,0xfffafd73d9fdb67e,0xc83cf7b3c9f1a2de,0xcfdffe3df1cda3fa,0xffffffdff9edbffe,0xfffffffff9ffbffe,0xffefeff7f9efaffe,0xf3bebf73f9f7a3fe,0x8dc3ea9ec1dda2de,0xafd8cd4bd94d275e,0xb6bf5ffbf9edbbfe,0xd1b60f32f0ec235c,0xd3b22362c8e92b52,0x46d3f3bad19cb24a,0x49da9ab198e7a24a,0x35ba4168d1c0a2ce,0xd9f0a25281692b6e,0x1a00652dd1842062,0x88202a408141221e,0x23418880c10020e2,0xb0137ee2a167237a,0xc32428c809d0258,0x90040002000,0xa0c20201002006,0xc4a22c2d0402390,0x25bb4d60c04c2b1e,0x1001000001000008,0x600180002800,0xd8000a240143024a,0x1020000082000,0x400400000000210,0x16000044c0482200,0x8014056088800840,0x2080011000000,0x800d00040001140,0x2301108209400212,0xa430f024c0a82042,0xc24040402200,0x12102724d1412ad2,0x5350650aa9c92372,0xeffff95bf1cdabfa,0xf7bbc5fbf94dabca,0x41000000000140,0x52a02c4cf16c2252,0x8976f28cd0282042,0x8ff63f5bd1692a46,0x88a8b52c0c02712,0x11635041d0402470,0x262521c0682140,0xdbdcfbbff1efb3fe,0xcff3febad1dfa35a,0x8fbefffff9ff275a,0xfdfff3f8f9cebfda,0x9b78fe73d1cfa762,0x1004000000000210,0x2982080510140,0x6e759f4bd9c5ab5c,0x24008400000,0x202006524124200a,0xcb10767cd1af2a5a,0x82ffeb3ed82c3a46,0x63fe9f5be9ec2686,0x1da637e761e7a26a,0xd97fffe2e9fdaf7c,0xdcbe2f76f9ed23fa,0xffdefb6ef96ca37e,0x2dfedfdaf9ed3ada,0x9ea26f7ac96c2a1e,0xf5bde472f9cfae6a,0xffdeef26c1e52b5a,0xc9f7ba67e1ef2356,0xeede1662c1a6a25a,0x931dc2fad1c1a3c6,0xd576fefee9ffa3ce,0x7c246cf3c94523fa,0xfbdcff3ef9efabfe,0xeecef633d17cb3d4,0xc4760724d9d4231a,0x95910c68d0a0a862,0xa80000000012044,0xe0008418d08020c2,0xcbde1213d8f42946,0xaa00ed8bc9f02282,0x95bd4c20616ca332,0x100302088500000,0xca000706912c2800,0x3880000240200,0xc50c210c0412480,0x1b2067b516d03d8,0x5000168040040270,0x800000800400200,0x89461faa718c835a,0x140000040008000,0x48000202000,0x43010000c50000,0x8865440c0182800,0x10880000800,0xc873ea0d8082142,0x9094c509004286c,0x0,0x400045048402000,0x53d4c42ad1ef3268,0xafffffffd9f7bf7e,0xafbcdb50d1feaa54,0x8400102090400100,0x22322500d060a9c0,0xcaabf777f9ef2f52,0xdf46bea7d9ada27c,0xaf7f5fdaf9fdbffe,0xd677defdd9cf23c2,0xdfeefff3f9fdbfda,0xffbfaffff9efbffe,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xab46fe22d53ca502,0xfff63fda99fcff60,0x16bdcf1b9f38e702,0xf77df3ffddfeff42,0xfffefdfbdffcfff6,0xfbfefb7fdfeaff52,0xeef3b7fbdf6eff36,0x179fc9f3d9ecff30,0x12b3d6e898bce7f2,0x9fee7f71dffcf7a2,0xdbbcff6ddabafd60,0xcfd2773bdfa8e5e2,0xefffcfd7dafcbfa2,0xf4ead6e48f6ce686,0xd9a54e6b9d7ce740,0xd39c7f16dfec7566,0xcf63ffbfdf7ef760,0xfb7dced0d8e8bde2,0xd4f777f6d13cfff0,0x9ffde45b1bbcff42,0xfb9effffdffcfdf0,0xcb5fff2edfccfd30,0xefdffe5bd824ffa0,0x9ede4fbb57acb774,0xddfff7ffdbeefff6,0xfffe3ff7dfecfd74,0xebfffb3fdf7cedb2,0x7edfbffddbfcff74,0xdc9fef7e9a3cff22,0xfdf2ff71dffcf572,0xfffefffadffcfff6,0xfb7cfb6edffeff30,0xeffbfbcedf6ef790,0x3fffcdf3dbfcfd94,0x7df4f7f69beefdb2,0xffbedf71ddbefd20,0xd6945d7f97bcb506,0x45dbbfd118f4e7f0,0x2d9fd7d39fecfff0,0x763df7efdbbce7b6,0xeb6de773df767760,0xff9c7f7ddbe8f7c2,0xc8c3bf3cdfe86764,0xeeffdfdfcfecfda2,0xbffe5ffb9ff8ff30,0xdbedb77ed9f8fd96,0xfaa73f79dafaa530,0xcfe66f261f28f790,0x6ffedfc31a2ebd66,0x7e366dc9fecb772,0xf330377951faf9f0,0xdb86fc4358aeedd0,0xcbdff7ef9a7cff32,0xefff1ffbdffefdd2,0x3ffecf6adbacfff6,0xd5ddf7f8d2aee716,0x69267ee29788e730,0xbf5c30eddfcf160,0x85ef7e99d86cf5e0,0xcff64f6097b8b7e0,0x7765d2699a24e590,0x9bde6d619320af50,0xff86ef7b1b7cbd84,0x4b82a29992e23540,0x3b9c5cd29f64ad02,0x10b87ee91dec3626,0xe9ef6de318fcf742,0xffb88f5e9dec6500,0xc847e22f9f08ad60,0x2d5fff5c5ae8f6a6,0xb4be1bc3dabcb704,0xf1fbfe6a5ffce534,0xdeb0277e58fce9a2,0xefdffbab9a7eff70,0x2eefbecf1b3aff54,0x6bde7e59838ed10,0x947cfdf9dbeca560,0x9b2e242c11388df0,0x6856f35a0fbced44,0x4ede1e1b5ae8f522,0xa3becf77dbb8f770,0xfcfffe77dbaefdf2,0x6ea42ff015acbd90,0xd9a25b7e9d28f552,0xce73ff705328f750,0x68be9f7b9d686774,0xf7f7f77fdfacfdf2,0xdb2afd60ddf0fdd0,0xcca37348187c8d32,0xcb7fbe3b5decfd20,0xafffdfdfdffefff6,0xfeffffffdffefff6,0xffffff7adffcffb6,0xdfb6ef7d9fbcef52,0xc80263985088ac22,0xeafa5dcb9d88fd20,0x963ffde39ba8e5e2,0x990e476a9f40ec24,0x9a8606421ae80522,0x4786de80996e2402,0x297e7e92db6cfd42,0xe825f569a482532,0x803074401200a510,0x6860742102300160,0xabf4eb2edbfc7d20,0xc461a111d406510,0x929ac3889a202540,0x100845800808d00,0x284e156848486522,0xcb81e6a48a44b560,0xa0602c2000884400,0x668058a1f001560,0x44406c010001140,0x19a617d08d642600,0x462226468b244e40,0x810290c018202520,0x490080040003000,0x411000405000120,0x640200002c0100,0x0,0x60008040000,0x0,0x2020000000600,0x50001000c00,0x10042a2c90208980,0xe181e93ad5ac6520,0x4274000910009110,0x1a092a9042a0b0,0x2404002008204400,0x200010020000,0x49d5cf0cd862ff50,0x201002db00006704,0x40100a005084000,0x514446210286400,0x1ac411200a20a440,0xca80a20695a0ed20,0xa846bd489288f780,0x6e9fcbcb9a3cfd74,0x74f7ff6a58fce772,0xfbbcbf7059f8ff64,0x220c24205030a520,0x800000000000a000,0xb65cdc57db38ed52,0x13100000a000500,0x7882460d080f500,0xa31675f85d72e620,0xc0462a3012a28530,0x6ed7da995bb86d60,0xbcbf7febdbe87d22,0xf1618ffc9f28cd60,0xfe872d70cbb8e582,0xed5e7ff29f78f530,0xaefaffc89af4edf0,0xef644f77cb786540,0xb0dcb674db4aef30,0xcb7e5de15bf8fdd6,0xdbb2ffeadf78ef02,0xefd67f5fcbecede0,0x1f98ff83cae8e720,0xd6efaadc1feaf522,0x1000014000088900,0xfbd7e73edafc6576,0x8eebfe719a48fd22,0x3b3c4f32d878ad00,0xf5babefed500ef14,0x80006d010ac08580,0x100400000002400,0x87fefa395facef72,0x2f58c31e1138ad12,0x4210142898c00120,0xd36ea722d9c0af60,0x1b61613d7c8af00,0x20100080020,0xa10040910000410,0xc92178390402430,0xd02020609d200042,0x8a1804300300bd00,0x8300c22180009c00,0x208014108c02500,0x20428008000142,0x0,0x4240d610870c504,0x840800800000602,0x697e1bd693aced00,0x531ccfd94aace726,0x400200040002000,0x4000001000000402,0x415496045508b534,0xebdff6ff9faeff62,0xe79ddfdf9bfefd32,0x2020000240000,0x82900c605d00b500,0xc288631e90c8a510,0x874eb21b12bce530,0x2bfedfd2d8f2ffb2,0xfcf7dfffdffae776,0xfdceff73dabeff72,0xffbfff7fddfefff2,0x0,0x0,0x0,0x0,0x0,0x0,0x910f4410a82540,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xcbf7f4ab2f926f63,0x7fcf1edaab8e7f60,0x16bdedea0fce7d71,0xfdf7f7f5a7deef63,0xe7fe7fd13bfeef62,0xbbf6ef7e37f67f73,0xbfdabfcf0fcf6d71,0xe7b74fd333b66f61,0xfcff7efdbf37fd71,0xd3bbfd629fb65d43,0xfc9eeb3d9bfe5f63,0x4ff3f75b9fbeff61,0x6dff5dce0ffb6d72,0xbcfe5efcb7de5f63,0xdfffcf60bffedf61,0xca227f260f9e8f70,0x8dd7ef6f9f8f5d73,0x27745f9806de4d62,0xa4f7fde38ade5d70,0xdffffff3bfefff60,0xffbfef7fbfffcf73,0x67a6f72b2f9e7f61,0x2ff67ed393f67b40,0xa6bafffb1fbe7f61,0xfa73ff6b1fde7d62,0xfffe7ff037fe7f61,0xababff3f9fbf7d73,0xeff7ffdf9fff7f63,0x9fbefff73fcf5f63,0xf6d6b37f1fd65f63,0xdbbefd6bbfdedf72,0xfbd6ef7f93df7d63,0xfdf39fff87feff61,0xeef69f79b7d76d61,0x7ff6fe692fde8f60,0xcbfcfd721bce7f73,0xdbb53f2f07efee73,0xfee63e2f06c64f63,0xaf3fdfdb969eef61,0xf4ffffeb8fffff60,0xfda6f76a1f8e7f63,0xf6befff7bf974f63,0xfef7febc9f9edf63,0xabfedf5b9797df73,0x8db9dff307ee6f73,0xd175f66687cef960,0xfeb2bff29edf7d71,0xac4742210a04d40,0xeb7fdfdf97de7f63,0xbeeacdb9879e5d61,0xb1fcf77f979ef962,0xf756ff758fdeef63,0xcac4fb2ea7a6e960,0xebf77fdfafc6df63,0xdfb9f7fb37d65f63,0xfc77f6fd0fde6f63,0xfffeff7b1fdf7d63,0xd9c0edbf8fdecd73,0xfae23ad79aee6d62,0xff3dcfc98b965d72,0x9f21ef88cb86f63,0xbffeff73bf3e6773,0xdbb7f73e9b8e7f63,0x4b423ea319c61460,0xab67fdde83eafd72,0xf4955668a70c3960,0xcdfb77679fdfff60,0xabb16f7d2bcecf61,0x2bd7cf1a0fce5f60,0xbbff5fd8179e7763,0x9093deeb17de5d63,0x95eeffdc3b9f5f63,0xdeb12f6a97866960,0xc7d0fa96177f5d60,0xc09c040396823a60,0xa29a274896d34b42,0xd955b74d33e26f63,0xf6be6dfd9bdf7f71,0xc9f6d7a7af821f61,0xeb469a5406868d61,0x9a9ed77b8a9e5f70,0xf5b7f678179f4d63,0xdfbcfd723fef5f63,0x4b30760a0b087d62,0xca793a9797978960,0xaf3f6f709f877f61,0x71145e88938e4d22,0xdfaffff19fbff963,0xd304af4e09625543,0x5fd73aad9ede7d71,0xbfffdfdb1ff65f63,0xdcfffeeebfffff73,0xfdfdfffabbdfff73,0xf99a777d8bc74f63,0xd5d1b7121ec14f60,0x83efdfd38fae7f73,0x959b7beb23ce5f61,0x95248e600f860760,0xca8520212ac46821,0xdd6e34ba1965f72,0x6fff1fd09f8f7f70,0x411a8ff11ec62960,0x0,0x640244001820160,0xd080020120004461,0x80740e0630c04960,0x1014442000004550,0x10440642000a0860,0x30074001225241,0xe100e67001035a23,0xa480100a83015a02,0x80100000820,0x102cb44004160140,0xcb80c40003904f61,0x800160000c00140,0x80080000020100,0x18104002002400,0x1400000002008000,0x0,0x9004080000080820,0xc00000005000,0x0,0x0,0x0,0x290010401024042,0xdc0c77932de4562,0x0,0x0,0x0,0x0,0xe184900c31861020,0x10000200000060,0x39c5cf005a64361,0xd1e43e6b1f934d42,0x2aa6d401e8c0560,0x8b82043100864763,0x40020000100100,0xff3a097b0f824f61,0xd0d616e8021f7d63,0xdbbcfd7117de7f73,0xe28404e601920560,0x2045b21a00064552,0x8adfd7530ffe7f61,0x14218ce004c05b40,0x42206502012a0060,0xdbfc776f8ff77b43,0xe15cf93a067e4d61,0x13de5f5117877f61,0x76bb6feb1fde7d62,0x506ca4748fa60760,0xc0863f5c1ef66962,0x4dc1f78e87de4d62,0x2d57dfcf058f5f70,0x89e103063d40,0xd231aa5516ee6570,0xd1de7f778fe67f63,0x8bffdb7f3fcf7f62,0x4fca5f971fd34f60,0xbe3d67ec1b8edd63,0xb737f6763bd64761,0x59ca3f420fce6f62,0xd99ccfae03565f63,0xcccadf1f8f375f60,0x299a470a2a8c6d63,0xf85bf25087867f41,0x8622d4104860920,0xf1046f1406122d60,0x201800100c04001,0x804ac158008c0b60,0xf4bdfce08fce6d60,0x8075a500010b1e60,0x21882c15001a6c21,0x80811001020400,0x2266d292169e3120,0x0,0x9de24942a4824860,0x40100000400,0x4480060004020860,0x410020800000d40,0x0,0x1000040000000060,0xa14043000000470,0x0,0x10100008000400,0x19888dc91f825561,0x8001006010880560,0x0,0x4400000a000000,0xefcfbe738fd6ff63,0x2b1ec75b95ea5b61,0xb0feaee806966b40,0x482645c100820740,0xcbe0c538861a4d62,0xdff6da0c16d3ad60,0x6fbfdfdabfcf7f72,0xb4def6e98ac6df63,0xfffeff7b9fffff71,0xffbfff7da7ffcf73,0xe001890884840900,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x461226ccd3047d78,0xcbdbfb3ff7bc7eeb,0xebef5f1bf7b475e1,0xb7bddff3ffac6f68,0xfd7ffff6f7947ff8,0xcffef7f3f7bc7fe2,0xa354f76ed6bc7472,0xf9e3fffff79e7df8,0xffaecf7fffbc7feb,0xfbdf96ead7be6df0,0x3bfefff2f61c3eca,0xff957e7ff5b46ef8,0xcf5fff17feac3ffb,0xa79e5f42d61c2dd2,0xe4bbf668f4147ee0,0xffeff77bf7967de3,0xffe56f50f5946be2,0xe7dbffabdebc7efa,0xebfeffd9f79e7f6a,0xf5ef7dfbd5bc7ef2,0xdf7def62d7be37e8,0xdf8fbffff6bc7ff8,0xcf56ea78ff1c75e0,0x8d53be47569e5690,0xb0b6fd63fc307dda,0xfd7cbfffd7b67ffa,0xfab67ff7dfbe77e3,0xfbf6fefff7bc7ffb,0xefdbffceff9c6de9,0xb6bdccfbff146ea2,0xf7befefef68c7dc1,0xdb767c7bfdbc7fc2,0xf974ef36df3c7ee9,0xefd7bedbffbc7de9,0xbebd4fe8d4bc7dd0,0xd5ff7fefffbc6ffa,0xb3bd6de0fd844dea,0xfbf7ff3ff79e7ffa,0xebdfff7af6bc7de8,0x6f7ecfcef7be7ffa,0xf4f6ffebdfbe7fe1,0xd9eff76bde1c7da3,0xfffb7f7ed79c7feb,0xffdfffbef7bc7ff8,0xefffffdfffbe7ffb,0xfefb5ff3f6b475e2,0xd9af6f7add9e3c6a,0xcf9a7d79fdbc6de8,0xcbcbd3bbd79c75f8,0xaffedc4bd73845e0,0x969e6f73d5346fc9,0xd1bc97dbd0bc7c70,0xff962f31ff344420,0xe9ffe73a75b43cf3,0xfddfde5ff79e36ea,0xddbecffbfdb47deb,0xd5fbf27fd73e3ef0,0xdbfe7f7af4be6fe8,0xf9c6fb6ad7b47de8,0xdfc7fef7dfbc6dfa,0x2fbccffcdebe7c40,0x35e7f66fdf9e6ce1,0xbb4a6d62d79c2c40,0xd8b6b36ddcbc6dc0,0x41569a0452902522,0xa9251cca788c2cf2,0x80f0d4fa80bc6da8,0xf7eb8f7af6bc7ec8,0xcbb43f7cf5bc7ee0,0x68cfefbcff946ce1,0x9fedff5afeb46fe1,0x8b725fe3d6bc6db8,0xf16dff22d5382d70,0xfe9a2f36de343fe8,0xedd4773bf68e6ce1,0x61240041c01401c8,0x90bacfe5fe9c3cc1,0x90511f2fe5b01548,0xc2d62d7652bc3600,0xf890f812d43e5cc8,0x43c95857ddb43600,0x1d8e4987d7b42cca,0xa5fffa5fffb677c2,0xcfaa3d11d6ac6ee2,0xf212614fd51c14f0,0x4e7137fef6be6dea,0xa2bedff4febc7ce8,0x30336ae0d13c2c62,0xdfeefff0ffbc3d6a,0xdfa0434357b22440,0xddfffef7df3e3fe1,0xefffdfdff7be7f62,0x34367eeadfbc7da2,0xffee7f79f7be7deb,0xfab6bf7ed7bc7fe2,0x8d6e77bef5b46ef8,0xb5e8259d49c35d0,0x3cffffe3fbbc3f69,0xf490806054003640,0x46843d20821c61c8,0x8a82d3a24e8024e9,0x39240a44c00e0500,0xbfbc0d6ad53e7de0,0xda50120856842240,0x718e1458560424e1,0x180c02352000442,0x800000100200400,0xd2260ce379982641,0x9400366a54a22d50,0xc882a540d4940520,0x1392a332551c2b01,0x80e00010c0802102,0x82ec58fb09c6080,0x1001000080000100,0x20060000143800,0x84805150830c0420,0x8120c40190300a48,0x10100000002001,0xc41010a010242ca0,0x4108200000200460,0xbb2667774ae13e2,0x50000040000080,0x2400000044000000,0x8420202014000240,0x0,0x2240110080420,0x100811000080420,0x208804042022001,0x34bf4df3d7946de0,0x120040044002800,0x0,0xa164e00672222d40,0xe9c61a0b77902e40,0x7214450801a414c0,0x9138ee69749e3c60,0x419c8000c2c01,0xf2f46302f79c4d40,0xedd81151c01c7d80,0xadbe6d22f7902f20,0xb0dc1fe1dcbe67a0,0x13aead70c6b03d00,0x280000140001400,0x802300014100200,0x8f5effd3f7ae2de0,0x400a4e4438c3420,0x8022422440c0080,0xdfb5efffff9c7fe3,0x88d210bc3a082420,0xaddedfd7f5bc2fe3,0xbcbf57ebd6b4377a,0xd7682720f8867440,0xffba6d56d59c7fa0,0xe306c72297100760,0xed53f58d59c60e0,0x6982fdfd4a86670,0x41ef77ead6bc2fca,0xdeb6bf7cf7be67f3,0xfbf7ffbfff9e7f79,0xdecf5f5ff7967df8,0x61d4a63889c5460,0xb4ea9e20d4144380,0xc30a3d02b28c3c49,0xfba29b1ed7946dc2,0xeb43b76ffd3c7c70,0xafbf4fd3db3e7d40,0x90f51fead69e2d80,0x9b38ff71d6be3fe0,0xe80844e010043d88,0x802328002ac2400,0xafdcdf5ae4946cf8,0x408c549100180428,0x5320002045100400,0x90810c2250101450,0x840000040000080,0x280400c041900400,0xc705112d6143800,0x8151a109d0046a20,0x8000030410000420,0x8000820400000468,0x0,0x2010420100400,0x2880000100,0x1020042091000400,0x8000000100800040,0x4000001000080400,0x5285440bc300460,0x2040320116140400,0x0,0x8108c002c5040400,0xcfdfbf3df6bc6df0,0x7bbecf52dfbc65f0,0x801042e0009c6840,0x609aadf144b47ce2,0xd924f37c753e0668,0xcfff3e37d33e6de8,0xbdc0d48c6166400,0xbd365efcf69e0ec0,0xebeeff77ffbc7da8,0xffff7fffffbc7ffa,0xebfdef3fffbe7ff1,0x0,0x0,0x0,0x0,0x0,0x802100081000000,0x0,0x0,0x0 },
	{ 0x0,0x880118083022002,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd63269440bb8270e,0xfb87ef73c3b627bf,0xafde9e9bbbd6261f,0x9fbdd799daa2271d,0xd42ff67c8b9a2f7f,0xfb9e6762bb9a273d,0xcb74fb2eabd0264f,0xeee3b6128b9e23be,0x7fbadf5b8beb2fef,0xf45ffafbabe4231e,0xb7beff73cbf2233f,0xdfb67ffedbfb267f,0x8fdef61b8b9f2bcf,0x834fd6428bc32132,0x9cde7ee7dbbc2f7e,0xffefef7a8bf927ff,0xff9cff7fbbbe277f,0xcfffffbf9b9e23ff,0x2ffefddeebbc233f,0xd7fd7f63dbcb277f,0xb3fcf7f7cbb22bbf,0xceaf2f778bc02f3f,0x6ce237b6cbc22f3e,0x4b7f5cdfa3162307,0x97f37ef09bce2fdf,0xf5ffb7fb8baf273f,0xdbf6ff7fdbbe2b7f,0xebffff7febff2fff,0xedff7fd7dbf92777,0xbfffdfdb8bf6277f,0xf777f6fb8bfe227f,0xffbe7ff1dbfd27af,0xcbb4f73fdbb223b7,0xeedebb979bf42b7f,0xff3fdfdb9ba627bf,0x75fff7f9fbbe233f,0xdbfb2d63dbf22f7f,0xfbbdff7fab9b277f,0xdfdeff1febf627fe,0xaffedfdb9bbe2f7f,0xb7b7ffeffbfd2f3f,0xf5eeff6b9bba227f,0xfbbe7f7edbfc2f3f,0xfefffffefbff2f3f,0xefffffdfdbfe2fff,0xbff69fe79bff2ff7,0xf0fdbfe4cbcd2b7a,0xfebf2feb8bdb27f2,0xcbdaff7efb902f9f,0xafd63c928392266f,0xbfbb5ff99bfa277e,0xf7fc17548bba273f,0xbabe7d6d8b9e2756,0xc9b1ffea9bcd2f3f,0xeddf5e599aec2b15,0x9bbae7bb88dd2f1e,0xf4f6fad0ebd82766,0x5aaeff728bd4237e,0xdbf6ef77ebff2fb7,0xbfdfba7bdbb22f6e,0x6bbdcfff9bfe2b5f,0x56ffb6fe8bbb2b37,0x9beef772cb1f260f,0xbb906f759bde235f,0xeda3a1beb88211f,0x8ada19c28286238d,0x54957ee6cbf8278d,0xad6e77728bf02a1e,0xffbeeffeabfe27ff,0xcd63ef1b0bbe2369,0xffffffdffbba273f,0x96933e7bdb93237f,0xf17cafb88bca2676,0xfffeaf7a9bb52fff,0xefdff7f78bbd2f1f,0xac31e4e0b94022c,0x1e047bff0b90232b,0xdcfc2f608bed23ff,0x7bbe7ff4dbae239d,0x83f1e136d3b2279f,0x4bf75cd50bba2faf,0x9c7fddf3db9b2bef,0xf487dec9cbc0270f,0x420c45628a282204,0xa98dcb4c8bae020e,0xe43b600f3962306,0x4194a0041862287,0xf5fcfcee8ba4275f,0xe0b9ec619ad2230e,0xb3a4d1fb8bd8232f,0x58c7ee5b8bfc2bbf,0x2f7e4f4bdbf823ae,0xb63bffc08be0217f,0x59aeb7d783b42eff,0xffbeaf769bff2fff,0xc886fea8a396262a,0x66fe5fcd8bfe2f7f,0xf5ef5feebb75268e,0xd9eca762819c2fa6,0xfabe3e7fdbe62e37,0xd9f5f33e8be62f9a,0x4a0e120101b02217,0x36027c801a412686,0xd080300208882420,0x5c302e4c0b2a2383,0x80640260b2000b04,0x122080000302,0x12944c8403860320,0x0,0x40060d208b060302,0x820010000004,0x8421c108ac02101,0x659819128bdc0392,0x43642e21b58022c,0x45200840a1142200,0x6000202088082000,0x40d4000410802414,0x2010000802000,0x4806000802402,0x40061040102200,0x4000a4c01802018,0x80800002000004,0x0,0xc00002000106,0x100ed0209480268,0xc88605e20bc30322,0x10841220a480000,0x90c2000b0800006,0x848905a013022a25,0x1200100802802000,0x480204001800012,0x89d4d4869bd0233f,0xc95806339a702705,0x68a82d920aa0221d,0xf560d2709b902386,0x1c14047003940214,0xc980e33d8bca2abc,0x89539b918b90273e,0x771bcf1393902766,0x4490f6ebba902327,0xdd2e2d708ba0223e,0x100400000002400,0x164a021303822a34,0x8dfedf5c8bb12b96,0x652509a0f840260e,0x1002140001302000,0xffb637758bc823ff,0x2f5deb06997522df,0x6f7cffdb8b94225f,0x9fbf7f63d3ae27af,0x996444685bd02027,0x5ebd37fecbc02b6f,0xcfebfb16cbf22b27,0x32d49f6a1b8a276f,0xef3b4ffabbd823ff,0xddfdf67f9bdc2f7f,0xffbebff1abbf27ff,0xebf6feeffbbf2f7f,0x7edfb7d5dbb2270f,0xcb0e4f128a902624,0x9175fc4b9bb02305,0x498a3543eb38225b,0x62c442069b00272c,0x894e901703962b0d,0xf28ecdb19b31274e,0xfdf5ffeafbf7230f,0xdebfbf728b9c237f,0xd82e058289c8026a,0x9d3920282882607,0xaffecf5b9b9e27ff,0xfcff57f9fbfb237f,0x8000258209880250,0x92b62e7ccbc023ec,0x0,0x808881702202304,0x9492026881802033,0x81382a608280230f,0x4084288088802209,0x49d6f0208ba9275a,0x0,0x401001000040,0x954105688bc00032,0x821c410081900050,0x8000080000802000,0x284502900300032e,0x89a0ed50cb84222b,0x100102000004,0x10010000100200,0xd2a0004289882213,0xeddabe4bcbf82f3f,0x4014080000002604,0x602c82420b900a0c,0x130234600103262a,0xdd8027758bc0279c,0xcdcafb8ebbdd2717,0xabd8ddd39bba271d,0xb6fe57fdbba02a7f,0xdafe2f6aeb9c233f,0xffb7777f9bf92f3f,0xe9ffff7efbba278f,0x0,0x20004000800002,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf62e3bffeb9cf7fe,0xabdefb3fdba5dfa2,0xe8fe7f796394d3ea,0xb6bdcff2fb34f36e,0xfdfff678eba5dbbe,0x5f3e6dfbfbb5cbfa,0xcbc4ef7ecb95d3ae,0xcfefbfcafb8cd1be,0xa5af5d52eb04d13e,0xddf37de97b05f5aa,0xcfbefd72fbbddbee,0xdbffef7f8bb5dfbe,0xcd7ffe1ff1b5ff2e,0xfa56d9cbd915e76e,0xfef4bfeffb3dd77a,0xddffefe9f314ff3e,0xfbfdefbeebbdf3ee,0xcfcfff9edbacd3be,0xef5fdfdbeb3cdf6e,0xdef77fe3e394dfea,0xd1fff57a8bbddfba,0xd6b23ffefa11d72a,0xeb77dfbff3a5df6a,0x4ffeedcbf915d72e,0xb7bf7f7ffb3dfbbe,0xfcfdbff0cbbcf3fe,0xfeb6fff7ebb5e7be,0xcbd7fefefb3cdbee,0xfffb7adffbb5dfee,0xffffeffbebb4dbbe,0xded7f6fffb95f7ae,0x7bfe7ff0db34d76e,0xbb24677ed114573e,0xfacfbffbfb34dbae,0xafff4f7bfb25db3e,0xbcfdeeeddb30d33a,0x9baeff71db95d33e,0xd9be6f7ffbb5fb2e,0xcfd7ffbfdbb5f7ae,0xffffcfdefba5f3be,0xfefffeecfbb5f7fe,0xd7feff7acb95d7ae,0xdbb27f76eb35f3ee,0x8adfff3feb305baa,0xaf7fffdfd9bcf7be,0xbcdb7ee3f39ceffe,0xddcdfffff3bdd3fe,0xfe9627104310573a,0xd5cfe37bf3b5d7ae,0xebfefaeddb0c537e,0x95eb6ff3f335d7be,0xd0b7bffbfb9cd1be,0xebfefd73f9a4ff3e,0xcadfffafeb35f3be,0xef461e7adb34132e,0xaf2d7fb98889f7ee,0xed67f66ef334d32c,0x5f3eed71ebb4dbba,0xcb95e37fd925f90e,0xfefebeffeb3cdb3e,0xbfbddff3e325d7fe,0x5db446fad305c5ee,0xdfbffff1dba5d3fe,0xdbb67f7ffb2cf32e,0xab4616338934d22c,0xa3ce5f52e32dd1aa,0x3491d4eb5b0dc080,0xcd20ff278384d1f2,0xcb976d97eb35eb2e,0x5fcfff3eea34d3fe,0xeffeffdbebbdd3be,0x94bb57e9d1adf9ba,0xf16ceff2fb15dfea,0xfed6bfd7ebbddfb6,0xc310d5704b25c3aa,0x98775fdbeb04f19a,0xdfbb4f5adbbdd3be,0x97bb9f78f234d33a,0xb78c7f71cb91d77e,0x8b86ff1f7b10d7fe,0x7cd63f587115c32e,0x77bf4ff3fb34fb3e,0x857b76e0ea84d1ba,0x6338bd72ebb4f386,0xe804b2614920c30e,0x28529c59430445a2,0xeadedfbeb85d17e,0xd0114876a194d128,0x70be7d61cb345302,0xeba1f73beb15d3f6,0xd946fab5db38d3be,0xebdfdfdfcb34dbbe,0xbc7957e4f230db2e,0xdab6a731c39cd5aa,0xc7b4ff268ba4d12a,0xdcdefeaae110dfa6,0x735f5f596a1dd6a2,0xa39b17e3fb20d39e,0x82482708cb8cd1e2,0x5ebeaf6eeb104be2,0x8381e13d73104726,0xa4000140004420,0x410462ac024c1a2,0x60260a8801d900,0x4814201189205080,0x8120882801044182,0xc2ca90110984c13a,0x9aaa4799dab0c122,0x800550a0c0208128,0x834f000d1049028,0xc098d110c230510a,0xc763b2cb892473ae,0x24c000928020c108,0x50404e003006182,0x422204004000912a,0x1004018050001000,0x3820040004102,0x28281424104d008,0x402c006088001122,0x120082060004002,0x210000241008080,0x4000800060000000,0x0,0x0,0x100580080200204,0xfabe2ffefb3dffee,0x83f3e379eb9cdbaa,0xaffffffffb9df3fe,0xbfab7f33fb8de9be,0x0,0x128008008020d000,0xcbf5f77fdbb5f7be,0x66d25e14d105c1a6,0xb121d5a5200932a,0xd554ee61d134533a,0x421434110184c11a,0x810034204904c080,0xe43be7edb145b2a,0xffbfddfffbb5d73a,0x71b4f6fee314e36e,0xd36c7c215b01916a,0x8030120080201204,0xdc2ba95e321d328,0xa14f58d0f925f2a2,0x8400100200,0xc044040005080,0xf6866fedf305c1b6,0x684602687100c020,0xae76dfd6eba5d3aa,0xcd7b92fe91ddf7e,0xb966da608b28616a,0xbaa63d20d221d7b2,0xcfdafb1ef994532e,0x619ffac56995d32c,0xb6bafff2cb3ddbae,0x99fefdf7eb34f33e,0xfefe7d77fb15dbbe,0x89f3f3ffdb34536e,0x38511e94c114f30a,0xc08085a08b041062,0xd510d2e06300c182,0x3b304f604b044766,0xe32860268b01c136,0xb7db92a7f3a4d1aa,0x9b0cd732eb04e5ac,0xd5f5f6efebb9d3ba,0x8ffffff3fbb8d7be,0xf7b2af7ffba4d7ee,0xfc0b60a6905c082,0x539e5f1beb3453fe,0xf6fffffffbbdd7fe,0xfdfffff7f38df9be,0x12a404e06a044122,0x604a7e344105c10a,0x8435e5250848316,0xb9b25c43f304d162,0x6124b0625301d0e2,0x308001000000c008,0x1270228020c10a,0x188080000018204,0x45148008020410a,0x914d2e70c214d3ae,0x382455e12b084002,0x8020001080001420,0xa000000080000004,0x2f024dcb8300cf2e,0x400200000000102,0x800140c0009000,0x100802080005008,0x295b12c00105d32a,0x2501424a8020572a,0x49612672da905728,0xb01b247088246206,0xc9a5f572f9a5c182,0x8486ba28e9005086,0x3afedf9acbbdeb1e,0xfefffefbfa9df1aa,0xdffea771e93cd73a,0xfa9e3fc74b94d7ea,0x57c3ce18e914f33e,0x40001002000002,0x0,0x0,0x0,0xe9ddd33af985d1ea,0x10e000100c008,0x0,0x0,0x0 },
	{ 0x0,0x0,0x8000100500000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xff142f9409869322,0xcbbdf7ff8bd7f37a,0x77c7be79c1d7aa3e,0x1a9ceff98ad3b3ea,0xff7fba738ad7e3e8,0xc7be7df28ad6f762,0xfaa8ff6f8ad7ffee,0xaddfb7cfcbdff7e6,0x7e3f4b75ca46f3ee,0x95fdfedcca5eb7e6,0xffffffe58bdff67e,0xefffef7f8bd7f7ea,0xcf5ffebb8bdfffea,0x6b7cdf5a88cff3a2,0xd4ff7ffb8bdef7fa,0xdffeffff8bd7fffa,0xffea7fff8bcffffa,0xedefffeecbdffaee,0xefdfff5fca5ef3ee,0xfdbfffe38adff36a,0xc99ff7608ad6b2e2,0x5ab62b7e8ad6f3a6,0xffd6f7bf8ad6f322,0x4a4ffe5588d6f36a,0x95bb77f6c25ff3e6,0xfdfff7ffcbdff2e6,0xffb67f7fcbdff7ee,0xf9f7ffefcbdffff4,0xffffffffcadfffee,0xbfbdfffb8adff7fe,0xfe77fefdcbd7fbee,0xedae2fe3cbdef3e6,0xf2e4b22e8893b760,0xfbd7bffccbdff3f6,0x7fbedfe7cbd6f3e6,0xd4f7f6ebcbdfb3ae,0xbfbefff3cadeffee,0xfdb7ffe98adfffa8,0xfbd7ffbf88d7fff2,0xafbf5fd3cadef7f6,0xffffffef8bdff7fe,0xffffff7fcbdffffe,0xcea5bf67cadff26e,0xebdeff2c89dfbaca,0xafffdf1bcadbf7e6,0x35bf7ff2cbdff6e6,0xdfedffe689d6bfe2,0x9abe2f758ad7f7ea,0xcdc78f338ac2f366,0xc3dfd7d38bcef3ba,0xbcfbeff7cad6f3f6,0xd0ccbff9ca86e366,0xf7f67d7c8bd6b66a,0xe9dfff3b8bcef3fe,0xefff5fdfcb5ff6e4,0x81bacc83cad662ec,0xdf36befa8bdff352,0x5ffe77e38ad7fa6a,0xbbf5e7668bc7fb72,0xeeca939f8ad7f3fa,0xb3becf5b8bdafaec,0xfdfeffff8ad7f7ea,0xdffefff3cad7f3fe,0xfea42f6fcad7f36c,0xcf8afa1b8bd7fa4a,0x628fcf428a563ab2,0x38b552618adab2e0,0x8b729f628a46b760,0xdf1a4d6289d69324,0xeed3df0f8b8ef662,0xaf7edf5bcadffe6c,0xbeb75dda8bd6baca,0xdaf4ff7a88c6a372,0xde3e3f710a56f3ea,0x69d0eaf589ceb222,0x3dfebbdf89d7f3aa,0x17224ab70acfb332,0xffff7f6fcbdff7f6,0xefdeffffcbdfffee,0xebdedfea82dfb7f0,0x6ca29eafcb5e32a6,0x92bb050b0ad5d208,0xc43ffa6f82c7f326,0xcffe6d718ac7f372,0x8914b72f885fb3e2,0xced630138355ab18,0x3f8c6f5188d4ea22,0x1468124208c2d260,0x5ac4e76289c5a366,0xc6bb67cfca5ebf2e,0xdfeebe3b82d6bb3a,0xbbbaef5b8a55f6f8,0xa4f357ec8bc4f2c0,0xe9ffff32cadff3a6,0xcf97f7f98954d36e,0xe181cf198ac3fae0,0xebffffcbcac7f3e6,0x8c1a21724b56226e,0x5dc02e4208560622,0xde2e3f6382def750,0xc4cbc53400d6b3ba,0x4a0a200085a080,0x25f05f788955f368,0x8400000008040000,0xa50434608208b740,0x80558b000a0a9242,0x40821e1008563022,0x820205000084d002,0xe35640081cb222,0x2b080c208280b880,0x2384ca0b01d4f222,0xcefe3e078883f382,0x3a8e43088256a3a4,0x0,0x8000004000020200,0xc081205109440242,0x4781be81cac9f204,0x1a0c8c4448020604,0x2004000000c00000,0x240268000221a0,0x4340002002020,0x0,0x208000000000600,0x0,0x140724040027006,0x4effa7fe03cff262,0xed92ff3c02def362,0x7f7f3fdd8bcfb7a2,0xdb90dffb89def332,0x2009208200a000,0x9f0606608a07b3a0,0x897efbfa8bd7f7f6,0xffffffbfcbdfffee,0x9ecbcfd988cefe22,0x508580700851d300,0x140000000600,0xfa94ff4f8ad4f2e8,0xeecf3f5b8b5ff3f0,0xbfbfeff38adffba0,0xf1f3f7ee8adff766,0x9591a7518bd5b700,0x4000520408028300,0x98682018846a220,0x8608858005492aa,0x10a010480000a260,0x5102a560c050a104,0xff96296c81d3b2a2,0xcdefff388bd7fbba,0xea7e7d5b8ad7aea2,0xf5bb55a8c8ceb36e,0xd064764a8145b2e2,0xce9a252080569828,0xc0466ef488c4b796,0xbdde9ffa8b4ef262,0xf7fb6ffe8bdff7ea,0xd5ffff6d8bd7ff6a,0xfefe7dd68adff7e6,0xc942ef3eca1e7a06,0x468202108194b300,0xc4824ff589d5a202,0xd56c82b88857f262,0x210000000a0a0,0x4302c10800868222,0xa573bb78cacef776,0x6dbdc63c2ddd30c,0xfdffffffcbdffbfe,0xdfbfff718ad3ffe8,0xfbbeef7e8adffbe2,0xdfd2fe93cadff3ec,0xffffffdfcbdff7ee,0xfeffffffcbdff7fe,0xffffffffcbdfffee,0xfffe7f77c2dff7ee,0xe5e6e302cace933c,0xed5d5fdacac47e9c,0x43250a002407040,0x71c1020080587002,0x4002304001848260,0x800853398042e062,0xc00000018000,0x200461200448202,0x76fcf6100a46a62a,0x8f6a77e38bdef2ea,0x44000000005000,0x0,0x93aa8de28a8cb206,0x9000164082d22020,0x6c0000720a08d000,0x0,0x0,0xb084f5b82ccb22a,0x422200000121200,0xd9186d40885e32c2,0x20660228082d200,0xcf53be3d8adff31a,0x9d0ccd5b885db7c2,0xf4f6ffea8bd7f6aa,0xba4ee77a8bd6f228,0xfef22f6e8bc4fb66,0x580002058142c222,0x2fddde538a4efa72,0x0,0x0,0x0,0xebc7bebb8bd6f366,0x43573e21cade7a16,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfeb63d73eddffeb6,0xe8f5fa6f8dff5b90,0x7b77dffd8ff3be92,0x8f9ecdfecdf3fabf,0x9cb2ef7eecd33b33,0x4f7eff63ece733b2,0xa3c4ff25ecd632a2,0x7fdfbb3bcfcfaeb3,0xafbecb7bedcb7bad,0xf5fdffffecf77fbb,0xff9eff72ede23fb3,0xdf927f79cfe73faa,0xcffbdebfcdd7fa8a,0xfb9aaddc4dd1ffb1,0xf6ffffff4df3ffb1,0xffffff77efffffbf,0xfbfebff6cfdbff92,0xcde7ffb7cceb6fb8,0x6edfff5beffbffbb,0xbdff5fffcdf37eba,0x456c777a69d01ea0,0x7abebffb4dd7bfb2,0xc992dfb34da75f9a,0x4bdcdb4bedc13bb8,0x97bbcffbcddb3fa9,0xf7eff77fcfd33fbe,0xfffefdffedffffbb,0xebfffffbefffffbf,0xffffffdfefffffbb,0xffbdcffecdf3ffb8,0xf7f7feffcfd3ffbe,0xfffe3f78cd5b3eaf,0xf9fdebffcffb7fb2,0xe972febf4dc7ae97,0x9fccdff9cfef2fbf,0xf4fff6eecdc3fe99,0xdfbf7fe1ecf3bfbf,0xffaff7feefeabfbb,0xcffffe3fedcbffbf,0xbbffdfdfcdebffb9,0xfeffffffefffffbf,0xffffff7aeecbffbd,0xfbbefffeedfbbbb2,0xfdd4dfbeccdbffb3,0xaffcffd84dd33bb9,0xb7ffb7f3ece38fbd,0xfcf0fffbe5c33ea0,0xcc6e966245d3fa32,0xdbfffefeedf77fbf,0xaffff75bcdc37ba2,0xb7bb7feacfd7ffb8,0xd3ddbfe8ccdb7bb8,0xfbba7f75cddb7f38,0xebf5f3ff4cc73f93,0xffcbdfd36dd75f9a,0xf6bedffbedf3fe3f,0xf5ff767fced3bfbc,0x7dbe7d72edf37fb7,0xfbd4ef3f6dc6ffb2,0x4bfcffdf4dcbbbbe,0x4ebfffffefdfbe94,0x54f572eb6cd33fbb,0x9bbfdf72cfd75bb8,0xd99dfffde8d7beba,0x8f69fa08ccd73e90,0xbf6567c0c9533fa0,0x2c1916e90487aa80,0x982aa7620cd37b12,0xf7bf6fd5cdd29a37,0xedcdfc0e8cc6fa98,0x2859dad045ca4a31,0xce35de54d422a92,0xc52caaf2eec22a18,0xdaa66f7fccc2be31,0xc9d7f73aedd33f20,0x4e5afaf9cde33a31,0x7ea2dd100c622b90,0xb1fe277aefd2ffbd,0xebeeffffcddbffbc,0x8b7ffbfacd86fab8,0x6f5f2e0ce7dbbe38,0x3ed06942aec32ea8,0x1e5b3eeecccbef3b,0x6a5c257948033eb6,0x38e900628cc35810,0x2a469feb6dd17e02,0x37b9df6c6d933a8d,0xf5b47eeeadc22e99,0x933aee71acc35a91,0xd3b4f7dfcde72f99,0x4febbebaccc3bf96,0xba6d9d5dccc37b9d,0xe2315e63c4d3fa24,0xdceeefe0edc33e90,0x5b94678748e35e90,0xcff7ff6f4dcbbea2,0x6ffcdf584cc32f38,0xf7fafff3cec32eb6,0x1334af48ccc222b4,0x58900a4a4cc25a28,0x900f202c4c20204,0x25ca5fd88cc11ab0,0x13204dc3ccc36e14,0x2121066888c21010,0x508605c4858a0808,0x304a72b8580b800,0x340074101030a0a,0x82180b428e800302,0xc56b28048030688,0x52221010a00,0x314441004cc20018,0xa513001cd921a80,0x1174a4045800e20,0x143156efccc223b8,0x63055c14c842b00,0x0,0x302a60244c11630,0x1be8ec5b4dc23e34,0x1606248008031a12,0x2e05400c022a81,0x4000040004002000,0x60080800000,0x0,0x0,0x504522045c30830,0xffb6bf73edd27ab4,0x81d7f31eecd3fe00,0x67dfde464dd29a9d,0x1ba007a0cc869310,0x288040004022008,0x120620600c0c0688,0x8093f9e36fcb1a2c,0x3cdf7f3ecfe3baa5,0x20424041000804,0xf5c416634de2bba1,0x72362d60c8531a94,0xc972e72e2c873b93,0x80e79e114c92ba97,0x9bbfcf7bcff2bd3c,0xfcffffedcddfffbf,0x8a3c3d632ca3ff12,0x4004028208010281,0x49430e10c8832b20,0x818a0a1200850e00,0x2a068408430288,0x100118000004a01,0x60902f6101c75a10,0xf3eff732efdbfebb,0xebfbbbdb4fdbfe31,0x40960fe345c71a0a,0x582257c2cc826a22,0xd2a827464dda3a9c,0xc0848c3549422e83,0x3899dfc4c4d2fe18,0x975d7f73c58bbe38,0xd75476ddcd9abf01,0xffbe7fffefdbbbb9,0xfbd7fb3ecce71fbf,0x8c916e218cc2bb0a,0x120488120422200,0x9022664c44c76e10,0x51843dd0c0809606,0x981eb2e64833a32,0xc4a224a85422a00,0x569b5ffbeec6feb9,0xf5feffffefdbffbe,0xfffdff73cdebffbb,0xffbfffffcffbfebf,0xcbfffff9cdf7fe9a,0xbf1edfdbcfef7f3e,0xf4fffeffeffbffbf,0xffffffffefffffbf,0xffffffffefffffbb,0xfdf7ffefecc3ffbb,0x86013514c820281,0x4dc502844003a23,0x1820c242e0420a06,0x2022c4340c21e0a,0xc970a7004ce32800,0x204000000020008,0xa2410840832202,0x500096084cc28800,0x14a332048222114,0x108001000800,0x2206000008021281,0x5fbdcfbb6eeb6ba2,0x1001042004800202,0x8000200000800200,0x202222400400a80,0x10100006000000,0x2028034241805800,0x41482c400424a1a,0x4182060048a00600,0xd0100e0068808304,0x8162ba014ccae89c,0xa91a031bcd8a7f10,0xd4f57ef8cceba2b1,0xdcd32f71c9c33f93,0x6cfe3f6449d71f91,0xc9eddb9fced71fb3,0xcfffdfdbcdfbffbf,0xa9aa8ffb8ddf3b9a,0x0,0x0,0x0,0xbf571bdac5c3de1c,0x0,0x108662081435a18,0xc804200004022600 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc1a8c6202378944e,0xfefe7ffffe7eb3eb,0xc9d7ff7fa07b93cb,0x7d7b9bfdfe5ab7c7,0xbffccde8ef7a9cc7,0xd716767f777f9bcf,0x7efe3d7b6f739bdb,0xb3f6ef7f7f7b17df,0xea5abfbb7a7a93cb,0xfffecf7bea7f99ef,0xf5fff7fb6f7b3fff,0xffdfeff27f7ebbdf,0xdbbee77eef7e95cf,0xceee7eaaaa5e96cf,0xfbfedfdfff7f97ff,0xfebffefaef7fbffb,0xffffff7ffe7ebfff,0xfbbe7f7feb7abfff,0x8defeb3bef4a1adb,0xea7ffed3fb5a9b4b,0xbdf397e7267f355b,0xfbfeffede25eb5cb,0xdebe3fff6f6fbddf,0xed4ff672f756918b,0xe3fffe6b7f72b2ef,0x9fbe4fffee7237fb,0xf3ffffffbe7eb5fb,0xffbe3f7f6f7a9ecf,0xfbffffffff7ebfdf,0xefff3fdfef7ebfef,0xb79fdfffee7e16df,0xfdbef67feb7e977f,0xebbefffbf77eb7ff,0xebfcff2eef7b97df,0xcdeaff5beb7797cb,0xefbfefdfe67fbbff,0xfbd5deeff77595cb,0xdff6fd727f52b4cf,0xffbff77fff7b17cb,0x7fefff3bff7b33df,0xfffeffdfff7ebbff,0xfeffffffff7fbfff,0x97febf7beb5e1fcf,0xff9e67ffea7e16df,0xcd5febdff7721d4e,0xeb7fdf5fee7e9ecb,0xffff77eebb7a314f,0xfbefefda767713df,0xd3be7ef3636e19db,0xcdffffffff7e9eef,0x7bfe9fdd6f7299db,0x743e77fbe263b0cf,0xd27fd760a6763c4f,0xff8efffdbb7f15cf,0xd9f7f77fe97a12ef,0x2efebf5b2e7a9859,0xaf4cef7a28563c5b,0xf7fdf7ffea72bedf,0xfffebffbea7bb7fb,0xf9fcef7efa7395cf,0xf9cfff792b7a93cf,0xfffccffefa7337db,0x55c5d6eb2b629cdf,0x9aaeff70fb72974b,0xebd4a7b43f5a13cf,0xceebfe07e2773b3b,0x2b7fd4de307e10de,0x14b4fc4b2b721169,0xbb8eff2bee4b11cb,0xbf9a2f726e5296df,0xe55d6f1b227233ce,0xfa911b531264108f,0xb6ff766fe776b5df,0x954b37d0f25e10cb,0xdfb12c7d2a6e9ceb,0xadcef63ea77234cb,0xaddefade6e7a1cef,0x579b4fffa37a94cf,0xc935ae5263661509,0xde8e7fd23a4394df,0xa9f7f77faa7a10ef,0xea569fd92e6791c9,0xfd564ce76d661a19,0x53fbfafe637a90cb,0xf328f4613777934f,0xa84ceb3a3b511602,0xc813b297266793af,0xa7dedffbef6217cf,0xfcf732f96a7a18eb,0x6bfe6f61ad7ab4cb,0xff957fffe75714fb,0xefd7ffd36b7b1ccb,0xa25887d12a52910a,0x1496f6ed3e5c94cf,0xc16e7fa9785e92cf,0xf7b82a4561629880,0x1582ea2b65563bc7,0x67ffdfdb6777bedf,0xbe7bffefea7e19ff,0x896db7f066769049,0xc6a0250060611049,0xc8c391002340101e,0x1c0444220210ca,0x8044002008001,0x4404959021403380,0x3000000000021000,0x89f0bb7b217615ca,0x40020000021000,0x1126ea200621097,0xd500280002421143,0xc10024020a021007,0x5200e402a3400023,0x885396002e40160b,0xc4441284000a9,0x9470365a6760105a,0x9a170d72af5e10d3,0x4002000000400040,0x0,0x888007480052100e,0xfcfffeffff7fb7ff,0xd96f277beb7a177f,0x0,0x2857c2026f4015c0,0x0,0x800010000001800,0x502020000461001,0xdfb425442b5a91cf,0xa78c7ca577531b21,0x29de7eae626a1b4f,0x13b06e19605a106b,0xc088250066020003,0x10120c00080a1041,0xa9c49022225214c8,0x6821201a0701082,0x330cc7102e57108f,0x954c8360e342140b,0x46ec204a205210cf,0x8174c7223264148f,0xce53bab0346693b9,0x7bbccffb6e7a97eb,0x34beceecb27ab9eb,0xb014af71ea4291cb,0x80130c246001c2,0x84c9388001731120,0x8810102402001,0x5cb4485000421087,0xd22975023643141,0xa8c32602c421449,0xdec3ffbcac6535cf,0xaf56cfd9257a15de,0x70ce561a06213c0,0xf19707182c709a81,0xc2b41382aa441243,0x8000f10000401091,0x69ffd9ce777e12cb,0x15fecff32a763cef,0x9f7c1d0aee7638cb,0xfffe7dfeff721deb,0x49eaff6baa7634cf,0x49c35a19e87600cb,0x5714d0261029001,0xb2bffe792647158f,0x418c3b2020461458,0x8d000220a000049,0xdd435e1ef8721183,0xdffecff3fe7717cf,0xf5fffeecbe7abfcf,0xfbdeff73ee77b3ef,0xfbf6776f6a5e15cb,0xcfeffbbfbf7b38fb,0xaa861fdaeb5e1ccb,0xfcb6fefee77a97df,0xffffffffff7fbfff,0xffbfffffff7f9fff,0x4fdaefaa6e523acf,0x294012810a021111,0x24215c612142350a,0x4041865861021081,0x14003020029000,0x42000400001,0x10124863000001,0x82010220400081,0xc080501032001001,0x5082044010603210,0x300a00202001540,0x2b4a594b226210cb,0x2b2ac91823721401,0x420002822001001,0x0,0x2100c0000040008a,0x0,0x8c0c5418a0420288,0x7484960822429b41,0x524b5202c461809,0xe901e40208521249,0x9fee9b136253908b,0xb655ddb227b135b,0x4480f4c02670121b,0x97a6e75ca65a90c3,0xffbe3ff46f7733cf,0x8ecffe227f5e11eb,0x77595fdefe57156f,0xfff37ffbfb7ebfff,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc5cded29cb3dade4,0xf7b67f7b8b3faff5,0xcbdbfb924b3cafd4,0x5cdb37390b3c8bfe,0xbebf2ceb097d2fd4,0xdefef67d483e97f5,0xcffc7ff1483cedf7,0xfbd5f77fca3effd4,0xb4ebbff90b3daed5,0x87bedffbeb7fefd6,0x74fffef8e97e9fe6,0xdfbeff72e97ffff9,0xfbb3afef687cbff7,0xcddefecbeb7fbbf5,0x7fffdfdfeb7fafff,0xfcf6f7eee83e9dfd,0xf9ff7f678b3fedf6,0xffbbfffd6a3dffce,0xddcbf33a083ebefd,0xeb76fe5fc93eaedd,0xb8be57e28b3cd9f6,0xb9ce777b6b3caed4,0xdabe2f5b0b7fbed6,0xcff3ff3ec97dbdd6,0xef9fefffcb7ffdd6,0xb7bb5ffe487fafff,0xffffffffca7fffff,0xff9effffc87effff,0xfbffff7feb7fbfff,0xffffffffeb7ffffd,0xff9ffffbeb7ebfff,0xf67fff7bc93cfff7,0xfffeff7be97fbff7,0xdbd6f9476b3ebacf,0x1edf9bcb483eafee,0xb7bfcffbc87eabd7,0xfff5ffffcb7fffff,0xd3feff70eb7e9efd,0xfbb7ff7feb3ffffe,0xdffbffbbca3daff6,0xfbffdfdfeb3efffd,0xfcffffe9eb7cdeff,0xffff7f7fe93fbff7,0xfbfe7f7e6b7ffcf7,0xcfefff3f6b3feff7,0x5ffe5f5d687dbf53,0xb4f7dfebca7cbef5,0xf57db77ae93cbbd7,0xdbf63df6683c9dce,0xc7e7f7fe4a7eabf5,0xebdedfffea3ffffe,0xbffbdf796b7fffd7,0xd3fdbef1c93e9e5c,0xeed67578ca7e8af1,0xc9fefb7b483e8af5,0x6b573357ca3dadd6,0x179fcfda0a3ca5d5,0xfdf5f7efca7cbeff,0xff9e3f6be97dbdf6,0xfbd6df7eea3caeda,0xf3f3fffa8977bdd5,0x1efdcf8bea3c6e7c,0xfdfef7e9493ebdf6,0xbdbefd636a3ea7d4,0xd716fbfc8b3e9ef1,0xcff7feb3487eafd7,0x6216cfdf683ea25a,0xb6b376efc87deff4,0x6deeff6aca3dfff7,0xd6ae7d364a1c11fc,0xd67fd91a887d8cf5,0x204c48c8882088d2,0xdcbfff7fea3fffde,0xf7fca7758a3e8ff4,0xd6b67c7f4a1ea6d5,0xc7c7710c083ca0f6,0x7b7f9bd6883efafd,0xa61bfefbc97c2df0,0xcd753e5a0b7d8ee4,0xe5be7dff4b3e9ef7,0x8bf0e36b0a356ec4,0x1dd31f82ca3e8c85,0x86d06ec2430c88c6,0x9040974b081402e0,0xfdfe7fe2c93dfdd5,0xfb85f312487c82d8,0xeffe3dcbe83eaaf3,0x2bbcefdac83dbedc,0xd0e2a37a083eaad0,0xff7f3f72893cacfb,0xfdbe7fffeb3cbf57,0xffdff267c83daef4,0x6fffdfdb483cbff7,0x6201cee80b3d2a64,0xf97adf72693ceae6,0xffbe3f77eb7ecff6,0xffffffd2c97dbbf7,0x6fffffcfe93faff6,0xdfe737aeb35a9e4,0x6104304808103080,0x8b021c00082ca242,0x40324000c84a0,0x39cca0208341250,0x1206e000088091,0x6d7238011ca8c0,0x52510008308401,0x8b9f09176035b7d4,0xc00200400208050,0x9105c00808a082,0xb33d86800834b20b,0x452c0450080ca066,0x980e302033c0c80,0x827a1a1801361a80,0x22200012001080a2,0x5570d6e3083d82f2,0x879aa910027e9730,0x8020224008242080,0x100240080060800,0x845008200840,0xfcf7feffe23f7ff4,0x110ba76a00361654,0x1004000000000050,0xe9f5d00a60363744,0x610000000162000,0x0,0x29642a600016ce45,0xfebf3ff4697eafc5,0x8fcaf7b6483e98d7,0xa6141c01090da846,0x32b34f0200369dc6,0x10a4080040048010,0x804002200200940,0xe3d6fa32e834e856,0x2e83be10e0368986,0x103455000834a064,0x84405220401c0620,0x442408c8002c0b40,0x9af49f2e683e59d4,0xc8963ea2493edaf4,0x8ebecdb30a3cccd7,0x55dc5e7b083c96fb,0x812abf60603f1060,0x4800d520020e0404,0x22440012001080a2,0x10050208100c40,0x30824248342a00,0x4220e522030f0c44,0xcb922e69e07c1c10,0x8b43fb1a493fc8de,0xa5cf858803d8ec4,0x4995ea1083cb0d4,0xa1002a0008340084,0x4081204049140844,0x8683890a033c9856,0xcbce9ddbc87dbed4,0xc233479a022f80c6,0xf5f5b7e8c93c0e74,0xfffe7d730a3fdbed,0xf1f5fb8e493ebcd3,0xcdce1299683daed4,0x82c4720825e844,0x4247402081080d2,0x8865541084e8109,0x3900e7254834a944,0x24868609023c00f4,0xc6fecf738a7caff6,0xddf5ffe2c83f9ffe,0xe7b6bf63087cbcff,0xef95ef6ccb34bcf4,0xcb57ae34e83f7d5c,0x891fdd53803edbf5,0xfefffefb6b3e9ff5,0xfffffffbeb3ffffe,0xffbe7fffeb7fffff,0xefcfff3eeb7cfffc,0x6000140000200050,0xba124f408038da65,0x1180816080308890,0x2830040000048c41,0xa010720402140140,0x1009a80002d0040,0x42250c40800ca51,0x5120002400103004,0xcd563d70483cac79,0xe9a4b112893889d0,0x2f7f1fdbcb36dee4,0xa00c34a0234aa41,0x1080807388300ac0,0x202c0000088012,0x68c4c204012c08d4,0xc48a101009088a52,0x20040008000040,0x50a5b06008388270,0x92b00d10012e8956,0xd191633b083ea6c5,0x2e5ffe096a3c0bdc,0xefffdfdbcb3da8fb,0x74f7e6e40876b9c5,0x651e3760083f83c1,0xfb9e7e7f4b3c9efe,0xf7ceef38c93deefb,0x9affdfdbeb3fbfff,0xffffdfffeb7fffff,0xfdffbfffeb3effff,0x0,0xc000200068200000,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400000020040,0xfdfcffff5f9acfdb,0xfff6efefbf9bcf5f,0xa9deff1e618acf57,0xafff7f595f9acf5f,0xeebfefff979bcf5f,0xf7f5fffd7f9bdf5b,0xfffebf73eb9bdf1b,0xfbd6fb7e1b1bcfdb,0x9bd7feae3f1adf9f,0x4fbccfeb4f1bdfdb,0x7dfff6ef7f9bcfdb,0x9f7eff716f9add9b,0xfebffb7f7b9acf53,0xcf56fe9f7b9a9f5b,0xaf1fffde7e0bdf5f,0x7cfef6efdf9bcf5b,0xd9effffbff1bcd5b,0xffbe6ff65b9adfd7,0xe3d7fbb73b1add93,0xe7c7f5adb0a9fdb,0xfefe4febff12df53,0xf73eaf62d79b5fdf,0xf89edffeb79adf5b,0xe97ffbb63f8acf57,0x7fffdfddde8bdf5f,0xdeffffff3f9acfdb,0xf7ffffffff9bdf5b,0xfffeffffff9bdfdf,0xebffffffdf9bdfdf,0xffdf7f7fff9bdfdf,0xffbffffb5f9adfdf,0xfffff6fd7f9acfdb,0xe7fefffb7f1bdfdf,0xfbfcc716d39a47df,0xc3dbbb5b5e83555a,0xff7fcfff7f9bcfdf,0xbfffffefff9bdf5b,0xaff77f60ff1a4f5b,0xfbf5efffff8bdddb,0xcffffe3b3b9adfdf,0xabffdfdffb9adfdb,0xb7fe5eeb939adf5b,0xdfffff734f9adf9b,0xfebf7fffdf9bcfdf,0xffefffbe7f9bd7db,0xebdfdfde7b93dfdf,0xbcfff7ef5f8ad55b,0x1bcca64b1318cf1a,0xfeb5f7e1959a5f57,0xc59ae139a313cd5f,0xd9df7fdf3f9ad7df,0xb5fb5ff3bb9bcf57,0xf1ad777fd79a4d5f,0xcfbe6b761319cf19,0xabdedfaa8b035f5b,0x2d7ffedc1f8a4fdb,0x8cdf07c7779a4f42,0xddffff5a5f1acf41,0x7ffefd73df9bcfdf,0xdbf0ffb2338b5fdb,0xea5bbbe7d31bdf5f,0x779fcfde3f9b4fdb,0xfcfffef83f0acf5b,0xdbbaff72971a5f1b,0xd482ed3ea38a8f1b,0xcfd6fe5d7f8bcf5b,0x83cfdf550a1a8f19,0xd0b6a4fc13185d1a,0x55fe6f623719df5b,0x7a9cffc43218cf11,0xe192ea235b8bcd0b,0x66dadc433d9b5d5a,0xb0b6dce0378bd5db,0xf27cef16039b5f1b,0xfef63f531f09cb13,0xafcffe9c3783df5b,0xeb7f3ff70f9bdfdb,0x85ba7fb95f1b4f4b,0xf7ffb568a71bdf5f,0xfabafd75379bdf5b,0x83b2fba7738bd75b,0x4fd55ff79b93cf5b,0x4bbec7776b9ad753,0xb487faea5f885753,0xabb4ad623218df1b,0xd95ce6281b00df10,0xffffffffbf9bdf53,0xee95cf791b0bdf53,0x37955f62f203cf49,0x9a9ebd737b0bc51e,0xdb93f77b1b9bcf93,0xcfeeb62d638adf5b,0xfbf7ffdf769acd53,0x74fdfeefd79b4f5b,0xe3ec37600782c7db,0x81122c351e084301,0x8ddeeaa9219d71b,0x6a8dded1178acd5b,0xd0127fe2138bcf52,0x5184c972121a8c1a,0xda0a2400320a83d1,0x4582f0e23409cf5b,0x8472340268ac312,0x3b30017003088719,0x10300a0405124c00,0x2a2240001820283,0x8920c72803124d12,0x21400a01000a8403,0x410800202004c00,0xb100929882834508,0x480024000300440a,0x101820004004440,0x0,0x410000000020040,0x400134800080302,0x42a0006012020340,0xa8048e0c12014510,0x8102f80a0700cd50,0xa448c41a06004710,0x8011c02202114c10,0x9110844030000100,0x800002000004800,0x0,0x140000002000008,0x84e9d460001a0c4e,0x21000a2041020102,0xffb06d773f9acf5a,0xf4f7f6201a1a075b,0xaffee4fc20b4f5b,0x310465a5a0a0d41,0x820c850004005750,0x4e462062110a474a,0x4097d30813924610,0xaa518eda4703571b,0x49bccc9264835912,0xd4e5362807095a1b,0x41002c0012001010,0xc9d8ff7a9182df5f,0xca66b67bfb888f43,0x17fccfbbdf8a5f5b,0x70ce7c6443824b1a,0x4a204540120a0617,0xe380c32c0a13cd5b,0x100822012001010,0xa854410e0108441a,0x422120800021910,0x979a0960070bd2d3,0x820c52e25382c145,0xcfcbffba539b5fdf,0xab5a5ed3008bc75f,0x30220442000a0c00,0x99430b60001a0cc3,0x5832066b500a175b,0xa117d621121a0749,0x29539a521200961b,0x14fb7d53491adf57,0xd5edb76a3f9ad7db,0xcfd6e47fb38bcf5f,0x89e6ff3b5e934f9f,0x6ff77e71771acf5b,0x308ac44222115718,0xd110823042088181,0xc49421f1810a4d52,0xb990c20c248a4751,0xc8105923104f03,0xaeffdffbff1bdfdf,0xf6f7fefc7b18cf1b,0xbf7efd71d79bdf5b,0xf79fffff731bdf5b,0x4fdff3037c1acfd3,0x2fd35ed1539a4d1a,0xfc76bee7b79bcdd3,0xdbfeeff2b71b5f5a,0xffbf7fff7f9bdf5f,0xffeeff2e5318df53,0x84490004004c41,0x88100081004400,0x6064b49012108710,0x800040002004018,0xe004300001804102,0x281d504802928650,0x40a66200001441a,0x4004040202004810,0x9f102fecbe984d99,0xc9c6f36116805d58,0x4283926b1782c75a,0x4a10c1480301c518,0x24000080100,0x0,0x4200200410020104,0x42c332414182070b,0x2abcdc5e73100783,0x1102060010a0c43,0xcd386c510b12475a,0xf7b6ffff7f9adf5f,0xeffffecfd71acddf,0xef7fdfdb379bdfdb,0xf4bffdef5b9bcf5f,0xd96fe77e3f9adf53,0xffbedf77f30bdf5f,0xc3dfef3b378bdf5b,0xeff7dedfff1bcf1e,0xfeffffefff9bdfdf,0xffffffffff9bdfdf,0x22810514408a0146,0xcb34fff4599a0f5e,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100000800004040,0x0,0x0,0x0,0xb5ffffecfdf5ffe6,0xf7fffffcfcbdefe6,0xf5befffbb4b97fe6,0xcbf7ef5f749c6fe2,0xfcd7bf53fcbdefe6,0xb29eddde74996ae2,0xd7dd76ebb4b86fc6,0xffbe6ffafcb9f7e6,0xcbd4ef3ff4b9ffe6,0xafdbbbdffdb96ee6,0xbbbfcfdfbcd9dfe0,0xf5feffe7ddfdffe6,0x9fbebff2f9fceee6,0xfb9dee3f7db97fe2,0xdfeffebffcbd7fe2,0x3fdbdfdbbdb97fe6,0xd6fffeeff5fd7fe6,0x9dbfb77b1cb5eec2,0xfbb23ffff5b87fe6,0x8c4fff187da81de2,0xefff1b093c205fe6,0xf57f5ef3f498e5e6,0xd8ec279a101926e6,0xcab62d7694f97fe2,0xedffff7af4dcfde6,0xcfffff4bf5f97fe6,0xdfbecffffdf967e6,0xf7fffffdfcfdffe6,0xfffe7ffffdfdffe6,0xfbfffffffdfdffe6,0xeffffffffdfdffe6,0xfffffffb7df97fe6,0xfff5fefffdb97de6,0xdffe7feaddfcffe6,0xebf3fffe9dfcffe6,0xee6fb769fdac7fe2,0xffbefffffdf97fe2,0xfdf6fffefdfdffe6,0xdbfeff73fcfd6fe6,0xfbb6ff7f5cb96fe6,0xcd5fbf8bfdf8efe6,0xebffffdbdcf9ffe2,0xfeb7f6fbf4bdffe6,0xdbeff7fb7dbdfee0,0xffbf7f7f7df9ffe6,0xef7fffbffdbde7e6,0x6b7edfdfb4f95ee2,0xb5f375f335b9efe6,0xfd6df75cedf94ee6,0xfeb63f5775b9efe2,0xdfd3e57ebcf97fe2,0x265fde6fd5b92d62,0x8efb4fe9d8b967e0,0x97ddff6334f8eee6,0xeffe7c6534fc7fe2,0xe9f4ffef3dfd77e6,0xaed61edf15b95de6,0x9ebb7feb3cf9ffe6,0xfcf6feee34dc6fe2,0xdfbeb7f0f4f9efe6,0xab95c77e7dbd6fe6,0xef733ebbddf1efe6,0xffbf5ffbf5b8ffe2,0xf5ffd7eb7cb9ede6,0xaffe7ff2fcb97fe2,0xff9e3fbfdcfc7d62,0x4fdbbe137dbdc9e2,0xabcfd5dc50d8e422,0x46f0f47898386dc6,0xe943ea6235e8dce2,0xfbb0ee7735b95ee0,0xdedaa2b83db9e6e6,0x2dbeff887cf8cee2,0x259455b23490a4e4,0xd97def7e35bd7fe6,0xde97a7dcc4b96fe0,0xafd7d3fdfdb97ee6,0xfdc1f5a14b0dee2,0x953855fa2cb46dc2,0xf7fdf6ae7c3957e6,0x25967ff950b9e3e6,0xcbc6f336d49c4de2,0xbf21ed75cb8dca4,0xbcbb6dcc14f8eae4,0xfcffd6fa75f9cfe6,0xd2682dc28c3957a0,0x20a4820400952000,0xfdde7f6b79b94ee6,0x8312dfd314f877c6,0x315d9ec8841865e4,0x9beeb772d1a8eee2,0xfcb4ed6675bc41e2,0xcbdfffd355f9fee2,0xaf5fdddeddb8ffe2,0x34fadeea74d97fe2,0x96e4454150106560,0xd6b53f4e7dfcdfe2,0x4ec6dbaa9c99f686,0x2f7e9f5fdcfccfa0,0xdb5317f470b167e2,0x5024c24505b90020,0xa0274018004ca4,0xebdff78814ecffe2,0x8000e7041204a22,0xa222000801280480,0xc07034829db04d40,0xcf34377415b86560,0x89708340083867c2,0xa008000300206842,0x20204c0042012c0,0xc44090000400,0x6a261460340063e2,0x0,0xa448261b34912ae0,0x8012044004004780,0x123226088cc480,0x8018002000180022,0xd010312a04983060,0x800100000000480,0x5500050010020,0x20004000002040,0x20044400c280000,0x4202100112000,0x8012600290080420,0xc80108004480,0x20040040000020,0x696cff6234d96ee0,0xffb62ff7fcbdffe6,0xc8d4f32f94bd5fe0,0x23549ad0b4b416e2,0xb42a4f6914996fe2,0xb8112618449408e2,0xda142130049c6040,0x8bd8df382491dfe2,0x8480000030808000,0x8928902140014e2,0xa414022114205c40,0x408029000080e840,0x9020208000045a0,0x8000621284004540,0xfbbedf3afdb97fe2,0x300c0c4a30006c40,0x4812005000314042,0xe088d30c14907cc0,0x144c00044001120,0x806010000114040,0xe4d47cb4341032e0,0x2260500300000c6,0xa3b43765d0b86ee2,0x8debffbafdb16de2,0xffadfc9a0b96ee6,0x16d2572014b14840,0x630c812014a0cc40,0x92902b50309940c4,0xc9c8652228b962c0,0x63d696b354d07ce2,0x96bbefffbdb97fe6,0xb53577e29da84ee6,0x7fe67dedbcb9ffe6,0xe9d7f7ab55bd6fe0,0xe3f607d730dd65e2,0x1b316f111c084ce4,0xd3edb772b1316dc2,0xf39e255270b8eae0,0xe9ccf71690bc6fe4,0xcfcb9f1399b96ee2,0xfb9ccfb1f098e7e6,0x97b7f6ed54b9efe6,0x922b3de3b0396fe2,0xdabef7695cf16de6,0xdfdfbe9314b1ffe6,0x3c7585830b966c0,0xc01280e44c906060,0xf96dff7b7c98ffe4,0x4b982f26d5f825a2,0x4908d02214a85c40,0x1c21810200061e0,0xb0542a300048e6,0xa0042810008820c4,0xd5a4275410914cc2,0x582400114885420,0xe2d61ac894b9d562,0x10000004800,0x12a044010908002,0x280000004400,0x48402002a0004440,0x408a1209840045a4,0x480001200000,0x804000004400,0x100020000000480,0x8000912c10a80862,0x8480080140106420,0x8000040000084000,0x42204a4d00044a0,0x162204084880540,0xcb9a7fbf953167a2,0xffdfffbffcbd6fe6,0xdfdeff5bf5bdf7e2,0xd6fdefe6fcf9ffe6,0xbfefaf7ffdfdffe2,0xcb36377b35f96fe4,0xccfdf3814b9efe2,0xeffeff5addf9dfe2,0xfffffffffdfdffe6,0xfffffffffdfdffe6,0xffbf3fff3dfdffe6,0xfbfffffffdfdefe6,0xa10010000,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x9965a571844412cc,0x0,0x0,0xabffffd77efdf6ea,0xdffbdff7fefef6ee,0xd7ffff7fb6aed7ee,0xffdefff7f6f7f6ee,0xfbfeff2ffcf6f7ee,0xefff3adbb4f6f6ee,0xfffffffbbefef6ee,0x94f3ffdd24be56ea,0xdfaefd62b61677e8,0xfbfce77ef6f6f6ea,0xdfffbbfff6f754ee,0x7fb7effbf637f7ee,0x7495fefbce87f76e,0x9ffafff36eff77ea,0xffbeef3feaf7f4ea,0xcfffff7b3ef677ee,0xffffffdffefef6ee,0xf6ff7ffefcf6f3ee,0xd9eeff63dcf6f7ee,0xfbb95fbef49654ee,0xeffeef1fee76f5ea,0x2b5e17cd6c5e56aa,0xb5b657eaa07452ea,0xcdecb670f6be7468,0xcc86376b8cd614c6,0xfffffffee2f7f6ee,0xefffff7df45ef7ee,0xbffb7ffbf6fff6ee,0xf7ffffffe6fff4ee,0xfffe7ffffefff7ee,0xfbfffffffefff7ee,0xfffffffffefff7ee,0xffbfffffeefe77ee,0xf1f7f67ffef6f6ec,0x7dfc7ff8b65e37ee,0xcb86e7ef3656f66c,0xeae2ae75441e366e,0x3ebfdffcfc7777ee,0xf1fef7fffef6f7ee,0xfffe7c63e6def66e,0xf9bdff362637f6ee,0xdecfff1bb4fe566e,0xfa7fdfdbf6f4d7ea,0xf7fe7efaf6f676ec,0xffffff7bb47ef7ee,0xebff7ffff6f7f6ec,0xcfdfffbff67e77ee,0xfbf3dfcffc16f5ea,0xffbf6fffbe5f76ee,0xd76fff768cf7d6ec,0xfebe3df8ee5651ee,0xede3e7ececfe76ea,0xef7ffedb7e3e76ee,0xb73f7cf2acf674ea,0xf7edd67c747fd66e,0xf5de7ff7b66f53ee,0xe904ffabf476f6ee,0xecffbedd661ef3ea,0x5ef8cfcaae97d2ee,0xf8f3bcfce6b675ea,0x7caefd7120f656ee,0xdbf2eb7ef67776ee,0xf7fa3edbfc7e12ee,0xafbfdfffc43ff46e,0xd4d5dfeae4b7f6ee,0xefbefd62fefff66e,0xf180e77c205671ee,0x4bdece3f969f56ec,0x2bbecf4bb63674ea,0x14e76660a4b4746e,0xa49ea070221542e8,0xff144d74545630ea,0x4ece76ba289e37ec,0xa1d08e4216b25566,0x1cc47502a014d02a,0xf9f7b770c656546e,0xfaad25fdde16776c,0xf5c3feb682b2d6ea,0xffdfdedae656f7ec,0x1eb1cf4a141cd4ea,0xd5f7bff9a6f6f4ee,0x7b9c7d7cb426d2ec,0x81c6f63cb49610ec,0x36837aa222a0d668,0xacf7a7adc6327664,0xf477ff7db676566e,0x4cbe47d38686f262,0xebb0e72f2cb670ea,0xae47fa937eb27468,0xa41eced33456f6e0,0xbcb774e0841e71ec,0x1fae2dc12474f06a,0xff845f7fe84634ee,0xdcffba9a2a92d26c,0x7bdedbdfa4f674ee,0xb6f6f6e8049776ee,0x296c3f608c3476ea,0xc2242c63a210556a,0xc5d7fb2ca455546c,0xeb7ffe5cbe3674ec,0x24207480284710a0,0x6008f3612422406c,0xfea60f4a647670ac,0xc154d6043c367664,0xfbfffedfeedcd7ea,0x4bb53d2665410ec,0x9148f44a2610c468,0x8004000004000020,0x6180e31c2e1e51e6,0x400000820000028,0xde200001068,0xd4a500a0a0c01268,0x54014000000,0x8800042040000060,0x40600011a0120024,0x1142c30000d542,0x304710e568e25220,0x8400014000001008,0x800000800040008,0x1800000101000,0x8040000000c0000,0x86811661661c7464,0x0,0x0,0x4806a08200c5260,0x144044c00163004,0x80400040000020,0x7d64766036d4f5e8,0xdeb7bffff6fef7ee,0xc4cfff1fa456d2ee,0xae5f5fdef455f4ae,0x94e9cdb3cc15f46e,0x20100000c000,0x81041900040ed16c,0x8bb06763e4bef0ea,0x480014000000,0x80a04979460504e0,0x404000000040002,0x280cb52204067662,0x98007fa4400c1462,0xc022960000126240,0x8e99c75a2c1674a8,0x4160e88200414e0,0x9212704004945002,0xcaa2d1262054d2ea,0x2c800000010e0,0x18000540001c01e8,0x5455fe6bfe56d66a,0x2a0560801c40e0,0x91945f5d024e1646,0x63d39e3a809654ea,0x2ac674dda4227768,0x6580a34040062,0xc91ca03824525260,0x54143c52e01412e2,0x8452b290244630a2,0x695f6e45263dd5ee,0x1eba7ffba433566e,0xf5feff6ef43f7066,0xffb67f7ff6fe76ee,0xe9ded72fe67472ee,0x6ed33e93a634d668,0x14991ca16204c062,0xd14bb6db34bef06e,0x65b2b56a04bf14ee,0xf316ef028c6775e2,0xc6da12eba4f05168,0x16140e02a0020024,0x7fbf7feff4f6f6ee,0xf5bafd73b63e576e,0xcd95bfb92cf6f6ea,0xcffffb1fe6f3d66c,0xec5fde5076d2f2e8,0x7cf4a452f617d362,0xf1ff7ffe7696f6ee,0xabbc0c65b614546a,0x81dd4b186416106e,0xa3755c8a61434e8,0x14137fa02686d662,0x20050020000260,0x20200000c000,0x81e433322426e066,0x2010108ea0022068,0x800010204001240,0x90145210000cd08a,0x40100001008,0x119002014060,0xa5802110290504a,0x410028140124024,0x0,0x4800000000801000,0x840000000020080,0x2420a6168005020,0x400100000100020,0x74001290009670a0,0x0,0xc103f92c2602d2ea,0xdfdfbe3ffeb6f7ee,0xaade97daa436f6ec,0x34b97efbf61670ea,0x9367fd5834725664,0xd3b67ff58eff576c,0x4d4acf2e3ce656ea,0xef7fffdffefef5ee,0xf4bffffffef7f6ee,0xfffffffffefff7ee,0xffffbffffefff7ee,0xfbffff7ffefff7ee,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x408300000090240,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xff7fff77f1feaff8,0xba9a677c71adb3ca,0xefd7ffbf717db7df,0xefdfdf79f17dafff,0xbeb27ffa71fd9fde,0xfbbef77ff1debffb,0xcbdeffe5f17d9ffb,0xebbbffbb717cb7ff,0x6fdebffbf1fdbfdb,0xf7bfffeb71fdbffe,0xd5aefe77f15ebffb,0xeffeede1f1fdafde,0xfbfdff3ff1ffbffe,0xddfffbfff1fdaffe,0x56bfcfdb71dfa7fa,0xf7e6f7ef517fbfda,0xffffff7371ffbfff,0xdf8eff7b71ffb7fe,0xffffff3bd1dfbffe,0xafdfdfdff1fdbffa,0x74f7d6ec71f4aed8,0xffedef3b717db7fb,0xfbb77ffef1dda7da,0xedffffbb71ff8fde,0x2e7c588a014d074e,0xbfde1bf371dd06d2,0xf9dfff1571febfea,0xdeae3f65711caed6,0xffffffff71ffb7fb,0xefffffddf17fb7fa,0xbdbf5f7bd1fd2eda,0xf3fffff971ffbfff,0xfffefff7f1ffbfff,0xebfffffff1ffbfff,0xeffffffff1ffbfff,0xbffffffbf1ffb7ff,0xbfffff7ff1ffbfdf,0xdffe7ffbf15f977a,0xdbddef7ff17d1ff8,0xe6fbbff950ddbf7e,0xffbfdffbf1fdaffa,0x7dfffefad17fbfda,0xdfffff62f15fbffb,0xfbf7bf7f717dbff2,0xefffff3f71dfbfdf,0xffffefd7f17fbffb,0xfcf57efbf1fd9fda,0xfdfbfffa71fdafda,0xbf947fbe717daec7,0xfdffffbf71ffbf7f,0xebff76cbf1fea7de,0xdfbb7ff7d17d87da,0xffeeff70717dafd8,0xfa9e2f7ff1fdb7d6,0xd7d7eb6b709c17d0,0xefdfbffff1ffaefa,0x96fb6fc0f1bd36fe,0xfbffbffff17fbfd7,0xbeae3f687115af58,0xebb0fb6af17dafdb,0x6cdf139f516d86de,0xb7fecfb7f07dbfec,0xd5e3bf7a51bd3ede,0xef7e7ffb717da6fa,0xcbc6eb03f1fca2fa,0x4ecab7735075a3c0,0xbd8efd7bd1dfbbf2,0xb1bff7faf17dbfde,0xaf5eedf5717dafec,0xf936f7fa71dd3ff2,0xcfffbe1551bfaefd,0xbf5fdfdb51fdbff8,0xf4f6d6e750fd8bfa,0x2d682f4a01dd0792,0xfeb2af6a51df87c2,0x8fc6f33ad171afcc,0xaf7c7f0cf04f86d8,0xf2926fe951debeda,0xc97eec50517ea6dc,0xee363d7351fdaff0,0x8c5ff3ea51dc8afa,0x2b5eded751d4bfde,0x3efdd7f0516da7ca,0x75a6b77ed1fd9bd2,0xc3debd7451f6077e,0xcbd9ef63d1ee37f2,0x4bc89a7871df8358,0x7937dd671f52fd8,0xf7fbf7fc50ddadd8,0x87be7df2517cbfca,0xf9d6ff1e415fb3ea,0xe939e8b717cb3d8,0xfadcd5351dfa7d8,0xd8b4874850fc3fd4,0xbf94ef4250cf27fa,0xdfa977de717f9eda,0x8c46be89515d835a,0x2b18ff9bf17fa76c,0x90b6d6ead0fda6c7,0xe0f6a761716597d6,0xdab47f7e513f17da,0x54c26f3b519c0fd0,0xeb3570caf11d3ccb,0x92330de7711e06c8,0xc510a843d0500210,0xc11a011180002,0x8f4eb600d11503d8,0xfffffffff1ffbfff,0x2fb66f72d07d86d5,0xf17dffe870d78fc9,0xce8020029000a250,0x90c5324902e0370,0x9d0125051a4a240,0x0,0x6000021040410080,0xb0042c2020010350,0x4850800851002082,0xa8620800101c0608,0x0,0x500886e1b1008754,0x90000c4000680300,0x0,0x80200300240008,0x8200100800,0x0,0x0,0x800040000040008,0xf987d52a612ea0ea,0x200000850000000,0x80846001410242,0x6144ff6370349743,0xfeb73fff71ffaffb,0xefbfffaff1ffb7ea,0x2fdefbda715fb6f8,0x8aabcf7b517fb6ca,0x8002000080200,0x4000280200410140,0x89c2e77b714d07d8,0x4eea7e2040c5365c,0x1001800300580308,0x956c924c503c23d8,0x3000246040150840,0x980a303400c2480,0x8000821050000080,0xaa2441a8101716c8,0x7000841001012248,0x5080294041242421,0x438051595061a352,0x8dc7bb83417daf5b,0x103494040102270,0xf475566a7055b6d4,0x7986936010451749,0xff946ce3513da65a,0xcccfa24e515796d8,0x8bdc5d99513c87d8,0x96d915285128a250,0xd860e36050cf86c0,0xf6b4277551d7afc9,0xcbe4972251d495d1,0xab5bff47d15ca7f0,0x86a3ed5951fdb5e2,0xd1fd777b5177a7f8,0xdfa67f6ed17f37fb,0xebf7f7ff51fdafff,0xedf7bedf71f7afe8,0x5e8ed56b71efb7ca,0xd526f3737155a7fe,0xeeae2570d1741a9a,0xfb97ff0751ff37c3,0x449216015108a252,0x27b84f60105d3fc0,0x1140f6f8f114a6dc,0xdb8e7d72717d96d9,0xfa34ffeb51b5bff8,0xcffefebbd15cbfda,0xe9deffdfd1feaf7f,0xd6f7fcfff17eb7dc,0xed5fff7b71bcbeee,0xffb36f7470bdbfdc,0xc0d3c33971d51b4a,0x69e0f12413c2a22,0x56933fe8017907d2,0x1180a90011400160,0x5c263ee651f5af74,0x8002422040003401,0x48824940111250,0x422000401010060,0x1288c4001ac020a,0x0,0x2200800010140,0x4140014e51110240,0x20040000000210,0x440270000700250,0x4002001000020,0x400200000042,0x50100800010060,0x130065250412540,0x417086ca515c2d0c,0x12000c4100000342,0xa881c16a5014ac48,0x9ec2bb1ef07f1bd0,0xafdcc952f1bca3ca,0xfcffffff717fa3fa,0xdfd5af17716da77a,0xfff7ffff71febfff,0xffffffbff1ffbffa,0xefffffdff1ffaffe,0xfff7f7fff1febfff,0xfffffffff1ffbfff,0xfffffffff1ffbfff,0xfbfffffff1ffbfff,0x100080041000000,0x0,0x0,0x0 },
	{ 0x94ffa8402a06520,0xac446a460220ad44,0x20aacfd990fd2442,0x8c761a69dab92d30,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd9fdf77adafdbd76,0xdabe35369ab5e5f6,0xdffff7b7daedbf76,0xefdfdfdb9af0bdf6,0xdebaf7fb9af3bff4,0xf3fff7fbdafabfd2,0xfbae7fefdafcbd20,0xdbdff7fe9aecbff0,0xedff1ef992b4bdb2,0xff3b7ffadab5b566,0xfff7f6f6d2a6add6,0xec5c75b390f9af74,0xdbf5fbbe9af5bff2,0xdddfbfdbdaf2bdf4,0xcdfccfffdaf5a756,0xf7fffefbd8f9bff6,0xfffffff7dafffdf6,0xeb9fffffdaffaf36,0xcfdffe5bdaf6af70,0xebdfdfdedaf6aff4,0xd4bbfeee9ab2ad50,0x957f77f9d8e4bd34,0xfb9f6ffc9af4ad74,0xe9cffeffdae5fdf6,0xce1f5f5b92b9bdf4,0xb0b947e0dab42d04,0xbb5e5f6b9af28794,0xdabe2ff79ab2fd36,0xefffff3fdafbafd6,0xeefefed992fdbf76,0xfdffdffbdafcffb6,0xf7ffbffddaffbff6,0xfffe7fffdafdfff6,0xebffffffdaffbff6,0xffffffffdafffff6,0xfffffdffdafeeff6,0xf5fff7fa92efad76,0xfffefde0d8f1bf56,0xcbfcef6f9ab1fd64,0xfffbbeffdaf3bd64,0xb71e46769a69ede4,0xf5fffffadafbffe6,0xfffffff7dafffff6,0xfdb5fb7edaf6fdf6,0xebdffebfdaf1fd36,0xefffdfd7daf7bff2,0x1cbed7e0da60fd54,0xfdfef77adaf5adc6,0xfb9467f7daf0af12,0xedd3ff7f9ae7ad66,0xeefedfd3dab0adf2,0x9dfe5bebd0f9ffd6,0xdb2fffff9afcef64,0xdaef39ea9a20fd66,0xc78af6a3daf5f564,0xbfdffedddaf2fdf6,0xbfefd7f29870b576,0xdff4b7ee9afb2d74,0xb8ac2d669860b564,0xe3fafb3e9ab3bde2,0x6dc63ba592b4e5d6,0xffbffffedaaef534,0xb5f7f7ec9ae4cfe0,0xffee7ff1da78fdf6,0xf3d4ef6edae5ed36,0xcec33e199af13744,0xee3c5f4b98f03466,0x75f7ffe8daf6bd34,0x8f7aff5290f1bfa6,0xff9dff6f9afebd74,0x85dcf62fd238ad44,0xaf5cc35b9aedadb6,0x9ed6f768daf4f560,0xc8a66e6892a0a520,0x4386646490902150,0xc5c7af9e9ae6ad60,0x475cc00a9070ba02,0xb61557639220bc56,0x91000e628064af04,0xfe863f43d834ed24,0x6d83f62bd2a42d14,0x2f56b35f98f2bd50,0xffab3f6f9ae4cd60,0x93f7f769da78bff2,0xf7fe7f6d9af5ad54,0xcbc6e4638a74a504,0xe7e7397918f0ef26,0xb7bfcffd9af4bd44,0xb56fb67b9aeafdd4,0x86a445619800fd20,0xa98de774daf4e566,0xdbdb3fcf9af8fd64,0xcbb4cb6390b4ad26,0x55f2ff7092a0ad14,0x5baaff63dabc2534,0xdfb6676fda70ffe0,0xebf7ffbf98e2fff4,0x75f9ed79aa69d50,0x50b4b2421a883412,0xd9aefff2daf8a5f4,0xde80420292342410,0xa88efb8e12247544,0x2b8a335398a0b416,0x42014a098c02000,0x2040227098608520,0xccf2247fda802d44,0x8f54f02490a0a510,0x87e3c09da81acc0,0x9ee87d7098e12754,0x30ed244a12b02544,0x933c0d61d8c0a100,0x0,0x2042001080800140,0x851a526390202142,0x42054090801000,0x8a02644188882404,0x81004e709a242510,0x450301190642732,0x808000000120,0xd2b07e7012a4cd70,0x9000052098400400,0x6888c132d0a23c30,0x384222398082806,0x8480110002420,0x0,0x540280202000,0x0,0x840280000208c00,0x0,0x47806bac8d12600,0x1f44702288c1bc44,0xfab3bf5fd0e1bf42,0xe9ffff32d274bd44,0x2d108ec992d8a516,0x1aa85f9ed0e2a794,0xd99c040082a42544,0x78a22c420030e584,0xe102b31b9231e512,0xef427e12d8a8a920,0x5218421812a02522,0x50c0143980508a80,0x440205488a008580,0x4002000200040,0x840000000202000,0x218430010222440,0x20880000000,0x4080055000382000,0x300b632c0702446,0x402fea89820a944,0xab5a11489024ad20,0x14d45ed3daa02d40,0xdb44a5c392b4b540,0xf7382f579ae42d26,0x9943ae08c8782c44,0x8846de1398f00120,0x828856b290b4a5a0,0x79cc76269afc2564,0xf2b42576d8f4ed00,0xca41f36bd2b4b570,0xafebeea98f8af60,0x2626d7789a62a552,0x9c57af7898f0eda0,0xfa967d7fd2f0ff64,0xfbfffb1bdab2fd76,0xffdb1fdb9af0eff4,0x86bdd6fe9aa0a572,0xf0dcb75d92fead46,0xc52c7c71d230fda4,0x8a52e5208a65a584,0xcdcb7e59d0b4adf0,0x633a4b53d072ad60,0xfcdffeac90ff6560,0xfdae1d229a70bd56,0xbab4fbfb9ae0af46,0xcdcbfe19d0e0ad56,0x6b7bcf42d2fce5e6,0xdcd75ffd92fff544,0xf168ab739ae0f552,0xdf846b6788f4ab42,0x8fdfef4f9af4bf64,0xa08b108398402256,0xf2b65febd0bdf956,0x500070021080a804,0x8800050008400100,0x4880921880a52404,0xa304108402400,0x8400042008600040,0x8a510d8400400,0x1240400190102012,0x8080339310283404,0xc8d21c0382b42540,0x8002000080800000,0x1100242010308480,0xa00000000008800,0x0,0x0,0x3800010888410000,0x2110000400,0x13380e5012b08520,0xc430637090a62520,0xcfde11019ac32452,0x335e5dca90a0e522,0x34b57ee8d264f732,0x9d6ff763daf4ffe4,0xfbd37f6e903d3d46,0xffffffffdafffff6,0x4fffff579af4bf74,0xffffffffdab6fdf6,0xffffffffdafffff6,0xffffffffdafffff6,0xffffffffdafffff6,0x9edf7fbf9ab6fdf4,0xb4bbe8241a344544,0x0,0x14090000000 },
	{ 0x0,0x9540c5080c86600,0xb4485693cdc82e02,0x4130ba49250e4120,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4045029000a0160,0xf9ecf7fefddeff63,0xfeb63f7e6f4a7f62,0xcdded4e05bda4760,0xfbfedfdfedce7d73,0xb6be6fff0f7eff63,0xdfffbdeb6dfeff73,0xdf9e7dfdffdeff63,0xebffff7a1fdeff63,0xfad6fedf9f4eff63,0xfffbcfffdf8efd63,0xffff766c4dfe7f62,0xe39e3ff3cf7aff62,0xf354ef3e1f7eaf70,0xcfdb7effedfe6763,0xeebedff97f1ef762,0xfdfff7feef5eff73,0xdffffff7dffeff63,0xdfbfffff8fde7f73,0xdfeaba0d4d087f62,0xfbdbdddacdfaf763,0xf6f7f56eaf4e7d63,0xdf5b07e04f0e7f62,0xff026bfd5eda6762,0xf5dff7ba8f5e7773,0x6f7fffdf9fde7761,0x34fefeeb37ea7573,0xbbbce76a1dca4d62,0xf2ae3fff7fea7f63,0xcdcfffffdffeff71,0xeffffffbefceed63,0x97bbffebcfde6563,0xdfffb7fbff7eff63,0xfffeffffefdeff73,0xfbffff7ffffeff73,0xffffffffcffeff73,0xbfbfffdbdfeefd73,0x977ef2edff7a6d62,0xdbf67d724f0e7f73,0xfbd6f76fddce7f63,0xeff7bfddcf9ef773,0xadbedf796c7c6f61,0xfdfff6eaddceff72,0xfbffff73efceff63,0xef8fff7f4fe8ff73,0xdfc7fe0f5f5eff73,0x2ffffddbcf7eff63,0xf4bcfff9cfce7771,0xfd7f7f774ffeef63,0xfbbe6f354f4cff73,0xdfc7debceffe7f63,0xff7fdedbbf6e7f73,0x64f696e34f4c7562,0xfdfdbf72cdfe6762,0xfebd6df20d6eff73,0xeff7f3a6435e7f63,0xefffdfff0fde7f72,0x2be5ffa4f6a7762,0xf3fcfffa6dde7763,0xd7ce3fb53dea7f73,0xaa1afa2acd3a6760,0x3dd6be58bb4a7d63,0xced9dfbb2fea7561,0xf17ffeffcfecef63,0xff2eb5eb8f5eff73,0x90dccb66071e0771,0xffcfbfcb5fce7f62,0x7fbf4ff62f7e7f63,0xf5fffee9c9de7f62,0xfbfa3de2fdde6f63,0xfb9ff73e5dfe6f63,0xc7d6bea34f8e5d63,0xa79b9fd7af0cff63,0x7cbd5efb6d4e5560,0xf97fad7a275c5763,0xfda42f67c8f87d61,0x6146e72a05cef960,0x321ed0820d0e4d73,0x101609820d084620,0x9bcdf66a4dda6763,0xfebc8f741f5ef562,0x67ccfb7f4f4a7760,0x6a066e0501a80760,0xedac5ffa3dce7763,0x9cffff69876a7f63,0x7dc635740d1e7761,0xc1e1ef3e8fcc6763,0x7fd637dd4f5af541,0x9ffe6d624ddc7762,0xbdf687f94d0c5f63,0xc8e825018b40f601,0xe3662002059a7b62,0xc87676ba00ece573,0x20b0c7010d8a4162,0x88f35aeb09de2770,0x8ed63d52c8e86451,0xfd80e52449427a23,0x947b28c4d4ee571,0x61d99b54156a4562,0x1495eee21afe4f61,0x813e3562251e4760,0x6e364f2c5baef741,0x8192c918488a4542,0x2b34de1e0c607743,0x28000000004040,0x8008200000000960,0xf6942db1094a7f60,0x20e0815002100,0x35679ff81f3a4f62,0x300858305020541,0x516c34ec092c7761,0x4000008002000400,0x806c902085003440,0x25965c0002c85c20,0x37bc056644b87541,0x4412c4e828405962,0x500160000444022,0x900001800044022,0xe0bda00080a4d60,0x80088c010040e820,0xb15fe24a094c3f60,0x100800000880000,0x210012020402020,0x0,0x804000000440,0x3404046001664062,0x100812000002060,0x8000010000003000,0x4000000000005000,0x2552100140466121,0x400200000000140,0x23486c600d004b60,0xfabe3f7fad5eff73,0xd7fff7a7cf7e6f73,0xa7527fdaac5eed60,0x302069002404461,0x9524046101ce4762,0xc1006c3c0b0a4442,0xe034920c0850f740,0x400636300d080c20,0x801187900f4c0460,0x9258785209c87860,0x100020001080000,0x80800308087020,0x884002124a006402,0x2608cb11410c6721,0x46d02200c406501,0x38c406041087541,0x6480a61002100d60,0x4002380030100120,0x8b514d500e3a6560,0xfc77fefecfde7773,0x90ef2768884eed70,0xdfb73f7f9f5eff63,0xdc5a1d0859084d61,0xa94edf460d485573,0xa47a5be00f0a5760,0x90fcf9634d48f162,0xfef2bf5069227161,0xc7f7f7d61d8a6762,0xeeffdb5b4fee6d63,0xa79a57fb2d7a7f63,0xd0e754c0010a4d61,0xf6fab8728d4a6843,0xebf6ffffddfe7773,0x6ffe9f7f5fda7f63,0xbbbeedf3efdef763,0x7567f7737f6e7f72,0xb667d42150a5942,0x8302eb0f35185960,0xccdab2736fa87462,0x9634442011522750,0x5df7dceacf58e762,0xed7ffdf64fba7763,0xfbbcfe6f1d5e5762,0x57d3ff2fff2e7d62,0x7b52979f9dde6760,0xf436defd1b5e7f62,0x936af77c0f7e7762,0xffba6f7f6d7afd63,0xceffea2add0a7762,0x2b6edbcb4fd8f541,0x901344e2804c2422,0x50008400000,0x4000008001000800,0xd40260008406141,0x2000800000000820,0x2008200000460,0x20100a0020004022,0x0,0x81a0f1031d6c4d42,0x4800060008080432,0xc00005000000,0x0,0x2050200004640,0x8000001000080020,0x0,0x0,0x80a8001000e20,0x41103f8048207e73,0xfa94e4280a4a4561,0xf02db4924487d70,0xe14696cacd4e6160,0xf7b6f6f9cfda7f73,0xfffffffbfdfeff62,0xffbefffffffeff63,0xffffffbffffeff73,0xbfffffdffffe7f73,0xffffffffdffeff73,0xfffffffffffeff73,0xfffffffffffeff73,0xfffffffffffeff73,0xffffffffcfeeff72,0xdfbe8fff4f9e5f63,0xdbfffffb5bfe6772,0x45423f4114300542 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x455766a0a9e7c61,0xf5ef76ea4ffefda1,0xdeb61ff2d7de3fa1,0xc95adb2afaae3cc9,0xefdfffdfdfde7fe1,0xbfbd6ffbffdebea1,0xd7f7f77fffbffde9,0xff9e7ffd7fff7fa1,0xfbf7ffef57df7fe1,0x8fd7fff6dfbefda1,0xfffffffaffbe3da1,0xfcfffffbffbebda8,0xff7ebf23d6bf2e29,0xeb97f77e773f3c21,0xfe8bfadc43b67d80,0x6fbfcbb38aafbde0,0xfdfffffefbffff81,0xff7affb79fbffee9,0xdfb7fb6f5fdefda1,0xcd53fea87e967ea0,0xbbfcdfdb4bde7ea8,0x36b9f6e0dbae3f80,0xdb7ff763c7bf7ea0,0xffbbff7fee9f3da8,0xf7dfffbfffde3da1,0x225b5fd7ffd33e00,0xbeffd7fbd7de7fe1,0xf1b93ffa5fbf7ce1,0xbfb7af7f7bbe3ea8,0xebfffb7fdfff7da1,0x6fdfffdb7fffff60,0xbefbcffbdf9fffe9,0xf7fffffbdfdf6fa1,0xfffefffffffeffe9,0xdbffffffffbfffa1,0xffdf7ffff7bf7fe0,0xbfbbfffbf7dfffa1,0xfdfffef9fef77ea1,0xefbe3d79febf3fa1,0xdbfdff2fdfde7ca9,0xef7efed9de9e7fa0,0x2fdedf7b7ffe3fe8,0xfdffb6fddfdfbd29,0xd7feef725ffe7fe0,0xf9a4ffff5bfe7ee9,0xcfffff6e7b9f7ea9,0xbffdffdbdeff7d81,0xf4bff6efefbf7ea8,0x27feff71b3b62ea9,0xffde7fff5bdf7fe8,0xffdfff3e5fb27ea1,0x5b7fffdbd7beffa1,0xa5ff7eebdeefafe1,0xffdea57effbe3ce1,0x9abe2d7be2fe6ea8,0xcbe0f3bd57de7400,0xadde9fddfab77f28,0xbfba5ffb77dfff81,0xd5f9b7bec79f3fa0,0xf3e46f73cf9a2fa0,0xa127e24c308e3ea0,0xefebfbd3cedefca9,0xffbdeff2fbbefd80,0xfef7d7ed5fbe3fa1,0xffbe4d63df9eae80,0xfbd4af5ffeff7ca0,0xecf3bfcf1f9f3d20,0xf6efdff2c5de5fe0,0xf57ff7eedade7ea8,0xfbbe7df17ffe3fe1,0xdbdfffbfd7be3ee0,0xdbd3febb56be7cc0,0xabfedfdbffffaca8,0xfcfeffe9d7ff3c21,0xdf6fef45d78e3da1,0x42ba2b7e4fde2480,0xc8c2cb18e39efd80,0xc0561606461e2900,0x945117c952be6d80,0xe54ffe73dfdfefc0,0xfebefff3dfbffee8,0xe4ab2a24e182d21,0xa01e5a1550962cc0,0xd7be4ff3579e76a8,0x7af977605ed62f88,0x4f9e07f45b963ca0,0x89c2fb7b4f9d7fa0,0x21d3164802852d60,0x3bbeff59defe3da0,0xdd57b4addaae3da8,0x2cb624725a9e3da0,0x8bd6c36c5d9e6f80,0xfec2121a78b62449,0x89441a102ce1980,0x74fe56f04bfffc80,0xc7aa6f6219bf7c20,0xce8c07338a9e3c80,0xcc8bf1a057833c01,0xab1fd7584eb62ca0,0x96f256e9cbbe3c20,0x2d6f35607a8e1d20,0xdebe7ec45fde7ca0,0xa91d6a0429a7c09,0x2d9e4b000216ac80,0x440010040000400,0x918c8f22482c3481,0x12a40050c0920c40,0xe42e02504a0c2480,0x965e08c68c2c20,0x841817f0561a0c20,0x5124f44856b39c00,0x212023402103400,0xcbc67f3f7f9e7d00,0x644120040043c00,0x24280dc1529e6420,0x10220b1020c1400,0x10041102040480,0x100020000002001,0x6867aa46ce3d2c88,0x200406c004921040,0x1a4040882002000,0x192a454010863e00,0x0,0x4ec2021082182000,0x2d088b9b560c3ca0,0x10806000040980,0x40040000021000,0x0,0x0,0x0,0x10001000024000,0x4d44f66217ffdfe8,0xffff7fffdfbeffe9,0xfffffffefffebfe9,0xeffe3fdff6bfbea0,0x480400000120000,0x819884020a003c00,0x400004000c00000,0x88d09b7156923c00,0x4346223056b2ec20,0x4808c962128e40a0,0x54743660c69e3e80,0x1200240040124000,0x8180a00200c02900,0x2002308818061400,0x480541846162000,0x3006020812000d00,0x20850042002400,0xd082932212ee2c80,0x88f086053a900600,0x8e98015a471e6ca0,0xf8f4beec77be7ea8,0xd088040081860820,0xc332bf6d1bbe7c00,0xac4fe32946d62c20,0x21365e0142963c20,0xce6d7c11abc1d20,0xb63aa7fc6a976e08,0xf73627d66a8e3fe1,0xf9c1fb0ec39e3f28,0x2b9f0bcffbbe2c21,0xa121e88548604a0,0x8100224c02128420,0x681205804a8c0400,0x8bb2fb27bbff3d00,0xeeff7ecf57bffde9,0x3fbeeffbdbbefee1,0xffff36ffdebfff21,0x6d3e1dea5e972ca0,0x252030612863d00,0x86563340ca9c2880,0xe3be89e38f9d3c28,0x51e7d6ed569e2ee1,0xa43a3575529f7e01,0xfb8ffb7ffb9efca8,0xedfef6ffdffe3ea9,0xa36edfdb45be7e80,0xb4f55febdebe7fe0,0xfdd627f85bdaae60,0xdd966f7d73f67ea0,0xdddfffbaf7ffbde9,0x69dfff4f729f7e80,0x1089d5eb608c2da0,0x0,0x0,0x800c32803322c00,0x20110800000,0x422000002082400,0x9021022002002801,0x124246049a812800,0xaba5fbed57946fa8,0x140120002082400,0x3000000000200400,0x0,0x8204002001000420,0xc000c0000,0x0,0x8002000200002800,0x2000020000000408,0x6f4ef7f19bb66ca8,0xdb95fd3f57982d29,0xced7f7b7dfcfec60,0x2948011a05140480,0xfeffffffdffffee1,0x9fbffffafffe7f88,0xffffffffffffffe9,0xffffffffffffffe8,0xffffffdfffffffe8,0xffffffffffffffe9,0xffffffffffbfffe9,0xffffffffffffffe9,0xffffffffffff7fe9,0xefffffff7fdefea1,0xb6ffffefdffe7fe0,0xfdefdf6f7fdebee8,0xfff67dfe77fe7d81 },
	{ 0x8ab58326d7142f17,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xbefffefff7bc677e,0xffffbffff7bc6fff,0x9eb0277ce71422df,0xddc7f603b390276e,0x2bdecc4af2a42b6e,0xbff7dff3d39c6fcf,0xfb7da77c921e6f5e,0xfffe7d61f7b66fff,0xebffff7ff7b66f4f,0xafd71f5fb714635f,0xbffffffff7be6b7f,0xf5fff6f9f7866bff,0x7bbe7fe2d6bc236f,0xfbafffffd7be6f7f,0xccdbffdff6bc2b76,0xefbfdffbf7be6fde,0xfddffffef6b66f7f,0xf7fefd67f6b66b7f,0xfbbdffefb69c635e,0xe7de7ebfd7b00b6e,0xbfdfdfdfd696637f,0xd4fff6eef79e2bd6,0xfdeb5f60d7822357,0xffb67fffd7be6bff,0xcdc7fff7f7942b7f,0xeeffdcdbd63c2f4e,0x773b7fe7f7a82b7f,0xf36bafe09294634f,0xdf9737fff7b8677d,0xefc7fff9d7b46ffe,0x3fffdf7bd7bc67f6,0xbebb7ffbd7b4675f,0xfffdfffbd7be677f,0xffbe7ffff7bc6fff,0xebfdff7ef7b46f77,0xfff73f5ff7bc275f,0xdffffff7f7346ffe,0xf7fefe7ff6bc6fff,0xffbefff3f7be6fff,0xf9d9ffffd7b62b7f,0xef4ff6fff7b627ff,0xb7bfcfdbd7bc2f5e,0x76ffdefed7ba67fe,0x9fdffff2d7ba6b7f,0xfbffbf3ff7bc2fee,0xdfd6ff3ad794675f,0xfffeffdaf6946fcf,0xf6fff7fff7b62f7e,0xfbf5f779f7946f97,0xff34ff72f79c2f56,0xebd7ff9bf6b0234f,0xfffedfdbd7b46fdf,0xb5b75f6b938a2fe3,0xf9dff77af79427de,0xd8be2f7ef7346b4f,0xe777bba3d616673f,0xaefffedbd7906f7e,0xfffbfffbf7bc6ff7,0xd7f5ff7af7bc6b7f,0xffde7f76f7866337,0xe8f6f77bf794236e,0x697f3ff2d7b86bdf,0xbffecff3f392675e,0xd7fefaf9d3342f5d,0x7bbc4e7092a467ee,0xcb90773cf6166b7f,0xefffbfdeb79e6fff,0xaf7f5ff3d7946f65,0x317f37fab296232e,0x6dfaff71d7962f6e,0xff967f7ff7bc275e,0xcfebfe93f7162b43,0x3bfee7d3d6b26bd1,0x76f6f6ffd7b467ef,0xdfffff7bf7be6fff,0x9fa8f76fd6342773,0x28c7eb96c780611f,0xe8d0b71f928827cc,0x1ba5362d3302216,0xdae53543f7302e1e,0xba973f61e72c2347,0xeed673dd97166f4e,0x88eebf6fb3842bcf,0x97baff7bb09c2bdf,0x91ee774af4b86347,0xcafe2562d7aa4b4e,0xc188e26fd0a82b5a,0x617712f6d3b0634e,0xedfafd72d7946776,0xd1b6dc6ad62423fe,0xe3164d62d5106374,0xea906522c4884a42,0xc882fffff7b40366,0x2f9dcbf4b2982f5b,0x9df656adf39e431e,0x17061df0d602016e,0xcdbeff7fd7942fcf,0x8ac2330392202941,0xbbcacfdbf7342f7f,0x86ff57e7f68a275c,0x933e77f8d79e2366,0x5fbe2f7fd7bc6ffe,0xf3c7f61bd6322759,0x301e0000a00048,0xa00053a83040382,0xfcccf67182b6275d,0xc282354c80042e10,0x8000d00082004104,0x2c206808f4042101,0x828a0d2aa0102b04,0x984026608210428c,0x9b04782192a00056,0xa004200c4102400,0x48020100100300,0x59c4c0d54b003d2,0x5e75c26a94902a48,0x1ba43600e6002b01,0x5800612010000160,0x4600040002202100,0x61b88d20d480214c,0x64010800342,0x4a98b51293902227,0x4100080002000,0xc10a0010002a20,0x891a8fdbf7b60b7e,0xc21444e4b630416c,0x800e024287002044,0x2a00400c4100010,0x88109000a0100a00,0xb40834050082215,0x983e0d629200270f,0x754cfa63f72c234e,0xfebeeffff7b46b7e,0xeffffffff7b66fff,0xffdf7ffbf7be6fff,0x1491640891206100,0x5008046246002004,0xba060da491a02172,0xe3d7e369f6842376,0x5ce61e1496a227e4,0x1011d4d3920c634d,0x10c012689200012c,0x40000d2090002321,0x8004202002002005,0xc08228210202a00,0xc4534a2082842647,0x500322d6910c0316,0x88a4028220204c,0xc1808216820a0a14,0x5b42bb36c6142ad6,0x2beced683802343,0xf09ee2fe96962b64,0x8121841ae4302204,0x59002d3290942642,0x48c2ef7eb786035e,0x219180992802107,0x94475763d780434e,0xd0daf7fef79443e6,0x74802d62d6900346,0xc8f0e00682106304,0x881a209000480c,0x0,0x8102282100200,0x1000004001000002,0x8002100890000042,0x8de5f41f3244372,0x9bafeff3d7942f76,0xfc7bd768c7b42b4e,0x5bbc7d32f79c63df,0xf394ef36d714675e,0xca5e3287d2b42756,0x8ff5d7f2f6942b7e,0x51f43eebd7202696,0xafdaff7392b46b7f,0xdb9cfb7ef2ac2f7d,0x8fdede9fd69c234e,0xbbdedfdfd68e676f,0xf6fff6ffd7b26fef,0x31be7678f2042b4e,0xdbb5b7f7f3942ffe,0xfdffff7ed79e6ffe,0xeefcdcdff6b46b7e,0xbc7bd763d7b467ce,0xa1b48228922c2a45,0x40220442a0046648,0x10844000000,0x165a5000202a0d,0x800205008000001c,0x203c3882002288,0x76343d7cf2942b16,0x8000c00090040040,0x101040000010,0xb010030080046200,0x0,0x18002000004,0x0,0x4040004180002008,0x828000230202000,0x0,0x10ab420d2026210,0x13841210d3002006,0xcc523e43f6a40164,0xeaa287cbd784255e,0xf4bffffff7b427ff,0xdffffff3d7be6b7f,0xffff7f7ff7bc2fff,0xedf7fffff7b42f7f,0xfbdfdf5ff7be2fff,0x37ffffedf736277f,0xfff9f77ff79c6f7e,0xfeffbffff7be6fff,0xfffffffff7b66fff,0xfffffffff7be6fff,0xdebbffebf7b46f7e,0xfffffffbf6b66f7f,0xfffefffff7b66f7f },
	{ 0xcbbfff7ffbbe77ff,0x0,0x0,0x2808000200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2100000c000,0x34bfdfeedbef73bb,0xfbffffefcbff77bf,0xfef6bffefbf6dfff,0xcfd7f7befbfc727a,0x2ffffff9f9be73ee,0xfdbadffbe391fbff,0xfb3edd7efbbc53fe,0xfbf6fff5fb985bf6,0xebf7eb3bfbfef76d,0x6cdb7f7bfb305fee,0xbfda87f2dbb453ff,0xddf677fffb347ffe,0x4fbefff3fb3e57fe,0xd9b5ff76cbb07b6b,0xcbefbfdbc9f87b7f,0xffbf5ffffbfb7ffe,0x1dff96fbe1865b7f,0xcaaefd61cbf0532b,0xdbbfef77dbf957eb,0x4fdefabbfbb05367,0xffdeffdefbfb73fb,0xd5fafe70d998d33f,0xdcbf7d6afbddf276,0xefbe3e7eebf4fb74,0xcccaffbaf93dd3ee,0xa7ffff59fbbb53fb,0xb63f7773d9fcdffd,0xfdffffeaebb5c3fe,0xfabf3f74dbb073fe,0xeffef7bfebbedfbf,0x3fdf9fdfebf37bbf,0xf7bb7fffdbffffff,0xfdfdf6bbfbd27fbf,0xffb67dffdbb4fb7d,0xcbdff72edbfe77bc,0xefdfbffbebfcd3fe,0xffbecd73dbfad37b,0xd87ff7fec9fc5ffc,0x7ffe3ffbfbfffbff,0xfbffff7febfef7b9,0x5fcfb6bfebf1d3bf,0xb5bfdffffbbef7eb,0xddfefefbfbfcf7fb,0xbffefff3ebbc7bbd,0xffb5ff6ffbfefbff,0xfecff7bfebfcd37e,0xafffdfdbdbbdf7fa,0xfcf6fefffbbbf3ff,0xf9efffeffbbffb7f,0xfbbe7dffdbbc77fe,0xfde7ff3ffbe0ff7a,0xebffffdffbf6ff3e,0xfdff7ffffbfe77ff,0xddfcfffffbfcdfb9,0xdebc29f78b9877be,0xef77fbfeebb7f3fb,0xfbffdfdffbffff3b,0xffbffffbcbffffff,0xd9bcf67bfbf3336e,0xda3635fdf9ca77bc,0x8bffff7eebedf77e,0xeff7bf7ffbfed33a,0xbffecffacbfdf7be,0xf6f7fffadba8df2e,0x7adc7ff9dbae73fc,0x9be6fb37fbf1f37f,0xffdf3e19dbfef3af,0xaf7edf57dbf37fbf,0xf4f77ee8dbde7b3e,0xdfbeff73ebf4ffff,0xfb9efffeebf45ffc,0x8f4bfbbaf9b8d7b9,0x5719dfdbeba17b3b,0x76fff7fafbff77bf,0xbfefffebf9f9dfff,0xea9433f4c9805fbc,0x83e7ffabd3b07b0e,0x25eedbd9cbb47338,0x35a559af5b11d34a,0xf47cb6528bd2d328,0xde977dfddbfc433f,0xdddbd6b5ebcc533e,0x6fdf9fd5e3b27b0a,0x97334c5a6ac84721,0xf72df72edbf45bbc,0xa212ff648ba8c72e,0x8bfdf37ee3b053fc,0x7c670e50e99066ba,0x8b888f4bc3b243ea,0x4196c861ca205208,0xe7f63ff0fbbef33a,0x1b06e73c8970f3ae,0x887e1653c943f20e,0xdebeddebebbcdbff,0x6cd767e3cb9c57bd,0xf38ea470c99e73be,0xef95fb7f7bd2533b,0xf4a939ac891133f,0x2947d993fbb4636c,0x76801460e0301320,0xc97c6d23db1f7b6a,0xff64fff7dba85bbf,0x947afabf8e4d529,0x2a1448d4d888d112,0x240019e1d0a41229,0x91785f70898a53aa,0x9c8683708b88517a,0x8080c96c81b04b80,0x637810068112498a,0x30000b8001004170,0x104c3001518041a1,0x814040000a000102,0x20080100000,0x8e8081181a10323,0xe53d05608b8071f0,0x8010005001100102,0x10c06c800806130,0x5150652c83a20a2c,0x121032028100130a,0x2904c04001100308,0x400002040001000,0x19ad18051120268,0x3000020002a00200,0x41000040001000,0x0,0x80140801800018,0x808002281105000,0x0,0x508c0088a001208,0x1400000800001,0x100a480008900210,0x682470200310528c,0xdab63fd7db91533e,0xc95e750b0bc10623,0xa0cac0209f00320,0x957f4da2dbbe53b4,0x52202d1061136202,0x9abe2d56f9e07b1a,0x8accf23ad1b441b3,0x4ec221348aa0fa14,0x8080d0800050062,0x0,0x482040508a001120,0x8300c3126b1c4122,0x600300001046020,0x8240028080030a,0x949056426081130f,0x900205600b983228,0x480005051020000,0x4142600a09400120,0xa5ba9d508b8a1130,0xf5b456e88b335722,0x907d2750f080c3b4,0xb12657a8d110538e,0xddc6fe96dbbc6bfc,0x398d8528bf0732c,0xc4f2176e89d863ee,0x53fefef48ba0532e,0x8e262f6f5a15536a,0xcbc5e603eae8538c,0x2292300183804112,0x4a944400bc00228,0x105c260408085382,0x42204080000064,0x153ca0081124124,0x4a461b55e0101b7a,0x33b497a259f11363,0xd03ab27561b5033a,0xeffcef73dbfd577e,0xebffff7ffbbef7ff,0xdefb7be7fb1453aa,0xcebedffbd3b4733f,0xf537fee9fbb453ef,0x6aaefc73cbb25b2b,0xfbffdbfecbc973ff,0xffefff3bfbf6d76b,0xbbdfdf5febf673ba,0xf6f7fefffbb2fffe,0x136f7d60f924dbf8,0x9ed0ef75cba873bf,0xf7f7ff3dfbd3ebfe,0xab548c0008bc0330,0x11ba1907e938439c,0x400516480b980008,0x9210075cc0340322,0x2802800100320,0xb555d40cbb05728,0x10043930102032a,0x80990d408b11030e,0x5ef46df18bdc67fc,0x20c00280001208,0x4002101001000160,0x94446061820138,0x5000202580b02302,0x4002000800001,0x8010202040000100,0x1800000004080,0x0,0x22088008c00001,0x10085002004108,0x1214000000000520,0xc967a21b8be05328,0xb26055c89c051a1,0xb6f6feedebfb7bbf,0xfd7f7f7bfbec7bfb,0xff9e3f7fcbfd737f,0xfdd0ffb0cbfa7b6f,0x6ebfdb95ebfe73b9,0x777bdfe7fbfc73bf,0x2966a522cb807779,0xfef67f7eebb5f77f,0xffffffbffbffffff,0xffffffffebffffff,0xffffffffebfff7ff,0xfd7cfff98bfd737e,0xff3e7fffebfbf3ff },
	{ 0xeb9fffe2cbc7fb7e,0xdedb3a9bc3c7d372,0x1868b3302028342,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa84cf5a83c4e30e,0xf9b80d39e1cde30a,0xf9ecff73cbc5c33a,0xfffaaffd8bdff37a,0xefdeeff68bc7d37a,0xaa5e3ddf83cdf3aa,0xdfbb6ff7e9dfeb3a,0xdf7db7fbcbddfbba,0xfffe3ff9ebd7f0fe,0xebf7fbffebcffbfa,0xdedf5e77e3cee37e,0x9b9b47c2eadee31a,0xf6fff7fecbdffb3a,0x6ffeff7bcbdffb7e,0xf3b7f7bf43d7c37a,0xfae3b73b8355d3fa,0xbe9eddb783cfe3ee,0x34237c65a1d6d31a,0xffbe3df1ebd4f3fa,0xca6ae77f81ddd13a,0xcfeffa3709c7c35a,0xb377dfdf83c7f17a,0x3497766d89d6e36a,0x99c6a763c2d3d85a,0xcfa6ffffebd7d37a,0xefcfdfbfe3cff37e,0xeb7ffdddab5ee3ea,0xb4e37feceb85fb6e,0xfbfff7facbcfd3fa,0xfebe3fff8bd7d3ea,0xebe6fb7681d5f96e,0xc85fdebf8bd4c3fe,0xbfb7dfffcbd5d31e,0xfdfeffefebdff37e,0xfbd6ffe6c3d7e3fa,0xe9f7db2ee10ff37a,0xeefabf7fcbdfc3fa,0xf7ffcffaeb55d36a,0xd5fef77fcbddf3fe,0xfbbe7deaebd7e37e,0xcbfdfffdebc7d37e,0xcecbbb7da9d5e3fa,0xff77cf7983cffb7a,0x7dfeffe8abc5e36a,0xdbfefff2ebd5d3fa,0xffbfffffebddfbba,0x6fd7ff97cbc7f37a,0xaffeffdbcbdff3fe,0xfffff6fcebc7fbfe,0xffeffffbebddd37e,0xdebc6fffebdffb3a,0xc7cfcb3bcbc5f3ea,0xbf5fdfdf8bdfeb7a,0xfdfbdffbebdff3fe,0xdffffffaebd7fb7a,0xde9e3ffeeb57f2ea,0x8fcbf73bcbd5c97a,0xdfff3fdbe3ddd3be,0xbffbeffbcbd7fb6e,0xdbbf97e3c3c5c36e,0x78f6bdf0abccd33a,0xebe7ffabebc5db7a,0xebdfbfffebddeb7a,0x9ebf5fdd01d6c36a,0xf5d7d7f9c3ddcbfe,0x63aebf6281cfd33a,0xfbd4e77ec3d5c36e,0xfe637f8b8b55f32e,0xaabfeff981d7f3fe,0xa4bb3af08287b312,0xb3bfbf70c9cfc32a,0xc8a0fffccb4ee146,0xe907ee270845837a,0x99bf9fd6c3d7c26a,0x74f6feebebcdf37a,0xfffefff9c3cff36a,0xddbd3f77abc5cb3a,0x2d8fdf028257935a,0x25b17d0c1c5934a,0xbcff7ffe8bd5cbfe,0xdd7df4714145c13a,0xc90604b081d4c220,0xedfbf7b9e954c9fa,0x2ffeff7fabd7f36a,0x76ba4feba3d6fb5a,0xfdfffef383d5937e,0x7884796281d6d24a,0xcbf6f73789cdc32a,0x64ab58408284d970,0x938805f3c3c1d32e,0x95b5fe5a81c7e3fe,0x6f8e2e31c305c12e,0xbb84c2068bd8d166,0xcc58b3828955c30a,0x1e9def5162d5c33a,0xb4fdd2df63c4c360,0x5ffa3f718ad4eb72,0xab88476e83c6bb22,0x4fde3f26c8d7d37a,0x22f693d3c3d4c3c2,0x342854e08344c042,0xc96875068057f362,0x8102724c83c4530a,0xc7deff1bc35ce2aa,0xaa044d802b0a400a,0x993f1c22e344d14a,0x92e1afd5cbd4d22a,0x12940b4c08dcc022,0xc596e7a801d6c3f2,0x224000c60244000a,0xc3125d2288c4101a,0xd2be2d1281c7db76,0xd84028460000c268,0x5bc678212b54c26a,0x2f0448000005c112,0x398051a43c8c002,0x0,0x808050000008028,0x200000008100,0x8140000240808000,0x60046108102520a,0x2240166003c48182,0x5baebc3101d4c13a,0x8104100000818000,0x48180000508008,0x11000020a008002,0x9421462481c44002,0x6000004008,0x8c2c0081804000,0x80100002008000,0x10114080002000,0x0,0x19306e428300c906,0xd68e2f6c8256ea62,0xd01081288081a004,0xa08280009448382,0xc735eff588d5b16a,0xa4241d288284f302,0x9614774c0a4cc31a,0xc9b4e66ae3c5d10a,0x40861e2008c4822a,0x0,0x41822608184c200,0x40202400c000a000,0x40031881005082,0x8643b0080087c002,0x280031100440202,0x50350220200c4012,0x17807d408a0dd22a,0x4480a2402800c02a,0xc316162aab546008,0x5dd6c28843a03a,0xa2137768a2c5d11a,0x89298862408dc200,0x4ba2272381c5d30a,0xffd6fb3ec9d7d36e,0xac5c8598b05926a,0x1ffcffcaab54d2ca,0x920424608005b102,0x5c282deecbc5d122,0xe8c2ea0e08d4c10a,0x70427b198b84fa2a,0x920005a081828306,0x40c0b73388d3c30e,0x202344001848020,0xa106aa028150600a,0x48422c0002c0802a,0x918a4c738082c82a,0xb42ca372a3dcc13e,0x38b63f31c3d1d3ea,0xfb4ec16fc9dde22a,0xf95aff5aa285d36e,0x7fbec7f9e95ce36e,0xb4b8def6cbc7fa6a,0xffb6ddf3abdfdafa,0xdbbc57ff8bd6e36a,0xdfd3be1aabc5d326,0xaf8acdd9c3c5d3be,0xd4bffefdc1cdd3ea,0xe1efdb6a83d4f84e,0xcb9eae728b47f3ca,0xcfdffffbebd7f36a,0xee5f124bcbcfe202,0x10402100848002,0xb300a66003468272,0xceb2414a69c5c22a,0x88015230005c03a,0x280e38808005a30a,0x60266c840468302,0x4020060040008200,0x91802c4009d40102,0x81a000370a868102,0x480000000500000,0x0,0x2020000008040,0x108000000004008,0x4000c02002808002,0x8000400000040002,0x0,0x6000008020,0x0,0x50842402c084a100,0x1800000008020,0x2ab803c14080c10a,0x2cf6d6e7a3d5ca0a,0xf73e677263d5c37a,0xdbbe2a76a8d0d96a,0x8044031b81508324,0xebcecdd481d4f3da,0x9cbbf7efabc7d36a,0xf3fcffef8bddfb7a,0xf6b62f564bd4c26a,0xebfffbbbabddcb3a,0xabdfffdbcbc7db6a,0x16b20fa68382f36c,0xf17df5fdc3c7e35a,0x4d7eff73abd5cbda },
	{ 0xdbffffffcecbfffc,0x6fdffefbcedb7bb4,0xffffffffcfdbfff6,0xf77776ffcfd3ffb6,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000e1805006a80,0xbabf5f538fcb7aaa,0x343676c0cdc22e70,0xffec7f67cfcb7fb4,0xfeb63ffdcf9b7ebc,0xebd7f2a7cddafe2a,0xeedfdefb4fcbffbc,0xf6d87f6b4dcb7afc,0xfdfff77fcfcbfffc,0xfffeffffcfdbfff8,0xebdfe76fcfda7fb6,0xc84b3e5f4ecb7a34,0x8a1e7b9689d35eb0,0xfbf7fff9cecbfba4,0x2ffe64724fdb7e20,0xdbf0ef2e4fdb7ba4,0xedce12abcfdbbe74,0x8fbecd6e8fcb7bb8,0xd1b3dfeac7cb3eac,0x5fbc3fe340d33ed4,0xcf1e6b764fdb3fa2,0xdf57bebbcfdbfab4,0xafffddd80f9bbfc2,0xfefff6ffcfdbff6a,0xdb3ebff64edb7e60,0xfbd5ffb34dd33ee6,0xafd7ff2fcfdb7e30,0xae7fffdfcfcb7e3c,0xf4bf67ebcfdbba60,0xbbedef7ccfd3fe2c,0xdfbe7fffc7c3fef6,0xcfffffffcfdb7fbe,0xecfffe9d47dbffb8,0x3bbf4ffbcdc9feb8,0xfdfd27ebcfdb7e76,0xf5fe3ff2cecb7fbc,0xebcff7fecfdbbef6,0xeffffe7fcfcbfffc,0xff1bdfffcfd3ff30,0xf5fefefecedb7e96,0x5bfe37f34fcbfeee,0xebdfef6bcddbfeb8,0xd9effaffcec27ef4,0xffbe8ffb4fdb7e62,0xdd7f7feecfdbffba,0xfbba3f77cfdb3e3a,0xffffffff4fdbfffe,0xcff7febfcfcbfeb6,0xffffdfdfcedbfef4,0xfcfdffefcfdb7ff6,0xff7fff7fcfd3fffe,0xf7bfffff4fdb7fb2,0xefefffba4fdb7ff6,0xebdbdfdf4fdbfeaa,0xf7ffffffcfcbfe7c,0xf9ffbf7b4fcbfebe,0xfebebf7a4fdbfeb4,0xcfdef7b8cfd3bf84,0xfd6fedec6c97eaa,0xe7bb6ffb4fdbfeba,0xf6fff7eb4fcb7f86,0xffde7ff64fdb7eb4,0xdbf7ff3b4dd3bfb0,0xeeef7ffb4eda7f7a,0xbfbdeffb4fdb7ef0,0xf7ff73bf4cdbfef6,0x5fbe0d7acfd3bbbc,0x9bfff76fcfd37eb2,0xefdf7f7fccdb7ef8,0x8f9fdfc2cfdb3e32,0xe8ff6ecacdcbebaa,0xdeb42d30cecb6aae,0xffbef7ef4fc3fffc,0x4bc3ae974cd27ea2,0x890edd5705d37be0,0xd4f676c4cfdb3b22,0x5deeff774fcbfeb0,0xdaa62fd3cdc2bef8,0xabd7ef2eced3feb0,0x2ebc5fc64ec1fa2c,0x4667d7f6cfc3ae84,0xff7dfffb4fdbfe28,0xaebeaf6d4fc37be4,0xe7d3ee7ccfc2ee08,0xafdc6f7fcdcbfeb6,0xbdfedfffcfc3ffa0,0xcd7caeda4dcb7bb0,0xb716efe1cecb7f34,0xe917f2834fda3ab4,0xea4b7a5d4ec936bc,0x604e4ca84c27820,0xb5d0066c4ec32244,0x602635f8c49bde64,0xb9ac36e4ecbfa64,0xe561a0f85ca3ec0,0xf39f4ed1c4caaf70,0x517f76e94fc22eb0,0x593abd73479bfab0,0xffa0ef654cd3b6e4,0x9fd7fe0a4edb7ee0,0xa2dcda5a4e5a3eec,0x341416e1061236a0,0x89bb6f7e4fdb3ebc,0xf63025c04ccb320c,0x82020a3046012e10,0xa3d8df59ccc27a04,0x34c054214c083620,0xe0ccb630c5c07e22,0x840087c148423320,0x6144c00000025120,0x686659d84e8abe04,0xbe0a1f1b47887e20,0x640651b106523b20,0x188c49c00c493a80,0x8151a13204c12210,0x68421e10c0807030,0x9b10ed88c4c926a0,0x8000040884802020,0x402004008081000,0xfbd4eb2e4ed87eb4,0x800930006002000,0x8008050180921200,0x66212c00cc27b00,0x2b8e77e1c6d37260,0x8210217004c06a10,0x0,0x401204410208,0x2404102008420200,0x0,0x0,0x4020040800000,0x103600c0820040,0x0,0x0,0xfeb6bf7ccfda7fa4,0x830a42002210,0x5188e408c832202,0xf6ff7ee84edbf7f6,0xd1bcf7eb4ec3fea4,0xb5102e0084c22b00,0x88402a0504020620,0x61480040802024,0x2044a444820024,0x14e0240207022aa0,0x40645041081810,0xc304a91e4ec21a20,0x641f71243022e24,0x4430004402200,0x10406000103002,0xc03aef7747c36aa2,0xc280e64a0ed2b3a8,0xeffff2374cdb7ea0,0x10044204487000,0xf4b3ddebcfcbfe90,0x170257ac1815a14,0x8c51724004332a2,0x7883dfafcecab220,0x7bfe53d9479a3efc,0x84b755c888c316e0,0xd108856246cbb2e0,0xe8d62660c5d27876,0xc357f7384593ffbe,0x8058ca430c027628,0x1982456a46c23fa0,0x885635080e022700,0x84063040c08804,0x1d02044002010,0xf402341c8c22640,0xd3eea1743c9ba10,0x44000000600,0x2ade75f04dc37aa0,0x9383c92444d0be04,0x865fa10084b73f0,0x8c9ddf784ec27624,0x5565dc6b47caae30,0xddfeedf14fd3bef0,0xffbeed7ecddb3eba,0xedc7f7aecfd2be70,0xffdedf5fccda7ab2,0xf6f7f7fa46cbfeec,0xdfefff7bceda3efc,0xc59f74724ecab2a0,0xfff7ff2f4ed37fea,0x23ef93d3cccb7a24,0xf29c55eacddb7a24,0xd96df7114f93fea2,0xdb342f304fc3afc0,0x8063e90b4c82f610,0xadedc1547c1eea8,0xa844224443200a,0x9181874244c1763c,0xa00247080889280,0xf9daf76fcfcbfeb8,0x50060000410600,0x0,0x21004604a000204,0x50000020200,0xb100020404802000,0x0,0x10001000800,0x2004001000800,0x0,0xe100510008002280,0x91900180c02020,0x41f5c160ccab260,0x58b28640074b2a10,0x9f677f6acec1ae90,0x9fb2fbed4fdbfea6,0xf7fdf5bacfdbffb0,0x9259dd3c6d23e00,0xb2f7dde04cd27efe,0x71efe760cdcbfa80,0xfbf72fee4fcb7ea0,0xe7deffff4fda3ea6,0xefffffffcfdbffb6,0x9fb3fefecfdbfafa,0xf12db7fcc582ffe2,0xfafe7debcfcbfe34 },
	{ 0xc9bfe72ee817bb9b,0x6bd77b5fea36b29f,0xb39d5db3a8523abe,0x9cff566ca8721aaf,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xebd7ff0e28723aaf,0xb42114fa850828b,0x726a47b38529283,0xd9feff63ba7239ab,0xdfbefff778763bbf,0xefe9f7ebaa12ba9f,0xfb5ff6dbfa76babf,0xffbb6ffe7a723b9b,0xfffdff7bfa763bab,0xfff6ffeffa53339f,0xe1f7ef7fba56bbbb,0xe2ff1b7dfa7232bf,0xbaaecfff8872ba9f,0x75fcf6eff8729b9b,0xfb9e6df26a523abf,0xe3d7fb1a7a563bbb,0xebcf37bbaa723bbf,0xfbbadf7fea72babf,0xd4f7f6f8ba7212af,0x97a6ff71aa723bbf,0xfb9f7f6f6a629abb,0x9dd7fe306853ba93,0xabffdfd53a52bb9b,0xdcfefff9aa56b3af,0xfbfeff7fda779bbf,0xfb30ffeee87212bf,0xffdffffeea763b8b,0xffffffdffa77bbbf,0x5fff7fefaa73baaf,0xdf6e6673ea53b33f,0xdbbeaf6eaa731bbb,0xfdffffbffa72bbbf,0xedffffdffa77bbbf,0xf7bbfff7ba72ba9f,0xf3fdbf77ba723abf,0xffd6fd78fa723b97,0xa9d7fb7eaa523bbb,0xfcfbffdbea72bb8f,0xbfdfcffbea72babb,0xfffffffefa77bbbf,0xdcb67d53a8539bbb,0xf3b5fb7ffa77b3bf,0xcfebb3f9ea733b9f,0xcfbfdf7ffa739bbf,0xfff5fff67a76bbbf,0xfebffd61aa73bb9f,0xffccf7fffa723bbb,0xdfffffbbea73bbbb,0xffffdfdffa76bbbf,0xf7fffffffa77bbbf,0xdfffff73fa72bbbf,0xfbbfff7fb877bbbf,0xefffff1ae877bb9f,0xbbff5bd9a8523bbb,0xfefe7ffbe872bbbf,0xfdfffffffa57babf,0xfffe7f7ffa77bbbf,0xfde7ff3f2a72bbbf,0xefdffeeffa731bbf,0x7ffa7dfb8a733b9f,0xfbfffffaaa533bbf,0x7edefff7e876babb,0x8bfef7b62a52b38b,0xffff3eceaa02bb8f,0xbfbdef7baa731baf,0xf6bef775ea52199b,0xf7ae3ff2e8723b89,0xfbd6fb7fea723baf,0xadfbbb1aea53baab,0xffbfdfdffa579abf,0x31f5eefce873138b,0xdd0aed635a52138b,0xffa57ffffa72baab,0xc5f321f6a52bb83,0xaafe9dcfea73bbbb,0xd49578f06873121f,0x9bfff773ea57bb8f,0xf796766ffa73babb,0xeccfff8aba7292bb,0x29ffffcaaa729a8f,0xfebe4dedaa56329f,0xdfdfff73fa739bbf,0xfeb63f7bfa53bb8b,0xcddafe65fa429a3f,0xadfeff5fb852baaf,0x87bb57f1ea72bbaf,0xf3d835f86a7299bf,0xbfaa77cdb27632af,0xabb7ffa7f8723a1f,0x49c216946a039b01,0x889a038182131a21,0x75c592fa2a52119e,0x47ba2570a8639abb,0x8a30610f8072328e,0xe3cd5a052a72308b,0x9fbacf12aa76339b,0x14131ee528533382,0xc87464208a531a83,0xf3b4e667e8729aab,0xcfde7ebfaa73338f,0x4876054862468289,0x5454748f8a471a9f,0xdfffbf77ea73bb9f,0xdfdf7f765872bb9f,0xef9efb9eaa72b39b,0x2beedd572a52bbbb,0xb4ff7fefea72b39b,0xffffef7ffa53bb9f,0xca862e712a529a97,0xc7f2e1aab8763a9b,0xe1275ed32a52babf,0xadf14e6808431b21,0x99bead68ea721abb,0x8a360ce0285192a1,0x84a11b8a850920b,0x4a00022228408201,0x8088040000421200,0x522aa0423200,0x98023030001080,0x8080c20a02503083,0xe8ca122020521216,0x4c00002021001,0x741656662a521a3b,0xd1ca7460a0569283,0x0,0x460206102002a00e,0x10000400000202,0xe4997e68b27232aa,0x20120a0000002,0x8000100000021000,0xc478c128a252800b,0x0,0xa01004a080121003,0x1000a6602802108b,0xd8162d1002039211,0x402130820020212,0x401c040108401b00,0xb6e6cce0e8421a17,0xb5fcaf6aba72389f,0xb066d6600523a0b,0x61d473aba2629a0b,0xa40636002040228a,0x100010020000080,0x82108080020201,0x24000000000a,0x20020020000200,0x4000000400000201,0x0,0x10001000400200,0x93ba4731a84233bb,0x49b4be5b0a423201,0x5f46a28722523083,0x110080420030208,0x949fdfe0ba563b9f,0x40ec250168020a9f,0x600644028508001,0xc9e6d51868631a8f,0xbb46d64268522383,0x9d327fe2aa62128b,0x80acf551a8021086,0x9d1a1fc0d8463a83,0xe0f48624c8423893,0x63cbda4028123a29,0x842a0de3a05292aa,0x8435743128123a82,0x4804042040400201,0xab75e32a48121b9b,0x4802100008021088,0x8ccb8d6242180f,0x6026025050021001,0xb2e1e716a721b9b,0x220424048a000803,0x159109188701283,0xa647f36052b207,0x5412b2692052121a,0x1bfaff72aa523bbf,0xfb96ff3faa769bbf,0x49d7be1f38723b97,0x9adedf1ee8523a8f,0xf49af663aa72babf,0xefefb770f8523a9f,0xfbb3af7e6a523b9f,0xcfffff3efa72bbaf,0xe7ff5fdbaa763b9b,0x19b34de1aa56128b,0xd340a5f42046129b,0xca94273c3a121011,0xe58ee6a4e252128b,0x4af75c91a8520a1b,0x1643622202b811,0x20040020000008,0xfa4240c8a8021082,0xebdfff7efa733bbb,0x842020030001080,0xa00500000,0x20800008001,0x448e0c408a520207,0x10082200802180a,0x4088910108423801,0x0,0x1001000080000002,0x0,0x20004040000001,0x40020080000002,0x208000010,0x2004000008101800,0xa6165610a523a8b,0xdf9ffffefa72b3bf,0xe0dcf9beca733aab,0x0,0x5c0536a2250120b,0xfdfcec514a563a1f,0xdcb62f24a052a21b,0xa5ded3b2a8769b1b,0xeefe7e1de8723b9f,0x6cbbdfe6e85312af,0xd7f4347ab8531b8b,0xdd242d6eb85292b2 },
	{ 0x89e7c73602361fd0,0xac0713cb62164a84,0x794c5a8237634c4,0xff7ff6fe6f7f1fd0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x202100260100084,0x8200100080,0x0,0x0,0xdbb53b65643e3de8,0x8ddffb8fe77f1df8,0x914455200541784,0x943fe4a160340ed0,0xfbac3408677c1348,0xdabe3f67eb760bdc,0xcdeff738ee767fd4,0xfdffffdfeb7f7ffc,0xfebb6bebe77e3ffc,0xd7ffff7feb777fd4,0xfffa6f55a5753fdc,0xeae6fbaa256f1f58,0xfef79edfe7771dfc,0xafaedff6e3745ff4,0xdebfafff6f7c1fdc,0xfbfefff3ee771fdc,0xdb9bef7f6f753dfc,0x6eebff5f677e4acc,0xef9fcfd7437f1fd4,0xf5f676ff677f1bfc,0xbff92f60297c395c,0xd3a4a3fa627e11d8,0xcadf3b9f63777fd4,0xa9bf5fd3e7757ddc,0xf4f9ffe7eb7e7fcc,0xdbfff77867765ffc,0xfffd2f7ce27c7bec,0xeff7ffbee3747fcc,0xeefeffdb677f1fdc,0xf7fb5fa3e5771cfc,0xfffcef7f6b7e3bdc,0xd6b73ff7ee7e7cf4,0xfff7ff7beb7f7fdc,0xfffffefbef7f7ffc,0xbfafdfffef7f5efc,0xf1feff69a37e7df8,0xfbfeff75ef771dcc,0x8a78fbc3a57f3cc4,0x6effbfd1a4775ddc,0x7bbecf7fe7771ac8,0xf47dfefe67771fdc,0xfdbe1d59a27639dc,0xfbf4f70fe7757fd4,0xe7fbfe7beb7e3cfc,0xffff4ffbef7f5fd0,0xf5f7bfff67763fdc,0xdbfefff1e7775ff4,0xdbffffff637e3dfc,0xdfffffffe77f3ffc,0xbffedfdfef7f3fdc,0xfeffffede37f7fcc,0xffffff7fef7f7ffc,0xfbb67fff66767ffc,0xeffffffeef7f7ffc,0xeffeffdbef7f7fdc,0xffbb5ff3e37e5f7c,0xffbefffa4b7f1f7c,0xfebe6f6ee37f1ff8,0xc7d2e7f9eb773f5c,0xeffffdcfe77e1ee4,0x153377fbe7777dfc,0xf7ffafd0a77f3b7c,0xfffe7ffee97f1dcc,0xcbf6f73bef7f1fc8,0xedc7bebf677f3aec,0x1d8f4fff6e7f3a5c,0xfffff67fe75f5efc,0xcbfe7d71ea7f1fdc,0xfbb7f77f637e6bcc,0xef7fbe9fe67e7adc,0xebbfcff3ef5f3ffc,0x76747fd6277e5fc8,0x930ff741e3721d48,0xfbb47fffe37e7f5c,0xcf67b29f6d565be8,0x8a1adbdb637703d0,0x5cdf562ee67c5bd8,0xb93eb77a67773ccc,0xf7be7f7d67775fe4,0xfddfffbbef767ffc,0xef7f5fdbef7f5fec,0x9696550ae25c19c0,0xd1ec3f5eef765e14,0x46ae3ffca9761dd4,0xefc0f33a257c5c7c,0xefdf5ed8ef7f5dfc,0xffb7fffbe37f7fe4,0xd5fdb77d6f741fd4,0xfe9e3ff5676e19d8,0xcbffef7f6f775fdc,0xdce7379869771de0,0x8a93410a81501c90,0xd6f6be7a2f761fc4,0x8407800240e0040,0xe9d4f723e0351e5c,0x46803ec26074104c,0x2d46450800745d44,0x945c7626007411dc,0xd33a376266351b44,0xda91ef2f6b763554,0xdfd7ff4a6b7e35dc,0x271edfdea3771ef4,0x50849449015e0acc,0xe1ae7fd3e3761f5c,0xfbae6c26625610dc,0xdff7f72ae3771ffc,0xaaeedddbe27f19c4,0xbeff56f2e377184c,0xd95ca6e6677419f0,0xfaae1f7f63761fec,0xcfffffffe3777efc,0x7fdfff7fef751ffc,0x87f14778e2744af4,0xd17daf00097c01c4,0x112025000140084,0xc009c11b22141094,0x2ec0081822341004,0xd08290081100880,0x8400200024440040,0x2c35c00072095c,0x808e10022720040,0x40120000250040,0xa9c0a0061580240,0x3037c66663740864,0x8f20af6043340fc0,0x2082c0441700800,0x100800000101000,0x810070308600048,0x11404000420048,0x128046025200050,0x8000000400400040,0x8020011802091040,0x0,0x0,0x40032821080048,0x4400000000101000,0x0,0x0,0xfcb5a7ea6f767f54,0xfbfcbf7bef761fe8,0x539257e266341b64,0xc99073762325184c,0xa9533ccb21361d58,0x48928b72617708c4,0x0,0x4000000200101000,0x9200420440044880,0x6000001000800,0x0,0x1400000000011000,0x998c5422207415d0,0x8c00b36860363ac8,0xced6b73eef761fdc,0x300850122101210,0x36f5f6e2e2761e5c,0x12000000240,0x5200008000141004,0x4881d02022660800,0x29ff9f58c0797b64,0xcc9259b7a2723ad8,0x456a752263641954,0x1080821400091440,0x8623eb5001361ce0,0x33c7931a02371474,0x4aa42a401110cc0,0xa14da40023051c04,0xaa844548e2241dc8,0x61346322a3541b00,0x6d4f3d01663648d0,0x321a4f42001d5bc8,0x20020000000048,0x668a0572a31c28f8,0x8304cf2460360ac0,0x2060328000761084,0x98c485a017009c0,0x19e706e86b561e10,0xca5befd0015712d0,0xd1aceb7fa17c19d0,0xcf673c73a37558d4,0x2f1fdf1263741f48,0xb4d55fe92a757dfc,0xdfeff76be37f1fdc,0xcf8f6fefe7375c7c,0xffdfffbfef7f5fdc,0xe3ffdfdb6b7e1ff4,0xf77b7f6be3773fec,0x59cc0500235608c4,0xf6363fffe77e1ed4,0xcd96b667637e0ec8,0x180112c808551080,0xa0087c3630500d0,0x0,0xb2942d6024771508,0x8142b2a024671848,0x8b424a02a41418cc,0x8001000010040,0x40700108440050,0x58a02e52836409d0,0x3200222e005400cc,0xc0000901100880,0x2d10482000204940,0x8c0040000106800,0x0,0x1004000000100400,0xa00020200101c00,0x800090000113000,0x0,0x822444000641d00,0x8b846f7ce37e4ae8,0xcbdfbba8677f3fc0,0x8a9edc40017719d4,0x5ca251a4c13409d8,0xc9676e662d5617c8,0x6ebebf70677e1fd0,0xc7fef33f25775fd0,0x6fdf3d5d63765fcc,0xa2b665f6677f3ec4,0xf3ed9fc8657619d8,0xf01e7d50247f1348 },
	{ 0xeba8fb8e1a8de7cb,0xc8cb1e5521ecc951,0xd3a496a120dd513,0x7567b2292a6dcdab,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1b7e3f9b2f29ffdb,0xa77ddfdb3f28ff5b,0xfcd6ea00888301,0xdbeeeff2bfa9dddb,0xd6bc6f7e170cc5d9,0xe7feef0f2fadc5bd,0x88865ed30221cd52,0x440100212008409,0xd9668f522b6cc75a,0xfa9eff7f1badffdd,0xedfef7fe3b2def52,0x6fdefffc3fecd7d3,0xbfbbeffb1b6ddfd7,0xdfffbf5baa2df7ff,0xdb9efd7c3fadd7da,0x89fef37f9a2c87f5,0xe2dfbfff9fadffff,0xffbe5ffbafedf7f3,0xfdf7fffe3f0cdffb,0xfbeefff91f2df5fb,0xfbdfffff3fedfff9,0xf3dbbbf91bedefdd,0xafbeeedbbaa9e7ff,0x34bf7fea3aadc7ff,0x9feb6d439a0dd7db,0xfff5ff7fbf8df7fb,0xaddebe1a1aaddbd3,0xbffedfd3bf2dffdb,0xf6bcf7f83badeffb,0xdffe7f7f3f2cfffb,0xdf937f7f3b6ce7db,0xf5d7ff5aae6cdff3,0xbbdbdfdbbf2cdffb,0xf5bbffff3fadffdb,0xff7fffffabaddfdf,0xffbf3ffe3f2dfffb,0xefdeff3ebfecffff,0x7ffffe9bab6de7ff,0xbffefffbbb6cdffb,0xf7ffffffab4dffdb,0xfffe7ffe3fedefd9,0xfbfff77f9fadfffb,0xff77ff7bbbcdd7dd,0xb7d9fee12b4df7d1,0xd7fff9fe9e0dffdb,0xfffe7debbf2defdd,0xbbdcfbffbe6df7fb,0xfffbb77fab2cd7f3,0xbeffef783bace7f9,0xfefff7ef9f6dffff,0xdffeff733fadfddb,0xffbeff7fabadfffb,0xffdffeffbf6dffff,0xffffdfdebfadfffb,0xf7fffffb3fadfffb,0xefffff7a1b8ceffd,0xffffffffbfedf7db,0xffefefba9f6dcfdb,0xabffffdf3f6cffdb,0xbfff7ffb3f2dffff,0xddbffcec1bedcfdb,0xdebffff3bfedffdf,0xcffbfbbe1f6dc799,0xebffffdf9f6dfffb,0xbcbe7fffafedffdb,0xddecb7bb1f2ddf5b,0xfbfe7ffebfedc7ff,0xebfef73ebb2ccfdb,0xeff71f7d3badefdb,0xf6dfdf633f2dc7f7,0xfff7fefe3fecd7f9,0x5e7e7d712b29c7ff,0xdb95db779b29cf5b,0xcfd7bf5e9e6de7db,0xbfddce591b2ce7fb,0xfcb43ee313cdcd53,0xcf233d521b28f598,0xfb72d77b874dcfd3,0x8bc2e6732a0ccfd9,0x8b1edfd30b2dc7d3,0xd4f6f7e83b2cffdd,0xe9fe7769ab0ccfdb,0xdfbf7f7f3fedd7db,0xefffffbfbfedffff,0xeffb7f5b0e2ce7fb,0x90a660230c5310,0x8f75ff6aab2cc7f9,0xcc082cf10f0056d9,0x81dde01633884910,0xebdfdfeb3a2dffd3,0xb7bb6dfbbaeca7df,0xd14d15630d6ce541,0x9b187d702a4ca7da,0x89e6feee334cc5df,0x6fd77e7f0b2dc79f,0xb2bfdfda1e8de797,0xb655fe72a688cd5b,0xcbee4f62bf4c6f48,0x8b16e12e0a4ca5e1,0x9883b7502e25ebd8,0xffff4d401e2cdf53,0x5573dee03a2cedc3,0xc1288c601f096814,0xc2b0e0422b004310,0xcfdbfe9a3b0dffdb,0x2f7ffbd8396cdd53,0x84d596e0060c4d1a,0x90eb6633aadcd59,0xfebe753fbb2cc5dd,0x95cbeb720e2deb98,0xaa7edddf282de5f9,0xbc3b7ded0f2cffb3,0x8d7daf7a0b08e3dd,0x4eb42fd60f6cf593,0xefffffffafecffff,0xfffffed3bfedf7db,0x83384458080823a1,0x6fd7183a6c83c9,0x12366574094c6523,0x71ddd37e3e6dedff,0x615ab4400e09e2d8,0x13904e198009c69a,0xf12c92f8310cd5db,0xc242000008010,0x80400000004800,0x121a000a002602,0xdc3d1583300cb10,0x34810ee6a64dc791,0x9006080001084450,0xfa04734e3788d5d5,0x8c080061a0c2410,0x8202000008,0x0,0x20860000008418,0x0,0x800800080010,0x2100024008008080,0xb80209121a0c81d3,0x50000008001,0x0,0x80000001000c0000,0xb5e5e40230c8115,0x95ff5ff73b6cfff9,0xc285534b0e04e7d9,0x3840264000c0190,0xebf5d72e3ecdfdf9,0x402120000000540,0x1800e80001001210,0x201000080080,0x8800040000008102,0x0,0x601080a08014088,0x800100080040,0x32421008042408,0xdf960df21f29fdda,0xdc0cd566060cc592,0x4fd77ff68b6cffdb,0x100153500a002318,0x74af5ffc3f0dcfdb,0x984224010c086011,0x694124418a82140,0x8ddf9e302d0967dc,0xe61896dc3a28859d,0x4369f7b1a6dd792,0x9720a4721a6dd7d3,0x90a80f02084c4790,0xe844b0843209c701,0x1890da920a084a99,0x54b272708b0cf49a,0x18503244088d6dc0,0xe80065620a081719,0xa0000000081,0x204393c00c24c301,0xf9b85180a4cde92,0x4026f4d8200dc587,0x68ea5dc23d0869d0,0x1940800401080150,0x409110000c8121,0x804a00288020,0x1100002824004100,0x332a300200895da,0x818cd3270608c73d,0x4fe3fe221e0c47d8,0x9a5fdd5b3b2df7dd,0xf6b9ffe80c2cc4f9,0xf9ceed63a96cf7df,0x36be2d68ba6cb7d7,0xc8b2f71a1e0cc7d8,0x8f7ddfda022cd5fb,0xf6fedcf19a0dd7fb,0xf0fdbef6220ddd1d,0xf4bd3fd10d0dd5b7,0xc7baff2e1b29f7d1,0xa9c73ec00b08b7d1,0x1c1997e9030cd779,0x2841002009200418,0x14002008000408,0x809030030a080358,0xc4ca4a1a1a21c589,0x0,0x0,0x4480a0101a208201,0x0,0x2461221020084d1,0x481202004808,0x5041000028000140,0x2010240012000408,0x8400001004000,0x803000000c8080,0x8008000000081,0x24f2b770120c85b1,0x8334044a0308c490,0xd38520300804a419,0x808010000a002008,0x8b18064a13448419,0xc43624ac0a0cd6de,0x710201a0010d4510,0xfebe3dfdbbadf7fb,0xcfcff6dc1f2dffd2,0xeddf7a7a3f8cc5d9,0xb5db5fe21eede5fb,0xd47cffd11facdffb,0x98b23d320e6c4510 },
	{ 0xc208f0234da20b62,0x85817214c904d22,0xb94cff18dfa9f62,0xf2f6a76dcdaa57e3,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4cdb3fbdddba7fee,0xefffffdfddbaff7f,0xdefbf7eedcbaffe7,0xd9ef277fcdfecf73,0xffba2f3749fa477b,0xfff7ffbfddfefffe,0xbee17554cba6de0,0x85a24d850d2a6963,0xcdefff6bcdbecfea,0xfb8fff5e49ba4fee,0xefdfebf9ddfadf4b,0xa97fdef9ddeedfef,0xb7fbdfefddbacf33,0x516ef796c5ba67f3,0x5ebaff75ddbadfe3,0xabfbfb2fdcfa5f33,0xefff7f7fddfe5fe3,0xbffeffe3ddfa5fff,0xdf7ff7ffcdfecffb,0xfefeff73ddba75fa,0xfbffff3fddbecfe2,0xedc3fe7fccfedfee,0xfffecf7fddfe5fee,0xfeffffea55fadfef,0x72b6f7f38dbadf4e,0xfbbe7f6eddfa5ffb,0x5ddebbbe5dfacfe3,0xefdfdfde5dfedf7f,0xd4fffffaddfe7fe6,0xdfeefffbddfedffb,0xfffe7f7fddbeffeb,0xefffff9eddfedfe7,0xaff7bfdf5dfadf6b,0xb7ffffe3dcfeefe7,0xffffffff5dbeff7b,0xffbf3fffddfaef7f,0xcfffff7fddde4ffb,0xeffffedf0dfeffff,0xffbffffeddbeefef,0xfffdffffddfedfff,0xffdefdffddfe7ffb,0xfbfeffffcdfadff6,0xefe73efbddfeff7b,0xbfff6fffccbedfeb,0xf5ffff7fddfeffee,0x6bee7ffaddbafdf3,0xfbfeff37ddba5ffb,0xedffff3f5dfedfff,0xfffddffbcdbedf73,0xff777eefddfe4dea,0xfbbefde35dfadfe3,0xefbfff7fddfeddeb,0xcff7febfddfedfef,0xefffffdfddfeffff,0x77bf7efe4dfe5fff,0xedcfff7f5dfe5f63,0xfbb77fff5dbe4fe7,0xcfdfffff4dbedff7,0xaffdffdbddfefffb,0xfdffffffddbe7fff,0xff7fffffcdfeffeb,0x7ebf2dfd9dfaff6f,0xc4cfff3fddbeffeb,0xfa5e7fd3d5becfe3,0xbebf6ffe5dfa7ffb,0xf3fdfffbddfecfe6,0xfffe7ff7ddbadfef,0xe9eeebaeddfa5dfa,0xcff37ef35dfedc6a,0xbbbfcfed4deacbfb,0xd1ffb6fa1df85fbb,0xf3beff7bddfa7de3,0xfbd5f77e8dfecfff,0xeaf3bff71d9aff61,0xffdfdf7bddfefdf6,0xb4ffefea51bc4f5b,0x871afff2d8beeda2,0xf884e72e4d8a4f6f,0xcff7fe5fcdbaff7b,0x234ec81a04307f4b,0x500976e309da4fe3,0xd6eff630dba7dea,0xffffffffddfeffff,0xffffffffddfeffff,0xe8ffffd31cba5fef,0xbceddb331dea546a,0xd968f773099a4933,0x85343c04b84400,0xedc9f626cdb89772,0xef5fdfcf1dfe4c7e,0x99bb4fed4dfacb72,0xd4c1d618189a0d40,0xed145d37ddbad426,0x8bf5a76c0da85972,0x6eeb0ed305fed8eb,0x911e5f7b4da24962,0xd950b659c5fed46e,0xab8e7f73ddbacf7e,0x43916b2f4d3a4d6a,0x204000c001200002,0x86cecf391cba4f43,0xf317a6ec019a4462,0x9391d7318dae467a,0xf8a16d1a0da2c140,0xcfc63e1e4c9a4fbb,0xff9fdfdf9dfedf6b,0xfcb7f7ec5dbedfe2,0xbdfcff7144ba9ff3,0xdafe7fefddbecfeb,0xed6cfebe4cda4fab,0x92f5c1a0d90472a,0x97bfdfedcdaa5f7b,0xd17d72d44cae9b67,0x128625670de869e3,0xeff7ff2e5dfe7f73,0xefffffffddfa5ffa,0xfffffffbddbecf7f,0x955cc61244ba5d2f,0x318207f0058281e2,0xc100a122012a0026,0x685516c28d9e5764,0xb2a2cf5a849a47e0,0xc02202f005381162,0x250100926040,0x88000000000000a0,0xa152b04e981a4de6,0x80904f200c8a0102,0xb2a0d0e0008c492a,0x88308f02400a63a2,0xbb81df755dea05b6,0x8100220304808160,0x0,0x4c0106181a20402,0x0,0x42000000300,0x0,0x101000000000a0,0x8000040000000801,0x93042760048a4842,0x9802194004ec4060,0x0,0xebde7e4fddfa5ffb,0x9efb6ffad1ba4f7f,0xf5ec766b8dba6d64,0x7900a9a009a8c060,0x69d0c23e45aa8b6a,0xa050300a042a0102,0xc00008004000,0x404144000928404,0x0,0x3004000408820100,0x0,0x0,0x124840000102,0x91ceaf600c8cc362,0x5380a72f0c9e5766,0xcdf6be4fccfe4d63,0xa9124fdb15ba6dea,0xf6f8fe61c4bedcfb,0x10ca252400a0932,0xc2a4223a498e812a,0x2fc7da81889ccfa2,0x88501418401065a0,0x50944ef201ba042e,0xf184245a19306d00,0x8243b4c44104f42,0x8bc27a029d920c02,0x200005240200140,0x3aa94f4bc9ba4d22,0x85a4104a05ac8022,0x202000080000400,0x480003020082000a,0x43004000080022,0x12004e80888a4500,0x160402e000180f2a,0x908552040380460,0x1000000400000201,0x20a1200040a0142,0x822010204106000,0x805061002a4026,0x3232047040900562,0xda9222268db0d7a1,0x9ffabac48cfaece2,0xab6445da85324c64,0xf0ecdc684daacdf2,0xdffeffb3ddfeffeb,0xebb63ffe9dbaf737,0x5c3fe101db8ddea,0xaa121688448841f2,0xb4ff7f690dba4fef,0xf9fc3ed655badf62,0x6c960f345c9acd60,0xcccbba3109124f53,0x12b54a0254940f20,0x61874c448124d2a,0xa000502000988420,0x200200001000100,0x8888001040280022,0xc09e321100fa664e,0x30404000904020,0x0,0x2016004001100,0x8008a2284940500,0x6388244024122,0x81405a201c8e002,0x2420020000880042,0x210011000880300,0x6002040000820800,0xc00001200000,0x81604c700b84622,0x80884420018a0102,0xa0070041000142,0x400004040000100,0x2380c10a85a84cea,0x10580604000b60,0x1026244201900603,0x100240200084302,0xfbb605e249a8cf6b,0xc966f6904dbc0712,0xefff9effddba5fe3,0x87b36fe84dfaf7af,0xd1cbf74cd9b84fa3,0xc9a6456198004d00 },
	{ 0x8200e21a02112040,0x27711ed034026746,0x958bcffbb632f4ea,0xf1f4f6582432e4e4,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x5001802040004260,0xffdffef2b4337fee,0x83de5f0ab21244ee,0xb6e5d6ec3011e7e6,0xd13e2c623222c4c4,0xd7a72debb631eeec,0xef4ffe32301366e6,0x2bccdfd3863027ee,0xba5668143276c4,0xf16f2f7e7433756a,0xfbaebe5fc413f7ec,0x8fc3e33f343262e2,0xa9b7bec33413f6c4,0xa6734f6a1633ffe8,0xddbffdffb433feee,0xff9e7ffbf6337eee,0xc83cfb8b94123fe6,0xeffffefbb433efee,0xdf3ecff3b633efe6,0x97d1366e34133ee4,0xcbf6bdf1741076ee,0xfbf4fb2f3032f6e2,0xcbc2bedf163377c6,0xebbecfffb433ffee,0xddff7eeb7613f7ee,0x76beff63b213feea,0xfb95e73e343376e6,0x8fffff3fb613f6e6,0xeffedfdbf433ffe4,0xfcfdfffe96337fee,0xdbefff7ff433fbe6,0xffbeffffb633ffee,0xdfffffff7633f7ee,0xffffffdbf633ffee,0xfdfe7febb633f6ee,0xddffffefb433f7e4,0xffbf7ffff433ffee,0xfffbfff7f633ffee,0xaffeffdff633ffee,0xbffbfffff633ffee,0xf7fff7fff633feee,0xfffefffff633ffee,0xebffffeff633ffee,0xffffffdbb633feee,0xbed5dfff3613ffee,0xfffffeff7633ffee,0xeffeff6a3633ffea,0xfbd6ff36f433f7ee,0xbffffbbf74337fe6,0x6fbfcff33633ffe6,0xfffefefb7633ffec,0xffffef73b633efec,0xcffe7f3f3633e7ee,0xfff7ffbbb63377ee,0xbfffdfdeb633ffee,0xf4fffffe3633ffee,0xfffffffff633ffee,0xffbf7ff7f633ffee,0xffdfffbff433ffee,0xafffffdff633ffee,0xffbffffff6337fee,0xf9efff7ff6337fee,0xffb6bf7ff633f7ec,0xffefdffaf633ffee,0xefffffefb4137fec,0xbeffdff736337fee,0xf57fff7ef6327fee,0xdbfe7ff3b433fee2,0xcb97fb7bb613f7e6,0x7fdfbe6ff4136fe6,0xbbbfcdfa3633ffe6,0xde7ff6f6b633f6ee,0xfef6777a7633f6e6,0xf3dffbffb632f7e6,0xeff79efef430fdee,0xbbbf4f5a7613f6e6,0x4673fef0163326c6,0x9b04bd52501152e6,0xf8a1f16e743242e6,0xcedffe9ab6337de6,0xbf5eb7da3631f7e8,0x74bcaeec34336fee,0xdde7b770a632f6e4,0xffff7ffff633ffee,0xfffffffff633ffee,0xefde5fda3433ffe6,0xb6f7d7eb34137ee0,0x851ad662342366c4,0x9e9e3ffdf412f6ee,0xcf52fe0b1433feec,0xf9ffdedf9633efe6,0xbefafff3f613f7ee,0xedffb77fb4337eee,0xdfee7d6a3433f7ea,0x8983836a401144e2,0x15c0020030004042,0x339ecff1b013e4ec,0x7655d4eb263376e0,0xcfae7f6834137fee,0xc920e7281633f9c4,0x94b1e0f06325664,0x23300610542142a0,0xf4345ef0b432d2a4,0x6198857054124eec,0xdd06af65243276c6,0x47cbdea1343177e8,0x8bbecfdab6127fc6,0x362d55ef16132ec2,0xc1e31fd2741017e6,0xfefd2f4f763067e6,0xcfc6efb7f6127fe6,0xadfdff435432e7ee,0x948774a2143264ec,0xb5bdae283422f6e4,0xfb9e3ff8f430fee8,0xa9d6f3af84136be4,0xb476362340067c2,0xf7fbcdf33032f7e6,0x9053532820122ec0,0xb4100140400062e0,0xc9c083342010f560,0xad9105114004dc0,0x549a156234326cc0,0x11de51aa1431c6a0,0x38452602600006c,0x9304a32450010460,0x888180130104402,0xb088e591010746a,0x10000fc484131220,0x5822354010126680,0x8000010000110000,0x4082781204104164,0xe802218182105060,0x0,0x100040000020004,0x0,0x8aebfeebf433f6ec,0x108000000104000,0x0,0x6840e060243262e4,0x7e9ebf67a633e7ec,0x1300002002c,0xaff73b4b3433f7ee,0xb4f36fe0361372e4,0xd369f77ff231deec,0x786516042127460,0x8912011a20006260,0x41000000204000,0x131f480214124bc0,0x2100020204002044,0x800350000136020,0xc058f130143246ee,0x4cc21000001124a0,0x480c100020002c0,0x90400620001162c0,0x9f56e4722232e6e0,0xe396f32c5433f5ee,0xeecffe8eb433efe6,0xfb9acfdbb43376ee,0x16fc74e0b413fbea,0x9010045030104462,0x4510203100136240,0xc0c29f3834323fe6,0xa9c849da3433c260,0x94750ee11412ea6a,0xc95c9250342042e0,0x3a83246e10323aec,0x120000000084,0x5020402000068,0xbb86170801242e0,0x4626f646361212e4,0x78123e02801066e0,0x40810880000060,0x10100000004040,0x280058800024ac4,0x8c34302112126466,0x8107402103346c0,0x200210110010022,0xa14a90ca040269c4,0x203540c000020cc4,0x50802c0084120444,0x20022c6000204ec0,0xd291a23e343166c2,0xcfebf6fff433ebee,0xba4007db7033f9e8,0xf4fefee6f4316fee,0x51aaf160303363e0,0x42a02426200324e4,0x26c5ab18340042ee,0x88dfd908941064e2,0x54ff77ed7613feee,0xd9ffaf6cb613e066,0xfb943fc43633476c,0xc964f7189422eae6,0x2df0a01341264e4,0xc01014000304060,0x1224272b1413e0e6,0x2aa03c4054127064,0x0,0xaff63f517433e7e6,0x7904a810412c644,0x144204040420c060,0x500000000140,0x20a400000024080,0x2142001000125004,0x1a1e8f53c012e9ec,0x0,0x8040200000010260,0x4004023400222440,0x0,0x3594650741270a0,0x401000000004040,0x1100a00000104040,0x8800832014010480,0x0,0x10000200000028,0x90080100004040,0x40a4700024105024,0x52880c0010112142,0x8102601c84300280,0x41810000003c0,0x4a08975b3411c244,0xf03af76bb43364ec,0xda362950243362e0 },
	{ 0x40000a0400101882,0xb461a595014d43a,0xbfbb8de3d8b494fb,0x432da60b0949403,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x981d33b50dc96f0,0xcbdffb6ff8bc7efb,0x271b4756500ca3c8,0xf6f5fefdf8bcf7db,0x596db76a58b43ceb,0xdd34270f70dc30f8,0x1ccc0718f09438e0,0xa29edd5fd05c97ec,0x3dbb5f6b7094b4db,0xc978bf0bd0d45dce,0xdcf4277ef084b5fa,0xc9eff7fef0dcd6ea,0x37ff12cb605c3eb8,0xbfba5ff7d8d4f6ea,0xd5edff7ef0fc7fee,0xcf327f6ef0943cff,0xe994ffaed8bc9ffa,0xd9f21ef970b4befa,0xfdbfeffbf8fcbefb,0xbdf8f66a68dc9ffe,0x69fe7df1f0d43cf8,0xeb8ce32cd034b6dc,0xc6fb7ad1f0dc96fa,0x85ffcffef0dcfb6a,0xf5f7f7eef85cb6ed,0xfbffff73f8fcfeda,0xff7e7673f81cf4ac,0x2f57ff2ef0bcb4fb,0xabffdfdaf0dcfeca,0xfcfffefbf8fcbffb,0xdffff763f8dcffe8,0xffbf6ffef8fcfffe,0xfffffffff8fcfeff,0xfbfedf5ff8dcfefe,0xbfff7ffff0dcbefa,0xf5feff6ff0fcbffa,0xfebf7ff7f8fcbcff,0xffffff7ff0fcf7fb,0xfffeffdff0fcfffb,0xbffffffff0fcffff,0xfffffffff8fcbefb,0xfffe7ffff0fcbefe,0xebf5fffff8fcfefa,0xffffbef3f09cffff,0x7ffdefeef8dcb5cb,0xf7edbe7df0fcbefb,0xffee7ff1f0fcb7fa,0xfbf7fffff8bcbfff,0xeefffffff8dcbffe,0xbfbf5ffbf094f6fb,0xfefffefff0fcffff,0xdfffaff7f8bcbffa,0xfefefffff8f4fefb,0xcfffff3df8fcffff,0xffffdfdff8dcfefe,0xf6ffffe7f8fcbffb,0xdffeff73f8dcbefb,0xff973f7ff8fcbffb,0xefdffebff0fcffe2,0xafffffdbf8dcfcfa,0xfdffffebd8bcb6fa,0xfffffff6f89cbffb,0xdebe3fff78fcf7ff,0xfbfef77ef0bcfefb,0xfffffffbf8fcfef8,0xbf9ffffff87cbcfb,0xf9ebb778d05cb6da,0xfbee7f67f0d49eeb,0xabf67336f8d4bcfb,0x69751e8250bc96d9,0xbd999ffbd8dcb3da,0xf577cefef8d4bee9,0xeab42db0705cb6f8,0xebfffb3ff07cbfee,0xeedfbe9ff854dee8,0x7bbfcffb70dc96ee,0x58ffbef5709416ba,0xd2babdf06034b5fa,0xf7a6eb7ff01c1ffb,0xeff6ff9bf0d4fefa,0x2f7ed6cfe89496fe,0xf7f6f7e8d0dcf4da,0xddffff71f0fcfefb,0xfff66f7ff0fcfffa,0xa5dffbebf0dcdfea,0xffeff5bf85cd7e3,0xfc9b5ffbf0f49eda,0x5f68f7e8509c76da,0xdaa67f6ca8a450ea,0x8955f32a709cd0a8,0xeafff9dff0dcbefb,0xdeba7fb9f09c3de2,0xb265bf79509cbe52,0xf7be3df1f8dcfff7,0xa9b7dfeff0f4b5fe,0xe9d65f54f0dcbeef,0xbfff7ffbf0bc97ea,0x963dffe0f084f5c2,0x6ffe2d60f09cb2f2,0xfbb1d36c781cb4f0,0xeaeb9e5f509cbaea,0xf7bddf61708cb6ea,0x349d1fe870dcb69a,0x9338edf5d05cb6f2,0xe2824063505c327a,0xcfc0fd13c0bcb4fa,0xebd45fd7f0dcd6ba,0x96395e64f0bc9ee4,0x194a79f9f0449428,0xfa92776e589c16fc,0xe4eeef3a70dcbefa,0xe97fdf5af01cb6fc,0xbfff7feef0fcf6ea,0xd9fdbf72d8fc3cfe,0xde34755ef8141eaa,0xadcfc606d01c9eca,0x1000360040000660,0x85b25638705cb6e3,0x50902050400026c4,0x928809001004148a,0xc308e377709498f2,0x6cd21a8778dc1cbe,0x10150d4540c01642,0xd4176662005c126a,0x48643d2040d4b038,0x8310830400c01402,0x8c3ba005084b1c0,0xa6684e3948141e88,0x30341e7ad024a0c0,0x32204561781416e8,0x1400094850040090,0xd83c0110501c88,0x9b1e1f57f894fff8,0x1000000440008000,0x401a030841000,0x8381050010149042,0x4002002080001400,0x0,0x0,0x600c80201040142a,0xde3607fdf0c495eb,0x8ccaf912d084a4c0,0x2d4a1e1e40dc36a1,0x46040007000,0x8098be10705cb436,0x35045be0f0701060,0x20410000080,0x1004020000020,0x4308cd41405c92d8,0x9054f249501080d0,0x28a02422101c0090,0x8940e120008016c8,0x8042824000143cc6,0x2800a0c040041403,0x3002906c904430c2,0xf8843d2298141428,0xe1b1ff79d0dcd67b,0x4942d810b8141b48,0xcf1bcf53f09cbee6,0x4642960240c41080,0x10225700010a5d0,0x41022d60500070c8,0xad16ea2c501cbe8a,0x68be99c0d0941c80,0x444d4e9500434c0,0x8024a67b309c56e8,0x48ac2744701424f0,0x8018532250148490,0x209580190009080,0x80220e81408010a8,0x8068766240a4a441,0x1020c4040841002,0x8800000000000240,0x7407405094a850,0x10808dc0404094c4,0x4002129050048484,0x6a3c2ce05094b7c1,0x30002704501003e0,0x2823021681c34f8,0x0,0x1000160200400a0,0x80890520d0108340,0xc90251e650941e84,0x3fd7371ad09cfeda,0xef76dfdad89cfef8,0xb4bfdfedf0dcfefb,0xb527f672d0f477d2,0xa3b0767c7014b7e8,0xd5f3ea0e5054b252,0x465800002401c0,0x36b2ffeb70dcd7ef,0xdd6ea75e70dcbfea,0x4e882f3b7014bdfe,0xe9dbf7b860bcf2e8,0x7f567a9fe0bcb6d1,0x80025830708c04a0,0x51715728509cb288,0x32143d31081c189a,0x80120b1450048002,0x8a5714465000b4e8,0x5dacb83701c8c68,0x0,0x4406020140010,0x8a10c100400480c8,0x42105a40942050,0x2180c6040549050,0x1004000040008000,0x59c012050108410,0x4000800000040000,0x200200840000,0x2e006300301891e0,0xb4b4f6a270d413e8,0xd9febff2f09c9efe,0x90200c4020441020,0xc1084000a000,0x4101001d0001000,0x24040000010,0xa1459f48408cb048,0xde9e2773f0d451e8,0x148612980581250,0x2a1634803004b4e3,0x93a24c72d084b0c8,0x919cbc045034905a,0x540e6de8509cd2e8 },
	{ 0x8000612401029062,0x485e19549140013a,0x3b88cfbaf0f9abf4,0xf5b63022d104ab58,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x240088000000,0xdffefffff1efbf78,0xcffff65bd1e7bb7e,0xa7efdfded1edb35a,0x949bfeefd16fbb6e,0x9566275ad1a02176,0x89102b289124b862,0x5840863001228a62,0x4c0f5e511166b45a,0x14af5eebd9e52268,0x8f79772bd8c9bfda,0xbebe3ff491e1bf6c,0xf9fcf7bfd967bb9e,0x6afafe9bd96bbb52,0xa7fce7f1d9e9af56,0xf7fefffcd9f9af7e,0x5fc66f7fd9feaaba,0xe9defbdfd1efbf7e,0xed5e1edff178abf6,0xffbeeffbf9efbbfe,0xd5fefafed959b3f6,0x5bee6fe3f9e9237a,0x99b0eb32d9d3ab68,0xebffffdfd9f5abfe,0x8fff4f73f9e3abf2,0xf7bdfefbd1fbbbea,0xfffefff3d9ffbf7e,0xdfbaeffff9dabbfe,0xdfebfe3bd9f7af7c,0xefffeddbf9f5bf5c,0xf4fefefff9ffbffe,0xfffffffff9efbffe,0xffbeff7ff9edbf7a,0xfff7ff3ff9ffbffa,0xabffdf5f99efbbde,0xbeffffeff9fdbbfe,0xdffefd7fd9ffbbfe,0xffbfbffff9ffbbfe,0xefffffbfd9ffbfde,0xfffffffff9ffbbde,0xbffffffff9f7bffe,0xffffbffdf9ffbffe,0x5fbefff7d9ffbf7e,0xcbf6ffffd9fdbb7e,0xffffbffff9fdbffe,0xfb9eff7998efbffe,0xffff7ffef9ffb77a,0x7ffefffb99fbaafe,0xfbffff7ff9ffbbfe,0xdfeffffff9edaf7a,0xbfffcffff9efbffe,0xf5f7fffdd9ffaffe,0xfffeff73d9efbbfe,0xffbfff6fd9fdbffe,0xdff7be1ff9efab7e,0xafffdfdbd9fdafda,0x7efffee7d965bbfe,0xffffff7bf9efbbfe,0xffb77f7cf9ffabde,0xefffff7ed9fcbf7e,0xafde5fd9f97fab7a,0xbdbffff3d9f9af78,0xff2dffffd9edbfde,0xffbf3ffff9efbf7e,0xfffff62bd9ffbffe,0x6fffffdbf9efab7a,0xbbbb6fefd1f7affe,0xd9fefffff9e9a3ee,0xfd7eeffc91f9abfe,0x8ba3f73ef9e4bb7e,0xeade3e97d9cd2b7a,0x979f5ff2d1aaab72,0xf7fefe7ff9e9bfde,0xfbbe2f6bd9ffbb7a,0xabfefb3fd9edabfa,0x4feb3e69f1ecbb68,0xffbdefffd9ffbffe,0x74a577edd1a7ba7e,0xa0f86d60c1c18b52,0xdd8eeffef1eebf7e,0xcdebf73ff16fab7e,0xbefadb5bd16b9bf0,0xd5f576f49965bb78,0xfdefff6bf17fbbee,0xfebe7ffff9eebbfe,0xffd7ee7ff9fda77a,0x6f5cffd3d1e3bb5e,0x15fffefbd9f1a1f6,0xdb707f62f1ed276e,0xfeb62f6f89e8a74a,0xc3f7fd6cf1ef2374,0xaa7ddfc39172bdf0,0xb7f3ef78d9edbb6e,0x268cb768c141ab66,0xb96efd64d1f6a2fa,0xe884d21a91b5aa4a,0x6bc21ab1918a9b82,0xf39feffbf9e7ab7a,0xd6b8f2c8d1458118,0x37ee6c7191c13b6a,0xcbd4f766d9e6bb5e,0xe043162990d70148,0xa6bfdfe9f9e7b77e,0xdcff3feff9efab7e,0x8018201051c08142,0xfdc4fffbd9eeab64,0xab7fbe13886fab58,0xebfedfd3f9e9a776,0x66ff7ce4d9ecbb4e,0x892c656a9120a34c,0x66b16450c1642248,0xcff6fea6d9f7bbfe,0x637f77dfd1f1a356,0xb477dbf199edabbe,0xe9feff6398f9227c,0x88342772c1842a1a,0xc192dbe6995b8af0,0x84f1a41d840ab34,0x111eedf499e69b1a,0x1071bf7d9155bb5e,0x500202090008110,0x8bf2f682d9e52958,0x812075088412a04,0x529b079d98c0a146,0x8026726e99e82204,0x224256080c89200,0x8310810c98c09800,0x608010099000a810,0x9c17450810208920,0x1444240801410360,0x8288284088412254,0xf018686089c00630,0x952d080c1650240,0x9398df59d9e9bada,0x702696dcd8902066,0x14e334ad8602840,0x84000000010040,0x0,0x0,0x0,0x4d0c2122d0c12046,0xf4a62fdb89cd275e,0xc182b1b29909a3f2,0xe1460c02e1a18260,0x3000470281608220,0x807cf742b14eae6a,0x2410044190400a00,0x42000841400200,0x808108420000a846,0xbba9cddbd9e5a3ee,0x14090000000,0x20080000208000,0x824442401046114a,0xc00a0010081200,0xacc54841482602,0x51040062c1202802,0x93b4be70f1aba2e2,0xfb30f7ffd9f7bb7a,0x8a467a2fc130b846,0xebdb4f9bd97dbb7a,0x5001402080640a00,0xc2052011650140,0x8abd1ec29849a9d2,0xcbd3f31ed9cd2a4a,0x69ccdbcbc109b966,0x20b852e2d1ca2742,0x47baf34af8588f50,0x6aae0f94b108b3de,0x839acb0ac964abea,0x80a100180400a00,0x202064e91002216,0x4000080200,0x4980265009402446,0x810050f011a08238,0xc0641c039029a320,0x16c1408c81510a60,0x90f41609d0a13264,0x4170261051242008,0x5981d32ad1c52758,0x10513290802c024a,0x98ce009902a2806,0x0,0xc00044000400210,0xc010c3a091192070,0xa8defe3391cdbb7a,0x9a1fdc4ad1c93066,0xf7fed7eeb967bbda,0x1f4ce61891f72a54,0xc7803365d1062258,0xa167e902d960ab62,0x206c05d8d121a14a,0x22a254aa89422d42,0xd126a7eef1caa342,0xc2a63754f1012362,0xc16ee776f1a5abd8,0x4cd2fa4fb976a34a,0x939a4dabf12bb352,0x80c4b75a904f224c,0x3020004041400400,0xe05070e290e12040,0x23511409110a820,0xc002002010020042,0x0,0x0,0x2000400000400200,0xceca3f1fd1e1aae8,0x2108802110882000,0x1410100a01000272,0x0,0x4000001000002000,0x10616000024a088,0x2f1945c3c1e42618,0xa072e8012da2ee,0x4808050040012040,0xd40c284291e40922,0x0,0x28880fd080019940,0x2880100000,0x20ce83601104812a,0x148c0140b0800a00,0x950d60000480210,0xa0a989a90402342,0x870ace37d9492646,0x170077991a4886a,0x8002204811800042 },
	{ 0x8a527aea6728f765,0x84800f4648240400,0x8460100884100,0xbd53e268096ee764,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x69ae4f73e5fe7563,0xa9e3f66f08202744,0xddcf9ebc8ffe7d27,0x9b3a8fd71f3e6f62,0x7cf6f8e25fac6746,0x53fdff79af2c7d22,0xffb3377edffce760,0xe7a6dfaacfea7467,0xed5fdbd71dba6562,0xdcb3f76df74af763,0xd5edb7da5b76ed63,0xdaa6757d0d76f761,0xe5c7ffbe5ffe7560,0xea7fbeddfbae6763,0x3fbbeff98ffce563,0xfffdfffffffeff67,0xd9666c6e432a6402,0x9fdf7379f2e7560,0xecefbffbcbfc7763,0xbffecffb8d7e7562,0xf5d4f6f11b7c7702,0xeb9e3d7117fe7567,0xeb7acb655d7e7762,0xefdbfbfffff87762,0xe7ffcfd7ebfaff62,0xf5f7bfff7ffe7763,0xfdbfff72cbfcf762,0xcbb4ff7fbf6ef767,0xdfdffe37ffeeff62,0xaeffdfdfffeaf763,0xf7fff4eecffeff67,0xfffffff3fffef762,0xfbf7ffff0ffef747,0xeff7feffffecf763,0xffdfdfdfbffe7f67,0xbfff7ffbdffcff67,0xdffaff7acbfa7f63,0xffb7ff7ffffeff63,0xeffffffffffe7767,0xefffffdffffeff67,0xbffbffffdffef767,0xfdfffffbdffeff67,0xfffe7dfffffeff63,0xebf7ff6feffef767,0xffffbffbff7eff62,0xbffeeffbefca7763,0xff7ffeffdbfeff66,0xfffefffbfffeff67,0xfbffffffdffeff67,0xefdfbfdffffe6767,0xffbfeffffffe7763,0xf5f7feeeeffe7f61,0xdfdbff73fb7c7f67,0xfbbffffedffeff63,0xdfdefebfdffe7762,0xffffffdbfffe7f63,0xbffffffedffeff63,0xbfefffffdfeef766,0xdfbe7fff5ffef767,0xcfffef3f7aecf763,0xebffdfdbdffc7f64,0xf4b77fe3cffe7f60,0xffeffffb5bf87766,0xdebfbffecffa7f67,0xcffeff3fcf6ef766,0xbfdfffcfeff8f767,0xafb77ffbeb7c7f63,0xdffcbfffdbfcf763,0xfe9e2f6f9a28f560,0x8be7ffeecb7e7f66,0xefff3fffcffef563,0xbf9ffffbdffe7f63,0xffffffffeffef767,0xfffefff3fffef767,0xfb96bf2e8b6cf763,0xc5e3be900b3c7d63,0xffbfcffbfffe7f62,0xf49dffefdbfa7765,0xa204008600520,0xdb7ceb7f0f7e7f60,0xdfd3ff3f5bfe7722,0x235a9fd98bfc7f64,0xdcfff7fbf8fc7563,0xdbfffff3dbfe7f67,0xfabcfff6ed7e6763,0xffffff9ffd7e6f67,0x6bffd3ddcb68e762,0xb4925ffec9f87f62,0xfb3ef6467bee7762,0xeeb027df1feeee65,0xc7cff3bd4b6aff63,0xcc77a2c0bcc7502,0x463df4e17f206762,0xf1e7ff7ba2fce767,0xfaeeef68ffea7762,0xcaedfbaedf7ee725,0x8675e50096e6626,0xeffccfd97f7e7763,0x90dc766a8f324760,0x5dbe2f62efea6526,0x318ce92e817a7723,0x3120165082020,0x64904d081d4a0524,0x1411502e05b80702,0xb214e9606a486722,0xd3b16e49ad4c6c62,0xc386b3a78f746742,0xaf084b420a2ce562,0xf6a6b6e01f2c6d64,0xf5ebf77bff686f66,0xfbbc673d0f7e7f23,0xef5ee70a7bbe3560,0xcfcdf1a8dc84762,0x1dbe4dbb9f6ab762,0xf96b2ffadfec7f63,0xdaffaf660fca7d62,0xc955eb39caec7566,0x69ccc1401ac4342,0x97234ff6dfa8e541,0x502422c000880160,0x7b63df0b3b82766,0x400c400650002400,0x4818160700704562,0x205c107000500,0x55b452028840a720,0x560472015026402,0x58600002002100,0x800b710100ac6523,0xcc80002404500,0x14e005ec03320720,0xb060420010a4063,0x4100008400000,0x8100000000600,0xb80a8b13c8bc6562,0xf09ce4f305baef23,0x43682e2101885703,0x7020040019084400,0x8000c0408c80000,0x60000008004000,0x2100080100,0x817026010b424160,0x8c008f404dc04140,0x80b28a224d026001,0x284417b0092a4160,0xb47c4ea3cdea7703,0x100502800604402,0x1a0465220b386362,0x81c3b88101324722,0x24c2028000082420,0x21884d5003487c20,0x1a1296e92b70fd21,0xe210296a89603460,0x8100021007000600,0x607a9280198e2720,0x27be4d40ca30a566,0x61c906e00a006743,0x912abd60d9207760,0xadabf3698f727d63,0xffffffbb7dee7f63,0xffffffdffffeff67,0x37b4ffe0e9ec7562,0x2124ae644d8e4242,0xc69009e549466401,0xedfffe9ddf786f62,0x3bde7fdfc76af541,0x553e77bab178f561,0xdbfdfe63bd6e7f67,0x1aa40f48018a4461,0x1d3930a4a684562,0x40923e10e8682520,0x4324af559727462,0x10820069a1007660,0x5112020401024161,0x88008268c0202444,0xa000000000240000,0x209a47c309084163,0x404000000880000,0x0,0xa2141a0000206541,0x40478239ca48e400,0xfbaa8ffb5bbc6724,0x180c44200487120,0x50a67543df046603,0xd800e0200000a520,0x48d7ff8a5f2c6d27,0x6fdf5ddb0fbe6f66,0xfcf77bec9b347502,0xf9ea3f6953e86f62,0xcbac7f5648f86522,0x8222020840a140,0x805e540308342760,0x808054e000304642,0xc072ac5009247702,0xda9e2e3285406762,0xef66273b5d3a6720,0x4ee5fed94f4a7562,0xf30154091ac6500,0xd04496d84be86743,0x202414480404404,0xdbddf19ccbfcfd61,0x62f11fd19bac7f60,0x6038c0800f004400,0x2004000000880000,0x896315008647001,0x2b05c1360d2c7502,0x9ce2f2d28bca6762,0x8008000402000,0x0,0x40008000002,0x529016200ac84522,0x480000000400,0x14440000402000,0x4c0400010404400,0x41b8042009a44900,0x200610001000e00,0x211001003000,0xa344e4889022400,0x5c0209004a00e700,0x0,0x301c3de9cd4efd02,0x6980f60602a4a160,0x95f7ea02bac4760,0x7aa5dfa9b6c7500,0x4426a32919005500,0x4006f4230b342e00 },
	{ 0x8110a10a10820560,0xecbbf788fb16e68,0x3410226304080d61,0x5082308004000541,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x517426fc83ba8f20,0xfbdfe47617c62f60,0x0,0x44013202001a0920,0xbfff5fd797fe6d69,0xf2bc5efe1fb66d60,0xe9fe7f7b9fff3d61,0xf6982d3c0f9c8d61,0x4fdefe5a0fb66d61,0xebdcd55396b62f70,0x3df95fff8fbeff60,0xf9f7f7fb9ffee770,0xfabc3ff696f62d60,0xdbe1f39a86b6fd68,0xffffbf5f1fde6d60,0xbe9b4ff912f6a748,0xfffff7fe17feff79,0xba3405621286a460,0xc8d7fb2e07b26d40,0xa7fb3f4b9794ad60,0x96bb47eb87de6d71,0xd6bdfa7897a7ad60,0xfff6fde2169e7d60,0xfbcdf62d97df0f60,0xeae3d3ff16ba5d60,0x2bbecff31befcf78,0xfdffffef97fe7f61,0xfffeff7797f6fd69,0xfbbeff7e8efe6d71,0xcfffffbb8ffeef61,0x7bffdfdb9fbf7f70,0xfffffee99ebfff70,0xffefff7b96b67f71,0xffb73ffe97f7ff61,0xfffffebe9fff6f79,0xefffffdf9fffef69,0xfeffdfff9fff7f69,0xfffff77a9fbfff70,0xdfbe3fff9fff7f69,0xffffffff9fffef78,0xffffffff9ffeff69,0xfeffffff9fbfff61,0xdffffffe1fdfff79,0xfffe7fff8efe6d61,0xdbefffdf9efe7f71,0xefffffdf9eff7f70,0xdfbfffff9fbeff78,0xfdfffffb9fffff79,0x7fbe7ffa1ef7ff79,0xfbffffff9fffff79,0xefffbfff9ffe7f78,0xffff5fff9ffeed61,0xf5f7feff9ffe7d78,0xdfbffff39ff6af69,0xffb7f7fe9fbfef61,0xcffffeff9fffff68,0xbfffdfdb9fffff70,0xffffffef1fbfff79,0xffffffff9efe7f78,0xffb7fffe9fbffd69,0xcdf7ffbf1eff7d68,0x8bfe5fdb9fdf2d70,0xf6bff5ef8eb76d68,0xfffefff79bb6ff79,0xffffffff97befd71,0xefdfffff9fffbf60,0x5fff7ad88ebe2d61,0xb4bb9ff79fdf6f71,0xf6fdbf6d1fbe7f68,0xf9127e3797866d60,0xebdefffa9fbe7f60,0xe9df7fdf1ffe7f70,0xffbfefef9fdfff79,0xffffffff9fffff79,0xfffefff39fffff79,0xdbbcef7f9ff3ef79,0xf5cbbff78fb7bf68,0xfebacfdb9ef7ef68,0xb59e7c688e9e4f60,0x80202c4100022c20,0xda84ffaa97bc6f48,0xcfff7a1b9f9f3f69,0x93fccfdf07fe6f61,0xf4d0fe691fb67d68,0xfba6ff710e96c578,0xff9a7f7f1e972f69,0xffdbfff89fbf7f70,0xabded7db8abe6f68,0xb6bfcdeb0fdedd60,0xfbefee6b9bb76f61,0xdebe3fff16bf7d69,0xc5cfff6f879ead61,0xab7ffedf9eff6f60,0xf7f67de70e366d70,0xf7ffb6de8fdf6f69,0xf7ce3de59eff6d60,0x8bd3e36796b76f79,0x6e5613d61f136d58,0x78ccf80128f2d60,0x7d1620a983960660,0x4692af2086322d61,0xeb50e6390ec22f60,0x4006060800104c60,0x2799cf9392c6df60,0x86b636e106d26d40,0x5ab00420108e0430,0x8aa0a62c029f4e20,0x9e4ffabf879f4d71,0x686d485005954c60,0x14b776e507be2f71,0x81787f7917bf3d60,0x6a99262c00324521,0xfdcffb6e97f76f69,0xdbfabd4b17e60768,0xb4ff5fec8c3f5d60,0x95a704c206da4a60,0xde3eaf730adf2f70,0x6cdfdf7d96977f60,0x6d7efd5a16deaf71,0x3fba7ffa82b71d60,0x756c276016963d61,0x208080c00060840,0x908c6a602324c60,0x118000100000,0x4400008000400,0x80cd030082920d00,0x1e057002928540,0xa100f4000b930c40,0x4002000004000040,0x1244b0003900420,0x420165004942900,0x0,0x0,0xc9c2940190861b60,0x2000800002000400,0xd012f69e12d72d60,0x8002004000100,0xb010010003802200,0x0,0x101000000400,0x2204e004040120,0xb762e7701e96c760,0xfeb6a7f51fff0f70,0x854efb171ad61d70,0x90dbf60ab72e28,0xf7b3e5f11eba3f60,0x9edc76be1fdf5f60,0x19d2f560169f3f40,0x8050a20000040420,0x6f4638000fb32d40,0xa79a45c313962560,0x84645461169e0760,0x7086054202090d20,0x8bced62d8ad22f68,0xc2a3a04000960f48,0x120e032000022960,0x3082062800b02e30,0xcd70fd120bf94d68,0xfebdffff9eb6ef71,0xffffff3f9fffff78,0xffffffdf9fffff79,0xbeb01ee89e9e4d70,0x1cb1fc012964d70,0xaa327fae068a1f60,0xfdafc820bda7f20,0xa70354f829e2d60,0x7cb44be81eb62761,0xd9f6af5c02ba7f60,0xf8beaf360796ae60,0xc593a20b02a66d40,0x714fded98e927920,0xb4ab07fe069a0d61,0x40b44002808e20,0x6000004100000,0x8030200201840400,0x18a016c010923920,0x31101b9608162d50,0x400220800b800000,0x67140118a0c20,0x946a23107304c60,0x60820800300140,0x2020001200122800,0x1425002006060501,0x812885a004d60160,0xc200860c02f20700,0x9c21e0918820f60,0xeb0e5f529fb6dd68,0x946762600e975f40,0xeb6fff7b1fdebd68,0xffaa2e4b06becf69,0x448a080008020c01,0x1204510a82100540,0x8811457012840d40,0x4008d20287880840,0x72362ec207b61540,0x6fccf2068bb72d60,0x63dbbec992aa1d60,0x100a3180100940,0x98bdff680c975f68,0x1204302402860740,0x8944f10612100d70,0xeb77dfde8ef66d61,0x431c4a1004964b40,0x880010800000,0x80200000200040,0x8a50100010860140,0x41c7be5886063c60,0x8006002000024140,0x40528000160120,0x488d56186126c28,0x4200430806800d40,0x0,0x2010028002000540,0x2a020002020500,0xa0360020a1460,0x8840032003802620,0xc00000840000,0x0,0x14200200900140,0x88778e588e1f3d20,0x283200100921540,0xcddff7b617dfef61,0x935fde630ea6bf48,0xb0822a6992b00d40,0x5024100004000448,0x4500480302168040 },
	{ 0x4181c0f090080724,0xcfffb40a9c9c7d80,0x46124d82c4206040,0x1088200c0100080,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x41c0900012041900,0x40184f73b3020304,0xb49057e9548e2cc3,0x0,0x0,0x2723a1396942945,0xffffffdfffbe3fff,0x34f374e0d41c3ce0,0x99f6b761d78e2fc2,0xba9c2d79f39e76c4,0xd0420f38d69434f8,0xfffcddded79647c6,0xf4f77dedf79e3f46,0xfdff7f6bf69c6ffe,0x1a8e3de5da325cec,0xd9f7df69d60e77fe,0xeb7fdfc7d83c7fe5,0x832207f0d79e3b34,0xd77fff435f3e2fd4,0xde94add7f29e7ff4,0x8adeffeaf6be2fc5,0xffd6fe3bf79c6d46,0x7f9dcff3fe247df6,0xf1fff76af7162fce,0x6fa67f419f1c7fe6,0xdb74ef4efe0c378d,0xcce3b61bf79e6fc4,0xefffdff3f7be7dfe,0xffffffefffbe7fff,0x9bb83df2d79c7dec,0xf39cff7fd79e7fe6,0xffffff1fff9e6fff,0xfbffdfdff79e6ff6,0xf4fff6eafe9e77ee,0xdffd2772f79e7edf,0xfbaefffffe9c7fe4,0xffcfffbbf79e3fff,0xefffffdff7be7fef,0xfffeffffffbe7fff,0x7bfeffffdfbe7fff,0xfffebfffffbe7fff,0xeffeffffff9e7fee,0xffffffdfff9e7ffe,0xf7fffffff7be7ffe,0xfdfffffff7be7fff,0xffde7ff6ffbe7f6f,0xfbffffbfd6be7fe6,0xefdfff5ff7be7ffe,0xdfbefffbffbe77ee,0xfffff7fbffbe7fef,0xfffe7ff3ffbe7fff,0xfbffef7fffbe7fff,0xee5ebeffffbe3fee,0x7fbf7ffdd73c7fe6,0xfdffffffff9e7fff,0xfffffff2ffbe7fff,0xffffef7ff7be6fff,0xdffffebfffbe7fef,0xbfffffdbffbc7fef,0xfcbffffff7be776f,0xfbfffffbf7be7ffe,0xffe67ffedfae77ed,0xefd7ffbedfbe67e6,0xbfffdfdadf9c37ff,0xfdb77fe7df947df7,0xf36ffffbdfae777e,0xffbeff7ffebe7ffe,0xfbd7fffff79e7ff6,0x8fff7ffbd7be7ff4,0x9dbb5ff5d39e75dc,0xd2ed7ffc57b83f64,0xffbe6f7ffe9a2d67,0xe1f7f3eefebe7ffe,0x3fd73fdbffbc7fde,0xffbdfffbfebe77ee,0xffffffffffbe7fff,0xfffefffbff9e7fff,0xbbf5ff3edf9c7fc7,0xcffffedbd71e7fef,0xbbbeffd3d79e7df6,0xbabd57c2d7a463e4,0x92bebd10bf1c2f44,0xf9fce3fed79e3fe4,0xeef6bffcf71e7fa5,0x6fdfdfd3f21c6555,0xf4f55e7ed6be376d,0xd5beff62d6966fc4,0xffb677fefe0e77ee,0xffdfff9ef68e7fce,0x6f5ff5d9d71c15de,0x8efedffbff3e3de7,0xd37ee778f61e2d9e,0xffbe37ffd79e79b6,0xcfc7ff3effbe2ff6,0x6cdcde89f60c3da4,0xf6bbfffbd7be7def,0xfd64f7fbd6be6fe7,0x6fbebff2d6ae2dfc,0xe9ffff3ef7be6ffc,0x4fc71e50d69c3780,0xcdbefefcf39637e4,0xc477becddf0636c5,0x65dc3531e59c6580,0x8080730ef6080d16,0x8d823e85571e04c4,0x4e975d587f9e2464,0x79b3c6e0f2943502,0x2110857056281744,0xf7b4b36ef60e2fc6,0x7bff9f17ff1e5fa0,0xab9cdddefe9e77ee,0x521e76a056802946,0x69627763d6063fc4,0xd696ebe6f3a625c7,0xcfdffffed79e7fef,0x6fff3fdbfe9e7d84,0x28b05fe5d61e3d44,0xd06b09e0f484664c,0xced92b1195180c4c,0xcdf2722cd6be2bc4,0x25836a1856063d14,0x741adca3f4942b30,0xdaf6bd76da9c7fee,0xdf9e3af7d49a74e2,0x9048622c60c01a4,0xc04b10409a042d02,0x2100630820400a0,0x3d5490609204006c,0x20004010000200,0x42942a029202402e,0x1680020034000240,0x0,0x6404d0010202220d,0x58c890092001506,0x0,0x200082000000,0x1800106c2000090,0xfef7feffffbe7fff,0x4565604ad20c1194,0xce10020654040282,0x108040000080,0x0,0x80040000080400,0x100086044802000,0xdfb63f7ed7bc7d7d,0xcb624b09c3126704,0xe9775bd6f5b43d60,0xb7bb65eaff2e3ee6,0x11c1540c901c2ca1,0x40200000802000,0x802f800540e0240,0x6110040444802204,0x91144f01548c1564,0x14c860802023c00,0x43fc7d71d71c6bd4,0xd9f0ff3ed79c7b34,0xccd9dfc7d6862b94,0x21044f58d6184784,0xc0140044220040,0x80272326182405,0xfbb8ffdf569c3f75,0xdfffffbfff9e7fff,0xffffdfdfff9e7ffe,0xd0944ce074823500,0x5900496156043202,0xe8946674f71e2564,0xf907c95ad6be57ec,0xaf65d9ad63475c6,0x9cb7dfeaf7983e24,0xd4ccb6d9d68e3ff4,0xb7b7355a77943dc4,0x89c5f973821c7f4d,0x834fc84013140184,0x8420141091001420,0x404500040082400,0x92160810502000e0,0xc8a0c328b20625a4,0x908002004c000400,0x48044000000,0x0,0x400200a082000104,0x3546510d2000552,0xa102044802000,0x0,0x4200200040,0x6100002000004,0x1080081012000006,0x40cf3e008b902c01,0xbfffffdfd7be7ffe,0xb2b4f668de863d76,0xdb76f772df1c3eec,0xdf7e6fd7f2342d66,0x8f468a00c28e0542,0x81e5c0380bc6144,0x2080000040080,0x13a0d658d0b42544,0xab2058000704,0x1503588c7020404,0x4400149c09e0040,0x20406000040080,0xe3a82f32fbb44f4e,0x6fe6f7e19aa4276c,0xc800800002001404,0xed635fcadb9e7f24,0xc338cf53f63c2cc5,0xc00040020000,0x2008140000081504,0xa155832ed6903d04,0xcc623608ba002d04,0x900485030040480,0x4080942bd20040ec,0x202c4120420e3404,0x8000004002000004,0x0,0x18480090100800,0x8401002010001400,0x100100002020000,0x34000244c1802010,0x2480410810000160,0x0,0x94614822b2141624,0x91980000400c00c8,0xca886f16d2946116,0x8208f00073120404,0x3093250db142504,0x1431f21093146b2c,0x52658570c7141180,0x74220560d6122540 },
	{ 0xcbb1d720a380673a,0x4910408000020c,0x880c15201126220,0x8600c0002000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x88000080000010,0x0,0xa352129b8ba0666b,0xa79e8fd4c390271b,0x0,0x0,0x0,0x5bcb7ebfabbc6f7a,0xffffdfdfabb473ef,0x94975dea89c262cb,0x3228fd728981231b,0xda347ebeab8064e2,0xdc4ff61e0bd142e3,0xefdfffdfcbfe67ff,0x35feffedaab83777,0xfb6feff0ebd873b7,0xdeb63f7fc1b06fea,0xcd87b35a83bc669b,0xefdbffdfc3b46ffb,0x86bbfefbabfd67ab,0xdfbfbfeecbfb6ffb,0xdee677ffa9d977fb,0x8bf5fbf7ebdd7efb,0x6bc3dfffcbb86f6b,0xcfaedffbcb9277bd,0xfffceef9ebf973f7,0x4f3e7d62cbf762bf,0xfb8bf37e8bfe277b,0xcfdabeba8bdc63fb,0xd6ffcd7e8bfd673e,0xffffffefebff7fff,0xfffefdf3cbbd7bff,0xffbfffffabff737f,0x4fffff3febfc7fff,0xbbffdfdf8bfd7bff,0xfeffffeeebfc7fbb,0x9bf5f7f24bfc7bba,0xfbd6fffdcbda6bdf,0xdfffffbfcbfd77fb,0xfbfedfdfabff77ff,0xfffffffbebff7ffb,0xffffffefebff6fff,0xdfb7bfffabf673fb,0xefffffff8bfa6ffb,0xffffffffebfd7fff,0xbfffeffbcbf46fef,0xf7fdf7ffabf76fff,0xfffe7ff7abfc7fff,0xfbfff37febfd7ffb,0xefdfbfffcbfe6fff,0xffffefffebfb7fef,0xfdfff7fbcbfe63fb,0xffbe7f7babff7fbf,0xfbfbffefabbf7ffb,0xeffbff5bcbfd776f,0xffffcf7b8bbe7f7b,0xfdffffffebfc77ff,0xffffff73abff67fb,0xfbb6ffffebdd6ffb,0xfddfffbfcbff7fff,0xffffdfdecbff7fff,0xf5bfffffabfc67fb,0xdfefffd3ebde77fb,0xdffe7fffcbfd7ffb,0xffdfffbeebdf6bff,0xefdffe9babb07afe,0xb7f7ffef8bbb77ef,0xffefffffebfe777f,0xfffebffecbfe7f7f,0xf7fffff28bf867fb,0xefffbfdfabfd677f,0x37f3cffbe9cc6f3f,0xf3fdb77a8bfd679b,0x1f4c2d2509e466b8,0xdbddff3babbc7b7b,0xeefe5ff7abec677b,0xbffeeffbebfa679f,0xf5fffffeebbb6ffb,0xebfeff7acbf56fff,0xe9bce37bcbf877b7,0xbfc3ff7aabbc6ffd,0xaf3f7ff28bbd6ff3,0xf1933fe2ebf96efb,0xdeba95728b9a63eb,0xff9e6b5eabfc77fb,0xcfd7fe3fcbf877fb,0xbfdfdddbabff6efb,0xf6fedefdcbf573b2,0xffec77f08bf676ba,0xdffe7ffecbbf67ef,0xedffff16ebf977eb,0xef9cdf5b89bc63ef,0xbafa7fe2abf86b7f,0xfdfef7728bbd77fb,0xfeffbffdcbff77fb,0xcfeef3ff8bb873ff,0x2fdfff918bb0677b,0xb7ba5f76cbb86ffb,0xffffbffb8bbe67ff,0xef9eef738bf963fe,0xebf7f77fcbd977fa,0x6df75eb8abdc6bbf,0xf1ce7e1ab950312,0xbd55ae6c8b98637b,0xd7be3f704b9d6bb7,0xc2d0e77e0bb56697,0x8fdb1759c9d07213,0xeab6cfe98bf87fed,0x20805040a8880010,0x351a7530a9a242af,0xf995eb1e88b9727b,0xcd4eb621c9ec3239,0x1fdfff5bebbd6fdb,0xc8bdd4a98bb267b9,0xb1eeb76acb9967f3,0xb420674ccb9c2a43,0xdfcfffffebbc67fb,0xcadedfdecb7e6efb,0x127e5feae9b8763c,0xb1405648e9922203,0xcd208a22019003a0,0xf156d32d8bb867bc,0x2b5d9d50cb146768,0x4f3a4feac99262cf,0xd93cf67bcbf367ab,0x28c3c40898c4218,0x8990f694890866ca,0xe2801fd989f26369,0xbec88dcb08c063e8,0x4c54130809042322,0x42080c1080002460,0xaa04230f838c2212,0xaec91a0089881302,0xc171e18c190225b,0x480164081002028,0x8000000000000208,0x4004100200101200,0x0,0x2f08c0880290220a,0xfef7feefebfd7ffb,0x177b360abc81310,0x120002800000,0x881900002102201,0x8003400a400000,0x20b036189882042,0x412090000180012a,0xf89caf6589b266fb,0xc7d4d5b3abf0776f,0x5435e188ba067e2,0x1a2000c82802400,0x8108044881004206,0x10a42d40880c6388,0x8012b90282902710,0x40161888192612,0x2080c0188500240,0x5116dccaaaa067a1,0x585ce581abd02619,0xc9f4cf3eabdc43f0,0xc65fb61b89147f32,0x833accf1a19463aa,0xa4fe77688bd8437a,0x8001a05003016002,0xc79635edabd067bf,0xeed6763fcb8b67f3,0xbff7dfdfabfe67ff,0x100004600b100000,0xc838a202819042b2,0xcbc7fbf28bb9677b,0xe047ee32c9b1720a,0x6bffdfceabbc637b,0xa5c61563819d2f92,0xdbfebf48abdf63ef,0x8c2044809044200,0xc1d4d72ca95a7602,0x26a304803ac0272,0x462014003920002,0x8000300a01000060,0xab0245640bb0522a,0x8a81134301b22322,0x50001280800010,0x0,0x100100002400,0x1405f080a01269,0xbb04e7298335774b,0x2100080000010,0x12c000009080008,0x24014081020008,0x90090560030040a0,0x4582e1488b966631,0x445bda3981a96b9a,0xabffdfdfebd56f73,0xd4b07cf98bba225b,0xd7f7f773cbd92337,0xfebc3f5fabfc7b9b,0xe9c6f62be382760b,0x21000000280,0x2a002000202120,0x816407208a806385,0xc600206009604240,0x444c609409984270,0x1100000202000,0x2020000001000020,0xd3ea6f62ab9c2733,0x4b96f5e48bde7699,0xa144d28c0bc841f2,0x2f7717d0abba7f3b,0x9f3fd77389f46b0a,0x1c56000b40420a,0x200200000000120,0x5b0463030b90260b,0xa20290e8a390270a,0x28ac4f62a9107369,0x201022e080002051,0x294a75600b100228,0x0,0x4100400442802000,0x20c85410be02208,0x0,0x800a240000890303,0x81002a0080006080,0x6000020809004000,0x0,0x2000004000004080,0x2008426080800041,0xfe382f70ab80641b,0xcce4d3a709f06231,0xefffbfdb8bfb737b,0x35114b12aad1429a,0x712cbff08bdb66bb,0x4bce3d430b946af8 },
	{ 0x10ad2008164c320,0x4682200c01540098,0xf29cc5b2d034513e,0x8400d01080400942,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4a8a1205c8d4c30e,0x210060080000124,0x4200000100,0x831a6ff009844212,0x8bbe377df9dc439a,0xd9d6fa5fabf4db3e,0xfbffffdfe3fddffe,0xbff5dee9d9f5d77e,0xc94822420144420a,0xfbbf6d78d9d4d92e,0xafefff4a73ccd72a,0xbbdfdedbd97ddbfe,0xfcfe7ffdcbb5cb3e,0xfffcfff8f9f5cffa,0xfabc7e6efb2cd38a,0xedc6ff7ffbfcd73e,0xaffefff5f9f4d5fe,0xbfaaffdee3fddbee,0xf37ff7f7dbddd7ba,0xffb6aff7f3dddfae,0xe3b9f3a2e9e5d53a,0xe9463e084bc4c03a,0xfbbefffbe375d7be,0xd4f6d7ebfbfdd7ea,0xf9fefffbdb55d77a,0xbbf4ef7edbddc3fe,0xeffbfad3f9ddd32e,0xffffc7fed3d5d97e,0xf5f7f6fbfbf4d7be,0xd7beaf73f3fcdffe,0xfdbdff7cebf5df3a,0xdbfffe3febfdd7fe,0x3fffdfdadb7ddb1e,0xf5fcf6fffbfdd3fa,0xff7fb777dbf5d37a,0xfb95efffe3d5d73e,0xefffffbefbf4dbfa,0xffffffdffbfddbbe,0xfffffffffbfddfbe,0xffeffffffbfdd7fe,0xfe9f2ffffbfddb7e,0xedfffffffbedd76e,0xeffffffffbdddfbe,0xbffafffffbfddf7e,0xfffffffbebfcdfbe,0xfffefff7fbf5df3e,0xfbf7fffefbfdd3fe,0xeffffffffbfddf7e,0xf79eeff6e175d5fe,0xf7fffeffcbfcdf7e,0xfffefff3f9fddf2e,0xfbf5eb0efbd5d37e,0xcad3bffff3f4dbfe,0xbfefcfffdbf5db7e,0xfcfffffffbfddbfa,0xffbfff70f9f5df3e,0xfbbeeffffbfcd7fe,0xeff7ff77fbddd7fe,0xfbdfffdffbedd7be,0x7efffefafbedd3be,0xfbee7f7bf37cdf3e,0xffbeffffdbf5dffe,0xefd7ff3edb7cdd3e,0xeb9effdffbfddffa,0xfdf77fe7fb7ddffe,0xfffffffefbfddffe,0xfebe7ffffbfddffe,0xdff7f7afe3d5db3a,0x495eff4fe3d5cb3e,0xffbbffb7ebdcc3fa,0xd1fffeedcbfdc1ae,0xce9e6d36ebc5972a,0x8bfcfb2feb54db3a,0x7fdfff7bfbf5d35a,0xffbecfffebf4dbde,0xf5fffffcfbfddf3a,0xdffe7ffbfbfdcbbe,0xfbc7fd7fd9edd77e,0xe4ebffebeb4dd3be,0x2b0ac75a0b55c12a,0x149e166d6344510a,0xb436a531484cd73a,0xdba0698c8b04c72e,0xcfcbf61bfbf5df7a,0x2bffd3dee3e5ddaa,0xd4d5bff8e3d4d37a,0xbbf6476afbe5d5fe,0xefbc6fffe314dbba,0xefe6ff3bf9fcd32e,0xaffeff53dbddd73e,0xb4ff776deb75db7e,0xda79fffafbddd33a,0xfeb6af7fe9fdd77e,0xc9ebff3febfdd5be,0x6ffebedb81fdd77e,0xdeb37b3acb4dd9aa,0xf9fdf5fafbf5d962,0x7fbc3d7cc3d5c77e,0xebfffffffbfddf7e,0xeede17dde97cd1ea,0x21bc4472d360c12e,0x77763ae9d9d4d52a,0x12a5f605058510a,0xa9f4e776db74956e,0x2ec39fe3e174d12a,0x80a54e52c1e4d23a,0x70be44c88b44d56e,0xcab79f5282cdd32a,0x3a28ef3c61d5c38a,0x8dc3b8a0e344d13e,0x6f7cc19b0bccc2a2,0xf0f4b6dfebddd1fe,0xdb6f7f7bfbfdd3fe,0x72902e145154c19a,0xdfdfff9ffbfddbba,0xebffffdbfbfdd3da,0xb4b17cebf855d002,0xa920065ac224591a,0x8c34097ef954da06,0xa9c4d76be374535a,0x4a3e1f4481541336,0x978a6f5188e5c83a,0xf7ecdf7ef3dcdb7e,0x5fae3dfcfb4dc37e,0x8372f3a88955c326,0x27181cc0c950d182,0xcb1cacebf374512e,0x84641642d8441122,0x5404000080009008,0x10201000000010a,0x431790c0a0d908,0x2104400000804102,0x0,0x1000051082001002,0x20020000040020,0x88e4c00003d4810a,0x2004000000420,0x843046e2a3145528,0x102004602004402a,0xd5066a4d0155833a,0x8210220201440428,0x480000000120,0x34200d60c854180a,0xc228a700c944ca8a,0xc8be3f6e6344c73a,0x4fdbf67ae9d4cffa,0xe5575edf6bd5c26a,0x86a01f61414493e2,0x92e47464c025d12a,0x9c0437e009440392,0x8016066881441212,0x28c0924109d00302,0x1aa084000800411a,0x77241258e1c4c13e,0x42c046260d44110,0xb444b2380d4d102,0x6cebba17c3d4d39e,0x92014ee3d14483f6,0xfaf57eebdbfcdb7e,0x9bee7d60f95dd10e,0xdb8e2932f125d126,0xdbe7f6b7c9f4db3e,0xad7f59dff384dfae,0x9000c6684104d302,0x5b70c7625380c31a,0xcfb5677ff1b5d73a,0xedcfdffbfbf4d3ba,0xeadccfdfeac4d17e,0x8193ae94b0c410a,0x8908266b01d4c280,0x36923104590005aa,0xcc832047080c028,0x61f91ed1db4497ba,0x23a2c501c1049532,0x18341a428041d10a,0x2004002060000102,0x8210000200804108,0x0,0x0,0x900200440004c20a,0x5f366e60ea44c72e,0xcbd2c30ef815c13a,0x80cbf8186855c12a,0x4402480201000,0x2000046001000320,0x100850000804120,0x891061288081cc24,0x437ed619f154d24a,0xdffaefdae95cd7be,0x509cf6f8f3548376,0xdbebe543d2e4d362,0xf9280d50d064c128,0xd04a7040304c22a,0x800300101040100,0xc1400a00004c428,0x94fa2f6259c5952a,0xca022dc28244c902,0x5056d28281041320,0x94e07416200d022,0xd4ca4de0c154c33e,0x0,0x2af66d5009ccc760,0x22822c0a04108,0x8eb7ff50cbfdd3be,0x84156222c060110a,0x183300e0d2040626,0x6aa66c627145c208,0x9b0200044050510a,0x842300000000528,0x28a420080040922,0x447190200054030a,0x12a104000c44048,0x0,0x0,0x200020000000120,0x5014020201404048,0x8000200001004000,0x8800000000400002,0x609080243044200,0x37882510244c21a,0x8140000000,0xc1ac2663f0545248,0xc6862574d1e447b2,0xcfeff3abf9d5dbae,0xefffffdbfbfdd9fe,0x31a4d3ee1d4532a,0xffffaffaebfdd37e,0xff76fffbfbdcd3be },
	{ 0x8000000200005000,0x6fc97e39c282b7e2,0x284442c280b120,0xc00006c040422c84,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x422211c240e0a6,0x0,0x0,0xa336a7208ad6f3f2,0xfbe2eb7ec8dffe42,0xcfffff9fcad7faf6,0xefffdf5fcadfff66,0xd6f77ce4cacff2f2,0x8d6fef73cad2f7e6,0x5a90af6ac082f2e0,0x57c6b7be4ac7a662,0x7b5fd75dcadebef4,0xbd9f5ffbcac7fbe6,0xfbffefffc2d7fbf6,0x9eb4aff4cad2bba4,0xeff7fa2ecadbfbe0,0xede6fe96cacab3b0,0xbffd5fe7cad3f3e0,0xcd7fa7294a96f3e2,0xfe963f7fcadff7b6,0x89c46605c00ad520,0x4b7c5e564ac38f66,0xbebffdd3c896f762,0xfdb6feefcadfffe6,0xefbe3d7bcac7fff2,0xd314fb36ca96f622,0xefff97eecacffba6,0x7fbfdffecadffff6,0xd9f5fffacadfbfa2,0xf7bae7704adbbef0,0xdbbaf76acad3fb74,0x5fffffa7cad6ff62,0xe7dedfdfcacffff6,0xd6b7d6f9cacff376,0xfdfff779cacffbf2,0xfbf6fff7cadebfe4,0xcfffffffcadffff6,0xffffffdfcadffff6,0xbfbfdfffcacffbf6,0xfbfeffffcacfffe6,0xffbfbfffcadffff6,0xffffffffcadffff6,0xffffff7fcadffff6,0xbfffffffcadffff6,0xf7fffffbcadffbf6,0xffdeffffcadffbe6,0xebfffbffcadfffe6,0xefffffffcadfffb6,0xbfbfffffcadffbf6,0xdfffffffcacffff6,0xfffe7df3cadfff76,0xfbdfff6fcadffbf2,0xcff3ff78cadfbb22,0xf3becff1cac7bfe6,0xfdfff7fac2cbffb4,0xdbbfff77cad7fbf4,0xffffffffcad7fff6,0xeffffeffcadffff6,0xbfffffdfcadfff72,0xfcffffeacad7fff6,0xffffffffcadffff6,0xfe9c7ffecad7f3f6,0xffffff3bc8cfff76,0xefffdfdfcadffff6,0xffffd7ebcacffff2,0xffffb7ffcadffff6,0xffbe7fffcad7ffe6,0xcff6ffffc8dfffe6,0xefdfffdfcad7fff6,0x1bbfefffc8dfffe6,0xdfb4ae6e42cffbf6,0x6ffa7df1cad3fb72,0xcaf2fbeec0db97e4,0xef57bfffcad7fef6,0xff3fdff6c2cfbfe6,0xf5fff6f9c2defff6,0x47b62ff1ca4ef7e6,0xfbfcff77c8dfffe6,0xeafa3f5ecad7bb76,0xadbfd77a8acff3e2,0x553ed7e888c6fba2,0xbd76ad624ac3ab72,0xeb16ff1dca4bfa66,0xfffefe3fcac7ffe6,0xbe9effd7cad2ff76,0xf6f4f6facac3fff2,0xfdee7779cacfff36,0xdbfe2d7ec2cefde2,0xcfc7ffabcac7d774,0x3f7dcbdacac6eb26,0xfdbedfebcadffff2,0x552fffefcacbff60,0xeebeafffcacff3f6,0xedfedf384ad7fbb6,0x4ffe7a41c857f736,0xaff1defccacbffa6,0xf97dae6bcad6ff66,0xfaaeade9cadff7a2,0xebffffffcadffff6,0xceff5e75c2d2fee4,0x879cef8bcacebef6,0x711c71f6cac1f7a0,0xe73e3dfacacbfb20,0xdb9ac7040206f362,0x8d6bb201c8c6dae2,0xfb79d3bc0c6f962,0xf5e436eecadefa72,0xd7baff73cadffef6,0xde3e7f2aca4fbe76,0x9c69e10c09297e0,0x31c5d504a86a612,0xb4b67eeeca8af736,0x8fb6affcc287df36,0xabb50626c0423830,0xffdfffbfcacffff6,0xefffbfdfcadffff6,0xb4ba5dfa8081fb44,0x12410164c280d020,0x803cae794a1e93a2,0x80dde2244287f146,0x8c771a40ca431a64,0x7005779884ffe72,0x95fcb5e8c2d1f7b4,0xfbbe6f7bcadffbe6,0x8973fbf2cadbfff2,0x7ff6be3cc253fea0,0x148c4b3b4243a3d4,0x5100765fc2cb9324,0x8211000000a280,0x19800a0d40d41206,0x64e304688c4d140,0xaa380da04a03a3a4,0x80121480428a3002,0x0,0x5006109048412e24,0x802020400848002,0xa29018400201d200,0x2a380768480e1666,0xc808077248c68270,0x911c1964cacab310,0x8809220800220a0,0x50004448001020,0x4221100c080c200,0x80786770ca405224,0xde122e78424ab2b2,0xfab9f72ecac5fb32,0x6777a09c2c4a2a4,0x1d0203b980c0baa0,0x240000008020,0xdcb65b4e8a57bb62,0x4081b2008080a120,0x40161420c092b126,0x8018010000410800,0x9441a71008859122,0x22a02dc04ac28a30,0x4200f10382018224,0xe6ca1f6bc2dff722,0x9b8a49088087ba22,0x51ffd67fc2c2efa4,0xdfaf7f63cadbfff6,0xc08244880c57002,0xcdcbaab4c0c6eb74,0xa201120042018400,0x4a6d0404802f2e0,0x92ec2739885a9a72,0x8bdc13e9820af766,0xcff6df1cc086aff6,0xa59dd94242c2ae70,0x94b1767942c2bff0,0x586005d2c842e220,0x26942d724a8ab462,0xc7f07b9988cdb2b0,0x2a4e3293020fb062,0x819a42034a423610,0xa9f8bb50ca86b1a6,0x80021900400090a0,0xb92026100022282,0x4d0a1988ca02f3a0,0x210c00000121020,0x90b0064008c8b080,0x69f276428a4ebe72,0x41d292008202a386,0x800093274a41b030,0x130040000808020,0x31000668c0001614,0x0,0x8b08600208181600,0xc152da3d82c39b60,0xa21e5f93ca57bb26,0xd6e77ee5ca53bfe0,0x934f2773c88bf0e0,0xe08034618085da02,0x8080600100038060,0x200000180004000,0x808204400041a200,0x404000008020,0x4800002080,0x4046000848020,0xc84ec404080a4a4,0x488264848c48500,0x1083546040428130,0xf2965d608243f3a0,0x800408208000c040,0x8a44800820292e0,0x5a1ccdd0c047fb72,0x90a0506040422286,0x290204002048022,0x2000400008400000,0x84d392cacac7f050,0x2c140101010,0x4e416c048408b20,0x0,0x1004000002000200,0x0,0x0,0x80040000020200,0x130020000028202,0x820425028000e120,0x880008040000a220,0xa5a0ac3c882b602,0x34e35372ca43da60,0xc1508c1042437230,0xda902e22cac2f292,0xc8c7f97ac0d2bbb4,0xe8fffad6c252faf6,0x973a55304acab242,0xffffbffbcad7fff6,0xdbde7c72c28ffde6 },
	{ 0x8b93a37c8cf23f83,0xa870114a07c1aa02,0x830954104865e00,0x95e87f4c0cc31a29,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x66000040022601,0x0,0x0,0xde3dae71cec27f93,0xb00733280871221,0x1af27e178dda7f81,0xaffecfdfeddffb93,0xb4b6feeeeddbff81,0xfbfe5ffbcde37f93,0xd7b22f518e83eb9b,0x6d56ffb4ae037e1b,0xabdd5d9b8cb6db89,0xf4feffebcdd7fb9b,0xf9cdbf72accb3f93,0xff96af6acef3be99,0xeff6fb9acfd3bfb3,0xfe7fded6cfd3beb1,0xaf0adf768cd35bb9,0xddfcbe7ecfc3ff99,0xfff63f7daea77fbb,0x89d7f31f8ce36e9b,0x395f7ed3ed873fb3,0xe13beff34dd2bf91,0xd77bffe98febafa3,0x7bbe5df2ecd3fbb1,0xfbdcff36cfc27eb1,0x2f7bff520de27fbb,0xefbfcffeeddafb9b,0xd6ff7ef98fe73f9b,0xf3ffee71efa3fbbb,0xffbfff7e8ff3ffb9,0x8fcfbf37cdd7ffb1,0x3bfedddfefdbfe9b,0xf4ffffed4fdbbf93,0xf7ceef72efc2bf99,0xfff7ff6f8dcbff89,0xdfdffffecfd7ff99,0xefffffdfafd3fb9b,0xdfffdffbeffffe99,0xfb7fff7feff3ffb9,0xfeff3f7fafe3ffbb,0xeff7ffffeffbff9b,0xafffdfdfcfdbffb1,0xbffaeffbeff7ffbb,0xffffffffeff6ffab,0xfefe7fffcfd3bf9b,0xebffff3f6fe3ff9b,0xef77ffffefffff3b,0xffbffdf7cfdfbfa9,0xfffffeffcfd7bfbb,0xfffefd7aefd7ffbb,0xebf4f73e4fe7fb9b,0x5ec3befacdc3bf03,0x7f7ddf7a6fcb7a93,0xf5fff6ebcfcfffa9,0xeffefff7afd3ffbb,0xffbfffffefdfffbb,0xdfffffbfefcf7fbb,0xeafddfdfcfbefe9b,0xffffffffeffbffa9,0xfbffffffcffbffbb,0xfebeff7fcffbff13,0xffd7efbecfc3ffb3,0xeffedfdf8ffbffbb,0xb5fff7fbcfdffbbb,0xffffffffeff7ffbb,0xfebe3fffcfdbfebb,0xcffdf7fbccc3fab1,0xaf7f7e4fcdf73b8b,0x4fff7fffcdd73b91,0xf385f660ccd27a93,0xffce6ffecde6fbb1,0xc3aef7bf8ef3bf11,0xa8cf99c92ec39a83,0xde9efffacff77f81,0xf4fefef7eedbfe83,0xeefaefd34fdfbe89,0xcbf6f37fce93bf19,0xffeb7e696fd3bb19,0xe6bfcffacf433f93,0x70548fc40cc29211,0x45603c0005803801,0xefa6fffb8ed37fbb,0x41c3d60b0ec33aa3,0x7bd6dfd6cd837ab9,0xdcfef7fdcedbbb91,0xffafff63efff7fb3,0xfab47f74cdcb3aa9,0xc8fefe284da1be83,0x1b64df598fc3db81,0xb6ffcff72fc3bb13,0xf9bfe7798ed5fb91,0xfebe3f7fcfdbbf89,0xcfdbff7bcfcb7eab,0xeffffeef8fc37fb3,0xb73b57fecec37e9b,0xf1eca5612dd39b93,0xfbf6ff7e8ff75f9b,0xebfff79acfdbffbb,0xc7f67c438cc7a281,0xbbbbeffbaffffa1b,0xdebe767f6dc3fb83,0xdfbeaf73cccf7f93,0xcb10f72eacc31fa3,0x5b5232970cc3da22,0x4fb6f5fa0cd33b81,0x1dfedee9ee88de03,0x97aced310cc0bea9,0xf100c1f00bc35b11,0xfddfdf3e8cc33fa3,0xab7fd7d3cfb6fa91,0xd4f1dce92cf2ba99,0x19eff7e92e52be89,0xfb9f3ff7aed7ff9b,0xffffffffefffbfbb,0xffffffdfeff7ffbb,0xbeff5fffadd3bbb3,0xddeea7434a503e81,0x66862f3605823e10,0xe9d2ff7befdb7fab,0x4efefe47cdf3fe83,0x4d057b18cd0bb0a,0xff78ef7fcfd3ff9b,0xce2eadf8edc3be89,0xca85d3122ec282b9,0xe3fffef3ced3bebb,0xf30481ac4c33a11,0x91b086688c823880,0xa2256000831a28,0x984c432008523a11,0xc8220a2008200b01,0x4400404888000,0x400100000808000,0x840b210008c0a201,0x4012008000001,0x0,0x0,0x74b5cef12e955f91,0x3222d50050f2a02,0x9ec4374c0ad1ba01,0x46e0000c833a09,0x80d0df5888436e91,0xb426d98a2c527a8a,0x4179f462cce23b01,0xd69fefd78ddbbe83,0xa9acf31f4dfa3b81,0xbd529b0b8c83be00,0x8e1269988ce66a89,0x8508000028000300,0x9a002f2e4c436ab1,0x8008da5004838a88,0x77d212844c83a281,0x839e46302fc00a01,0x642654c000d23e00,0x2a142c2205023810,0x2984c7280cc332a1,0xc957b2330fc2aa09,0x2b3ccb79ade27a99,0x91d706c8cfa23891,0xfffeff73efdbff9b,0x5284ee2e2c923328,0xa24432bbcf8b1b01,0xa12e489acd321b09,0x8699c4ea0842da81,0xe98c476808d25aa9,0xaa9763d48dd35a8b,0x5fc3a63e85d67a01,0xa9ee5ed18fd3ba91,0x3ed76be7afe32f9b,0x4128a6080c583900,0xd0162e208d430a00,0x6c4e51140cc2a290,0xa2769dd20fe33a89,0x3dba476a8ef0ba83,0xf44dbe618f831ba1,0x410020404420200,0x8280420080131200,0x40a301806103600,0x2090006000202,0x4410800002a001,0xab3c75714cc73f99,0xca54531e89c05e01,0x4cc290388d829991,0x200001c000011a00,0x10020000020008,0x60008000001,0x822002120cc11a00,0x382d60288f23a10,0x8a50040a0c031a00,0xf8f4d26085cb0e8a,0xdeec0560c5c67a83,0x8290200080021a20,0x5296000d402b00,0x10e000003000801,0x8810010000808000,0x2010040000001,0x0,0xa002c86048021001,0x19475e0088d61712,0x888294d924971200,0x1000001080020000,0x43a2753184827a18,0xa1049250c8420a03,0x808171005801a81,0x8fa8d75aafaffab1,0x7194d8604dabee01,0x44961d72a4c3a201,0x1280c1220ec00382,0x623f410ec14a11,0x4040008000001,0xa0500080024800,0x400004000010800,0x2020160206103001,0x2100008200000,0x2210000401010a00,0x4008402028000201,0xc08250000115800,0x7000000204800600,0x0,0x6208140004022220,0x19d15dc18cc39a91,0x2204040008000282,0xba0cb7ad8cbaba09,0x8983e72b4d92ba8b,0xebdffa5bcdcbbeb1,0x9fa2cd524fd3bf89,0x3935a6788e8a9e09,0x12616400a200801 },
	{ 0x6389e50a60711554,0xa000000a004410d0,0x8110400008421000,0x444020020500080,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80831e28227b104c,0x80000800100400,0x54f17678645c15dc,0x10a184060001108,0xb995224da555114a,0x4edf7f19e67e15de,0x7fffffcff27e97fe,0x3282d6e826721146,0xdd7fff73626e15eb,0xffbf2ffe717b155d,0xfffffbafef7797de,0xefff17d16a7a15de,0xbcbfcfe16e7a13ca,0xddeee643e45a92ce,0xdabc3d78337295d8,0xc9f7ff7b327217ff,0xa3ffff78237317d6,0xb0bb6ff3a77b15de,0xd176b5d4667a16ce,0xff963766677614de,0xcbd1ff2ae86211ee,0xaf6e0a51a2721258,0xffa9fdf3a957164a,0xf5fff47d237e13de,0x2e3a6d60f273136a,0xdbd4cf1ee37391c8,0xdedbff4fe276156e,0x9fbfcff2a37f14df,0xb5bf76eae75f17fe,0xd9ef6d777f5217da,0xffb6ff7f6d7797df,0x4fd7bf0bea5b16fa,0xfb7cdfdfef7215ce,0xedf7fefdfb7615de,0xffeff77bff7a13fa,0xffbe6f7ff37716fb,0xefdffe1bff7f17ff,0xffffffdfff7f17ff,0x9ffffffbf77f97ff,0xfbffff7fff7f95ff,0xffbfbfffef7f95de,0xcfffffbff67f15ff,0xebffffdfff7f97ff,0xfffbffffff7f97fe,0xfffffffb6f7f17fe,0xfffefffff77795de,0xebd4feae6f7315df,0xeff75fdfec7797de,0xfffffffee77b97fe,0xfdffffffff7f95ff,0xffbefff2f75e15ff,0xfbd5ff3ff77615de,0xdcf3f29b3172175e,0x8fbffffbb67f13ce,0xf5f5f6ebf77617e7,0xbff7bd71787f11da,0xffbfff7f6f7795fe,0xefffffbfef7f93ff,0xffffdfdff77b17fe,0xf4fdfee87f7f95ff,0xffffe77bbf7f97df,0xdfbfffffe37f17ff,0xffdfff3ff77797df,0xefdfffdfe67e95df,0xf7ff77ff377e15de,0xfffffffeff7f95ff,0xfebe3f7feb7f17fe,0xffffffb96f7615fa,0xeb57de5be27395de,0xbfb37ffffe7715fb,0x50fcffe0277a14d8,0xfefe3feca87f10de,0xdbffff7f2f7b155f,0xcdcf7fc73f7a11de,0xbfbdcffbab7f13de,0xfdfef779f77e15fa,0x7ffefd72e37a95df,0xcb94e7af247b97fe,0xdafbffdb267e15fe,0xa79a9fdbab7f13df,0x72d61dce78721478,0xa17ffe22617211ca,0xfb98ff6fa77a11ce,0xcdd2f695377217ca,0xabfe0b4eaa7210fe,0x3437d7eefa7e91cf,0xd3feeffbfb7b11ce,0xdfbf7fffe77795df,0xdddffd32a57a12da,0xefdcfe5be77397e3,0xb49f1eee226611de,0xff7ff77e6a7e91f3,0xfebe3d66f67e91cf,0xcfd6ff7bf37695cf,0xefd7fe57bb5f11cb,0x3fb27f7efa72154b,0xd1ecf4faa27b15ee,0xf7b6bf76227f14df,0x89fdffbfb77e14ef,0xea4f3679ab7a97da,0x123addfba27711d2,0xf6bef6f3fe7311ef,0x5efe7df2277e17c7,0x89c1e1610073155a,0x5ec62e3aa45615ca,0x6d205348005a1446,0x59b48678637213ca,0x57a8ff60246a91d8,0xffb76d7ee67214de,0xfde7b31b267e95ff,0xad7f5c922b6e10ca,0xecf676e1027e17ce,0x9d6f6d76237711de,0xdb926f7eab7611de,0x5fdffebe7b7796ff,0x7fffffdfff7f97ff,0x9697436031701050,0xdff9a778ba7a17ea,0xfabf2d02f073165a,0x6182e3386472904a,0xa90b1ae0a20214ca,0x7b0ad7fb325611ce,0xdb5c7733235a14df,0xde9eeff5e27693df,0xc2b2fa676f7a91df,0x4adb7f50ea7610cb,0x8a3ecd5962561558,0x8010002022400001,0x5a04200100721046,0xe00000001640,0xc8032a303210105c,0x0,0x89e9076a237610af,0xa519a037ba7e94ce,0x1004102000431100,0x82300300001150,0x10bdc4226301052,0xf09deee8b25212ce,0x2040000000082,0x88000a2800521021,0xc9334f0e205291cf,0x30eded3b263107e,0x4000200500000,0xd9aebe30237a13f2,0xfbb22ffff57e11ce,0xfdc7ff362f7315df,0x7bfeff48237a15de,0x8c306432a3611448,0x9200a640227213d0,0x3200c4030001248,0x6000000050,0x6e877808705b1154,0x8998c08ba362004a,0x844030692c621068,0x709c29a060541487,0xf356ef7ce373116f,0xc372becaa75613d9,0x37ef4f7a364395d7,0xfcf776fce06a95fe,0xfffffff7ef7717da,0xdb98ffae665915c8,0xea46050f3974105e,0xa81eca52615215d0,0x94fbd6e17a73115e,0xb0e88f4021461159,0xcbb26ffee87315da,0xab47ff1621771552,0xab57dfdf667a174e,0xb4bb3de3a77b10cf,0x12ca430005814da,0x72a227667170117b,0x49cac710e0531448,0x83bfee40a47a90ce,0xffbbcffb367310fe,0x91c635007c421000,0x74061560a27406da,0x900800820420040,0x0,0x4002a00121004,0x200e00a000621000,0x6830a04264500041,0xcb80f7263163164a,0x605236eb6f541089,0x4400020000080,0x811502002011050,0x320d406c410040,0x4f80d24922701160,0x49878e9b296317ca,0x600e4c12626001ca,0x909464a060551049,0x41002252220000ce,0x85002208600046,0x0,0x835c4e50223010ca,0x0,0x90492600a542000a,0x8000020010001000,0x638c7606325211d6,0x12a20220129000,0x108422026421100,0x0,0x723636d4225710ca,0x8194877c29531548,0x4802580122021150,0xf8c6dd9677710cf,0x4082244a0701000,0xb020d4800721440,0x0,0xa053b0d9e056104e,0x60230140a0740006,0x1000208400011000,0x2c0200021021000,0x5004000000400060,0xc1230a04002001d2,0x8108041220700408,0x10400000080040,0x200002000000000a,0x80004000000140,0x40108824400020,0x208480121420058,0xc1408a0201000c2,0xa10c2610a610005a,0xd28326362472115b,0x89f6f326e154905e,0xefdeffc1a17315da,0x8b825f7a627e10ee,0x98fc364110781740,0xa09642052002c },
	{ 0x888060460110e205,0x80d01801081e00a0,0x1228c83810c0884,0x942026020211e01a,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x98c20f39533de7e7,0x203004a0083401c0,0x5600002c0010,0x1225d500101c002,0x22104306404c8001,0x800820100340080,0xa216ced31a1d60bc,0xa4c45429c0342a00,0x86625e05b4d6c10,0xd696f3f60f3eced5,0xedffffbfdf3eeff7,0x277afe51ca7fefed,0x859437efce3cefed,0xf9ff3f72db7decd6,0xdeb6776e8d7febda,0xefdefbbedf3eebf7,0xbdfffe6b9f3ceff5,0xbfbbffd7de7deefe,0x7035b667433ecbd1,0xfbf6fdf69f3ceff0,0xe8fdf206db1cee8d,0xe6e31504b3c6e98,0x9f9f5bff9b3deaff,0xfddcf67f0f7eeed7,0x71b4fffa8b3ccdfe,0xfb84ff3e1f2febb2,0xcfcffef8df3deff7,0xbb1e45fe483dce3c,0xbdfeffee4f7defd7,0xdbeebf52df4eed55,0xde96e7771f3deffc,0xc9e7ff37df5feef2,0xfbfecfdbcf3fefff,0xfdffffe9df7feffe,0xb9feff7b5f7defff,0xffbfff7fdb3fefff,0xfffffffedf7feffb,0xffffffdfdb3fefff,0xdfffffffdf3eedd7,0xff7ffffe5f3fefff,0xfebe7fffdf7fefff,0xeffff7ffdf7fefff,0xffffffdfdf3fefff,0xffbbdfffdf3fefff,0xdffdffff5f7fefff,0xdffefffe5f3fefff,0xcbf2ffbfcb7febfb,0xafff7effdf7eeeff,0xbfffffffdf3feff7,0xffddf6ffdf3defff,0xf7beff7bdf7deffe,0xfbfef7ff5f3deede,0xeefab71b9a3eeed4,0x8fbc5f5fdf3f6cb0,0xf57ff6fb5f7ecfb7,0xdbaeff731b3cefd7,0xfbffffffcf7fefff,0xdfffff3fdf7eefff,0xff5fdfdedf7fefef,0xfeffffffdf7fefff,0xfffffffedf7fefff,0xfbff7fff9f7defff,0xffffffbbdb7fefff,0xffffdfdbdf7feff7,0xffbaffefdf7deff7,0xdffdff7bdf7eefff,0xffbf7f7edf7fefff,0xcdffffbfdb7deffe,0xcd4f3e5f9b3eafbf,0xb4bbffffdb7feffd,0xf5fdbf79db7ceffb,0xb7be7ffc0f35eed4,0xcbc7ff1eca366ee4,0xfed63fda4d3feed0,0x8fbbdffbdf3dcab6,0xdff7f77ddf7feeff,0x4b3ef5728b3fcbf5,0xebf9f33fdf3dedfd,0x2a5eaf5b4a3fc6fb,0x7ff8efc1de3ecbdc,0x3140874a0b1cac81,0xf3bed570cc7eeacc,0xdfbbfbff5f3feff6,0xeedeff1f9f3feed1,0x2fff97d98b3eeff1,0x5effffe8cf3deadb,0xddfeff7a5f3deff4,0xbffe6f765f7deddf,0xecdfff32db3fefd7,0xabda3f978a3dcef4,0x5cf651e31f2cc8e6,0xd92cf7e1df7eeb3e,0xf6b6bf7acf3dcbf5,0xf9d7feebdb3ecedf,0x2cd7fa6d8c3ceaf6,0xb7bf67fecf3fcda4,0xddfdf7fbdf3cebf6,0x797cfd725b3feec9,0xd1dfdf0e1b6ce9d8,0x8f5317da4b3cebc4,0xfebe5fea8f7deaee,0xe5a7f2ff4b3ecfff,0x55be3fe2833eeed6,0x9991fb6c1e26eb58,0x4dd63e06c1144c24,0x8711c5f11f7de468,0xbde957ebcf7ecff4,0x9fefaf77c93ecfac,0xfa85b7fece3ecfb7,0xc34ad2290a3c6a9c,0x6714d7921b30e2a3,0xf6d4d47d5b2cecd6,0x14a07400026c892,0xf6ac17745f34efc0,0xdfdffebfdf7defff,0xefffffdfcf3feffe,0x3ef757ed8b7eeed6,0x93c4ae720d5dea84,0x42202c414a242400,0xd151971d487c2ca0,0x4b4e1f0b497c0ce4,0x9cb61eb45e3661c4,0xd3ceb5f24c2ce8d3,0xfe9e5d709f2dedf0,0xcac6fa4e4a7ecae9,0x1ad71f401a57e22a,0x1f2adfdf893c6ee4,0xd77ea6a85f3cecd9,0x2221540001c8630,0xc334f102da3dc84d,0x8801eea882164a24,0x188090008102088,0x0,0x100800080000800,0x42000004800,0x87cbffd11f256f18,0xa1c0fdd1a3e84fd,0x220e8417044404,0x0,0x80000408000080,0x46021002004900,0x2200c00013000800,0x21b2546043344d00,0xfde4ffef1e3fefd8,0xdebe3ff95b7fefde,0xcfd3ffee1b7eecdd,0x2eded8dc9a3caed4,0xbdca4dba8f2c4cd8,0xd0008a0000208810,0xa40640009042c00,0x83a8d1228924c918,0x95f1003030e0e84,0x1000010008000200,0x0,0x20050008080220,0xdbdeef37cf3deec9,0x8dea3e170a36cfca,0xd1cdd585a3d47c4,0xfdfd17f85f3dee9d,0xdfffff77df3fefff,0xe6ace56e8e3f4ae0,0x5180a227013d4844,0x8a5a5fc85a3d6c90,0x96bc4eec5b3deb95,0xb9c80fe00d75e045,0xdf8eff2d4a3ccfd7,0xe9d3f3134a3ea88f,0xa951845242142004,0x22840309002880,0xaf7d2330c9776884,0xcca62740131c4b94,0x80201362822c0810,0x41185a9981384e90,0xb4aae5ec1f3dcbe4,0x816c42300a3c82c8,0xc2cef5f69e2deaf9,0x100850000040044,0xa080108002240c00,0x28220d400666c200,0x0,0x10087440082c2014,0xa080c026000c44d0,0x4a402c404d10e802,0x0,0x500c002040600,0x9008144006244108,0xd111cb3c037cca92,0x40261081044c00,0xc010091208168222,0x49120210b040844,0xcc35020b040c48,0x2800400080c0008,0x480000000100080,0xa010400002040004,0x90500000008002,0x51480300080940d0,0x3000010408302000,0x802d0088101c880,0x1408001002c0040,0x53ae7654934c2c8,0x0,0xc4222d6403162a80,0x8c4812a401e01b0,0x40020000240000,0xbb817f789b7eeed4,0xc31044600a3c81c8,0x44086c020a2c0080,0x8100041002040140,0x42284004084010,0x8008000008080,0x40040000008002,0x801029400e104160,0x19223051a0e8085,0x0,0xc00000080004,0x420000080000800,0x6008200000,0x10142a640a34282c,0x6002020002004110,0x9bb9df5adb7ecfff,0x5223e255b2489c3,0xa04225784814ea14,0x9bb22e4ac35ccc5e,0xcedaf7fbdf3ceefc,0xafdffbfb5b3feefd,0xbe3a46f15b3feed2,0x10040618050c4884,0x0 },
	{ 0xc0002a010208853a,0x606060801000502,0x61440000122040a,0x9290000050080d21,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x488e1285c19acf6f,0x5154ab238001799,0x91040000008001,0xdfe82d72b28adf1b,0xd944020240908529,0xa10000008030a,0x73bdcfc0910a971e,0x31a1040040100721,0x94497590901a4fa0,0xdb363faf71bacf77,0x57cbb690c2825d57,0x89dff95b633acf73,0x449a77eb93b287fb,0xfdbefffedbba4f7f,0xf7bc3f5ee39adfbb,0xc9f5f7fffa3adffa,0xaf7eff4ffbbadffb,0xffbf5ffffbbadfff,0xe03717ebdabac7f1,0x7a9674f2291847ab,0x8afaf3ba909a575f,0x6effdeb8f09a57fb,0x9fdffddb792adff7,0xfffff6fffbbacfff,0xdfee7de3f19add77,0x8bd4ff2cfa3a457e,0xfffffffffbbadfff,0xe7bccf77fb1a97fb,0xff7bfafc0b9acdff,0xfefeff70339a9fdb,0xffbdf77e79bacffb,0x9fffde9bfb9acff3,0x6bffcfdbd9badffb,0xf6fff6efdbbacfff,0xffffff797b9adfff,0xffbf7ffffbbadffb,0xffefff7efbbadfff,0xefffdfdffbbadfff,0xdfb7fff799badffb,0xfb7fbffbdbbadfff,0xffbeaffffbbacf7b,0xeffff7fffbbadfff,0xfffffffffbbadfff,0xfffffffefbba9fff,0xf7dffffbf9badfff,0xfff67fecdbba9f7f,0xebbfff7bfb3a9fff,0xafff7fdffbba5ffb,0xffbfdffffbbacffb,0xff7ff6fff1badffb,0xffbefdf3e9badfff,0xdbd5ff3fdbaa97ff,0xaccedb5fd13a07fb,0xffbbfffefbbacfff,0xfdffefebfb8adffb,0xdfffeff3dbba9ffb,0xfffefffffbbadfff,0xcffffe9ff9badffb,0xffffdfdffbbadffb,0xfefdfffefbbadffb,0xfffffffbf9badffb,0xffbfff7ffbbadfff,0xefdfffbebbbadfff,0xebfedfdbdb9acfff,0xfdff5fe3fbbadffb,0xfdefff7ffbbadfff,0xffbebfffd9bacfff,0xeffeffbde3badffb,0xf35e57db1abacffb,0xa6936fee399acfff,0xfd69f77f419a57fb,0xffde7fedfa9acffb,0xebc6efb3a29acffb,0xfff7b7dfd93acff3,0x979edfbb39ba97db,0xd77ff7fbf3badff7,0x7ffe7fd2fbbadf7b,0xfbdfe76ff19adffb,0xefdebefffbb2dfd7,0xb3ccf3f93ba1f5a,0x7d97dee8c21a0fbb,0x89ae2de2ca9a0f78,0xff9aff7fdbbadffb,0xcfdbfeabf9ba9ff7,0xbdfeffdbfbba5ff3,0xfefd5eefd13a4deb,0xff6eff76ebbadffb,0xfbbf77fffbbadffb,0xecdff35ae09ac73f,0xabfe5fd3c992ddf9,0xe4365767832a8f9b,0xd93c2ef2798a113a,0x9ca63f7e539a9f7b,0xddd7f6aadbbadd7b,0xaf5f3fc65abacdfe,0x3cf3efe7fb9adffb,0xdb79ff7b799a87fb,0xfffe7feecb1a9d7b,0xe925e377529acf7b,0x667f1ef9f112dffb,0x9fbf5fdbe38adffb,0xf17bff7bd93a5fff,0x5fae7f6a439adf77,0xebddff2ff9badff3,0x8957b99ad92acd6b,0xefeeecfc3ba957f,0x15f75dead0185ff9,0xdfcdff635b9a9fbf,0x9bb4ecef592a8ffb,0xcdfdfe1719aa9dfb,0xfbedfde6a1a8ff1,0xfefddef6fb1acfff,0xc1e7ff55109a9f7f,0xdf9e677cd3aa5db2,0x7d43cffed33acff7,0xabdedbdfe9b21f79,0x3d3a99f3c9aa4fe9,0x9129ade8001acd0f,0x22842c21900a0443,0x8080e620000a456a,0xa0181283d0180731,0x95b66e6bc99a8f73,0xd0fc2d74c308877b,0x93fc7f1541aa8deb,0x8880f73a51388dfb,0x4b571ec76092dffb,0xd3934cfec9aacfd3,0x8df17369d01a1670,0x4bb82d30418a0fab,0xcbd0f33b703a0deb,0xd8829a69409a0d13,0x804884301880102,0x0,0xa00000020000100,0xa900400101200901,0xe9cff3b34b9a5fdf,0x2f5fd6d9b18a8fbb,0x48000000300,0x4200000120,0x400004000080002,0x0,0x0,0x418523060020542,0xf76d77da6bba8ffb,0xca9b2f3fd31a8f3f,0xc9fbdf399b3a95d7,0xa9569f6e512a877b,0x1bb25f62993a9f11,0x19fab70011184f23,0x902076041320d00,0xe8c672f2839acf53,0xad31e15ea08cf1a,0x1084410202000538,0x40800004800,0x303c5440100a04a5,0xbc4e51111888f3b,0xccfa36dac82a07f3,0x2cfdd7521b8a97fb,0xfdfff7faf39adfff,0xdffffff7dbaadf7b,0x79c677fea89a9dfb,0xcdd7be0cd8bacd72,0xe99a870a4288057b,0xe4f67b645b2a0f7b,0x354c8762618a1fbf,0x8a16330860980523,0xca060e93418807b3,0x8115109200970,0xa022145a2108951b,0x8300040000028901,0xf63a2576000a9f73,0x80bafb2849b24d43,0x214ec25050180fb1,0x27b155f24918c523,0x10c0020010080500,0x1dafbe2ba1acdf9,0x8000d08088080502,0xfada5f4f488a8def,0xa88c11900289458,0x600e0000003a0,0x290000004500,0x83001119800a0c47,0x110000001010,0x200001a001008011,0x8060176221a20120,0x10982c1040028070,0xdc06534e80180703,0x41101201400001a0,0x44000000408,0x0,0x100236208080540,0x100000140000020,0x2cc1508081824c06,0x804a418108040a,0x0,0x48022800080700,0x880829010108480a,0x8300240410080402,0x2100088040820100,0x0,0x10c4140080100601,0x484202020012,0x9001218020007b2,0x40020000080002,0xf9ba8fd3891a9f7b,0x14764464a8020f13,0x4800000100,0x0,0xa8901287401a4423,0x810036001080002,0x4000800000020800,0x200300009000102,0x420260080000052a,0x4c420e011108050b,0x1000010000100400,0x8000002000100400,0x62094000080442,0xc282214000109531,0x8000e00010001410,0x1880e12100a0d07,0x8030068108020152,0xd76ffe70d89ad531,0x92a67774381acd39,0xeec7ff3a899a5f7b,0xa2ffbe5a590a9f73,0x96fb4f6bc99a8fbf,0x50648668620a0732,0x421225200218173a },
	{ 0xcab0eb7a3cdefd46,0x10060200828200,0x20000080c000000,0x50c4b7d13c91f5e0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa97ab1de24b27ee4,0x671c4c00a09326e8,0x6014408000122284,0xdb42fc6114afd266,0x200000610014800,0x8c089620008646e4,0x5084020000040,0x5096f6fa35227ac6,0xcfc52e23143a7362,0x9ae22e6514f9c582,0x64ce31221c334ae2,0x6bdeff813cb8eeec,0xa61bfee6bc967cc4,0xfb6de77c3dfbffee,0xfebe7f5e3cb2ffee,0xe9d7f72bb43a7fee,0xefffffdfbdffffee,0xffbffffbbdffffee,0xd373bff83cbbf8ee,0xff9a3ded3cbafbe6,0xabc3c33221ba72e0,0xefdf3edebcfbffe6,0xbdb37df235baf3e6,0xf55ff6e9bcdbeee6,0xcdee7fe114ba7fee,0xdbdfffff95bfffea,0x6fdbbffb34feffe4,0xcc7f4548149558e6,0xfdfff7febcff7fee,0x9f7fed5334bfffe6,0xd1b8e7febcdbf7e4,0xdfffffbbbdfa7fee,0x3fdfffdf3cfbdfe6,0xfe3f56fd3cbf7fe2,0xdbeef77334bf7fee,0xffff7fff3dffffee,0xeffffbbfbdbbffe6,0xfffedfdfbcffffee,0xffffdfffbcffffee,0xffefff7fb5ffffee,0xffbfbfffbcffffee,0xffffffffb4fbffee,0xffffffdfbdbfffe6,0xbffffffbbdfffbe6,0xbdffffff3dbfffee,0xcffe7ff7bdff7fee,0xfbfffbffbdffffe6,0xefffffff35ffffee,0xfffffffbb4ffffee,0xfdf7ffffb4ffffee,0xfffeff6bbdffffe6,0xbbfee77634bbefea,0xfff3bfbbbc3b7dee,0xefbecfdf3dff7fe6,0x7cf7f6f805fff2e6,0xeffffff7bdfbefe6,0xffbcffffbdffffee,0x6fdfffbfbdbfffee,0xeb7bdfda1cffffe6,0xf6ffffff35ffffe6,0xdfffff7b3cfb7fe6,0xfffeffffbcffffee,0xefffffbfbcffffea,0xffffffdfbdffffee,0xf6fbffea1dffffec,0xf7bff7fb9dffffe6,0xfebeaff73cbfffe6,0xeff7f3b7b4de7fe6,0xbbd77eef35bf7be6,0x9ffe7fff8dbf7dee,0xffeff7fd1c3b7fea,0xffee7d453cbaece6,0xeaf9f3293cbe6eec,0x6fdeb7df3dfbffe6,0xdffd6ffe3cf9ffe6,0xf53ef67e3cfb7ee6,0xcbae1f60b4bb7ac2,0xf8ddf77f3cb3fdc6,0xee77ff9f15fb7fe2,0xbfffeffbb4ba7eee,0x71ae66d810946ee8,0xdb2bed433cbf5bee,0xfb95f7bf3cbbfeea,0xcfdbb73b34bff266,0xfffcffdfb4ffffee,0xf6f7feef3dfb7fee,0xd7efed433c9b7ee6,0xbbb7ff7fbcfbffc6,0xafffffbb3dfffee6,0xafffdbdf3db77fee,0xb9fe3defbcddf7e2,0xffedff78bcfb37ee,0xfa99257f3ca6efe6,0xadf6f77bbcbeeae6,0x2ddffc512c92e7e4,0xbf5fcffbb4bbffea,0xd378ff793dbbffe6,0xfb9eff7e3cbb6fe6,0xc984f7b635bbefea,0x9ef73f5ebcdbfbee,0x47becffbbcbaebee,0xff74beed14eeffe6,0x4fea67eb04bb7ee6,0xebfdff2f3dbfffe6,0xee93feeb3dbffbe4,0x3e6ccd753cbe6ae6,0xf0b065fb1cb726ea,0xc3a12d031ca66be6,0xf793f3ff3dfa7fe6,0xcfddfea3b4b336ee,0xbecefcb94bb17ee,0xb0d1d428809262e4,0x11e20fe334b1d162,0x5bbd6ffd35faf6e6,0xdfcfffae38f2e7ee,0xafdc7f93b4972ee2,0xb5d615e20dbcd162,0x114c066880b99280,0x8a04256004126b40,0x82018602002260c0,0x80808e86042027c2,0x232807400c391028,0x6178be180488d6e0,0xd2aefdfd1cfbfee6,0xa9deff2a3dbedae4,0x46d27f1a392a6aea,0x81aa85c208f25aae,0x5062f6f01cfade40,0x5c04a5d004b811a6,0x8000120001220004,0x88018604041002a0,0x401000004800000,0x0,0x102000000220000,0x8000b3804111062,0x0,0x2080000200200,0x8521106214b10386,0x112c202c30b10682,0x0,0x80800000002040,0x100100000084,0x140414c034110462,0xfb64c5b83c966fe4,0x6a0205700891caa8,0x9d0521401b850c4,0xed94100b242a0,0x303208221891cc82,0x9041820020114ac0,0xd6b63fd5349afee4,0x89dee12a3c986ec6,0xac5e5ef81c9c56a6,0xd0acdca301232c2,0xd5b49a5114333ee4,0x41223d6000336a66,0x8b4cfb6a28fe52ee,0x6c95b5b18ba6ee2,0x628cccf114f2e6e6,0xf5fff7feb4bfffe6,0xd5baaf71a0fa12ee,0xfe9e6bfe349f69ee,0x5df7fe8d35bf6de6,0x2afeedd33cfff2ee,0x74503df691bae6ea,0xf9b0b721343127ee,0xbb96b771149f796e,0x88d1d303301322c6,0xa51c2488c923004,0x3220816904126064,0x813a776a10117be2,0x4abda7700094e6ee,0x2956e71ab4f8256a,0x65dfffdb3c396fee,0x363a65bf30f2f8e6,0x98501059202a44ee,0x5784b7e00cf8f842,0xe938d20224ba6342,0x60869a2115391200,0x8c04000c808000,0xe4346c62049271e4,0x152c1d4000d36244,0xb9acbe6e14fbd2a6,0x5c7ebe86841333e6,0x6000000280000,0x9001806210024e80,0x19640720cd19880,0x6506002a4240,0x4080100100101000,0x870824105a86000,0x810000010010000,0x122584000986068,0x3100000100000484,0x202000100020,0x0,0x1020102000844a0,0x1000040000200002,0x40840020000024c0,0x29ded2321cb9d820,0xa9631800009062e0,0x322101810200482,0x51a82752383864c2,0x900420000112000,0x180001000280200,0x800010000000042,0x7198b7d3b4bf57e4,0x8000000000000084,0x0,0x4800d06810884440,0xe04230d38494606a,0x2abec8400c1b6ac6,0x1100860200c02040,0xc0c12000000042e0,0xe312624c14b14162,0x200200000880000,0x410000000120024,0x140010041c000002,0x0,0x280042000006840,0x1c4c00c24ae3242,0x140000000280,0x120000000101000,0xf825228210860bc6,0x7020019201386020,0xafead6a514bb73e4,0xea5f7ed9b4b977e0,0xf81775a1cbbfca2,0x5586176e1c1b7bc4,0x821a2c4018b9c482 },
	{ 0x8191c20510d45a28,0x4122400101c0a28,0x22908702145cd242,0x716ca75bf4347768,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2840915024105020,0xefbdcf49f47c5fff,0x546c9665bc1cfee2,0x9d40424180205442,0x529160119010168c,0xcd73f733fe70726a,0x6f9cf53c05456ea,0x75b77ff9def474ea,0xefef67fbb4fcf6ea,0xfa36bf83f43c5e78,0x8681f000141510aa,0xebfcdfdbb47c5eed,0xa37cddb3a41c7eea,0xdbe5e7fff6bdf6ec,0xfe9325e7e65c7efc,0xffd4f7be76f076fa,0xeffffedffefdfeff,0xfffbfffbfefddfff,0xfd74f75cbc5c5fef,0x5fae7f77be547dff,0x8be4e352765c14ea,0x6ed79fdff4741fee,0x5b9ccffb74d5f4ee,0xd5f1faff7c305eed,0x47be7df3d61d1fff,0xdbfbef7faefcff6f,0xf9dbfffb34fdfefe,0xbfbdc7f33e5cdffe,0xd6f77ff2bed4ffee,0xd9fef763fe387eef,0xd9f5fffff49cf6ee,0xffd7ff17b6fc77fe,0xaf7fdfdebcbc7efe,0xfebafeed54dd76fe,0xdf77fff3f4f57eef,0xffbe7ffefcfdd6ff,0xeffffb3efe7cffee,0xefffffdbfeddfffe,0xffff7fef7efdfffe,0xdfffff7efefc5ffe,0xfebebffffefcf7ff,0xedffffbefe15fd7e,0xbffffffffefd5ffe,0xbf7b6fff9ebd7cfa,0xf7febffffefdffff,0xeffefff7fefcfffe,0xfbffff7fbebcdfff,0xffffffdffef4feff,0x3ffffffbfebdffff,0xf5fff6fef6fdfffe,0xfffe7ffbb6fdfeff,0xfbddefb776bcfffe,0xcdfbbbdd769dfffe,0xbfffdfdffefd57fe,0xfb7ff6ffaef55cea,0xfffeef75f63dfeff,0xffbffffffef4feff,0xffffffaffefdffff,0xbfffdfdff6bcfeff,0xfeffffe7fefd7ffe,0xefffff7d7e7cdffc,0xfffe7ffff674dfff,0xcfffff3ffefcdfff,0xfffeffdffef57fff,0xddbefffffefdfffe,0xffeffffdfefcdefe,0xfff63f7bbc5ddfef,0xfffbf37fde14f6ef,0xffff3f5f34fdfd6e,0xbffb4ffafe5d55ea,0xd3edbf7ab63477fe,0xbff67f61bcf5d7fa,0xdb9fff3af6d4567a,0x6f765e5f6e7c54fd,0x3abbcd637cbcdefc,0xdd74fe7efc5dd6fa,0xfffe7ffbbef47eff,0xc1d4ef62b6745ce6,0xfedbbebbbebcf7fa,0xffbffffbf67dfffe,0xb5d17fe2fe985e7d,0xdaaa0d519614d66f,0xfbbef77ef63cd6ee,0xc9eb9a319434549e,0xbbbfffdbf4fcf6fe,0xf6f7f6effe3d7eff,0xdfffff7bf49d7eee,0xfb967ffef4dc76fb,0xe3c7df7afcd4dceb,0xebfedfda7efcdffa,0x97ba47fbb47c5eeb,0xfb7dff5edebcd6f7,0xfae2ff65fc7556ea,0x8fd8ff37f4dcdeee,0x3ff6bbd8b67cdef4,0xc5bbdefff6d4dfed,0xffbefe737cd4dffe,0xf89e3f6ef67857f6,0x828ef38ffe155efe,0xeedd1f92b49414fa,0xfbecff594545ff2,0xfd7fbafd7eb9d67c,0x7ffcef7afe3ddefe,0xdb56fa2fecd47f6f,0x84f212589c14367e,0xebb0deff1c9057fe,0x47536e2d61436f2,0xbfdebf74b65577fe,0xcd86ff3bb430f6ac,0xcbf7ffbbb6bcfffe,0x11c4b4c201c163a,0x36d4366a2c10f6e6,0x90ab67411e1452e8,0xdbef7ef6d631d7ed,0xefdbffbaf6fc5eff,0x6adfdf51b89cdcba,0xb53adac3549d56ec,0xb7ff2f6a72d45dfb,0x12940f7c54101666,0x8372fea27011166a,0x8080d220005660,0x80804c6020001468,0x933e8760a03cd462,0xb447e6c2615366e,0xebb17f27cc81f76a,0xee4d4e10161036c4,0xd89ac763a2f0d679,0x312406633004d420,0xe7bc2fe2409c56e2,0xb02820d90141802,0x32a0060000420,0x0,0x2020000004000,0xc2208041402,0xc006912002104840,0x8480c00002805040,0x0,0x48000004000,0x858b2b7fc8d4547a,0x0,0x1800000801000,0x604612d000401481,0x4b7c230201c58a8,0x1000260004100154,0x529e3cd12494566e,0xc1c4d77e38155cee,0x450a394144146278,0x82222560d2141468,0x452e198000785058,0xd3b4ff74f4195668,0x8b67d2cff6d4556b,0x46e636dab2581467,0x994c0290101c28,0x8008a20002000268,0x416e55e8d4141cf8,0xc290fb2e340014a2,0x6bdf36bb2a50dcea,0xceafcf7934755c6a,0x79e293ec3654576a,0xffbefff3341c3eff,0xdb97fffff47cfffa,0xf9deee1bf6bc5ee8,0x43dfcd5eb47d7efb,0xb5f85ee8f41474ba,0x9ce363bb414582a,0x3902041c04901a40,0xc023d108009c5e28,0x200001000004000,0xc0507c0001cc340,0x514af7402430f42c,0xb62c710434d242,0x4584010614140028,0xcab69c411634c4c2,0x94385f406095166a,0xc4f4140a141454cc,0x30bcb5c084941456,0xb009322b01450a0,0x80b00a02240410a8,0x8040000000400,0x454d4c0e0001062,0x2920c3014141412,0xcbc0f03a36b456e6,0xd9429b4a74e414ea,0x25bec842001c128a,0x80808ce028001420,0x2404000140012,0x3000000008001000,0xc061ee2014dc14d0,0x84688410290544a,0xec0001006d4,0x10083088000c4a0,0x0,0x800408020000100a,0x802000000104000,0x0,0x40a00000000040,0x4080218080204460,0x8110c92834b454b8,0x2219dc8834905460,0x8603433270d01c20,0xb0522570041c3456,0x10a340020104020,0x90083021010cc00,0x0,0x9119cd91a6b1d2e0,0xa020220000105642,0x200240100044080,0x4004000600001481,0x580e16b554305ca8,0x1a020b4000941450,0x24000001008,0x180ad8030801428,0xc9002401a014d022,0x503080801c1002,0x210961b08101748,0x0,0x1180250000101a08,0x228660500145028,0x8b77f6bde0d456ea,0x20120450208410a0,0x100000200001008,0x1848a746303456a8,0x1a20074eb019d560,0x2ef87201e414123e,0xefded9dbfe1cd4fa,0xbf3a4fbab4b4d6fb,0x24020804000484,0x4a40200414210420 },
	{ 0x8b50c000c0c82100,0x826617426124224a,0x4c2888001080260,0xfde38e70d1ec327a,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xbfbfdffff96fb3d2,0xd53e33eef1cd27da,0xda263f7099cda3fa,0xe2a8ff62f044b2d0,0xa8ef3e0bd9e523c0,0xf3f75fdaf1efb3e2,0xd674f4e2d1e7b062,0x206f4768b10923c2,0xd381af7391c4a342,0x8801602a00c2042,0xab5eb7dff1e7b2fa,0xf5fffff7f9efb7fa,0xfdf8fff6f1eda2da,0xfebf3ffaf9efa7fa,0xc787f922d9c6b0ea,0xefffdfdff9edb3d2,0xefbfcf6ff9cda7e2,0xffffbffff9efb7fa,0xfff6fff2d9efb3f2,0xc9d6d73ad0e7a372,0xe8ed1ec95166b2ea,0x1fdecf93f1cdb3fa,0xd7fff7faf9edb7fa,0xffbe7f72f9e7b2d2,0xfa57eb2ed1eda3f8,0xbefeb7eff9e7b7fa,0xbfbeddcfd9eda7d8,0xd4bcffe0d16da7ea,0xfefaeff3d1eda7da,0xdf2ee37ff9edb24a,0xfefef7b2f1efb3da,0xfbfedfdff1efb3da,0xf4fdfef6f1efb7fa,0xffeffff3f1efb7da,0xfffe7ffff9efb3fa,0xfdfffbbff9edb7fa,0xfbffffdff9efb3fa,0xf7ffdffff9cfb3fa,0xfffffffef1efb3fa,0xffff7ffff9efb7da,0xefcfff3ff9efb3fa,0xeffffffff9efb7fa,0xbfbfdffff9efb7fa,0xfffffffff9efb7fa,0xdffefffff9edb3fa,0xe9f7fffff9efb2fa,0xeffffffff9efb7fa,0xff9feffbf9edb7fa,0xfffff7faf9edb3da,0xffbefffbf9edb3fa,0xfb52fb2ef9ed2648,0xddff3ffff9edb7fa,0x3fbfdf77d1efb3fa,0xfcfeffffd9eda7fa,0xbf9efff7f9efb7fa,0xdf87fbfff9edb7fa,0xfffffefff9edb7fa,0xffffdfdbf9efb7da,0xfeffffebf9efb7fa,0xfffffffff9efb7fa,0xffbc7f7fd9edb7fa,0xffffffbbf9efb3fa,0xebfe7f93d9eda3da,0xbdfffffff1efb7fa,0xf9fffff2f9edb3fa,0xdbb6affff9efa3ea,0xcb76dfedd9efa7d2,0x6f7efe9df9edb3da,0xbefbdffbf9edb3fa,0xd9fffe7fd1cdb3fa,0x8f387f6dd9eea2ca,0xe9fffb1bd9cdb7fa,0x6ed61f75d1e7a7fa,0xb39e5efbf9cdb6da,0xf5fffefff1edb3fa,0x7fee3df2f1eeb758,0xf9d0f32ed96ca3ea,0x4fdb9ad9f9cdb248,0x7ffddffbf9edb7fa,0xfdfffe6af1efa3c2,0x888c2d62716d12e0,0xfbbef7dcf9cca372,0xdfdbfed3d9cdb25a,0xbbfedf5fd9efb3f2,0xf4f4fffcf1cdb3ea,0xdfffff7ff9efb3fa,0xeebfff34f9eca7fa,0xcddfd72ad1cea7e2,0xefeffddbf9efb3da,0xb7fffffbf9efb7da,0xffffff75f1eeb3fa,0xd8f6bf75d9efb3fa,0xfd9efbfef1efb3f8,0x2f56abf8f1cc23fa,0x1fbb6f6fd96cb3fa,0xf8feff7ed16db3f2,0xfbf6bf7cf9ed23fa,0xebffffaaf9edb3fa,0xecc71fdad94db1da,0x4a9beb89d0edb2d8,0xd77dbffbf1adb3da,0x7ebefffbf9edb7fa,0xebd6f72fd9edb7fa,0x8fc3a20bd0cd26ca,0xafbfeffad96da7da,0xdc77bedcf1c333ea,0xafaeff73f9ed37d8,0xd792f77f58e7b2fa,0xcb7ef116f844a602,0x2b95cfc7d947a2da,0x16e44d4091252260,0x9da7ce5099ef24ca,0xdbea257ff1ed83fa,0x8dc7df8ad9e433d0,0x237ff752d0ed3750,0x189f0febd1c5229a,0x7100528891890240,0x82020609080a040,0x886a720a49432242,0x2210040040048040,0x95020566d1c4018a,0xc400003090008040,0x4d003f4159e5a140,0x8ab15622484c2658,0x24a3057d9c02350,0x8014449a80042148,0x84917440504283d0,0x4226060260841080,0xa184f930d12ca2d0,0x86603011010a2342,0x249005829008a002,0xd09c120080412440,0x80dee520d184a270,0x820000410000,0x0,0x40200000042,0x80040800000218,0x200140008000,0x0,0x60010400c2060,0x404000174100a04a,0xcf1deacd14cb3da,0x532ef742d1eca772,0x5a8b2fd4d08eb2d0,0xcfc6ffb7d1edb6f8,0x1c800a0d20c48350,0xa25658818c2068,0xd0e2967b7166127a,0xf4123950f94822e8,0xe9c4640091cc02da,0x6e5a1682c8cca440,0x33807a2d048a642,0x10008dc0102002c8,0x4884213101668270,0xdbf7f70ef1efa3da,0xa9c3bb4981ed2752,0x6b9fef7af16e33f2,0x5dd0377ef8e8b6d2,0xf7b63f61f9eda3da,0xdbbcfb7cf965a3ea,0xc9d63e9bd9c1236a,0x2f1fdbd9d1e6b3ea,0x74bd46eed9eca76a,0x9b2ef772e14c3378,0xde963f4051c5b36a,0x6fc6ef1ed9eca7fa,0xae7fc0ac061a340,0x242e041410200,0x9400008011400400,0x20108610d0003040,0x0,0x818134050402282,0x45ae7da4d1ada3fa,0x512a376bd165b7e2,0x4e8e35035864b0c2,0x8176300951240142,0x22f29208500c2042,0x800120000200,0xd858828000cc8648,0x2ae246050a42272,0x4104600240009042,0x800004c01080202,0x8700cc1b71a4a248,0x6411481040a462,0x2200256141488620,0x9a062625c1c103c8,0x900198041048000,0x869d1909d1c9a1c2,0x8890800000640308,0x43894058420250,0x102d1000090340,0x808840008080a040,0x140000000000210,0x0,0x8124202000402640,0x4082004001010060,0x68008328510ca212,0xd0080140012250,0xb6a8545291492542,0x5020054010003082,0x840060400080c0,0x0,0x21000740d0040048,0x1fba45ebd1e4a3d2,0x64317278006ca252,0xce00200251409248,0x8200010461000200,0xb460e00c10020e0,0x20004000000208,0x14050000000,0x0,0x1280254018408288,0x8fc233825184a29a,0x421c4748d06122d2,0x84300162400ca140,0x21285d73c0a4b2d8,0x128c8e48d00701c2,0xcf63ee2ad964a0ea,0x14e811041002232,0x660045290c08340,0x5120b6ded160234a,0xd00c200391410142,0xc9e3d26898ed22ca,0xef7fffdbf1cdb7ea,0x9faeedfbf1e6a3fa,0x300c260208400350,0x480ea122c0850040 },
	{ 0xc182f20b18266144,0x2a7c9ec30a2c4700,0x458852398302c06,0x92cf0648d8306534,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x9b8bdffbdabcefe0,0x10001409c0202000,0x550306090009080,0x900401000020a100,0x48b920190302600,0x2b6c17811828c720,0xa48c4628903a85b0,0x8967237292ba0570,0x733064445a202560,0x8fcbfbbfcab8ed26,0xbbafffdeda08edf4,0x3dbbd5fbdbb8bf26,0xfbfdbf2edabefde2,0xdae6bf6ad1baadb6,0x8dd4b3bad22ced56,0x69ffffd3daa8cf46,0xfebfeff39bbefde4,0xfdffffffdbbefff6,0xfffeed69d8b8ff76,0xc9d3fb2b9bbcf760,0x4cd73f659bbc6556,0xdfbbffdfdbbcfff4,0xf7f7ffffdbbcffe6,0xe3fe2d7adbb86d34,0xfbe5e3ae9abcfdf6,0xeddbfffedabcedb6,0x2bbfef5ad9be6552,0xf5d636ee9aacfd66,0xb6bebd719abeef86,0xf636ef5ddbb8ff60,0xd7febf67da38fd64,0xfbffdfdfdbbeffb6,0xdef7fffedbb8fff6,0xffffffffdabcfff6,0xffbeff7fdbbefff6,0xffffffbbdabcfff4,0xefffdfdfdbbefff6,0xfefeffffdabefff6,0xff7f7ffad8bcf7f6,0xffbfbfffdabafff6,0xefffffbfdbbcff76,0xeffeffdf9bbceff6,0xbfbfffffdbbefff6,0xffffffffdabefff6,0xfffeffffdbbefff6,0xebffffffdbbcffe6,0xefffffdfdabefff4,0xbeffeffedabeede4,0xfff7ffffdabcfff6,0x7ffe7df3d3befff6,0xfbf4eb7fdab8af34,0x8dcfb7d9dbbcfd96,0xfffecf7edaacfff6,0xfdfffffbdbbef5f6,0x9ffbfdf7dbbaff74,0xffb0cb7adab86534,0xffdffebfdbbeeff6,0xefffdfdfdabeffe6,0xfefffffcdbbeeff6,0xdbffef7a9bbafd76,0xfb9fefffd2acff34,0xefdfffbedabcfff6,0x7fffffdfd3a8ff66,0xbfbfffffdbbeeff6,0xfffffffadbbeeff2,0xfe96b77f9abae7f4,0xffd7f79fdbacefa0,0x2f7efed9d8b8ef76,0xfefefbffd33cefa4,0xf565a7f9daa82f46,0xdec465ecd8b8adf6,0xe9ffffbfdbb8ef76,0xc2d63e39dba8ed64,0x969fd5eb9abc6534,0x7f76f6feda38ef74,0xefbe7ff292beaf22,0x7385f36e9a386764,0x7bdabfbb9aa8ff30,0xd7becf7adaa6ff74,0x71be76eb5baebd24,0xfbbe6d7392bebf04,0xfebc7ffe9aacf774,0xf7ffe1b9b38edf0,0xbed2cddbd8a2fd80,0xb4d7deeddab0edf4,0xcffffff1dabcfd64,0xd7ae3fffdabcf7a4,0xff8fb2bbd234b760,0xebf9ffd3dabcbd64,0xfc9fffebdabaeeb2,0xdf7faffe9bbafd62,0xfebebfffdabce7e4,0xe9d7f737da3cefe6,0xeb7fbe47daa8e766,0x8772cbfb98b8afb0,0xdffd75eddb3e2d54,0xf63e7f7fdbbcefb6,0xf9f2f73edbba7d22,0xecde1fde91aedf32,0x993e4f52d10a2724,0xfd6da2c898bc6f26,0x13a877e298bab514,0x9a91b77f9aa0ed44,0x94a7b565aa8ef26,0x6aa9dfb9da242d26,0xf5fffffedbbceff6,0xffbffff7dabefff4,0xfaa66ec6d8ba6d14,0xcc6e6e3ddba8f510,0x51b6cb57d1b8afa0,0xf4fdf5eedabcfd44,0x393ec662d2b43524,0x780414a090002e00,0xc8e649b098348550,0x30815d9da226560,0x1c787ce290804530,0x9282ad40da982c26,0x9aa4264cc080ad00,0xc902d62858ace586,0x2100109092084400,0x8012006050802600,0x8018200200888920,0x82122e0408000502,0xe914963a4aa40520,0xd777ea1da28ad40,0xe014c4411a002412,0x402d1e000082524,0x0,0x100030800002100,0xc4230888820a800,0x830e440a92342104,0x440548690a02820,0x10000002004,0x0,0x41020890004120,0x0,0x432202010880520,0x50048000000,0x0,0x800060490000400,0x28821a0149ac2004,0x44a606e790a8ad24,0xd976af64d2884d52,0x88ac213ad030a120,0xffdefe4398befd36,0x4460008002400,0x22804228020bc00,0x6940021880280500,0x71063da01208e5b4,0x808cd2815880f110,0x40801c100820c480,0x5d8a4d03daac6562,0x90c2a20590a40444,0x1c0005000820c100,0xa9d6e37f9abcfda4,0xcdd27adb9b38d760,0x89cef709aa8af66,0xf5fffeffdabebd26,0xbffcff77dbbefff0,0xcb94f3b4cb382666,0xcc5fb29c9388d520,0xb3d7dfdbdaac6764,0xf1ff7eefd8bcf9e0,0x9b60afd5c82c6760,0xd6a4234f10286514,0xc96323101028ec20,0x6a68d94ad8a4bf44,0x48508611022a484,0xd26cbf44d8300f76,0x4606340098086402,0x4110000080080900,0xf0414100248600,0x369a43309a886544,0x8310050890384400,0x4020480000d100,0xe17a92384c20,0xa42124e1822ad80,0x1218c203902085a0,0x4406a0541aa82402,0x5eb06fd19aa4d760,0xca32017a9a3e4500,0x62020810288420,0xcac0c0a022c2520,0x800010600a300402,0x8100a700003881e0,0xc024b33290202600,0xcc40b00010881600,0x0,0xe41450359aa84546,0xb76f2f589ab4af02,0x5384794640a82720,0x800000400000600,0x168001082000700,0x0,0x100080000202000,0x2014c90080500,0x2c88431250326520,0x4856388399880c10,0x14a2446890203700,0x8000200080100000,0x6b3c7d76dabc2542,0xcb81ba6ad838cd76,0x406485888008c02,0xf1bfdffa9bbaf766,0xf7fbfe7bdbbcff74,0xebfe7ff3dbbeed76,0xcade23045838c530,0x44d23a1ad2306744,0x2410008008004600,0x14c1162010882500,0x88053000b02500,0x19a040190242804,0x147000010886004,0x188c90040202604,0x8010002000008c04,0x30802624c0800504,0x82202d4c5a248510,0x80a600498880100,0x84c41290204510,0x8010002200080422,0x4143846012200750,0x8aba0d7a99a8ad24,0xe886d02d9ab8e924,0x2fffde9fdabeef04,0x379b4ff8da38ed24,0xc158572ac2accd00,0x6902140298100c02 },
	{ 0xca3cd92e45067360,0x8bf7bfdf57967761,0x37a8d0400dd44342,0x2400106a0128a260,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc30a4d234d067550,0x141326a2b49e2160,0x0,0xe22453f014960160,0xa9467e99178e5762,0xefb7cf55077e7f73,0x1c9554e1450e3621,0xf96e7673b9b66362,0x99d561f11de25571,0xffc7e73391be7f63,0x95e454a1cce6d60,0xfdfeffeb9fa67d61,0xfdfcef7bdffef761,0xdaae3d2c8d866a63,0xefcfff3e7fd67763,0x2f7ede5a7e86ed70,0x96cb6fe2e74e7762,0xf5f7ff7ffffeff73,0xfffe3ff50fee7f73,0xebf7ffaf5ff6ff63,0xfff77f5fdfdeff63,0xffffffffffe67f73,0xf7f7feffbfee7771,0xfff67ff39ffe6773,0xfbf7ffff77feff73,0xcfdbfffbdffeff63,0xb7becff30f9ef763,0xf5dffeefffd67763,0xdb1eef77bf8e7760,0xfbbe7ffbdffe6f63,0xcffbfe1b5df67f71,0xfbffdfdfffdeff73,0xfffffeeffffeff63,0xffefffff5ffeff73,0xffbeff7fdffeff73,0xcffffaffffceff63,0xefffff5fddde7f63,0xbdfffffbdffeff73,0xff7fff7eefdeff73,0xffbebf7fffdeff60,0xedf7fffbdffe7f60,0xef5ffffd9fdeff61,0xffbfffffdffeff73,0xfffffffffffeff73,0xfffe7dfffffeff73,0xebffff7ffff6ff63,0xfff79fdf3ffe7f70,0xbfbfdffbfddeff63,0xd5e7fee7fdee7d71,0xcbee7f7bdff67f72,0xfbd4fb3a13cefc71,0xeecfbfaf5fde6f71,0xaffbdffbeff67f60,0xf5fffffedffeff63,0xdfffef77ddfe7f71,0xd989ff32ebe6ff62,0xddcfffabffdefd61,0xffffffdffffeff63,0xf6ffffffdffeff73,0xfffeaffaffde7f73,0xeffe7ffe1fd6ff61,0xffefffbfbfde7f73,0xefffdfdf7ffeff73,0xffffffffdffe7f73,0xfdfdfffbfff6ff73,0xbeb6bff7fdfefd72,0xefe7feb79fd6ff60,0x9cdbfb5f1db67f61,0x9fbedff61fce6f73,0xf7b7bf7f3dee7e63,0xdfee7b6ad3fcf763,0xcbf7f3fbdfd67f60,0xcfc73fde99f6f563,0xbaa8effbcf9e7762,0xfd7676f79fb67770,0x7fbe5fe2bdee7d62,0xfbe7ff2f3dd6ff73,0xffffffffeffe7f73,0x3fb0fff9ddde7162,0x68775fea0fd67f63,0xb7bff7720ffe6f73,0xf738ede155e23571,0xcfcffe963dd65f62,0xae7edd5c17c2ff61,0x56bbffe8ddfe7f73,0xf1eefff16fce7d63,0xffbe6f6f95967773,0xeceffe0c9dde3761,0xbf5fdfd38ff6fd63,0xfc2bdfe785ce7f61,0xfb5daf7e9ff27573,0xfefb2f7e77de7f61,0xe9f7ff3fcd5e7d63,0xef7e19cbd5b67d40,0x8cfa1dfb1fde6f61,0xb3fdbf7d3be6fd72,0xffbe7fffffeeff73,0xa9dde33f5df66f63,0x3bf37ff097be7f62,0x328a26dc1cc65102,0xbd77e3f01dc65d63,0x9e2805fa0db60560,0xfad7e63a7bce7460,0xc96437df3dc24c62,0x86bacebf4dd67763,0xfdfffffffffeff73,0xffffff773ffe7f73,0xd290e76419864e60,0x1fd6b62a8df66771,0xab5e0fd30ef64760,0x409cfeea4dc63360,0xf9ceff7bbfd6f761,0xd692077a15963160,0xc4d3aa38578e6521,0xa5f5fc663c66d61,0x804a100a43062d40,0x9ac325c295167d21,0x8782a66021860570,0xc824dd3c13ce4e60,0x6bfd1b51c5b2af60,0x8010000005000000,0x9280281722084423,0xc1005100100,0x380508167406420,0x40080603850c2020,0x800084001002200,0x750414d2cd407a00,0x1000040000024022,0x0,0x0,0x0,0x0,0x0,0x0,0x800e0001020100,0x0,0x14000080020,0x8402004005020000,0x0,0x200200004040,0xbe81e51874a7463,0x7832d7e00d845760,0xb124a24881127942,0x8004016008940500,0xc5c0d214040a4560,0x80462a6280866640,0x1408400009400200,0x8010201108403000,0x210001400004821,0xa026ca0684846762,0x443320010424560,0x2330440a108a2960,0x910200600a4e0420,0xc38270780556a160,0xea98e62e1dce6462,0x5dcbbf3f4dee3d70,0xaf3e4ff85fd63560,0xd4d55be88d027551,0xffdeff73bffeff73,0xea98f73dcdaeff61,0xcff7be3305967f41,0xeadecddfd7966f63,0x64befee61cd6f570,0xfbfef671edc87763,0x8400396080002820,0xc842b3b88606f560,0x1200000040002000,0x14066e864d066142,0xa54d27482d527522,0x6c9c2c7324863f70,0x0,0xa20d504000283461,0x23256480f425542,0xc53c3d6b8dc4f360,0x6d3025621d9eed61,0xe980d332078a1760,0x2943361001960640,0xa020e0810860160,0x9c540480810e2860,0x8605c004446060,0xa214e56e02960741,0x8008308100424820,0x8c16474039c40860,0x1180180a09c40640,0xa00000000840000,0xca86e26414b60551,0x100800000000410,0x18002748088c4462,0x54f5fefdaffe7f63,0xbfefffff7fde6f73,0xfb9eff7dbffeff63,0xc040090040823200,0xa0151b5008486c62,0x0,0x8000060001424000,0x8a10024400021540,0x45cc500004424470,0x1632a344a66651,0x120a0f1340c66221,0xa0248d5018f2a640,0xee9609f06d966763,0xc9daea670dd66f63,0xf583e0187066d42,0x8bbb7ffb7fd67760,0xc479f367d586e462,0xfbfefff3fffeff73,0xcbf6ffff9fd67f60,0xc7cffe2bcd7e7563,0x8880812903021740,0x5088826009c04640,0x0,0x8000004002000400,0x8020000000840000,0x12410084002500,0xc15006040126142,0x462a5c088802640,0xf292094c08d66341,0xa0c7a8000a420560,0x290000180426000,0x5a0003a09c4a430,0x84c9221028527860,0xd2a63f6e759a7961,0xc75fe07b95c67b60,0xebdffa5fcfd67f62,0xbeb24cbe55a26160,0xd3befffaffdeff73,0xdf9efe72cf4e7f72 },
	{ 0xcbd2ef6ef69e3f4a,0x403a126b47d92c08,0x939ec433d7280fc0,0xb0d2463877982648,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x42805000a012800,0xbc70da61f71e2c48,0x87163522dd940ec2,0xcf956f77ff9c3ce0,0xafc6febb77bc7df9,0x2b29074244d06460,0xb4f9bce0742c3c42,0xd0c0646212147240,0xd99c6d77573c2fe8,0x4842fe36d63f6e01,0x135552527aac2ce8,0xb4fdd7efdefa6fda,0x91ed2f7a5ebffe08,0xffbc2de4deac2ee0,0x4feef7aadeffecea,0x6afa7ec2f6bd3f50,0x5fba4fffee3d2ca8,0x35ced768df9caee9,0x71b67e55de1e3c50,0xf37bef6bfeff6fe8,0x4dff3f15dedf3cf0,0x9fbbfffefedf7ce2,0xd6fff7fdfeff7de0,0xeffe357bfe9f3dc0,0xfbf6f7ffffbf7fe9,0xffd7becbff3eeeb2,0xfbbecffadeb77ee0,0x75fff6ffdebcfde0,0xf6bafdf3dfbfbee9,0xdfee7f7ffffe7feb,0xcfffff99df9f7ee3,0xbfffdfdfffff3ffa,0xffffffeffebffff9,0xfffffff7ffbe7eeb,0xffbeffffffff7fea,0xffffffbeffbf7ffb,0xefffffdf7fff7de8,0xffbfffeff7fe3ffb,0xfbffff7a7fdebffb,0xfebeffff7ffffdfb,0xfffff7ff7eff7ffb,0xfffffefffffffff9,0xbfbbfffbfffcfff9,0xfffdbfffffff3df9,0xfffeffffffbf7ff8,0xebf7ffffffbffff9,0xffffbeffdffefffb,0xffbfefffffffbeea,0xfdfff67ffffefd70,0xdffefde3d7bc3cf3,0xfbf7e3ffff9ffefa,0xefff7e9fffbe7ff9,0xfffffffb7fff7efb,0xddffffee7ffe6fea,0xf7eef7f3febebefa,0xbbefff7effbc7fc8,0xcf5ffe3ffefcacf9,0xffffdfdfffffbffb,0xfefffffffffffffb,0xefffbfff5efc3e69,0xfbbf7f7d7efe3feb,0xffcbff3ffef7ffe0,0xff7fffdfffbf7efb,0xfefbfffffffeffea,0xd9fde77fffff6efb,0xfebe3ffef7bf7eeb,0xcbf7ff7f5eef3deb,0xffde1ffff7ee2dfb,0x96b3b7ff76be7de0,0xf7edb7794fff7de2,0xc72635f1dedf2cc2,0xeb8fe33ed6e03c61,0xc2b35e40f6a43e40,0x9ffb6ffb77d67de2,0xf776f77ef2be7fe2,0x7ffe7df25ebf7f68,0xdbd2e7ffdeef3d68,0xe9e6b6f3d7df3ee9,0x4e9cd7df5ebe3ec3,0xf5ffffe17eddaf41,0xbfff7f76ffbe3ed2,0xdbacf76edebe3f41,0xd7dbbfbb5fbf2db8,0x93dacf1bce9aad18,0xf4befee5de1ebfa3,0xdbefdf62d69c7cea,0xfbb3ef7bfbfe2ee2,0xdffeffbbdebe3efa,0xafffd7dbffbffd21,0xb4e705e8f71a2c4a,0xf9bdff7fdedc3f88,0xf6bebf6fdebc3ec8,0xcf5fff365fbf3cea,0x489a3e725eea264b,0x96bacef2d6cdbeea,0xf76cfff1ff9a7ce8,0xfffe3ffefebe3ffb,0x817eeb6976fe3ec0,0xfbd77ffb5effbddb,0xfdbecd71539e3ce8,0x72c4916943143439,0xdf8e3c30de802c60,0xcb9ee37e7fb2ae68,0xc85eb20b759f6ce0,0x27ff5ffa7cfe7ef1,0xfffffffffffffffb,0xfffefff7fffe7ffb,0xffb7695ac6be7de9,0x8da3f31ad6daa460,0x325313c0e6801450,0xa8500438149648c0,0xfbcf777bdfdfffaa,0x4eba2f75762d3f48,0x9145ff224fb42482,0xc5c5048069424e0,0x3b0f7804e952980,0x82a8a56aaa106d08,0xca800520442404c0,0xc112873a43d429a0,0x22319cc182072ca0,0x11841320a842480,0xc1ee06ca579a2e40,0x80244c42902480,0x8c1012002b00400,0x40000800001400,0x88440081b00c00,0x201000200400,0x600044082200400,0x800000844000000,0x8000000102002000,0x0,0x0,0x20040000800040,0xb720094854a00240,0x4080000000002800,0x1000010040000080,0x10001000000420,0x2100044000000,0x84000008000400,0x0,0x17f4c4056303cc0,0x2000002040106000,0x0,0x280c0042000400,0x4080c20444041480,0xe3e83c18e922d00,0x10024206100400,0x9480841040013c00,0xf886fde2422e3c88,0x41c2a042568c2420,0x0,0x211480002802441,0x4400a2e060003c80,0xc9bc37327d9e2cc3,0xa85c420062b06c88,0x9efa3fd7f7be3de2,0xe79f5ffbf6bf3ff9,0xb6f6ffeffead3fdb,0xbfbcff77dffc7dd9,0xffb6ff7b7efe2ec3,0xcdc3b733f6367cfb,0xdb7fdfdbf7963de0,0xfd76fee4df3e7ee1,0x986aa7e0c6232cd1,0xc9806c2244101410,0x5000110a46122c00,0x210000000900000,0x8c54c8aa020d00,0x4100b46056042880,0x2143468d6081c40,0x3800844000000,0x229398ca74002d48,0x900430006000460,0xd02c35708ac22c20,0x4a222c40562c2d48,0x8970faf07e3e7e68,0x6fff3feb7f1ffd62,0x2040617ca003400,0x2404026e40de4802,0xe6da2120821c2c20,0xc094832c061a3c00,0x1860000144020,0x2622dc103b040448,0x1d5404081a0c0480,0x979c25225e106a00,0xab92ef4fd3fc2d6a,0x8160c12a56802240,0x4a00030840822840,0x1cf3f6e07fb4bee9,0xddcfffdb7fdd7fe8,0xdfb8ef7ef6b67da8,0xf9dafb325f3c14ca,0x8ebcc8508a982d48,0x280040002002800,0x100054001a000040,0x9b008660569c0440,0xc180d122c6982c80,0x9411c064a003410,0x20b0200842042080,0x90682753d6843fc2,0x6aee2d74df1c7f13,0xcbccf646769f2c82,0xeaf6dfce479f3ce8,0xefbecdebffbaaf08,0x943cd642cf1a2d00,0xfffe7ffbffbffffb,0xfbfeff7fffff7ffa,0xf8c2fe9576b62ca9,0xb317066072a22e41,0x10c1040054100402,0x9238240070222051,0x20000200040080,0x0,0xa210440005140020,0x448524182023c01,0x56352ec2242,0x8200010442003410,0x100a322246c02448,0x1ac495047143d80,0x4143d32354924c68,0xc075aa3054340468,0xdf14bfee76be3ef8,0xedbed275dfb82fc2,0x2affff5ffff63de8,0x9fbf7fefd7bb3de8,0x30251440002c32,0xcd467b61ae906400 },
	{ 0xf9feef6a9b966f2e,0xe7f62a9a9ac72726,0xc38cfd99ad12f37,0x8c1054e0aa9c2214,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80400080040000,0xf6f7feffebdc27f3,0x9b64ad629b8e2716,0xfd0cf7ff9b8d6727,0xefd73eafbbf82b3e,0xbbe2dbdb9b9c639e,0xf6bfb4ee9286234f,0xe9ceffe48bbc230c,0xc98617379ac26b07,0x88c7e30bc3816f8e,0xb5cdd4a9b046f04,0xbfff57bbbb892fd7,0xd17aaffa9bdc6bb6,0xfa92abe09bd46bbf,0x49d7d6be93942f06,0xae9f7bc39bf52f1e,0xf7eb5ff39afc6b3c,0xd5e8b7c8ba982717,0xb3362d60cbdc2033,0xcadefbab9bbd6337,0xeed51a559af22334,0xbfaedf7b9a9a6b9f,0xf1f5b6fd9b196f7e,0xe7feadfafa9f6f3e,0xfb9fef6fdbbe6fdd,0xffea3ffdbbde2f36,0xefaf5dfbd3bf673f,0xf5f7feefdbff6fbf,0xfffffd70db9c6fbf,0xdfe7ffffdbbd6fbf,0xc7eebe1fbbdc6fcf,0xffdfffdbbbff6ff7,0xfffffffffbff6fbf,0xfffffffbbbfc6fbf,0xffb67ffffbbf6fff,0xdfdfff3edbdf6f3e,0xffffdfdfbbfd2f3f,0xbefe7ffbdbec2ff6,0xfbfffff6bbdf2f37,0xcebf2ffffbfe6f9f,0xffffffffdbdf67fe,0xefffdfdbbbdf6f3f,0xbffbffff9bff6fdf,0xf7fffff9dbbf6f77,0xfffeffff9bff6f7e,0xfbfffffffbbf6ff7,0x6bdf9ff7fadc6fe6,0x9f9ecfbb9bfd2bb7,0xf5b67effbaaf2b7f,0xffbe7d73faec6776,0xfbf6fbef9bbb67ff,0xf7fbffdfd39f2f7f,0xefbfdf7f93af2fff,0xf7fffffefbff27ff,0x9ffcff77fb9f671e,0xfeb7fffe9b9c6fbe,0xafdefa93fac56f7f,0xffffdfdbbbff6fbf,0xf7ffffeedbfc6fff,0xfd6cff3a9b9c6f37,0xdb8fbf7fbbdd2fff,0xffc7fffafbdf6ffe,0xebffffdffafd6fbe,0xfdfffffffbdf6fbe,0xfffffffcdbff6f3f,0xfffcffffdb9d67ff,0xeff6fffd9bdf6fee,0xefff5edffabd673f,0xbffb7ffbbbed6b7f,0x97fe7f4a9b942323,0xbbbe7f7acb992b1e,0xc9bef3669b8e6776,0xffda7a8fbb98673f,0x3fbbdf7aeadc2bd7,0xd514b6768bc4675a,0xf7fefdf39b9f6fb7,0xc3d5ff2f9b8d632f,0xf6cabf4b9a3c6355,0xeebfcffb9b1c673d,0x7f7ffff58bfe633e,0xdfddff77dbfd6777,0xedfbff7fbbd46fb7,0xefd7febfcbb86b5b,0x3bdf99429aba6313,0x9dff7eaa8bb12516,0x11ecbf2892862716,0xf99ef77e9bb46f07,0x67def71ffbde2f56,0xab8dffdbba9e273f,0xf66357e99b872f56,0xfd7faf3b9bb92f9e,0xfebebff3fbff2fff,0xefc3e36adaac273e,0xafde7e539b886faf,0x3e2a6d539bb0623e,0xfb3dff7282be2f5c,0xfafe7ffddbdc27b7,0xdbb9ff4fbb9d6326,0x9be71bfffada6f1f,0x98b9cff29a05430e,0xf23672739a5c0306,0x6e862fe99bdc292f,0x9a81a7279291233f,0x2cc124183840140,0xa3becf72dbd4671e,0xfdfd7ef9fbfe2367,0xdfbbaf779bbc6f7f,0xd9f1b73792a86e8e,0xa34cbc02a2912133,0xbb4c07da19402787,0x740262e082112107,0x656ea77182ba6f15,0xc7ae39678b9e0312,0xced29103a3902712,0x20a1542880c2281,0x49743789a902323,0x8010258092000334,0x44062ec993a02a17,0x6cc4426480042205,0x8f6dac288802312,0x8038082102005,0x4142701b8102215,0x4084018c8a222004,0x280600209002000,0x8080000000800010,0x80800000040200,0x5855300098a02310,0x50a404108b100302,0xdb91ef0682102b14,0x4483fa3388142387,0x16ccc082002102,0x42202a080002402,0x0,0x30a00e629a08092c,0x8060a428a802050,0x0,0x298c46089002208,0x2100002000000,0x820000000a00,0x0,0x35354408a080140,0x588080010102,0x0,0x10001000002400,0x8950d2008a040300,0xa44837c81a482e04,0x0,0x60080000100,0xca9eac7092862730,0x890a400098506301,0x60870890930e0b02,0x1900010283000006,0x841422420b000304,0xdfa6fd629b812b57,0x8104ab4cb2002026,0x6c7b3e919a8c678c,0xebb6c7f893972336,0xfcfffeee9b966b3e,0xa954ba029a99636e,0xdbb4f37bbb9e6737,0x4fd7af1edb932336,0x9a0effd3fbd42f96,0xdcbd7ee98b996377,0x690c27e292844b36,0x10306a029000611d,0x6400c02888c4200e,0x64014498082004,0xc23204620ac12022,0x100140880802002,0x2810255c81242312,0xc0a0880019800822,0x40000800010002,0x3490600282140214,0x149048080000018,0xc9e2254182306304,0xc3fcf59bbb556166,0x2600000090100100,0x1028a900200,0x10e270c08840220f,0xcd242c6a8ae0610c,0x4000240080102000,0x49c3121093122007,0x8021406082106040,0x10e412088a400014,0x9e0201508280268e,0xdb83fb33bb302f0d,0x10920000802420,0x300c0202262200,0xdcb6f2f59a962b1d,0xb37b7f7b9bde2b17,0xebb2657f988c6b3f,0xe7eefe0afb6d6b1f,0x2edd5d9abbc62336,0xb8aacbeb9a9d2327,0xca69a542bbcd2317,0x5a0e2c70dbcc6800,0xc8c2820490146381,0xc075168393840324,0xc010052098400842,0x99eeb7499b596b0f,0xdfcebd46bbd42327,0xcbb6f1bbbacc6736,0x6fd71ed19a980b26,0x4c86081100000,0xc4222010a0846402,0xfffefffbfbff6fbf,0xfbffff7f9bbe6fff,0xc5d8102f9b000b36,0xfbcddcbbadd2707,0x2410088a040000,0x9a98ad709ac44b0e,0x1804000888004200,0x8320040210090206,0x102014002000006,0x54845420c184630e,0x890a5f5a82276a0e,0x5090a3449a802316,0x444a7ab8b820231e,0xa2bccf4a921c6324,0x110a464293904b12,0xc0c0270282a42404,0xdff63ff6db9c6347,0xadffff33baa4271f,0x4efdde4f9bd12fcf,0xbd39cf739bbc2b32,0x800024218a0e0000,0x3dfe7d739b1c6f1c },
	{ 0xebfffffef1bddf7e,0x6d4612996198c3bc,0x3fbceffbf1b5df7f,0x8c121288f000c00c,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf4ff76effbbaffff,0xb3392c55e93457ae,0xffff2f7febb4f7ef,0xa661203d110d32c,0xa9da6d5f81b0d7ed,0x76f7f6e6f1b0d3bf,0xfdbbef7efbb5d3ef,0xd7956975d1a38735,0xa0d2ec42e8885198,0x3b5e595ec084d06c,0x28ba4be9e190cb2e,0xf9acfe3ac39bdb6d,0xde3a2f6ed13bdb2e,0xffffffeffbbedfae,0xefffbfdff1b0dfbe,0xa3ccfe1cb20fbaa,0xd1cdb660f1b2732e,0xff9eabf6e1bc533d,0xe394b372c080f37e,0xea4f5f5ff1b4d32a,0xaffbff7be90df72e,0xb4a3f3e9e998c338,0xdfbebfd3f1aeffbf,0xfbd5f72cf9badbfd,0xefeeffffd9bbff6e,0x3fbbcddbeb95dfa9,0xf5fdfeefe1b7f33e,0xfffe7f72c9bbf3fd,0xdfb5ef7fd1b7d3ff,0x5fdebe3bebb5db3e,0xbffbffdbf1b4ff2f,0xfcbfffeffbb5f7bf,0xfffffffbfbbfffff,0xfffffffffbbfffff,0xfffffffffbbfffff,0xebffdfdbf3b7dfef,0xf7be7fefeba7f7ef,0xffffff7ffbbdf3fb,0xfebe7ffbf9bfd3ff,0xffdfffbef9bffbbb,0xfffffffff9bfffff,0xfffffffff3bdffff,0xfdfff7faf9b7d7bf,0xffbe7ffffbbffbff,0xebfeffeff9bddbff,0xfefffffff9bff7ff,0xfbbfcffff3bbf7fe,0xddf7f6fbe1b1f7ff,0xdfe63dfae199dbff,0xf3bfff3ef9b5fbef,0xcfffff7bf9bff3bf,0xbfffdff6f9bcdfef,0xf5ffffeffbbddfff,0xfdfeff73f995ffff,0xffbff7fff992ffef,0xcff7fe1fe9bdd33f,0xffffff5ff3b6dfff,0xf6ff7ffafbbeeffd,0xdffef7faebb3d7ff,0xffbe7ffcf1bdd37f,0xffdfffbfe3bdf7bf,0xebffdfdffba4fb7d,0xbffffffbf1b6dffb,0xffffffffd19fdbff,0xfab7bffff9b7f7ff,0xefffffbef9bfd7ff,0x6ffffffbe3bff7be,0xffbf7ff3fbbcf7fe,0xff7df7fde9b7d37b,0x7384adefe1a1f17b,0xabd6f3fec93edf24,0x6cdb5ff9e1bbd36b,0xbb3bcfa9e13cf36a,0xf75fdf7ef383d7ab,0xdba6f5e361b6c3fb,0xcbe4ff6ef937d3ac,0x4dcf3691d191c33e,0xbffbdf5bd0b0dbfb,0xf5fefffaebbfdbee,0xfbaffdf2d9b0ff7f,0xc294ffbef13ed72e,0xffdfbe9fe9bed3af,0xaa5a97def9a4d33a,0x809ae670a3025a68,0xe36a2762d924532e,0xfa973fd4d184d3aa,0xcdcffe36f8bddbeb,0xaffd7dfe1b1f329,0xc5f6ffedf19df22d,0xd1afff7ff9b87f7a,0xee8a0d5e6192f17a,0xc7dff6f6fbb3f7bb,0x4bdfffd9e9bdc3ee,0x9dbf5f6fe9bff32e,0xf7febf7af9b6d73b,0xfffe7ff3f9b9dfef,0xebf0f737c3bdd7ba,0xeed316cff19dd364,0x85a3cf12f0a0d209,0xd7c0f6784088c33a,0x77d625f2d9a5db38,0xda30dbfecb35f7be,0xb5b9ec8d198d000,0xffbfdf45f1b0df7f,0xaef776fbcbbdd33f,0xc88ebc51fa3cc7e4,0x620065002001c108,0x54cbfe38f3305f3a,0xa51817d3c0a5d73c,0xf43b76f46985c12b,0x7faeb768e83c4f2e,0xffbf7f74f9b5f3ec,0xdfe7ffd9419ff7eb,0x7ae09dd189bc5568,0x1811e84041108342,0x5daedcd441b8c369,0xc200260241100302,0x2e726ae110912e,0xbc71ab9c992cb7b,0x8e122df04181d14a,0x822860000108,0xc1a0047ce200d024,0x184c31851900120,0x41180001000060,0x692d7e0d0a0c19e,0x3934f2d609964371,0x0,0x200400000000102,0x21000004002,0x100001000800101,0x400002001000020,0x88102012018040a1,0x900400208084000,0x40200000100200,0x14000004002,0x48001000020,0x12043460a080d002,0x4042000000160,0x0,0x0,0x4000200800001,0x200c0000201200,0x0,0xdc20e01e000d004,0x2c48a944a0904384,0xc228067051029228,0x80a42504ca005004,0x960c3c604080d1f2,0x8b00a032d121d342,0x8000a0e61029208,0xb2924000e020c208,0xc022900058004102,0xceaced4881b85727,0x9988d7b0f390d738,0xcefffabbe9b1d37e,0x2f9ecff8d191d3ef,0x30e0c568c190832d,0xed4cfd61eba7db7f,0xdfac775ee115cb2e,0xc6cb5630c1a0f33e,0x8a1e4d49c131c328,0x34a12ca950805289,0x1b68e622c180e22d,0xcbd2679f7021f30a,0xc2a9d215d13058e2,0x1484848c0041922,0x81a806424020c328,0x10800600008c020,0xf88403228280d21c,0x4080c08080904008,0x215090400100c10a,0xab324ffac9b6f6a4,0x814004400900c120,0xd2a2014e83a453e8,0x80085208020c008,0x22681f532891c31a,0x805920181c209,0xd7a69078099065a3,0x25280c0000a06334,0x8953d18c7121c7ac,0x20421a8a40108328,0xa0048100100c128,0x5dfc7e0bf330d30e,0x0,0x4280c20ae1909149,0xcc3af1ae113832a,0xc21f52dbf184d128,0x74d550a8c1141566,0xccae2f60e3b0daa9,0x898e7726e1b3eb8f,0xeffeff3ae9b7db7b,0xbfc5b51dbb4d16b,0x97bf5de6d134d728,0x51ccf64a8134c3e8,0xca2c1b4af1a1d32a,0x89dbf72a81b2d37e,0x31a10c1001020,0x22a1cf8b89384624,0x92082415c131832c,0xf3b02c7001a6d3be,0xa957f200e9a2c37b,0x9c21a01c0a0d308,0x100c410288005120,0x34002042085000,0x50b62c5001901729,0xdadfff7dfbb2d3ab,0xa8e39710818c51bb,0xdb9edffad1b1f7ff,0xc0108620010c4128,0x281500190030b,0xc00008020000c008,0x18302100848020,0x8802490041105080,0x100452684080d300,0x2804466051040160,0x52022a016111c000,0x90206a02800051a8,0x2752165279804188,0xe082312001804103,0x332ce6415301e328,0xfb982f7ff993cfbf,0xc1e7ce2ec1b4db7b,0xa1eb5d7c995d36a,0x843a6f73e112e378,0x712c26080a88417b,0xff3e7f7ce3bdf7fe },
	{ 0x4bbdf3f20bd4c38a,0xeb6606868bc6e06a,0x7ab8c9e388d6f96e,0xdc766ca689d6cb2a,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xeffcf36f8bd6f362,0x9eeafd728357eaf6,0x92b530678a44385a,0x0,0xa754401803c48202,0xd6f2b6e583d7e3fe,0xed0a6f6383d5e3aa,0xcd96392303d2e322,0x86cf59d38bd6f372,0xaf18dc8f8354e2a2,0xf69354a18354f3f6,0x716d266083d5e306,0xfadf2f748bd7f3e2,0xe9c5f3a709d6f3ea,0xf25eb7d28bd4e3e6,0x969047e081c2d1ca,0xb7fca75983d7f3f2,0x3fde2fe083c7f3f2,0xeba3f52b8bc6fb6e,0xeac31ea78b45a8aa,0xf79fdffb8bd7f37e,0xfc7fbf7f8bd7f2fe,0x53a8ef6a8346b3ea,0xfbb7ef7e83d7f3ea,0xda5bfeff8bd7f3ea,0xbfbfcf5a8ad7fb62,0xf5d73f7b8bd7fbfa,0xfffefff28bd7f3fe,0xffacffff8bd6fbfe,0xdfd3f6bf8bd7f0fe,0xcfffdfdb8bd7fbfa,0xffffffef8bd7fbfe,0xbbffffff8bd7f3fa,0xffffffff8bd7fbfe,0xffffffbf8bd7fbfa,0xffffffdf8bd7fbfa,0xfdbfffef8bd7fbfe,0xffffffff8bd7fbfe,0xdebfbfff8bd7fbfa,0xefdffffb8bd7f3fa,0xafffffff8bd7fbfe,0xbfffffff8bd7fbee,0xf7feffff8bd7fbfe,0xdffe7ff78bd7fbfe,0xebfeffbf8bd7fbfa,0xffff7fff8bc7f3fe,0xdfbf6ffb8bd7fbfe,0xf57fdeff8bc6f3fa,0xdfbe757b83d7fb6a,0xfbb8ff3f8bc5f376,0xeffbbeff83c7abf6,0xffffefdf8bd7fbfe,0xf5ffdfff8bd7ebfe,0xffbfff7583c7fbfe,0xfbfcff6f8bd7fb7a,0xcfd7bf9f8bd7fbfe,0xbffeffdf8ac7f3fa,0x7fffd6ff8bd7fbfa,0xdbcafff98bd7fbea,0xffbe7ffe8bd7fbfe,0xead6ffb781d7f3ea,0xbf7fdfda8bd7ebf2,0xfebfffef8b86fbfa,0xfdffffff8bd7fbfe,0xfebebfff8bd6f3ee,0xefdffffe8bd7f3be,0xefffff5b8bd7ebf2,0xf7ffefff8bd6fbfa,0xf1ff7ff98bd6f33a,0x7f3ebd630bd6a3a2,0xabffffbe8bd7fbfa,0xefdfbfff8bc7fbea,0xbb9fcfbb8bd5f1ea,0x531ef2f38b44f3ee,0xcfbced7b83d7b31e,0xfbfffbf682d7f3fa,0x8e7b7bdf8bc6bbee,0x77bedf758bd7b3a2,0xb5f6ffeb83c7e3fa,0xdffeff738bd7fbfa,0xefbcdfff8bd7ebbe,0xcffa9f298b56f9ea,0x6f3e5fda8156f3f2,0xb4fefeec83d6f3a2,0x9fecad6383d7b372,0xf3967e768346abe2,0xf5dffefa8bd7ebe2,0xaffedfde83d6bbfe,0xc5ffffef8bd7f3fa,0xdb6fa6db8bd7a3e2,0x7e3c2c7d8b56e1aa,0xc1c7f73f8357faea,0xef7fffdb8bd7fbfa,0xbfbfefe28bd7f3f2,0xd1fefffa83d7f3fe,0xfefe7fff8bd6fbfa,0xabbcf3be8bd6f3aa,0xefff9f7f8bd7f3f2,0xb21ffdc08b55e3a6,0xf534d6f68b57d3b2,0xaaacff2382c2b962,0x8306d26a8b44300a,0x81660f4882c4c052,0xb14cfef8ad7e17a,0x86f1f6e889d4b2fe,0x9276ef228bd5e2ba,0xc906f3388b469260,0xc943f237825672a2,0x80184a420146a0a2,0x94bc8eee83d4f33a,0x5940024009c42012,0xcf3e7ffc8bd4f3ca,0xedf7d8958356f2ba,0x1abfff598bc7c372,0xfcd777ff82d7fbe6,0xf1fcbefe82c6f366,0xc2362fc48285c902,0xcc2f20301d621b2,0x2c36e2c28b54400a,0x101b4de080469132,0x4000a64288542200,0x610004c00068220,0x144a52000c4a022,0xa800e0202522022,0xc10480000568220,0x0,0x0,0x2212400080849002,0x8c83280282859002,0x222004000200a200,0x943364e68a46e16e,0x10ac0c608294d042,0x2801014000c0,0x0,0x1a8200a000,0x8000040000020200,0x0,0x8080080000004,0x4020000040200,0x0,0x120100b0300a0a0,0x50a0040003840002,0x440008000101c000,0x40c00800818000,0xa413160108542022,0x0,0x4069052883462100,0x460e3d640306a232,0x0,0x8d41e100145c260,0x81ce8c98211f102,0x712ca2e90856a33a,0x73be7f6381d6fbfa,0xb90ffb2e83c7e26a,0x8ffbbfdb83c6fbee,0x26be4f618bd7fb7e,0x1430080046002a,0x192ead208ac4b342,0x9a82eb7c8bd6bbfa,0x82c3972b80567260,0x8158cc190046a12a,0xd4b37ae483c6f3e2,0x4826768254aa02,0x30464310a028122,0xc89db2289867ac2,0x40d0815980862aaa,0x2440b380529340,0x1024000400040212,0x90010000060030,0x0,0x941104409460342,0x29b459b88b87a3b2,0xb4a470600356b030,0x645000402102,0xcb803b6f82d6f162,0x60c41fd00bd4800a,0x441c44c481446298,0xd5acd2518954aaaa,0x398615e38917f332,0xf31477230ac7816a,0x802258000002a2e2,0xa020c2230054b030,0x829516420150a1e2,0x508002000040102,0x8184002400062220,0x94410800205a212,0x6b0cd1108204aba2,0x71547882c3e160,0x98025508244832c,0x9bbf3f718bc6da6a,0x8ddbff3b8bd7fbfe,0x2b5ed3d189d7f3ae,0x142147e10bc483aa,0xc85420420a54900a,0x6898afbe0b5660d2,0x894ffb2b83c7a3fa,0x3886d2f500d6b242,0x87120c360854a2a0,0x2065360c8bd6f122,0x828c388000559012,0xa9d7f92b8b46ebaa,0x6c4b82108144f0a4,0x800480000458000,0x4000200002008000,0x4000000200020002,0x4000000200008100,0x142000a00808200,0x3b8442c03d6b372,0x11d206a08a02aae2,0x10b0014000002122,0x491fd3798356b3c2,0x840161003042002,0x1481200c48000,0x45446e38284a3a0,0xc80026008300e202,0x14202e448086a162,0x3038e228804a102,0x108000040200,0x100000200040200,0x14025008200e020,0x9ab493368386e3d2,0xcb97a6b28bc5f3aa,0x687e96c58387e3ba,0xdfba4ffa8b56f3fa,0xf1e9efeb8896db42,0xdfb63ff68bd7fbee },
	{ 0x68dbf38b0f473e08,0xeaf63f53c5c33a14,0x17dacdfa435a7a34,0x5df7defacfd3fab4,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf95d17622fcf3a99,0x5bac2d71cfce3e10,0x700051fc44ca1214,0x6cd92a647cbfa8c,0x2bbedfd3c8ca3e3d,0xf4bfdee44fc97e10,0x53ff67fb48ca1eb4,0xdeb02f6e47c27e30,0x346a95246871210,0x1affdf9fcfcbfe34,0xbebb15ebc5cb7e1c,0xf8c4f67acfcb3a3d,0xf2992f5eccdbfe20,0xc9defb9eeecbfe08,0xafd7fb5bedc23a0c,0xde1a5fba86ca7e18,0xafffaf7beecabeb4,0x4ea675e2ecd31aa9,0x8303ae6b0d527e10,0x63523feb65dafea9,0xfdbecff7eec3fe35,0xd6f9b27d4dc2beb8,0x447c37304cdebe1d,0xfbdefb7fcfd3febd,0xffcfffdde6db7e9d,0xbffecfebcfcfde35,0xeffee7ffcdcbfeb0,0xeffefd774fcb7e39,0xfbdfffffcddbfe99,0xcddffe3b4fdffabd,0xbfdfdfdfefcffe3d,0xf7fffeffefdffebd,0xffffff73cedbfebd,0xdfbfffffcfdffeb5,0xefdfffffcfdffebd,0xfbffffdfefdffebd,0xbcbfffebefc7febc,0xdffdff7befdbfeb5,0xfebfa7ffeddbfe3d,0xeff7ffffefdbfeb5,0xefffffd3efcffebc,0xbffffffbefdffebd,0xfdfffffbcfd7beb5,0xfffeffffefc3febd,0xdbf7ff6eefde7eb9,0xeeffffefefd3fe35,0xbfffeffacedbfebd,0xd7f77ff2cfdb1e34,0xfffefffb4fdffe31,0xfb96f307cfcbbeb8,0xfe7bb7dbe5da7ab9,0xffffdffbefcf7ebd,0xf7f7ffffefcf5ebd,0xfffeff73efcffeb0,0xfea6ffffe7cb7ebc,0x6fdffa1fefcbfebc,0xfffeffdbefcffabd,0xf6ffffedefdbfebd,0xd9ffff7befcb7e31,0xdf966f7f4ec97ea5,0xe7efee3bcfdffeb0,0xa7defb57cfcbfe38,0xffffffffefdffebd,0xff7eefffefcffeb5,0xdfbe7fffcfcf7ebd,0xefcffffeedcbfeb8,0xefffffffc4cffebc,0xfebfeffbcfc7fe25,0xf77f9ff9cedb7234,0xdbc67dee4fde7eb9,0xfbfdf77ecfdf7ebd,0xbbdbbfdbcfcbbe35,0xfbbccfffcf9efe3c,0xf721b8fdcfcfbe95,0xfbbefdfaefd7febc,0xd9a4ff3bcdc77eb4,0xbdf3b39bcfd37a95,0xaf3fdffa4fdf7e9c,0xddf476eeccdb5a14,0xfbfc6d73cfc77ebc,0xffbeff7ee7cbfebd,0xcdffbe3fcfdb7e39,0x7bdcdddfcfcbfa21,0xd645d6fbcccbbabc,0xd1eebfa04fc2aa31,0xfbbebffe4f8e7e3c,0x4caec1aecc2b690,0x6fbcfbd3c6cf6a24,0xbcbbffeb4fcffe31,0xf5e7b7f2ccc3eea4,0xfca52feccc837ea1,0xcb5ac77fa4cbfe08,0x2f7f7e6f67db7e18,0xf6faffe9cfda7eb4,0xf3efff7fcfcb7e98,0xd7f27f65cfc35e94,0x7b5fdb2e4fcb4ea4,0x4ff7bfd88fc37ea9,0x7aa0a4e3ee470230,0x98753614244b3a01,0x97bc2d68e7d7feb8,0xa9ccc116cd422e30,0x876716204483ae10,0x8a9e4f52cfc13e84,0x3092547445c20a14,0xdb9877c00c4e1619,0x4a82b724c59b7808,0xc945db21ca430e24,0xaa27dfdacec936a0,0xd414166dc1824e14,0xa5dbe770c5823e31,0xee043f6e85cb4a08,0xc1e6e93ccd8b3e94,0x736116d140c30e35,0x3c751ffb6c8a7691,0xfbfce6caee4a76b4,0xc20404a28ac81280,0x84603084c420020,0xb100184802280,0x14885281ac882200,0x952c977325c37a11,0xd8040710c4c20eb0,0xaafffe6b4ecad6b4,0x6ec10edfc8020634,0x100650040a3000,0x20004000400800,0x1280004040400220,0xe130d120c0c25a00,0x4723082c8c20814,0x0,0x0,0x804092008020608,0x100020540081400,0x142100000121200,0x0,0x0,0x0,0x0,0x0,0x90040140000010,0x100408ae68805600,0x800400000400200,0x540c36f58cc3aa00,0x200200080800000,0x10e000000081280,0x20004831a0828200,0x8500d00284828280,0x7f9e3feeefca7ab4,0xc02020400a400200,0x8050014a048e0210,0x9892ac8245c32a00,0x7621b3f264833a24,0x6ffc6df0cecffe28,0x3d7762e4d4a7a10,0xece6de524cc8de1d,0xefa64de0058ffe80,0x4012606440921600,0x1b9a1d70c68b7e2c,0xd9b6df3b4fcb7a3c,0x4cc3ea134dcb3e18,0x2b1debdd25c36ea0,0xb0919ceaeec24a00,0xd0e30f490d0a3e81,0x85ef2084873a10,0xdcc34e0f41cadea0,0x2bc6d551ec83ba0d,0x16420b420e0a3208,0x240080800000,0x80000b0000000680,0x0,0x90d108044027400,0x34a24ec005cb9808,0x5475d058ce4e3684,0xc690090046422600,0xebbefb6f4ddb5a30,0x6ccd1f4804cb5e10,0x23288240c8420400,0x11c2964ac8c08024,0x5214019040826a00,0xb0046f2cccc30e2c,0xc0d212b8468a3604,0x148e5008441230,0x3400000003000a00,0x20240306822200,0x102020480810200,0x200200000600,0xa6b1412c5801a84,0x14b056684d836a30,0x97432170cdca9828,0x92222a4044c23498,0xedc6eb3bcf8afe10,0xe41d9042c70a3a14,0x811000041020800,0x808600006001080,0xaa123c1c46c38208,0x8fdbfa0bedcf3a39,0xab7f7ac36cc36e21,0x13880b1020c6aa00,0x20448568440a7600,0x2802d0c46400608,0xe9eee22f4dcb3a10,0x4c4b3a0c86c3a608,0x380c00840c01280,0x424424004503200,0x250108801200,0x8310ef2e6ec20688,0x82328a42101220,0xc4841000a001600,0x64040004422200,0xb58ce5344cc37e08,0xd0b8f31c4e437e1d,0x786700a8b805a80,0x800000800800200,0x1436746ac5ca7810,0x1210010004820008,0x3210080442030238,0xe586040046420a18,0xa04860204c3b000,0x40c00a040841280,0x802846014c02220d,0xdbb527f4ccc7269d,0xeed2e392e4c35e0c,0x6ff5dffecfcbbebc,0x82aacffb4fd33e90,0xd3bc9e46eecb3e99,0xfa1627e24fcefab4 },
	{ 0xc9f7f33ffa7f9aff,0xeef7fbdaeb7a9bef,0xefbf4dffe97e39ef,0xa7040400000188a,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd8fcfeda6a7618ef,0xdabff772eb771bcf,0xcdacf441a8650a6b,0x4fd9be9f6a569049,0xaf5cdfdc2a761b6f,0x74e6d4f9ba731b4b,0x936ef7188a4e29cb,0xfab421f3a87a92cb,0x49c6be323a7a32cb,0x8b5ed7db687a9bc9,0xb43777efea731bcb,0x8168ff53ea5610cb,0xfbbebf72eb7238cf,0xcddefbbfeb7610cf,0x6df7dedf2a7b9bcb,0x2e947de968783bc7,0xf47df67c7a729beb,0xd2a6f5ef7a761b8f,0xfbdeffaeab7e19db,0xecdb1efb6b7e9bef,0xbfbfffeeea779bff,0xf79cf3f9f876b3db,0xc3ae7dfaea6e1bdf,0xdbd6ff7eea7a93cb,0xefdabfdb6b773bcf,0x2fbfdfdb6a73bbff,0xf8f7d6eaba76b9db,0xdfdead723873bbdf,0xfb8affbcaa5b9bff,0xdbdffebbb87ebbef,0xfb7fffdffa76bbff,0xfcfdfffafb7f9bff,0xfffffffbfa7fbbff,0xffbfffffeb7f9bef,0xffffffbfeb7f9bff,0xffffffdffb7f9bff,0xfdbf3ffbfa7e9bff,0xfbfffffbfb77bbef,0xffbebffffb7fbbcf,0xfffff7febb7ebbdf,0xfffeffdffb7fbbff,0xffbffffffa7ebbfb,0xffffffff6a7ebbdf,0xfffefff76b7f1bcf,0xfbf4f77ff97f3adf,0xffff3fdfeb77bbdb,0xbebf7ffbeb7f1beb,0xfffdfffefb7bbbef,0xffbe6df2eb7f9beb,0xfbf7ef7fab7ebbcb,0xfefffeffeb7abbef,0xfbbecf73aa7b1beb,0xf5f7dfebfb7793cb,0xdffeff77fb7abbff,0xdfbeffffeb7a9afb,0xdb96becfea739aef,0xdbfedfdbfb77bbff,0xfcfffefeea7fbbff,0xffffff7ffb7e9bef,0xdfbef77f6b7abbcf,0xfffdef3fba7a3beb,0xeeffffdffa73b3db,0xf7feffff7a7f1bff,0xf3bfbfeffa7fbbcf,0xbeb6bf7d2a739bcf,0xc3e4ef6b2a799acf,0xabfffed7ab5ebbf3,0xe6bbc73228521919,0xddfd76fa6a539abb,0x71f63f6228539bcb,0xc3d0f77afa729aeb,0x6fff3b5f687e9bdf,0x9801dcdc4a5418c3,0xc5f633fd2a6f1acb,0xdf9efd796b78b9cf,0xfb94e64362509bcf,0xebfa3edfe97e385f,0x2bbecf7deb5219cf,0xf4ff57eeeb7e9bf3,0x951ef7f0e876136f,0xfe97ff6feb7e9bdb,0xd7dffe3b6a7b1bcf,0x8afddfd7aa729a9f,0x757dd77faa7e38cf,0xddeef7e26b7693c9,0xfbfde77f79723bcf,0xefd7fbea7b7f9bef,0xe37fdcdf2b763acb,0x74bf3fe33a7bbbdf,0xfb77ee7b3b779beb,0xdf8a2f5f7a7810cb,0xefdff7ffeb7fbbdb,0xbffffef4e977b3cf,0xbbbfdff6bb7fbbcf,0xd93fbd6a2a721aef,0x5ab63c746a72bb8b,0xe302f3e77a769987,0xeaf66e8beb7b9acf,0x171fcffe2a7e98f7,0xcdd9f6740a7e19cf,0xd6b02d40b87212a7,0xe1f4e4200a621007,0xcfda0e3e7a7312cf,0x9b4855b2852109e,0x64fe066bea73130b,0xe49e4d72287c00ee,0x7204fc013056114a,0xd6443e09285e134b,0x431285c329501bca,0x34c00ec4a86419c2,0x330a0761285a91c0,0x200100101000,0xe8d3f3f86a721abb,0x2a18dfdb1a5f1bcb,0x34235324a86a2a56,0x902425426860808b,0x200250200501800,0x47c2073ce852100b,0x80a01541810110a,0x226680a621107,0x1db404000a561203,0x4870217048621a83,0x42080000002,0x2153320828629006,0x2004bc0a1400001,0x4040821000400242,0x4022080002400009,0x830038009010,0x48400050a8008102,0x802094038001180,0x400200001400000,0xa080052108181147,0x40002401200880ca,0x623800a8400200,0x8010000000006,0x3022400008701010,0x100002008000100,0x0,0x2001c00001501000,0x0,0x80020e920842108a,0x3880502008101302,0x64121b802a62100b,0x400212000400007,0x24125f010a600083,0x24000080040,0xa0008b208a080842,0xcb727c5068721ac9,0xe1b0e6222b7e10cf,0xa9741dcaa97201ce,0x4498c5a1084219cb,0xd15dd569a85212cb,0xcf9c3dc8ab7a114b,0x8b87a7771a7e114f,0xcf5bba2b6a7ab04f,0xef3e4ff82056988b,0x4010504202521041,0x3bb4af303a7b1bcf,0xdbbfee3f2a7f9aef,0x8beb5e41005218af,0x2414465208600140,0x8014002000601882,0xc16ff6306a339303,0xfb226750ba541bcb,0xe8acd22e8870b3cb,0x165640002244134b,0x80023d410060020a,0x2200242428500080,0xd0100b520864094e,0x602200010104a,0x4a080002500002,0xb43854e24a669801,0x18b414025840100f,0x160a286008431360,0xe1d6faeaea7c38fb,0xaece036128700a8b,0x0,0x75602100469029,0x499615621a62110f,0x0,0xc806001100109203,0x48840b00a00094a,0x34605a822213020f,0x202c50a8020081,0x80a00000400068,0xc4c286082a0200cb,0x931c4dd16872810b,0x9c1272015850198b,0x41c8454042551249,0x4692ab5c28761b29,0x659adb38ba7592cb,0xa9573aca2872b23f,0x8490422119601182,0xb18601088a400a42,0x9191013008521182,0xef45f58ba87a91cf,0xa9ca1211887838c3,0x898cbf30872129f,0x1000a76020420282,0x1844e9a148521083,0x4828f2224a741981,0xcfc52e18e85693c7,0x34100100a0000086,0x2202402000008a,0xc20024010a021850,0x8000113848541881,0x0,0xc80000641806,0x0,0x0,0x8000004002000002,0x800020000000202,0x2000800000000082,0x4451522948423303,0xa500a6100a001901,0x8081498008201a82,0x102801060020001,0xa24e1409005002ce,0x1020044002001003,0x8040a8408060000e,0xda912f3e2b7f1bcb,0xc356ef792b7f92cf,0x2bfeda593b7c13ef,0xbfbb6fefea76b9cf,0xf17ffd62aa7e9bcb,0xff9e7d63fa7e9bef },
	{ 0xe9f5fb1ee37cdfc6,0xefffffdee77ffff6,0xdf9ddfffe71f1ff4,0x1425b4ef203fa9f6,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x385ca7d8013e3dd4,0xfb3eedf2653f19d4,0xcd86f72fa3665dc0,0x41e55b8a663f14e4,0xe39ecf52223cb7d2,0xd4b47d7ee33c5f74,0x9eefff76a37f9fc6,0xd3928f76a21c9ed6,0xa4d9ea30252d9fd6,0xa82dbd9633d1fc4,0xf69f7fe7e13cfdd6,0xdbfcef7b627f3df0,0xdea67f7ce77e1df4,0xefd6f7f7a77f1fe4,0xefffffdbe77e5fd4,0xcebb5ffa667eff76,0xdfffb75ee37dddf6,0xfd9e7fff657e9fd6,0xebfffffbe77ffff6,0xefffbfffe77fbff6,0xbfbf4efae37f9ef6,0xddf7ffffe63d1ff4,0xfdbe3df2e77f3de4,0xbbffef2f677fbed6,0xeecbf7fbe77f1ff4,0xfbbedfffe33f9ff6,0xffffffe9e77edfd6,0xdbfeef76e47e9fd2,0xfbbefffde77f3ff4,0xdffbff7fe57f7ff4,0xefffffdfe77efff6,0xf6ffffefe77f9ff6,0xffffff76e73fdff6,0xfffe7fffe77ffff6,0xffffffbfe77ffff6,0xffffffdfe77ffff6,0xffb7dfffe77edff6,0xfb7feffee77ffff6,0xdfbf2fff673cfff6,0xefdfffefe37d9ff6,0xffffffff673f7fd4,0xffffffffe77f7ff4,0xf9fdf67aa23d9df6,0xfebe7ff2e77f3ff4,0xeb9bffff677e9ff6,0xefdfbf7fe77ebfd6,0xdfbfefffe33ffff6,0xff7feff9673ffef6,0xfffefdf3e77fffd6,0xfbf7f73ee37ffff6,0xcdfbbf9fe77d1ff4,0xefbe8f7b437ffed6,0x91d5b7ff667f5ed4,0xffbffff0677efbf2,0xfbb6efff677fdff6,0xcfeff6b3673ffdf6,0xffffefdfe77ffff6,0xfeffffede73fbff6,0xfffff77ae77ffff6,0xf3864775643c17d4,0xc7cfffba657f9fd6,0xbefcd3df233e19f4,0x7df7f7ffe23e5ff4,0xf9fdfffee77fdff6,0xfebe3ef2e73c99d6,0xc5fffb3ce33ddcd6,0xee7f7acba77fdbf6,0xc1374ff3e65f3e94,0xd08df66aa53c1cf0,0x8be66f65a77f1dd4,0x8b35fbfb673d9cf6,0x2fdbdfffe43f7dd4,0x4c01493202141580,0xf597a6fea37e3f54,0xc7a87f53a52c9fd2,0xdb96d31e637edce6,0xea67bbdae27e1df4,0xabbe6fffa33d5ff0,0xbfffde78673efff6,0xf3f8af70643ed8f2,0xffb7b3ee617f9ff2,0xcfceff5f673f1e74,0xe3dfdbd2603f30f4,0x14fcfeefe37e7ef4,0xdf7efe7ba77fffd6,0xffbf7f7ce73ffdf6,0xc6f3ef72e63fded2,0x6ffedfdde35d3ff4,0xf6bf5fefe27e9f52,0xd9acbf5ae25e9db6,0xfeb6ffffe77ffff6,0xedffff7fe77fdef6,0x7dfefefc613fdfd6,0xb6b9dffb67769f52,0xe9d597e8277f93d6,0x7ee67ffe677ebff2,0xeb87f76e437e9ff2,0xecbc9e87633e9e46,0xb6bdfefce63e3df4,0xf5f5f67be77f9dd6,0x6eb02d50623c8576,0x1880890c00100470,0x801b98a2007500d0,0x2400001a24041440,0x51507668211d99d6,0xb398ad1022371744,0xba94373663349d86,0x89428010000c8412,0x89020fc382151d94,0xf4106660213c5f14,0x9627a68015d2c54,0x12b40f78403711f0,0x20c47080204c04d0,0x2840d162148042,0x4023306010e0840,0x1106a060e7040800,0x4a002104021c1040,0x450a020005402c0,0x5d4c40621802d4,0x0,0x3564562a635412d0,0x6010b59124461450,0x954d626415c0ef4,0x2a4b9659003591d6,0xb020040040300880,0x8000000000040040,0x60702443802e8852,0x100810440041400,0x4080840001040880,0x808054000160480,0x4b4240001040cd0,0x8308ec01242e14c0,0x30002400c0040c00,0x8020c00000011100,0x9ac8c1200101450,0x80000800100004,0x0,0x31000a14001c16a4,0x0,0x0,0x80100000100004,0x40640124108006,0x12800940211c0200,0x8100106125000140,0xa04005000340e00,0x402002040000,0x91c1224800340e80,0x1a0405cc61140244,0xe946e1ba631d1dd0,0xe5de1bdb633d3dd4,0xe61cf772647f35d4,0x94d61270242c1d44,0x93bc2d61e21cd8d6,0xdbc2bb7e617f9ae6,0xbff2b68b673fdf56,0x1f194d02e2345bc4,0x22034802040e40,0xc99e77b0a57f1de4,0xd91e4336e21c2cd0,0x89acc603607710c0,0x210060a007a1060,0x367736e8c0361fc0,0x8bc2563641893c6,0xfb977f37267fbff6,0xd481eb12003c11e4,0xaa94344ba2240c00,0x8014102022341140,0x4000221060200040,0x9a342c5062521254,0x8000080000080020,0x1100002040000,0x1318450201341960,0xa02496eb647e19e4,0x47e46e2d653c1d64,0xbc3ef06627f15c0,0x9432a38603d1424,0x410030000240c00,0x8400044000101040,0xe3d02d21263c1454,0x80400000100080,0x482200000140420,0xa28c8c381660c40,0x51745466a2164840,0x8064546060103054,0x41200062000880a2,0x4200200400241100,0x83091402002c1480,0xb4914fec677c04c0,0x492c320860380050,0xe292ef86661d1790,0xfadcc11ac11c9df6,0xa6d68656607c1df4,0x4222a6860182c70,0xf164f270006f1950,0xdaaa2940032e8d52,0x6986c31b243d1dd0,0xb361e5a60367f04,0x11200d10603402c0,0x8100044a001413c0,0x1986ade222350ec4,0x280e00200181150,0x24cf1fa1405615e4,0x2840300018042,0x0,0x1a645022421140,0x8390d323625c2664,0x800008080000800,0x6025f1064001510,0x241608000815c0,0x91b8a105001c06f0,0x4000852000181320,0x41a00000089482,0x804a400400c0,0x0,0x203402003400c0,0x520e449000362404,0x44c64026611900c4,0x400813c122208042,0x8400202000101040,0xed42ff63633f0ae0,0xca89273627345bd4,0xfbd5fbbee67dbfd6,0xfff1f63647e17f0,0xdfbefffbe77edfd6,0xf32bdff8637f17d0,0x3bfe7d6de71edad6 },
	{ 0x8852f0ca880087d8,0xaff7ffdf7f88cf9b,0xf6ffcffeffa9cfdb,0xf777feeb7b08dbd3,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf1ed227a4e29e7d9,0xd6beef774f28cf1b,0xeb9fff7f6f09efdb,0xc9d3f28a530955db,0x6f7fcf5e3f09f59b,0x74f6f7ef1f09ef5b,0xddffbf6eff29ffdb,0xdfdc35723a09ebdb,0x51c2f6161320cf5b,0xabfabfdfffa9f7db,0xd4fb5feb7f21ffdb,0x4964fb606e01cfdb,0xdeb63f776b29cf53,0xe7c7ffbf4f21cf5b,0xfffeffde7fa8ffd9,0xb6ba5fff3fa9cfdb,0xddffbffeff29dfdb,0xd7ee3d745aa1d7db,0xa9d6f77f7a29c75b,0xef775f3f9f21f7db,0xdf3ddff32fa0cfdb,0xfff7ffff3ba9dfdb,0xdee677d10f89cf5b,0xebf2ef7f3f28ffdb,0xefffffff7fa9dfdb,0xafbfdfff7fa9ffdb,0xf5ffffffffa9ffdb,0x97ffff536b28ffdb,0xfbf7ff7fefa9dfdb,0xffefffbf7fa9ffdb,0xbbffffdf7fa9ffdb,0xfebffffffb29dfdb,0xffffffffffa9dfdb,0xdffe7ff7dfa1ffdb,0xedcefbfe7a28ffdb,0xefffffdfff28efdb,0xffff7fff7fa9ffdb,0xfdffff7f7f29dfdb,0xfebeff6fbf29f7db,0xf9f7ffbfdf28f7db,0x7bdffedf5f29efdb,0xb7bb5fffdfa9dfdb,0xdffdffefff28dfdb,0xfff67ffdff28efdb,0xebf7f7ffab21efdb,0xafffbfdfdfa9ffdb,0xfffffffbbfa9ffdb,0xfffffefd7fa9ffdb,0x7fbefd7a7fa9ffdb,0xfbd7ef76ff29f7db,0xffffbfdb1f29dfdb,0xafffcffbce29ffdb,0x71febffc5b29e79b,0xffb2aff1efa9dfd9,0xeffe6fff7f08dfd3,0xcbfbfe8f9a09ffd1,0x267cc7596a20819b,0x6cf4f6632ea1ef8b,0x8541047a8008cd03,0xa1942e200820cf9a,0x1200f28a0b814b01,0x8efe4a1f1b01675b,0x9417d6e82e20cb51,0xfdeff6ebbf29ffdb,0xfe1e3f6f3f28e7db,0xef9ef7e92f29c7db,0x2dffaf6f69a8b353,0xcdb9577b3b89dfd1,0x9179bf1a0f08c749,0x468a59ec8f08e582,0x88f6fffa6f08eddb,0x6e42b2595b01c55b,0x9f44ae4320cf12,0x9cdb70540b01c7d3,0x5a7635d27b28fb9b,0xf3d7ffbf3fa8f7db,0xeecf7f5bab21cfd3,0x6efeefebfb20f5db,0x74ffffff7f29efdb,0x8efee5713981c79b,0xfda7ff6ebf28dfdb,0x4fcfeb7b6ba8f75b,0x2b5e575b6821d759,0xfef7ffec1f08f7da,0xff8f7d639f09ff1b,0xffb77f7fcf29ffdb,0xcffeffbf7fa8efdb,0x8affdfdfcf21cf1b,0xd895d5e93b08cfd3,0xdffdfdb24fa8d7db,0xffb73ff77fa9efdb,0xeffbfff77fa8ffdb,0xbffffec7ffa9dfd9,0x7cba6de64d284d9b,0xf462f7fb0ba9f7db,0xfe7e7ff77f29d753,0xab8cf7b61b08cf9b,0xefc73ffc6f28efd3,0x93b8efbb5e20cfd3,0xd0f3b6e30aa8c51b,0x6eba7f588f20f7db,0xd9ca43650908f599,0xed4a321923084793,0x6f9cd8db2b88d3d1,0x907c564b4928e9d3,0x9eaebff17f20c7da,0xc889ceaa0f09cf50,0xcbd7968e3ea9cf53,0x8a8a01420b209312,0x301400a000080301,0xfd8ffffb4d08cf9b,0x9a043f6c8a01d758,0xc130dd08830085d8,0x2b64884a198887db,0x15730fc229216902,0x9141366012008513,0x4c062c203100c993,0x400200000800a010,0x8d0849002088311,0x122088240208300,0x406100280a000300,0x8c50284c0400c311,0x804e3660d084102,0x2311184e0b21d392,0x298c80009001250,0x51f586280b08c51a,0x2054149002102,0x10020000000300,0x41135008208140,0x200040000000210,0x60c017401400c301,0x82b00d100a214d00,0x402c448203080710,0x9d4ffeb87388d75b,0x100041a22000112,0x0,0x1400d2010000708,0x6080240102008018,0x8000000100008001,0x0,0x0,0x10a15221000c51b,0x9c0605008a088580,0x8810c06809200883,0x200020200080310,0x4090050003004100,0x48e050001004d08,0x1092f5500c21c281,0xe100a3320820a659,0xc46014009080742,0x294c6f20201d31a,0x80405252010885c3,0x18ae64400c20e68b,0x69dcd2390e08f7d1,0xfeebff9dfb08ef93,0x6999d7ce3f01cf9b,0x1141060c0a0846d0,0xf6d8ef752b08cfdb,0xdc8427662d09e79a,0x108340208008891,0xa60ecb4983884f93,0x48414024001c112,0xb2c776808005fdb,0xfb1eff7dff00efdb,0x8cda3b1a3b08e913,0x8054220088583,0x218255fa0b01cb0b,0xd5dae772c9080f83,0xf69425f57b89ef5b,0xa81050022100010a,0x214ad8012c0045db,0x923ec7a3da209d19,0x6d07080f206300,0xed3c0d728a288f8b,0x8c0042452004900,0x2006000000008082,0x8480012080400,0x8840024048000189,0x5a8000420000c391,0xb2d4e2624821e7d1,0x42d8b2271a08cd49,0x4214420020008180,0x347d73668f01454a,0x86a63c001e006442,0x4000411020008101,0x884040000c004100,0x6009cd4ac200851b,0x853406a04208c911,0x90c00620800305a,0xc298673608088f83,0xc5bcf9aa1b21dd1b,0xabde59995e21c75b,0x949606a00880b553,0x4a40f5021a00cb51,0xf6946d602b28c39a,0xc1d1c2048820ef59,0x331233424a01e5d0,0xb2024d620a00db10,0x30b800200a802502,0x71284d430a21c759,0x10012000000,0x1710d21a21c810,0xbb0c9d33b015b12,0x20c4144003004093,0x4ce66551c800a79b,0xdb92836c5b81c79a,0xa000000001004000,0x0,0x440606880808a111,0x38a6103b20050a,0x4c08008014080,0x4507ba034909c111,0x2a0009c00c00c180,0x1004000000000100,0x84d042a1700c093,0x100200100209010,0x4000060408200100,0x140014020880540,0x843a04e65400c913,0xb119ca300900c390,0x909209c00a080718,0xe4c6ef3d3928979b,0x3f7ffffa2b08efd3,0xa7ba6fff5b29ed9b,0xd5fcb67b9f08ffdb,0xcad465735ba88f4b },
	{ 0xca5a72b23dbe4d40,0x6fcf13fb1cbbed66,0xfffffffbfdffffee,0xfefff6ffddbfffe6,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2049035200aa080a,0xf7bffff0bcfffde2,0xffbffffffdbffdee,0xbff6f69b34bffde2,0xefffdfdefdfb7fee,0x54b67ee934b47de2,0xb33e77e13cfb6ce2,0xfeb53f6bfcbbece6,0xcfc3ffb33dba6846,0xabdffdd3bdba6de2,0xf4f77ffff5befdea,0xfbfde7fafdbe6d6a,0xfefe2ffbbdfb6ce6,0xcfdcfb1a709a6c66,0xeadeffd134beed6e,0x3ef76ff3bcfa7d6a,0xf9ffb7e99dbaed6a,0x7f7e3d70f4ba4fea,0xc9f3eb2b35ba6de6,0xefeb3bfffdfecfe2,0x9bbfdffbddf87fea,0xdff7fff9dcbb7de6,0xdbbeff7335bfede2,0x9bdfff7ffcdaf56e,0xefdbbf7ffdbaefe6,0xffbfeffffdbafdea,0xfdf7ff7e7df7fdee,0xfbffff717dfafdea,0xdfbffbfebdffedee,0xffffff9fbdfafde6,0xeffeffdbfdfaffee,0xf4fff7fe7dfe4762,0xffff7ffffcfbefee,0xffbf777dfdbaffe6,0xffffffbffdffffea,0xefffffdffdfffdee,0xfeffffeb7cbefdee,0xff7efffbfdfadfea,0xff9fafff7dbffde2,0xffdfffbebdff5d6e,0x2ffeffdffdfbffee,0xbfbffffbfdfefdee,0xfbffbffbfcfbfdee,0xfffefff7f5faffe6,0xebdff7bfbdbf7d6e,0xbff73fdffdbefdea,0xffbffffffdffffee,0xfff7fffffdfbfdee,0x5afefdf2fdfbff6a,0xdbdeffbff5bbff62,0xeafffed7fdfbfde6,0xcf76e7fb7cbb6c66,0xd5dfffeb6daecb62,0xff9c7d705dbe7f64,0xd7bef7f654faed60,0x8dc2f21b1cba75e2,0x2b5b5ed20c9a4d62,0xd4374a2ed8aacd26,0x940ced75f42a7c62,0xdd84095e51324c60,0x9c423393bc80c520,0x2b7cdfd875badde6,0x6bf0fe270be64e2,0xf77cff52d5bbdeea,0xfab63ff47dfacdee,0xffaef37f7cfeef66,0x3fffff6fd5bbf4ee,0x3e24c6bf35ba8546,0xd17f3779fcdbcc62,0xde9e7fee75facf60,0xa3f6db6eb9ba6562,0xafc738c305ba75e2,0xebb439115ba6162,0xf362fa7b34b8f5e2,0x49fe77723cbb7d66,0xfbd4ff3fdcbaeeea,0xeffffffffcbefde6,0x5e95ef2b7dbb5d6a,0x7cf7ffeb7dbf65e6,0x973baf72fc964de2,0xfffef37fbdfa7de2,0xedcfff5f5dfafde2,0xbaff9ddbfcba4de2,0xf6fdfefbfdbb7de6,0x9fcfe73139bbcc62,0xffb53f7ed5fb6de2,0xdefeffbbf5bffce2,0xedcfdddf3dfae6e2,0x94f73ffb35be7d62,0xddecfe7e74ba5c62,0xfeae7ffcb8bee66a,0x85fbcbabd5b24fe2,0xefdefeff74ba7dea,0xffbefffbf5bfede6,0xd756bf6bddbbe760,0xff72be6d1cfbcfe2,0xe3d6df2a9cfa4d6e,0x6df7929d3d2b6dee,0x9aada7fd74be4f66,0x551060a10c380560,0x5c48694b348b4de2,0xa1144300208608e0,0xa4d396ce50dbcc62,0xc11495804aa5842,0x7deff7fedcbbffe2,0xdffaff55fdbaffea,0xffa6fb367dae5f62,0x40d2d20190024d66,0xae8fc9933d205f40,0x643550733492e042,0xaf6eff73f5ba7f6e,0x1b84523414120442,0x1243f30a543a4f62,0x202140000920040,0x4471612040a5046,0xc060a69038ba4c06,0xf294351244b0dc62,0x4580a12234aa0924,0x204490921080880a,0x190581014c10440,0xc1180e10049a0440,0xbaae45e03c9d9822,0xa982e13af4906d60,0x22505ed0a0924d44,0x238c4518048a5960,0x400a06014091020,0x52c144011204462,0x8b00b60c14931442,0x80c2101030808042,0xc01004005200100,0x2020030000000,0x110c296015316400,0x80000400080002,0x952ae0004aa00e2,0x200001004001000,0x0,0x8001000000000a,0x32010012040,0x1000020000000140,0x488a120b10da0442,0x8010000010000400,0xa104342805284142,0x2a2042000984420,0x4441300014004540,0x2806da800db00160,0x410e452049a4442,0x660ca72205b8f560,0x5b2e7de6c4aa2f62,0xc881836054c80560,0xeee75d523c324d62,0xe194ec8309bece2,0x645292602c98e560,0x220005a0a4a80442,0xb3e6af3614daede4,0xcecb1f95fdb9ec62,0xadbe5f74bdaaff62,0x604124a004a88102,0xaeb26d6254bef8e2,0xd8066e0804b804e2,0x8d4aba3144fa1e60,0x3e49d95a4cba4462,0x24e4516830caa06a,0x1068d30b0184560,0xcd8133cd94985422,0xc098580080124540,0x218cd4810284d4a,0x76a3742214ba6546,0x142e658109264e2,0xffbe3f77b5bbffe6,0xe9f6fbfefdbb7f60,0x92de1e5774aef062,0x983a4f5a30916c42,0x1008a76214086062,0x200110280002,0xc1aa90183482c440,0x400260000844040,0x7344a00103a040a,0x14e120900140,0x92ac0f4194b84d22,0x8972cb6a34924d44,0x43701192208a00ea,0x10042100c085002,0x420444030104c40,0x124140000a8e000,0x0,0x40a020020020840,0x129e450940d84420,0xd498cbecd08e4d22,0x8848780214190462,0x88008b08442870e0,0x804509080098cc02,0xaebcd80530ba4f6e,0x4220c61000221e2,0x40254050140a1424,0x93ae2c3414ab6866,0x25924207103a2e62,0xf056c65a14929ca0,0x83ba0fe10dfacc42,0x404000000020800,0x728e2d54d0f9c462,0x80000160008a0004,0x0,0x938f89aa748a6462,0x8472f8f03d205d42,0x400c00020082000a,0x91040a2411a60444,0x8408000104021800,0x420000084000,0x40440020084040,0x832f35601ca855e2,0x480003a000824042,0xc453b49110ba6966,0x2886814430a90002,0x3004000400a20040,0x8808244211a420ea,0x9021c4801aa0800,0x480a22210904042,0x802000000000120,0x20014610020406a,0xd771ae08b5325542,0x9aa52c40b4824c40,0x69dfffae3cbefde2,0x2adfffc71cb9d5ea,0xb6ba7efe3cfe4d6a,0xfdfdfffcfdbb7d6a,0xfefe2d709dbc4d6a },
	{ 0xc966fb2264d07cec,0x6eef3ed2e4125ea4,0xffffffffb6b3ffee,0xf777fef7b6337eee,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfffefff7f6f3ffec,0xffbffffff6f3ffee,0xfffffffff6f3ffee,0xebdfdfdea6d3ffee,0xf6f7d7e3f633feee,0xbfec7f4034126fec,0x1bb67a7c60f37eec,0x70fbeba0053f6e6,0x2b78f7dab472f5ee,0x34b6eefaf473f2ac,0xddf1a7d6e433d6ac,0xca9e377fa633feec,0xcdfef723e2d33ee8,0xedbe5fdd24b377ec,0x3ebfcfe39632d6e4,0xe53ce6f7d632e6e6,0xef665deb843076ec,0xe9effbebb653feec,0x76ff7affe6b377ea,0xcfff6febe6f356ec,0xf7fdf7fff6f37eec,0xfffe7df3e673f7ee,0x93d4ea3e647377ec,0xafd6dadba0f3f6e6,0x7ffedffbf6f3feee,0xf5fffffff6f3f7ec,0xdffefdf2f6f376ee,0xff76fffe363376ee,0xcdfffebff6f3f7ec,0xffffdfdff4f3f7ee,0xbcfffee6f6f376ee,0xfdfffffff433f6ee,0xfbb77f7ff4337eee,0xdff7fffff6b3f7ee,0xffffdfdff6f37fee,0xffffffefe6f377ec,0xd57ffff7b6f3feee,0xffbfff7fe2f3ffee,0xffdff7fff6f3f7ee,0xefffdffff6737fec,0xbdfbfffaf6b3f6ee,0xf5fffffff6f3feec,0xfffe3f7fe4f3f7ec,0xfbedfb7ef6f3f7ec,0xeffffffff6f3ffec,0xfffeeffff6f3ffee,0xfffff77ff673ffee,0xdffe7d73b6f3dfec,0xfbf7ef7ff4f3f6ee,0xffffbfdf76f1d6ec,0xffbecff7b6f2ffec,0xfcfefefbe633ffee,0x99ceb776a47356ec,0xce1ae7ee767277ec,0xd574fe042072f160,0x8b5e1e5a941177e4,0xcad694fc06b3f4ec,0xd94e6631843356ec,0xcb9e752e24f337e8,0x5fd2e73ff4f2f6e8,0x4beabf43249336ac,0x3c177febe653f2ec,0xf79eff7fa6f376ec,0xdebe3ff4369377ee,0xc7f7fffef6b23eec,0x33dcfe99a43377ee,0x2cd13000014ec,0xa1efbf39a4f336ec,0xd7963f7ca0d276ec,0xc9c4e72a34d116e8,0x6ed41753a65272e6,0xc2b4dd864306ae8,0x76ebfaff74123eee,0xcebc25e234127aac,0xfbd4f76ef453f7ee,0xefd3bedff47377ec,0x6fbfedd7b613f5e4,0xe5fefbfe34f2f6ec,0xbbeefff37673f7ec,0xf9b9ff7da4f356ee,0xcfdefffb36f377ec,0x7b1e1fd7269276ee,0xf57154fc467356ee,0xdbfe7ff34452f6ec,0xfbbe7e7be6f376ec,0xeffffe36f672f6ee,0xa9fdbcfe4c2f6a4,0xb7be5ffef453feec,0xdbf6edfdb4f377e4,0xfca227721692f6ec,0xcfca7671c45373ec,0x5fde7adfd63376ee,0xf7bb7fe36433f2ee,0xf9bdfe40e693f7ee,0xf7d67df0e672f7e4,0xe9d6f3c9b4f336e8,0xbfee7f5160f3d6e8,0x6d6dfe5c432f6ec,0x7424d264a0d1d020,0x52406561041122a6,0x5cc004000202c4,0xa85b94ebd651d468,0x8042412001046e8,0xc973166ac43376e6,0xdffcdf72f472d268,0x40042201000002e0,0xed61e022051126a,0xa08080002010060,0xd83ed6e1d632f4ec,0x1266a640a05252e4,0x5f387f75b4333eee,0xc4180118244012a0,0xcd8224134334264,0xc53e0d23801336c4,0x93248008a00222a4,0xb104008000110240,0x6080811204007488,0x8e08041005032e0,0x9eab6f5ac01376a2,0xa208060000021640,0x51841810043240a8,0x28c0c10a8413348c,0xe4f7941e42175e8,0x3bc4fc1a4007ee4,0x0,0x25a27c00a42050a4,0x202010200026000,0xcbea9c51b43266e4,0x404474300102568,0x811103084001280,0x8601000114020,0x12002d1420c215ec,0x810012000c020,0x804000000000060,0x0,0x400010000000240,0xca60000a04121480,0x804100400330000,0xa808104000105008,0x1044108840000168,0x9063322209018ea,0xd112025a0022d400,0x4180e016261234e0,0xf21212c2143270ac,0x30020ac414515020,0xf9d18de2e632f06c,0xcf12b52f5633e7ec,0xb14238ac20224e0,0xadef169ab4727cac,0x8c1043c0020202e8,0x322cfa68201256e6,0x428205314011d020,0x9314d1aa00b372c4,0xac4bb398203332ee,0xc39ecf74e6d376ee,0x141020800000068,0xe304d62e0823428,0x4f941127001142a8,0x54af281266234e0,0x2620d5c0269016e6,0x94bed621041132e4,0x8d2e54000432d2ac,0x83bc6734a01156ea,0x6044c01434000068,0x8ad7dc82209214ae,0x3ebff7eba6d3feec,0xb9ece7f1b633feec,0xce9a253ea0535cea,0x25d4c18cc41276e0,0x81521a0760401468,0x6aa2d322431c36c,0x856eb77a0292f6e4,0x428475301432d028,0x88458228009330a8,0x40c1081020001560,0x809410b04107024,0x40144006200040,0x4d6820600410d028,0x8180e402560004e8,0x101100121200,0x8144e00021054e4,0x51224284422402c,0x84950460829252c0,0x82023a0504825644,0x0,0x847e8d4280a27668,0xd41402f14000c2e8,0x85622678043352a4,0x98280f60243242e0,0x308080100020,0x2308129000421be8,0x92431060004012e0,0xe1501e20963118e8,0x46922a6504605ea0,0xcd55f010841274e4,0x304e4ae0a470f8e0,0x131347a286f3d4ec,0xd391a16004f054e4,0x77c67c62801376e4,0x0,0x4800c0000021088,0x200c20000102040,0x201042040002a0,0x409a098040331028,0x881410c208216e4,0x1118004011028,0x0,0x2024204330004,0xc1825102009016c2,0x1ac6230524133620,0x8ec3a20bb47252ee,0x236c491200123ee8,0x1000040000401000,0xc802204000220640,0x14128c64a01020,0x0,0x0,0x4088456000135200,0x412073512423c0e0,0xf58425778432f6ee,0x8f7ee7f3b4f277ec,0xa736d54060f3fea0,0xf53acff3f633f8ec,0xf5fd7ffee4d2feec,0x768e3545042376ec },
	{ 0x8b84ebba799c15fa,0xa48ab207901c148a,0x14a9ddeed11e9ac8,0x95041422919c001a,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfffefff3f9bcbfdf,0xfffffffff9bebfff,0xdffffefff9bcbfdf,0xeb7cefde719eb7da,0xd4f76efcd83cbfd8,0x9fefff7a513eb2ea,0xdbf37e57511cb7fe,0x5ec3e332d11c949a,0xe7dfd75b599cbfda,0xfcaf7befd91eb7de,0xdd6eaffaf93c9fdf,0x9eb82df05116befa,0x61e0df38591c95d0,0x2bffde595194bfd3,0xf7becfffd91cb7de,0x23f33779590cb0ca,0xda9e7fbc511ebdea,0xe9feffced1be97cf,0xfef77efb7994b6fa,0xffffcffbf19ebfff,0xfcfefeff793eb7ff,0xfbfefffad9bcbfda,0xdbfcff6b50bebcda,0xe6d232bbd11c9ad2,0xefbfcffff93ebffe,0xf5ffffebf99cb7da,0xf3feff73593cbfc8,0xfbbfff7ff13cbfef,0xeddfffbfd91c1fff,0xfbffdfdff1bcbef6,0xf6fefefaf1bebfdb,0xffffff73f13ebffe,0xfff7fffff1beb7db,0xeffffffff9bebfdf,0xafffffdfd1bebfdf,0xfdffcfe3d99eb7df,0xfbeffffff99cbfdb,0xfebebffff9bebfde,0xffd7f7bfd9bebffb,0xff7ffffbf9beb7fb,0xf7fffffff9bebfff,0xffffffffd9beb7ff,0xff9eff7b71beb7fa,0xfbf7fbfef1b6b7fe,0xefff7fffd9bebfdf,0xcfbffffbf9bebedb,0xd5fffffbd1bcbffb,0x7d7e7f7bf93c9fda,0xfbd7ff3ff93c9fdf,0xca4f7f53d99cbe7b,0x8fbfcf50513e9edb,0xd7f4bf7a599c9ede,0xdeb66d65d13695ee,0xae10ef65519eb5e0,0xefc7fe13f10e16c3,0xa29eda1e51b416ba,0x7cb574a6c01cb5d9,0xfd423fe6593496e3,0xbb163f69519cb6c9,0xefcedeb9d994bfdd,0x4b5f5f9159be3eda,0xaafe57e3d11cbeca,0xdf7cbff2f9bc1eca,0xfabe7d7271a4b7cb,0xebd7ff7ef9bcbffa,0xfbdeff5f59bc97c6,0x36bf4de05994b7de,0xf5ed3e79519cbbda,0xff92e576c9b4b5de,0xcbeacf7f788cb7da,0x7fde1fd2f83c969e,0x9eb26c5250143499,0xd5fc9fffd194bbda,0xcbaebd41591cbdf2,0xebb6ffbed1beb7ca,0xdfdbfeb351bebfda,0xefbfcfebd91abeda,0x79ec77ef511c3ecb,0xf4b8af7061b43c6b,0xdea459fe791c968b,0x9fd7f21ad91cbde4,0x8cfecd4ed1bc34c0,0xf4aed77dd13c96ca,0x1def3fe3d19cbfec,0xfebc277ed9bc97fb,0xf957ef1ff91ebede,0xaf7fdfded9bebede,0xd7f65ff7d1beb7d3,0xdbe58f7af19cb7ca,0x8aa73f6e51343eda,0xdfebff6851bcbfda,0x6a7f53db7194aada,0xc39aeff6511417ca,0xf9efa77f71be97da,0x4fbaedf95136bfca,0xe9e7eb2ed81cb6ca,0xeff75f4bd11cb7ef,0x16310c60512416ce,0xfdf57ebbd93cbdfa,0x559e0d4298349452,0x9200022440941800,0x8910000100a00,0x213e4c4658143683,0xf13712da10a487c4,0xbbbebd70f934bbca,0xed18ffd25194a41a,0x45c23a3108163cda,0x841ace81f02494c0,0x4c11661011217e6,0x856ffe42501cb8da,0xc812401c00140688,0xc1f5920791a415ce,0x8004400290001008,0x100010090000000,0x69a5f068591c1098,0x42102a0010140484,0x90af128500c954a,0x841f4141841248,0x108a2dc5d1009450,0x8149500000260142,0x44020000000c0484,0x103a02051040088,0x8fbfe31d13630f2,0x1621044040342842,0x54e1f4f4411eb2c2,0x64282422111690d8,0x20020050004040c,0x1210024000a040,0x1a2000011000400,0x8942402e0041480,0x804000041000000,0x0,0x8a80000140140802,0x204000040040000,0x1401000400040140,0x0,0x0,0x4c22e20400401c8,0xe35012c3d1243442,0x9c11252350b60950,0x9ac8f770519683c8,0x40823dc059049042,0x4dd6d7325116b4e2,0xab5eb647511ca85a,0x96a3cf6f5094bed8,0xc577fe7ad1b4b4da,0xc6dc6c6551362cdc,0x94cd72c4924b6ca,0xeff7be31501cb7dc,0xbdbbdff3518691e8,0xd222c85058101c9a,0xcf8e5c72711cb7c2,0x4b96ef36511c36fb,0xc6ce164e7184954b,0x60bccec2d1b696da,0x2034000008441,0x4a82044101100342,0x8f12764ed196324a,0x2184b00200041488,0x65c7580850148cc1,0x78713429503ea2c0,0x91bcaf62593490fa,0x9bb62f7751b4b3d2,0xc7cfc502c184348e,0xbc4c958f104348a,0x9dbb77ef5194b4fa,0xd9eff6fbd136b6eb,0xff9dbfee7194b7de,0xc1f7d30230141eda,0x39c71bdab1249442,0x2438f071d118b098,0x846d006810149a4a,0xc2825451108c83c8,0x8000c04800940a40,0xa641bc21094a740,0x22206e00100148a,0x53242010503030c0,0x7ff67778513cb6cf,0x8910842018101410,0x41308040840082,0x886800280001008,0x1000204440000150,0x1008400100c1080,0xc0022041400c3008,0x8eeef8095134b3da,0x1c28c52411c1098,0x14314468501698c2,0x812cbba251849062,0xaa98291111148d4b,0xae57000c110003c2,0xabf63a5151a494e8,0xb4975ceb613cb572,0x500021a001101202,0xca320a5059149602,0x48d21508701421d2,0x2a171c8151103689,0x98b34528590c91c0,0x31ec140851008742,0x4002041000000,0xc080480051000242,0x4482000040042000,0x640420050101090,0x0,0x84000000000400,0x90002164114143b,0x80814019001400,0x1014010001000,0x20610ee0401c85c4,0xd7ee4d60511495ca,0x42830406412e1040,0x4cc3fa19701c9bda,0x31ecf13d03c32c8,0xf6bef6eed11ebfca,0x2036056048041340,0x80000008040,0x6004008001c0800,0x10000400002040,0x0,0x80a4262870040480,0xdf8e2c75519cbfca,0xceced36e793e9f6a,0xaddfbfdf793cbeda,0xb7fa5ffb593cbfda,0xdbbc25d87134bcca,0x7a523fd67014bada },
	{ 0xc8c6e3efd0e8b7a4,0xa0d2b95490e0a772,0x8b8ccffb90a0ef30,0x5114007090432210,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1a868e4090202c00,0xebb4fb3eb879277a,0xebffffffd8f3fdf0,0xefffdfdaf8f9bffc,0x34b2166790c32644,0xfb6eaf70d8ebbf76,0xf3a65dff90c2aff4,0xdcc2bf6ff069af3e,0x6f7e9edbd8692e7a,0x33aedffbd8f3be52,0xff69a4f6b0c3bff0,0xfa9727fff0f3bfd0,0xd2d7e34a90e9ef58,0x2b4f36da90f1377a,0xe3a4fc798f993c0,0x99ede6e3d071b524,0xcbb2ac608060af8a,0xf9d7fe2f98f8abf2,0xeeff3f79d86bbf7e,0xffbffffff8fbfff6,0xfff7feeff8fbbffe,0xefbefff2b8c9bf76,0xabd4e33ed8e0275e,0xdfcbbed3d07bff3a,0xaffbdffbd86bbf54,0xfdfff7fe90a8bffe,0xeffcfdf5f8e9ef7c,0xdbb6fb6ef8d9ef72,0xcf5ffebbf8faaf76,0xafffdfdff8fbff72,0xfcf7ffefd8fbbf7c,0xfffffffbd8fbbffe,0xffbf7ffff8f9fff6,0xfdfffffff8f9ef76,0xffffffdff8fbfff6,0xbe7f7fe7f869bffa,0xfffffffbd8fbbff6,0xdeae3ffff8fbbefa,0xcfdfff7fd8fbfffe,0xebfffedff8fbfffe,0xfefffffff8f9bffe,0xf7fff7fbf8f3ff7a,0xffb67dffb8e9efee,0xebfffffef0fbef7e,0x6bffbedfd8fabf76,0xffbeeffff8e9ff76,0xffffde7ed8eae776,0xfcb62f71f86bbf76,0xd3f5ff3eb8f8fffe,0xb7dffedf98f9efee,0x9bbecf5c98eaff8c,0x7efef77598c9efd2,0x818cf5709050ab52,0x9cb0a76bd861b364,0xcfeffebad8f3bff2,0x9dfdd5bd0837f70,0x5416fae0d0e1af70,0xd8e7ffdb90f9bf66,0x9bbe7f6e98e8ff16,0xc9d7f23ad8e3c760,0x63fe5b5798e9aef4,0x14fbffc7b8eaf736,0xf9f5ff7a98f9ef4c,0xffb73f6d90fbbf7e,0xcfdfffec98f1ef6c,0xefd7bed9c8e1afe6,0xfaaf77a9041a7f6,0xf77eef77b0d96762,0xf7b63dfff8e1f774,0xa9d0f71e98f07f70,0x19f51a4b90a1cff0,0x15be4fda90f12942,0x96edf67290e9b7f2,0x7f7ebf7bd8f1ff64,0xdbd5ff7ef8f37f3e,0xfdfbbfdfb8d8eff6,0xafbe5f1290e86f76,0xf5a6ade2d0f1f77a,0xbfbeed73b8ebfffa,0xdefcf7fa98fbe7e4,0x8cc6271498682778,0x2b9ec95190e2af00,0x30bac7eab8e0ff60,0x1fe7bd69d8e8aff2,0xfe962f7798f8bf76,0xe847ef5eb869a766,0xebf6fe9f98fbef76,0xfcff5d6798e1ff54,0xfbfc7f7a98e3bfd2,0xf2bf3e76b8f1bf64,0xbdffb2eb8c1afd0,0x3f7dfed19069b77c,0xbe9a5ffed8f1af56,0x595df79af861ab5c,0x4e8a3d22c8e8af08,0xc9aaf7bfb0ebaf76,0x5afe3e7198c92f72,0x98b9ddf398e02f46,0xa723e261b8403e14,0x10c152090b0a120,0x8944e00218410360,0x8a731280c0612e00,0xaffe9fd2d8e9f560,0x512e066e98810560,0x87931d609089b7c0,0xdbb6aeb698f1b750,0x40c01080010040,0xae6ed9cbd8e92b12,0x34673ec890418f30,0x9b6de16298e06318,0xc800213090410a08,0x8082a22080612600,0x2182315998608c10,0x8010000080000800,0x406ca02090c02214,0x10082b05d0c02660,0x2084020408410040,0x80e085080800160,0x3a41009009a344,0x340800209200,0x2010000080002000,0x10051a080002444,0xa015a508050a286,0x988050288400e10,0x514091c030010270,0x20344008480208,0x5310c00088802004,0x8103be02b040b344,0x0,0x10002000000210,0x100484200480600,0x0,0x0,0x10000280000800,0x1890a6c48820f500,0x20280100002400,0x8000004000000140,0x604000400010248,0x6c5057c0c8612650,0x2474555230e8ae90,0x7b6ceffaf8eab706,0xd6962d739829ba70,0xc0e2db0a98e0a492,0x2b9f9e4098ea2f30,0x3281434a9861b412,0xf1cc96d298219f58,0xcca645e14028ad00,0xa9e7e376b0fabf30,0xe9c2137fd0637766,0x2dbedfe1f0e1eff4,0x604cd66180a02542,0x528a3d400861a750,0xd024036690c0ac80,0x374af21b8879af60,0x2e4a52d0ca2340,0x1040601020b900,0xc58f05288e0c520,0xee94676ed0c1ab60,0xaf46be290861f760,0x820a1c9698e2af02,0x96bf5decd8e9bf60,0x314f75f2d8d2af18,0xcf82676b90e0af12,0x69ced872c040af92,0x6b5616d3b8e0bff4,0x3cd79ffbd0fbaf50,0xda61d7c0d8f08d00,0xf7bc2f7cd843af6a,0x8065a210609a10,0x2f89b4398e02644,0x2308c519b0a03602,0x8000246380102460,0x18c409428020b122,0x2350e0028089a500,0x612c08020a500,0x2c5805208001b400,0x9d1c046248602612,0x80866568d0412756,0xd9c5af2c9060bfe0,0x2002208a00202402,0x6008481010082022,0xf1540cc2c030b522,0x100080000400200,0xd3344a0c88492562,0xc8c39e9090733700,0x4010050000002520,0x94241568806087a0,0xf8ef782986aa712,0x12b82f05d0102250,0x6d86831e9060275a,0x2f5eb7dd90e93f70,0xbdd457698861af40,0xf5edbed598eabfd0,0xa84236490c12444,0x4d8c110a860a6e4,0xab4e1a569840edc0,0xb8b24fbb98e1ae54,0x1200d10080202900,0x200010000400100,0x100600000400500,0xc48291080212250,0x0,0x5424261090202406,0x82252000400600,0x18202700c0400500,0x2048800900010270,0x0,0x1910000090003000,0x835aad30b0c08a90,0xc810010200216140,0x2dc6c91498298760,0xa10a089288e9ae40,0x845074a010688e10,0x802404000000b002,0x2410000080000000,0xacc0400990009570,0x24000010800000,0x3002010000880300,0x1908426a9020bc00,0xca002585d0212c56,0x8947a228b080b70c,0x277fdeff98e9fb76,0x36b2ef5290e9bafe,0xfdffb77a98f9bf52,0xefbc7ff38861af50 },
	{ 0x2ac7eb20493a2e61,0xeee63f574b7eff72,0x8f1acd53ba646524,0xd1c2957185ac6626,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x840000000000401,0x200040000880000,0xfbbdfffffffe7f67,0xfffffffffffeff77,0xbddedfdb53fe7d63,0xfefff66f9fcc7e67,0xff7fef72df6ef547,0xdf526e3d9b383f61,0xcdffbbbe9bfaed77,0xabdcdfd3fbf87d63,0x863a546b25ac4760,0xdb7dff72d97c7f63,0x763eaf7e4dfc6d62,0xefc3da3adffc6742,0xef7efbda4fcafd63,0xb5b14d2a972c6e62,0xf1ffb7f5876eef63,0xf79a3d71eb2ef663,0xabdfcf3a1f6e4f63,0xefff3fcedffef563,0xfffffffffffe7f73,0xfff7b7fbdffcff73,0xfffeff72dbfcff66,0xcbf7e32fcff8ff73,0xecd79fbedfeef767,0xffff4f72fbecef62,0xffff7ffbeffcff63,0x9b6f7c753f4ce526,0xcbfe7bde5feeee21,0xcddfff332fee7f64,0x6ed6cf5e3bfefe62,0xd6fff7fbdfacff73,0xfdffff7bfffefd73,0xfbfffffffffc7f77,0xffdfffffdffeff77,0xff7fffdfffeeff72,0xbfbffffffbfeff73,0xfbfffff1fbfeff77,0xfabf3fffdfeeff67,0xeffff7fdfffe6f63,0xffffffdbdf7eef67,0xf7be7ffbdffc7f73,0xfdfdfffafbee7f37,0xfff67df5cbfef777,0xfbffffffdffeff67,0xffffdedfff7eef77,0xdfffdffbdfee7f72,0xfcfffefffffeff67,0x6fbe7f7a5fbeed62,0xbbf2eb6f8beeff62,0xea5bfaffcf2e7f76,0x909ecf43cf7c7f71,0xf5f186c81afcf512,0xa51ebd620bac7e22,0xb02476648d7c6c30,0xdddfff2e1bf86f67,0xbb0ecbd79b4e6d70,0xfcf27eecebfc7d63,0xf9cd7f427f8c7f73,0xdaee7feadfaaff72,0x4953b71a03fe7f23,0x2b7f5fc96f4c6773,0x34bedfe75bfe6d61,0xb9ffff76ffeefd63,0xfefeaff6ffeeff71,0xe7f7ff318f2ef771,0xbffffedffffe7f67,0x97bbeffadb6a7766,0xd7cd34fc0beef467,0xefbe6ffe9ffe6d63,0xebf4d6ab4d7c6e35,0x3957fff2cb78ff33,0x9f3fccf36f48f563,0xf5fe96f94f7e6f33,0xeabe6d78ca747775,0xfb9cfbb6c9fe2d62,0xea5bdfe71f7ef773,0xefbedffbebeee773,0x557f5ffdc9bc7561,0xffbdfff2dffe7f67,0xffae3d5f5fbc7f63,0xc7fa96501f064661,0xbffedfdbef7ef563,0x96ff7dea6a687d61,0xa97ee7731aec4f67,0x9b84a675c8ac6560,0xd7d2fe92cac86763,0x2f7fd9de1f2cf762,0xbcf4dfffcbeeef67,0xebedff7bdfee7d73,0xfabe7fde7fecef63,0xeddfff6e4f6efd77,0x2edf57f3077e3f32,0xbfb267ff7d7c7763,0xd95d37b3917e7f62,0xfb966d70bb7ef767,0xa97eaf63c74c6d72,0xffdb3b5ffbfe7d67,0xffbfdffbff7e7d67,0xbe76867c59723e22,0xebfe6ffb8bbef524,0x7350e37e87fa7940,0xedd29a02836a7160,0xcfbfff7bcffe6767,0xf5f7e37a03bc7d62,0xd1b2bd324dee6d33,0xeb1e72cb12340f66,0xe1e7be201ac87d24,0x82fc97d39a7c4761,0xd080b2ec49342260,0xd9eeff73abee7f73,0xc39a29034f406541,0x8a03ef498d44b620,0xa9f61e5449f86f67,0x54508040006472,0x8008014007000400,0x9210221001205461,0x4480112002c4020,0x98462a4002340500,0x820000800300600,0x41212102404401,0x40880d4102240550,0x0,0xb8a61e0071087422,0x3190d0a2490a6822,0x9122887083c85402,0x2250000200444,0x822040042020,0x510020084020,0x10034002080504,0xf1ffb77edfeee767,0x0,0x0,0x0,0x8380f02812c5540,0x34f21ec40b2ef542,0x802f084090006540,0x20242000880820,0x100c00280002060,0x10000400000400,0x4023d2608fc86102,0xd100b48a830a6f27,0x840000008400000,0x80a0000209005401,0xf7d19468fece662,0x388051848006462,0xc4cd241000642600,0xded67ddddb7eef21,0xfbffd76fdafc6f72,0x65e63a190bba6662,0x2eb4c7fec8786521,0x302ea2e4a90a7d62,0x8826056001467860,0x1081010d07300470,0xa690f3ae18747410,0x61f4e4242386426,0x71b7675a0fecf761,0x97bbbd31378a6573,0xd62677ec4b666f61,0x4dd3de18976e6661,0x338ecd438fecfd67,0xf6b777eb6b887f66,0x80eb324005f87240,0xb3b42d6d5b387620,0x8eebaa38193e4760,0x67d21b9b4bec6e23,0x147d51328f74ec62,0xd93caf7acface473,0xcebc2fff9b6e7d60,0x40a002000303000,0x2842809209225402,0x12a80dc80a646161,0x0,0x7000b00001c85000,0x8000824000000660,0x6cc609c0c92c7020,0x0,0x92001408808c2020,0x228e24700aec6502,0xc10c613382a82500,0x8d403e018b4e5420,0x6000080140003000,0x10c6538cd2bc6523,0x9f325d6028747d60,0xf3a07a17d32a6762,0x8e47a6838b2e4c21,0xa378d88689ac7420,0x74c4042841ae6461,0x1100a70281ba2700,0x91811d10892e2502,0x85e6df928f407d62,0xef7e7217da48ed62,0x638cff3807c6d21,0x915c7f60eb6e7560,0x881a093807025e20,0x8102821081021400,0xaeff9fd7872eff63,0x8ab04de823bc5563,0x9090560850887c25,0x6600002408044421,0x48040308050c5161,0x60140801407040,0x808a0520a0086820,0x1090402808480102,0x0,0x100040008004000,0x0,0x2e06040800085d20,0x4100000000000400,0x8c0cac32214a6060,0xb2040508082c2062,0x587380248404461,0x2000010020000060,0x402130080886002,0x4080000000202000,0x0,0x8080600000000406,0x804001020000,0x3000000000000404,0xa124ac004524b000,0xca1224624024fc23,0x4d4ff7a8896cfc60,0x3ffffef2cb6c7d66,0xffbfcffb6f6eff73,0xb3fd7f7ffffeef63,0xd8be7df9afec6d73 },
	{ 0xd983f3634f32ed60,0x3fd7fe5dc6963d60,0x119cccb353922d00,0xbdb7d6686ff62d08,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x597fd26ac7be6f60,0x8008a36205b61040,0xfb90fffe9ebe5f60,0xcfffffbfffffff68,0xef7fdfdeffbffd68,0xdcb9fef0fafe7d40,0xd5efad611d9e6d60,0x9bfe677f4fbe7f60,0xeddffa2aeebf7d40,0x2b5efddbdf9b3f68,0xb47a7de35fb63d28,0xf1ccdf62beb77d40,0xe69b27f116ff3d60,0xcdeaf31a9fda6f68,0xbfebfd6f2fe3f68,0x17bfcf78febe7d60,0xd178f6db1fb62f60,0xdba67dc537da2f40,0xebb6ffff9ffe7f60,0xeef77f5f5fb63f60,0x7ffffffbdfff7f68,0xf7fdf7fbfffebf60,0xfab63ff35fbe6d48,0xebffdf7fffbeff60,0xebf7bffbdfffff68,0xe7bf4fffdffeff60,0xf1e6d7fe26befd60,0xf7ba7ff364b23f60,0xfb99fa2b0fbeff40,0xffdfbfbfffbfff40,0x3b7fdfdfbfff3f68,0xfcffffffdfffff60,0xfffefff3ffffff68,0xffbeffffffff7f60,0xffffffbfffdfff68,0xffffffdfffbf7f68,0xf5ff7fffffffff68,0xf9fffffeffbe3f60,0xfebe7fffdfbfff68,0xeffff7feffbeff68,0x3f7fdfdb7fbf7f60,0xffffffffffffff68,0x7dfeb7fbfff6ff60,0xfbfeedfdfefe7f60,0xebf7ff3fdffe7768,0xffff7fdffffe7f60,0xffbfffffffbe7d60,0xff7fff7bffd7bf68,0x1ebc7ffadfb73d60,0xfbd5ef3d9fbf7f60,0xefdbbfcfdffa3668,0x9fbfff7bffb67f68,0x35e606ec37a23d60,0xefecfde3f7bebf60,0xfbfee7be17baff60,0xcfeefc87a7be5540,0xad7fdfcb6bbe7d60,0xd6d5bf6e5f9e7560,0xbeeebf78959e6f68,0x9f4c3f6fc5fa6f60,0xede7ff1adab6bd00,0xeabfdfd63fbebd60,0xfdff7fff7e9e7f68,0xf5ffa77c76fded60,0x5abf2f37f69f6f40,0xeff6ffbf7fbffd68,0xff5f7fdbbef6fd60,0xbebe6f7affbe7d68,0x4e7587294eb27540,0xdbf67f77fffbbf68,0xc9baf6afdfba3d40,0xe8fb9ef3dbbabd60,0x8bbecfaa6eff3f40,0xe677dffbdab62d68,0xdbae3f7afffead40,0xbbfcf76fd7be3d60,0xedebffd99effff60,0xefb28f5f7aa6af60,0xfdfff6f9b7be7f68,0xfffefff3ffffff60,0xffbeff7fdfb7ff60,0xcfdffedffebf7f60,0x2d4e77d8c7965760,0x743d5fec56fe5f68,0x5bef7d615f956f60,0xfb947ff6de922d28,0x97c6ef399abb6f68,0x6b7fdfd7feba6f60,0x94bf766a1e365d60,0xf9c4b75a9696ef60,0xfef9af7fffbafd68,0xedf7ffffdf9e7f60,0xad5ebefe9e9e6d40,0xb6b8cefa53b75d60,0xf0fdff6c7ff76f40,0xcbba6fc5deb67460,0xd9ffefae7f9e6d68,0xedff3fffdff77f60,0xe9dffefcfb77f40,0xd4b2d2fc5e9a7940,0xce06e57a571a6d68,0xfa9063e6169a2d60,0xcc23ebad7d22c60,0x3f3ec76bf6def560,0x75d78ef2b7b2dd40,0xca167c60cb963620,0x5d85a4fbcf1f6940,0x191bdb15479a7760,0xd2815392123d08,0xf4313fc0d6122d60,0x7dcded7266b32d40,0x7983a1150972c00,0x4842000001300040,0x80393642133a9d40,0xa001002002128000,0xcb6c6e46559e7460,0x4272010122d00,0x8022c80004000140,0x20e0804814921d40,0x220050004300400,0x1049827004920e40,0x8924396054842b40,0x40000102020a0020,0x2200000800200,0x488000000200540,0x20004004000040,0x8480d00020a0440,0x190203060aa06040,0x8490300182848c20,0x2110020000920200,0x2006020010802800,0x0,0x802002002000,0x0,0x8adecdc903f62740,0x54d54df016ba1740,0x805b076006b21720,0x0,0x0,0xa000014000c24000,0x6004100000,0xf0706e2026968548,0x4804258000025420,0x3c4a23486a23f60,0x88c2314104123d00,0x202058300860e40,0x5322010006123900,0x4210000e02080440,0xebf4f2b694b60d60,0xeffe1f5feebfff60,0xbefd5dfe9abeef40,0xec15f66d1ed3ed48,0xcfec2e6ad7983d68,0x78c4120ac09e2f00,0x212322e9b922d60,0x3084c1000221420,0x3dff026e5e16ad60,0xf2be355056b62f40,0x4400e34e34a61440,0x695fb2b117df7e60,0x3aaecddacee6ed68,0xf6fe3deddfbfbd60,0xdfb3fb52ce323d60,0x2808240002228840,0xc156fb544f8e6d40,0x8b7cce4e0fba3d40,0x3ed4f5eb9ff36d60,0xe9548e181cd43f60,0xdabe297217bc0560,0xedcedf080eb6ad48,0xec1bbe696fb23540,0x802491202926000,0x20040002000400,0x7882054010b23440,0x4c20406300420,0x52015054800440,0x2000004000000440,0xd350fe084b9a2e40,0x0,0x8b10c1281a8a0460,0x300001000400,0xa0440014900000,0xf1d73e7b43967d68,0xda28ad418bbe7640,0xc080b174121e2560,0xfceceab07962f60,0x2b8e8d9e0aa62d40,0x24cce64096962d40,0x460a551028a2c08,0xc491121006373c40,0xc4c6f0bc4fb26d60,0xa3e9959a1b963d20,0x9475f04197840c40,0xfbcd77627f9a7d40,0x1a8daf20c2882760,0xdbd2f7b253ba7d48,0xefffffdf1fbf7f68,0x1d885d5b97b2ad60,0x5b3e3c4a96ba1f68,0xa110202040001420,0x400800000440,0x80000a1010a00040,0x0,0x420002004100040,0x200010000104000,0x0,0x2004000080400,0x0,0x8084024842102500,0x81002c0102922d20,0x92143b8557b26d40,0x100900100020c20,0x90ca0108120540,0x2024004002001400,0x600500010a0500,0x10020000000440,0x0,0x100001000004040,0x840110001120400,0xa30803400a802440,0x9b1a001006822560,0xcecb5f2c5b902d48,0xffdfffdfdfbfff68,0xf7bb7ffeffffbf68,0x50cc376fc6963d00,0xdffeff6b5f9e7f60 },
	{ 0xcba1f7fbf7bc6d7d,0x7fff3a85d2a460e5,0x65bccd9ef6ac21d6,0x9cb6e651d2842e4f,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfffffeeef7be3bf4,0xdfcba572d6002b77,0x4000400814040040,0xdf7ffefff7be7fff,0xff5fdfded73e7fff,0xf6fa7fefd79e6bfd,0xb377f77cd70e277d,0xcfb77ffdc7243fff,0xffffffffd7be6fff,0xabfeffdbf6242ffc,0x1cdf5fe3d634211d,0xd5f5ffe9f61e6b95,0xdeae3f76f7842bfd,0xcffaf7bdf73c73ff,0xefdededbd61e6fff,0x96bb6de5d636095e,0xf66f36c1d3b26af4,0xda942c74d60c49c4,0xc9d2f37df6b62bfe,0x2fdf5fd7f7846fff,0xbebf7ff7f6967fdf,0xf77fff7fd2b67dff,0x7ffe7f73973e3fff,0xfbf5fffed7bc3fff,0xffffffbfd7be6ffd,0x6fbecdf7d61e6336,0xd5f47fecf6a46ffc,0x95b96d53f6ac22df,0x9a95d36fc6ae23dd,0xdffffffff6be6f7f,0xbf5ecfdbf72e6bdf,0xf4fffffbd7b67fff,0xdfffffebf7b67fff,0xffff7fffd7be7fff,0xdffffffef7be7fff,0xafffffdbd7be6fff,0xffffffffd6be6fff,0xfffffffff7be7fff,0xfefe3ffff6be7fff,0xffd6ffefd7be7fff,0x7bffffd6f6bc7fff,0xfffeffffd6be6fff,0xd1effffbf73e2d55,0xffdefdfed68e6fef,0xcbffdb3ef6ac6fee,0x7fff9fdff6be7bdd,0xbfbfdffbd7ae6fef,0xf5fff7fef6ac7ff7,0xdbfcedf3f6342b75,0xf9dcf77ff6a66adf,0xfcfbfef9f63e61fe,0x2fffef7fd53e77fd,0x70949eead28e29ce,0xe7feff70d7b67b7d,0xffbf6ffff6be7ffd,0xe9dede93d72e6fdd,0xaffe5f5ed7b6367f,0xfdfff7ecd62e6fde,0xa94bb772d70474fc,0xdf8cff6ff6266fe4,0xd4ced912f63420cf,0x6ffcdfdbd6967fde,0xfeffffebd2be6fff,0xf1ffbffbd6a46f7e,0xdf9f7ffef68661f7,0xdfeeff7ff7ae77ff,0xaf7ffedbf7b63fff,0x9fbeffffd7ac7fff,0xb3fdff70d6b427fd,0xfbfe3f7cf23e6dfc,0xebf7f73bd62e3bf5,0xfedfbfdbd6be6eff,0xbbbccfbad7942bed,0xd7aaf77bd72e2d5c,0x5ffebd60f62e3bcd,0xdbf7e73ef23e39fd,0xedebbe8ff6a42bef,0xaffedf77f63677ff,0xf0fbff7af72267fd,0x9ba2ff73f63e2d5c,0xdbbefffef6be7bff,0xee4fff3bd72436ff,0xaf5a4d5a920c3d86,0xf9f44ef0b62c3969,0x92f6ef6ad60635f5,0xbbc6fe17932e2d54,0x7ddfff2ef7b47ff7,0x4b7fdddbd63e2ae5,0x9c9fd72fd69469ef,0xdb7eff7bf6be2377,0xdfbfbfffd69c7fef,0xeffefffff6be6f7e,0x2addbaddb2a62b46,0xb0938febd63473e6,0x95c8dd65d22429c0,0x7a9e3ef1d60e6fd4,0xdbf5fbaad60c79fd,0xf7fb32dbd6b43bf7,0xa5185fe356242362,0xf7f1f67cd61e69bd,0xd28e7debd6bc7bf5,0xfb8c636ef624637c,0xe8c6bab2d2266f6c,0xef9e5fbbf62e2f7f,0xb4f47642d6042fed,0x921a3d70f60c6bec,0x4b0216ad829c28cd,0xc01021c92000134,0x6b4b991bd70625ec,0xd614f444d6b42174,0xdffcaf72d62429ff,0x7395200dd6046285,0xa500e40286002044,0x8667c10d20e22a4,0x1685601920c0160,0x4800a480d6040180,0xc6902b00c2042095,0x7008b122920428c4,0x600002800,0x8010080000010,0x50000800040,0x20400041000,0x8000822040202000,0x98062748d6142da4,0x2ca9de28f4a47ae0,0x5000a00090004040,0x400c712080942860,0x1d20001100c2001,0x4480001000242860,0xc28180046100,0x0,0x400000c0000,0x2088c002408e28f0,0x304a302540e3880,0xa10b481291841240,0x58314360560420a6,0x918aa718d4804144,0x0,0x42000000200040,0x100080010000020,0x1000040080004000,0x28082000002001,0x2002000040080,0x40c0a000c0002980,0x101207900c2200,0x4c100242080,0x220d0882102d04,0x50242c60d6004154,0xc0c074a4c610201a,0x5fde3e39f48c7aef,0x8b9857abd4806264,0xfd7dfad0d69e31f4,0xeaaefdc1d7046d7d,0xda80b30e44061296,0xce9bbe8bd33e43ea,0x247a1272d6042914,0x5536a2f2d62c00dc,0x9208256154040220,0xcb28e36ed60c21dd,0xefdeba9fd6867d67,0xa31117dad6042514,0xf4fbdce9f6a66b76,0x506f8dd2d60e523e,0xeab42facf71e013e,0x60dbfa06d61420d9,0x4ededdcad21e6afc,0xbcf7dfecd706677d,0x616f37ead68e69c4,0xda362f51560c2b5e,0xcdcef67ef68c7954,0xabf67edff69e7fef,0x5aa4dd292064168,0x1080000010000040,0xf6082f4cf684246c,0x8100f5aad00421a8,0x294611d091204870,0x11094fb2d2806064,0x8576e728d6022e5a,0x2010000002000,0x800411090000240,0x895232289200254e,0xd084360f68440c4,0xb5f58c68f6242ba8,0x4b2e4e2192042e24,0xb2386640c6024156,0x4043200244060080,0xab9ecd8bf234255c,0x841214a882242554,0x142243854002041,0x9bb24474d6042157,0x8750f4aa920c694e,0x2305d7dad22222d5,0x363e4cfad60e2b9d,0xfb77ff73f2b62ff7,0x5ea62fed92166b65,0x8bf7ff7fd6be67f5,0x2a59cecbd6a62dc8,0x81024f6356062343,0xf024872006042562,0x48182cb0c4040090,0x1d0420812043824,0x1010940020c0204,0x400000a082000100,0x406120192006200,0xd887dd2f610234d,0x0,0x0,0x120c000080204040,0x100066044041002,0x6b7e7d63d73c2ded,0xcf2c55e6d494637d,0x100040044000000,0x200000880000010,0x0,0x4020000044000000,0x9280220000002c00,0x10810000100,0x204191c256040066,0x8400080040000,0x1200000000002400,0xd8808b52448c28c0,0xcfd4d627d62469dd,0xefdffedbd79c2bff,0x3fbe5fffd6ac79ef,0xde7e9679d216237f,0xdf9e3f64d7ae6f7f },
	{ 0xe3c3fa77a9f4633c,0x77fb3fdb8b9e5bae,0x5086036688c43224,0x94fc36a989b4332f,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4480482002840220,0xfcfff7fe8bf677ff,0x93bfaf56abfc5b3f,0x100042080340220,0xdffbffbfabfe7fff,0xffffffdfabfe7bff,0xf7fe7eee89be5bff,0xfb7def7bab9e7f37,0xeffe7fff8bfd7fef,0xfbffffff8bbf7bff,0xaf77dedba9de5f2f,0x909756ee8b946bb6,0xdbfbf77e89dc5bef,0xfe323d6389be73e6,0xfdffffff8bbe7f6f,0xfeffbf5dabf95f77,0x309b4ee28b9150a3,0xdf7faf7c8bfe7fb3,0xdefe7df78b955b73,0xeb97fa278bfc577d,0x6ec78fd509bd437b,0xb7ff4fff8bbe77ff,0xffffffffabff7fff,0xfffe7f738bff7fff,0xcbfef73e8bfb4b27,0xe8dbffff8bf45bee,0xff9ecffa8bfd7b77,0xb5fd7fea89b657be,0xffbfff728bfd7ffe,0xffffffffabff7fff,0xc7f7fabfabfc5ffe,0xbbfdffd78bf46baf,0xfcffffffabff7bff,0x9feffffb8bfc7bff,0xffbe7fff89bf5ffb,0xffffffbf8bff7fff,0xffffffdfabbf77ff,0xfdfff7eb89b67bff,0xf3fefffbabfd5fbf,0xfefeff7fabbd7fff,0xfffff7ff8bfd7fff,0xffffffdf8bfe7bff,0xbfbb7ffb8bbf7bff,0xf7ffff7babbb7bee,0xfffe3fff8bbd7ffe,0x8bf7fbbb8bfe5336,0xeff77fdeabbc7ffb,0x1eba6feb8bfd73ef,0xffffffffabfd7bff,0xfffefff2abfc7ebf,0xebf6e77fabfe7ba9,0xfeefbfdb8bfa5b6f,0xbfbfdffb8bf87fff,0x5ddc7eee89fc57e7,0x93afad32abdc5bea,0xfbde9f3e8bff7faf,0x8bd6fb1789fc53fe,0xbbffdf5eabf47b79,0x76b7dee98bbe5f26,0xb9e31d708be8636a,0xf7be7f348bbc7ffa,0x84e2b71a8be663a6,0x6bde3dd189b61bf2,0x3eff7ffe8b9e63ff,0xfdffff77abff7f7f,0xfebe6f7e8bde77aa,0x8493f287abd30b36,0xcbf79f5b8bf61263,0xc6b3ffd78b9d67a3,0xd7ac37fd8b8d7b7b,0xfb722fc58bfc53a1,0x89dffbfe8bbe47f6,0xeff3bfdf8bbc5bfe,0x9f5ecfff8bfc7bff,0xd7fff6faabbc7fbf,0xf2f6ff718bfd72f2,0xcbf7f7beabf55fee,0xeffbbfff8bff73ee,0x6fbedffd8bfe7ba2,0xd7cfd7eda9fc5bbe,0xdbabfc56abfc7bbb,0xf79fffff8bfb7fef,0xf5d7bfbe8bfe4b7f,0x610fde9289740b6a,0x10b7666e8bbc42a1,0x93ffef738bb45bbe,0xcb867f6f8194737b,0xffdeef678b7e6bac,0x2afedfd189bc7bbe,0xa7fe4ffb89b64fa7,0xfb7daeff8bbc7332,0xeab627678bbd7eee,0xffcfff7fa9dd7beb,0x24ffdf728bbc43ae,0x9eebefff8bf67fb6,0x93fcf64689800b36,0x7fce772689fd5327,0xc9a1f36e89f85bab,0x6df3bfdeabbd53a7,0xbdbadfb88bac7b6e,0xf676a6fca93853ae,0xdf22edc08ba943aa,0xf39cfe0689ff42be,0xee4fbb5e8b8e53f5,0x7bbfc7fb8bfd4bfb,0x7cffbfe6abf87326,0x8a82d620b904188,0x880000200800002,0x4022808098a0209,0xa30c051f89b84390,0x945497e18386032a,0xdffdaf7dabbc6bbe,0x700e3ca10b90426c,0xc77bfb3b8b9363fb,0xe990279e09c00a03,0x34abdfe8a3a6233b,0xcb7c672809b076be,0xca902529811c0a0a,0x127028a1800282,0x275cdfd8a9a073fc,0x13329c089804923,0x0,0x490c036081b04100,0xd200f92283b003d7,0x8804000602800200,0x8a804d0803940328,0x9000000000100200,0x908240201900201,0x2989e10a0130132c,0x800050818a801100,0x203156c00a880211,0x904809000208,0x40200800001,0x200000208400000,0x10100000040200,0x2002008180200202,0x112009688a800222,0xc4220702a0140202,0x201028480a404208,0x4000202800000,0x108100800202,0xc502000b00201,0x4008300002800001,0x4002000001000020,0x100040000000280,0x80010a040200,0x100480201000120,0xf1f514888ab04283,0x0,0xd97bd2ec8bb9723b,0x6b7697d7ab9a037f,0x1f99cff289b41b6b,0xc4df36e88bc47aa0,0x8bbe7e718bfa533b,0x80042e0e80840ab1,0x0,0xa020006081004084,0x59f3ae6e8bf47ba9,0xcbbffdf08bbc73ae,0xde8d61e78b987322,0xfd7fa230bbc7362,0x2e5e55d089904b72,0x80d0146189385268,0x37eef7758bfc773b,0xf59c6f75a1b47bbb,0xffc76eb28bbe4bff,0xaa971f4188e97aab,0x5cd454bb8bb277ff,0xf5ddae6c8b9273d9,0x54222d56abb80320,0xc77ed31489d07b2a,0x3b57720babb41363,0x921a7ff08bf81337,0x4404814081005800,0x4986304089620002,0xcb9afe3b8bb843bc,0xe2ca48db8b305204,0x411878b00b884233,0x1422604001104240,0xa2615a0031c4320,0x8164d22001300360,0xc4b52d889384206,0x5980c9200b880112,0x94f434ea8b805b22,0xcfbe3de08bb84fb7,0xd896b70c8bbc7b70,0xe4e7e01abb9428f,0x8077454a8b800386,0x24a0e86909541300,0xb874f77189d94ee7,0xc6852165a18443ba,0x6840e712ab88603a,0xdb3edf5f8bb4436f,0x9c5614e589f40322,0xf96faf7eabf97fff,0x9cb6057c819057af,0xcbfff72eabbc73a7,0xebdddf4b8bf07a67,0x36b85e6b893c7e66,0x9155947289344b20,0xf0024218b805130,0x8844431809d22320,0x41111c90881c122c,0x3b384d000986632b,0x160067081821462,0x1280280080100e20,0x8002009000044a0,0x0,0x12092180200122,0x2068000800030,0xa586070891c0342,0x810000480a800001,0x102880001004120,0x290d9100ba82320,0x1a40a081900014,0x840240009004002,0x88005240810040c6,0x6044619089114286,0x2c21e1083901328,0x2008500004288,0xc929b00209206360,0xb8102500008063a0,0x4102c12289846322,0x2b73be4da3bc53a2,0xbffffffeabbe7bff,0xdcfdfff989fc5bef,0xfffefffe8bfc77fa },
	{ 0xc9b2f31b5bd4d1da,0xebdf1fdf7bc5d32a,0x9baecdcb6ad5c33a,0xc77a7ef80bc4c23a,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x85ab412a82c4e10a,0x7cfff6efebc4cbfa,0x9210a571e3d4fb8a,0xebb7db7ffbd5fb3a,0xcffffffbfbddfb7a,0xef7fdfdaebd5fb6a,0xbfffff6ffbddc3fa,0x79beff62fad5d93a,0xffdeffeffac4f32a,0xffdffffffbddfbfa,0xefdf3fdbf9d5f9ea,0x909337e3c3c5e392,0xfd7cfffe69d5d0ba,0xbebe2f5cf905eb9a,0xeffffdaefb4dd3aa,0x2f5efefbc9cce33a,0xb6ba77f5cbc4ebfa,0xf6fef77b7bcdd3ba,0xfefe7f7ffbdcfbfa,0xebb7ab7febd4c33a,0xefef7edff3d5e3aa,0x1fdfede3fbd5e37a,0xfff7fffffbddf3ba,0xfbbe7ffbdbccd17a,0xfbfdff6ffbddd37a,0xefd3be7b63ddc3fa,0xefffdff7fadce37a,0xf5fff7fffbddf3ba,0xfffffff3fbddf3ba,0xffbffffffbdddbfa,0xcfdffeabfbd5f3da,0xbf5ae7dbf8d5e37a,0xfefefffffbccfbfa,0xffffb7fdd1d4ebfa,0xffff7ffffbddfbba,0xefffffbffbddfbba,0xefdfffdffbc5fbfa,0xfffffffffbddfbba,0xffefff7efbddfbfa,0xffbefffffbd5fb3a,0xdfffff7ffbdddbba,0xeffffffbfbd5fb7a,0xbffffffbfbd5fbba,0xffedfffff9ddfbfa,0xdfbeefeff9cdc3ba,0xc1e1fdfeebdce37a,0xefdf7e597bd5c37a,0xffffdfff7bd5d33a,0xf5fefeffebdddb3a,0x5dfe7df2fbd5e1ba,0xfbf7eb7bfbc5c32a,0xce6a7e8bf9c4eb2a,0xaffecffbfbd5eb7a,0x54d4fffc79c5c1aa,0xf9fe7762f9dce32a,0xebb0770e63d5c32a,0xeff7f7937b4dd32a,0xebffdfdb7ac5d17a,0xf4b6dbe8e1c1d9fa,0xd3fb2f78f144f3aa,0x9f1caff77944c34a,0xd8f6ff9e6984d3aa,0x6aeefedb79ccd3ba,0xfefbffebf3d5cb3a,0xffeffffbfbc5cb7a,0xfafe3f767bd5c3aa,0xcf7ff7f3f9c5fb3a,0xebffdedff9c5f3ba,0x3fbb7ffffbdcfbfa,0xf5fcbf7dfbdddb6a,0xfeae7fe3cbdcfb72,0xcbbdfbffe945fb7a,0xffffbfdffbdcdbba,0xeeb6fff3e3dddb3a,0xfd7cfa7ff3d4db7a,0x8bee7d68d1d5f13a,0xbbfcf77f6bc5c37a,0xeed3bf9ae9dcfb2a,0xfb9e4ff1e1d5eb6a,0xf4ddf6eee3d5f3da,0xdbbabf70f9d4e33a,0xed97f7fde3ddcbfa,0xdfdffefff3ddfbba,0xff51d35343dcd27a,0x54bf36e051d4c37a,0xd3fff756f9d4f378,0xdf957f25f1d4ca3a,0xd8c7a2075bc5c10a,0x6f065a487244812a,0x9df77febc9cce23a,0xf95cad78f104f13a,0xdf85277c6345d32a,0xc0d6d1367ac0c12a,0xff57dbca4bc4d1ba,0xfebcfd9fbdce32a,0xd1fdf65bc3d4db1a,0xdb1277e4e1c5f08a,0x8b46dd2febcce31a,0x28d6de137bc4c13a,0x1f9e4f05d38cf33a,0xadecf77e49d4c30a,0x477efde0e34de18a,0xf9ffff3ffbddf3aa,0x2dd2bb39cb4c734a,0x2138cd7bf9c4e08a,0x4d70d6ea71d4d10a,0x93e1532a305e022,0x99bef5e901d4c32a,0xc5809c0002c18152,0x20348584104130a,0x949555a951c4933a,0xcbaa7fe34144c34a,0x82a43e4bda05eb02,0x8002c30040d41002,0xb98cc418044e83a,0x10500000404048,0x40060370e0446202,0x2240000044002,0x2080000040200,0x57564970c48302,0x4a240840900c10a,0xf04694560bc4c220,0x528a246d7001d11a,0x0,0xe232604004d31a,0x10200d41c080e100,0x0,0x4880002000,0xb3a1292ad3c4e96a,0x1c1080404040,0x6000003000000,0x51400001000200,0x1220a40003040010,0xb024834050001112,0x0,0xa01004000800800a,0x0,0x80468248880a102,0x20280200440012,0x41200080042010,0x108000080002000,0x460600001848108,0x888500608080c008,0x0,0x690414706145c208,0x8c1cc4083002012,0x8007c1d1040a00,0x100d82602b81c002,0x4040245040041308,0x8011d2006000c128,0xb61e504005c10a,0xf9bdad7beb05f12a,0xcc76ab106144d30a,0x7cbe75e3c3cdf36a,0x519112008044a028,0xdd22e22e0c0c91a,0x6022012040c0110a,0x543264806084c112,0x10831200080c102,0x52907be6c3d5c30a,0xcff3feb5ebddc99a,0xebfcdf57e3dceb5a,0x10b28e68c1c4c248,0x99ef7d4b5b84c13a,0xdfb66ff3eac5f30a,0xcdcffebe7bc5d37a,0xaadf5fdbf3d5f13a,0xc5baf36bd3d5c91a,0xdbeff662e1c4f93a,0xf6bf1ff2f3c4ebaa,0x4a5477a4fb85e00a,0xeff39fd6ebd4d9fa,0x84bb6a7b6ac5d33a,0xb7fce76f59c1d32a,0x9063e208084e912,0x8880062400d0c102,0x4d470ee8c0c0c82a,0x1086b70800dc822,0x2100002880002108,0x44ba00708344430a,0x94082040088c110,0x42000000004002,0x1298c420d880c02a,0xe18235684184d31a,0x89e4ef50e38cd10a,0xfaac3f1449d1c2aa,0xc1c05203d9045008,0x9100100c0c002,0x47845606040c10a,0x31ee4662c104f32a,0xc69c3f62e2c4cb0a,0xe9e7f3166bcdc33a,0xbd7b7dec3d0e37a,0x6eb40de372c4d33a,0xf975fe7ac1c5fbaa,0x3c163ff0f1d5e16a,0xdfdbf3eacbc5f33a,0xafcfbfdaf9c4f16a,0x35aa6fe2dbc4cb5a,0xba5cbf09e1d5d9a2,0xb29c2df08144e35a,0x8005e38b8144e39a,0xaa4d024a8354e13a,0xa98ce5359c4431a,0x242e6272705dc108,0x720464014350510a,0xf83442226144021a,0xc40200000458000,0x4488004a00a,0x3000088060001000,0x4818081203004108,0x8e2c054e41c5c1d2,0x42400a2150d40012,0x0,0x188153000000,0x0,0x0,0xc44ec0007104c2a0,0x9ede5ad08944e3da,0x10984c41e000e02a,0x7cc60bc90cf30a,0xc406244980c1e138,0xcf6bfb7f4bd5c35a,0xaa5f184641d59308,0xf6baeff7ebdcdb7a,0xd6ffb7e17bd5c33a,0x614655eb01d4c19a },
	{ 0xc9e7f2ebc8db76a6,0xb6fa7b53c6dabe60,0xb5bf0f7744caf5be,0xf676d67cc6cbb230,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x55c24da3ced3b622,0xfcfffeeececbfff6,0x840820018093b140,0xfbbbf36fcedbf6ee,0xdfdfffffcedbfbf6,0xffffbfdfcedbfeee,0xd4bfdfecce8b97fc,0x7beff663cec2fbe6,0xdbe076efcec3f7a4,0xefffff3fcedbfff6,0xffffffdbcedbfbee,0x442f5ee7c69ab7e4,0xd17ceff3cedbf66e,0xee9427fe4ecad36c,0xcfccffb3cec2f3f4,0x6b3fbe5fcec2b3ce,0x695dc704e9bb262,0x516df7624ed2b276,0xffbeffffcedbfffe,0xe8c5fbebccdbfb32,0xefff9fddcec3ffb4,0x3e9fdeaeced2f370,0x77eff7fececab6e6,0xf5be3d73ccdbfbae,0xdbf0f73e4edabf5c,0xffffffdfcedbfff6,0xffbe4fffcecabff6,0xfdf7fefd8ecaff66,0xdffefff2cedbfff6,0xffbeff67cecbffee,0xcffffebdcedaffe6,0xff7fdfdfcedbff7e,0xfe77feebcedbfbfe,0xff7ffff3cedbfbf4,0xdfddff6fcedbf776,0xefffffbfcedbfffe,0xabffffdfcecbfffe,0xb7bf7fefcedbfffe,0xfffeffffcedbfffe,0xfebeffffcedbfff6,0xefdffbbacedbfbe6,0xeffffefdcedbfefe,0xb7fb5fffcedbfbb6,0xd7efffeaced3bf7e,0xf6ba3f74c6d3f7a2,0xe9def7facecafbb2,0xefdbffdbcecbfb9c,0x9ffbcffbcccbf7fe,0xf777fef7cecbb7ee,0xdfbe7fe3cecbfbf4,0xbbfdcf3c4ecbf784,0x6e53ffd3cecbffac,0x6f3ecffb8edbbf60,0xfd7576e84edbb7f2,0xdf8aa772ce88b6ac,0xebb4f7fbcec3fb90,0xdddffea14ec3fbf4,0x8f5ededbcedbbbd8,0x54bff6eec6da7a34,0xa7477fe9cecbfac6,0x99142f344a52bf30,0xd343a73a4ed2fa30,0xabdddfdfced3fbb0,0xb7fbf7ebcecba766,0xfd7fffffcedbfffe,0xfa9c4f72cedab762,0xebe7dfbfcedbfe8c,0x27ff5fdbcedbfbf4,0xb73a7ff7c6cbfb26,0xddef36fbced3fef2,0xb28c7da1cec3fb84,0xcbe5fbffcccbf6fe,0xebd69e9fcec3b720,0xb6dfdffbcccbbbee,0xf5e6f6efcedbf3e6,0xfafe3d71cecbf7f2,0xfbf5e77dcecaf7e6,0xcfdbbed9cecbfbb0,0xf53f4dcb4ed2f3fc,0xd4f6ffffcedbbab6,0xfbbf7de1ccdb9a24,0x8bb05f3ecedbfae4,0xdd7efebbcedbfff4,0xebbfdfdbcedbff7c,0x37ffc4c08acbbac4,0xd9f4b77acecab3b0,0xce3ebf5486da39a0,0xfde7bf9c46cbbfe0,0xee185b494e8ab6e0,0xaee227fbce83b6ec,0x59ecf55244dbffb0,0x9eb8277cc6cab77a,0xe9aff31bce8bb6e2,0x69fbbfe3cec3ba6e,0xbffb6fcbcedaf3f2,0x797f376bcec1db1c,0x7aa67d654ccab354,0xe9f3ffae4ec3aaf2,0xe1671ef8c492d76c,0x5998dfffcedbfff6,0xb1fffe48cec2a320,0x5fbe76f0ced3fb60,0xf9d4d9374ecafe20,0x4101be12c8c21a20,0xaf9ecfdaced27ef4,0x4104a6c2084232c0,0x840ec40ce829204,0x8b2224664c8ad010,0x48c5a000a022300,0x318893d2c492b04c,0x10be76c9c0d39130,0xdbeb657acec3f39c,0xb3b61f74c69bf262,0x3ccf3328c8ab660,0x98cc2518a839240,0x86b254e846883004,0x4f41b7720ed9b222,0x56943f2cccc3a220,0x2996d6364ac3fe30,0x846d88884c22122,0x480044c0c002600,0x300c042044121400,0xc1cee062c4820024,0x0,0x200000280001000,0x100009c448808204,0x8000100080001000,0x2a0d604ac22802,0x10404020000,0x5012080040a242,0x200040004000008,0x1004008001000,0x93c8e4524cca7462,0x8000400000020200,0x4000442008004,0x2a04094000121280,0x402000008100,0x1000004004020000,0x12200760ca81a204,0x5000000002002000,0xae504800000a3220,0x80100400010,0x400cf53046929144,0x4a26217a4c8ab5a0,0x8013da014e8aa286,0x2f4afe4b4ed3db00,0x9f325f388ec8fa20,0x46406004001d004,0x46020104004222a0,0x4200410a42c08200,0xcd5c1e4846c2a602,0x77aedffbced3b6ce,0xd626b2f4ca83d204,0xcffefff3c6dbbffe,0xb880c77c44caad30,0xeef3deb5428bb6b4,0x4a0040000022180,0x56f61eebcc83f3a2,0x8b1a0c024cc10210,0xd980777cc4cb834a,0xcefebe238ed3f3fa,0xbaf7df4ace83b38c,0xf4d7fef8ced2fae6,0xdeec7562c8d39900,0xf3baafe6cecbb370,0xa4d6eb3ac69ab686,0xb3db17dbcec3fa72,0xa5f917efcecbf6be,0xf9fcf7374ecbfb76,0xfe9e2d77cec3ffb4,0xcffef3734edabba6,0xeaffd8cf0ecaf290,0x400046204822040,0x940c8ca6ccc838a0,0xc84e752286d29230,0x8054932884585014,0x2412368004127020,0x2a8cd840cc2a780,0x3186221886820200,0xfabe6dd24edabe86,0x25c420202087220,0x4819600064a2840,0xa42100e1060292e8,0x26802008200,0x89a47c424680f6b6,0x921a271006493e02,0x4d14e20460a5220,0x2804409b8098d000,0x5424c4a08e81f200,0x81182768ccd21c10,0xcb58af0bc6c2f83c,0x6dd6f702c68bf426,0x6a5ff3d7ced2d2ee,0x16b37ff2ced3f634,0xd974f64b4edbf1b6,0xefbe2efbcecaf7ae,0xcdc7ff3fcecbbfee,0xfffffefbce92f3fc,0x94924ea3c6c2af24,0x73c836b8ceca32a4,0x3a866f74cacab652,0xe9f3e33ac2db9666,0x20b394a0c8ad220,0x9a9a9f584edabb94,0xcc152e2c46c38a4c,0x68c3840064a2a00,0x9000116044c00204,0x8135004c20240,0x8409cd6044c22204,0x3000000000002020,0x14052040c08804,0x8800000080800000,0x1548e0008401200,0x828080282101204,0x0,0x20004040000004,0x0,0x842020e0642a300,0xad545ac80618f148,0x342005b208c1d200,0x10aea500044222a2,0x529614d0c8d297a4,0x608ae13aca0a9734,0x2591bda46cabfb0,0xfffbffebcedbbaf4,0xd9fffe684ecaf724,0x5bf67bdec6cabb2e },
	{ 0xebe4f6ef8d737b9b,0xee561edfca473e9b,0xcd1adcf016e2b611,0xf5edf6fa8fc76fbb,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x9f8adffbeb733ab3,0xfdfff7ffffd37fbb,0x8001000000004800,0xeb99f76baf737fbb,0xdffefeffdb723bbb,0xabe8fd3c9f33a83,0xf4b77efecfd3bfbb,0xdd5b37f05a63be29,0xffb67f7f5fd7ffbb,0xefffffbfff73ffbb,0xefffffdffff7ffbb,0xc53a5bff09d35a1b,0xf9e4e7eacce2bf1b,0x5a973f7adb53bebb,0xe9c7f7ffdad37eb1,0xaf5f7d5b49737b9b,0x96f2ee2a28733683,0xf56e767d8b737fab,0xfebe7fff9f737fab,0xebd7ffba7cf6fb1b,0xeeffbfffdbf7bf8b,0xbfbedff7ed73fbbb,0xddffb6ff6bf77a83,0xcbaefff3ca53fa91,0xfbdfff7e2dd37eab,0xbfd3ff9b1fd3ff9b,0xbebfdf7ffff7ba9b,0xf5d6f779ddf3ff93,0xffbffff7dff77b9b,0xff8fef6fedd3fe1b,0xeefffab609763eb1,0xaf7cdfdfff537fbb,0xf4fffffffdf3ffbb,0xfffffffbfff77fbb,0xfffe7ffbff737fbb,0xefffffbbfbd7bbbb,0xefffffdffff37fbb,0xb5ff7fffcff3ff8b,0xfbffffffef77febb,0xfebe3fffffd7ffbb,0xedfffffffdd7fbbb,0xfffffeff1f777fbb,0xbfbf7ffffdf37fbb,0xd5feffff4bf73f93,0x7bf6ff7f6ff37b99,0xe9f6f37faf737e3b,0xefcfbf7b1df37fbb,0xedaedffbdc773aa1,0xfdf7fffb6ff37fbb,0x7fbefd7239f37bbb,0xdbb7ff7fcbf3bf8b,0xaef3bf5909733f01,0xefbccffbc9f71a91,0x10f7f6c089c25a81,0xbb1b35704d73beb1,0xe3267b664e52be11,0xcaf6faa71a737bb9,0xfbdecddfcd52fa11,0xd4b5fee90ef27e1b,0xdbecf73a0853fb3b,0xdeb73f6ead573fab,0x41ceff1b0e763b13,0xeffcfd5bcdf33f93,0xc7be5fb78df33a1b,0xdff5af7299773fbb,0xdebe3ff7bf777f9b,0xeedafeff6ff3fa9b,0xaf5ffff71e737a3b,0xfefb6fff4b73ffbb,0xfdff7efbeff37eb3,0xdfde3f7f9b73bfbb,0xebfcff620fe33f9b,0xebf71fd7cbc35e1b,0xb5fedde28df3ff31,0xf77dfffb8bd35aab,0x5bfeff6b8b73ffab,0xf9f7fb2ebff7bf9b,0xffe3fffd5ed3fb91,0xaefe4f5e79677a93,0xf7365efb8bd33b1b,0xbff6ff71bdf73b93,0xdfb5efef7dd7febb,0xcf5ffe8f0ff7bf93,0xffecd5a0cf73b1b,0xf69726e828763a1b,0xb5efff744f723fab,0xfa872f744c673a9b,0xd9c6b3980a521e8b,0xeb9bdddaccf73f0b,0xb4f6776e99e33b33,0xdffef7fb7ff2ffa1,0xdcbe2f78c9671b9b,0xcbc4f777cdf37f03,0x3d76ff5f3b77babb,0x77ba5ffdce737a9b,0xd5fd1e7b18727b99,0x6db2fd4c48723abb,0x8ba3ba7bb9f23bb3,0x64e9fe518e4223b1,0xd054f6158533a8b,0xf69e7ee54b53fb01,0x508e6d631a721713,0xf184f7741b537f01,0x4043ae38bb461a13,0xeabce7f98af23b1b,0xf576b66f5a52363b,0x0,0xc8ec672288c35a01,0x52104000021011,0x20191500492320a,0x541404340a023891,0x28ba8f304b42be03,0x8addaf7999d33ebb,0xb8cfffce48f27e11,0xc8d29f41c8021a09,0x404124000640600,0x2108020008009901,0x482a2c620d023910,0xe344e1309af35a01,0x210cc21009010901,0x2409642848601201,0x806848000001,0x2040000800800,0x200400000021000,0x0,0x421014000500220,0x9400200000400202,0x280040000804a00,0x0,0x40020000021000,0x0,0x230d669a04202b0,0x0,0x0,0x4c3301008031a80,0x0,0x90102a0008600021,0xb110ac2031413a80,0x926002a8c222000,0x41c1e01200c61a10,0x4010034100425820,0x0,0xdc40324208521001,0x3080200100400280,0x8fc6aa2229d27ea1,0xab9f7d414e733a0b,0x22321cfa0e813e01,0x0,0xa214253008701a03,0x700a0040021011,0x480a0000000020b,0x43a4d5b22f127f01,0xbe569ff848779a0b,0x5bfaff726bf33f93,0x8b95e73d8a43b281,0x81ca1a9146562209,0x180c400060020a,0xf8d456fa0dd36e30,0x9ffabd204d12ff9b,0xa9a0af2f05523a39,0xc9ca90518dd23aa9,0xa99adfdb38621f91,0x94bddee368537eab,0xcdfeef6a4f437f2b,0xcb9e3ffcd8437f3b,0x4a46eb1259f35a81,0xef7ede5adf527a1b,0xbcff7fef5fd33bbb,0x497cfe7a0c533eab,0xcebe77f39ff3fe8b,0xcddfff3fca77fb8b,0x5e5e7f5b6bd37f0b,0x9a284c1809010f01,0x11248e0000509201,0x4a80256028549281,0x4040e00100021600,0xe7ed0e5298139391,0x613445008421b00,0xf46e32604a461aa9,0x65a27d600d82fa23,0x0,0x8c0b00080031800,0x2cb283d30f62bf01,0x44000021000,0x108542201521800,0xfb00430f8313ba21,0xc2062908224a13,0x23440a4008021901,0x54a2bec60bc37a9b,0x96ef7f9c9f21f9b,0xcbdfff64eff27e81,0xc8d2eb600d23ff81,0xedfffed12c73fe83,0xbc3f77e95e73fa31,0xf3edf6f44af31a23,0xfabe9f75ce763bbb,0xebf3ffeefdf33ab3,0x8adfdec70dc3babb,0xdfbfcffb5f573a9b,0xf9feb77d4f773ba1,0xff567f6549737e93,0xeb5dff2adbd2fa81,0xa9551bd02a733b0b,0x6bbc6777c8b71b1b,0x9c9030410912190a,0xc1283f40ac836a88,0x920025448002100a,0x0,0x212c4ca101533a22,0x6184444008701800,0x1200014001021800,0xc000c11408001e01,0x1870048000681,0x201005008600001,0x8002000101000,0x4800000008000100,0x8404006801710a01,0x9c80018801a00,0xa551d5420a521681,0x0,0x620023200102081a,0xd7ae3c7548f71e8b,0x5002202208401b00,0x38cc8031a561211,0x95325ef43cf77fbb,0xd7fff7ffdbf37fbb,0xfaee7d72dcd27e8b },
	{ 0xcbfef3bfef7f79de,0xaece1e39e37630d6,0x9e7bdce7a6749fca,0xf47f76682e7411de,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xefbceffba77fbdde,0xf5ff7f7be77efdce,0x0,0xe888d52a207e11ca,0xcffffebfe67eddde,0xa67edfdf627659ce,0x7efdfeeee77f7dde,0xbdefff72a67f1fde,0xdf927f7b2e7731da,0xffffffffef7fffde,0xefffffdbef7fbfde,0x3ebe7fff6b761dde,0xdd3dfffaee761cde,0xeab03d536976d9de,0xebf7f7bfe17ebdda,0xabffbfffe37f15de,0x33aac7fde37f94ce,0xd6edff72e27515ce,0xfffe3df4e77e3dde,0xd9cfffbfe77f5dde,0xefffbffb667f97de,0xf7bbcffbea7f9ede,0xf57dfe6be16f99ce,0xd6be3de2eb7e5c5e,0xcbece76eaa7f5ddc,0xad5bbfffaa773dde,0xffbedf7fe77e5dde,0xd4befef7677e1dde,0xeedfff76e67f94de,0xfb8dffeee7765e56,0xc7ffffbbee777fde,0xefffffdfe27f3fd6,0xfcfffefeef7fffde,0xdffff77fef7f37de,0xffff7f7fef7ffdde,0xffdfffbeee7fddde,0xfbffdfdfe777ffde,0xb6ffdfffef7f5fde,0xfffdff7eef7fbdde,0xffbfbfffe7777dde,0xefffffffe77fddde,0xebfffffbe37ffdde,0xfdfaddffe377b7ca,0xfbedff7f677f5dde,0xffde7f7fe77fdfde,0xcbfff7ffae7ffbde,0xefd63e1b6b7e1dda,0x1fbeefebe35e5dce,0xfbfffefdef7f97de,0xfb3e7de2a67f9fde,0xeba7cf3e2b7f35ca,0xfbda9653eb7e21ca,0x8eb7cfe26f765fde,0xd4d1f6e1e77733de,0xdbbbbe73a276b1d2,0xfbb4ff7ee87ffdde,0xcbf7fe1d6777ddd2,0x2ffedfdba27e91da,0xfcf8d5f7ea7419ce,0xd9e23f60a26e004a,0xbf940b61207299d8,0x6fc5ee82257190d0,0x2bfe59c3227d14ce,0x24e7666ea2753cd0,0xddfcfe8a62761d56,0xffbe3f5fef7fb5da,0xcfeeff7eaf77ff5e,0xabffea7b6e7695de,0x97fb7fff677c5fde,0xfdfffffe6f7efcde,0x9f7e2fefe7765dca,0xa9bcdf7a67761ade,0xf1de2e2baf7d174a,0xadf4dfff6e763944,0xd73fee6b6a7695d6,0x5b8efc43a276b5da,0xebf4ff2fe77e35de,0xeefbbf5fe77f35de,0xbfffcffeef761fd6,0x7bff556b637f91ca,0xcbffbd70e87695ce,0xd9bdffb7aa7700da,0x8dfffe25827719d6,0xaa5ee7dfa77c15d6,0xb831df61255a1754,0xfdaff760ee7715c8,0x8b6ebf72e276594e,0x8dcfee2e6f7715d0,0xae5fde59e27714ca,0xbdf25be3e27715da,0xf9fcff5bae7e17de,0xfbba8f5fe775b9ce,0xcbda7f3fa17c5cce,0x35f7fdeaf655cde,0x3fbfcffb6776dd4e,0x9775f56be37f5dc6,0xbf963df8e67e30de,0x6b56ef7ae36414d2,0x6fdb1fe925729f5e,0x1f5ccfe1a06e91de,0x10eae72e00e11c6,0x4bee6d62a1763cce,0xcbd2ebb6e37f17da,0xd85b920f4276354e,0xa3bc4fb3a23e95d2,0xb4fbf7e1a2775f5e,0xab8ca761e0640a5e,0xc580a0430074135a,0x940800020020006,0x3002810120221900,0xc0201461212e0040,0x21224fc108761146,0xcbbfff6eaf7f9fde,0xefd7f7bf6f76dd5e,0x829c5fc16276594c,0x8432d303205315c2,0x9d34b6eee27d5858,0x80c2fac627401c8,0x61c6421048460048,0x4400040040000,0x2020000000000048,0x1204010002600040,0x280000200800,0x38ce5a9623632cc,0x4032e28002010d0,0x28888c0204200ce,0x0,0x84000000000050,0xa0840003a0120400,0x40430a30e0153054,0x944c0000420086,0x404000000080040,0x20080000000240,0xf02480a008600140,0x8154a60e663606d0,0x42000a01411000,0x6668a5406c631d88,0x0,0x8080000000048,0xa004000100400048,0x4820000008,0x102122101000,0x142540020268040,0xca88656226740046,0x6294d6a462361cd6,0x2ffebbcd29773dc8,0x8000442024000040,0x8600000061400,0x782034482c461148,0xc9e3c29663565fc6,0xf4008002010205c,0x1ec3c909a07318c6,0x8435f66ba2761856,0x605a253061340842,0x8b80ff3ea377ddd8,0x5cc71ff9667719da,0x29344c01226600c0,0x140102060000082,0xb7b23d50271c7dca,0xd7867e6666379954,0xcccfffb12b7635de,0x8a4edf8fa167dcde,0x349d7febeb6759de,0x9d602610e0543496,0xc0922e1c666000d8,0xfcdfef1aa2765bce,0xabffdfdb637fb7de,0x8cb37fe3e776365e,0x71afdb6aea6f1dda,0xfef63f7f667f3fda,0xefd3fffbe77f9fde,0xaf7ffedf66761956,0x84162124e020014a,0xc03a0510ab605c42,0x754e25482764008e,0xcae1b28a625e014e,0xc0565e01686608d6,0x85b25a3222743848,0x545406f2206249ce,0x3aa5f226076a8de,0x62908200e22018c0,0x8d4a837123761442,0x29b68d8aa25e684a,0x1001704024101000,0x80a0c6123501540,0x8091501000541144,0x8033a22620c1548,0xa90e8940e4540006,0x44b23668227409c2,0xed6afffaab7699de,0xfbbf6f7fe37e5dde,0x88d6b62e2b678242,0x8ffd595ba77f10d6,0x3a8057fba07d11de,0xd46aff576775995e,0x189e1dece37710ce,0xeffeef62e7741fde,0x8ede7a1b2f59154e,0x3e0e1ff2e76815d6,0xd77db463a25717c6,0x6a14e933e86e02c8,0xa926030e23181450,0x6fdabfd22276bd42,0xf2f5defa27e91da,0x84d5d072616710ca,0x2040000001100,0x8a002546025417ce,0x241b68000341d4c,0x400c30000301142,0x0,0x4002000011000,0x8300400020620000,0x0,0x5d2158022515ac4,0x92900e80023415c2,0x286e0102241982,0x1200020200600800,0x4042420800080146,0x906001824020006,0x0,0x8418253000721844,0x2943812e14700c2,0x83c862224401ca,0x2000814022000008,0x8faa27e6666ec8ca,0xfdfefffbe77f5dda,0xffbc2feee6767dce },
	{ 0xd1d7f3ae5e3deedf,0xebfbbe519f3cdffc,0x3fb95dfb8f1dfff4,0x7596e6ee1b1dcbb3,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xffffcff18b3dffff,0xbefb5eff9f3dfffc,0x0,0xfa19fb3a4e5deffa,0xcfdffffb9f7defff,0xeff7df5adf3deffa,0xfcd7f7ff9f29dd59,0x7dee07689b3ccef6,0xdfbcff7fd71dddff,0xffffffbedb7dffff,0xabfeff5bcf7dffff,0xffbfffebcf7ddffd,0xfdf5b77ecf3dcedc,0xd2ae7d7f9f7dcdf6,0xeddeff1adb3dcfd8,0x8e57dfdbcb3def7d,0xffbb5ffb9f3cefdb,0xf57e77c99b3cfede,0xfbbe3f7edf7dcfd9,0x89fef33f9b3dfedd,0xefdb3e7f4b7ccff7,0xbebfdff7df38effa,0xd5f7f6ffcf7dffd8,0x76fe7d72df7ddfff,0xbbf7ef7fdf7ccdff,0xc9f3feb60f1defd9,0xfdfedf7ecb5deff9,0xfdfffeffdb7dfffe,0x9fbffff2db7cffff,0xfff67f7fdf7dffdc,0xefdfb6f78f1cdffd,0xffffffdfdf3dfffc,0xdefffefdce3dffff,0xfbeeff7adf7defff,0xffffffffdf7ddfff,0xfffffffbdf7defff,0xefffffdbdf7dffff,0xf7ffffffdf7dffdf,0xfffffffe9f7dffff,0xffde2fffdf1ddffa,0xe7f7fb2bdf7dcfbf,0xef7ffffbcf7dffff,0x9ebbfffb9f3ddfd7,0xffffffffdf7dfffb,0xfffe7fffdb5deffd,0x89feffaa9b3deffd,0xefdbfe79df3dff7d,0x9fdfdffa9f2cfdf9,0x95fcbffb8f0dffdc,0xdf2e7de38b3cffdd,0xeae7b72fce3cdff9,0xeefbbfdc9b1dfeff,0xc39e8f6a5f7ccf9a,0xd4f7bceadb1cdffc,0xaffef731df5d5ff4,0xcbf4bbffc91ceef9,0x8dd6ff99cb1cbfff,0xf9ea85dacf7dded3,0x36bc766a4714dd9e,0xd9eefe07c83ce8c7,0x9bfe2f7ddf3dfbdf,0xeee79e22cd1ccf76,0x28db9397431cde17,0xb12e1fe10a54de19,0xf13cf7591e3dc7dd,0x9eba2b4fcb7cefd9,0xe7f7f7ef4e7dfffd,0xbffffedfd21dddff,0xd5befff94f3ddfde,0xfffffffacb7cfeff,0xfff6fd6e9f1deedd,0xebdffbffdf3dffdf,0x4ee23b5d9a4cefd5,0xba331f7acf34cf35,0xd1e3f7d8de19cff8,0x8bc67d61df7ded7e,0xfbfeff3f8b7ddfde,0xcef23f7f1f1dfffc,0x6f9ecdd75b1deffe,0xf49c9efdce3dcefe,0x4bbebf73ce3cdfff,0xfba3f36f9f3dcfbf,0xebdff79f5f3deff9,0x6fffdfdbdb1cfefb,0xf4bedcf40b1defdd,0xddcfef72cb3cced8,0x82b6e372095d4f1c,0xfceefb9e1b3df7fd,0x7afedf5a1e6ddfdc,0x55576ffe8f2cddf6,0xfb1ce772cb38e7dc,0xfe962f6a8b2deffd,0xe977f76ecf7ccfff,0xac7fdea14f3ccfdb,0xb59a7ffbdb5ceedf,0xf1ffe778df3defdf,0xdff62f72571dca9d,0xa964da4e8a0dc7f8,0xe61317fb471dc1d4,0xdabec7f31b7dcfd9,0x141ff6ea53348d85,0x46324540820c0504,0xb1f2f3169f6cff1c,0x87dbfe3a8f3cdfdf,0xbfbaee718f3dccd1,0xe43747e4ca1cdefa,0x9beeff73cb3dcf7d,0x4284f7b4483d44e8,0x14630041a294010,0x20080552401c0084,0x3c503750cd1c4658,0xc7aebde0130cde5c,0xfbffffffdf7dffff,0xffffffffdf7dffff,0xee7ff5da9b3dfff9,0xb0dadd6b0b1c9ef2,0x59c6d542871cd7ba,0xdba63deb4f3cedbe,0xe943fe520b6c8dd8,0x42114008208440,0xf2be0ff24a3ca6df,0xd406011408640214,0xcc61560c11c0ab0,0x4066480000c00,0x20800080020,0x4f05060120c0450,0x0,0x208250008000040,0xa8806042080ca00a,0x230024010c801,0x20049080a0002c0,0x1004500000008418,0x8071280000000210,0x600000000210,0x80d2080081004c40,0x0,0x400002000100080,0x0,0x0,0x800000200008010,0x84000ce120d4020,0x90000a020c0010,0x7000203008002c80,0xb01427611314c094,0x8782f7c48b08cdd0,0x85f1d518f1c4f58,0x16906cd2831c8e79,0xd56fbf548a08c7fd,0x84296089204440,0xcb77f72a5a3de736,0x4008005108304a00,0xfebfefff5f7cffbc,0xc120d468001dc7d0,0x2ec3960810cdc19,0x5134000609186214,0xea53beef8b3ccfff,0xc21206610d00c618,0x14120a02080540,0xdaac00401b01d213,0xeb76672f4f3cdadc,0x89c7ff0b8f2cdfd9,0xaf7ecfda5f347f94,0x65b552e50b1c8efc,0xc1ae05601339ce18,0x3982656c021ccf93,0x74c3e233cf1dcc97,0x6bffffdecf3dffff,0x95ff5fc79a2ccefa,0xb9fef73f4f3cffdd,0xdf90bf5fda3cffdc,0xedffffbe9f3ddffd,0xeffffedb9f3dffbb,0xba074ff35f3cc7dc,0x7043332e860cd470,0x62263c620b3ca498,0x8080822282001418,0x4cc72ebcdb5ddebd,0x3086c5e5090cec92,0x94bf72709a5c6c88,0x58ba3c521b1c3ffc,0x214020e08008099,0xca4393511b35cc9c,0x80e4b90021087ca,0x440114408380280,0x280014008042290,0xc000330686044e98,0x44c27c120a383718,0xb1c428e0a04879e,0x14d654ef825c9e16,0xddbebff69f7d6f9c,0xfab5bf73cb3cdff6,0x7bd7ee32491de2d6,0x4fdfffdacf3dfffe,0xfef4fffbdf7ceff4,0xff7dfefb9b5dfeff,0xb2b62e64121c96f8,0x6dcee336853dceb1,0x2940b0550a0ce594,0x9600c508880884e1,0xd709af79935c0f80,0x300234200008cc38,0x88c1eb590b08845a,0x5fda3253d3144fb8,0x51804b2109097890,0x9232400008294050,0x0,0x3144802005144084,0x985b3a4b1a1c97d1,0xf99c1994a1d608c,0x6e90909c8d0,0x108080004004440,0x6800612401104c88,0x0,0x6490680100c68,0x34005640082c4c08,0x0,0x2300242500009010,0x2004000000008080,0xa0c0400000890d0,0x12a18840080ca486,0x880c74130d28c886,0x48e02e00020c489c,0x9019102030c4c24,0xa850e2518b008e51,0x10024fc0000cc8d0,0x97ffbf7f9a1ceff7,0xffbe3df19b3dcfff },
	{ 0x8bf7ef5f5dba9ff3,0xebffbfffddbadfff,0xdf8feffbfdba0f3e,0xd8f6f951582a8fe7,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x284800c88c285600,0xefbefffb5dbacfff,0xb4b356ee5daa9f6f,0x0,0xf8b1c93e6cba1d62,0xefffffbb7dbadfff,0xeffedbdfddbadffb,0x74b4aeec1c384702,0xd7feeffbb5bac7f7,0xcbbf6dfa7daa1dfb,0xffefffbffdbadfeb,0x7fde5fdb7dba17b3,0x7ebbfffb8db857eb,0x716eff7ae5b8c7a3,0xfff73547ddaa5dfb,0xefd7ffeefdba9ffb,0xaf9cbeddc5b847a3,0x3fba4ffa75ba57e7,0xddffbffdfdbadf7b,0xfefefd7ffdba9ff7,0xebdfe73fddbadfef,0xcbd61fdbcdba45ea,0xfb8dfd35dbadfab,0xf7f6bf7e7dbadf2b,0xfff66ff3edba1fa3,0xd376ff3ffdbadfeb,0xffffdfff5dba5fef,0xebbfcf5ffd3a5fe3,0xf6f7f7fffdba5fff,0xbbfefff5fdba0fe3,0xfffff7ff5dba5fff,0x5ddfffab5d9adfff,0xffffdf5f5dbacfff,0xf4ff7feefdba5f73,0x77befff37dba8fe3,0xffb67ffffdbadfb3,0xffffffbeddbadfff,0xffffdfdffdbadfeb,0x9fffffef7dba4ffb,0xffffff7bddbadfeb,0xffbfbffd75badffb,0xffffffbffdba5fef,0xeefeffebfdbacff3,0xfffa5ffaedaadfeb,0xfffffffffdbadfff,0xcffe7df6ddbacfef,0xebc7f7fed5ba9f7b,0x7fe77ff9fdbadffb,0xbbbeedf37dba8f27,0xd1ffb7fbadbadf27,0x4fbe7dc2959a8f23,0xbb9eff3ebd2a1fe3,0xeeebffd3d5ba1de2,0xbfbedfd3fd3ac7eb,0xfc7f77fbfdba9feb,0xaf3ab5f64d3a0fef,0xeb2acf7fcd381fe3,0x41cef6020daa4da3,0xf1acb1319b845a0,0x448bc26c149a0522,0xf7fff7fbddbadf73,0xde9c2f7f5dba07eb,0xe636d62a311a0762,0x649a1e4185b85523,0x6f25ae50d3a1da3,0xfb2707d8cd2a0def,0xd6bf2feefdba87eb,0xffddf7a7f8ba8faf,0xafdf7f4fdc3a1fe7,0xb6bffffaedba4feb,0xfffffffef5badfff,0xfffe7d77fdbadffb,0xfbfffffb7dba9fe3,0xefe77ed37dba1ff3,0x779b4f794dba1733,0xddfffffafdba4ffe,0x55ae7df165badd63,0xeb7cf3aae5ba1fe2,0xaec3bfdcbdba5feb,0xaebfefd77d3a5ff3,0xf632f6fa7dbacf2b,0xdfcffde3ecba5fff,0xe79cf76e0dbacbef,0xdfdfbf9ffdbadfaf,0xbfdedf9aedbadfa7,0x96f67fe8ddaa0fe7,0xcd7ffffbedaa5fe3,0xfff737757daa4f6a,0xdfe6ffbad5bad7a3,0xafffdfdedcba8fa3,0x84ff5f6be5ba5d6e,0xf9ff8f5afdbac7ab,0xdb7caf723d3a9fe3,0xddc9f7f00d3a9deb,0x2f5feb7b7c8adf33,0xb5bbfff254bad76f,0xf57caf98bd3a5f73,0xff567d5e9c9a47f3,0xa904d326e4ba0f43,0xeef3cd9402a096b,0x5b1dcd63752a4fab,0xd15ffe69112a4d4f,0xa63ae5508daa4542,0xfbc4fb5f1d3a4f23,0x47cbbe8f2daa0516,0xa3ccbf16db2c703,0xfdded3efcdbac763,0x9b201c4300288323,0xf1102d24180a0b63,0x810d20331028103,0x8886c810048202b3,0x10b44d40749a0102,0xc5feef77f8aa8feb,0xfbfffffffdbacfff,0xffffffbffdbadfff,0x88448483d8100520,0x16105ef104a21b4b,0xb0102a0401088303,0xa00242c000a4402,0x6044051a40020b03,0x0,0x91204470881a0562,0x8010200114800100,0x20480000440a,0x800202088000500,0x0,0x8008000000100,0x8106206a14921400,0x58bebd2244b85760,0x10408000100,0x0,0x2200000000100400,0x52c008000300,0x10001020000100,0x8000200040000020,0x0,0x800200000000201,0xa000000402,0x10005a000084c03,0x0,0x1044020004001402,0x6000000820000,0x20040004000100,0x1000054000000120,0x1a142712012a88e1,0xd902e609458a1b4b,0x6c5bfe3b0dba05eb,0x4a444600da80122,0x5b52345af9320737,0xff367f72f53a4de3,0x8c5c17c249815a0,0x47d70a03042819e3,0x91b2cbf435a005e0,0x1404a661452057e2,0x21b04e32053209e0,0x9bf4cf6fdd3a0df3,0xe2fbb3c395aac5c3,0x8a0a4720608201b6,0x550450a201a200ea,0xc096244184aa0d02,0x80813c84280500,0x8bf5a80378a88d2b,0xe72c3d7413a0f47,0x94b0e4e7552817c3,0xd9eff760ddbadff3,0xb2b137764cba4fbf,0xecc3fd2215a815a3,0xaf7e5dcefcaa9fe3,0x383c57e3c49a8f6b,0xd9ec7f6afdba9feb,0xfad77f7f7dbacfe3,0xd9f6db77753a8f2b,0xefdfdf543dba9f73,0xbffa4fffcdaadf67,0x91240d00401ac741,0x6a12546005381d0b,0xa0e461030c8a094b,0x8841a00100a0724,0x823c06612c089921,0x97b356640c1a4743,0x410a051050200740,0x9990eb6854ba0732,0xacde3aa189b88f2b,0x3b9e45f2493a8f13,0x14304ed451200132,0x3210503018000700,0xc989ad2c448a1b62,0x143b00204920c00,0x215d11e4028836b,0x3564146405a21f43,0x1a6def7b40ba0da0,0xf793be74799a87a7,0x8483c6082d2a0102,0x27fef751fdb245fe,0x54b5dee34db8c7a3,0x99effe7a1dba8f9b,0xd0120742140adf83,0x4e802331080ac9a1,0x6f2936d380502,0x3200249012a0022,0xc260a402a0903,0x839205400c8a0a21,0xd992460314b807a1,0x40160f40013a0d64,0x30b02448480a0106,0xf56e867815badc6b,0xcb82144060228107,0xa804802000080920,0x40e23942ad185502,0x2d82810005081223,0x240002600c00010a,0x82004400000040a,0x4190807001288181,0x8840200004000140,0x20000500800010a,0x42444e024821981,0x4082004000020052,0xda166f5c4d3a8f77,0xc9d4d22c402803a3,0x261d48d00c3851e0,0x94857ec455ba15b3,0x91502f604d908721,0x1ca4356504b80703,0xe094b314588a07e1,0x255e265cac1a0d6b,0x10a20f4009288303,0x5736976c25b8970b,0xfaf67df79d3acff3 },
	{ 0xebfbf7bebc9afcee,0x3ff77fcbbcb7fdee,0x844c00001260a4,0x974795ceb492eaec,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe8763907b496f2ec,0xffffcffbbcf3ffee,0x94da4eec189a5fc6,0x9330a8103432f06c,0xc9b6fbfa34f3eeee,0xefdefbbbbc96ffe6,0xef7f5fdb94bbfee6,0x3cfddfe6bc9fffee,0xf547bf723c9afde6,0xffbe3f6eb4f2fce4,0x5fd7febe349efeee,0xebde5fd93cdb7ce6,0xacba7dee94da6eee,0xf16bff72b496feee,0xfb8eaf6e34ff6eee,0xedfef6bbbcd378ec,0xfedffff734b3ffee,0xbf32effbbcf3ffe4,0xd8fffffd3cf3feee,0xffbe7dfdb4b6feee,0xebf6ff7a3cffefee,0xffffbffebcf3ffe6,0xd9bdefffb4976eee,0xffbefefdbcffeeee,0xbb7677e834335eac,0xdbd5efbf3c9bf6ee,0xefe7fffbbcfb7fee,0x4f9ecfdfbcfffbe6,0xf6ff77eebc9fefe6,0xffdabf73b49afee6,0xffffffffbcffffee,0xcfffffbfbcbffeee,0xfc5edfdba4b76eee,0xfcf6fefa3cdfefec,0xffffff7bbcb7ffe6,0xffbfffffbcdfffee,0xffffffbebcfffeee,0xfbfeffdfbcbbffee,0xbeffffffb4fbffee,0xffffff7fb4bbffee,0xfbff3dff3cffffee,0xeffff7bfb4fb7aee,0x6bffffffb4ffffee,0xffbbfffbb4ffefe6,0xffffffebb4dbefe6,0x7fae7ff3bcdb7eee,0xd9f7ffaf3c31efe6,0xefffbbfab4feeee2,0x8f9edffbbcb7efee,0xd132fef930b6eec2,0xffbe2f73b49ddfe6,0xdbd7efee909b6fee,0xffef7fbbbcb3fee6,0x4f7cdff13c335fe4,0xfdfdbfffb49fffee,0xb7887770149b76e0,0x4a896371149ce82a,0xcdd6fe8a3c136a66,0xaeddddca34faebee,0xa57fdec0b4d574e0,0xfbeef77a14def7ee,0xbbb63ffb94fbedee,0xa352d32aa09ae2e4,0xeb7fdfdab432ffec,0x99827dc3a41262ec,0xff77a7f89493dfee,0xf91e3f7bb4bb6eec,0xcfd3ff30b4f236ec,0x6f7fff4334b3ebee,0xb29ecfb3349b6eea,0xf7fdff7fb4f37eee,0xde9ebd66bcb7ffe4,0x69abf76bb0defeea,0x25de975bb4334cee,0xffba6ffb3cf7e7ee,0xbc5fefe9bc90dee6,0x7dfeff63bcb3ffee,0xbbddd37eb4d3feee,0xeaebbf8fac9bfeea,0xbf9fdfff34b7ffec,0xf5f776fd3cbbfeee,0xbdfbfd72b49cf2ec,0xff9cfbef3cb3feee,0xcfcffeb634f2feee,0xaff7cddb3c9bfeee,0xd4feffefb4b77fee,0xc4bffd70b49becec,0xff8e7f5d34bafaee,0xcddb7fa9349dfaaa,0xe3f5f4934326a66,0xfdbef7fbb4bfffee,0xd8fcaf1234137dec,0xca88272f343264ae,0xcff4db3e349afee4,0xaa8fdfd6bcdbecee,0xb71b7fff34b3ece6,0xf1fc9eda34befce4,0xff8e3d6ebcbaf6e6,0xfbf6ff2f3cf3fee6,0x2ef6bed93493eeec,0xdfdbcfffb413ecee,0xf4ffbe2404bb76ee,0xf53a5f5914136eae,0x83d8c17610d35aee,0xcee79e7734b3f26e,0x2b166914b411e2e8,0x3067877eb0d3e6e4,0x8d0a7460801270e0,0x2b1a67fdb43272ec,0x40003611300048c2,0xa218c200040263e0,0xfef756fbb4b3f7ee,0x49a69d60349b26c6,0x8bb66b273cf0e6e4,0xe5f7ff3ebcbaffe6,0x821e5c1734916de0,0x2a800906000,0x90340cc020107140,0xc600204084100084,0x4948d610841303e4,0xa101090200000c6,0x80040020000040,0x140c852c94106a84,0x1004040000123000,0x0,0x80480020808040,0x8008807030000820,0x447cae401490c0e4,0x5bbc6d2234dcece6,0x2b8c600204325ae4,0x480100114002004,0x0,0x4000d04820c2c0c4,0x0,0x0,0x40004a3000828280,0x140010014000880,0x400200000900000,0x8406040910910482,0x8100921000127800,0x2211004001060,0x31c4e0014c86060,0x8020502114110042,0x8008224210101802,0x44000a0000880060,0xc650400d00927ae0,0x0,0x24e0a240009194c2,0x5850936cb480c2e6,0xc290a60fb4123466,0xdb8eff6f14fee6ee,0xefff9eff3cbbffee,0xa639640200db64c2,0x420000010000002,0x4cb655d130936fec,0xdb5c9b3e34b1e2e4,0x448a13103499c8e4,0x20504440c204004,0x810000014000000,0x1a0ca80418946800,0xc2831a020492e006,0x1efe1924b66ae4,0x6bdf5b17b4b776ec,0x5420d6e194b7eae2,0x93ef6f7994d37ee4,0xffb7afffbc937eee,0x87cb1a049c56a8,0xef7ddf5b3cf37fee,0xbe7f3ffb3cbff6e6,0xf15ef64234ba6ee4,0xdbb20f6298dcfee0,0x7f0e7631493eee4,0xafffdfc93cff76e6,0xffee5dfbbcbbfee6,0x50f0970034ab4860,0x281e30220084e4e8,0x9878162308462e2,0xae8e32cb94b236e4,0x3eb7cd6314327ce8,0x348e34a8148a6cc6,0x20383d60242202e4,0x100803000808084,0x46031e2800b27884,0x20af8b613400feec,0x684066110b2ec40,0x80040420000042a8,0xd18020010091e028,0xc45ff88d24986ae6,0xf29ed5d3b49feee4,0xbcb292ecb4b2ffe2,0x37820f70b8b2e460,0xf6a62e77b4b67a6e,0xc8dfb633b49f7aee,0x4f7edcdb34937ee6,0x75b3ffeb34d8feee,0xd1febfe9bcf2efea,0x8022478809062a0,0x8a4120210006cc0,0xc402a0a008662c6,0x80031000901000,0xc5448e6b3480e5e4,0x82002c4010114440,0xc3e2f36a349372ee,0x480a320508106860,0x18aac4104307cc0,0xc110146814886c00,0x4140010000400,0x100800000100020,0x840140010000142,0x5394d40b41266e0,0x410202e01ce04400,0x5020004010104800,0x610010c00400,0x840c00014000800,0x800010010010000,0x10042060000220c4,0x2000002000020004,0x180000500104084,0x6866000a4223020,0x8f384d50001368e6,0xd0ba5fc134f06ee4,0xc888a7203480e4e0,0xab042730b491d8e8,0x2183c0010093a240,0x39cfaff9b4ddcbe0,0x18244f7084ba42e6,0xd1901472141125c4,0xdb9667ff3cb3eec6 },
	{ 0xc9effbbf721c77fa,0xeecabad37214dfe0,0x33095c90701014b8,0x95d9f7f2b014d6f1,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x244c02e00142b42,0xeffffefbf034feff,0x9fbfcfdff2fdd7fe,0x1890a48c50b49ec8,0x91188f70105cd498,0xebb3ff3f929cfffb,0xcfd7f7effa75b76e,0xbb17dd579a34d6ce,0xfc3f7efdf2d4fefa,0xdd43f7cbd01df6f8,0xd6a13fe6f01dbeee,0xcdd7ff167055f6e1,0xaf9fdddff0b0f7f8,0xb4875feeb0143ce2,0xf768f764585c54da,0xfe9f27fef2d4d6fa,0x795ed73d90b4bfcb,0xaaff3f5dfafcff76,0xfe9befbbd274fefa,0xd5fdffdbd2bdf7f2,0xfffe7f7ef0d5fffa,0xcbfffbefda7ddffb,0xfe56ff5eb83df9ea,0xb73fffe7983cf7da,0xd577f6fbd2fdffeb,0xdbf23f0260b45e6f,0xdbf4e73e7074fef5,0xa7f3bffbf2fcf7f8,0xefffefd3d29cbffe,0xfc77f7fdf2fdb6fa,0xbdfcff77f255fffa,0xffb7effff2fcffff,0xcfdfffdffabdfefa,0xeedfef18da7cfeca,0xf4fffef6d0fdfffe,0xffffff7bfadddffb,0xffbe7ffff8f57ffa,0xffdffffffadcfffe,0xffffffdffafdfeff,0xbfff7ffffa7cfffe,0xfbfffffff8ddffff,0xffffbffff8fcfffb,0xefffff7ef8bcfffb,0xfffeffdffaf5fefb,0xfffbfffbfafddfff,0xfd7efffffafdfeeb,0xffde7773fadd7ffe,0xcbfff33ff01d7ff3,0x7febfadffa3d7efe,0xffdfcffbf2f5dffe,0xb4b67a79f014d6fa,0x7ffeffe2b8fddffa,0xf998fb17ba65b66a,0x5ecb1e5f50d597eb,0xaebfdfd6d874fffc,0xf5fff7ff7a9ddffa,0x9bfae74750b4feeb,0x81201364501c1482,0xcce3fa29506cd6c8,0x5f9ecfdbf87cfff3,0xf4f54779f03cf6f1,0xfbffaf7b521cd7fa,0xfbb7af76b0357feb,0xefe6ea1e50dc1fea,0x6f5fdfdbf0fcdfff,0x957bdde3123cd2ca,0xdf48af66281c5eef,0xd7be3fe498f556f8,0xe1c6df3e501c97ce,0xe8ff7e595075df72,0x979ffff7185dbef3,0xf9fcf67ef0dc97e6,0xf7febf72d0bc77ff,0xfba8f7aa00bcf6cb,0x6fdf3adff014d6fa,0xfabe6dfbf2d4fffa,0x9cbe766f5abcd6e8,0xdbbcbd63f8f45efa,0xeb95e73e3ab05f68,0xdbdbbf7ffaddffff,0xfffd5fddd8fdbcff,0x51d63ffad294f7c3,0xd7eaef7072b4ffe2,0xfefe7f7eda5cfecc,0xdcd2fe8fb25d97ef,0x9eb7cf5af09df6f8,0xf6f5fef97254f6f6,0xffffff6afa94feea,0xfa947efed815f7ec,0xc9d6f3faf06cf7ea,0xaffeffdb701df6eb,0xd6bf7ffbd0f4ffeb,0xd97faf6cf09cf7fa,0x7e3a2d70708cd6eb,0xcf7fff7ff27dd6ef,0xabd6ff5fd0d4fe6d,0xfefbfffff0fcfff9,0xd73eff7ff2f556fa,0xe73e3d76f8d4bfea,0xebf7ff1ef2f4fdfb,0x7fef5fd1f8fcd67a,0x9bbacffb82f5fe6a,0xdd7eb23a7075f7ee,0x65ce75625055b6ea,0x851e7300005b5ea,0xaec63edb701597e2,0xffb1cdf3f81c7f7a,0x30b40e42d034d4a2,0x85b827409258d440,0x5306f3209031cce8,0x8141901098041480,0xc08a48001020,0xb29286ecb2a0d560,0x5388a060f0103462,0x8bac0a443050d6e8,0xed16e7476a10d77a,0x6f085f1170151aa8,0x30202de0500092c8,0x5204000080001140,0xf4961efa70153ee8,0xc0d09201288c1168,0x69041e80501068c0,0x90050010801080,0x8000100010000080,0x0,0x300001200004240,0x2a04915050180481,0x4c101a040055468,0xc01070603020d4a0,0x4002000000240,0x0,0x400300000008440,0x120405010140240,0x0,0x0,0x420082261214a040,0x4100840000000480,0x0,0x0,0x0,0x20000210000080,0x1080000004000,0x8488000040406,0x150000240042,0x712ca2482010dee3,0x32000e0430101628,0xe10048e0120086c0,0x110000100800,0x11024010001082,0x4842a500501022c0,0x1a14046020d45020,0xd9d6e73fb0bc73fa,0xedd7de035a3c76fa,0x3698df42b0d4d4f0,0xd510c86230405662,0x5bae47e23090def8,0xc9dcf316f2b45ef8,0xde4a3b5f7a1d95e3,0x11c040009016e8,0x4f15660509499e2,0x2210100000240,0xc900550c20941d68,0xe74db6011201f768,0xab4a07ca301c47e8,0x50800e6c001416d2,0x917d266ad098f662,0xe2beaf96f095d6eb,0xc9d6af6350b8fee5,0xebfbdfdff2fcffeb,0xf5bffffffafc5efc,0xb7fca6f2129cdeca,0x3ac635a270103ef0,0xe9c6eb33f0a0f4ea,0xa87fdfc8906456c6,0x3fbfdffb32ddd6fe,0x53a817fad0b116e0,0x83027001000424c6,0xa1fdf32ef88d36ea,0xbfd7ff9fd01cd6ee,0x4bbe8df73274d7ee,0xb427b221523054d0,0xc0022552f2501068,0xdbc5f324d05c168c,0x8863b3b93074d6e7,0xef3fddf5d2d0fee8,0x90900e48088c1242,0xa88254230511432,0x83301b1312e0d142,0x5dd7be85f034f7eb,0xa2eaca9080145c8e,0x5cf37fee1294bfda,0xddbaa75a5894f648,0xcb9c267e201c3768,0x88e2f2c3f255b6e8,0xebfcdddff05dd6fe,0x84bb6eff123c5fdc,0xd189e752b000d5e2,0x80860340103412ca,0x4080000000040002,0x8060600000c1248,0x1620044010000480,0x13488768d024f5e8,0x1466861d020f440,0x4381ebbb9098b54c,0x23473f4bd834bdc6,0x18e59301034f452,0x6000002000110030,0x4024300050040082,0x91a1c07040c190e0,0x2082180050101208,0x90c600000941202,0x1d100204000102a,0x602400b0000020,0xc00000440,0x8423f1200141c8a,0x4400000000400,0x680104150180481,0x40400000001008,0x800822000040480,0x85888401509810a0,0x800200001400,0xc000466012009440,0x20026010001082,0xcc942870001496d8,0xc1b0f96a1058d7c0,0x96ed3c6504d15e6,0x93304fd3325cd0e2,0x140d448805416c6,0xf7962ff5b0dcb7f8 },
	{ 0xdbaff77fd9efb35a,0xccdffe97d9f5b76a,0xbdde6f24d96dab5c,0xdd74b270d9fc2304,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe9d4ef3ef9ffae6e,0xffdfffeff9edbbfa,0xbfbfcffbf9ffbbfe,0xbef7fefff9fdaf5e,0xdbeccf22f9ed2b4e,0xe804496a516c0372,0xefd7febbd9ffbfda,0xbff5fdfd9fe235e,0xf7ff7ffff9fdaffe,0xd967ef74d974ab4a,0x4fbb7fffd9e42bd6,0xefdffebbf1ed275e,0x2ddcdf58d9e92a0a,0xb4bf1deff1edbfea,0xd0d6afebd9dc267e,0xfab72cbad9e8b3f8,0xeff7ff3ef1efb3de,0xeffeffdaf9fdab7e,0xbfbfe7fbf1fdb3fa,0xd1f9f7fbf1ed3bda,0xfbbefff5f9cdaafa,0xebcef3efe1ff227a,0xefcf3f7bf1edafde,0x95dfdfeff9efaffa,0xff7bfb7bd9ed3bdc,0xdfbe3d72d15faef2,0xdbffefbff1fdbfee,0xfcfbbfdbd9ef2fde,0xffbffffff9fdbb7e,0xfdf6fffff9efa3fc,0xf7f6fd72f9fdaf7e,0xebfbf7bff9efbfde,0xefcffebbf9ffa7fa,0xeb9fcfdbd1fda37e,0xfcfff6fff9fdbffe,0xffffffffd9fdaffe,0xffb7fffff9ffbb7a,0xfddfffbef9ffabfe,0xefdfffdff9fdbffe,0xffbfffefd9fdbffe,0xf9fffff7d1fdbbfe,0xffbe3ffff9fdbfde,0xfffff7fff9efaffe,0xeffffffff9ffbbfe,0xfffffffbf9edbffe,0xfdf7ff7bf1eeabfe,0x7ff47ff4f9dfaa7a,0xcbdeff7ff9efa3ec,0xefff5fffd9fdaf5e,0xf79fdeeff1bdaa6e,0xf57ef8fcf964bbde,0xfffe6df3d9edabfe,0xdbfeff7ed8fcbf58,0xa8f35aa9c1c42bf8,0x6ffccfebf9edab5e,0xf5f5ffeed9fdabdc,0xdfaefff3f9efabfa,0x79867f09d1fda05a,0xd7e3b1bd0ef23c2,0xeffedfdfd9efbffe,0x9efddf61f9efb35e,0xf5ef5f70f9f4a64a,0xfbde73bdd18dbbce,0xeddfffb3b9ed2f5a,0xefdfffdff9ff3f5e,0x74d73fefd9ecbfce,0xc9edf77af1e5a7f8,0xcaa6af70b9eca34a,0xedebf3f9d1ecaa7e,0x79db9e81c169a5e6,0xb7bbefe2f9cca7de,0xddfffffbd9ef3b6e,0xdffe7d67d1cf2ad6,0xebd7fb6bd98b3966,0xefdf3f53d1ffa36e,0xdb9c6fbef16d2b52,0xffb7f27cd1f4baee,0x6ffead72d9fd2272,0xd39cff67d177a36a,0xefefbffbf97fabfe,0xf7be4f73d9dfbffe,0x75eefffbd9efabfe,0x9ffe5f72f1e4b25a,0xdbb6ff3ff9fcab1e,0xdfd6febbe1fc2776,0xe35fdf5bf9ed3b4a,0xf5fff7faf9efaf7e,0xdfefff7bf9dda6da,0xfe963d7ed1cfb7d2,0xcd5ffe59d1eead5a,0xaa025d5bf1e5af6a,0xfcfd7fefe9efaf7e,0xd97be7cad1e5323e,0x91263de0e0cc3d54,0xcfa9f72bd1dfbf6a,0xadfefed3d1eda2e2,0x36be7ffbf1f7abfa,0xfd7d7729d1e5334a,0xda9e7f61c9cd2ade,0xebddfba6d8fdaf4a,0x67e73b59f0dc2b18,0x1a28877bc9890ad8,0xb1827ee9d9442124,0x158e3f61e9a5b2f0,0x39c4ef2fe9cc82fa,0xbedbbbebf9afb2ea,0x20244ab1d18c3b40,0x943142bcd0cd3204,0x201e0500c0802006,0x7106641d90970868,0x84883a01c12c2352,0x0,0x3009496850c48348,0xdd3835c2d9e7a37c,0x7a9c2d7499cca148,0xda422f02d140b750,0xe88cc5880442a6a,0x8531d66188e40368,0xd570062c918c204a,0x90ac2d60c06d2060,0x80daffb0c9612742,0xcde3e50c8cda3d0,0x82048000402204,0x8048260a88c12644,0xa001c441480310,0x4be2f12ee9cfa14a,0xe8800204402600c0,0x80804010808040,0x80100408d1800000,0x12000010040,0x1000010000080200,0x0,0x489400a80650c40,0xbcf4deffd9f82c06,0x90002d6091900800,0xfae9322fe1df2b6a,0x0,0x2100001ec1012040,0x25b100e890900000,0x9000000040000010,0x4020005000640400,0x0,0x0,0x402306880002946,0xa210096000c40700,0x50920c21808d2040,0x1006238c10000e0,0xaf52b94ed9d4aa46,0x121400a81500108,0x20248240008c2208,0x3630205800cc231a,0x897ad768d9edb27a,0xa8c632dbd8643096,0x35061962c1e42214,0xb3f7f66579e5835a,0x6d2cfdf0d9f5a272,0xc1d1804ac1692344,0x7ffbb7d3f9e5abda,0x2d3e6f44c84d3948,0x10521caad1c03900,0xc2b80d2061452362,0x88243041910d824a,0xd6c2f631d9e9a3c0,0x10b24152c10e8240,0x74906eefd1e7bb6a,0x91884572d1703202,0x92947631d84f234a,0xeee8db1af9d52a78,0xffddffdff9ffbefa,0x9cbfffeff9efaffe,0xb1fde66ad9cd2b5a,0xd2867dc249e4b34c,0xcb2dde04d9f125e6,0x661f9a36e94823a2,0x6082e5b416d2258,0x1dc4744cd148b4a0,0x2140bc2061ee2342,0xf124612b6124a042,0x2941980041002042,0x6ea8583d9750b58,0xf0f596d0e0e0af50,0x91800f20e01c2080,0xba00490090058048,0x9129188612040,0xea3c4fe2c9fd26c0,0x3085504648448160,0x1064701c12c2008,0x81880521d9802000,0xcec99613c10caa56,0x6f1acf8bc9fd2342,0xf473f6ecf9cd2772,0x8508724ac164b1c0,0x82a46f0cd140a2ca,0x6682e82ae1c11e5a,0xebfeff52f1eda25a,0x1cb35fa9d1c53162,0x80100060c0002080,0xd024253240483306,0x8900122051022040,0x650821a88e02140,0x411420491880202,0x1764372af0acaa08,0x2f4e7f6df9ef234a,0x2894c20e51010252,0xa7dededef16da9ea,0x1800cffaf8ec224e,0x40108ec8c100a042,0x2c07401c0102212,0x8004142000400700,0x8801ca02c8242000,0x8004b50c0a42244,0x4260480b0900200,0x81882e00c94c2042,0x200180002000,0xc029020c0252000,0x44000002004,0x54000508c0510288,0x8200000040008,0x1000020000008040,0x802102080012004,0x154100051401100,0x10300142c1410002,0x9080000080800800,0x90844601c0808880,0xc5f07b12d9c320f2,0x60efa71c1e42d44,0x8a048880840940,0x121546080412250,0xdbde3f70d1e5ab7c },
	{ 0xcbfdfffedd7efdb2,0x66d73e7b9fbafff0,0x8a9eeff352c8bf26,0xd0b3fed4508af720,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xebffcf7fdffefff2,0xafffbfdbdffcf772,0xe7fbcf779aa8ff60,0xb4dffe67dffcff32,0xdffeed639abefdf4,0x4001c020cd0c2820,0x1bffbe1fdf2eedf4,0xefffdfdbdfbeff36,0xfedffffedfecfd76,0x856ef7fbdfaee7a2,0xff9c3f7e9ffebd32,0xdfeeff1bdb7cbf64,0x6bde97db1a3cefa0,0xf6ffdfef9fbcffe6,0xb1bda7729abce5e2,0xfefe3df75afcfd70,0xeffffffbdffef7f4,0xffffffffdffeffe6,0xffbfeffb9ffcffd4,0xff7fb7ffdffcff32,0xfaf66f6c9afeff00,0xe9d5f7b1dd78f744,0xcedf9f6f1dfaf532,0x3fbf4ffadffcf7b2,0xbff7f7efdf3efff6,0x49fe3f72daecfd30,0xdbdecf7f9fbeffe2,0xefdbffff9afcf752,0xbf9fcffbdffeff60,0xf4ff37e69becfd32,0xcfbefd73df7efdf2,0xfba9f73fdf7efdf2,0xd9cefe379ff8ad56,0xf77fdfdedffefd72,0xf4ff7feedffeffb6,0xdfffff7bdffefff6,0xffff7fffdffef7f6,0xeffffffbdffefd76,0xffffffdbdffcfdf4,0xffbfffffdffcffe4,0xffffff7edffeedb4,0x5eae2ff31bf8ffe0,0xeff7ffbddffeef76,0xefdeffdfdafefff6,0xbfbbeff3dfbcef52,0xd5e7bf6bdd7cff30,0xfe7e7ffbd9feff70,0xab87f7fbdaa8ffe6,0xffffbfffdffcfff6,0x9fb3cfd8ddec7536,0xf7fefefedbbef754,0xcfbe7ff29ffae520,0xdbd1ef6eda3cffb6,0xea7b97dbdbf8ef32,0xbfffdffbdffcfd36,0xfdfff7ffdffeffe6,0xdbbfbf73db30ffe4,0xd5a43eef1a2eff22,0x47c7b6bfcf08ff20,0xafffdfdfdbfefd32,0xf2fbeee09decf770,0xc16e7fe0da24ffd0,0xfbd7767ddfbcf7c0,0xeddffffe9f6cfd76,0x2f7fffd39feefdb2,0xb4b7d063db7cfc00,0xfffeaf7fdffef7e6,0xff9e2ffcdd3eb562,0x4f6ff73a5ffebf00,0xecde7e5b59ecf784,0xb4bedffadb78fff0,0xf7efffebdffcfff2,0xffa67de49ff8f7e0,0xd9dffaff5abcefe2,0x6d5b5e154ff8ef20,0x369fcffbd978ef12,0xd595f167d9e8f722,0x5ffa7d73dfacad36,0x8b16af261afce730,0xffffbf9fdff8ffb6,0xffffffffdffceff2,0xfcf7f6eddffeb7b2,0xbbee66639ba8fd24,0xffb0f7dfdff8fd32,0xc9d7fea7dbf8fdd2,0xbffedfde1afaff92,0xd4bfffffdaeeffd2,0xf1efef78dffee7f2,0xb3962f444a68ed20,0xcdc3eb9b98ecf510,0xe9ecd5d31b6cf530,0x9ebffffbdffeff70,0x9f4ae7625a7cf570,0x12a22d75982c6660,0xe9dffb1bd8fcaf32,0xa88f9fc29a2ce5f2,0x9edfefefdfbc7572,0x9d9df7fa18f8af70,0xfff67f77dfecff66,0xebdfeb5fdfecfde6,0x1cce3b791fbcfd62,0x16decfe38034e7e0,0xfdf4df7217b49570,0x58ae5df392bea774,0xe8c4830e10209560,0xa06ab4d94aa0a540,0x292c445843382f20,0x44b80ad010088d24,0x4220240015020400,0x4000820e00200440,0x0,0x838a881a08086702,0x5000008049084000,0x3a21208830d400,0xdb966f3ed548b710,0x6fdff32edfecff66,0x4ed85a1b97b4b5a4,0x24bb526a8f482102,0xfbff9feacff4fd40,0x8280204005000600,0x80c0a00108404000,0x9ea4100d00d100,0x24a35f604d206602,0x20800002000,0x80010005000000,0x2002080000400,0x448080000880400,0x1208400000,0x840020200804a000,0x8000040000402000,0x0,0x0,0x0,0x74d5b6ed9032ddd2,0x9300b4509088a590,0x4000001002000100,0x0,0x0,0xb0b8d46812205732,0xb90a37224d6cec22,0x5aa0065088ec6000,0xe0218000ca002402,0x2340c0000086100,0x1024004010001404,0x840000001080000,0xce02280048887000,0xc312084888a524,0x6b7e7ac3d80ce5f2,0x989a4db3db7cf702,0x120150a90284050,0xd1822860d82c0430,0xa104eb225f00b412,0x4e4b284a9800f540,0x643d7429824a0a2,0x9918b94181202f00,0x8bcef5425b3eb520,0xe9d0f6769f6ef1a0,0x58437e98d8c8fd10,0xfbecfd9d8f8f730,0x84f69aeacd70b560,0xdba484325d3c7502,0x10040d2008344400,0x1cda2a115bcbd70,0xabe88f5bdb3c6d60,0xa09454689d200764,0x1b6675799f8c2d30,0xce345feadaaab672,0xedfeef7adafc6566,0xab7ddf1fdf7cff76,0xf5bfddef9fe4ed30,0xd17c766cdfbcafc0,0xfa9e0b9a8868e724,0xe37cd7784f7ce522,0x650505a18a89c40,0x103844611080ad10,0xd0b0d758d8fce706,0xcb9e155c98aca5f0,0x8100649180002740,0x2a10b2c3112c4930,0xdccaef709fa0f570,0x34ac88405200a560,0x3ab024609dfc3720,0xb32242029800ef30,0x9841968398243452,0x7c8edb438a6ce500,0xd45552ead824b524,0x834a7c410f24fd04,0xca24d32ecac8ed84,0xc7e6fe3bddccf520,0xaf7f9f5b9b68cf56,0x94ffdfe4daacf7d2,0xd8882c4309308d42,0xdb806c661d740520,0x81e4c81602aca700,0xa362575a58a4ffb0,0x14a2fc4058a8bf60,0x1300080081002800,0x98262410302402,0x4482000005242000,0x9740502c9482520,0x36104f619a602400,0x13287f2812bc9602,0xff1e3576dd646570,0x8ccb004a202520,0x4046023991080522,0xa5b446618b84ad40,0x4d0330108008d02,0x4400244000a88500,0x2880402110303400,0x464826e912ace500,0x3b2e290090686404,0x4880000a080,0x0,0x9404014040002600,0xc7c1ba311d24fda4,0x210845ca00300606,0x2000400000404,0x20024808402400,0x9a82674ed02cb590,0x1402020890002400,0x0,0x90000000004100,0x100886a00804d524,0x0,0xc14208109022a400,0x4080002000100,0x600002810008810,0x1200050600884500,0x7a92adf49af23742 },
	{ 0xc1d6fb7bbfa65573,0x4c4f3b7197f64f63,0xfffddffb1ffbef70,0x4732ea630ac77d63,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfbfcff7eb7ef5f63,0xfffffbdf3fdf7f63,0xafff4f7a9fde6f62,0xf5f77febbfff7f63,0xcf3efff1bfdfff61,0x6882c02a36926f70,0xbbd7fe6f97f77f73,0xfbbedfdbb7ffff63,0xfeff7fffb7f6ff63,0xd9ffff7abfde5f73,0xa38261801264d60,0xfdc6f7fb37f77f61,0xe53575aa3f0e961,0xfebf7f6b3fbfff62,0xfff5bffc1fde7f61,0xdfb67fdebf9e7d61,0xefdfff3fbfdeef73,0x7dffffffbffe5f73,0xffbffff99bef7f63,0xfff4f7ff3fcf7f73,0x5fd6fd758f86ed70,0xebf5f73e86df4d60,0xe85f9bd72b9f0b60,0xdffbcffb87ff5f71,0xfdf7bfff9ffe7d63,0x6bfe7d6096aaef73,0xfbd4f7373fff7f63,0xbfdb7eef8fae7f63,0x7fff8ff3bfde6f63,0xfdf6ddeb3fdf7f71,0xffffff73b7eeff63,0xfb92ff7f9fdfdd63,0xaff7da730bdf5f62,0x7fffdf5fafbf3f60,0xbeff7fe7bfdf7f73,0xfbefff7b9fffff71,0xffbeff7fbffeff73,0xffffffff97ff7f63,0xefffdfdfbffe7f73,0xfdfffffbbfde7f63,0xfbffff55bfffff73,0xdfff2fff9fff7f73,0xefffff3fbffeff73,0xfffffeffb7f6ff63,0xb77f7fff3fbfdf63,0xfdff37e93fe76f62,0x70ae2fe111e65d63,0xc314f52c0bc74b62,0xefffba5f37f6ff63,0xdf9ecfffbffeff63,0xf5f6fceb8fff6d71,0xd2be7df01fa75d62,0xfba7cbb6839e6f61,0x8f7bba779f9e5f61,0xfffeffff9fdfff63,0xfdfffeffbfdfff63,0x8fae75731fd6ff62,0xdf14ffba1bf67d72,0xcffeffcf1f9f7f73,0xfb5edfdb0ffe5f61,0xf4d67eb011d35961,0xbb6de772bfd7ff61,0xe7b63f569797fb62,0xcdc2bf1c8a8f4d61,0xef7cdd5a078edf72,0x945e5bef8b068f60,0xf1ffbff33bfeff61,0xffbf3fff9fae7f71,0xc7adf7bb17be5f62,0x2b5afbde1fe65d73,0x959edffb3f9e5561,0xf3cfdf7f9fde4f73,0xfffcfdf43fee7d71,0xe1f6f71fbfaf5d71,0x4db7e1313a65d63,0xdd8fcfcb8fcfff62,0xf55cbce29fe6ef62,0x5bfe3ff00fdf5f63,0xfb93fb2d8f9e7f72,0xffdfbffbbfff7f62,0xfffeff7f8ffe7f63,0xf5fff7fe1efedf71,0xb1bbb7738fae7f60,0xff96ff6f9bf67d61,0xcfc7ff8b9fee7f72,0x6ffedfdfbeafdf63,0x76fcfeee9fd7df62,0x9d78df7aaa8bef63,0xdfd1bf3f13de7e73,0xe7e6ef3a97d6ed70,0xbfffff5b9bfeef71,0xf7ffffef9fdfff60,0xf18cf7fd07df5f61,0xff3e29528b965771,0xcfd7ff2b87d6dd72,0xbb7edfdf37d75f63,0x3feeeffb15c67f61,0xd7ecbe793bde6f62,0x782e7fe413de4f63,0x8bf4ff7e3fd6ff63,0xefe7a1c51ef84d61,0x30aa979b2b02ef61,0x8410780180160960,0xb7383d6807065541,0x854c20020404961,0x202221200004061,0x7bbec98617c27673,0xcd173eb4939a5960,0x83a2bd409a8ecc11,0x4100008182300140,0x8102a00001120340,0x20004604000900,0xc0501e8002122540,0x30c276000970920,0x88a42d2a03868b20,0xebc47606b7875863,0xa1e164183842920,0x8090560881021d40,0xfdffeffe37fe7f73,0x92a00d0200960d01,0x800c20020000060,0x810601000e4421,0x1322040801124120,0x4000001000010800,0x80010020000020,0x88818a0000084861,0x800000400000900,0xb90205128b86a400,0x91c0b62c02085473,0x40000404000,0x42001020000,0x200010800000,0x0,0x4876161001845822,0x280000010800000,0x100000100000900,0x0,0x1200000900,0x0,0x24000000020020,0xfa9a6f3e17df6d61,0xc7d6fabe3ec66d60,0x63f49fd812feef70,0x343c57fb07962d60,0x99f6ee6e96d67d62,0xcaba2d5204725d63,0xc924c30f0d9a0f60,0x82ae3ecb15865f70,0x360259c0978a4a61,0x15142e1887135940,0xbb040400061a0960,0x69d2dd9ab3c26762,0xea52125004921940,0xc97c588074e4921,0xb4a8ba7d1fc2df71,0x7f7e7f7393beff63,0xfbd6f36eb7ae7f62,0xeeea3a19a2da4f73,0x6ab8cf5b8e2e6d72,0x4503268829a2f60,0xc3bfaf709fd4fd71,0xc59c2ffe8e964f61,0x4fcbee1b2b8f3b60,0xb01d5603828641,0xf455e6a812ee6760,0x91aae7c396cefd60,0xfb1067203ed26f61,0xf5ce162824067961,0xa33e5f5e87cf4f61,0x7fff7fedbfdf6f73,0xf2eaa7608e865f61,0xceb265eb1fd67f60,0x8184f3ff0f071d60,0x8564a0402165d20,0x400200010000100,0x5c44f4548fc66740,0xc104c22021a0d40,0x8b00830a03842660,0xace50b6816864d60,0xbee00c200ac66141,0xd4f4f46b1fbeef60,0x4a3e0d4aa08b4f21,0x838c420706002d40,0xe0a220b80864c40,0x1e945108140a0d60,0x74beb6eb17d76f62,0x9710ef40178ecd60,0xfb967bebae56fd71,0xefeffebbbfdefd73,0x30abd9d21fed6d71,0xf5f7f6e9be5e7d71,0x9d5c8f78868e6f63,0xf6943effb6567d61,0xefdfef2c1feeef60,0x2f597fdb07d35f63,0xb4b25caf86caef60,0x516cbe0c178ecf61,0xea3e3c6e00a64962,0xcc646008a50e500,0x888e9eda1fc6d463,0xb0bb6ca011035f60,0x6480e010900,0x5c462d6a1e824b22,0x4380c31283c20c40,0x2050129703825562,0x72180380040a4061,0x418b32001921550,0x802004202084001,0x81010802042020,0x66026000b884040,0xb3c4d0806c26d40,0x8010364101a06c41,0x1218400001025400,0x8024214305025900,0x47ca3c1680c64f03,0x410000000002800,0x1484eae010814d70,0x8500240001021140,0xf797a7ee1bce5773,0x9846c2140742d140,0x8008000404000,0x120000000020040,0x8020044000064402,0x20004000004001,0x0,0x160064803022920,0x420660000820860,0xf120060004000d40,0x22187c140a5e6922 },
	{ 0xcb7af3b6d79c7d33,0x67ef1fd9f6bc3de0,0xffffeffbff9c7fe3,0xf77efefef7bc77f1,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40ae284890144380,0xbbfdef36f69c6cf8,0xe7cbdbfb779e7ffb,0x7b9ccfbbf79c6ff2,0x73ffffffffbc7fe8,0x753af563d7947fc8,0x5000c00003300420,0xdfdffebfffbe7feb,0x8fdfdfdbff307d2a,0x74ffffed7fbc7efb,0xf9fef77bffbc7fe0,0xffb4ff7edfbc7fe0,0xdfb7bb9e97b43fe0,0x7bffd7d9f5be7de0,0xd4be577e7e3c6c22,0xf9eefffedfbc7fe9,0xdfa73f71ff3c55e8,0xdbfffbbaf7be7ff0,0xfbfffffff7be7fe0,0x77b7dfeefebe7df8,0xd1e8bfeff5bc6ff8,0xeffeffffffbc7da0,0xebeef3eefebe7ffb,0xeeffffdaf7be7df0,0xbfbfefeaf6bc7fe1,0xfffffe7dffbc7ffa,0xdffe7dfbdfbc7fea,0xfbdff73fdebc6ff8,0x6fcbdfdfd7be7fe0,0xfefecffbdfbc7ef8,0xf5f7f7fdd7be7fea,0xfabefd72d7bc7fe8,0xffbffffeffbe3fe3,0xdfcffefff7b67fe3,0xfbfedfdbffbe7ffb,0xf6fffffff7be7feb,0xffffff7bffbe7dfb,0xffbefffeffbe7ffa,0xefffffbfffbe7fea,0xebffdfdfffbe7ffa,0xbfffdfffffbe6ff9,0xffffff7bdfbc7fea,0xffbf3fffffbe7ffb,0xeff7ffffffbe7ffb,0xffffffffffbc7feb,0xffbfeffbf7be7dea,0x9c6df76fd6947fe8,0xd2563c32d4982721,0x6b86eb269b1c5520,0x215e9749d4986f60,0xb5ffcffbffbc7cf0,0xfdffd6fff7bc7ffa,0xd3beffebffb67fca,0xc394e3bed7bc7d68,0xffeb5fdfffbc7feb,0xfffffffbffbe7ffb,0xf5ffffffffbe7ffb,0xdcfefd52d6b46ce2,0xebadf7fade9c6fe2,0xc9defe13f7bc6fc8,0xbbdfdddddfbe7779,0xf6b36feef7bc6fe9,0xbbea7771d69c6d6a,0xdf2467ecd79677c8,0x4ec5ff96f6946de3,0xab767e10d79c2c60,0x9c57feebbfb47760,0xe1ff3e72d2bc7ee1,0xd6b63d77df3c7661,0xeff7f782f7bc3deb,0xee9ffe9bd7bc6df0,0x8dbadff2dd1c3de8,0xf5eff7f7f7be7de9,0xdfb67fdfffb46fe2,0xebf7f73ff7be3fe0,0xcef67afef3b42dea,0x9737dff7d79c7d68,0xf843b7f1d4243f73,0xd5ee3f78f69c7f20,0xd3deff3bdfbc7fa8,0xeae7bfdef7b67df0,0xe69fdffbfebc7de1,0xf5f6feffff9c7fe1,0xdd9efdf1f7b67ff8,0xfb9eef7bf7962fe8,0xcfcfff9fdbbe7efb,0xab76ccded61c3de9,0xdafe7ed6f7343cea,0x996f7f71f4167fe9,0xfba435b4761474b3,0xf1f7dfdad7bc7d62,0xaffdff57ff3e3ff9,0xbdbffffbdfbe7fe1,0xffffff6bdfbc6fe1,0xf8fe77fbdebc7de9,0xceffff2fd63c65fb,0x6affffcdd7be35e3,0xf43b57c9d3bc55e8,0xd3edde34d09c7ef0,0xbb8e7f61d53c5fc2,0xcbf6f1fbd7943d69,0x61d33ec1f6842f68,0x19bad75ad61e3de8,0xf497b6eb93bc6d60,0x108aad4074220420,0x8b246165d21c6002,0xb9000e0554040460,0xeffdc5eaf63e4f60,0x3090802014001460,0x6a02c4284b42408,0xb8002f1c56107cc0,0x3b80008000468,0xa480140d4903c40,0x34a646a8941c1640,0x99af7360d49c3dc9,0x30842760908428c8,0xcfc2a71d93342de8,0x4be89d43d2146dc2,0x5ef1d289ab42d68,0x1000700240100420,0xd8242d61c6945580,0x4800c304020c1480,0x21bb64b57922c60,0x12104820c0044400,0x200608003c0400,0x74b23de4d6a421da,0x28b6890840168,0x8523080542401c0,0x80c80a02982400,0x0,0x4200240050002050,0x5000c02844843000,0x800000100100000,0x0,0x400a044851c0000,0xc800000002003400,0x81000080000100,0x0,0x0,0x1002000c0000,0x0,0x844070002882400,0x46208003141400,0x26999dd0d7160720,0x90507256040488,0x8100002040020408,0x0,0x40078e0250182c80,0x80011d080a42400,0x92284ce3de3c26c0,0xc575b67beebc2ca2,0x4116103041900458,0x2005c3e0403454c8,0x3570161093302841,0x8f988809d3043ca2,0x883e222002282c00,0xdfbe3d79d1bc7de8,0x93decb32d4b43ec1,0x7cde3e5fd7be3d73,0x6b1fdfd7ff3e7ce8,0x249e3a28d0346e40,0xcfbe7d71f1146df8,0xd78ddb79d6bc35e1,0x87c4a20ad5b82d20,0x94e05cb5e0e3ce0,0xbe75def9f73c3fe8,0xa1cbff68f7bc6de8,0x3024770dc103f09,0xc4a8d99042802c48,0x8bdeff57d7b66df0,0x24bfffebf73c77e2,0xf1e6aef8f5bc2f63,0xde9eaf7effbc7fe1,0xe9f6d71af69e77ca,0x1c011a4882942500,0x16317c42d68257c0,0x5045f5b9d53c36c0,0x82c67d6092987700,0x8845d260f02c1500,0x888180104302401,0x34f54ce4fa2c3da0,0xf5f6f66efebc3dc3,0xba7aad709e3c5728,0x820043104c800440,0xcf6ab61af21e6709,0x1884c518500c2c61,0x747376eae1bc7cd8,0xdd801a22838c3401,0xfaa26b7ecf363c82,0xefdfff73f7bc2feb,0x6a7ece53dfbc7df8,0x56d4def0f4bc6d48,0x596edff9d7be6d63,0xffd67e67d53c2fa0,0xefc7cf7576b43cf1,0xab4cdfd176bc65c1,0x8b725e6bd6bc7ea0,0xd158d7f2f4ba2fc0,0x9284020002140480,0x8584a001448c00c0,0x24a1012f4900040,0x412681486300440,0x92501600d4900062,0xfee2e548f33427c3,0xc1f7d37bd51e7de1,0x2044084812002800,0x18aa4d5bf69474c0,0x413042e844822840,0x504152091002c00,0x20410000020,0x40020008000c0400,0xc9c4e20d6903440,0x1104bc291b42440,0x8e82244016182500,0x4808208280144080,0xe864021c14b01440,0x98880490904100,0x40447ea0168c2460,0x9b40042050040462,0xa210230470902c10,0x440e00080340400,0x10400040000002,0x3808010113000500,0x6004100000,0x8057164100400,0x81c6054b92241560,0x6010904a502005c8,0x108650d0100810,0x140152894040860,0xb2407d4286b43441 },
	{ 0x41a3cea2ab88233b,0xecdf9edb8b9f2f7b,0xcefeefeb9bbc2b7f,0xdcfffe7f9b9d2f1f,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x76bc35f88bf52a5f,0xfbdef77e9bfb2fff,0xa4621ffacaa32317,0x3bbde7d99ba2232c,0xfcffffffdbfe2bbf,0xfffffdf3cbff2f7f,0x548422f90bbc0bb9,0xdeefffbfbbfe27ff,0xbbbfffdfbbfd2fbf,0xf7bfdfeabbff273f,0xbd7fff76bbfe2fff,0xdfb6ffffeb9c2bff,0xdeffbf9fcbff2fbf,0xefdffedbcbdf239f,0x952797f78bf023ae,0xf8ffff7aabdc27fb,0xdaaebf67fbfb2796,0xedffff779bdb2f7f,0xefffffdfebfe2f3f,0xbfb4fffb9bec2bff,0xf7fbf76dfbdc273e,0xe53a6d7dfbcd2f4e,0xcbeff77ffbfd277f,0xffff3f47fbdc275f,0x3fa9cff39ba72bf7,0xf77cff7dbbfe23ff,0xdffe7ffacbfe2f3f,0xf9f5ff778bda2337,0xffdfbf7bfbbf2fbf,0xffffcffbdbfa27bf,0xfff7ffefdbf82f3f,0xffbeff93dbd927ff,0xfbbceffffbab2f7f,0xeffffffffbda2bff,0xbbffdddedbfc277f,0xf4f7f6eebbff2f7f,0xfbffff7bbb9b2f7f,0xffbf7ffffbff2bdf,0xffffffbe9b9e2f77,0x6fffffdffbfe2ffe,0xdfbffffbdbba27ff,0xff7efffffbff2fbf,0xfffffffffbfe27ff,0xefffffffcbfe2f7f,0xffffdfdfbbff27ff,0x3ffbffdfabbf279e,0xd3cdfdff833c2b66,0xc5ca3b4093b223a7,0x89bcc2061bc027c5,0xe3523a6213182744,0x2dffd7fb9bb82327,0xf57ffefefbb627df,0xcf367df39bd0277f,0xab92f76e9bba232f,0xfbfbb7dfbbfd2fff,0xfffffffffbff2fff,0xfdffffffbbff2fff,0xdfbebff3fbba2f5f,0xfb5e7ffe8bff279b,0x5ffebf13bbe6272f,0x1acccddbdbca23bf,0x76ddf5e9cbc1271e,0xb94fef62b39023ce,0x7c3437f40bc2226d,0x8943efa98b70274e,0x2fdadfd99bdd2756,0xbdb77f62a990233f,0xff6eff79cbe82384,0xdea0af74dbac2717,0xc9d4f7baab9c23b7,0x3afefbdbbb9b2b1f,0xd7bb7ffbdbdc23ef,0xd9f5f7fcbbf42a1f,0xdbdebb74cbe2270f,0xc9d3fb6e8bba2f3f,0x2c923f088b58230a,0xd7facdd1cb9d275e,0xfdf5ff7f9bbe2e7e,0xdfbe7fe28b9f235f,0xabdcef689bf22f1e,0xfbf7ffffcbdc23af,0xff5f7f2b9bba27bf,0xddfff7eefbb423bf,0x9ffefd67ab73272f,0xcd9effefbbc9272f,0xcfcfff1fdbdc2f7f,0xafdeff5ebb8e23ff,0xf4f7deee8b9f273f,0x11c2a57ac3a8231e,0xff9e3f7d9bd8273f,0xeffbff188b9f277f,0xebf6dfdbbbfe2b7f,0xb6fbfffffbf623df,0xeb7fef76abfa2737,0xfebe2fffdbff2ffe,0xcbf7ff9eeb9b235f,0x6b5ffbfb8bd602df,0xb79fcfeaebba274b,0xd57ca4d08bd92f0b,0x4e963737abc4221e,0xe392fbbd8bce2faf,0xdd6f5ed6eb982776,0x9b9a6ffa8b9627ea,0xb716be42ab8022b6,0x200012082000204,0x8af1af7e9bbc223c,0x80426c5151922225,0x3ebd55e1cbd02746,0x34b7f6e883a02f7f,0xd9aa2c508ba82218,0x56ac20738b8822aa,0xb00880b83822402,0xa9d40c008bc00047,0xe4b65ca8ab1a020a,0xf96cee7b8bb42b5e,0x8b0cab3083162608,0x23c8e7328b4426c2,0x810cd409114264c,0xf216ea9b3e230d,0x4004001009800001,0x9c964d25bb84035e,0xc0802926a104232a,0x1800030002000004,0x508205a003360200,0xb6040200138302a6,0x0,0x2002000000104,0x60230009000030c,0x800804008080000a,0x1000200580800041,0xc4001009000000c,0x0,0x0,0x0,0x8060022002802208,0x1082c0182820041,0x8800631003802028,0x200200800002,0x0,0x0,0x0,0x80004000002001,0x2200000002001,0xa31ccdc4aa800345,0x100030a8400000,0x100040080000008,0xa0008000200,0x628882802020,0x1e04050200220c,0xbf395ffab2852f9e,0x80500082042600,0x20004001100000,0x8bf9da080b802fa2,0xee7f9bdbabee2bf3,0xcf9fcffd8bfe2ffa,0x9220204082802106,0xe66225608292221c,0xfbd0eb348b992fee,0xcffab6f39ab62715,0xe7be5ff7fbf0263e,0xd567f2ee8bd12abb,0xfffefd72ebd82e3f,0xd98837a68b92233e,0x2ae22b2800306,0x39f5d12c2a0270a,0x74f85a68cb9a2797,0xdbe677f2cb98232f,0xcbc4277389f8236d,0x96c3ae128b8023e6,0xa37f5edacb8c233f,0x94bb587b9b9a267f,0xd5a4a7739b182754,0xf7a62ff6dbb2271f,0x7d6ab33a3962331,0xaad7fedf8bd42f1f,0x241850460ac02503,0xf101d76e1b94220f,0x50438208310200c,0xa836422603902223,0x46823e1083802113,0x80800002800000,0xc104d06a8b8c035d,0x621a55728b8023dd,0x810020802810002,0x48519a100195020a,0xae29482803b02a44,0x9656035180222e,0x1220d600b182500,0xf1a0be1e0b96226f,0xcdc6be1dc3be221e,0x8abf17d3abd22795,0x9dbf7fea9bb6227f,0x93cb7f63cbd0232f,0x83346ce482a82b16,0xcffbffbefbfd2fff,0xbfdfffdf8bba236f,0x9d9efbfbcbcd2b46,0xf9bf72688b8827b7,0xc220045282800319,0x8947a61492942206,0x29c31c8a0bc0030f,0x2004d4a09080328,0x1044aec88a882220,0x648a7de19b89225e,0xe1bffaf78ab82f3e,0xec562e3389c82f73,0x511ccf138bd8272f,0x1082068481022206,0x4142010100a04,0x3084400400880290,0x8801a0013000006,0xc11c0f7082802385,0x90466888410a0a,0x8080244009802200,0x400024882002608,0x4c49b01203902228,0xc05448a400300,0x400000480000002,0x2944932083a0242f,0x800003440201000e,0x4000021000800a02,0x10054001800002,0x0,0x2482000280000c,0xc89805568b802329,0x4000440801002308,0x8452080009080008,0x40144600101034a,0xd1f657209392033f,0xb2166765cacd2306 },
	{ 0xc3f1f6a3d914d7ae,0xeef7bedfeb99f39e,0x7bb95d72cb94532e,0xd43d77636b0cc1ea,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfebe1ff9eb35dbbe,0xdb56ff3efb2dd1fa,0xffeafaf6fbace3ae,0xafbfdffffbbcf7fa,0xfdf7fefeebb5dbfe,0xffffffe3fb3cffee,0xfefe3f6febb5d72e,0xffffbefffbbdfbbe,0xeffedfdbdbb5fffe,0xfcfffeffebbcd3fe,0xed76ff6bebbdd7aa,0xef95bee4ebb4d33e,0xdfcfffbddbbdfffe,0xfbffffdbfbb4fffe,0xdcfe7dfaeb35d5be,0x99cdff49fb145faa,0xbab2afcef9b4f7fe,0xcfffffbffb3cd3ba,0xfdffbff7fbbcfbfe,0x1dbafffbfb94d7fe,0xdd7e64fbfb05d7be,0xf79effeed38df9fe,0xdbffff7fcbbdd7fe,0x6fcb3e7deab9f3ae,0xbebffffff835f7be,0xfdfdfeffebb9db7e,0x5fee7de0e325d37e,0xfbf4f7bfe9bdddfe,0xeecfbffbfba4f7ee,0xfbbeeffffbb5f7fe,0xfff6feeef3b4d72e,0xcfbebf72e335d7ea,0xf795fb7fd98df7ae,0x5ffbffbffb85d9fe,0xffffdfdffbbddffe,0xf6bffffaebbdf7fe,0xdfefbfffdbadd3ee,0xfbbeffddfbb5fbfa,0xaff7ffbbf3b5dfbe,0xabfedfdffbbdf3ee,0xbef7ffffebb5dd2e,0xfd7fe7ffebbdfffe,0xfebfbf7feb35d7fe,0xeffffffffbbdfffe,0xefff7feffbbdf7ba,0xb7be7ffbfb34576e,0xf13d7ef9d335d33a,0x7ff43c7183a4d28e,0xc954d3b7c315d7ae,0x88c2124dc11450aa,0xbebdffdbeb09dfbe,0xfff7bffff33cfb3e,0x6bfe7d7ad98c7d2a,0xb381ebb6eb84d33a,0xeef73ffff934d33e,0xffbfeffbfbbcdfbe,0xf5f7ffeffbbdf7be,0xbbbfadf0518cd32e,0xb6acffeee981f5fa,0xe7d7fe19f30dd3ae,0x3a96cb9feb24f7aa,0xf473f664dba4f33e,0xe5fdaf79e33dc3ee,0xfb99cf7adb24db2a,0xd9dfdb2cfa31d31e,0x2bd3f95beb04d18e,0xbcbf7feafb3c539e,0xd33e7fe0fb24e12a,0xdef57ff6cba5d7be,0xefc7f71ffb1dd3fe,0xe9ffdfdeeb24d1ba,0xdcbbfff2fb3c532e,0xf5edfffff9b4df3e,0xf6b67dffebb4d3be,0xebfefafffbadd72a,0xb5c21200d900d12a,0x9e9556ece891d7ba,0xfdf7f7fbfb35fbbe,0x7fbe3e73eb21d3be,0xfbdcfb7ef984dfee,0xafd7bffbeb39dfee,0x8bbe7fffeb98d39e,0x35b5ffebfb3cf33a,0xdbeeff71d334db3e,0xdff6fff6f335f7be,0xd9cffe13fbbdd33e,0x631edd1beb25d5aa,0xb4f65cedeb84f96a,0x592cf846831079f2,0xcbb26e6fdb00d19e,0xf6d2fe1ad114d1ba,0x6f7ddcdfe99cd72e,0xbdf75ff7f0b5f7be,0xfbbce7f1e935f37e,0xdea62fedfb05fb2e,0xedfff7bcfb34ffae,0xf9ff9fd1d3a4c7fe,0xbcbe4dfbfb94d7ae,0xc77cbf77f325f32c,0x8af6bdfce9b4c39e,0x83ba61026801f326,0xc5db3673e985e5b6,0xdc0def8a31447ae,0xd3208361eb20c13a,0x468803a0811481b2,0xd819eb2ad398c3a6,0x48c136845005c1a2,0x820a0010802c4120,0x35750e568830539e,0x87340f608a204206,0xd4204042c104d000,0x4880021482004086,0x2082010ed100d008,0x3454a880148b0e,0xd3feefa0fb84f18e,0xfb363f76fb84d18e,0x8188c6035a80c10a,0x4f52cb190b28d39a,0x301203214900e102,0x9108b645820c412e,0xf3106b3c8badd18a,0xa8263620b044206,0x610016d089044096,0x4ab4cee4384c18a,0x94450240e208410e,0x200000400006000,0x0,0x0,0x200081428080c006,0x24020004000,0x4000008008004000,0x102010440108200,0x0,0x420080000004,0x2840004000,0x80020040c0085004,0x100042001104200,0x0,0x0,0x4000200040002,0x0,0x84050000000,0x10100000004080,0x208050040004008,0x64004124000813a,0x0,0xa00000000008020,0x908430243045008,0x6000000000a,0x1100465080001102,0x604452026000052a,0x0,0xc00048005000,0x2d4793c0c820f7a4,0x1e9eccdaeb3453f6,0x446cf2504100c532,0x21c64e204304506a,0xebd4c766db19d3a6,0xaccfbedfcb84d3aa,0x26fedf737b15f3fa,0x3af637eeeb95d36a,0xd3b6b573e185f1b2,0xee922546d3b0f21a,0xc153fb224300c3ee,0xeeffd7deeba4f3ba,0x249c46686b24c20a,0xd1ac06685100e120,0x1ab4662ec8a8e9a2,0xd5eb5e3ceb30d1be,0x7b5ed3dac8b9d37a,0x1c826dc47300e20e,0x6020a731e101c20c,0xf2b62675e105d5ba,0xa1084205c104d32c,0x2f5f5249c314d06a,0x3120050201006160,0x91ded475c025933a,0x6e847463e3a4d1aa,0xe8e5d7aafb10dbe6,0x44636885301c000,0xb0804468a9044086,0x842622f18930c30a,0x40262c20d820412e,0xd394ef4e5335d386,0xab425a8373045726,0x9e01eb48e924e10a,0x414124684924c32a,0xda3844116110e220,0xfb147e3ac810777e,0x4fff3eb3db2cdf7a,0xeb7a9fcbfbb5f36a,0xf0395ee8c9a5d32e,0xd6e276ad325d7b6,0xde2e7b74c820533e,0xcddfffcdebbdfffe,0xecfed94bf9b0fb8e,0xb4f752c361b5d7a4,0x116ae568c994c366,0x363c3ffdc8a1d17a,0x8020d081888c4226,0x66feda5acb28490a,0x1104350890c5800,0x110e3502c2a1c020,0xa3800544f0218204,0xc9f4d2a2e924d18a,0x66953e2c5984c1a0,0xaa98ef528a30572e,0xa42320604a04802a,0x560c20724100c13a,0x91840326c900d326,0xd9e9b710e81dc31e,0x820e07005000011a,0x810000040008000,0xc0ba1c00d130c000,0x88240400010041a8,0xdd0b902cb04f206,0x30243008000536e,0x1018526cc110d106,0x20e4468d0040464,0x80050401006080,0x0,0x200000800004008,0x100744a68135c30c,0xd94c34e9db05912a,0x97a5096ae9a0c00a,0x8800020100040122,0x2c7e9ac85200d12a,0x37c347915304c18a,0xdd94f6035b90c3a8,0xa122f546905d38a },
	{ 0x8980d2a6085291a0,0x25e6df418b0fb360,0x2c304ff082d6b260,0xddffba5ecbd7b3e4,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfffe7ffbcbdff3fe,0xfb95f73e82dff7ea,0xef6bfffd8bdff37a,0xefffdff78bdff7ee,0x9ef7f7eb8bd6f77c,0xfbbeff758bdeffea,0xdbbf637f8bd7ff2e,0xdfe7febf8bdffeee,0xbbfedfdb8bdff3e2,0xf6b3ff6b8bd6fb7e,0xbdffff7f8bd7f3fa,0xdeb5ef7ecb5ff7ee,0xfff7fbbecbdffbee,0x2fdad7918b56fbe8,0xb4ef77fe8ad6bbaa,0xf5fdef728bd7e77a,0xdbbfb77e8bd6f7fa,0xeffff77fcadff7fe,0xefff7f6f8adef3ee,0xd69befff8ac7feba,0xf9ffbff7cadfffee,0xffbc7ffe8ad7fb6a,0x89e6ff3f89dffe4a,0xef55375b8253fea0,0x67fb4ff3ca5ef334,0xf5fff77e8bdff3fe,0xdfde7ff382d6fbea,0xfbbdffbf4bdef2ee,0xefdbbedbcacebfae,0xffbfdff689dff3ee,0xb5ff2ff282dfb76a,0xdfeaa561c2defade,0xfdffff6b8bdffffa,0xffefbe9bcbdff3fe,0xffffdfdf8bdffffe,0xfffff6efcbdffffe,0xffffff7fcbdeffee,0xffbffffecbdef7ee,0xffffffffcbdffffe,0x6ffeffdf8bdff3ee,0xbfffdffbcadffbee,0xfdfde7ff8bdffbea,0xfebf3f3fcbdffbee,0xeff6ffffcadefbee,0xffe75ffbcbd6ffee,0xffb9fffe8adef7fa,0xf7efbffc8bd7f772,0x77b277f4894ef36a,0x8ba3efbd8ac7f36a,0xef7ade7b89c6b3e4,0x5dbf5fe38adff3c2,0xfef7fffb8bdefff8,0xdfee7f628bdfbfe2,0x88f5eb7acb47fbe6,0xcfc3db7a8bcef3e2,0x67ffdfb3c35ff7f6,0xfefcd6ff8bd6ff3e,0xb3d7b7d04347f75e,0xfbbe6f6f8bdefffa,0xcf5bb8128b8ff2d2,0x2e5fdfcb0bcfd3da,0xd42df6f68a5ef36a,0xdd7fbfeb8bd7f7ee,0xffbc7fbc83d7bb66,0xcddff7b689cff7e0,0xabdd57dac9de3aee,0xb5f9ff6a8b5ef7ee,0x9f6efe7e88deb3e2,0xff9e6f768bd7ba6e,0xeffffe7f8ad7fff6,0xeefedf57cbd6fbe6,0x9ffb6effc1def3fe,0xf5fff7facbdefbfe,0xfbe6fde18ac6f2fa,0xc1bafeea8246f352,0x2ff7dad9cbcdf7ee,0x3b9ecfeeca5ef6fe,0xfffdf77dcbdffefe,0x7bb61f638b5ef3e2,0xe9c6e73f8ad7f2f2,0xeef3ffdf8ad7f3e0,0xefbfefe3cadfffee,0x7dff3fefcbd7fbe6,0x9ffefd61cadfb3f4,0xfb2e6ee78ad7fbd6,0xe6d7f223cbc7fb76,0xffffdfdecbc7fb76,0x9cfe7ffc825fb3ea,0xcd1e95008ad79340,0xdfb667d68b46f2a2,0xdac3fa1cc3c7e3f6,0x2afedfde8bdef3aa,0x921617e28bccb74a,0xdd6ce66e8ad5fb30,0xdeae36768a46feea,0xcdfcf33acbc6bbec,0xa9fddfdf0bd7a362,0x3eb84ff889d6f3ee,0x9cccb57ac2c7a154,0xcfb02d5c8ad7f7e8,0xebd6f32f88c7f6e2,0xaf5236178a86a350,0x1309479e88c65300,0x36e3f67749dfb306,0x1aa2144002d68222,0xae1a6260047826a,0x8d0128082845a00,0x6c9645240a538360,0x800551e100503002,0xfbf2bf718ad7bbf2,0xffbefffecadbfbe6,0xa97e961b8adfb702,0xab5ccbce88d7bb72,0x5010be5680d0f0a0,0xdca4aff088d6d226,0xba943f0580d7f6b4,0xc92687108082f920,0x1fb5cfdf89d6d36e,0x302a580080462240,0x5222bf408202e3e2,0x8800010081001200,0x81409c218854300a,0x241811c101540250,0xf7bbffee49d7b3fe,0x412a89508006f062,0x200000008020,0xc00000008020,0x0,0x0,0x911400800800f380,0x0,0xc00020200,0x0,0x480000440000,0x4002025002060212,0x8000000000020002,0x4000001000081000,0x0,0x0,0x0,0x810000018000,0x0,0x4000000400005000,0x50000208c00200,0x420204a01828202,0x20024002042200,0x4100001000200,0x8004000400028220,0x40001008000200,0xc90809230887f6a2,0x4020846002089060,0x0,0x20300008ad100,0xac60b4880c7a390,0x51c087081524242,0x28646000460202,0xfbea7d708ad7f7a4,0xc397ff7d0adfb7aa,0xcee3ba2182d6726a,0xffaecffb8bd7b67a,0xf5d49aeac8c4f7e4,0xd95efd728bddf3b0,0x9894ffa281dbd308,0xcfcbba1f8adffaee,0x337ddfd2cadff36c,0x80c8d6600a86a2f0,0x8a3ce4628354a320,0x7ae02f2780d6d360,0xee43e78883cbfb8a,0x834830db88d6f0a0,0x38f71ead8a97b626,0xd1ef174a0b563372,0xd393372ecadebfee,0x854df09480d0b782,0x2e5e781b8817f3a0,0x9e3347b38bc2faa2,0x80e53e6a084db340,0x128629c000c0864a,0x6ac6610b8842f22e,0x25c2da08c8879204,0xd2b80d2000568240,0x9974f6cb8b80fba0,0x2ba3f0088d6a26a,0x8ad4e228084e1042,0xc08008008500200,0x7bac471789c7e260,0x842650610b529240,0x58006c41800ab380,0xfa0269128206bbc0,0x4fc3fa1b8ad3f762,0x1bf9df5a8adffbea,0x34254dea8a9fd2a2,0x416ebf6289d7f6e2,0xf19077758bdcb720,0xffcef7eecb56bfe6,0xeaf6dedfcbdebe76,0x75937fee82c7b37a,0x43ba7778824ff3c2,0x80004c500042a120,0x810020030808d020,0x888380900028322,0x802247520a54b222,0x802193088a05a300,0x852005410a023220,0xc879872f89d6f338,0x8511b410056d068,0x5f104dea8b84a602,0xc121200008541008,0xeafe256083dcf328,0x4102e02009048222,0x9682360b8ad6f122,0x620dc2d20047a238,0x208012018241c008,0xc00afd608844a74a,0x48a740a28802b744,0xae673a9b08ce9318,0x24b108000a180,0x7475f4e38a46f3a8,0x851001001280,0x800010002000200,0x80c80000848000,0x802000000000600,0x1c1400020086a060,0xdbcbb662cac6bb4e,0x18861d6888c4f120,0x210b20a80560214,0x2010848600561042,0x4a2050280c4a102,0xc02404c288044200,0x7a84fd350b46a3ca },
	{ 0x8a503662ccc3ae00,0xcefe1fdccce3dfb3,0x3bbeeffacee73fa3,0xd567b6ffccc3be96,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfdeefff3edd7ffbf,0xf3d6ef2fedc3fe3c,0xcb7a1fdbced3bfb6,0xefbdcffacffafb9c,0xbdb67fedcdcffbbf,0xbebebf736dd2bf3c,0xfbbfff7fefcfffbd,0xceca3e0fede1fa22,0xeb9ddf5bede3febb,0x37ffdee34de7bfbd,0xff667ffecde3ff9d,0xfbbb2f7f6ddebea9,0xcffffffecdcffebc,0xea3ffbcee9f27fb6,0x1cb27f61efd3bf1a,0xfbebfffbcfdfffbb,0xcaafae7bccc32bb2,0xefffffbeedc7bf83,0xffffffdfefd7ffbd,0x8d3bd77bcdf7de97,0xffffffffeff7ffbd,0x7ffeff76edd3feb8,0xfbfef7fbcfc73fbb,0xebdb3edbcde3eab6,0xbffb4fbaedc73f91,0x95fff6fbedd27ebd,0x4ceeff63cc9b5e9a,0xebe4f77f4cdb3b3f,0xedf3fffeedcbffbb,0xefffcfdfcfcfffbf,0xfcff77ffcdffffb7,0xfffefdf3efd7ff3f,0xffbfff7fedf7be9f,0xfff7ffbfefdbffbb,0xafdfefdbefcfbfbb,0xf4f7feffcddbffb5,0xf3ffaf77eeebfb37,0xffb77f7fefd77ebe,0xcdffffbbcfd3bf3e,0xffffffdfcff3ffbf,0xffdeffefcfc33fbf,0xfbffff7bced7faba,0xfeffeffeeddbffbd,0xffdfffffcdffff3e,0x7ffeded3cdd37a92,0xc4bbd73aecaa22b5,0xf73cfed8afc33e99,0xbef6ffffcdc3eebb,0xe9bef76ecfdf7ab9,0x7fee36f32dd3bf99,0xbfffcffacddbff35,0xff7ff7dbcfe7ffb9,0x5bfefdf3cddb7f37,0xcb82673beee3cea3,0xfecbfeffeedb7e86,0xe7bf4fd2cfc3bf1b,0xfdf5fffbcdfbfeb8,0xebfeaff74dc77fa2,0xe9a67f77cfc6eb8c,0xbec7ba22cdda7ba7,0xafdfdbd2ccd37f3a,0xf4fefef2efcbbf88,0xdf7faff2cdd73ebb,0xdb937ba76cd33f85,0xbed3ff1e6dc69ebf,0xa9ff9d5a4ed77b94,0x34bf64facec21f30,0xf95aff7bcdf73f29,0xfebe6f7fedcf7ea0,0xdfdef7f24fc7ff9b,0x6bffff536ddffe9b,0xb7bbfff7eef3febf,0xd1e99d746df23a8d,0xfbae2d7dcc877f92,0x8b36f2afced33a2a,0xffffdfdbccf7bebe,0xffb9ceeeccef7eb3,0xf5dfffffcefbffbb,0x5ffe7f73cdfb7bbc,0xfbfeff6fefffffbf,0xc9fffebfedf73f38,0xb7ffcdfbedf7bf39,0xfcbffffdcfe3ffa6,0xdfbeff62cfdf7fb8,0xf2beffd76ed37fb9,0xcdebff3304d7beab,0xbfffef5fefe7fbb2,0x7efaf66f4fc76e13,0xa9ceaf716cd7bf9a,0xeb347faf6dd23f9b,0xefcfee77ccc7efba,0xe77fda51cfc32e19,0x1c931ff1edebbe83,0x9b25a55accc23b2c,0x93b52d420d895888,0x41c6d3ac4cd03e90,0x4edfffd36cf3bf3c,0x8e5e5f69c8d37eb0,0xd9ddfe7b8fe33a28,0x7eee7dfcced6feb4,0xd1d1f7768daeab03,0xaa760ede4cc13fa0,0x30b46c374c643e30,0xd9efa7f90da53b9a,0xbe7c6d706ce2fe20,0x5152f6018c973a19,0x60891a8980533211,0x292c55d2cc923804,0x5490f4f108831b8b,0x800150240020024,0xcaac2f5bc9d39ab0,0x0,0x1e0431348d41b00,0x190610c002282,0xcced0c60cda21e92,0x3e5c2e60ccc23e12,0x2481688080c20202,0xa055c00c4901220,0x80100821c6802000,0x20080000101000,0x9014004808901800,0x1428b604dc02a12,0xf300040c00212,0x162148fb44d15e24,0xa2b793288e82a100,0x55061a424cc34b10,0x803a1b0745817e00,0x6e428f4205d1fe00,0x3080030144c1aa20,0x9494a244c7c13a20,0x4042081020800,0x3004080004880000,0x5832a0080c62229,0x0,0x7174bece87c32aa8,0x2022086048810880,0x8000538688032810,0x4954f40a8c232238,0x8100001204080220,0x346941400d886e80,0x8080210004012200,0xa9144148c8a6a81,0x4000040008400000,0x28118400c424800,0x14aa4c2805804a20,0x1000004004020000,0x2014008000400010,0x1820004808000,0x50521805813a00,0x8012064140432610,0x8010000004080000,0x8080000801000,0x0,0x8d14e042cc20a01,0x8400004002000,0x9c62b0b80cc00a0b,0x80363d5046c3aa36,0xfbb6ef7fcdd7bbb7,0x5c8aa2fb4fcb1a88,0x6f9fc9b14ecb6ab0,0x647776f349d73e3e,0xebbe0fe02cd33e92,0xea80eb0ee4d292ad,0xe5bf28bcdcb7f21,0xb9ecf414cc3af91,0x10104ee088818800,0xc5aa150289c70b10,0xdf906f760dc1eb00,0xeda6ed2a0cc39a12,0xe35b92daeddbbe90,0xfeff5fbbefc7beba,0x9078df5444c72213,0xde910f52c4c28eaa,0x704e0044052120a,0xeb0038d34d622918,0xb62dd2f1ccc23e30,0x3020d6294983f804,0x9c047d604ec30610,0xc994d362ece96b14,0xc492274480523210,0x369df77844d63e38,0xd464bb7447d15e80,0xb0865018ce3fa18,0xc95442960e931290,0x40308000010202,0x3806404c020118,0x461064d068022e07,0x44a0054088424e80,0xbfa04342ccc0ae10,0xcfdebebb6dd7bebb,0x23fddbdaedd7fabe,0xd4f4dc720fcd2a80,0xb7bf7f6bacdb7eb1,0xfbb6efefccdefea3,0xd6d7bf0eade77e92,0x2a76dfcccfc22f87,0xf5fb5fe3e5d27e1a,0x416004080008,0x2900a0404888201,0xc2805b96e4c32a12,0x840068040040630,0x890005102c002380,0x804022424c012c80,0x806b0000c800211,0x40812885c00800,0x6c6801085823800,0x1d688fd045933a30,0x4810d0002c402a00,0x14840822000,0x180801808400201,0xaa4b90c2c8d25a82,0x2501005040000222,0x10118cc004423a08,0x10a050048e00a02,0x2102242080d01810,0x8152f8020c412280,0x8a004c1a44022231,0x3407f26048c01e12,0x8cfe87604c563b13,0x420060020c024800,0x1808520048600410,0x2004508048022084,0x810000000004800,0xa15706084cc12720,0x56041f2548521310,0x889013600482a600,0x256ec8180de33a08,0x82cfd384c21810,0x6018040410020,0xf3163f674ce66eaa },
	{ 0xa1c5e3aae05b99c7,0xc8d23e29ea52b253,0xbede4fcdee7616eb,0xf7bef6ff6e5a90ff,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6dfe7ffbef73bdff,0xfb76ff3efb77b9df,0xefebffebef7ebfeb,0xffbfcffbff7fb3cb,0x14fdeee86e1b94cf,0xdffffff3ff7f15df,0xff9f7f2fef7b9cef,0x1ca10110068804b,0xefbfffdf6a7e1fef,0xc6ffd7e37a7a1ecf,0x9fffff7cbe7abfef,0xffbe7fff7f7f9fdf,0xffdfff6bff7f97ef,0x8bd57c5a6d571549,0xf4bbffff7f7e14cf,0xd5debf66ea5f9dcf,0xfeae376e6e72196f,0xcbfaffbb6f5fbdfb,0xebfefedbff7e37cf,0x36934ffb247814db,0xd1ff77fb6f76974f,0xff767d7eeb7bb3eb,0x8bfffffdaf7637eb,0xffff5f69ef4eb5df,0xff9edfebff769bcf,0xd47656eeaa6a13cb,0x7bbc7df3e27217ff,0xebf5ff2e2a7fb9ff,0xcffbfe162e7697df,0xbbafcf6eaf72378b,0xf5f7ffefff7b3fdf,0xffffef727f7fbaff,0xfffeffff6f7e97df,0xefdfff1bff76b7df,0xffdfdfdeff7e97ff,0xfefefefbef7a9fef,0xfb6dcf7af2733fcf,0xdbbe7fe57e7f93ef,0xefffff3fea5ebbcf,0xfffedfdffe76bfff,0xfffe5fffff7fbbeb,0xfdfcffffbb771fef,0xffbfbfffef7fbbff,0xe9feff7eef5abfdb,0xefdfdfef7f779aff,0x4e93cfe16a7eb5cf,0xd16697edb65e33c3,0xefd6777fee43bf8f,0xc3f3fb97ea7793eb,0xefdf7edbfe571bdf,0xb7aad7fbeb7796eb,0xf5ffb7fe6a57b3ef,0x9bee7d61aa7297cb,0x7bffd22f6a761b8b,0xcd5bbfdbfb60b1ef,0x32bf4edf2b7a14ff,0x75f6ffec6f7eb6db,0xc7bebd31a2661c43,0xef86e7ffee539ecb,0x8ddbfaaf2b761bcb,0x6ffeffcb627397df,0x55f7ffee6f7a90cf,0xfdee3ff2626b10cb,0xbe3cffe36b7b136b,0x4d5fffb8aa66912f,0x2bfe7dcbae77385f,0x9ddb7fffba7e10eb,0x976faff2fe731bd3,0xfebf3f7fae5b97ef,0xebdfe37ffb7295ef,0xaffefedfb77ab5ef,0x94fbeffffe52b5bf,0xddfdff7fef7733ef,0xfbbe7d72ff7e93ff,0x8bf7fe67f67e96ef,0xeed57edffb7f1ccf,0xde9dc7eb706a114f,0xfffbffffff7e33ff,0xdfee7ffbff7eb7ff,0xfbffffffff7fbfff,0xaffbfffbef7f9fff,0x2f3fcfdeef7e1beb,0x7df7f7fa7b5b37cf,0xdbbfff73bb7f93df,0xdb9effbe665fbddb,0xdecfbf3b7f5b1fcb,0xeb3fdfdbea5b9fdf,0x3cff5eea3b76b0df,0xdf77b7f06f7316cb,0xffbf3fff7f7fb7ff,0xcfc7ffbe2c733ddb,0xabffdfdaae663dc7,0x6457d3ffa34d1beb,0x1924c46228461207,0x8c18a760e86a164b,0xe9c6f21022629887,0xcc7e9e53fa721deb,0x558adf3fea72124b,0x32ee376d2272b3cb,0xfa765efdaa7219c7,0x8b5ff3eeef52951b,0xafff7e5b3f7a31ea,0x1a2d35a28721182,0xf5d797ebef73159f,0x1b2a1f6026629486,0xcbd5b22eea4793cb,0x4dabe0b6b42128e,0x679cc9316a4238a7,0xacdd22804843300b,0x97103c53b052304b,0xda90876c22421c49,0xbc4ca022a401681,0x4840010002001101,0x6ee47e82b7e14ca,0xc1ae2f6869721a8b,0xdb06db7c25431bc9,0x54c0a73027429387,0x31c5e5727721d99,0xcf0dd61a25a02cb,0x91098000600a00c1,0xea0215832a601801,0x8001e32861619801,0x822a98ca207090a3,0x20884de8ea602201,0x5002016ea2408083,0x1ac27c621a62320b,0x82747970b2721122,0x2400000024000000,0x804000421010,0x9000000000421200,0x24052002000046,0x8100820010401048,0x480200002001402,0x100080001400000,0x6000023100480060,0x30000460400a1041,0x4200241200411220,0x4008100100001080,0x2002000000600000,0x0,0x200010000011000,0x1000080000001001,0x4000040024000000,0x4000100020008000,0x2000220000080,0x100046048000001,0x800010000000082,0x4908c00428141,0x2b00181921521088,0x1259302942000b,0x8080000020020000,0x200200020008000,0x2004000040000001,0x409080822501800,0x4b3c4600ea60002a,0xf46432087052100f,0x22542a4429093,0xeb98ed24266014df,0xcc53be7fb352b56b,0x28e8b336952144f,0xb3700668a77811ce,0xf7beef60fe7a93ef,0xdb91fb9f6e7a11cf,0xfadfff1fef7e91cb,0xab7fdfde617711ff,0xf0c6f5e6eb62948b,0xf96e676af37a96d9,0xdb3e7fe7aa6612cb,0x8f564f1cef6218e1,0xaef7ff53ea7b90db,0xb7badfeb7f639ddb,0xf96df67f787b10c9,0xc1a22f40226810c9,0xefcffb32eb70baaf,0x1f76cecce47611ef,0x33384759ac7e12c1,0x5004bc0808661408,0x4f526d696e6213cb,0xcbd6df6e6e52b1ef,0xfadfbfdb6f7b97db,0xbbbcca786e7a1fc8,0x1340d27263721401,0x4aa86d42265213af,0x1304033218603001,0x84198a1210609007,0xad32050267441402,0x153d002a2312140b,0x978207402000908f,0xa62c45682b42104b,0xcfc63613a272b1cf,0xffedfd72e7a9fcf,0xb4befff96f771eaf,0xbbefeff22f5797df,0x6fd3257662421783,0xdff6eb7fff7b92df,0x21e8cd1a2172b00b,0x2a002008000001,0x100040000400008,0x8ec60f600b72b0a3,0xb41c2342a5230a3,0xa2172e4040621501,0x78005000143100b,0x52487771285212cf,0xf2023220ca420043,0x816000a0600000,0xcec32e402263116e,0x3316c7f06856208b,0xc424504222609c83,0x0,0x1d0800e02422049,0xe8ea7e52f87e174f,0xb884520224c10ce,0x484040002080040,0x3054122c001007,0x6208c13c186a1441,0xfc7af3b6a7a12cb,0x2144040a0409442,0x109486e0204614c8,0x79ee7f7bf7721c8f,0x4a912d34a201118e,0x4041a010a2021082,0xaa15c5460260115f,0x869556e8017a31ca,0xd984b4220a503046,0x18000c0124401420,0x260000001049,0x48cc4cc9a76908b,0x4caa59872c60978f,0xe1228f58226a14c6,0x4294311c64021101 },
	{ 0xe9f3fb7309164ed4,0x6ed0be01483ccfd6,0x9a9bcdfb883e2c7c,0xf7fdfffbea3efef6,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xffbe7dfbea7ffff7,0xfbb6ff7feb7dbfff,0xeefef7ef4a3c97df,0xffbfef538b3feeff,0xf7fe167bea7efff7,0xdffffff5eb3deff4,0xebdf6e2fc21e1dd4,0x43cc1001011e0e30,0xefffcf5e6b7cbfe5,0xfefffeff4b7effff,0xfdef7ffbeb7daff7,0xffff7fffe93faeff,0xffffef3beb7dbfff,0xabffdfdf8b7caffb,0xf5bf7ce76a3fefff,0xdb67b6f8037f8ff5,0xfab73f7e6b3caef5,0xcfdfff3feb7fbef7,0x6f7f9ef76a3e7cf4,0xbfbb5ff38b3caefa,0xf5eff77feb3eecfe,0xefbe7dfde93efed9,0xebfdfbbecb7efff7,0xebffdff3e93caff6,0xdfbb4ffbe87eef57,0xff57767c093c89f6,0x7dfe6f7aea3dbfe7,0xfb84db36c87cfbcf,0xebdbde7fc97eadf4,0xefbffff2cb7dabfd,0xf6ffd7fa693cefdf,0xfbfefd734a7eb3f7,0xffbeffffeb3dfef7,0xefffff7feb3ebff7,0xff7fdfdbeb7ffff7,0xf6ffffffc97ebff7,0xd7eff7734b3ccfd7,0xdfffffffcb7fbfff,0xffefff7fcb7cbfff,0xfbffdfdbc93efff7,0xb7ffffffeb7dfff7,0xff6effe3ca7fbffc,0xffbe3fffeb3ffff5,0xeffffffeeb7fffef,0x6bff9fdfca7cbfd7,0xbffbdfffc93cafd7,0x51aeb571603e9e55,0xfffc7fff6b7f1fd4,0xabdff36e0b3cadd7,0xffe77ff7487fafef,0xbcfecffb6a3eeff4,0xf47fe6f4e93eecf6,0x6ebe7df0687cfff6,0xbb4c6f6ae93dcfd1,0xcc7b1ed9c83ddfcc,0xaffecffbc83cbcfa,0xd4bffeef4b7ed9f5,0xa96e7d61893ebe66,0xdb9cc7e7683c97f4,0xc5c3f958697e8ed2,0xc3dfcd48cb75f9ca,0xd477ffeb487caeee,0x83222b406820a041,0xfbbe2bf34a1da7dc,0xcfd7ff3aeb3eebf5,0xbffefccf683fa0f5,0x14ee5fe2e93dedfb,0x9169f74b0b1daccd,0xfe3e2fe6e83cbff1,0xcfe6fff64a3ffcf4,0x4f7edf5dea3c0ef4,0xbefbfffaea7caed7,0xf3fbbffa487fa7f7,0xf8e6ed62eb3ca8f1,0xe9dfefefe93ceff1,0x6df77e5ecb3e4b54,0x97fbdfdbeb3cffc7,0xfff7ffff6b7ffffd,0xffbefffbeb7ffff7,0xfbffffffeb7fffff,0xfffffffbea7effff,0x6f9fdf7b4b3ebff3,0xfdd5fffa6a7d15dc,0xdfbeff74c83ebedd,0xfb96fffae97fdad7,0xedefbef5c93caff7,0x7bdf5f5f693dbef7,0xb4fc7eed683caadf,0xb3ecff704a3e9cdf,0xdfb67f7fca3ffff7,0x9cc7ff7ec87c3ef0,0xef5fff5a487dfef7,0xbef51feb6a7efcf4,0xff9beeea486c8ddd,0xcab7aff70a7ebadf,0xc984e72d4b36aa4e,0x3ff92c38b3ecb40,0x3ab8d9f30a3ebfd5,0xd36dbf39c93d2cd4,0x93beaf7d6a3ebec5,0xe9e7f94f4b3cafdb,0xfe971ecbe87c2fc4,0xbbacdf7bcb3e2ff8,0xfddf97f8eb3dbae8,0x2a847403403e0154,0xfbfff37feb7efff4,0x85f7db49093498d4,0x821ec7f30a3fdaf6,0xd530f67f4a5c27f8,0x950221600b1c981f,0x8ccb20008382044,0x47c60a0948362644,0x81200f1000080074,0x8054a80900a842,0x412e4b62083c38c4,0xb6140144083e0a94,0x4782a000081c009c,0xf25485a083c16d0,0x8022b0180208882,0x106cef71823c88a1,0x4096010048102200,0x20222a260210140,0x1cd60008082050,0x208060008082004,0x40040000008082,0x400040420800a080,0x4000010808200200,0x9126800108246,0x22a80400000c8086,0x4164142a002c84d6,0x780c2d2000389893,0x802000040004,0xc0189140140180,0x8590080000360400,0x502cd4d4087cb09b,0xb038042088302800,0x4320007280148c51,0x604220088100a40,0x0,0x0,0x2100060000000,0x0,0x0,0x0,0x80100008002000,0x0,0x8000010008002000,0x80400000100200,0x8580283082c00c0,0x4114060081890dc,0x1002100080004,0x200100100020,0x8800200348100440,0x60a3200601000d0,0xb1cf41b413e88d4,0xd1f49a634334d8c3,0x5a2024400010a1d2,0xfbd0ef4e4b7ecbdc,0xafeb9880ca3d0df4,0x3e004f804824ac01,0xd4f7576e0a7faff7,0xff247c604b7ca8f6,0xfba8ff3b683edef4,0xbfdffef18b3e9ecf,0xeedf5ed6ea3cbff7,0x103e5420081e22d4,0x84dc2571ca148e93,0xffb02747ca1cabd0,0x5fd6f798cb3feee5,0x2dfb1d4b8a3d6bfc,0xbc131deb481eb2c0,0xf7b2af20c23dbad5,0xb8b2e762c8188fd0,0xc59ac379801c88b5,0x28675ad1403487d0,0xf4320fffe97c3e90,0x16cd368031c98d4,0xa00276c003e87d0,0x81720b1dca3489c4,0x6a563fdb8a3eb9c6,0xe92cf02091c2ac0,0x53507069c810aac2,0x29402f22603482b4,0xd070672c0924e8d0,0x4000202000240040,0x65064012001091f0,0x8010004200260040,0xc7bc2e60c834e8c0,0xd4003708406c0a50,0xcec23c1bc83cfef0,0xeebfcddbc93eaed4,0x966646e0893e4d54,0x9f7ffff2697fdffd,0xdf9a6f3c483dbed6,0xefc7fbbfe83cbbfe,0x85b18514a1c1c80,0x9081512009240a40,0xd800001040108c80,0xd485201000341050,0xc050c00000248452,0x8004000080010,0x80324e23411c8ac2,0x0,0x4212310000080050,0x20008200000,0x84cb0a46c03cc8d0,0xb5bfadf3683daeff,0xc084166a001ca0d1,0x88042560483c1040,0x8210400000180480,0x8622290600e82d0,0xc90854000349141,0x4200080040,0xa1da209248880,0xeaa5e32c083e94cc,0x8972bea06a1c8f52,0x621881c2081c3080,0x549476ca483caef2,0x808d4060001c14c0,0x9a94a364082c8ad6,0x4280400208380210,0x3f40541030c8c80,0x0,0xa0040480002c00c0,0xd4840e66c15c8a83,0x89448b0ae8100a50,0x102ca40240804,0x27e864f34938c9c6,0x44245c2a80150d64,0x6a408c08900a880 },
	{ 0x4200600200094d00,0xf9565b575b181f19,0xde9fc76b1b9b5f5a,0xdf7ef6eb038ac79b,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfebe7dfb7f9acfdf,0xb3c6bf62268b4f5f,0xefe7bf696f9add5b,0xe7bfcffe379bdfd9,0xfcfef7fd7f1adf5b,0xdffbef76579bdfdb,0xfbbfffff7f9acf5f,0x252300a658acf1b,0xbf7fffcf7f9bdfdb,0xc7f3f6fb7b8a870b,0xfffeff737b9bdf53,0xffbfff7fff9bdfdf,0xcdffffbfbf9adfdb,0xefdf5adfbb8bdfdb,0x3bb7ffffd71adfdb,0xd9fbffd42f9b4f5b,0xfebc3dfe7b8ad7db,0xefdff7fbbb9bdf5b,0x3edfffdf379b4f5b,0xefab6feb5f9adfdb,0xd7effdfbff9bdf5f,0xffbe5fe77b9bdf97,0xcbf5fbff7f9adf1b,0xefd73fff6f93cfdb,0xfb1fcff30f9ac75b,0xfdf7fe7f3f9bdfdb,0x6fc87d59369b5f5b,0xfbd5ff6f571bdf5b,0xefdfffffb39adfd7,0xffbfef377f9bdfdb,0xf5fffefb171bdfdb,0xfff7bff1eb9bcfdb,0xdffdffffdf9bcfdf,0xdfdfbf1fff1bdfdb,0xefffffdfff1bdfdf,0xfefffff8ff9bdf5f,0xd97fff7b771bdfdb,0xdbb77f7fdf9b5f5b,0xcddfffeafb9bcfdf,0xef5fdfdfff8bdfdf,0xddffffff379bdfdb,0xff6fff7b7f9bdf5f,0xfebfbfff7f9bdfdf,0xcfdfffafff9bdfdf,0xfffefefdff9bdfdb,0xbfff4ff2ff9bdfdb,0x5178b46b33881f5b,0xdb967d735b9bcf5b,0xebd5db7f378bcfd7,0xeedfffd33b8bdf1b,0xbffbcff11708cf5b,0xd573ff7d668adf53,0x5e7ef5f3978adf5b,0xfbdfef6e731acfdb,0xeefbfe327b829fdb,0xcffdfff86f9bd7db,0xbfff5fedb78acfdf,0xa794bf30030bd71f,0xff86e773f21a4f5a,0x4dcefed8d703575b,0xabf6dfdb6f8b5fdb,0xf4f77fefbf9acfdf,0xcdedef4b0b8bcb83,0xff9cbf5e1f83df5a,0x7d88cf187b824b1b,0xebfbffdbff9bdf5f,0x35ffdffb5b9b4f5b,0x9deefe7a7b12dd5a,0x5ebfbf7fbf1bcfcb,0xcdf6ff32539add5f,0x1b7fff5f979bdf5e,0xbfbb477b3e9bdf5b,0xdbf5bf7e6d9bc75b,0x779e3d756f9bdf5b,0x8bd6ff773b8bdf5b,0x6fdefc57079acf53,0x9f9dffff3792cf5f,0xfffdfffbff9bdfdf,0xfffe7ffbfb9bcfdb,0xfbffff7fff9bdfdf,0xeffffeffff9bdf5f,0xffbfdff77f8b4f5f,0xfff73eff4f9a5fdb,0xbebe7df51f1bcf5b,0xfbbfefbedf0bcf5b,0xefd7ffbf7f9bdf1b,0x3feedfdf4f98d71b,0x91fa7eec3e9adfdb,0xcfed67fb6b9bdf57,0xdff6ff6f1f9bdf5b,0xcdffff3eff1acf1f,0x4bfecf5f5f1ad55f,0xb59679f7b28adf5b,0xddcdc5c2021acf5b,0xdeb77fff7f8adf9b,0xcddefaee1b1bdfdf,0x6fdf5bdc3393d79f,0xbebbefff7f935d52,0xd3f0b743538ac31f,0xd8b67f741f0bcf9b,0xc9effbff179bd75f,0x3edbdf43070ad71f,0x90975dba0913cd5b,0x75b4fef9b31b475e,0xcaae3d70db8a4d5b,0x9980e716de0a451e,0xe463df5812035f1f,0x6eb68f927301cf18,0x308254c0618a8101,0x291e24601288050a,0x500426080600410a,0x58da3fbfba11cd9b,0xa0015d400412551a,0x62262804015993,0x214fe76aaf9bdf52,0xc0007c0010004d4a,0xc849561036104709,0xa1c14020c02c51a,0x84d556a82d01431a,0x108860203114410,0xc9ac7de05911cf13,0x820000000402,0x400000802000200,0x4914020800a0442,0x440400380800051a,0x204456440008113,0x90c2246908010b,0x2020000004800,0x8414000100412,0x2800000408,0x40200080000400,0x0,0x0,0x10020000000100,0x41fcf2e8a709555a,0x8088d2003924408,0x880c280381014610,0x802100000000381,0x101914020001518,0x2000002002000008,0x2404080020c00,0x0,0x0,0x100400010000400,0x22004000080112,0x24000100400,0x40a9044202020253,0xc10202000112,0x48161a04010a000a,0x400002000000012,0x901c50255500455a,0x102100010000100,0x100840008020140,0x10040000000100,0xd5bcff34e9a5fdb,0xf3c5fe7a521bcd49,0x5f96bef2798a5f5f,0xc901eb6253885552,0xf1da064f909346d5,0x104d421403440a,0x99d4d6620b08471b,0x9fc276435b82175b,0xfba5f76f5b02cf1a,0x8fdbff231b8acf1b,0x6bf6cfca3792df5b,0x74b95d61f19b5f1a,0xda2ee5721b0a475b,0x34b00b74080253d3,0xce7aff5a0e9a4f4b,0x83fefed7d31ad79b,0x330476a03004f4a,0x49e82d084093cf1b,0x7a9faf3cd61ad55f,0xe342dd2a7b88df53,0xcace5c19211a4f0a,0xf7fe5df33a1a0f5b,0xf42ea6f18a0a9f5f,0x78143434f2020f09,0x83bab221050acd43,0xeff71f573f8acddb,0x913ec7e33f925fdb,0x5814220818000540,0x87821f405b024f03,0x8180622610080412,0x48c2921101084f09,0x831244a0108d112,0x508056420f004313,0xfb32af72038a5f5b,0xf0243f361392055e,0x5fc397b63b8ac75b,0xab97c7dfcc0a4fdb,0x96be7ee8238a8f5f,0xfb7377f36b8b5f1f,0xdf92f77f5b9ad71b,0xefdfff7e971adf5f,0xa2109141020b4059,0x0,0x844127000300830b,0x1885010001820110,0xc0c0130201824310,0x21080000400,0x8400001000010,0x2404044008020144,0x608e0de083088713,0x808b0020100500,0x8744e00218a0b13,0x469fcde2738adf9b,0xd223e6fb4308cf9b,0x5d262e6a0118455b,0x80d0024204184500,0xe040b412000206d9,0x36808410210a0902,0x308d002012820300,0x2a262c20028a4f9b,0xfe2cf37e321bdfdf,0x4dd0de10138ac395,0x808014009000102,0x54d5f58efb081793,0xb97cff72571acf53,0xd21461140002cb1f,0x1907f66454084759,0x4a80194820080709,0x441415831008c903,0x8849660040080581,0x7c952bda131bcd5b,0x8110c9113000cf01,0x44000000080100,0x88434a32008110,0x50120d08000a0b01,0x5e9e77740b0bdf5b },
	{ 0xe8ebeb38fcd15ce2,0x6ffd1f4c15a84ee6,0x1e8fddef9dfdace2,0xf537f6eeecfcece2,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xdebe3dfafdfddde6,0xfb8afbfff5fdfee6,0xeeeadedb54bdefe2,0xafff5ff37db9efe6,0xfd7d77fe3df967e2,0xfffeff73fdfdffe6,0xffffffff7db9ffe6,0x2956f00f1cb9f9c2,0xbfffdfdf7dfdffe6,0xf7f7f7fefdbdffe6,0xfffffffffdfdffe6,0xfffffffffdfdffe6,0xffffffbfbdfdffe6,0xeffedbdfbdfdefe6,0xffb3ffeffdb8ffe6,0xe5edf772bcf97fe6,0xdabeaff6fcb87fe6,0xcff6ff3bb4f9efe2,0x2bfeb7d394f967e2,0xb6aadffb75bd7ee2,0xf9ffaf7a3df8dfe6,0xc7fe7f76bdfccee2,0xd99ffb7bf4f9efe6,0x6ff736dffcbdede2,0xfb2fef7b5cfc6fe6,0xbaeeff6eddbdffe6,0x7dfc6fa39dfd45e4,0xfbd7ff3ff4bcffe6,0xefffbefbfdbdffe6,0x7fffcfebfdb8efe6,0xf5f7f7eefcbcefe6,0x9ffafff1fdb9efe6,0xfbbfffffb5bdefe6,0xeffeff3f3df9ffe6,0xffffdfdff5fdffe6,0xf4fffffdfcfdffe6,0xff6fff737dfdeee6,0xffb7ff7f9dfdefe6,0xfddfdf2ebdf9ffe6,0x6fffdfddf5bd6de0,0xfefff7fbfdbdffe6,0xffffff7bfdf9ffe6,0xdbbcbffeddf9ffe6,0xe7f7ff3efdf9ffe2,0xafdebfeff8f8efe6,0xfebfeff3ddbdffe6,0xfdfff7ff3db9efe6,0xdffe7df5bcb9ffe6,0xebcfff3a7df86fe6,0xffff1eef3dfdffe6,0x8baacffbf49d7ce2,0xf55676fd3cbdeec2,0xdfce7f70b5bd7fe2,0xfbfcff7ffdf9ffe2,0xeefebffffcfc7de2,0x28bccafb34a546e6,0xb4331feb3cb947e6,0x7bfebf72d5bcece6,0xcf01bf6f1db867e2,0xcecbf62b7cd86fe6,0xaabfdeda7db96fe2,0x36f457e0b4b8efe2,0x9b7fff5ab5b9ede6,0xfb9a7f5c9cb0d6c6,0xdcd78fbe35f8efe6,0x8bdefecf74bc7fe6,0xadb6fffb95bd8fe2,0xb97cf667f5fde7e2,0xde9e3ff1d5f9eee2,0xff7cf33195f9cfe6,0xb5e3e69fde8efe6,0xb6b25ffb7db96d62,0xffd2b77a54b9f9e6,0xfbb6bffc74fdefe6,0xcbf2f363fcb94ee6,0x4dff1e819c9c6ce0,0xbffbfffbf5fc5fe2,0xfffffffff4bd7fe6,0xe7b6bf7935f9ffe2,0xfbfffffffdfdefe6,0xfbeffe9ffdb9efe6,0xefbffffffdfd7fe6,0xf7ffffeffdf9ffe6,0xffffbf737dfdffe6,0xffbff7fff4bdefe6,0xcffffea5d4dc7fe2,0xaf5d7fdf7c917fe6,0xf4f77fea74f9e6e2,0x896cbfb8a5396de2,0xffbc6b727dbde7e6,0x8dc7afb0fce85fe6,0x6e5e9cdb94f877e2,0x309e1cef95b8eee2,0x7aaef5e358316ee2,0x9eda2d51dcdd4de4,0x4b8af7a79df94fc0,0xefdffedfbdbdffe2,0xffff6ffbbcbcfde6,0xd1aff37b1c997ee6,0x5bd67f3015bd6fe6,0xebe7fb2f14957ee6,0xeeea3ffbdcf9efe4,0x9799cfd23c99eec2,0xb576f66e7db9fbe2,0x67aa2df8bcb8e7e2,0xfbd4f1343488f7e2,0xae42bf59f491ede4,0xffa64fb33c8ccbc0,0x84f5e6c01cb949e6,0x928d21c011a087c2,0x8180c80405b08480,0x4dca1c1414986d60,0xa08ddd20db8ed00,0xd290746054184fe2,0x98d2104040184ee4,0x5ebc4571141942a0,0x46008330141166c2,0x2bfece5635b966e6,0xa0d194ab10d84d60,0x8001820000200286,0xd286b324cdb86ee0,0xcdd0d5b255a0bbe2,0x6cc318581488a7e0,0xc3914ce11089c842,0x5c449443353803e2,0x4a86041020d802c0,0x800002000006000,0x860a0900106060,0x28b23b82b0808dc2,0x4104007230004484,0x1884040010880282,0x80030500c40042,0x84ea515191382d20,0x864840140143a0,0x95b510080490c2e0,0x0,0x800000c38004000,0x8020000010000400,0x8010170204d00000,0x4001200000,0x200410000002,0x0,0x0,0x0,0x301a02230099c802,0x0,0x8002000110000,0x187900800196000,0x40020a00310040,0x12420000204042,0x5000845040086020,0x4a12040088904540,0x1048020040008c0,0x808408000c000480,0xffffefe3fdfdefe6,0xfff6f7fdf5b97fe6,0x577e7df9bdf96fe6,0xbbf2ef331cf8c6c6,0xa421e0ad0194460,0x814014001282,0x32008269748054e6,0xa50e7d640cf87ee2,0x73b6fb7c30b967e0,0xc277b23ff4f97fe2,0xa3dfdd1b3da9cce0,0xd60516e9f4b567e2,0x9a4f7638c98ace0,0x9abc02460d306a42,0xedc7fe0e34b8cee6,0x3fed9f581c99fec2,0xb6ba4ffb54d9f7e6,0xfbefff7bfdf96fe6,0xa8be2f50c5a8dca0,0x83d9faf03cd8e5e4,0x1e1c024011b98ca0,0xbcfbdfbb9df947e4,0x108400a124901040,0xf12c3f6c3c0847a0,0xcad0db0804b94fc6,0x2cf73adf3db56fe6,0x87364f7254117fe2,0x911cf22bb0b96462,0x1816045045b04820,0x4290c00214802cc0,0x458021000a08042,0x480843f10099e040,0xd052b34204dd7120,0x1be4d6030b8ece0,0xc2304006141158e0,0x7fc6ba3b5db973e4,0xbb1ec7de94d9cdc2,0xc49276f235b0ffe6,0x976ffff37cbcefe6,0xe3b6e60f95b85fe4,0xe957f71a1dfdefe2,0x62040880084420,0x4ae102210848da0,0x80407750009866a0,0x2802d42140027c0,0x89040a0020112040,0x10340010112000,0x688140890842442,0x940070a05a003a0,0x4204e1e0049982a0,0x8124ae0011b0a400,0x651300180b90400,0x20a4c55290984560,0x61365278841175e0,0x6010b53004bc4342,0x8000200008004000,0xc002023090006d20,0x882090010904820,0x34a006ec10e02760,0x9ac020401c096060,0xc18442bc20186ce0,0x4a433681cdb0e6e2,0x104e4000880120,0x84d054e850b9e6e6,0x95be21603c986ec2,0xc0843b5211a1c684,0x48d5c316149975e2,0x4a061001b4011680,0x18155a144b04462,0x4840166084806d00,0x18282f4c14989282,0x8141a80010980aa0,0x6de8090004ea0,0x190c25000306442,0x2150934840980,0xa804758208004ca4 },
	{ 0x30ae3289610566a,0x6b52db69a80f756a,0xb4b5ef533eb6d4ee,0xb4f6160d94f650e2,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2100020000020,0xcbffeffffefef6ee,0xcf4bbfecacf2f4ea,0x2fbbcf58caf670e8,0xfcfdf6fb3cf6f7ec,0xafb6ad73aede57ee,0xdfadeb7fb6f7f66e,0x6bfefa9fea7ef7ee,0xbfffdfdffefe77ee,0xffffffffb6fff6ee,0xfffffffffefff7ee,0xfffffffffefff7ee,0xefffffbefefff7ee,0xefdeffdfb637f7ee,0xb5ff56fbfcde74ee,0xf869ffb2fc74f46e,0xfebe377a6efff6ee,0xdbdbffffa6fff7ee,0xebffff57befff7ee,0xbfbb6ffea6f6f7ee,0xf5fdf6daf45e77ec,0xf3b67dec2436f6e6,0xe9effbffbe3ef6ee,0xeff7dfddfc7674ee,0xbe9f5ffff676f6ee,0xbe76feeeb6fe77ee,0xeffe7d7bbeb6f6e2,0xfbe7ff3e3ef7f6ec,0xefffbffb7efef5ee,0xbfbf4ffbfef6f6ee,0xf7fffffefeb7d6ec,0xdffffff3fedef7ee,0xfbfdfffffe7ff7ee,0xdfdfff9ff4f6f3ee,0xaf5ecfdfb63676ea,0xdefffefff676d7ee,0xfdfffffba6f7f7ee,0xdffefffff6f6f7ee,0xfff6fefef4d6f76e,0xfb5edfdbfe76f6ec,0xbdfffffff4fe77ee,0xd9fffffff4fef6ee,0xfeb6ff7ee6d5d7ec,0xef9fff7fb6b275ee,0x5f7ffe5de676766a,0xbebf7ffff4b772ec,0x77febfef9456f5ee,0xfafe7ffefcb6d6ec,0xcbd7fbffeefcd6ee,0xeef77adaeef6f6ee,0x3faf5fd966f6f6ea,0xbce67f7bbe1ff66e,0x79bc4760663677e6,0xebbfd367b63776ee,0x8ca33550632f66a,0xafbedde3a47676ec,0x54f4d6f1961776ee,0xefbefff2f67ef46e,0xdfc4e5ffb4325026,0xdfcfb717fcd6f76e,0x33ea4d58045574ea,0x1cbf64624456d6ee,0x913ef773207dd4ea,0xc954362ec446f1e4,0xc24be31a64f0146c,0xafdf7d93e6bef3ec,0xb5bedfe3e632d2ec,0xa1ed7779861ef6ec,0xde8fbfd0a4f6f2e8,0xc1fedf287436f66e,0xa94ecfcfa4b65562,0x9693dff68c1456ce,0x1ae31f79ce16546e,0xbf9e7d3dc63677ee,0xe33eef628a1ef6ea,0x2fd6b6a79434f6ca,0x838ecff016367468,0xd55bbfefacf272ec,0xfffe7df3feff77ee,0xfbfe6f3ef43676ec,0xcac67f7fa4bed16e,0xe3bedff3f4b672ea,0x75f7d7ffa676f2ea,0xcdbfff73bcb6f6ec,0xff9ef7fffeb673ea,0xdbdff79bb4f6f7ee,0xefdfffcbf63ef76c,0xfcffddef96f7f7ee,0xedefef72e6f776ea,0xdf967f7bd654d4ea,0xddf7f73ab65554ea,0x4a5edfd95416d7ee,0x446a57e936b670e8,0xd1e6f75e20d676ea,0xd3a92f67a697f26e,0xce02f6828405d26e,0xafddc9cf240574ae,0x9ebcfff9acf7746e,0x8179b06224433464,0xeb06bfe6acf677ea,0xe9bffb6eec16f4ea,0xfff7fedff6fef6ee,0xbffffffffefff7ee,0xf7fff7eefebff6ee,0x64ee6460b4d2566c,0xd9d4c76e341e56e8,0x87421e2314cd35e8,0x85fe8576081ef46a,0x3020004020000460,0x402953040801428,0x90000100000010e0,0x862e15a0525024,0x2b5047582676766c,0x341510218402502e,0x812f0448804e12e0,0x3200824a00161202,0x42200008801000,0xefffdfdffef7f7ee,0x804881400440a0,0x61025808441408,0x800020000000028,0x2446680814230660,0x2e71e60f614f56a,0x2a03510006e208,0xe061176894961682,0x243e0570c004156a,0xe3b9ffbfcafe54ee,0x40001000001400,0xa000140000,0x48000040002,0x4000a00080101008,0x80800000000240,0x4821280002210e0,0x0,0x0,0x0,0x20020400001402,0x40200004000004,0x21009458044412c4,0x2020000000000140,0x80617240808010a8,0x7c8908607480146c,0x308c402b4100028,0x200004000040002,0x4000002000401000,0x2008000000c000,0x3029d0200052e0,0x60804200400a0,0x1000020080001000,0x88000000040080,0x100900800801040,0x44004002000,0x80706080340414e0,0x2150000c060010c4,0xbfbdeffff49e77ee,0xf5fffffffefff7ee,0x7bfe7e6234d3f46c,0x89e00b1b2c46f06c,0x43422c4800104028,0x2108080000423004,0x20d966f244b45468,0xffbeadf1f4b7f6ea,0xdfa7ff3f9676f2ee,0xcf76fb0f047e12ae,0xabcecd5afcfcd56e,0xf6a686eaf412f2ea,0x979b7ffbea4cd46a,0xb394057420c235e8,0xccfeef1bb67777ee,0xffdfdfdfecb6f6ea,0x953f7fafe072f46c,0xf5f7f776b67ff6ec,0xc2e25614426502e,0xa30eca29b443366e,0x88811b5b36041660,0xb65aef7ae4fe54ec,0x859c05008004524a,0x8104394c84541024,0x5393ffe3a4f2366a,0x86a137ab43374a4,0x8c4e800012520a,0x10080000c000,0xa020c4024401020,0x4010642000068,0xa85000ca400616e8,0x3d78cff9621556ea,0xb13148a000535662,0x61e23de0a4a4f56a,0x99c46f3a202272ac,0xe7c6362a161454ee,0xb76cd9d23c1e17ec,0xd47777e8a6be76e8,0x8c727f6734a6766a,0x52b06607a49550a8,0xeef6c201a41d546c,0x2807b50a4067026,0x820410200801248,0x59c068408406f0a8,0x2082c632014c0a0,0x4fc080028080d080,0x9530ac38414d08c,0xc034904801e4024,0x416885528006d0ac,0xe02265c016221648,0x81b0f293025d146a,0x42c09652a044142a,0x992acf91a674546e,0xdcdeb6f4e2b7d4ec,0x5ba4058044ec542a,0x8106e06eb6104264,0x444a10a0000054e8,0x21e0380200410a2,0x2c08000224000,0x4f06500800631e8,0x8a0008001002,0x844bb748a41e34e8,0x3b4205c8e81230a8,0x86307448144110ea,0x800205628452520c,0x302026400001268,0x746a602041650a4,0x23f51cd18ed635ee,0x62e18e820d414e8,0x218c25f20802742a,0xd0006824c404302a,0x40001000401000,0x210014034000020,0x809804a3201264e0,0x1040a65086220060,0xfe342571f425d468 },
	{ 0xa9543516715c07c2,0x802101010000400,0x38beefe371fdafec,0xde79ff7df1ffb7fa,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd9ddbfda716d03fa,0x0,0xe2d5d31051bcb2d8,0xcedbffdf71fdb7fe,0xeffec7da71942eca,0xbdfffffdf1fdbfdf,0xaede6df0517faff2,0xfb97fffff1ffbffb,0x5fffffbbd1ffbfff,0xbfffdfdf71ffbfff,0xfffffffff1ffbfff,0xfffffffff1ffbfff,0xfffffffff1ffbfff,0xffffffbfd1ffbfff,0xab7ef9dff17dbffa,0xebfa5de1717f17df,0xf8fce771717cbfd0,0xfebfff7e71ffbffb,0xf7d7dff371fda7d8,0xeffeffdff17daedc,0xdfbbfffb517db7f8,0xf1fef7edd1ff8eff,0xfb9ebf7751dfbfde,0xc1f6fbfd515d87fa,0xef7f7fff71fda7de,0xff9ccfffd1ffbffa,0xfcfff67df1ffb7ff,0xf7ae7d73d1dfbffa,0xfbbbfb7f71debffa,0xfe5bb7dbf1f50ffa,0xffbfdfffd1febffa,0xf7ffdefff17db7fe,0xffbf7f7071ffb7f2,0xffffef7f71ffbfff,0xdfffffafd1fdbffb,0xffffdfdbf1ffaffa,0xfefffffdf1ffbfde,0xffff7f7b71ffbffa,0xdbbe7f7f51ffb7de,0xcffaff3b71feb7dd,0xefffdfdff1ffbfdf,0xbfff7fef71fdb7fb,0xfbffff7ff1fda7d8,0xfe8fbffef1fdbffe,0xefdfffff71fdaffa,0xebdfffdf71573efe,0x9ff3fffb717fbfff,0xd7fff77ef1fdb7fa,0xf7227df0d1753efe,0xebfff37f71fdbff9,0xfefbbfcf71ffbff8,0xbfff7dd671d63efd,0xd4f777fbd17dbfd8,0xffeeffe271ed97ef,0xebf7773e71fdbfd3,0x8ea92bb71142e10,0x819fcf79d15fa7e2,0xfdd2beea51df8ffa,0x17b2ffe1d1f59762,0xfa93e71e115d87d4,0xf5ffe3f5157bfd9,0xafbecb50517d9fd3,0xdef5f6f8517fafc7,0xc7fef77251fcb7f9,0xd29c2d5c5114a7d8,0x8f92fa0070340258,0xa3dfddda51f587fb,0xfdbf5fef51fdaffa,0xf5ec2e78f1fdaf4a,0xbc2cef6e51dc2fd6,0xeffadf3651798ff1,0x223fce6b51571af2,0x8eaa56f0505d9758,0xf5eb37eaf1358fd7,0x5fbebff771dfbff8,0xc3dffbff51febfdd,0x6fffb65351ffbfde,0xa7f6c9ff5175af79,0xfd7fff6ff17faf7b,0xfffe7ffbf1ffbfde,0xbbd6ff77f17fb7fc,0xcdfe3e3971ddafda,0xbfbfdf5f51dfbffa,0xddfed7ef717eb7da,0xfbfffd6371fcafdf,0xdbbce62a700ca778,0x9fe79baff1ffb7da,0x3fffdf5bf17db6d2,0xf4b6feecd179bff8,0xf9eeff70d15fb7ca,0xdf9f7fff71ffabda,0xffdfeffe71ffbfff,0x69dd5dd9f1fd2ee6,0xeeffdffb717fb7f4,0xf9bda7fe715faf5c,0xbeb62f6d514db7da,0x4b5bf732506da7d9,0x2ddfbff6517587f2,0x16127ff671ff1ffa,0x90fab759317d9f5a,0xfbbe77edf1cc87fe,0xe9fffffe71ffbfff,0xfffffffff1ffbffb,0xfffffffff1ffbfff,0xbe7effff71fcbffc,0xf7be3df9f1e5a77e,0xdbf7ef7f717fafff,0x860e3211d1350ee2,0xe9af4fd170d5376a,0x4124271d030aec0,0x41882f11107d0b58,0x10000b0000210242,0x463d80850100290,0x6844b831115244a,0xd274e95151b248,0xf0864570d053be41,0xdab6b77cf19dbffe,0xa14ed204518cb140,0x1a4ece9b717997c4,0x100010241040000,0x24000011000000,0x400c00a020400680,0x8120060820340200,0x480c1a5c4104a36a,0x410000100601,0xf021546990580fd5,0x202012451a4000a,0xd9d4df7af01c3ed1,0x2044b948510f036a,0x80008a040100090,0x94648440501102c0,0x0,0x840030041400200,0x2046800000580a00,0x8400000200200,0x0,0x100500100180200,0xdc042a025124b152,0x42020110848040,0x727f0bca5130aec2,0x4794be05054bd88,0x5440100080c0,0x0,0x0,0x14000010250,0x1430400000248042,0x9000200111002040,0x40a0000008240,0x4110c00441008080,0x2080b0800044014a,0x8083c62150503700,0x20270241081218,0x88043402405c0208,0x200000200100800,0xe006188040411300,0xfeffffeb51fe9fdd,0xd777fefd71dca7df,0xbb367f32516fb3da,0x5306372771592342,0x8b4bbab95167b7f8,0x0,0x2044c010188200,0x13228743103c0f58,0xf3beffed71bcaffe,0xc0d6ffb651dda3d8,0xb5ede17718c97d2,0x4b0717951509bd0,0xf38bb7e35159bbd2,0xdb967f7d51f537d8,0xeddeef3f517fa7db,0xafcebd5fb01cbed9,0x9ad05ce8d11cb698,0xb175ff6a71e5b773,0x72961de651e6b6ea,0xc4d8f1b7015e0b7a,0xbcc89105001b2d0,0xef3c57fa51bcaed3,0x80607648519c8200,0xf6005d7e5104b56b,0x6104e00510189645,0x46863880106c0680,0x48140400180,0xc810a22910188ac0,0x44220c60410083ca,0x2154000811180610,0x8a421b6a5045b3e8,0x91b04b6051749672,0x32566041408142,0xcd22bd42d14423ca,0x4220418a01940158,0xcfd7be1b514db3f8,0xbb7fdd9b511da7d0,0x14b33aeb717da7ca,0xc9fe2561714da6ca,0xdf9e7ffff1fdb7fd,0x8a4ce88a50fc8350,0x40161883203026d0,0x2100012040202040,0x502d34e25145a24a,0xc000251100610240,0x4848962870341210,0x25312c0208d8240,0x86324f607105a4c2,0x11ae4d00510106f0,0x74b20d4050d42240,0x8970d3e1615d87fa,0x4881160131508360,0x1591c580101ca5d8,0x30ff1ae1719da652,0x49247621416da3e2,0x8018404010088680,0x4881930151208240,0xe70a4d935120a3c2,0x414200050000480,0x402254040040248,0x4000240d51000640,0x1002121111850040,0x82030000420240,0x41b2f6ff716ca6e8,0x10a074040340a10,0xd314aa134144235a,0x8052161b511ca7f2,0xa2561f4a507c33d0,0xa4a37eea7144834a,0xd05026d0701586d1,0x96141d4841848661,0x6906c00251d40202,0xa54008081100a040,0x14b37ce8715dae14,0xc810008030000680,0xffb6666c517c97cb },
	{ 0x81568b4490219550,0x10884a2310e4a3a2,0x4ce4dff89867b452,0xf77ff76ed2faa572,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xb0114ec21a26b544,0xfdfff6efdaedbdf6,0x80010000020,0x6184f13680acad40,0xcdfbff7fdaf2fff2,0xfe374cb29830bd50,0xf1b3befbdafbad66,0x1fb23d509a70fd00,0xdbfeffffdaffed76,0xdfffffffdafffff6,0xffffdfdfdaf6ff76,0xfefffeffdafffff6,0xffffffffdafffff6,0xffbeffffdafffdf6,0xefffffffdafffff6,0x6ffebd5f9afebf76,0xfeff7ffbda2afde6,0xfdffef7a9af3fd54,0xffbf3f7edaf8aff6,0xefdffffedaf6bdf2,0xeddfdecfdaf4fd56,0xfebedf7b98fdff24,0xe6ed376392b2adb0,0xfabe7ff692f2b5d2,0xe9f7fecf98b1bdb6,0xffcf3ffb9af1afb2,0xbfbfcfdbdaf4bdf4,0xf7fff77bdafdfd76,0xdffe3ff39aecf766,0xfbf6ff7fdafbbdf6,0xfefeff7f9afdad46,0xffbfcff3dafdef76,0xfcfffffedaf7ef76,0xf7fffff5daf82f76,0xefbeff6fdaf1ffc4,0xeffffedfdaf1eff6,0xebffdfdbdabcfd70,0xf4fffffcdaf4ff76,0xffefff7fdaf6fdf4,0xfbbb3f7f9af1bd66,0xfdeffbbedaf4ffe6,0xbfbe9fdb9af22f74,0x3eff7de2daf7f770,0xfffffffedaf5fde4,0xdebe3fffdafaff76,0xc9d6f7e79ab7bdd6,0xbeff5e5e9af1f772,0xbebbffffdabefde2,0xfffff7ffdaf6fd56,0xfffe7fffdafffdf6,0xfbdfff7f9afffff6,0x6ffffffedafffff6,0xbbbeffdf9af3fff6,0xd4fe7e7b98e5bde6,0x7fde777bd8b8fd36,0x33b68777d0b2a526,0x9c8b1e49dad0af22,0xee3edff09af0af70,0xbc50aeec98e5bf46,0x96a9bd6512f4ad46,0xdfb2bf7fdae6bd66,0xcffffe6bdaf4fde2,0xebce87d7da51b546,0x76fe57efdaf4bd62,0xc366afe89060afa6,0xdaf72f7bda34bd36,0xefdfff7bdab4e5e2,0xef9edbdfdaf2e5f4,0xfbbf5fe7daf8ef66,0xdbe6ffffdae0ade4,0xfebe3fed92a7fd70,0x877b7b129ab03d54,0x2c76f75f9af0a554,0xb7a0d6fbd8a12d12,0xf5ffbf7bdaefbf66,0xfff63ff7dafb2d56,0xe9d4f31f9282ad40,0x4be61fff98e7bdf2,0xfe9f7ffbdafc3f62,0xf4f7fef6daf4b564,0x56e67d6392a42534,0xfbadfb4e9af46752,0xeffb7fff9affbd62,0xefadcff39ae2af04,0xf5f7ffeadaf0fff2,0xdbbfff71dab5bd66,0xfb9ef73e98f4bcf4,0x1be3bba390a43d76,0xbfff5fdedaadfdd6,0xf6fee7f0dae6b730,0xd76ab7689ad1bd06,0xdf9c7fffd8fcbff6,0xefffff7e9ae1fff6,0xe7fc5f59dae42f24,0xf5ff5df7daf5fde6,0xc93cf77a9a78a766,0xf3302b66d836e570,0x79defe22d8e1eff0,0x75fadbc1d8f4ad74,0xa67b5fe7daf5b5d6,0xb5ffff79da38ac76,0xf7fe3ffddaf7b574,0xfbffffbfdaf6fff4,0xffffffffdafffff6,0xfffffffbdafffff6,0xfefffe7fdaf4fdf6,0x5aa82d60d0602606,0xfbffffffd8f9ef56,0xf3c2374bd2e66534,0xffffdffbdafef5f4,0x54812dea9ae0a574,0x128ac1088404740,0xc989e72692b4bd60,0x200200002004,0x22d6cd4482706700,0xf7f5fc6d9ab4fd64,0xd1ee2c00d0a02406,0x808014892000404,0x10800b1088100000,0x2064c0682a42504,0x94b5243018c0b560,0x91a3a58592a0a542,0xa89a5d719a70dd00,0xa8078a299a80b562,0x1000020980402604,0x1001129090800040,0x810000000400400,0x2184090800000,0x8800110002200540,0x0,0x8282040090008110,0x8c14549092202544,0x40000000400,0x0,0x846109188c82600,0x0,0x48008400000,0x0,0x0,0x300080190000400,0x210c000080008800,0x48418400010,0x1088002000240400,0x0,0x0,0x0,0x250160088408900,0x0,0x4000001010000400,0x0,0x404c00a92002124,0x10090000000,0x108024018c00000,0x4c04520698c00000,0x910d70890222402,0x340c0002800100,0xb6ee4df398e9ad56,0xd7fff6fbdaf4fd76,0x7b767ff098b4fde2,0x9b56eb2e9a6fad74,0x884a520198c01570,0x8316d9e11024ad72,0x50563ae090b0b414,0x9b2235079030bc50,0xfb40ff2d90bc2400,0xeff7f60e90bead64,0x8d1c5b529ae0a500,0x84f676e3d2b1fd20,0xfd7f7fe2dae5bfe4,0xfa9e7f24dae0b700,0x8ddbcf1a9ae5b524,0x377f5fdbdaf7ed64,0xfbff7feb9ab5ede0,0xfffdfffb92f4bdf4,0xfef67fedd2f0bdf6,0xc79af71e92a83d32,0xa956ee4a9af0a560,0x45a66df49a262550,0xd2f93de0da70a710,0x42163c71d230a552,0xc803b2a292202500,0xc20a1090000520,0xc89a4da2dab02500,0x15d3b67a9a30f520,0x709e3c5192209d80,0x110e2208080a902,0xe6c3db88daf0ad64,0x2322ce909ac82d30,0x107a56e0d0802560,0x11aa3d628230b546,0xc9106f6edae53544,0xcdc7ee098af0ad46,0x75f5edfd8d0ed10,0xdefc76fd9873b7c2,0xddc6f530d2a42d74,0xd7bc7fd792f1bf50,0x41d5c52898f82c02,0xa441a4002602342,0x451000000242002,0x80652c4080a48c80,0x8000202010000030,0x8970c8039a70b400,0x856566098402160,0x16405dd14264b500,0x8118002080008504,0xbe363704d0e6e540,0xc8767f7e9a67a7e6,0x60920c02d0202120,0x80c6d40d0202514,0x81515e299ac0afd0,0x48960d309a622540,0x8121e9028031b960,0xc1da3ec198a8a594,0x1980c0000242104,0x71a702689a240d00,0x1000004010800000,0xa85e5229808a092,0x1800080000800,0x81080c039080a540,0x847172680a60af00,0x8c00438810303100,0xc8102e7282602700,0x1c7fe02d0a4b540,0x6254514b90e02766,0x84d47b22d022a904,0x4100801010208010,0x82a2054290202404,0xc163f6a19a326d60,0x62121042c8e40000,0x499c520c8782040,0x70458550d220b500,0xfa9e3f4e90b2ad72 },
	{ 0x8020e01429084e63,0x210190c20d442640,0xb09b4fe0fd287562,0xf6776ffe8dda6d72,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xb4ffffeeed7eff63,0xf5f6f7ff8ffe7f73,0x201000000000a000,0x4000408004000,0xeedbbfff3dde7d61,0x7bbccff9fb5eff62,0xb57e4eae0fdaef72,0xbffefdf3ffeeff62,0xfb9677fdddfe7f73,0x5fdffebfffeeff73,0xbfffdfdbfffeff73,0xfeffffffeffefd73,0xfffffffffffeff73,0xffbfffffdffeff73,0xddeffffeefdeff63,0x6fffdfd3dffeef73,0x34b76df64fe86f43,0xfbefff6beffe7f63,0xfeb76f660f7e7f61,0xefdfef376fde6763,0x3f7fffff4fdeff63,0xbfbbdffbcf7ef763,0xffddf7de9ddeff63,0xfb567febbe5e6d62,0xe213e24b49ca7160,0x6ff76f7fdfea6772,0xffffcffbffdeff62,0xfffff779658e7f63,0xfafe7dfb1f7eff63,0xfbd4ef264f5eff73,0xfffbfffbefdcf763,0xdfffcffb4fdaf771,0x55ffffefcfde7f63,0xfffe7f73effeff73,0xffaff775efcefd72,0xffffff7fffdeff63,0xefffdfdfdffeff73,0xfefffffbdf7eff63,0xcfeffff2dd7e6f63,0xff9a2f7ecf5e7f63,0xef9ffe3eefbaff63,0xfffffddfffdeff73,0xf5ff5feb0fda7753,0xf7fefffecbfeff62,0xdebeafffdf7a7f73,0xcfffff3f7f6a6f63,0xeaffdfffcfcefd63,0xf7f367fa4dee6f61,0xf7bcb77b9dce6f73,0xddfe7fff5fceff73,0xfbfffffffffeff73,0xfffffffffffeff73,0xffbedffb0f5e7f73,0xb7b6f2feeffc7f63,0xeffeff624f4e7f63,0xebc5ef17ad4ef760,0x8df3bf4de95a6f62,0x73be87120860ff62,0xd3c5c742054aec70,0x3faeff62adde6f73,0xfbb7ff7f5fea7773,0xcdfefe7fdfdc7f63,0x2bdf4ddf817a7763,0x1ef55eeeaf467763,0x9f6fb65a57aa6f63,0xff9c6f7dddcaf773,0xefdfff93ed3e6d60,0xa9da5d12cd5e6d63,0xb4ff77eb8dfe7f63,0x99ecff72976a7773,0xfeaeff7e7fee7d63,0x49cff77ecd7e6f73,0xa3ffdb5b39ae5f72,0xf4ba5ff2ed5e7f63,0xd5fdf7e9a1fafd62,0xfb767764ad6e6d62,0xabebfeffef3efd63,0xabfe3fefbfca6563,0xb7bbfff3af7eff63,0x94f7f7eccdfc6560,0x7f766db3edcae563,0xdbbcfb2f29fa7761,0xedcbff7f8fde6f63,0xefbedffb4fee7f73,0xd7fff6fbeffe7f63,0xdf3eff71cd3e7f73,0xe314fbff4b5c7762,0xedffbe1bbdfe7d62,0xefefdfde2fcc5d63,0xdcbec4f0ae4c6562,0x914aa7702dba6763,0xdeb46b7e8dee6d73,0xe5f6fb3bedde7f73,0x2ff7dd59c7de2f60,0x143fcef36f5ae562,0xc17db7f2fdce7f63,0xdcb6177ea79c6563,0x6edce7627848f563,0xa37b3bd35c46ff63,0xbfbf7f7a8d5e7f63,0xd3fffffdaf3ef763,0xdfb62ff6ed5efd63,0xfbffffbfeffeff73,0xfffffffffffeff73,0xffbfdfffffeeff73,0xfff6defffdea7f63,0x13ee6f614c1ae761,0xebc4f96a0d9c5b63,0xac5fbb134f4e6d73,0xffbfdffbeddeff73,0x74d1fefb0dfa6f53,0x4a3a2c620bca4622,0xfc84ff1569ca7463,0x1000020000000820,0x69dd99540cbc6142,0xf6ffffefdfdeff63,0xa06d272424ce4760,0x808d08480b064442,0x8120800000040421,0x1131658a1882d22,0x10824d0015020140,0x5000900000000740,0x3c950d54cc486560,0x2114118800080d62,0x36e10038d427b41,0x466005000940,0x2024000008400820,0x6080052005003100,0x80800002008a0040,0xa12020302004740,0x0,0x2890122012084700,0x4082044000880622,0xc894275200085663,0x30a0084002100,0x200020000000400,0x548008040000,0x108150004080560,0xc2102b1201124d33,0x8000200008040000,0x5859f553050e5c23,0x84848060094a0540,0x42304089026000,0x0,0x4002c00001001800,0x0,0x0,0x0,0x0,0x84200249881000,0x806200448004412,0x0,0x0,0x8a04112c00984160,0x82c0653424cae161,0x28c216d904dc4162,0x75b9cfdb01ce5673,0xb4e7f26fcf3aef62,0xf946de21fcee762,0xf2a8e30605084573,0xccc27e3dcdde6f72,0xe14c64c008a4960,0x1cf35ee68f1e6772,0xf5be3f63cd18f563,0x8995636f5f4c7f62,0xefcfffbbbf587773,0xaffedf1ecfdeff71,0x94b5d6e5cfca7c42,0xb63eb77109da7d62,0xf7bebf6b2d6e6d63,0x4cdeffa48d447571,0xebfddfdb9b5a7f72,0xfeffffffefdef763,0xfffffffe7ffeff73,0xfebf2f7fff4eff23,0x4ddb97209c407741,0xbbffffdfdffe7f73,0xffaa4dfb1fde7763,0x6030832009004e60,0xa1001b004204143,0x8044280080182160,0xc8c27e048aa86e73,0x24a3480000426500,0xc414126007100540,0xd4aa4d402cea0561,0xc208610008884c60,0x8c212174a080d60,0x1236065a4bfe4562,0xf7b7f46acd8a6543,0xbb7efff2df66ff71,0xf692ef6b59567f73,0xefeebfb7fd6e7772,0xafded7d21f5e4d23,0xe47474e84ccc7542,0x6bae8d622e48e761,0xcebd3f7e07def560,0xe042405004084760,0x14420004100140,0x10100201080900,0x40200004001,0x82820b40830a2400,0x7140128a040,0x2441630010a0550,0x92428108484540,0x109544800044560,0x52083d541c04a560,0xeaf5f3ef8dbe7f62,0x42c2122d4c586120,0x1899cdca494a4f60,0xb5a7f26e399aed61,0x4a0604200c080550,0x43c4e5320c886f61,0x6e22802484c5160,0x20c480020082020,0x24008040000,0xc9982442c0ca6541,0xdbc0677e4e5e7560,0x944168845006c32,0x31ad2d2c80e2722,0xf4932ea40fda6542,0x812002480008a430,0x2b02c00050e0160,0xec48de9c882c7473,0xa5a180901483320,0x34a0c50000342f60,0x800025020c40e100,0x4220041601084b40,0x4002802808004402,0x0,0x808010200084021,0xd410905549487220,0xc39e2dd0494a6d62 },
	{ 0x8000930803908000,0xc01f6810960d20,0xbfffffe2dfdefce9,0x7df6a6f656bf3ca0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xffff6feb7fff7ee0,0x7ef67e6a5a9e7ee0,0x200c0044020408,0xca90430201c02c40,0xffffbefffffeffa9,0xffffcf7adb9e7d69,0xa4d757e917933e68,0xbbfcfff3cbff7fe0,0xff9f7ff7fbfefe68,0xdffbffbfffffffe1,0xffffffdfffffffe9,0xf7fffeffdfffffe9,0xffffff7bffffffe8,0xffffffffffff7fe9,0x5bc7f7aa479e7c40,0xefffffdffbffffe8,0xfffedfffdfbfff69,0xde7efffedf9efea9,0xfefe2fff7ebfffa1,0xcddffffe5ffe7de9,0xaa5ededfdb976da8,0xdfc3dfbb5ffa7e68,0xd1efbf734ebe7c28,0xed1a7feecef73da0,0xe38cf32c4a262ca0,0xfff73f9fdfb7ffe1,0xffbfcffbdfb67ea9,0xddfff7fe5f9f7fa0,0xeffeff73ffbffca0,0xfb95ef7efefe2ea9,0xfeeffffffffe7ea0,0xeffecffbdf9eefa8,0xb5fff7efdb9e7da9,0xfffeff737f9f7e28,0xf9beffff66ff7fa9,0xcfffff7f5fffffa8,0xffffdfdfdfbf7fa1,0xfdfffeef7ffebee9,0xbf7cb77a5fdf7da0,0xdbb66ffe6f9f2ca8,0xeffeef9afebf6fa0,0xeffeffdbfffffee0,0xbffbdfefdfbfffa9,0xf9fefffedfdeffa8,0xfeb6bf77ffbe7fa0,0xebf7ffdfffff7fe8,0xcbffff7ffb9e3ee9,0x9fbfffff5bfebfe0,0xf3ffb7e351de7fa1,0xfffebd7fefffffe8,0xebfffffffffffee9,0xffffffffffffffe9,0xfffffffffffeffe9,0xddfff6ff5ffeffa0,0xfbaa7d61f7d6ec20,0xabb5f3d6de9e6c80,0x6e43377fcadf3e01,0x2a1bcccecad63d80,0x15e7f66cd6b6bda0,0xdbfe7f726f9f7ea1,0xfeb7fffeffdeffe1,0xdf7bfeffff9f7fe9,0xef5bdfd7ffbe6da9,0xf4b6fdfbdf9ebfa9,0xfbf5f77bdb94fea0,0xfa34ef7e5fbdbda8,0xef63ff1e5ebf3f60,0xebacdfdf73de7fa9,0x649ef3eb5296fee0,0xd5ae27f04fdd3ce0,0xfebe2f7fce9e7c60,0x8dc5e837ce9f3c20,0xeffbffffdefe6fe0,0xbf9b676a7ed66ea1,0xf779b6ff4e9fbfe0,0xfbfe2f7dfb9e7ce0,0xc3bbfa150bbe3e80,0xfffe7f7b7f9f3d81,0xfbbecf7a569fbde0,0xd4f77efefa973ea0,0xdbfaa5727e9ee680,0xf9fdfbbe7ffeef80,0xddffbf79d6bf7d80,0xffffdfff7fbebfe8,0xfff7feeedbfebf28,0xdbfe7e737b9e7ee0,0xdfa6ffdc579e7e01,0xcdf6febfdbde3da9,0xe272cc8fc3be7c80,0xd4dd7eefde9f7fa1,0xdbfeaf7bd7d67de0,0xffb4be77dbde7f81,0xfde6eb2beffe7da0,0x3bfeffdac79e7d61,0xffa25f6f1e9f6da0,0xf170664f1f9e2d20,0xffbe3fdfdf977fe8,0x8084f3324e366ca0,0xd946fffbc7b27ea0,0x9fb7cfffdfdf7f88,0xffffffffdfbffde0,0xdbf67dfecfff3fe0,0xebfefb2ffffe7ea0,0xffffffdfffbf7ee9,0xffbfdfffffffffe9,0xf5f6fbf9d79e3de1,0xeab675630a9e6ea0,0xcbd2c36a43863c08,0x48c2fec947963bc0,0xa7bfdff3dbdefc41,0xff1fe9c3be6f80,0xc1a4ff6052ae3d80,0xca85d50045943481,0x4a92da02e286e488,0xebffdddbfff6bfe8,0xd5bebffc5bb67680,0x5101004002902400,0x519c585240e02e80,0xe402f0a090fc4c80,0x1415d4850970ca0,0x1d069bc78a8daf80,0x91d6bf425e907480,0xa8121e040147080,0x182c041509e2d80,0xb04040080480,0xb0824f8030d64400,0x6000924081860800,0x4810001000100408,0x0,0x8042000200802041,0x2010000822020,0x8041226910840200,0x10041110820800,0x3950c77210bd2c61,0x5892a2850862260,0x20c4802ca03ac00,0x247156699b8b3900,0x20080000800200,0x0,0x50000010800000,0x8080a00820a00,0x8010000a000000,0xa04202082400,0x80284003080400,0x0,0x480000800040,0x0,0x0,0x0,0x4d0000000810840,0xc20a1900901420,0x80040002000400,0xc110a0124a002d80,0x774e77788797bc20,0x303e90a12ac3620,0x2fd71bdeec9e6ea8,0x1fbdfffb3b977ea0,0xf1f7d5fed7fe2ea0,0xe2ae75211f982d08,0xf880db3cc29e64e0,0xcfc6be036edc7d60,0x124e0210865600,0x456086f451862ea1,0x932a7d61d79e2ee1,0x8306fd2857b72c60,0x89d7d6b24e963f60,0x6fff5f41ca9efea0,0xf4f677666b9e2da1,0xfb6cfff0179676a0,0xc2b03e1363b27c20,0x4e433eba52d67620,0xaa3a17583bae7da0,0xd47b7fefdebfbc20,0xfdced7dc9a9eace0,0xf6b22f64569ea981,0xe158f5e8569ffda8,0x871c7a004b82aea0,0xbef7eff253bf7ee8,0xd175fee95bd66f28,0x805041100b902000,0x4958d00266163d00,0x84e1600419c04c0,0x88b0868801962061,0xd085b66992962920,0xa8604425d930c20,0x8b004120000d2880,0xa5403a071f141900,0xa6ffcf7233be3e20,0x75e55efacf963fa0,0xc5920f7062d66ca0,0xe8a967ff59963fe8,0xcdfff619dfbe7ca0,0x2bf61fd63bfe7ee8,0xf4bf5ff4df9a3f21,0x4b0efd2102963c20,0xc3953d674e3a7c21,0x2002000900000,0x9ec0940400b0d80,0x20b3626a4e167c00,0x990c0528820d2c88,0x8280058100062d20,0x8044bf260e942d28,0x10034040002c00,0x4004fb006322c00,0x0,0x286bc40163cbd01,0x89a4f20a82fe2d20,0x2cc20aa1008a3c20,0x8b9cdfd144b63e01,0xb467f7d843be7f00,0x40885040128c2620,0x80e4c300029a3c20,0xc0e3ba10de8c0c00,0x2000008040000400,0x22026110802800,0x50207502c6100480,0xc9beef6f5fd63f20,0x84c3380000903d40,0x7086811214123c20,0x90284f62169c3c01,0x800740102826c00,0x888073200e9c0400,0xeb46a204cb1a3ca0,0x38c1d821b882c29,0x1010140004100000,0x44020840040080,0xa11406104206a880,0x80f13a433e6420,0xa00004042000480,0x290052010960080,0x2004000800200,0xd2025952229e3e20 },
	{ 0x8013198094224a,0xbffa5fc8a3146b53,0x9fbdcffed7bc67ff,0xf7efb67ab7942bfe,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x480000040200,0xbbadefebb7b4276e,0xb530f77c961c2b65,0xffbebffbb39e6f7d,0x600082004004,0xeffbbffff7be6fff,0x7fbd5ffad6a6275f,0xfdff7ffed79c2f5e,0xfbfeff73d7b667ff,0xd60caba0c6a4635f,0xeffffebff7be6fff,0xffffffdff7be6fff,0xfffffffff7be6fff,0xfbffff7fd7366b7f,0xffb7bffff7be677f,0xef97ffbaf7aa236e,0xaa5f5f5fd7b66747,0xbebf7ffff6b46fff,0xd9f6f77ef79c67ff,0xfebe3ffff7b46f7f,0xedfffb3ff6be677f,0xbffffbdbf696635c,0x9fbb4ff2f7b423ff,0xf93dfcf9d73e2726,0xda0e3f62d68623ee,0x88c7fb76d7042363,0xeff7bbdbf784277f,0xbfffdd5bd7b66bfe,0xfdf7bff7d79667fe,0xdfbe7d79f7b42b77,0xfbfeff3ef79e67fe,0xee6ff7dfd73c6f7f,0x7fbedffbf6bc67ff,0xfbf7f7eff7bc23fe,0xfbfe7d77f7bc67fe,0xf89ee77ff7b64b7e,0xeffffebfd79667ff,0xff7fdfdff7b66fff,0xf5feffefd79467f7,0xdd7f7ff3f6b46fdf,0x9fb67bfbd2b62fde,0xcddfff2e973e23d7,0xfbfeffdbf7b46b7f,0xdebfdffff6b4635e,0xf7ffff7cd7be2ffe,0xfebf3ffff7be6f7f,0xffffff3fd7bc6f7f,0xaffeff8ef7bc2fff,0xbbbb7feff69e2fff,0xf9fd36faf6b66bef,0xe7defffad7b26b7e,0xfbfffffff7be6fff,0xfffffffff7be6fff,0xfbbffffff6b46b7f,0xdd7dfe7ad69c6f5e,0x5e24ad4092302f44,0xe990e32e93046f2e,0xb2fa9e11d7346a20,0x2592d903930a23ce,0xfdffb7ffd2b66fef,0xcd2efc73f7362b7f,0xff9f7ffff7b66fff,0xdfdffebff7b4677f,0x63ca9fdfd7342b76,0xdcb776ecd6b6274e,0xb7a8b749f1342f7e,0xdb94ffe4d78422d7,0xaef7f793f70627dc,0xab525e17b2326b67,0x24b23f6bb2a06f4f,0xdbedbec496b60bc6,0xeebf3ff1d6946756,0xe19af39ef2b42356,0x976bedbf616237c,0xfab3fdbaf6b263df,0xf1f57feed734275e,0xbfec3f7fd69e2f5f,0xebdcfbbcb684234d,0x6fd7be19f61a2b6f,0x9fadcf9bf5a42f62,0xf4f7f7fff7366fef,0x134c2d60d6042765,0xebe7e76bd7362742,0x7fdfffdff71c6fff,0xffbecfffd6be2fef,0xddfefeeff790676f,0xebbeff70f5942f76,0xfb8777eff794636f,0xca32feaad414235e,0x3be64d48d6006b44,0x74d6eeecf6a82bc6,0xd9ae3df7f7104b64,0xdab67fffb69e4b6e,0xeccefb6dd4b02773,0xebffd7dbf6a66f5d,0xb6b7dffbd786675e,0x99ceff77d78a234f,0x7a9a2f4ef4942746,0x8fd5b624d5a42a4e,0x2fffbec9d6be63cf,0xbf9fff7bd7b4636c,0xb5cdd56bd0bc6b7e,0x7ffa2f52f6946f7e,0xc9feffeed7be2ffe,0xeffffffff7be6fff,0xbffffffbf78e6fff,0x75f6f6ffd7bc2fde,0x53aeac6193a8235e,0x98f1cb3ed710275e,0xedcfb15ab3a4276e,0x7f080b39b38c274f,0x84bf0278b3144316,0xb96f7e73b2bc0b78,0xfe98bdfef716615e,0x5fcbfe2bd6342f6e,0xef7fdfdff7b46f7e,0x642c4ad0d6102350,0xd37f2f63c2126716,0x9a8c436ec5842b42,0x4688c62093000292,0x80e0fd0c2146202,0x420000040200,0x4860ce62d7002342,0x848043081900292,0x8003804180002b40,0x148080082202041,0x9292024890a02a42,0xfa741c0191b00252,0x88200d4080042040,0xc2045766f2042671,0x4de20141c3802102,0x8804180002400,0x140010200040,0xa200d0110002440,0x8004000080000008,0x0,0x44210200040,0x0,0x40200080002000,0x0,0x4044001091000200,0x18038480104100,0x240411600600004c,0x1200a18081002302,0x30060010000300,0x8400000002001,0x0,0x1001000010000020,0x0,0x0,0x2000800202000,0x800180000040,0x20204800200034c,0x100052004000048,0x100048400080004c,0x8902e938d6002158,0xa652320fc680245e,0x8338cf51d2a62746,0x94fd777c929027ef,0xe2627d42d6002334,0xa9d3772fd3386b5f,0xeaf3ba8fd6a4225d,0xcb3dcfded3942fd7,0x305326e996a4214e,0xdbbbe766f294275c,0xfffeff77f79c2fff,0xfddefe33f7be6f5f,0xb9265992d3860366,0xbcd77e6af6be6f7f,0x9f38ff42b2142f47,0xca3077f656842f66,0xcfe2f3b4d598234f,0xa2dddf5cd7342b55,0xbdfa5fffb21a2fcd,0xfdfcbfeeb3bc2f7f,0xc2a73fe6f7342b4e,0x8590a313d3b02257,0x1008020180000050,0x361a6c2252882356,0x50a0120cb4080348,0x201401208000620c,0x8112120010200340,0x3527840d4a022d8,0x39a6f7ab3842f66,0x8450023491102020,0x43222f6283100316,0x8b040464e6040216,0x2c5e224ac6900254,0x39ecfd7d70e4b5e,0xbd77fe6bf3bc2bd6,0xcb37fd72f7302f47,0xf7107771d6380a4e,0x4ecedf13f69023ff,0xeadecf5fd4b4276f,0x5420066993046107,0x37826de093942346,0xdaa82f61d2042307,0xc800c11392044160,0x800010080040000,0x2cb356888084260e,0x9018a07453a00146,0x8a024098004020e,0xab57f02892b442d4,0x108024044800040,0x8012408190002048,0x1084550083242202,0x91040820c200001c,0xe986ff2bf6982736,0xa0420650c004220c,0x1a2bcf9bd6a42b56,0xd7feb6efc7b427c7,0x80c050070082200,0xc202b44080900b48,0x67d27ecff7b82f17,0x0,0x11202000900240,0x2100010000200,0xc110432cd610011e,0xcb42b2a2f0a02d5b,0xa000c80290800042,0xd2148802800106,0x31c26752b10c4336,0xab9c6f3fd736631a,0x4282c20002102700,0x3471040908c0262,0xb04028a0004010,0x10402646b0000b10,0x90042530c4046602,0xe4a3f114f71423a7,0x1029880a2000104,0x84a2004090006205,0x2002024012000020,0x1000a4042020002 },
	{ 0xcbd7c35b890473ea,0x27447a5a8ba852dc,0xcfbdefcbfbf9fffe,0xf57ff6ecdbfd53ef,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6af77fdffbb4d73f,0xdfbedfdbfbfe7fff,0xdd70aa70d9b87b78,0xfe3e7df3cbfb5bfb,0xbbe57f7fdb2557ee,0xcccffebbcbfa5fef,0xefbfcff2fbc47ffe,0xf5b67eee8bdd736f,0xdbfaff71f9b2d7bf,0xdfaeff7ecbfc57bc,0xffffffbffbffffff,0xffffffdffbffffff,0xf7ff7ffffbffffff,0xfbffef73dbfd77fe,0xdb367f7e7bf8777f,0xedffffffebec77af,0xefffffdbfbf7dbfb,0xb7ffffebdbfbffbf,0xd9fdff7ef9b5f7af,0xfefe6fffebff73ff,0xffffff7ffbf8fbff,0xffdfff4ecbfa57aa,0xb6bbfffbebb5ff3e,0xffedbff8cbfcdfbd,0xdee67f67dbf579ef,0xfbdeeb4f8bffc3ff,0x7ffebbfbfbf17b6b,0xbfefcffbfbf1d7bf,0xfff7f6faeba94fbf,0xbfbefff3cbfaffbe,0xfbfdff3fcbff7fff,0xeefbbffbcbfe777e,0xffbffffedbfcd3ff,0xf5ffdeeff9bfffff,0x9ffe37f2fbd87b7e,0xff36ffebf9f673fe,0x4fdfffbffbf67fff,0xff7fdfdffbfff3ff,0xf7f7fefffbb75ffe,0xddeffff3fbe1ff3e,0xffbf7f7cfbfc5ffb,0xcfd7ff7ffbff53be,0xcbfeffdefb91f3ff,0xf0bff5fbdbb867bd,0xfb7faf76fbfdd3ff,0xfe9e3f7fe9f5533f,0xffffff3cebfe57bf,0xadffdf5bfbf3f7af,0xbbba6797db75d33e,0xf7fca7d5cbf457b6,0xffb67ffdc3b6fbbf,0xcbfefffffbfffbff,0xffff7ffffbffffff,0xbfbfeffefbdf7bff,0xf4f6fefffbfd5bfb,0x468633608b9a1329,0xdb56f72cdbb9772f,0x5cc7fb11e3a0f634,0x6a8c85989e05321,0x74fd3ffcfbbcd3be,0xbafc7d55dbdd7727,0xff9effbffbb67ffb,0xcfdffffbdbfefffc,0x8a1ecdd78ab9777e,0xf7b7ff6df9fd737a,0xf9eebf30d9e3577a,0xdebc2f7e5b9247e6,0xcbd7faa6e9b1536e,0x6ff65b9bf9dc537c,0xd5defb6bebb44b6e,0xf3fcff74eb96dffc,0xfeb52ffef9f273fd,0xefdff727dbfed3eb,0xaffe5ec7e9dc73af,0xfeba777efbb4d3fe,0xf5eff7e7eb94d32f,0xd9562d7069a6f265,0xc994f2ffd95d456e,0xebdf5fd1fbf6f17c,0xb29ccedfc9b47332,0xdcf7feffebfe737f,0x6fae7c61dbb453fc,0xdbe6cf5cd9f073a7,0xee6bbefbf2a4ffff,0xbf3fdf52db9473ae,0xfddffefffbf775bf,0x9d3aef60e96cd32c,0xab40655f8bcc633a,0xc9fef6bffbe45fee,0x2a76db8d89f80363,0x54fef6e8c9ec5369,0xd5ca3d72c8e41374,0xd29c7e06c997333e,0xaf8eff8afbf05b08,0x696f5dcbc9f072e7,0x3c3779f309e1637e,0x7d6cff72f9b9fb3e,0xdedc2fd45bd557aa,0x8defe73acba0d3b7,0xe955d74feb827963,0x23284ff2c3805b2d,0xf7fdaf7febf87bf9,0xfef46dfffbb953fa,0xc3ffffffcbbaf3bd,0xffdf3ff3fbf577ff,0xbfbf7df7ebf75fbf,0xf7fffefefbfc7fff,0x32aebf61899853af,0x9bd0493ee8ac5221,0x4c4313a84bc85312,0x8c25056a8919032e,0x90802ee8dae8510c,0xc96c2a0083b0130c,0xe0b4e3028bd80a00,0x4d82b380518440f1,0x959c6d888e4132f,0x9482162401044764,0x8940d430e8e8130d,0x899047e0c100580c,0x10250000000,0x21010485b004000,0x10b944e003924268,0x96082400d1001040,0x9082110261800118,0x184120409504000,0x800200080200000,0x8222002849004040,0x8010055048801201,0x202041051000060,0x200100001200,0x66c603c9c9c0710e,0x20805080001120,0x20204101044020,0x220020214801130a,0x20008000200,0x0,0x0,0x40040001000008,0x8061204001906000,0x8100200003004008,0x0,0x800080001000008,0x20040001000020,0x2003200b004000,0x89b3337f88b47bfc,0x4201420080001328,0x10024480141020,0x8140020c0101102,0x9b18521389d0431e,0x200200040000000,0x8180920501100328,0x80202018124c028,0x500cd9b619042a1,0x4200042000200200,0xa047180086442a0,0x8000f122e904c32c,0xbeebbfd3dbf07fa9,0x1189015a09944320,0x92f7a757eac97f37,0x11387d2288b946b2,0x99d4af6dfb59534e,0x6deb38b3ebb45b35,0xbf9f4ddbcbf0dbfc,0x71121eefe130636e,0xd9bef674dbf877f9,0xef9ff77febfff72f,0x495bff13fbff73fa,0xefdccfcbdb907ffa,0xd471a6f48b7f73a2,0xb34037d3eb837a0f,0xcba23fdd8b805337,0xc42bc639e9b047bf,0xeafad35beb95d33a,0x95fb5ffbdbf27b7d,0xd3ccaa604b99f20a,0xfe803d6e8b81dfb3,0xc7c2e61edbd01256,0x463304980b0132d,0x2924e4009b80338,0x91b4166b88c10f3a,0x24426402006520,0xc9f6e60afb38436e,0x66985fd2c950432e,0xacde55f8fbf152a2,0x52642ec8ba07770,0x2d7e7d63dbb4737b,0x184e02201806159,0x820818a104200,0xfdf8c733e9f0730c,0x65fef76cdba05f28,0xffb9b741dbe8d332,0xcbf6f36ffbfa7bdd,0xcfefff7efbfff7b7,0xab9ecfdbebfbe67f,0x241166841305200,0x81ce57408b644228,0xdb9e7f37fbd45b38,0x6140b08082105110,0x7b3ccddafb94537d,0xaba456a0b146228,0x91200c4489200330,0x928c2b0080a00b18,0x50c20880201200,0x8188cac2010043e8,0x1a13031b09206268,0x8008140000800020,0xd5005d66c9124b24,0xca91c312e1b01329,0x88d61fd109d5f223,0x8c87d7e1ebb9d13e,0xf1f5ff7b8bf9e33a,0x5b2a0f028bd0236a,0xc3126420408470a1,0xc8cb9e67eb84562b,0x2110040001204200,0xc0105000c0301000,0x883a6541c1301220,0x49226612d0842a2c,0x84f22290ca00510c,0x8504d4080a00121,0x588061100102,0x9222a50088c44d20,0xcb903f3cc97853ee,0x4481221251060128,0xee02dc538b6a4304,0x24c3d76a49a072a9,0x3204a660ca00610c,0x9602084062840229,0x8983b0048b980228,0xc0e2d94261a41631,0xf23a4fc3595a438a,0x50283610c3b10002,0x800084201800060 },
	{ 0xeac5f2e589cdd22e,0x24ce1ff281ced346,0xb7ffdfffcbd6fbfe,0xfcb6f3ea8bc5e35a,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x85bfa8083c6916a,0xffffffffebdffbfe,0xfbbecf538bd7cb6a,0x3dc0927983d6f972,0x796e77f3cac7c34e,0xfbddff3fabddfb6a,0xeacfbfdb8bdfc33e,0xbf9acf788ad7e14e,0xf5dfdfefcbd7e36a,0xdfbffff3ebcdc33e,0xfbdfffffc3d7db6e,0xfffffeffebdffbfe,0xffffffdfebdffbfe,0xffffffffcbd5d3fe,0xbdffef7fcbdffb7a,0xefbd7fffa9cff3ea,0xfdebff3eebdffbfe,0xffdfdf5f83dfda7a,0xfcbf6ffbc9d7e37e,0xff7ffffaebcfebfe,0xfeb6bffecbdfd3ea,0xefdeffffe3cfe3fe,0x7f7affd583d5f33a,0x3dfffdfeabcfe1ce,0xffffffffabddf37e,0xffd6ffe6abd5d3ee,0xebfeff6febdff33a,0xfeffff7ccbc5f33e,0xbfbfcf7bcbd7e3fa,0xf7fff7ffebc7f2fa,0xfffeffebebdffbbe,0xfbfefffeebdfebfe,0xcffbbffbcbd7fbee,0xfffddfffcbdffbee,0x54f6f7ffebcffb6a,0xcffefff3ebd5d36a,0xffbeffffcbcffbfe,0xcff7ffdfebd7d36a,0xff7ffddecbddeb7e,0xdcfedeefebcdc37a,0xfffe77f3ebc7c3ee,0xdfbf3f7fcbcdd3fa,0xffffdf9aebdfd3ea,0xead6dedbebd5f3ea,0xbefedfebcb5efb3a,0xdfffff7febcff36a,0xdea53f7fe1defb7a,0xedffdfebebd7d3fa,0xaeffffedebc5cb8a,0xaf9acf3a42c6c32a,0xf7fd7f79a3cff96e,0xffbe3fffebddf3ea,0xebffffffebddf3fe,0xefffbfdfcbdffbfe,0xebfffebcbd5d366,0xdcf3fefb89d6f3fa,0x57be5d7283c6e3f6,0xdb14ef2c89dfebe2,0x62cfb6096985c16a,0x2fbec97a4147c36a,0x22056eac1d0e30a,0xffffff73abd4c1aa,0xe7b9f7ffcbd5c3fa,0xffcfffbfcbc7fb6a,0x1b7e5f5bcbc5d1fa,0xf57a7af589d6f39a,0xffbef772abc5db3a,0x9b247776cb44935e,0x69dffb6ee3c5f3ea,0xaffedf5bc3d5e36e,0xffffdfffebd7f36e,0xdbfdfedccbc7e36a,0xca9d2ffecbd7f37a,0xe75bffbe8bd5c36e,0x8f5f5f4ac3c4d33e,0xbeb55dfecaded37e,0xd6e535f383d5f06e,0xdbee7f73e9d5eb2e,0xdb81d20f63d5c24a,0x6f46065a42d7836a,0xbcbacd6b0bc5c22a,0xfcfdd3facbd6f25e,0xffbe7ff0a3c5d36a,0xfbc4f79ee9c5fb6a,0xaed9fac7abd4f16e,0xefbe4bbe8bd7f30a,0xbfffbeebebd5f32e,0xcfaefd732bddc33a,0xd7be7f33a8c7d02e,0xcb4cbe85c0cdca6e,0x8afc9fdb89d6d3be,0xd4b9e6e2a384eaac,0x90cbf79c905d10a,0xcb90767fa0c5c33e,0xcac7ea00ab86ca32,0xaff75ccf83d6d346,0xd6bbdaebe1d7ca4a,0xdb6f377bebd7c3fe,0xceb67d72a985e23a,0xef76e77acbc5fbca,0x7bfedffbe3d7ea6a,0x339afff2eac5c11a,0x93d2a7d2c9c7a25a,0xff86f77eebddcb6a,0x8bd6fb3bebdfd3fa,0x6b5fbffbc3d5f3fe,0x1fbccffb63cdc3ea,0xf7f7f7ffebddf37e,0x5bee6de24bdcc36a,0x8a0283228250b020,0x8febaa5aeb0de30a,0xe21e4f0acb85d14a,0x50801c8003540082,0xd1702a4000558208,0x421261244981c08a,0x4a43ba908205f116,0x71e4d9768c5c202,0x4a3562400c18028,0xc0a6b6608140e12a,0xdf9e3f58e9c5fbe6,0x6c033120bc4411a,0x2106dcd00041812a,0x905e410350434a,0x5342067a8350ea40,0xa0344f00018122,0xc100881040018100,0xdc33e488bcdc87a,0x22082e98003a142,0x940835380805c12a,0x3000280000440002,0x851d1aa8a00b322,0x80580a4700848228,0x1498859309d58262,0xc621667d0bd5c20a,0x70ce74602384c132,0xca02616e23504312,0x4ccb1e344354c092,0x15046000444008,0x5000000000040002,0xcd4cf52282c7fb62,0x1000080c8080a000,0xa1c4ca2103d4020a,0xebc60bd3abc4c39e,0x401000000808000,0x480805202881c002,0x0,0x500206000800800a,0x204080208440008,0x4000102100540200,0x804000000102,0x0,0x40a006080404808,0x0,0x18600008058000,0x1100050082001002,0x640001f88a04a822,0x100001001004000,0xae7fbed9abcfe34a,0x75bb63fa83c4d31a,0xdec3f67f8ac5937a,0x5a3e7c33c3c6fb4e,0x89f5c70ecb84e362,0xe83bb028287c322,0xad9edf52ebdce36a,0xb4d57efa49cec37a,0xf7bf7f76a3dfdbea,0xfeb2ff6fcbd5db7a,0x6fc7ae8b8bd6f33a,0xfe7edfdbe9d7e2ea,0xb4f6dfed09d4c37a,0x9bcdfb7acbc5c22a,0xf336fb7cc3cffbfa,0xc1cffe9bc1cfe16e,0x9ded5d3cad5d32a,0x9cb67d99e9c4c366,0x73eeb76ac1c7c2da,0x9faa3f478b07e27a,0xc9f3a03603c7832a,0x100607408050080a,0x81364ce28154c33a,0x0,0xa8366472e285d122,0xc002a0200a440002,0xc5d33652c104f252,0x1283459e81c5ca2a,0x472ffe6a8ac4c11a,0x922c0d82830fe192,0x200400800008108,0xd40323140c1817a,0xf9264d24c244d34a,0x9cf77fec8bc5cb5e,0xd8aef567c3c4fb3a,0xdb9eefbae9c7f36e,0x8fd7fe2febdde35a,0x6bff5fd78bd5d34e,0x10a404488080d002,0xcb4d6400815cf222,0xdbfcaf7e83d7d25a,0x40d2f3298144e312,0x27fe5fdb8bdcf32a,0xb8020d938105c00a,0xa51a0a4088c4c816,0x50100408240,0x5300e31c83c4c30e,0x208001100c4801a,0x8080300038040,0x40001004000,0x48023c30080bc072,0xdb00f32ac8c1c00e,0x5b5f1e534bc5c30a,0x4aaccfe3ebd5c36a,0xcd5754c1e005e942,0x490008000104000a,0xc0c1280105c302,0x8cc3ea0041d1c132,0x0,0x5584305ac887c152,0x938c9771c14dc37a,0x10230581405802,0x44ac2100103c1e0,0x1008014001008208,0x4a801024040,0x90c35200184c302,0xda200c42e2c1c82a,0xbc8008080a100,0x2a8ecd4a8984e322,0x9c90f5ab82c4f30a,0x90100e0009008202,0x4e8826448145c052,0x0,0x75e5e5888c6cb4a,0x93b27f7281c6f3ce,0x8200480008104,0x76b2246083069332 },
	{ 0xc082401b47020848,0x6800060440828600,0xe7fffff2cfdb7eee,0xc122a2a806814e00,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe3d7f6eecfcb7efc,0x7effbe9f4fdbfeee,0xb8badf7bcec3ffe4,0xf4efb369cdcb5ee4,0xefb67f73cedb7e7e,0xfbf7ff7fcfdbfef4,0xcfdffe7fcedb37b4,0x6fbfcfdbcfdbfefc,0xf5ffffefcfd3fff8,0xbfb8aff1cfcb7eac,0xfbbffffecfcbfefe,0xfffffffb4fdb7ffe,0xffffdfdfcfdbfffe,0xbcfffefccfdbfffc,0xfff77f7acfdbffe0,0xffbe7effc6dbff36,0xffffffffcfdbffbe,0xffffff5fcfdbfff4,0xffbfffef4fdbfeb4,0xffffffffcfdbfeb8,0xfabe2fff4ec37fb0,0xcf8fffbb4dc3beb4,0x2efff9d7c4d3fee2,0x9d3b777bcecbfeb0,0xffffffffcfd3feb4,0xffbe7ff94ddb7ebe,0xeb4effffcfdbffb0,0xeffbbfdfcfdb7af4,0xb7bfffbbc7cbfefe,0xfdfef47dc7dbfeee,0xfffe7ffbcfdbfefc,0xdbdfff7fcfdbfffe,0x8fdfffdfcfcb7efc,0xfbbfcffbcdcbfba0,0xfff7feffcfcbfbf6,0xfbffff77ced3fefe,0xdfbfffffcddb7eee,0xdfdffef9cfdbffb6,0xeb5e9f5b47c37aa4,0xd6bffefccfcb7bf0,0xfbff37fb4fcbffe6,0xffbeff7f4fdbff2e,0xdfd7df3fc79b3ebc,0xffdfffdfcfdbfeb6,0xbd7ffded4ddb7ee0,0xd143ff58cfd3bf22,0xdeb62f72ccc27ee2,0xc9fedbf6cfda7eba,0xfdfeff4f4edbfee0,0xbf3a7fefcec36eb4,0xd2e7f7ffcdd2ff74,0xffbe7feecfdb7eb4,0xebfff7f7cfdbfaa0,0xffffffff4fdbfffa,0xdebfdfebcfda7f24,0x96d696f04d9afe10,0xfffe6ff24fd2fef4,0xfb187756c0dbbf64,0x4f4b22eb4cda5284,0x4e775f704fcafec6,0xf6d55fe74ed3bef0,0xdbfeff724fd2febc,0xffd2ff7f87db3ee0,0xdfdebebbcfcbffac,0xaf4edac2c7d3bfe2,0x50fe56e8cdc1da18,0xfd6ffff6cfdbbeb4,0x9b181f7e86db7ad2,0xaff7fffacfdbfff8,0xcbdbd7c7cecb7aac,0x9dbf5ffdcfdbfea0,0xf9ecbe784fcb7e24,0xfebfbf7fcfcb7fe0,0xe9fcfab68fdb3ea0,0x1ddedede4fd87e28,0xffffdfffcfdbfffe,0xf557f76fcddafe86,0x7a8636a64cc8d6b4,0xabc588588f88faa0,0x4ec7be0b0c8af622,0x9db9cfeeccd2fe22,0xf97ff97fccdbfeb0,0xff2eede2c7dbfaa6,0xdbd6efaecfcbbfba,0xacdfffffcdcb7fb4,0xafffcf6a4fd37eb0,0xfdf7f7facfcb7f30,0xfffefff3cfcb7ebe,0xdbbccf7e4fc37ea0,0x4dcfbfb547db7ae4,0x9bffcf5b4fdafec0,0xd437f4ed0fc36a72,0x91e3b76247cb7ea4,0x9a8c667e47cafa6a,0xcdd7fe37cedafff6,0xebdfde5fced3feec,0x5c977feb46db76a6,0xd16ca7e04f8b2fa0,0xfa982f624ec3feb8,0xe3cbfff78fc37ea0,0x22d113db45dafaa4,0xb21f4ffbcfcafe82,0x58c4a7400cc23a42,0xdf1e3d62c6d37aac,0xc9b7fbbf4fd37a8c,0xcefb3fdf4f1b7fa4,0x9f3ddffac7d3faf6,0xdfdbf7e8cfdbfeb6,0x5a7275728edb7b20,0xe993ff2a8ed27aa2,0x8c18928b84825220,0xde7757f605c3feb8,0x11ee0400c4822030,0x1236091044c02830,0x6808640944c25e20,0xc419b62041d33860,0x47921f124cc32604,0x84b966604d822e34,0x1048ca60c98a9a04,0xd79c633c0dcb9e00,0x800020440022080,0xe12f90141cb9a30,0x44a2109400a00,0x9700664040827112,0x529e3feec7cafaac,0xcf50f01a8188be02,0xb58024202c1f282,0x91bac7cb4783beb0,0x2173276b0dc9fa00,0xc4122db2c4c21e04,0x81b3f7774dcb3fb2,0xc86020000407080,0x410440001400a00,0x810000020200,0x14000002080,0x800202804881000,0x42220040000410,0x403c04a04382ca00,0x844808081200,0x8062400020222c0,0x100400000101000,0x602200008010a00,0xa80014001485800,0x404000040000400,0x26004005002880,0x20404800000,0x40210804420200,0x0,0x0,0x0,0x4000200001000800,0x8000300284802000,0x273100046800604,0x8092048a44828044,0x98c10c6047807800,0xaa8444040410634,0x801810000410800,0x2e50100804404a08,0xbeef5da8c48abb64,0xbd9d74ef4fcbbea0,0x7b647d71cfcbfab4,0xfbd5f97acfda7aec,0x7fdbbbebcfdbfba4,0xbf7f5f2dcfc26fa0,0xdcfbf7ffcecbfeb4,0xfbfe7d73cedbfff4,0xfa8effffcfd37f7a,0xdfefbe174fcb7ea4,0x2a7f1fca4fcbbaae,0x549c94f4c593faa8,0xd9eebf534dcb3f40,0xdaa475a70f8b3aa0,0x4dc0821846901288,0x2bfe5f894dd3feb0,0x15ac22e0468a6e50,0xd9082f6a4ec3bf28,0xee84277f4dc9bea0,0xcf92fb124ec3f2bc,0xb5a7b408ec93aa8,0xb201050840c81610,0x49645242412600,0xce080d0406022620,0x88c5f32d45c3ff0c,0xef621ab8cdcb3b8c,0x11847a8068a3222,0xb5d676edcac3bbb0,0xc2360d104fd27a20,0x8b18ea2f4ec132b0,0x8001120904c17080,0x1a16cc2906c27e28,0x211656ec46d33e40,0x6fb6ff70ceca7e34,0xeb8fd7bdcf8bfae2,0xcfdfffd74fcb7ff4,0x131adfdb0ecb7e40,0xf4225ee4478b38a2,0x968f2424dc3bfa0,0x8b966f3f43cbfea0,0x40861b0e06c0f810,0x8e204e0004123600,0x1656d703ce822310,0x4900f4438481a620,0xc828040248420204,0x538cd21186c27b00,0x421001040a1000,0x1641f20d825a00,0x8058773206923142,0x4008010200404810,0xd160a32f0cc232a0,0xa812160106ca2a90,0x4c0b5ff34f82ae60,0xb4f576614dcbfe44,0x7beef522cfc3ee30,0xc320c10a40c14a20,0x44e220802402222,0x401801000085000,0x75fd36eb4fca7ffe,0x9faabf774edbff3c,0xcb88762d4cc236b0,0x4343b22304c32e20,0xa9d9bd3c6c12e80,0x41c482044820000,0x8d45ee0846537aa4,0x9a862c4048c1be20,0x98c1d282004222a0,0x2b54cec00788ba28,0x88ac923405b3a84,0xc50cb670c6826ea0,0x9abe64704dca3aa0,0xd80cdb0c07c8aa80,0xc9559b4ac0c3faa0,0xb5faefef4cd37ea0,0x18d03440448a7804,0xc324f563ccdb3f30 },
	{ 0x8810120020021088,0xc5517cc8a721a87,0xfffffffffa76bbbf,0xff7ffffffa76bbbf,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xebfff3ffea729a9f,0x7f7f3ffbea57ba9f,0xf77eca76a25612ab,0xf57ffe7aea7233bb,0xfbbe7f73fa76bbbf,0xfbfef73ef873ba97,0x5edbbfdb2a773bbf,0xeb9ecffbe872b39f,0xd5fffffefa77babf,0xdfbbff77fa57babf,0xffbfff7ffa73bbbb,0xffffbfffba73babf,0xfbffdfdfea573b9f,0xf6fdfffbea73bbbf,0xbbaef77bea733a9f,0xcbbe7f7fa872b3bb,0xcfdfff3fea73bbbf,0xabfffbdfea77bbbf,0xbfffdfffba53bb9b,0xddeefffbfa76baaf,0xffbf2f76ba563bab,0xcfdff7f3ba72bb8f,0xbfdfffdfea729bbf,0xfebfeff9ea563b8f,0xfbffff6be876bbaf,0xfffeff7eea52bbbf,0xebf6fbbfba773bbb,0x4c5f7fdfea76bb8f,0xbfdfeff3fa72bbbf,0xddf6f7ffea77bbaf,0xfffebff3fa77bbbf,0xebfff76ffa73bbbf,0xeefffffbfa77bbbf,0x97fadfdbaa721b8f,0xfffff7fefa53bbbf,0xfff7ff72fa56bbbf,0xefbeff7fea72ba8f,0xfdbbe11aa73ba9b,0xbfffdfdbfa529a9f,0xb6f5ffe9e877bbbf,0xfb6fff7ff8763b8f,0xdff4fbff2a733bbf,0xeff7ffbeba77bb8f,0xefff5f5fea77b3ab,0x95d7ff7ffa73baaf,0xf97caf73fa773a9f,0xfefe3dfeba769bbf,0xcfdbff7baa73bb9f,0x6fffde5b6a569abb,0xbdbaffffaa429a9f,0xd7fcff7faa773baf,0xfffeff7fba73bbaf,0xfbf7fffffa77bbbf,0xfeff7ff1fa5733bf,0xf7befffb6a773bbb,0x9673f7307a723283,0xde367dd3ea563a8b,0x9bf4ce36f852338b,0xf44b1e19a8621a87,0x2c8ddfe3a85212ab,0xb4b85ff8aa729a9b,0xeffeff72ea77bbaf,0xffe2af7ffa73bbbb,0xddf7febfea72babf,0xeb9fdfdb6876bb8b,0x5cbff6f66a429b8f,0xfdffff63f873bbaf,0xf79ffffefa72babf,0xfdf7ffeee877bbbf,0x6dfedfc97873bb93,0xb5fe7ff3ea539aab,0xfb7ebfe2fa739a93,0xf7bfbfffaa77babf,0xe9fff73ea876b2bf,0xfbf5fed72a721bbb,0x1ffbeff73a523bbf,0x91c3bf6f2a621a8b,0x794636000840b227,0xcbfefbfeb252babf,0x46461b38a8623abb,0xb49fcfeb6a501bbb,0xd97fe66ee8569baf,0xdfbebd729a763a0f,0xeb7eea3eba523a9f,0xcff79e8fa872bb8e,0xe3becff13a723b9f,0x789667d8a8529a1f,0xffedff73fa72ba8f,0xf7bfff7f28763b9f,0xc7e4f30fb2709aa7,0x9f9fffd37a729a9f,0x12bf7fe7f8563393,0x10cf7e2aa52b2ab,0xb3947fe52a72aa8f,0xd8d29b0de803928f,0x2af897d22a62338e,0xd4b01d72e0461a23,0xf17efe7aea73ba8b,0xd282b65eaa523397,0x8fd4f37dea701a8f,0xaed7dfdfaa62ba8f,0x34f74fe1a8721a09,0x91793f23e8523b8f,0xfaae3de22852ba9f,0xd292c22ea8721baf,0x65df1e8db8779a97,0xbbaccf736872bbbf,0x9ff6b6fce872ba9f,0xdbbaff72ea5233bf,0xc3d0f34eaa52b209,0xab7bff41aa7212b7,0x43c0561a0721221,0x8110290a101901,0xce9d2de26a46139f,0x523c37e928601bab,0x8140b0800043021b,0x2b2a0f4a2a52018f,0xd437a06ca242829b,0x332a8b2888621aa1,0x210640100121201,0x40c24a00b0021000,0x8a5c7a4aa0621887,0x147615e0a8220382,0x90000208a0020000,0x4a20f40804222a2,0xe757f776e8663bb3,0xbb4fbb4b28511b0b,0x801801022002108a,0x41002038806202a2,0x59002d94a8629801,0x4000100000000202,0x320100021011,0x1080840a88620080,0x810080020000,0x0,0x840420c10011282,0x363104002500282,0x20120000008,0x102080c0a8020882,0x4010201008000082,0x200100100002,0x0,0x2100400000021082,0x40048000400021,0x40000001001,0x0,0x10200021000,0x800080100001001,0x48040000001,0x0,0x0,0x0,0x140024020020201,0x2010000010200,0x8000004000000201,0x514040600001,0x6000400200,0xa14010880a022011,0x37fb7ceeaa72b29f,0xf7fff6fffa52b28f,0xfffe7d79ba72babf,0xd384effeea72ba9f,0xc94bf6f068721abb,0xbbb6cdd1fa529abb,0xd6775ffce873babb,0xdbfabf72ea7332bf,0xfa3fe7d728779b8f,0xcbd7fa1bea52ba9f,0x7ffedfdbea523bab,0xfefefffde877bbab,0x39ae6769ba72ba8b,0x8bb41d7e2a621a9f,0xd9d2fb2ee253b2af,0x6b10dcdb68621b8f,0x35bb5fda7a521abf,0x7b4fc66a28723b2f,0x82b63ff8ea521abb,0xc14cf82d6a529a9f,0xfe7fbf5be0563a9f,0xb3a84fd3c873129f,0x905c5eec28023a81,0xe27e6f64ba72ba23,0x8b27eb43ba52138f,0x604f5e90a8622a8b,0xb8354ea228529a0b,0x1142844020031201,0xeb342ed2e852329f,0x88c10b1288028213,0x8256b2c2aa42128b,0xe63c4efb2a72310f,0x53c035aa8121289,0x97062fe3027312b3,0xfaccaf36aa729bb3,0xffeef23b6076b2af,0xafdfffdeba773a9f,0x74b757e96852989b,0x2848627320421292,0xa60feeaa852329a,0x84dc00000402226,0x40001020000080,0x8810144820129082,0x412012422800030b,0x2862060880202b0,0x4010020080020220,0x0,0x10000202000001,0x2032420a410203,0x8886350020609a87,0xd800982028129081,0x89523c13a8109a8b,0x393883b2aa12120f,0x543174d188561a96,0x8418008000282,0x8910270a280212a8,0x800101620420200,0x2c002008400210,0xfff77e7cea7733b7,0xdfa5fd5768763b99,0xc020004e80620213,0xa94ff6010a721a83,0x82861a00441221,0x300696812053920a,0x8d6a7df2ea731b9b,0x52800942a8021980,0x90d00000521200,0x2c411c4008431a01,0x24f846400850988b,0xd310cff1e842209f,0xdf022966b8421aa3,0xe7aef7b62a56ba9b,0xebfffadbe8529aaf,0xd7fa6fe3ea723abb,0xdb74f77f2a773a8b,0x421025020840082e },
	{ 0xe804ab0263741840,0x88c2381002620140,0xbffb6ffb227c17dc,0xff35a67ee7355ef4,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x82026704403718e0,0xe9fffbffe77f7dfc,0xfcd6ff5f677f5efc,0x141dc393a23c1e5c,0x79ffffffef761f5c,0xcb6e7df3677f1ff4,0x8bfcf77fe77409cc,0xcfdffff9ef7f5ff4,0xbfbf6ffbe77d1ffc,0xf7fff6ffef777ffc,0xfffe7ff3ef7f7ff8,0xff9eff7fe77e3fdc,0xafffff7fef775fdc,0xabbfff5ee77f3cfc,0xd4f7dffd637f5ffc,0xdffffff76f7f17dc,0xdfbf7fff6f7f5d7c,0xdf7bff3be77f5ff4,0xaadeffdbe77d1ffc,0xffffffff67767ff4,0xf9ffff7e677f1ffc,0xdebf2dffeb761efc,0xdfefffffe37f3ffc,0x7ff1d59e2761c6c,0xb7beeffbe37f3ffc,0xf5ffbff9ef771dfc,0xff6e7ff7e3773fec,0xabd2fb7be37f5cdc,0xebde3e77ef7779cc,0xbfbbeffbe7765fdc,0xfdffff7fef7f5ff4,0xfffefff2ef7f3ffc,0xfbf7ffffef7e1fd4,0xffffbffe637f1ffc,0xbfbfdffbef7f7fec,0xfdf6ffffe77f5fdc,0x97afffd1e97f5ddc,0xffb7fe7de87e1bdc,0x81cffe1bee761cfc,0xebfeffdfe77f1df4,0x76fffeff6f7f5fdc,0xfb67ff7b6f7f1ffc,0x9a8cefefeb767ff8,0xfb57ef3feb7f5adc,0xffdfff5fe77e3ff4,0xbcff7fefe77f1fdc,0xf57dfffbe77f3efc,0xfeb6bf7fe77f5ffc,0x89d1ff2c6b7c1ed0,0xedbfdf5be77e7ffc,0xb61bfff2677c12dc,0x977fbfffeb765ddc,0xdfbcbd726b757bf4,0xe9fefff7ef7f3fdc,0xcfee7fcec07f1f70,0xedfbcfeb6f7f1fdc,0x942f3e6be37e3cc8,0x6afeedf9ec743fdc,0x8114e73ca5771b54,0xeffbbad7eb7f1fe0,0x8ffdcf71653f7fd0,0x75fff5ebe7773fd8,0xbfb6fff0ef7e5ffc,0xffbeff7feb775ffc,0xefe7fe1ee7765fd8,0xb18cf5aa77c1ecc,0xfc7736e663761ee4,0xb3af5768e2661ddc,0xd3bfbf77eb7f5fd4,0xc7f7f77f637e15d4,0x1f5f5f53a5763fc4,0xbc1fc7fbe77f5ff0,0xdd3cff73037f3edc,0xb2810561627c14c0,0x4fdbeb3a437616cc,0xbffdff5bea7f7ddc,0x982a5dfee3771bc4,0x52ee9d7466775a58,0xfe566d5d637e4e74,0xabcef7e6af7f0b4c,0x6fd7bed56c7f1e54,0xbfff6ffb67755c7c,0xf26b76f06d7717dc,0xfffe6f71e97f5dcc,0xcbe4e71f63341bc8,0xfedeb779637d3cd4,0x2ebedffb6f770edc,0xd5be76ee63763dcc,0xeffe7f74e77f1dec,0xfbaff77de57f1dfc,0x8b4fb72f67765cd0,0x2bdddfd7677c1cfc,0xbeb4dccce3741f40,0xc566abe863741bf4,0x438d372d687c1764,0xc9c6eb9e67741ac0,0xebe4de4a62767dd0,0x3cd6d6ea2d523d48,0xf7cfe7f308571d4c,0xced9236ee6365fd0,0xebf2f23ee1765ddc,0x6aff7add647f5d4c,0xbebfcfff603f37d4,0xf97d2779c2773954,0x7e9e6df3ef7c1ddc,0xc1f2e77a677617c0,0xefd73e5b6b7519cc,0x9bafdfbba6753fdc,0x95fdfa73623c5ad4,0xdebe6ff26f7f1fdc,0x8b42bb24657714cc,0xe0f83c1b6a7e18f0,0xa3824d82663508d4,0x1080066001600800,0x812011222041080,0x9210090c64551044,0x4421a08035408a8,0x67071d10627e084c,0x2461142022301800,0x8a845000184a40,0xd801b6d237b19e8,0xc053fab064541454,0x9be1df5a257b1dd8,0x15f55f8340761e44,0x9318300000340044,0xc004354200384840,0x6100d528806c4844,0x67d81e81e070484c,0x8da223e0026d1948,0x200460000000,0x5a04101202140284,0x4080800400520040,0x4002000000240000,0x8210030040500048,0xb094b51042344c80,0x98c0d60e0347944,0x4020000001400,0xa02320000240854,0x82880060102000,0x3060246280104c04,0x1402d0000180340,0x2b0220080110844,0x818c9002e0280840,0x2110049080100844,0x5090940080210840,0x2004022000000,0xb39c35b0033418c4,0x8120c60a40141184,0x10040200302800,0xa040100000,0x0,0x280010400200a40,0x200020000000140,0x10020024000000,0x0,0x8242001100a00,0x82045c260524c20,0x804002224000050,0x0,0xb7f74ef263764ad4,0xfffffffeef777fdc,0xfffefff3e77f7ffc,0xdbdef76f637e5dfc,0xef73ff5be6644d48,0x8e9fdff86b771dd4,0xf767f6efeb7f3fe4,0xeff6fd73eb777ddc,0x598ecf0c0074245c,0xcdcf7efee3741bfc,0x7fdfd99be37f3ffc,0xb6d15e2a627e14d4,0x81df6f7def767cdc,0xf2bc3f7e66760fd8,0xcdf6eab26e7635c8,0xaa6dc443627c1f50,0xa570dfe3655636cc,0xeb67bf7b6b7e5de8,0xc1be2dfd677f3ed4,0x4982e523027e6864,0x777d5f59e27e5ddc,0xab324fabc07c0b4c,0xdcf8261327760f40,0x80a6fb6342343e50,0x8852230e615200dc,0x6cc31fd0e3665dc0,0x34bd06302c46144c,0x8053040400721800,0x951a3d7100741d5c,0x3bd7e3b66a7f1f48,0x88438b60c0760cd0,0xe884550022c0848,0xf5777feee6767f5c,0xbd6fd70e3561ffc,0xcfff77de6b741dd8,0xcfaefe02275d1fdc,0x3bafdfdbe7761adc,0x74ba76e9e25c48d8,0x41670320c0150ac4,0x12903a6c80521cd0,0x8840000100460244,0x2062080824460000,0x18100c02a5040840,0x810230606c501448,0x8000020000001100,0x484052640360250,0x60001000800,0x2000010000240000,0x8050040060101000,0x8100044002240040,0xa845c00000621148,0x420e16500076007c,0x41d4c0362343820,0xd496701463340884,0x4484646260361c44,0x82560020001915c0,0x408680901240a40,0x401d80160101144,0x36b4c468017a1944,0x407af51027351050,0x904710083102840,0xfd0320a08741c4c,0x814d05e425210c0,0x24118060407d0848,0x494e2570006e1d5c,0x5000020200440004,0x80430a0940360048,0x4002d84122700080,0x2c30d36963340e64,0x612e87f8aa563cc0,0xd414a23060764884,0xccc6f3ede67f5dc8,0xbfdeffdf42765f5c,0x8902ce5a674615d0,0x2044260000580248,0x7f962d1023762ca4 },
	{ 0xd3f7f31c3f2cd5d1,0x65c2b21a0a28274a,0xb91247533b05f3d1,0xdcfef7d09b09efd3,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf7be2fff3f29ffdb,0xdbbafb3b2f0defb9,0x4ada3a151b0cc393,0x9ebfdfe3be6cc5db,0xf16bfe7e3e2df7fb,0xdeaefd729bade7db,0xd3d5e36faa2dc7f9,0xefdbfb7fbf6dffdb,0xffffdfffbfedffff,0xf5fdf6ef3fedfffb,0xfbfffff1bb6dffdb,0xffb7fffdbfa8ffdf,0x6fdfff2f2f8dfff9,0xfb5bdfdf9e2df7db,0xbcff5eebbf6dcfd3,0xff6ef7ea176dd7fe,0xffff7fff1beceff7,0xffcffffabfecd7ff,0x6fffffdf3bedf7db,0xfffb7ffb3f6defff,0xcd7fa77b3f6de5fa,0xffbffffdbf2cf7ff,0xcfdff7b53b6defdf,0xef7fff4f1e6cffdf,0xbffbfffb3fecffdf,0xf7efb7fa3f2defdb,0xfbbe7ff70faddffb,0xd992ff3b1fedfd72,0xeffeb65a3b29f7d0,0xffffcdff3a6df7db,0xffffffffbf6dffff,0xfffefffb9fedfffd,0xfbfde77f8fadfffb,0xbefffeffbbedeffb,0xafffdffbbfedffdf,0xf4fff7fb9fedfffb,0xfabe7f722b2dffdb,0xfeb7fffc3f6deffb,0xcfc3fe9d3eadefd5,0xaaff5fdbbf6df7dd,0xf5ff7ffb3facffdb,0xbb6f7f7a0b2dfffb,0xffb46dfb3beccf59,0xffcfffbc8badf7ff,0xabffffdfbbedffdf,0xf4dff5e3af8df7fb,0xfffefffdbe2dfffd,0xfae77fef3fecffdb,0xe5c9f2192e2cc7d9,0xea7f3e583f6def9f,0xbfff6ff99a29cfdb,0xfdffb77bbfeddfff,0xfffe7fffbf6dffff,0xebffffffbf6ddfff,0x8fdf3fdf1e6cd7f9,0x1fbfefdb3f4cdfff,0xf7f766f81badffdb,0x7cfe3ed21faceddf,0xd8f4e70e2ca1fd53,0x6b5fb34b9a29cf7d,0x6ebfef433ea0d7db,0xddffff638feddf5d,0xffffbff6bfedfffd,0xfbffff7f3fede7df,0xfde6ff3fadecd79f,0xaeffcddb3a2de757,0xf5f7f6e93f0cf7df,0xefeb7d631e4cfff3,0xffdf7fff9e6dffd3,0x89f3fe8bbf297f58,0xbdf5d4f1e2dffdb,0x84fb3f639bccc790,0xfbfe67eebbadfd99,0xbaa33f6b880ca5aa,0xc9feff3e8a09e75a,0xb6c3a0dba4ce559,0xb6fe5fbb8b2cd69f,0xb33aff441a0ca71e,0xdef86070080cc71f,0x8992db270b2cc753,0x2cf3bed83bacc5f9,0xde1d77de2b2cd7fb,0xf0f4f6f31f65f7da,0xceaefd4a8b0defd8,0xdbb6eb3c1bcdeb39,0x4fdf3a5f3f6de5f9,0x8fbecbf58a6de7da,0x78fd6ffc9a8ca7db,0xd9f63d773e89ff9b,0xfbd7fb6f3b6ccffb,0x8fcffe2fbeaccfdf,0x219207da2f2843c9,0xf6f6a4e60b2ce7db,0xc778a7393b04e598,0xba902f4c0f28c3d0,0x8fc5fe039e0dc7d9,0x6adf5bc2174cc4d1,0xda121f680e6de6d8,0x7955ff531a6ca7dd,0xeeb42f7e0a2c97d7,0xebdcda0c1e6cf5b3,0x6a5fdfcd1f4cc5db,0xa7b64ffb3b2dcf13,0xf7ddb7ec1f2ddff0,0xef9e3d741f2cffdb,0xe0d6f36a8b6c45dc,0xefce3e9e9a0de7df,0xb5afcffbba8dfddf,0xf4f6f6eb8e8df7bf,0x64ba4df29b2c85d3,0xd950a922070c4d50,0xcffa36218a2ddfd9,0x611430802081194,0x897252ea000dc5eb,0x808c04221a08a20d,0xc1104012000c0504,0x6c3fe100a89cb91,0x3b3c4e9b8909f57b,0x34ea56e83a8c875b,0x55188082548,0x21a0f18090c4e04,0x4989c2300408d599,0xa8500c4a1b08845d,0x24b64c841f0c4e11,0x995cfc76142cc5c7,0x369e26740f01c892,0x2100c210020891d1,0x226c8046008c8113,0xb1007520a808580,0xd107167a0e08cebb,0xd2b02db21b00cf51,0x40021ca00300d018,0x266b1dc8024d4919,0x262601280a08849f,0x2400102012080010,0x84080040500,0x8220526000009618,0x4a12d9020c84d2,0x805cd400300c013,0x6480a280100,0x2018050002004b00,0x0,0x0,0x1844010a0108440c,0x30f054ca004c8511,0x69fa2e601b2cd891,0x480c093002208418,0x0,0x8994c400b204899,0x2108200000,0x5100004008000440,0x20040000800100,0x0,0x80800000008002,0x818240281a00a191,0x0,0x901408600508c101,0x9c2200300808519,0x20068221000c8512,0xbef9ffebbfadefdf,0xffffffffbfedffdf,0xfffefff3bfedffff,0xfbffff7fbfedffff,0xeffbffdf2dadd7fa,0x97becfebbf8cdddb,0xfcb75eee3aedfbd7,0xc33a7d620b8dcfdb,0xca0cfe2e1b0cc3db,0x87daff1b882c84f3,0xefffdfddba2df7fb,0xf4b6feee1e6cffbe,0xc55b77720a2cefc9,0xdb120b76128cc7da,0xcdcfc32a0b88edd9,0xef5a1e5a930895d9,0x5a2dd6b206c87b0,0x108c004008208441,0xcc902354004dc9a3,0x8f201710280ca580,0x480008000040,0x3f0e1ff72e2cd792,0x491cf66918a8b3db,0xb31429349a29cdc3,0xe389f30a3b4d5dd1,0x64d884410c0c4450,0xc00a4560200c8311,0xb302a6780a4cff9d,0x459c3570120dd191,0xcb82c3cf3a08b71f,0x815bb26a000cc793,0xc30c4c301b08c1da,0x7755d2693e0d6ff7,0xa51a9d100e4ce3b8,0xe82befff1b2ddfdb,0xd7dfbe320e6de7dd,0xab9edd930a6cc7f1,0xb5bcc6682e29e759,0x892d0f609b087410,0x84006008002408,0x2c04004084004,0x8114012040500,0x4025b0129086380,0x91085a081208cc86,0xa02351000008598,0x0,0x10040000080080,0x212003401b20c002,0xf0bca6112a09c753,0xdb947f320b64f7db,0x8940804006004c00,0x2e1317448b09c701,0x52818f81804c9491,0xdd67b6ed0acced3b,0x50160430286c2104,0x254000000088181,0x2c200800a002,0x901e5009086480,0xf09147620a0c1719,0x803424751a0c03c0,0x8020021004084080,0x801ac02080d6500,0x831a5c822a08a706,0xd0228af00909c381,0x20401200000c010,0x0,0xe4c602300c08c798,0x8140041800054408,0x361f57721a25cdd9,0x50e0354002089017,0xd5963b489a2c8771,0xe057d77f0a6cf7f3,0x18c50e0500248591,0x9a02048a0c886144,0x1080000001000010,0x3a002c00082c8010 },
	{ 0xc1c6d30d419ac82b,0xa7de3f5f05f64cee,0x9a1ec5520dbec436,0x959034681dc6e452,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfbb63dfc5dbacfea,0xcba5f36b4db65f27,0x26fe9e5addfaefe6,0xc49ecfb041bc67fa,0xf7ffff6dddbe7f6b,0xfffe7ff3ddfaffef,0xfbb5ff9e5daadfeb,0xffffffdfddfafffb,0xffbfffffddbeefff,0xf5f5fffacdfe7fe7,0xdbfe7f775dba5ff7,0xffbdff7f5dfa6fff,0xc7c6bf975dbe5f2b,0x8b3dcfdf4d9e5df2,0xfeb7defbdddecfea,0xdbcffff3d5bedfea,0xf3b4e73f5dbafdf7,0xfdf7fffeddfa5fff,0xcbfffb57d5faed7b,0x3f93ffeb5dbe4ff3,0xffffbd705dbacf7f,0xfebe2ffa9dfedf67,0xcbcffe2acdaadf67,0x2b5ffefbddba5ffb,0xf6bacffb4dfa4ff3,0xd97dfd524dbcc733,0xfff67f5c5dfecfea,0xeb97ffeb9dba4dfb,0x7f7b36dbddbe576a,0xd2bdcfb7cdda6f67,0xfffdf7ffcdbadfee,0xdfbefffbddbe7fff,0xfbf7ffefddfacf7f,0xeffffffb5dbe5ff7,0xefbfdffb5dfeefff,0x75fffffbcdfacfeb,0xdffeffe2ddba7d6e,0xff9aff6f5dbafffb,0xcfeffe3bcdbad7ef,0xffffffdbddfedf7f,0xf6fff6e6d5fa7ffa,0xfffe7ffbddbedfe3,0xf3bdff3f5dbe5dfa,0xfbf7ef3fcdfa7f72,0x7fffffdfddfecffa,0xf7baffeb5cbacf6a,0xff6fbff8ddaecfff,0xdb9e2bf708faefea,0xcfc7fe379dfa5e62,0xa5dbdb5bd4da5f6b,0xb2bb6fee453a07ea,0xfdfcffffddbe4f3b,0xfffe7fffddfeffff,0xcbfffbffddbeffff,0x2f7f3f5fcdb6cffa,0xdbbfcffacdbe7fff,0xd0deb67b4dba6362,0x92e677e88da8cf5a,0xf1958f26d13a0eb7,0x2f5a93db5dfa4de2,0xefbeef5a55ba5ef7,0xdddff77adcbadf6a,0xfffffff7ddfeffff,0xffffffffddfefffe,0xcfefbfbb1dbeff67,0xb97ecf92c5be5f4e,0x34ca5feec5fa7f61,0x39e6676a4d3a6f5b,0xfe3c2f7b54decf67,0x474bee3a54dae5e0,0xaedfdfd35dbadffb,0x34a46761d9aa4f23,0xb9e85e6808f20fea,0x3eb43f7dcdb8dfeb,0xd7dbf2b3ddfe8f67,0x5ddffee1458acb3b,0xb118cef241e2cd42,0x944c9dd00dfccb03,0xd7a0fd44158add65,0x8ba2db0e4daac9e2,0x6bf71ffa0dfe5ff3,0xafbfcfdbcddedfeb,0x4c7f07ec49b8cdea,0x6eae2de35d3a5dfe,0xfbf7ff7fddbe7fe7,0x5ff6badfddfacf6b,0xafbecf7a59bacd6f,0xdc7457e04cb8c732,0xfbb6bf744dae6ba3,0xffa8f76f0dbae7ea,0xcff7fe3f5dfacfff,0xaf5edf5a4dba9fea,0x8b65eed8dba470b,0x5fea777b05badf76,0x9fb22d1fc5a85d62,0xc58fea3a0dba5fe6,0xebfcdbc95dea5fe2,0xbcb65efb0dbaeffa,0xdbfff7e2cdbaed6a,0xfe902f5e9cbacf63,0xc7d3df3849ea4f6b,0xaaffdfc64dba7f62,0xbd325d7e44fa4fe1,0xf16e87704dfa4de6,0xdbbe2dfc95f25fea,0xebb5ffe29dfedfeb,0x2f5f1e4105fec1e2,0xfb9ceffb4dbe7fea,0xfef776ff9dbeff7b,0x4fa62d60c48acd7b,0xdbb67206100a4ff2,0xc1c1d21244d68544,0x30e4c104c82d942,0x48010000400,0x618ac30d0c24c60,0x191c22400c2000e,0x87c6fe924dac5f03,0xa0005e1a002203e2,0x54a7a4a28dea0da2,0x8541ac01cdd20746,0x181252041888122,0x840021008080320,0x950040840084d20,0xb190150240840122,0x802032604c820960,0x4000280200000320,0x804022030d804002,0x1006025001280002,0x38214bc081b84642,0x1088266300baca21,0x4a84266c14885c22,0xc982db1a8dfa5fe2,0x40020000020800,0x0,0x8000000040000100,0x402625c88c1044e2,0x8c20004100012,0x0,0x208e090045100120,0x110406245800102,0x414002005000102,0x800030400900400,0xa02300305a80340,0x41440110003004a2,0x843132c20cdced62,0x80760758508a4d02,0x1200200018004400,0x8152260344a40112,0x2100000080000400,0x0,0x60001200000,0x0,0x102820008020140,0x0,0x80003100002080a,0x80f00c409c800502,0xf2002dccc5928d03,0x400061b000224922,0x60800e0000c24840,0x9fbddfe64dbadf2f,0xffffffffddfeffff,0xfffefff3ddfeffef,0xfbf6ff7fddfa4dff,0xffdbbfffcdbeffea,0xff7fdfff0dbaff73,0xb475f679c5badfa7,0x2ab22f02458ad96e,0xf298e72e4ca8d9e2,0xceefb73b1d9aed22,0x6b3fcf57cc9a4f64,0x3cbb56ac4dfcc52a,0x1b2e3f614c925c6a,0xd3a427631dfa4f62,0xc4cae92a0cb8cb2b,0x7d6df5adcbadf6e,0x35fed5e94cba4f2b,0x636aa730cdcc17a3,0x1a3435b349ba49e2,0xc200009404a0,0x4d765c94549a7dbe,0x30aaefd05d0a45e2,0x456d152350884d4b,0x9061b605c8a4f00,0xc080711149080102,0x6842c844080a41e2,0x41c4ca019a44520,0x143022f9c8ba47a2,0x401c2c70810a0762,0x8a9e531bc8280720,0xa4e2322f04ba5d42,0xf1c5e780dea6342,0xdc90762a5db45f20,0xcf28b4015498ed2a,0xfbbe7fef45badd6e,0x94ff6df4cb85d32,0x7b7a97dbc5be5faa,0xfeb7f6f6ddaa5f7f,0xc9d6456a0c824942,0x8b810e448caccd00,0x2004400401000901,0x10c40000000301,0x400804600ca04002,0x4042256088084d20,0x44082df00cf08d22,0xc080401004820044,0x101008000100,0x84280460049a1404,0x8085b5350c387d20,0x33de7f654dea7963,0x88002a4201004422,0x20d61af2919a6c03,0x28cc36884a21645,0xc64264a05fa7de2,0x90011044001100,0x8102e2201d104442,0x48021f8000926740,0x610e4640498a0320,0xf0f210015d8ac522,0x5280204444281302,0x200400000004020,0x86803a000490d842,0xa618d19254a87f23,0x80816840000302,0x220857204002412a,0xc8082c4ac4a80760,0xa81c200480801a2,0x8000014200020804,0x80d699f305fae3e1,0xfcdc04320898cb21,0xdbb48b4a4db85332,0x811263284d026902,0x41484800d00000a,0x8200040841000102,0x41000000020040,0x4c4c2040019a0422 },
	{ 0x952f06a143066c2,0xaf11266a74102cee,0x4212cefa32107ae4,0x90fd2e3312125ac6,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf7be3f77361377e6,0x944c26d341047e4,0xaf66364e263272ea,0xbb3c4de2720067e4,0xf7fffffbf633ffee,0xfffefdf3f633ffee,0xfbeeffbff633ffe6,0xfffffffff633ffee,0xefbfcffff633ffee,0x7dbdff6bb4337fe4,0xdbaabff776336fec,0xff9bfffff633ffee,0xcfcfff7f3633feec,0xaf5ed7dbb612fde6,0xf5bef7f93433fee6,0xfffeef77b613ffee,0xfbf77f7df632feec,0x9fdfff3b7633f7ee,0xebdeffdf3433fee4,0xa57e2aeb32126be4,0xd1aaf7f3743367ee,0xfeb02f6e34127eee,0xcbf6ff39341377e4,0xede7fedf7433fbe4,0x7fb247ee3631dfee,0xfdfebdf91213f3ee,0xfbb665f516326ee6,0xebf7fb7a763377ee,0xfffebfd6b6337eee,0xbffbdff3f632ffee,0xf7ffb7fdb433ffee,0x4a847583943374e6,0xbad6e338063366ec,0xef7fb7dfb63376ec,0xbfffcfdff433feee,0xf5fbffefb633ffee,0xdbbfbd71b4136eee,0xfbbdfbef7433f7ee,0xaf4ffb9b96137dee,0x3ffedfdff63377e6,0xf7fffeeff633e7e6,0xf7eff77a3633ffee,0xf7b67ff7f433feee,0xcdddff3ab6333fe4,0xe7dfdfdc363377e6,0xfffff7fff631ffe6,0x3379affa663372e4,0x7ebe3ff73613eeee,0xaefc772a363375e6,0xbf7fbed7f612fb66,0x9fbb7f7e3633e7e6,0xfffeffebb633ffee,0xfffefffeb633ffee,0xfbfffbffb633ffee,0x2f7f1b4af61363e6,0x199f6efff433eee4,0xd5f6fedeb412f76c,0xffbe7df336327fee,0xdb56ef6714337744,0xe35b360ff4117be6,0x8effcf7af012eee6,0xf5fff7e9b433ffee,0xfffefff7b633ffee,0xffbfffffb633ffee,0xebefffdbb433feee,0x2ecddfdf3613e6ee,0x54d5b6f0361076e6,0x97ae3779f430f6e0,0xcb7e7b7f3432d3e6,0xe8d7ff0d141246e6,0xe3fcce583233f4c6,0x40b657a1301366c6,0xd594047410131e86,0xd0b12c7080106eec,0x8bf0d36d703376e6,0xef9f2e589433f6e2,0xa5223162a212e6e4,0xfffffffff633ffee,0xdb5ef7f03613efe6,0xc976ebb2f63366ea,0xefff5f5fb413f7e6,0xcfbfdffbb6327fee,0x1200100810104002,0xcab648780013f5e8,0xebf4e777b6327ee8,0xdfdf7f7f7633e7ee,0x7ff9dffb76337fe6,0x757556ee36124eee,0x9ebaec7114136eec,0xfa94fb2ab413e7ee,0x4fd7bfbf363357ee,0x3aa7ef40163342e6,0x96d377eb0431e7e8,0x91f9f773b413f7e6,0xfb91673734333ee6,0xf5ceef3a363377c6,0xea7f5fd296337dc6,0x3eaedfe5041377e4,0xb948f7fa34132ee6,0x9b84047424127660,0xec84d0266412eee4,0xfffeffda34127eee,0x9e327ff164324ee4,0xf0be3739a632e6c4,0x86166469d6333366,0x8bffe77eb433f6ee,0x4cf71f5ff6326aea,0xba3d7eebb631f6e6,0xb4fff7eef633eeee,0xdbf62de1b433ece8,0xfbd6f68636132762,0xc67b3613563346ec,0xca928d821201e7c4,0x305110a412014042,0x33eefe636430f660,0x50808958160106e0,0x458101814012268,0x8f9cccda6610e764,0x1084548aa01002ec,0xcd4ef573a41236ec,0xce062dc442002660,0xc492f132a41166e4,0xa5cfcd2b413326c,0x34a62d6400122ec0,0xd1d0a5c830106fc0,0xf392271c30116c42,0x400014a020022060,0xa2900f40141236c4,0xb0910d6230116de2,0x50002081120006c0,0x5ed0025e54136ee6,0xac0802210000ce4,0x6443158804314284,0x2100c0c010000568,0xd5fa7eee9633f764,0x4528014010304ac0,0x8900000002020240,0x9b5ab297b43262ec,0x2920029814000a44,0x451506010022aec,0x4d8c4c22843374ac,0x4000400000000084,0xd2e00a16026a44,0x214400000104480,0xf2b854d332126462,0x2056296010064,0xc010a21c0410d444,0x200030800101060,0x0,0x0,0x0,0x202000000022000,0x30a010010040,0x849105110114282,0x2220002830000960,0x5020240410010242,0xc082004020204,0x5088f0900030d4e8,0xc000040214000804,0x8498e2a2302344c6,0xfffff7fff633ffee,0xfffefff3f633ffee,0xfbb7ff2ff411f4ec,0xfffa7efff43377e6,0xffbfdffff633feee,0xf0ef96f8363376ec,0xdf96b770301366e8,0xdfa7f77eb633fcee,0xdddbbf97b633f7e6,0xabdedbce3613e6e4,0xf5a464f0361376c4,0x31e7a7f1b4336ee2,0xca906e71140226ee,0xebc2fba536335ee4,0xab3d0fde361357e6,0x9fce57ebb4135fe4,0xcbabfd613633ffe4,0xd286ef7b36336aec,0x89c0101014001368,0xae50a0d8221260e4,0x5bb86d73363267ec,0x81080642121042c2,0x5a0e096030114c60,0x8191b04004323164,0x14110000020c000,0x90a10400041054a8,0x5534f67814316ee4,0xd2e2ff1101377e6,0x8b90e7323013f2c6,0xaf63304b14307eee,0x2f165f53b43362e6,0x996e79e3a41376e6,0x92222d4054014466,0xc798fff664126ee6,0xdfcfff0bb633fee6,0x8bdbfd58b4307dc4,0xf6ffd6ecb413feee,0xf3b8ff72b433f6e6,0xd6842078301164e0,0x46c3ac024602d164,0xa15e144084326046,0x5304bc0142200aa,0x80402600000322a0,0xc82004421010c402,0x8140b1014011cee0,0x500042000004c0,0x2204c042004060,0xc506a346262224e0,0xdbe47de23631fec0,0x8006000c04220244,0x6e460299343172a0,0x8b18ee52201347c4,0x7c4c8661202226e4,0x20900050401062a0,0x10000000280,0xe008300001104a0,0x26e6cf4012134ae0,0xc23240141324e2,0xb045860042061c4,0x8584616890121aa4,0x8e7a7e1b001123ee,0xff7f1f5616336fc6,0x4000001000204000,0xdaa2542b0136264,0xca02414a1412246c,0x104108010100022,0x740401000122480,0xc0148170000460,0x9a844002942044e2,0x8a16016c103154e0,0x0,0x0,0x80460004102004,0x930815230321482,0xcd000000001202a4 },
	{ 0x1e30060245072,0x7a471adac01c96aa,0x89104e717004f0e8,0xd574b77bd0949fea,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf7be6ffdf8dcf6fa,0x4880731a103086c4,0xebe7135bf0b4f1c0,0xfebb4fb7f0dcd7fa,0xf7fffffff8dcfffb,0xfffefff3f8fcffff,0xfbfffffff8fcffff,0xfffbfffff8dcfffe,0xafffd7dbf8dcd7fe,0xa4fec6e950947ee8,0xdbbefd73f8fcb7fb,0xff97fffcf0fcbff2,0xcdfeffbff0dcfefb,0xef7fffdbf8fcdef6,0xfeffdfef78dcdffa,0xfbffbf7a7054bede,0xfbb66734f0d4b7e8,0xcddfff3a78bcbefb,0xeefedf57f094f7fe,0x18cd5581e0a43068,0xd9fbef7af0bcf6fa,0x86b7bf7e50b4fefb,0xefcef23af08cb7ea,0xe81efe8138e496e0,0x97daeffbf09c3fd8,0xb0d7ffec7854bffa,0xff667fe2f8fcf6ea,0xdbd7ffbff89cbef9,0xbdfefe7bf8d4f7fe,0xffbfaffbf8bcbffe,0xfdfffffbf8dc7eef,0x4fb6bffbd8109fdd,0xfb7eff6ff0dcbdfa,0xeffbbefff8f4f6fa,0xffffdffbf8fcbffa,0xfdf7ffe5f8dcbfee,0xbffafd71f0fcb4ee,0xfbccefadd838bfd9,0xafdefea370bcbcfb,0xffffdfd7f8bcb5fb,0xf6fdf675f8fc37fe,0xffeaffe578dcf7fa,0xdfbe7f7db8fcfefa,0xdfdbffbef09cbc79,0xc7ffff5bd094beda,0xbfff7ff3709cb7fb,0xfffffffff8fcbfff,0xdfbf3ffff09cbeff,0xedeff71ff8fc9ffb,0x5b5f7f73f85c74f8,0x95bb7ffff0bcb7f2,0xf5fff7fbf89cfffb,0xfffe7ffff89cfeff,0xcbffff7ff8dcbffa,0xdfda3fd7f8f4d7f9,0xbf3ffffff8f4feee,0xfdf7fefb70d4f7f9,0x6dfcbff0d09cbfde,0xcbfff72ff03cb6fa,0xcfdbfaff709cbffe,0x47f7dff970d4fefa,0xdcff36eef0d4f6fa,0xfffefff3f0fcffff,0xffbefffff8fcfffb,0xefdfbe03f0f4fdfb,0xefdecdded09c7efa,0xf09c56eed0d430da,0xdb7dbc71f8f496ec,0xffe66f6fd8fcbed7,0xfcdfe31ff0dcb7fa,0x6bfa2d5bd0fc1ecc,0xb43b5ed3d02cf0ca,0x11cda578f08494e0,0x76a02c66d08474da,0xc157da2570d416b2,0x9511fcaf09c9498,0x96964d62d0949e89,0xfffffffbf0dcdebb,0x600c041140a03468,0x102000040040000,0xeffe3fdfd05cfefd,0xcf9ecdfff8d477fe,0xf424f2c8d01c1c8c,0x7bbe65faf0dcfcfb,0xf9c5eb1ff09cb8b8,0xebc3bebdf0949fef,0xff3fdff3f0d4fefb,0xfdffffeef09cfefb,0xfbbebf65d0dcfcff,0xf38ceb6f509cd7d8,0xcfffff6ff8dcfffd,0x2bf9dad9609c9ef8,0xd4747434505c9698,0xcd7337eb709cbe76,0xffbc6fec70b4a7e9,0x68c6633af0909eb1,0x1ffcdfdb709cb5ee,0x9cbb5f7ef0149feb,0xd9faaf60781cf3ea,0xdea32d6ed8b4fed2,0xcd96f70ed0d49ca1,0x2f7ffadbf09c1fe6,0x172eff7370b43efa,0x906e3ff750f4b6fe,0xdf3e7fffd09cbc5f,0xc9f6fffaf01cb7ee,0xeacb92b3109096c8,0xacfd4feff0a4f4e2,0xdd76f65d70dcfcee,0xdbae7c437010bcfc,0xdb937b3f5090d499,0xcb53b69b7094b4ea,0x103cc772701417e2,0x90332eeed064b7c0,0x9a88254370348842,0x48c9001a00141496,0x495af29238d037ea,0xe0eea93700494a8,0xc4b4dec8d0541686,0x124052020143c40,0xcb86f771f0c45428,0x8210a202500014a8,0x20801035040108a,0x841152735084b4c8,0xdbd4f652900cd7ce,0xfc971750d04094c4,0x89527a1398949e82,0x885a120a300420c2,0x820401000108840,0x1014160290501008,0xc0600210148000,0x8106020450141400,0x8002300280801008,0x992c4d33103497c5,0x8002022010100000,0x9085c2000140e0c,0x1000c10010840080,0x431208b0141022,0x40d4c20402014c2,0x304226c04070b150,0x8001012070941000,0xba14d73050141482,0x49061220101028c0,0xb34091bf0b894e0,0x252104205010a028,0x875200901802,0x880046010148000,0x280a000d0001000,0x0,0x0,0x1000000400100010,0x40100100010,0x10840000080,0x5d21230809062,0xa000040000,0x100a41240c41080,0x32f02640301016e2,0x5802000101808,0xa806020030001428,0x83344a90500014a8,0xfffffffff8fcffff,0xdffeff73f8dcfeff,0x739be39370bcbdf2,0xefdf7fddf0fcbefa,0xfdbf7fddf0dcbefb,0xbdf5feee50dcf7d0,0xcfb7bf50d094b6fa,0xeaa5d36ff01c16ab,0x4fda3e13f014b62b,0x29dfd8d7789c1778,0x14ae76cef07c96f8,0xc1ecff22d01cb4d8,0xdaa43f2eb014f1ca,0x47cff20678943672,0xa79bd9d670bcbede,0xb0535fe1584c30ea,0x13c82d0070941168,0x4e972db0500ca6f4,0x8764fb74701417ea,0x688bba78705c32f0,0x971a456370bc94f8,0x520074685050d8c8,0x0,0x102a30b103012d0,0x400000800002040,0x128c050c708c14a8,0x631146a608434d0,0x8006800041402,0xc00040001000,0xaff31884f09cd07a,0x26364f5250d4bc50,0x79fc56eef0dcbff8,0x85aced72b89c94ab,0xf88ddf3ad09c1a7a,0xb53deabf034b6f8,0xebdfcfdad094feff,0x9c957664d0dc55fa,0xc9ef7f71709cfdfa,0xc08e2d40d000b042,0xd986e326d0dcb8f8,0x623fd489401cacc5,0x8493de61509495c8,0x6568f432401ca262,0x8004056c10141480,0x89c6e208f09c3ce2,0x802000010008000,0x483a0f83d014b4d8,0x3108f648509480c0,0xc1807800501012a2,0x7004e132208c1568,0x8142060a90009448,0x8e0b4f9a901c1eca,0xb75eaeeaf09cbce2,0x42554100141682,0x80040108004034c0,0x16422a014084b0c0,0x230004000101498,0xc4545cec78d4b649,0x40250000805140,0xc1102111580096c8,0xc47ef5a3f08cb6d0,0x6a5bcc5ab094b4ea,0x14010020000c0000,0xd280200060002020,0x80100000000c0000,0xc00010100000,0x2000008000000401,0x524000001242,0x1060d41008501810,0xc202245000c41480,0x80000200040002,0x808040000080,0xa212416848109850,0x0,0xb10098008001403 },
	{ 0x88000e2080048848,0xa9e140188e80370,0xf4b067a090eba380,0xa5ff76e9f16eaf5e,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfffe7ffff9ffaffe,0xebd4f777d96cb71c,0x23ef1fd7f8ffbb5a,0xf6fffffad9fbbbfc,0xfff7bffdf9efbffe,0xfffefffbf9ffbffe,0xfbfffffff9ffbffe,0xffffbfffd9ffbbfe,0xef3fdf76f9ffab7e,0xf5f716eed1e5bafa,0xbb3eff7399ebaff6,0xffbdfffff9efbbfe,0xeffefebfd9ffbb7e,0xfffedfdfd9ffbffe,0xffffd7efd9ffbf56,0xe97ceffab9edabda,0xff93ff7fd9efba6e,0xcddfffb999eeb37e,0xfbfeffd9d9ffbb7e,0x941e716798eda3d2,0xf17c6ffbf9ebb77e,0xf6be6f34f96dbf7a,0xcdcbf7ffd9dbab76,0x3ddcfbd5d1f5af5e,0xb7b36ff691fcb376,0xd1f097d0b1632b52,0xfabe3ffd98eebb5e,0xebd3ff3b90cebf5c,0xf7ee3e7dd9fdabee,0xbeb4effbd9f9bb5e,0x97fff6fff9efab7e,0xfbd6b7f0d9ffa36a,0x8ad06a0a90e52368,0xaffffbff91eebb5e,0xffffdffff9ffbffe,0xf4ff76eed9ebbf7a,0x3fbaff73f1eebb7a,0xfd8fffffd9ffbf7e,0xcec7bf3fd9e7bbfa,0xbbdfdfdfd9fdaf7a,0xf6f5def6f9fdbffe,0xffffff73f9efbffe,0xdeff27ffd9edabb8,0xfff7ffbfd9fdbf7a,0x2bffdfdf98ebafc2,0xbffffffff9fdbffe,0xfffffffff9ffbffe,0xfebe3ff7f9ffaffe,0xebf7fe36d17fabea,0xeff7fe5ff9e7bf6a,0xbffbfffff9ffbbfa,0xfffeb7fbd9ffab7a,0xdffe7dfbd9fdbffe,0xcbdefb8fd9edbbfa,0x87df3ed6d1ffb7fa,0xffd7cfe7d1ebbbfa,0xf5f7f6fbb9edbf5a,0x653e5d78d9feab7e,0xd9dcf72ed9efbb76,0x8efbb35fd1ef237a,0x93bc4f7ab9e9b338,0xf777f6fed96b2b76,0xfffefff3f9ffbf7e,0xffbffffff9efbffe,0xdc7e68fd1ebbf68,0xaf5f5b5e99edbafa,0x749cb70099482212,0xf5fff771d1ebabfa,0xf2bc7f7fd9f58f5a,0xcd8fdf0e91edaeda,0x6bdfff5ed1f9ae7e,0xe826593a0964a75a,0x11aad730d1188ac2,0x4aa0213a908ba350,0x8b71f60290f3bb60,0x840d288904c0308,0xf6fa47e391fdbb7e,0x14ed72ebd1a80a72,0x4b0677e2514a225a,0xc282e942d9d0a142,0xe04e9801d178aa72,0xb3ba5f3a91e5374e,0xcf33f472f0e12378,0x6cb475f1d9cbaf7a,0xfba6d33ff9fb2778,0xaff3dcabd958af74,0xffbbdffbf9f5a37e,0x77f176ead1eaabb2,0xeb77f5f3d9e7abda,0xfa9aff3dd9efaffe,0xaffffffff96fbf7a,0xe9df59d2e1aeab4e,0x743ff0e5d16dbf5c,0x9fadf7f0d12cab7a,0xfbb47f74d1e8aaea,0xccdee31c916aab4a,0xaa9fdfdbd9e32f56,0x149f59e290d00372,0xf54cbf66f1dfa77a,0x76be3747f9ebab52,0x6694f2a491df034a,0xab7fde7bd9d9ab7a,0x93335ffe99ebab52,0xb1eebff3d1e9321a,0xffd67fffd96fab7a,0xea87f76ad9fdbb72,0x69d3b6d491df8a1a,0x27fdcfc3d17faafa,0xf576f6fff14fbb7a,0xda6c7df3d9f9a3da,0xcb4cef7fd1c8b3f6,0x63c0bc92e968aa04,0x2808c54351252048,0xcad2a2c019650062,0xc51c78708861a140,0x8818c30a99409002,0x82388218618060,0x92364e00b9e62254,0x21a8c02819410270,0x8188204780c18540,0x8a804344b1052248,0xe88560299c02016,0x1d0800200240208,0x1e05340a00a2260,0x4008e453996488a0,0x3800060409480208,0x6060488402004,0x6218140010082256,0x1010240090010040,0x4040900020080200,0xd302200091048908,0x800108810008810,0xc0121e92d8480e18,0x1409058211000052,0x30122e00b8009252,0xa5343d70d983b064,0x9010eb2008480742,0x8cc8108010402414,0x84880019400000,0x1080248088440008,0x820e0160a8012244,0x1004242088002800,0x2106420291400258,0xe3dff96f1f8036a,0x30106740a1c08688,0x8246c21d1802042,0x4110220481010230,0x41010801400102,0x0,0x0,0x0,0x8000010000008800,0x0,0x46000888000000,0x8000002000000042,0x4028046988a22044,0xa20d40d0011060,0xa0040002000,0x845a0e48d0402e30,0xfcb3cfe399fbaa58,0xddfff7fff9efbbfe,0xfffefffbf9ffbffe,0xfbdfffffd9ffbffe,0xefeb3bdbf9cfab7e,0xbfbecfdbf9ffbb72,0xf5f73ee9f9ebbffa,0xa6baad70d9d9ab7a,0x508c9adcd9408b7a,0x4d52d288d963a27c,0x26d8d5cb914fbb5c,0x94972420d1401200,0x81b627b09160ab42,0xdb9c7e7f91febffc,0x71dfe8229022bbf2,0xeb7fdfdbd9eeba5e,0x20260a428040234c,0x82484c010ac8208,0x4a963e2298209a50,0x478e70e4d1c00b2a,0x2a7018c01100b298,0x41b0e552d164925a,0x9048b55098648a10,0xc21e0c31b0082220,0x8100e2402104220a,0x0,0xc68090830040,0xd46034fcd9c6a90c,0x8002340080012214,0x8220010001000160,0xfda1e53d1b5a34a,0x2fff5ff9d1efbbfe,0x95f337ebb1e5b35a,0xd4a52470f961b3d0,0xdba86f2ed9eda75e,0xcfcafb7f99f5bfde,0x3bfedfdbd97fab7a,0xd0bf7feed969ab7e,0xddfe7ff2d9e9b7f2,0xf6b6afd490faaf76,0xe0defe2490d0ab52,0xca3e3603d9e7b24c,0xf4da5d6bf1f5ab50,0x51010e400100214c,0x62b82de2b01c9a98,0x100800000208000,0x0,0x8b2245e3f12d3210,0xf5968770d1abb360,0x2880026280692a00,0xb83bb1ed966a352,0x60779251d9e02708,0x449888209180002a,0xf547726ad149a3f2,0x1024a000d0402200,0x8000100000002004,0xc6c238ebd1f42358,0x619840800120ab46,0x941754ebc1f4b352,0x2954502250222262,0x4082619590a42038,0x246fa00b1632b40,0x821e044290492362,0x145064e0b0a82200,0x440210400210,0x20000520c0800000,0x8002080100008a08,0x120080100000,0x2cd0302098c02004,0xc000081010021000,0x4202008000402204,0x8000100041000000,0x416000010800010,0x8000012000060048,0x24048000002212,0x240080100000 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfff6fff7fffef767,0xcba7efb7cdfcff66,0xabdedfdbebcee563,0x96dfdff3e36c6561,0xfffffffbfffeff67,0xfffefffbfffeff67,0xfbfffffffffeff67,0xcffffffffffeff63,0xbfffcffbdfbeff66,0x74f7feed5ae47766,0xfffeff62cffe7d63,0xffbffffffffeff67,0xfff6ff9fdbfef567,0x7bfedf5faffe7f63,0xf6fbdfe42bfcfd67,0xfdeeef7e0bfe7763,0xffd5ffffebfe7767,0x4fcfff3ecff87f24,0xaf7ffed7cfeafd67,0xb7bf5ff87f7afd60,0xfbfd77fc9bfeff63,0xdabfbff5bbca7762,0xcb57d7b65bc27762,0xdb5bfe74cbbc7727,0xb88c7b90a886664,0xd7edfffddbfc7763,0xfffe7ffddfee6723,0xf9cff7fa4feef563,0xebfedfffdffc7767,0xbfdfffff7feef763,0xddfefffa8bfeff67,0x7bbe3df35bbef760,0xcb95b75ecfee7720,0xbfd3bbd35ffef763,0xfffffffbfffeff67,0xf5fffffefbfc7767,0xfbbfff73dbfe5767,0x78c5e7960bea7605,0x6ad78f17facc6561,0x7fffdfdfdbfef767,0xfcf5b7f99b6cf567,0xdfffffeedffcf766,0xffbeffeddf7e7562,0xeffffffffffeff67,0x63fdcf5bdf6e7767,0xfcfe7feffbfcf767,0xfffffffffffeff67,0xfebe3ff7fbfeff67,0xadd3f46b5b7eef66,0x2f6ebfcefdca7663,0xaef6dffbdfeef747,0xf874bffe4ffcf566,0xffbebf7defeeff67,0xe9deef7fdffe7763,0x6f579f51da7c6561,0x9fffcfffcffcfd67,0xf1dfbee9effc7760,0xcf267d798dec6f67,0xd9d5c33e9d287f65,0xeeefbbdfdffe7d63,0xeeffcfe9cbf47761,0x7f779afb4bee7762,0xfbfffff6fffcff63,0xfab07f7b8fee6f21,0xcd87ff8e4f4aed20,0x246088b48086140,0x50f4beeedfce7762,0xc5cbb769593e6723,0xd900833c49286760,0x6dc7ec288ffe7564,0xa3ffd3dbcfee7763,0x7df37debefc87763,0xfbf5d7f29dfc7f67,0xe40c2522090a6d41,0x8f28f2b048486704,0x6b7fd7d59bfe7764,0x3bb54f38c96c6d65,0xc8e1134805884140,0x5200745d47203720,0xe2b1fb6e8bf87521,0xa8f7ce1292746507,0x3ebb6cdb49287760,0xfc7ff2fff9fa7761,0xeefefd7afffefd63,0xebdce776ff7a7563,0xfffbbfffdffcf763,0x67bedff39ffe7f67,0x9d7f7fe85fbc7767,0xe39aff71ddfa6543,0xfb9ff7a6dbee7743,0x1df39f39dbee7562,0xabff0fcefdfc7767,0x409686e102786461,0x91487e5109a86f63,0x8d3c2f7d19fa7f60,0xcdf2bf3a9b3afd66,0xaadedf5a59667c42,0x8e871f6a8f6c6d40,0xdb643e5a5d7c6f26,0xf2be7d72fbaae767,0xeecaf36ac8bcf546,0x2b5e7bde4fec6562,0xbf7867f98dfa6742,0xd44cf6f4efa82f60,0xe3b67d6edbfa7661,0x6acce32a09be6562,0x6fff1ffb5b6e7e27,0x68e8f8111aa0f60,0xde7e7eeb7ffc7563,0x8d74edf3098c4f65,0xbbb4ef2f4a6c7767,0x8e4b5fd94fe87f66,0xbbbe5f5d2feaff63,0xa477fae2cb6c6d06,0x9f76fc70cf68e763,0xea8a640000bc5664,0xcfc6ff8213ee7765,0x3f3f3352dd60fe63,0xb0f2dfc7db4cfd63,0x994b2770ed68e422,0xca081e6a85406e03,0xedd5ea3f0fea6763,0xcef23f4789ea4560,0x8bb16fbb0aec7763,0x7964767390bc2764,0xda3e7673da2c3746,0xaffeffbacbeaff67,0xac771b5bb9627765,0xb43b24440d424162,0x54bc2e065f406722,0x56f875544040f462,0xcbbae97b8fae7760,0x606806d128600660,0x2e8dc5411b425463,0x9525006a09ec2c60,0x0,0x0,0x800020000080020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4200004001,0x4008000300000,0x1800000004040,0x84080101200c00,0x0,0x0,0x0,0x80400000404000,0x0,0x0,0x10020000000600,0x0,0x102000000000060,0x0,0x800000801020000,0x100860801380400,0x2806440118a0500,0x8bbbbb27152ce520,0x8000520002000460,0x80c00040a000,0x1100000000000060,0x444000404060,0x400004000000400,0x4c30045003000,0x0,0x2100004002,0x0,0x0,0x0,0x0,0x0,0x84c008084100,0x0,0x0,0x24000000400,0x0,0x20020000000400,0x718086021801520,0x8a80f00041403000,0x8082191005646002,0xa4b6c67083883660,0x3074720589426743,0x21200d0042246101,0x110030d03422400,0xc0c619510d646e00,0xd912cdda9b2e7562,0x6496d0e0112eec00,0xa31eb63008747665,0x9392ef7c5bfae541,0xc5ce328f4fb84562,0xa3dfdfdbdbfcff62,0xb3bc5460db487722,0x2006056102080760,0xd2b43f32d97c6d62,0x4c59941001884f40,0xef7ef401d8387764,0x308a4741013a4720,0x8110420109005003,0x3800050001004921,0x0,0x0,0x148936e85b026503,0xbffe97f9c9ee7762,0x0,0x0,0x0,0x0,0x40100000200040,0x2010000000000400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2000200000400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8188662286f64d20,0xffdeffff9fbe7f79,0xcbb5ff33969f7d70,0xe49a3c8a8a0e3d20,0xbfbfffff8ff6ef60,0xffffffff9ffeff79,0xfffe7ffb9fffff69,0xfbffffff9fffff79,0xffffffff9eff7f69,0xffbfdf4a1ebe7d60,0xf5fff66e9ed76f69,0xfffffff39fbeef70,0xdfe7ff7f8fffef71,0xcfffffbf1ffe7f70,0xfffedfdf9fdffd61,0xf4fbf6ec97bf7d68,0x9beef77a9ebefd68,0xffbfff751fb7cd60,0xdde7ffba1efe0f60,0x2f7fdf5b9bfecf61,0xf4be5fa196bb6f60,0xddddf7fa97b76f69,0xbfaf67769ff6ed79,0xcbc1f7779bb5ae61,0xa5fededf9fbf3f68,0x162b4fe500f67f61,0xf2fef7fa9fde2570,0xffeeff771bbf6f68,0x98ff62e07d30d40,0xf7ffbfdf9fb6df69,0xbe7ecf7b0ffe6f68,0xd6f7f7fd97ffed69,0xefee7deb12be7f61,0x3bd5e33a9ef84d68,0xef77fbdf9efe6f71,0xffbfcfff9fffff61,0xfddfbfef0ffffd61,0xd7fffd721ebe4d70,0x7aaf77fc97e76f71,0x4ddffe1b8ef67d60,0xabfcdddf9ed64f70,0xfcb8f6ee96faef68,0xfbefa7718a9fbf78,0xfbbc7f6f9bde0d68,0xefdfffba1fbf2f79,0xafdecfdb9fbe6d70,0xbdff7ff79afe7f60,0xffffff7f9fffff79,0xfebf3f7f9ebe7f70,0xedfed39f15b61d61,0xeb6bdfdf83de2f60,0xfffffff79f9f7d69,0xfdfdaeca9ebe6f60,0xfff6ffff1e9e6f61,0xebffff7b9ffe6f78,0xf9f71edf9ef6df68,0xbbbdcffd1ebe3f78,0xc177566201974e60,0x4bfe2df31fda7d70,0xcbf47f2b16d6ef60,0xeecf9f530ffa3f70,0xbfbfff5b1fff7f68,0xdd75366d86ba0f78,0xf7bbff7287bf7d71,0xdb9cf5ff86bf2f79,0x8787ee0b1e964d68,0x89df9d521b9fed68,0x9081f06c10826820,0x19f4be0916960540,0xc9cca77c1e926f70,0xc4e8fa1382960d40,0xffffffdf9fffff68,0xf7f6fdff9fb7ff60,0xfb4ca76e12d63c60,0xffbeef7f9ffeff61,0x8ff6f73f1ef6dd79,0x49deddc30e1a2d70,0xfb27d3094920f60,0x95efb74a07f7ff70,0xff2c7e6416b62f20,0xe9ffff2f97f66d60,0xafc61afb9adeed70,0xc1b647f00ada2d70,0xf4d3f6f99e974f70,0x3b3e4f730fff7f60,0xfbdfef2e9ef7ff60,0xefffbbfb9bbeef71,0x3f9fcfcf97b66f61,0xbdf7566e16fa6f60,0xbebebfc212f63d68,0xcb22f77517be5f61,0xeddfff3f1eb62f78,0xab72ffdb0696ef61,0xd4f576f583be9d71,0x7daa27e20e360560,0xcab753fe9f960d68,0xfde6ea120f964d70,0xaeffdf5a1fb65560,0x9c9e67e316be0770,0xdd397e520e9cc561,0xfe9e3d6c07f74d60,0xc956db3e969f2d70,0xafefdbdd0eff6f70,0xf6be5ffb9ffedd60,0x860dd56286926f40,0x8eb2654216df3d70,0xd3c5f7fb9e964f69,0xef63ee8049e4960,0xf9fc36b17b64520,0xd77ff77816ba5f60,0x7edc35611e972f40,0xeb9ee73d979e6d78,0xcdeb7ef396356d61,0xda9ecff2029e5f60,0x1477d66a06b48d60,0x639ca51012961f20,0xdabee6269a9e0d20,0x9dfffe1e9eff6f60,0x8fe1cf9b97fbaf60,0xf4d6f2ee96b6ed60,0xd5ea37e197b64f61,0xdfce75fa87fe3f68,0xd167a7360e375d60,0xc98edd9b069e2f60,0xe5977eea8f366d60,0x68ccf7531fbf0f61,0x12828364832a2d60,0x68c6cb2a08860f60,0xf59fa570e320550,0x36915fba022a4d60,0x65efbf6a16be6d60,0xdab65c0c9abc0d20,0x6bd4a373169e4f61,0xeedf3fc486bedf60,0x9a12944906d22b40,0x1430f13780a32f28,0x8200000000060000,0x800010000040020,0x0,0x0,0x0,0x0,0x202000000900000,0x2100000000140,0x0,0x4002000000420,0x0,0x8000200000000820,0x6000000800040,0x0,0x810003a100c12820,0x8000004000000408,0x0,0x0,0x0,0x0,0x0,0x0,0x2000420000060140,0x40020000004040,0x1010210000021820,0x9520024a82900d50,0x20052000842800,0x2004010400200540,0x8020108000120100,0x0,0xc110000000000410,0x0,0x0,0x200210800000,0x0,0x4000001000010800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1104800a000400,0x0,0x9200004000020460,0xb40200100000d10,0x8204080100062820,0x888840880020d00,0x14ac0240001a1020,0x1422e4002810c40,0x1280010000900820,0x12100e02080c20,0xa6fedff306afaf68,0x1d76386d1a8e2f40,0xcb3e3f56863bad60,0xc92eff588616bd68,0xc9dab6db060e6d68,0x2fffffd717b76f70,0x90b614e0139a8d61,0x14abf0805860d60,0x8094019010060d01,0x4bc292340e320d40,0xa1460210020a0c20,0x20f4002002408,0x28420000020160,0x400004000020800,0x80000100020800,0x0,0x10332be184220d40,0xc0911061801a0920,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x21000100100,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd3be6f22f7be7ffe,0xfffe7fffffbe7fee,0xcbf7fb3ffe9c7ddf,0x6fff1f7dc79e35a6,0xffffcffff73e7ffe,0xfffffeffff9e7fff,0xfffefffbff9e7ffe,0xfbfeffffff9e7ffe,0xeffbffffffbe7ffe,0xbf7fdf7fd79e7fec,0xfcf7fef7debe77ee,0xffbefff3ff9c7fef,0xff97ffffff9e7dff,0xcfffffffff9e7fbe,0xeadecfdfffae7fec,0xf6ffdeeff7be7fff,0xd73faf7edf9e7fc7,0xff9f7ffed73e7fe6,0xffc6ff97f7962fe7,0xebfeffd2d71e1dec,0xddbf5f2bf68e67dc,0xff7eff7fff9e7fe7,0xdfff3f7ef7b47fcf,0xebf7c306f7be6e6c,0x63d7bbd3fe8c6fd7,0x5ebf6df7d79e776e,0xdffdef63d79c77ef,0xfbfef7fff6bc7fce,0xebb6f3dfd79e6fce,0xbefebef9de1e7dee,0xdffdcfffd79e7f4d,0xfdffffebffbe7ffe,0x4fde3f03f53c57ee,0xebffff7ffe9e3ff6,0xeedbbffbfebe7ffc,0xffffefffffbe7fe7,0xccf7fffddf363fe6,0x8bfeff73f3be7ffd,0xdb87ef3fff3e2765,0x8feffe0bdf9c7fe2,0x2bdedfdaf69e75cf,0xdff57eecdfbe07e6,0xbffffffbffbc776c,0xff3e7e7bf6bc7745,0xeff7f6bbf73c7766,0xabffdfd3d78e3fd4,0x9cbfdcefcb9e3597,0xf9eebf7e769c3fe6,0xcab63ff0fe0c3f1d,0xc9d3d722d49c15c0,0xadeeffffd71e75c4,0xb7fbeffbdf9e27e4,0xbdfdb67bffbe3df7,0xff3e3fecf79c37e4,0xebf6ff7fff9e7fcf,0xfddfffdfd6be3fdf,0x3fffeff3f61e7717,0xbcfe7673d5be67e3,0x6bdaaf61da2c67e6,0xfbdee77fff9e2fff,0xfefafffbfe9e6ffe,0x2bbecf5af79c23ef,0xc1d3c772f61c6726,0xfbf6fdf2deae36f6,0xfbb6ff7ef7be37ed,0xcdefb894f694355d,0x8a5c5c9adb0e4724,0xb0d51ca4560c3544,0xe9fefe6ad68c3c46,0x80182f78d30c6124,0xc84bf7b9d69e6dcd,0x8f5ffe51ff9c3ffe,0xe7fe5fdbfe3e7de6,0x944df57ffaae36cc,0xe9b70458d60e15ce,0xc1c3fffbf7942f62,0xea965dddfe8a6544,0x97b14db2fa822dce,0xfced3624799a3f05,0xc20630c0b2043404,0xdbfefaaeff9e75ee,0x74e623f0cc942cc3,0xcdd8ddf9d79c7d34,0x17e7776fff9e3d74,0x7dbefd73f71c6fc7,0xf9f6ff3ff6862fed,0xeecbfebbf71e77d4,0x7fffd7dad7343fd4,0xf0f21eebf68e2fec,0xdbf43d70fe1a440e,0xef9eeffdf68e3f7e,0x8a5effbff79e3fdf,0xbf7fdfdbfe1e37e6,0xd49c70adf6827f66,0xb1e7376af7b61de6,0x6fb43fffd49e3fec,0xe1d6df3bd79e2365,0xa7f9d7dbf2ac7fe6,0x14ba4d8656a4294c,0xa32c3eeb571a3d54,0xb69e3f60d7b67f7d,0xcfcafb02d79c3fe6,0x3adfffcff78c25f6,0xb7fecdfef28c7fff,0x61382e21d60c03c2,0xbb367f7dfe3c55ee,0xc1e6f267f6842d7f,0xa6db9693d79c0664,0xf19f4bf29d9c3d44,0xd1e3f629d21e6cbf,0xcfbeff72d2167da4,0xabd0f33ed69c2da7,0xc85336bbf71e35c4,0xa9ae4d68f6967464,0x29525bfbda8c7de4,0x9f88fff0d6021d4e,0xfbf2ff7eb2966d4e,0x4b863223de006c0d,0x3a8f5b9cd2907f15,0x953706e05a0e3e26,0x53e72d75d6a672de,0x7aae5feedf9c6dc3,0x845ff00bb3822c4c,0xedfdd757f2ae3de3,0x446355e2f2ac23e4,0xf3b87ff3d59e6dac,0x870e0868d6944164,0x29c4d228d18421a0,0xc173800120c0064,0x90026cc0402428d1,0x214d3719c4044140,0x53423d20c6802444,0x8bb3e223d6044b84,0x825e1042f1140184,0xa100542408005d4,0x84002260800c1400,0x0,0x0,0x0,0xc000080040000,0x0,0x20040000002800,0x0,0x200010000020,0x0,0x0,0x0,0x0,0x0,0x800080040000080,0x0,0x2004080000100,0x4020244044802000,0x0,0x0,0x0,0x0,0x804000020000400,0x0,0x2100010000040,0x18428202002904,0x0,0x44322541500e0480,0x8b104143560000a6,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200100200040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6000000010000200,0x0,0x0,0x22441b200010c,0x22162540008c0,0x28161050b01c0440,0xb0024f49d6001606,0x40060040000184,0xcdfe3d62f614694c,0xb990e524de8c25c4,0xc8e3af38deac6fc5,0x10074030080500,0x90b17269d6102646,0xdf9b7760f68e35d5,0xd704fc60d32631de,0xfdcef23bf7bc3eed,0xeefedfd8f3be7fce,0x78b07420d39029c4,0x804f2450d2862422,0x48042491d0807450,0x4012168044042880,0x20d3905890a82544,0x91c0a052040182,0x8110006210002a20,0x0,0x0,0x0,0x841052e836102e04,0x9228dc14da0414c4,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10a0844000000,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x21000000300,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd3beef728bf577ff,0xfff6ffffcbfe7fff,0xebfbffffcbfd7ffa,0xff7f5f7bcb9e6fbf,0xbfffffffabff6fff,0xbfffff7fabfd7f77,0xffbe7ff3abff67fb,0xfbfffbffebff6fff,0xeffbffffcbff77fb,0xffbfffff8bfd6fff,0xbdf6ffff8bfd6fef,0xf7feef77cbbe6bfb,0xdb97ff7e8bbd6fdb,0xcfeffe6f8bbd7afa,0x3bbedfdb8bbe66bf,0xfcf7feffebff7fff,0xdbefee7fcbfb7bff,0xf7b77e7debf86ee3,0xc7d2fffacbbf7bab,0xafffffdbabfe67f7,0xfebf777b8bf967ff,0xf7fcfffaabfd67ff,0xeabf2f6fcbbc77f9,0xcfffdf39eb9573eb,0x2bd6dedf8bfb63bf,0xbef26f7eab9c6ffb,0xf7fdfffeebfa77bf,0xdfde7fa68b7c677b,0xfafffbfe8bdc73fb,0xefdf3fffcbbf7fff,0xf7dffffe8bf46fff,0xdfffffffcbff7ffb,0xdfae3dfbcbff73ff,0xfb97eb248be86bbb,0x6edf9eafa9db6b37,0xdeffff7febff77ff,0x75f7fffaa3be6fff,0xdffeeff38bdb7ffb,0xfb7afb7f8b9a73ba,0xcffeffbbcbb77b7b,0xebfeff5f8bfa77df,0xf7fdf76febfe67fb,0xdcf5ff7ee999639f,0xff9f7ff7cbdf7693,0x4fd6ffbfab9a6ff9,0xabffdfdfeb9863f3,0xf5f7dfebebdc67ff,0xf5cff7f98bf07ffa,0xfeb65f6e8bf673fb,0xaffff7efebbc6fff,0xaff7dbfc8bc46ffe,0xbefb5fee8bfb73fa,0x7d7df77f8bd472ef,0xfebe3ded8bf373ff,0xf3fefbbaabf477ff,0x7def3fd9e9fc767a,0xbdfcfd7b89997ffb,0xf6bef7f88bf07bfe,0xfffa7f708bfa63ec,0xdba4fe2fcba863a2,0xfeefbfcfcb946ffb,0x7fbfcff2abfe732b,0x70ffffeb89fe777b,0xdffeffd2c99667bf,0x5fac5f7f8b326643,0xc0fbf22389bc67e9,0x2b5b1c528b806227,0xf4b4f6680b4a7b25,0x2627708b186043,0xcdb2777e8b7763ea,0xc082e3ba8bdc3f3a,0xa7bcd89d09a16743,0xb4fb0fe58be977d3,0x596e2d02aa902735,0xf6342fde8af06ff0,0xe181d7264bc866f8,0x3f8edecba1b463b3,0x8ba2cfb88ab82213,0x9967b4708b982773,0xdb8e7c718bda2b52,0xc9c1f77b8bb8637c,0x4ccbdc558b8c62e2,0x569acf7aebb1737b,0xf4e3b7fdabb86ff3,0x5f9e7ff14bfd73eb,0xbbf4cf3eabb6667b,0xafd3bfc38bc56bfb,0xeabe6ff3cbd86f7b,0xb7fdb76acbb567ef,0xfffef5b1e9b8633b,0xe2b6f2ca8bd863aa,0xefdffa93ebb27bf7,0xabdddddf8b9467bf,0x34b749e6c1902f72,0xcbdc2770899077e9,0xdf303f6689b56bae,0xcc4fef18a91872c9,0x3b2f5c1ee1847fbb,0xbcf65f2e8bae6b77,0xd9ffee68abf962eb,0xf2a63e62a9a07748,0xc1f6fa2589d223da,0xe9feffdb8bb4335b,0x9f9beffea99467ff,0xd330b64b8bb872ba,0x7ea67d698b96233b,0xe90ef3aa8bdc3f5b,0x3e823e9b49d87ab1,0xb4abcdc08bd06b19,0x746f72ea8b08636f,0x6b5ee5e1899462ff,0xf92cdb2681b43232,0xefefdac3c3b4673b,0x49ecd89c820264e,0x50d534428b802279,0x2a80f4001900374,0x5954013209e46660,0xca5a32a9ab3e77f1,0x2977dfd38bb266ff,0xf7feffea8bfb673f,0xd9a5ae62c9386309,0x77bebf6aab9c36bd,0x868aa2288b5c620b,0xa5a15ca03815a62,0x3d1c03e28b81666a,0xfee42f148bd02339,0xdaeab76e83a06e3a,0xc26646268b902224,0x18ec84181326a02,0x9f225f60090063e4,0x9180240003104208,0x10a8034009202401,0x88c4010c82802708,0xaf4809488bc0222a,0xfa1d6dda89dc62ba,0x1496b9810908628a,0x1004000000000208,0x600009000000,0x2020080000040,0x0,0x2802000000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x88c88008a480202,0x10304cc28aa82003,0x1080000000800001,0xa00080000008,0x0,0x0,0x204000001200,0x824a80012003,0x60900c420b880209,0x4000400080000040,0x48000014a002000,0x21000000a000000,0x400100100002300,0x0,0x0,0x0,0x0,0x0,0x100020809420000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000000008400000,0x479231540b0062a8,0x100a28401900070,0x85214050bd86000,0x198310281282213,0xd261fe398931677e,0x1647462aba6033a,0x5ad16706a93063b3,0xa03bacc88806203,0xa98df1a0bd14327,0x1c41d2280b840328,0x95186d50818061f8,0xc30c777fab3662fe,0xcc1ff3beb982743,0xbfbeffdbebff6fff,0xd0d5f6fe8bf42f13,0x4192a543093063ab,0x8f1a24240bd01338,0x80000221a2800000,0x6a5e1d5aa9d04261,0xa122056a43027760,0x0,0x0,0x0,0x0,0x8830434080042710,0x9564350609c8272a,0x4800000002800000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd7beef33fbfdd77e,0xfffefffbebfddffe,0xebfdff7be9eddfbe,0xe8ff7ff5fbf5d9fe,0xfffffffffbf5dffe,0xffffb7ffe3fddbbe,0xfffefffbfbf5d7fe,0xfbfdfffffbf5dffe,0xcffffefafbe5cd7e,0xefbffffffbfdd77e,0xfffffffffbfddffe,0xdfbffff7fbfddffe,0xfe87fb7aebf5c76a,0xdfdfbefffbfddf7a,0xebb7df5ae9d5d37e,0xfcffffebfbfdd5be,0xc1bf3f7ee9f5d37e,0xdfbf7e7efb75d33e,0xfff7bf3de1d5d33a,0xebffdddbcb7dd3fa,0xf4bf77eddbf5d37a,0xfffdb77ef9d5d3fe,0xfebf6fe7f9f4dd2e,0xd5efeeb4c96cd53a,0xae7e5dcfc9c4d53a,0xb6fbffebfadcd96a,0xfdfff7fbfbfddfbe,0x75fe6f7dfbd4c3ba,0xebf7ff3fd0fddb3e,0xefffff7ffb75df7e,0xffbecffbfbf5dffa,0xfffffffffbfddbfa,0xfffefffbfbfdd7fa,0x9bf4f5f6f964c77a,0xaf7fbfdfabdcd73e,0xefbfdffffbfddffe,0xf5bff7ebebd5db7e,0xf6feff51caf5d22e,0xfba67f7fcb7cd73a,0xedfeff7df3fddbba,0xabdfffdffbfdd32e,0xf6f5dffefbd4d36e,0xfff6ff7afbddc37e,0xfbb6f77dcbfdd32e,0xedc7ff3fd1f5dd3e,0x63ffd7d7e3f4cb3a,0xbcbf7fedf34c5fbe,0xf9edff72cbf4d77e,0xdc3e2fead175d13a,0xc5f7ffa6fb5cd7fa,0xeeaffaebf1fdd32e,0xb7fffffffbf5d77e,0xf37dfff9f3f5ddbe,0xfff67f7dcbf4df6e,0xebfdff36ebf5d13e,0x7bff1f1fd3edd90a,0xb7bfcfdbd1f5db7a,0xb42edae8c354c35a,0x69ba6c51fbdcd1ba,0x93ccfb16f1f4573a,0xffdb7fadfb74c35e,0xc892c57af3f8c102,0x7465d76bd3d4d07a,0xdafefd63fb54d3fa,0xfb9ce737ebfcd72a,0xc9c7bf03c165c7fe,0x27d6cf57f1c4d12e,0x86f6d6ebebfdd16e,0x81eefb63cb6c50ba,0xfa121f62fb64c32a,0x5802810a6048514a,0x2b1edf54c0f5c23a,0x94173fe38b3dd52e,0x736f075ee164d96e,0x7ebe3dd5e935d91e,0xc1bdf626db2053be,0x940081001484082,0x94f15ee0434c511a,0xf7eda65dc0c5c93a,0x52822de5c155c376,0xe9a9c130c8d45168,0xa8163e455b45c32a,0xb29de762d354c33e,0xd5fef6d5f9c4d30a,0x7bfe7f63f3d4ddee,0xf1d5d36cf354d3be,0xadd7baabdbc5d13e,0x7fbfefd1ebd5d3be,0x14f67fceebe4c13e,0xcb2cfd33f1d0d16e,0xfbda6f2ee330d7a8,0xe6eff7bbebe4d36e,0x33ceddd8db64c15a,0x347edeede344d3fa,0xf10fa7f3d9d4db3a,0xfe32273cdac1c90a,0xc9c0cb3cf1c4411e,0xccfddfdbcbfc511a,0xd7fe7de7d175d3fa,0xf93c6e807045c31a,0xface257551c4d37a,0xcdd2b16bc944d3fa,0x2ff7b43d1d0df7a,0x9fbbf7efebdddb1a,0x536b376acbdc533e,0xa7143d3dd1e4df6e,0xeb97eb26f3fcdb3a,0x6ec4bfefcbe4d05a,0x8592ffe24bdc51da,0xdef7f2f7ebf4c76e,0x10c3d32e844d31a,0xd964fb0ed041d10a,0xf5d39a59794cc63a,0xe7dccf51d3d4573e,0x14fc0e6048ccc122,0x8f867d6181ecd3da,0xffb4e372dbdcd32a,0x595236026105c72a,0xbfafdf92e1d4d33a,0xf4f26c68d8ccd52a,0x9945ad76c0409136,0x92b62f25d8545358,0xc843ee225041c11a,0xafdecb59c0c5d02a,0xa9b24da680a1c12a,0x5d687756e1f4da1e,0xfabe3f76fb55c3ae,0xccc99618e8a1c01a,0xab5184650444132,0x622cf52c004d11a,0x508026688104930a,0xefdc61a3c0f5d01a,0xda50630f8075931a,0x21410628f0e0d008,0xc09446fac20ddb12,0xd0d5e6cccb54d3be,0x2010000040020,0x0,0x0,0x0,0x0,0x8080000004008,0x0,0x0,0x0,0x0,0x0,0x4002080001000,0xc0001040000,0x60000000004002,0x90000040004000,0x8080000000400008,0x0,0x0,0x1012000000802a,0x0,0x8008a84041041200,0x80010000400002,0x200200001004000,0x2100000000008020,0x2010050000000,0x840e000480219012,0x12005000000120,0x0,0x2c110090044000a,0x0,0x0,0x0,0x0,0x0,0x0,0x420000003000000,0x24000002000002,0x0,0x2220240085000,0x24040000000,0x0,0x0,0x8200000080000004,0x0,0x0,0x0,0x0,0x821449c08044d358,0x8bc68a20cb55d102,0x28622048f204d202,0x104041280001104,0x95e9fffbdb14d7be,0xb2c3522e0e5c52a,0xd9897e1dd1c453de,0xc3c03ec77155c322,0x8fee4f136345d12a,0x64f6f6f9f17043ba,0xcbee7f73fbfcd33a,0xcb9aefbff37ccb3a,0xaefefa3bebf5c12a,0xffffdfdbfbfddffe,0xfe15b7e7c374c76a,0xd77e93eaf1add23a,0xff4a7b76f3c45b4e,0x86d28a204154c102,0xeb7ec9c3ebc4d31e,0x2420101040120,0x1008e2455000d108,0x0,0x0,0x0,0x30404048001002,0x9202031081240a02,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1080000080000800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x93beef7382dffbe6,0xfffe7fffcadffff6,0xebffffffcadffff6,0xfffffeffc2d7fff6,0x7fbfcff3cad6fff6,0xfefff773cadffff6,0x7fbe7ff3cadffef6,0xfbffef2fcacfffe6,0xfad7bedfcadff7f6,0xfeb77ff9cacff7f6,0xfdfffffecacffff6,0xdfbfff73cacfff76,0xfbadeb3f8adefaa6,0xefceff3ecadeffe6,0xabffdfdf4adff3f6,0xe5a255fcc8c3ff34,0xd97ee77ac0defbd2,0x7abfff7fcac3f7a6,0x94cdff3bcacbf6e0,0x6b7f97d3ca86f3e2,0xbfffd2facacbf3e4,0xfbdfff6ccadbfff6,0xefd63d6ec2d3f3e0,0xeff6ffb9cadbde74,0xe957de55cadaf2f4,0x3fbfe7fac2c7f7f4,0xd3bdff77cac6f3e6,0xef967d93c2d6ffb6,0xabfdf27fcac6f766,0xcbe75edecacefb20,0xf6ffffefcad7fff6,0xffffffffcadffff6,0xfffefffbcadffff6,0xfbd5db7fc8dffb34,0xefe7ffffcad7fff2,0xffbfdfffcacffff6,0x7cfff6facad7fbe4,0x1f1827f1c2dfb262,0xfe99ff7fcacbffe6,0xeff7ef3fcadefff6,0xeb7ddfdfcad6fee6,0xb6fdf669cac7bff4,0x9bafe767cadbf7f0,0xdbb4ffffcadabf70,0xffffffbecad7ffe6,0xef5bfb584a96f366,0x1dbf5fafcacef2e6,0xbbebf77acadfb3f6,0xfe9b1f74c0d2fb54,0xea86fff6cad7f766,0xa3fedeffcacfff62,0xbfb32fefcadffef2,0xf37cbe73c896ffb2,0xffbcf7fcc8d3f7f4,0xa9f7eb67c8d7f776,0xe25d1ffbc2c7eaa6,0x9fbfdff34adbb6f2,0xffd5ff73cadbf7b2,0x7bfe7d73cadff6f6,0xdbdeefeec2c2ffb0,0x6e8e9f4fc246f726,0x2abecd93cacfb6a4,0x75f7d7fccacef7f6,0xd7faef71cad6f7f6,0xf39cf75dcad6bf26,0x8bc7aa17c8dbf7c0,0x9a4fdd51cac2f2f4,0xdcfdf6fbcadff776,0x9d7f677bca57fff6,0xd3022f05c8d6f324,0xc080a73bca83b024,0x2845ddc202c7a252,0x964b57ea8a41ab60,0xd13eefc38acafb22,0x4a842fe68a45f6a2,0xc286f0fac2c5ae22,0x156740042402a24,0xb6924ff2024bfec0,0xd064fdec48cafe22,0x8b107544c8d233b4,0x8934bbb9cad3f774,0xfe5e36a9ca4abfb4,0xb8abcd6acac6f6a6,0xf17ed7d8cac7eff6,0x5bb63f73cac3f7e6,0xdb76fb37cac2f722,0xeb139fcecadbb9e2,0x339ecfbacad7be36,0x1ff3feeecaceffb6,0x9bb4bd71ca5bfbb6,0xfbbefd2bc2d3f3d2,0xe7fffbffcadffbe6,0xaf1fdfceca43fb36,0xf4f7b6e54043f3a6,0xfe627d71c8d67ba6,0xff76ae70cacfb3f2,0x5dc3fbbac251b594,0xa9f6dfd0c0cfffe6,0xadb66de24acbbff4,0x91e5b5c2c893b382,0xbaae2d68ca43bba4,0xc100d665c242a320,0x6f9e1ed5c8dfbba2,0xc7ba5debc2dbbbf2,0xfdfddefecadef7e6,0xbf86f57c8203b760,0xe8eeffabcadafbf6,0xcdde1a4cca13fbe6,0xc69bcffbca92bf66,0x9276deebcad7f336,0xffba6f71cad7fb76,0xd914e5244842de24,0xcefb3b49c2cfffa0,0x1e576f5a4acb9360,0x777157efc8cfff66,0x76be8ff0ca47fbf6,0xd38669e28acef720,0x8be7de3ac2d3fab6,0xbbdfdfdbcadffff6,0xfef774e0cac7b3e2,0x996c6e3a8883f0e2,0xbb042900c08bb326,0xcec1df2a8a9ef2e2,0x24001520200a062,0x2daaf16b8051f380,0x884e706240c58234,0x963a3de8ca43fef4,0xc0c278034a4a3482,0x245bcc190843a722,0x34b264b18295b222,0xd33c44bb4a02f2e0,0x2b467c6088ce1a06,0xf156eb2748dbbeb0,0xad4e020e8842ba24,0x5f084a300cbb6e0,0xd585d264c2d6f120,0x0,0x44000040000020,0x0,0x0,0x0,0x0,0x0,0x80800008400000,0x0,0x0,0x0,0x0,0x100800000018000,0x0,0x0,0x0,0xc2040000010,0x0,0x6000000008020,0x0,0x0,0x0,0x0,0x2300000028120,0x84c846818201a220,0x404a81000862022,0x48027440c2002a04,0x202c00000400e00,0x0,0x0,0x0,0x10001000002000,0x0,0x4842101042005204,0x0,0x8000002000020020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x820000008020,0x0,0x0,0x0,0x3c5e0960c8125204,0xe390e72a48caf722,0xebfa34d6cac3bf70,0x110bc00002a3c0,0xf0f5f65ecac7f3b6,0xdfc42d02484b36d4,0xc174234fc042ab44,0x8bd39647c0cbfba0,0xe03e8ea3ca46db72,0x14f62ee042daf230,0x9bb40d6188c3eab0,0xdb9c7ff4cad6fff6,0x4d5afa1d8842fbe2,0xafdecfd3cad3baf0,0xf6ffffeacacfff76,0xdd8e5d7a48c6aef6,0xdf9ffffbcacfb3f6,0xc1ceb62e4a42ab96,0xad4355c7c247fbe6,0x30801d0182840a04,0x8c19a2124285d306,0x0,0x0,0x21040000004,0x0,0x50002002080,0x0,0x0,0x0,0x0,0x0,0x0,0x4000200000008200,0x0,0x0,0x0,0x0,0x0,0x0,0x6000109000002022,0x1002008000200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xb204a70144026811,0xff36ffffcecabb99,0xfbffffffefffff9b,0xefffbfffefffffbb,0xff9edffbefebff93,0xfef77ef9cfd2bbb1,0x7ffefffbcfdbffb9,0xfbdfed2fafc7ff39,0xfffbbfdfcfc3bfb3,0x7ffedfffcef7fb99,0xfdfcff7aefdafb9b,0xbfbffdf2eedbfb93,0xddbefbff8fdffa92,0xcecfff1f8de3fe91,0x2a5e17de0fd1fe93,0xb6bfd6efafd37f91,0xffaffff3afebbebb,0xff966ff38ec3b219,0xcdc7e7178ec3ff89,0x2b9e7d5b0ff2fa93,0xfcfe4ff3afd3bb91,0xdf3eff7aefd7fbb3,0xfe16fff58cca7e91,0xcffbffbfefd7ffbb,0x67dffedaadcb3f89,0xccaadf7f8cda7bbb,0xd6efff69afc33f83,0xfbeeff66cdeb6f9b,0xabc74b2d8dd3fa01,0x96fbbf53eec33b39,0xf7fbffffaff3bfb3,0xffffffffefdfffbb,0xfffefff3cfffffbb,0x9bf6cb0ecfd6bb91,0xcaeebe50cee23a81,0xffbfef78efe3bfb9,0xfcff5fff8fc3ff93,0xd4995df2e5c33a29,0xfab7e7feefd3fe91,0xdffffeffefff7fbb,0x2fffffdfefdf5b9b,0xf6bffff6eedbff39,0xf7f7bf5a0cf3baab,0xfb98777dcff77bbb,0xfddefe92cfd3feb3,0xeffeffdfcff3ffb9,0xbcfbfbef8f92faa1,0xfdf7fff8cfbbbba1,0xfeaeaff92ee3ff93,0xcff3f31b8fe3eba9,0x2b191fdf0ee17f01,0xb3badf7b4fd3fe99,0xdf7dfffbefd7ffbb,0xdffe757b0ff77a89,0xe9dcfb3a2dd79ab3,0x6f77beffcfd7bab9,0x9f3f7ffa8cc2ba29,0xf75dbef9cdf2daab,0xefbc6ff3cdc3fabb,0xfbdceb4fcecbbf9b,0xff4aff5fcfc2fb9b,0x4f9fcbffade7fb9b,0xfcff327f8dc33bb1,0xbbbebf738fe7feb3,0xeb36e76fafc73e93,0x8dc3f61b8ec3ba81,0x2b1cdd940c961e81,0x76b6f6794fb21f9b,0xd9ef7f7a8f97df93,0xbbb829064cf69281,0xd1c6e9168cc2ba81,0x4388df920d036a11,0xa4b9f5e24f823e91,0xf1ff66aa0f8a2f11,0x9cb6bffdce8aff1b,0xa9d6c20f0fc97f1a,0x1674108e801901,0x841034684a819b91,0xf7ffbfffefffffbb,0xca887c6d0ef23a11,0x89c1e10a0c927b01,0x6c7e3bcb8cda9ab1,0x2ebdeff94dc5da8b,0xf9ffffff8fc7be9b,0xfff6ff73afcbff9b,0xfbdeef2eadd3fe3b,0xc9d39e5be9674b83,0xdfbfdf5f8ffbff9b,0xfdb35ffecf837f91,0xffecfdf2aefbbe99,0xf9bced2605d1ba08,0x8efff70a0dc3bf0b,0x8bdfdf9b8fe2bb89,0xd4f776ed8fc2be83,0xdae6d66ab42bb01,0x9fba6fdf8ec6bf1b,0xfcaef3358ec78b39,0xe3f76432f8fba03,0xffd65f608df3db83,0x537aff788ceaa38b,0xfaaeaf63c6c77a13,0xc592c60604c69601,0xeffefa4feef7b2bb,0x94937fea44cfba13,0x722cf7798fc2ae91,0x582c65620cc29e89,0xebfffa3e2cfbfe0b,0xedf3be7e4fcebf13,0xef3fdf76efcabf11,0xf5e7f67f0fd9ba9b,0xf3b47d70cfc17b8b,0xa9c5ef9809a18e83,0x8a533b1acda35a81,0x6e16efc84cd37f99,0x84de57a045c4be21,0x9ffefdd48ff3ba99,0xe2866e2a4ec09b11,0xdbdafa79ce9bfe99,0xeb1fdfdb8fdffab9,0xf5ffdeeeedcffe9b,0x9fed2d528ccfa333,0xd7100e22cc0a1881,0x80a23e088ac39b01,0x104c0008808001,0x89120fa38fa20800,0x914aa74085c23a11,0xd285031249010a03,0xcddbba848eaa3901,0xaf49f74a0dabfa89,0x8b8cdae8ec22e21,0x90e3077a4a239b93,0xd2820c1288c31ab0,0xabd5eb368e92b891,0x49801b0503431a01,0x26d0d5890d821801,0x911e1aea0d833e98,0x0,0x0,0x8020000000010200,0x0,0x0,0x8000040000008001,0x0,0x40001001000800,0x0,0x2020000004020000,0x0,0x0,0x0,0x0,0x24000020008,0x804000000000600,0x0,0x100000100001800,0x0,0x0,0x0,0x0,0x10808000100,0x800010000004800,0x1000070080030800,0x9012028005422a00,0x1a00250008401a81,0xd200000602800281,0xc0081006002000,0x0,0x40a0045008000630,0x8200008000100,0x9000016809801a01,0x0,0x0,0x1220280c012280,0x8421001001800,0x804020800010202,0x0,0x44000802400201,0x0,0x0,0x0,0x0,0x0,0x0,0x18c820008023101,0xf6bc7fd7cdd3fa1b,0x890632f00c90db09,0xaf77b3de6f82fe9b,0x9013054a0c007801,0xd1f5a02e0d8b7e09,0x5fbefc488c82ffa0,0xa1d4c7260c527a11,0xe9c7d757ed621e89,0x5929c9d3aed2be01,0xf5f676fccfc39e99,0xbf2aff73efc2aa83,0xe9d9ae6e682226b9,0xcfd7ff3bcff3bb93,0xafddcfdaceea3bb1,0x7eff5ffbcfdebe9b,0xffef7772cbd7fb8b,0xfbf67ffecffbbb91,0x6944a50a0d4a2e03,0x2f3f7ed2aee3fb21,0x302809010c062288,0x5260aa5005816800,0x80010000000201,0x2000406008c01201,0x0,0x0,0x240208000110,0x0,0x200000001800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x14000100002,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6f5dedcb60311150,0xbffbffff663117d4,0xffffffff677717de,0xfffefffdff7717ff,0xfbffffffff7e97ff,0xffffffffff7f97ff,0x9fffdfffff7f15d9,0xf5f756fd3f7e11de,0xf7be3ffb7e7e14df,0xfbcfeffff77b17df,0xffffffffff7f97fe,0xbb3fcfffe67397fe,0xf5d7deeb737a15fe,0xffbdff73e27e95fe,0xfebfff6e367e14fe,0xdffffe2fe67615ff,0xebdecd5f735a15fe,0xf4fefcf0607f95eb,0xd9acef78a25611da,0xefbaef77657b91df,0x8fdfff8baf7b17fb,0x2afeffdb6b7714de,0x77ff57eb7a7e95d8,0xf9cff77be57f16da,0xd6b627f7aa7615df,0xedefffbbb67a17fb,0xa1dedc53337611ca,0x3cba4fffa27a93d2,0xddfcb652245316fe,0xebc63ffbf27a155e,0xcbfff63a385e93ce,0xfff717daef7717fa,0xfffbfffbef7f97de,0xfdfffff9f37f15ff,0xfffe7dfae77395ef,0xfb7ef77f677f95de,0x475f7fc3a05e95de,0xef1fdf76ba7215df,0xd5b6f6ed237615ff,0xdffc6df36f7297da,0xfbbfffffb77e11ce,0x9ddffebd377717ff,0xbbdfffdf277795f7,0xf4bffff2af7f15fa,0xd1febf6aa77691df,0xdf95bf6ce37294ce,0xfdffffbfe77717df,0xef5ed7d1237a94ff,0xdcfbffef737e956a,0xddfffff8e47691de,0xfebeff7ff27797df,0xefe7f3bfef7717fb,0x6fffffdba27e1777,0x3ff2d77e247715de,0xf0ed37e3627a12c8,0xff9e3fff6f7f15fe,0xc9fff77fe77f10ee,0xe9ff7ffbef7f95de,0xdf3deffbea7615df,0xddf7be7ef27b17fa,0x5fee7debbb5e94fb,0xf9d4bf3f627312ff,0xce7bb7f76c7b15d7,0xbfbfdfdafe7f17de,0xfe94d6eaa77a11db,0xdfbaed52267a96de,0xdb9f7f6fe67610ee,0x49cffb0ae15017fa,0xab77dfdb6373167a,0x3af97660a272115a,0x8090210002001642,0x40080d00004c004a,0x405196a000621549,0xabdcd65f297314de,0xf5b75febe356154e,0x9feda5ea357415ce,0xc6a61fe6327615c7,0x8949f91262761450,0x2fd84a1c816215c3,0x9d885fb9207410ca,0xf1f7b766e067105b,0xcfb63ffe6f7f125f,0xebf6fe7fef7e15ee,0xffffbfffe37b95fe,0x3fbcefffef7f17ff,0xfcf6feffae7396df,0xfbfeff73ff7f95ef,0xfb70fbf7677f10fe,0xc9dbbfdfb77a93fb,0xf778cd7e7364107f,0x9ddf16e86a7b91db,0xcfbefd636d5617de,0xdb9ff7752276916e,0xcff3ff93ee7795ff,0xeff65d87a77713de,0xf4b6d6a6ef7215de,0xfdff6e74237e94ee,0xbeb12d06b27615d3,0xd9dffab2676f91cb,0xeb6ffdd9e27a15fa,0x8c971f6a7a7295de,0xd1681fb8b15414ca,0x848a2949204114cf,0x2596f73e617515e8,0x6dff5edfff7714df,0xb6f35ffa277e13ce,0xd0fc777bb07a11ce,0xab265fe5207313c9,0xcac6f37e327495de,0x6fe77ed3247e11df,0xf92cd42a05614cf,0xd53776f0267b91ce,0xdf9e2f71277314ca,0xc395f110227200ce,0x424b9e4b225811f2,0x8b3e4c7c287295d2,0x75be37c6a27710df,0xdb2abf60e17315de,0x82a06c720174114a,0x8973fa2be61817d2,0xaffecfdfe77795ca,0xb7ff7fe6ef7795de,0xd92db672227a11db,0x8a04622464601058,0xc000500224600050,0x8a775ed0307210eb,0x651106000601050,0xa46ca708007b1458,0x8a04212902531446,0xefeff7fbf77b15d4,0x6e567fd9a26b11de,0x30106b02007411c0,0x8000025060421400,0x80004000500000,0x4190f20368481668,0x2115000440006,0x883c6ca070721404,0x9c75d311a07317fe,0x0,0x8000080022000000,0x0,0x804000200040,0x0,0x100020000600000,0x0,0x0,0xc00000600000,0x0,0x8404000820121000,0x20620062000000,0x0,0x0,0x0,0x0,0x0,0x0,0xc0000001100,0x0,0x0,0x0,0x0,0xc480000a2000002,0x8002083200501100,0x4424142820620040,0xc427540a21811c2,0x44020810101002,0x0,0x2400000000400008,0x0,0x0,0x8040022000000,0x2100002000002,0x10040000001010,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2000002000100040,0x0,0xcf167eb7277f10e3,0x895fda28227014d0,0x8b4e0e4022721060,0x7fbfcefb675317de,0xfaf7f67c6a7f97ee,0xdbeefd63237e97dd,0xe354f73ea37210db,0x4edb3e7b2f4217df,0xb3c4754207315c0,0xb467dec2637a11de,0x8dbe1ff0a67a905a,0xf0207366235215e0,0x4edbfadfa77f90ee,0xffffdfdbe57717df,0xffffffffef7f95ff,0xfdffaf7fad7f17df,0xdffe7fe6377e14ee,0xc47b3b0860721256,0xa1dedcda257e15e6,0x2300422a040024e,0x8040040000600000,0x0,0x8000100000200040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6000000040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x108000020020000,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x100640288002800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4040200004004800,0x7aa53ff60b7dccdf,0xefffffffdf7defff,0xffffffffdf7fefff,0xffffffffdf7fefff,0xffffffffdf7fefff,0xfffeffffdf7fefff,0xfbffffffdf7fefff,0xfeffbfffdf3feef6,0x1f9eff734b3febbf,0xdff5f7fedf7eeff7,0x56ae2d7b5e3fefe1,0xfbfff72fdf7fefff,0xffffffffdf7fefff,0xffbedfff9f7febf7,0xfd7ff6631f3ecff4,0xd7f6ff715b3fefff,0xf9bdb77e9e3dcef2,0xcdeefe3b8f3cebec,0xfb588dcf863dea96,0x54b773ea0e1fe2cf,0x8ffbf77e9e3dedfd,0xcbbe7fff1f3fe8dd,0xdff7f3ab9f3fcbdd,0xffdfdfdb9b3eeffe,0xbdb77dfbdf3feffd,0xffffffeedf7eeff6,0xdaf6fff55b3fefd6,0xf9e4d39c4a7e24fc,0x3b5efed14b3d6cbc,0xbfb2ffff5f7deff2,0xff5fbf6bcf1fcdfe,0x7fbe7dfdc53ecfe6,0x8baedb7f1e3f63dc,0x2fdf3ffb8b3eefb7,0xffffcfebcb7eeef6,0xf7fff6790b3feff6,0x5ffe7ff3df7eeffd,0xfbf6fff7cf7feffd,0xeff3ffffcf7ecfff,0xffffdfeadf7feff7,0x8f7d57694d3deaf4,0xdbfe75735f7ccefe,0xbfd4e76e4f3defdb,0xefdeb73fdf3fefff,0x3ffddfcb9f7feadc,0xf4be7fe7cf7eefd2,0xfbffb77a5f7deefe,0xdfb7bf7f9e3eeefd,0xfdf7ff2a5b7eefdd,0xefff5741da1e2ff4,0x9d3f7ff2cf3defd6,0xf5fbeff2df7eeffe,0xeebe3c769b7ec8d4,0xe7dffb1f5b7fedfd,0xffffdfdd5b3cabff,0xb63b4ffe163ece95,0xd0ffb75f1e7eeaed,0xfbde7f534f3feedc,0xeb95f73fdf3feffe,0x6edf3eff8f7feafe,0x59bdcf71c83c4ab4,0x99ffb76e1b37cfcf,0x6fbe77e18f3eeeef,0xbbfcff7fdf3deff4,0xfb53be8b453d6cd4,0xa7dccc7f9f3dcac6,0xe73f76f81f3c4f9c,0xcbaaf7430f2fefd8,0xde80ff764b3fc3c7,0xdad7ffa59a3fcf80,0xe1bfddd14f3ee08c,0xb4df5fb24a3fecfe,0x8986094009262800,0xb000252c120fc090,0xdcb90aa002368384,0xaec4d560b3dce56,0xb5badde71b7ee7c4,0xdd9ce7744a7c22c4,0xc6b0055e5a1ca7ea,0x383e14a1a05e4c0,0x286faa030a3c29c4,0xb4b2f6c0ca3caad4,0xd17dfe6e4e3de7d0,0xffde7fb2df3defff,0xc9c3f3ae4f1eede7,0xefffffdfdf7fefff,0xffbfffffdf3eefff,0xbd7c76ae0f38cadf,0xdffc7f738c3ccbd7,0xf386fb7ecb3fcfe6,0xcecfbefecf7dcf89,0x3e3ccd584d34ce80,0x91de14fc4a1eeafc,0x9edeebb45a3c24dc,0xfbbeffff5b7caff7,0xeb7ffe3fc634cf97,0x8f7fde5e5e3ee9b7,0x77feffed0b7e8fd7,0x8dced74a4a7cce95,0x2bb06f26de7c4c08,0xc7d2f53a4b1eacd5,0xee06d5470b5defaa,0x35ba4fe30e3cedde,0xf76c876a033c6cb4,0xff8805580a34ce81,0x82f08330413ec8b2,0x69da9ed15e3fc2b0,0x15fa07f3ca3dedb5,0xfffe96fbdf3c4eac,0x42b63d034b30ca82,0x8120f358ca354ac0,0x48001e48810c8c82,0xb3b24f52071ece14,0xc417e47042362b84,0x9abe04025a3d4e94,0xcb34ef088a36cac2,0x843b2300b3c6840,0x3fbcff99cf3deafd,0xfcf47eed1f2cede1,0xbffeff739f3fefef,0xf716637b5f3deffc,0x1db5c0b011fca95,0xbbdfdfdfdb7eeef5,0xb7f9d7e4df3fcf9f,0xf9ffaf7c4f1defef,0x119253358838aada,0xcdc59410482ca093,0x245b13c0032cc1d6,0x77475fab8b3cea92,0xc102a04a411e4880,0x2ca605f84b144fa8,0xc976f729cb36aac6,0x2b4621c84a31c1a2,0x1a8241ac4b346a00,0xd1b137401b2ce0d3,0xef842c500508c8d2,0x90481040c214000,0xff21940083481c5,0x110909309306280,0x60265067081e82b2,0x40200000040100,0x0,0x8100080000800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc0000004800,0x804040000008012,0x100800080010,0x0,0x210000000040400,0x0,0x0,0x0,0x0,0x0,0x20400080004,0x0,0x4904000001800c0,0x9000000000080080,0x57982e000a1cea89,0x800010000040100,0x11000000c0000,0x0,0x0,0x18200008002040,0x4000200012000000,0x850000102080210,0x24000080004,0x100100100020,0x100000200080020,0x0,0x0,0x0,0x204004004000,0x0,0x0,0x0,0x0,0x0,0x0,0x30405248240200,0x88bffe3a4e3eea98,0x6f5f37fb1f3def9c,0xbcbe5fefcb3fced6,0x5df6f77d5f3eeeb5,0x2c9e5df1071ceff4,0xcbd5e72a8f3e69c4,0xe0c7da5b1b3d6c80,0x2bbe5f5cdf3eebbe,0xb5feffef4f3feffe,0xdfbdef75cf3feed8,0xefb0376f4b7fcfd7,0xdecbff9f4f3febf7,0xebffcfdfdf7fecff,0xfefffeffdf7fefff,0xffffaf7fdf3fefff,0xfebebe7e5f3ceeda,0xdcd73faac41ee854,0x2d4cdbd64d3eecb0,0x4000030180280804,0x24c008280040,0x0,0x0,0x50000008002000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2020080000800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x41014240108082,0x0,0x0,0x400400004004000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4e45021412000e0,0xfbed7fe151bacfff,0xffffbfffd9bacfff,0xfffffffffbbadfff,0xfffffffffbbadfff,0xfffffffffbbadfff,0xfffffffffbbadfff,0xfffefffffbbadfff,0xebfffffffbbadfff,0xfffffffffb8adffb,0x9f9bcd7bfbba5fff,0xf9fff7f3ab3adffb,0xfefefdfbdbba8ffb,0xfbffffbefbbadffb,0xfffffffffbbadfff,0xa7bedff7c9ba5ffb,0x6d14136bc93805ba,0xcdbeefe15b9acf37,0xffa6efffdbbadf7b,0xc96ab639088a07f7,0xab8e5b58d30885fb,0xf0f7ffe7798a9ff7,0xe9faef69199acf7f,0xc2a27fa691ba87b2,0xce47ef29eb8acdfb,0x2bdfffdfd9ba5ffb,0xdfff5ff3fbba4ffb,0x592caf7871029bfb,0xfab7bf7ffb9a1f7b,0xcbd2e764499acf73,0x7edf3e52db0acfb3,0xb7bf77f253b29ffb,0xf9efb4f17aba9ffb,0xfb967d77f99acf73,0x2800e33a28180712,0x7fdf17e7eb9a8f7b,0xfff9cffb9bbadffb,0xb7bc7e6e3bba17fb,0x5ede7d61e99adffb,0xcbf3ff7ff3bacf7b,0xfe5b7edb4baadf7b,0xe71ecff7ca8adfbb,0xbc77feead99acffb,0xffbefd7319b2dffb,0xff8dffffebbacd7b,0xefdffe3f78ba8dfb,0xaf5fdfdfbb9a8fff,0xf5ffdeebcbbad7ff,0xd9eff7fbebbacffb,0xcabebf57410a4f97,0x4d57ffbec93acffb,0xefffffdbfbbadffb,0xbebfffeff91acfff,0xe97ef769012a5ff2,0xfa3c2f3ed28a8d7b,0xeffffbff5bb2dfff,0xade969dd32a075f,0xffb3df7a61aa5f5b,0xf4efbf5bf93a9fba,0xfffefffff9bacff7,0xabddf77e59bacffb,0xeeef7beb739adff3,0x5fbfdf70f99897fb,0xf577e7ffdb2adff7,0xfff67ffbebbacffb,0x7bd4c72a0b8a4fbb,0x2cd3329bd08a0d29,0x5bb8453c288171a,0x7df6b778d91acfbb,0xddbaff71393a5df2,0x7bbbff2e593adfff,0x3fcef61f89ba0f2b,0xa4c15db288a4d02,0xd4ecd6a143aa05fa,0xdd96206662020b10,0x1c30a345a08007a1,0x2858500011821110,0xab1e4ed1f2924fb1,0x3ff33f6b4918d7fb,0xf9269fded11a8d73,0xced235d8808a1d93,0x20c4f3f4020a89ba,0x4bcb1a5903a8c9fa,0x849bde93413a85eb,0xdbfe97094aaa9ffb,0x7f8a7de2cb9217f2,0xebc7fee7f9badf7f,0xfffffffffbbadfff,0xfffeeffffbbadffb,0xdfb7feed1aba8fff,0xcb4ebfea839a9ff9,0xdbf3e71a619a8ff3,0xe6fbffebcbbacffb,0x3b9e4f5a993a5f7d,0xa5fe16ec90aa857b,0xdfee7d71fbbadfff,0xfbf6f26ffb3a0fff,0xcfcbd4b1c9084d1b,0x8f6f590023129df1,0xf7bbffff799adf7f,0xbb6fb77bd13a4f7b,0xdd367f5e131acddb,0x88d1baa0113804fa,0xb75ccfdba01a87ef,0xdcde76efdb9a4fff,0xf5bcaf64f30a5dfb,0x8ba6266ca008dd3a,0xa167c008c1184f23,0x29a9a4f6912054a,0x37b54feb6bba5ff2,0x9cf79668309a4f69,0xcf92f6306918057a,0xcb95ebd37b2a8d7f,0x40c2081020001110,0x838a0f9a1b28577a,0x15ef94e0909a0522,0xcdbefd62d89a4ff2,0xca10430e802805ea,0xa043ba98819a1453,0xaf9ecfe3399acffb,0xfcf7ffe9ebb897fb,0xfffeff727b9adfff,0xddbce3e7689a9fdb,0x8a579a0a41088712,0xabbcdfdbd12addfb,0xf4c97ee5d11a9f73,0xb15b95710b08957b,0xcb1e0ca8290a81b2,0x816ac0088908050a,0x2e939fc600928fbb,0xb6160460019a0d42,0xd600d171408a052a,0xca082530c11a052a,0x41c68c0401089513,0x24644c4008000f63,0x878b6772091a8faf,0xe77cd7f903328fbb,0xfffe7ff3f99a9fff,0x6956bb8e001a0f1b,0x4e5938d399880dbb,0x363fd5e2a03a8de9,0xfefcde7f7918cffb,0x0,0x8000010001000002,0x0,0x1000010000080020,0x0,0x0,0x0,0x0,0x0,0x100004000000000a,0x0,0x4002000000408,0x0,0x0,0x0,0x0,0x8000010002000010,0x0,0x108000000000402,0x400000000440,0x0,0x8000200000100400,0x802000000210,0x0,0x3000084000820800,0x40100000080100,0x28b4b5028b02ddf8,0x0,0x0,0x0,0x0,0x800440201200040,0xcb022a6240124f13,0x488208000108403,0x0,0x2004000000020800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000402000010,0x0,0x0,0x0,0xc1804b56901a0ca6,0x8090f843cb228d30,0xff5e9ffecb8acffb,0x2bbacfdac1ba5eb3,0xfddfb671eb9a8f7b,0x47be2de051ba5ffb,0xcbf4dfaad9bacfeb,0xc94b9eb9411a4fbf,0xf7bcdfffc93acfff,0xfdf7fffafb9adf7b,0xcfbbfff3fba8dffb,0xb8ccefff7a8a1ff7,0xc77ffb33e9921f7b,0xbbbfdb5e3a98cf7b,0xfcfffffffbbadfff,0xfffffffffbbadffb,0x8a34af2a910a1f5b,0x800020000000402,0xe6c10009200160,0x8002000080000400,0x800000200000100,0x0,0x0,0x0,0x0,0x108802000104400,0x680c353041284c20,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x210000000004400,0x0,0x40211000100000a,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xbbeefddb41336ee,0xfffff7efbcfbffe6,0xfffffffabdffffee,0xffbfbffebdfeffe6,0xefffffff3dbfffee,0xfffffffbbdffffee,0xffbfffffbdffffee,0xffffffffbdffffee,0xfffeffffbdffffee,0xebffffffbdffffee,0xffffffffb5fffeee,0xffbefff33cff6eee,0xd7f6ff7d34bbfee6,0xf4aeff6334ff7eee,0xcb56e3bebcfa7fee,0xcf7b3efbb5be5fee,0x8fbedffbbcbff7ee,0xd1f7f6ea0dbaffe6,0xdffffff33dbfefe6,0xfb9ff3af35ffffe6,0xcfdffe3734fb6fee,0xab5f7bdbb4bb6dc6,0xb4f5defa9cfbffe0,0xfba77f6334d25ee6,0xfabe3ff7a4fbffe6,0xffffff7eb4b7ffee,0xefffdfdf3cbf7bee,0xbfffdffb3cbb7fee,0xb97eefe19c9b776a,0xf6962bfa1cbfede6,0xfcd3dfac94fbfae6,0xaffe9fd73cbfffe2,0xb712fffab4d34fe6,0xf5feff799cfaffec,0xa2da7d6eb0bee6c6,0x8ae43fee303a6ae6,0xfefefadb35faf7e6,0x3f91eff01dbbdae4,0x96f2f7ee1cba7fc6,0x73be7f631cbefee6,0xfbdceb3a34feffe6,0xaed3fbdb349b5de6,0xa79ddffb35fb7ee6,0xd7dfdfff34fbfaee,0xfffeedf33cbbfec6,0xfeb9f77f34bb7ee6,0xefffffffbdbfffee,0xbbdffddbb4ffffe6,0xf7f6fefb3dbffeee,0xf7ffaffa3dffffee,0xffb77f7eb4bbfbea,0xfddfffbd3dffffe6,0xefffffdfbdffffe6,0xffb7ffe735bcfeee,0xff7def7fb5f77bee,0xf69d2f461cffe8ea,0x4752df913cb672ea,0x23cbbafb341375e6,0xbffa7ffbbdffffe2,0xf1ff7ffb15bb78ee,0xfffefffc3dfeffe6,0xf9fef30d94bbeeee,0xeed73efdb4bde7e6,0xffbffdffb5bbfee6,0xd5fef6f334dbfeec,0x5bae7f6314be7ee6,0xf3d4e73e34d854e2,0x8fe21e10349a03c0,0xfeffef533c9ffeee,0xfd77fffdbdfb7fe6,0xeb71ff7315b8f7e2,0xfa95ff2b3cbaffe4,0xaf5ff33e1dfb7fe0,0xa392dd5b3c987ce2,0xb6ddd6ebbcbbdfc6,0x4922af420c1250e6,0xebd07e360cd3fa62,0xc8db7330bcbc7ee4,0x8ef95fdab4b7fbe6,0xfe7ff7e3b5feffe6,0xf7c6c6761cf2ffe4,0xda3a1759adb95b42,0xae87f76e94bb5346,0x7bdede498dfe72ee,0x808ace1a849874c0,0xa0301c2a84da6100,0xf7862f6c34f8be84,0x8b7ff3aa3080eec6,0xefffffdebdffffee,0x7fbfefffbdff7fe6,0xd76e77fb35fb7bc2,0xeb3e6d7a34bf7fe6,0x9b26ef2e34fb3ee6,0x4edbbada84f37fe6,0x3aa1cf48148bc86e,0xf5f77dfe3cbeefee,0xfffffdf1b5fb7fe6,0xbdad6fff14ba7fe4,0xff9efaabb5f3faec,0xc99699511cbd6cc4,0xf4f3766685bfe2e4,0x9fefad7a9cf3eeee,0xbf9d2f6e0cbaf3c6,0xcb53fbba2cf87bc0,0xe1ff9ac49c1a74e6,0x94d67df294bcf6e6,0xb9ef8f6835b3efe4,0x4a963f75053933e6,0xfebfff7e3cbaffc2,0xe99fdc80499f4c0,0x20aacde100bb07c6,0x904d2461042226e0,0x21025f0209038c4,0xa9e4470604927bc6,0x4d14f4a04126ae4,0x539d2e01cb9e2aa,0x90c0d25220946cc0,0x92e4d6005b337c4,0x2bd6c71e3cb8ffe0,0xcc62321f3c9d17c0,0x2ebdcc5b1c3876ee,0xf5fefffbb4b3fbe6,0x93ae2f4034a7aae2,0xda222fcf3487fa68,0xccf6961f149b6ee6,0xffffffdbbdffffee,0xb77d56e19cbbeeee,0x7d7eff6a3cdfeee6,0xdb12fe7b149a6ec6,0xc8d6f612048aefa4,0x26e134004aa4840,0xe6f677e424937de2,0xae0260140006c0,0x663a2de10cd85020,0x8bd5be04348e25c0,0xa376163024d36ac4,0xeb046e820bb7142,0xc706ac6a14ba7de0,0xa4886c5d10ae2640,0x8082000f0482e060,0x2d4030018c123484,0xf7ff57f23cf9fec2,0xec58a2b104db6860,0x0,0x0,0x0,0x200000280,0x0,0x200010000110000,0x0,0x0,0x0,0x0,0x8010000000101000,0x0,0x0,0x0,0x0,0x0,0x80010000808000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x12000000280,0x0,0x10040000102000,0x0,0x2010251010030262,0xc004652010132560,0x8000020204004020,0x0,0x0,0x2004000110000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8010048000c84000,0x40162d7214d90402,0x182e2621cc8dd40,0xeed33c523c127ee4,0x49bec9eb1cd8fde2,0xf477eec63cbaeee6,0x6b3a2d6110137ee6,0xbb95bf213dbb5ae2,0x8ec3b6381c3377ee,0xcbdd5fb0dba39ee,0x7575c7ea3ca6fd46,0x1c6cbf3034bb7fc2,0xf980276f3cb86f60,0xc0c3fe2934bb74e6,0xcbdfcf5e3dfb7bc6,0xfeffffffbdfeffe6,0xfdefa7ea3cbb7fe6,0x4000241011280400,0xc0000101000,0x48081134102040,0x8032082000d42280,0x700100000c80,0x8040000004040,0x0,0x0,0x48000000480,0x84684d200c2803e2,0x0,0x0,0x0,0x0,0x8001000010000400,0x0,0x50000000280000,0x0,0x1082010c900000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfde7ffdff0dc1efb,0x7fffdfdbfe7cd7fe,0xbcbf6dea74d47cfe,0xfdeffff3fefdfefe,0xffbfbffff6fdffff,0xfffffffffefdffee,0xffffffdffefdffff,0xfebbfffffefdffff,0xfffffffffefdffff,0xfffefffffefdfffe,0xebfffffff4fdffff,0xffffbfdffefdffff,0xbbbfeffbb6d45ffe,0x97f7ffeffe5477ec,0x77befd71f4f5f7ee,0xfbdeef7cfe9d76ee,0xadda76f7f4d4f6ee,0xdbbfdffffefcd7ff,0xfcb777feb6f154ea,0xbbfe75f3dc1dd66b,0xfbb4f73ff630d67c,0x4a5ff711bcb476e2,0x8e7f57d83e3557ea,0xf4f7fffef4fddffe,0x9beef761bc74767b,0xfff2bf7efc3cff7f,0xdfffffbefefdffef,0x6fdfdfdefcbc7ee6,0x9e376febbc15d5ed,0xd9f9f77e5c5157fc,0x9bc77f7ff0d01efe,0xcff7ffbf36fd7eff,0xb67ef7cab4515efa,0xd7dbf7fe0efddffa,0xf4fdbffcf670fffc,0xffb62df1b4b0f6ea,0xe990d64476b4147a,0x2fffbe4bf454dcee,0x819c4f33b01c5ee8,0xfdfffffff6fcd6fb,0xdefe6db2f615d6fd,0x895feb2d7cf4dcfe,0xa6debbfbfc7cd6eb,0xf7fe5fcbb65ddeff,0x75f5f7eff4f4fefc,0xbbbeff7276b4fefe,0xfbbcfb7efcbcd6ff,0xeffffffffe3dfffe,0x7f77dfdfb4f5def7,0xf5fffeeffcf4ffed,0xe97f7f76f6dddfff,0x9b9c7637de3cd4ee,0xfff7ff5afcddf6ff,0xaffededf5c9dfefe,0x95af7ff39cbcfede,0xd1f2bf7ef450d6ed,0xdebd2f77b65cddfa,0x6ffff77fb6fcf6ee,0x2e763f5b7cf4d4ed,0xb21b4fe39255d7ee,0xd26337ff14f55cb6,0xf29a2f7fe6b4def4,0xeb8dfb22e63d5f6c,0x6f7f3fdffefd5cee,0x9dbedf7bbcdc5e6f,0xd1f7b7ff3cbcfeee,0x7b9ebd617474dcfa,0x8988c30d843c662b,0xffdbbefd363c5fff,0x893fcf532c54de7a,0xf5beecec8c197e78,0x85e72f708494d4fe,0xe226f76ea63cdffa,0xabc7b68aa0f4dc7a,0xab5bdddb28905cab,0xe4f7fee024b454f8,0x85cf7e62dcc4f1ea,0xb2102b70a05c1d68,0xddd79f1eb4d474ea,0xa9fcfecb54b5f673,0xecc277e8b490546d,0xf7e8ff756c14d6fa,0xe93b75ca434f4b8,0x7c6ee7ab4f4d6e7,0xaf7fffc5d6fc5eff,0x88bb4fdb841c3c9e,0x7544543bb0145ee2,0xd6802f4074214568,0xcbf4e5b7d4947c7a,0x6ed75efd749dd6fe,0x97bddffbb6f55cee,0x9c557e0d04547668,0xdbe5e72d69c1cea,0xbbcdef7e96f474c8,0xceaa13bf3c5c15e9,0xbeadef4ad01d54ea,0xf6df3ef476fcd6ea,0xbffffdf2f6fcfefe,0xfaac6d7af474fefe,0xdef67e8b7434767e,0x6bd8c9d3a434d46e,0x11b27ec6e094def2,0xf96f57695654d4ea,0xdfbe7f725c94dcba,0x5fdbdbb0b4f47e7f,0x6fdfdf5174107d7e,0x14df7fff543dfefa,0xdb7367f796d57cfc,0xeab5ab77765c76fc,0xc4d3d04a3615766d,0x5c7effce34bc767e,0x881c5efaacc4142e,0xb025b561b67453e8,0xb3265e70e01056e8,0x794eff2afc9cd6ee,0xe84711fe303c42b2,0xc958d61541474ac,0xd0e27761408414e2,0xca86d80d2151268,0xb995df30d434de76,0x4b5336c9cc5c1ab6,0xafdb4f68e43456ee,0xbc73f7eff69457ee,0x9ba7ad70fc545826,0x43a442a716155566,0x8848120194041402,0xffffffdff6bdf7ee,0xb4fd7ee5f494f6e0,0x696e2f761cb45cff,0x96162d60649054a4,0x2e84630a34b854ad,0x6f77dfdefcdc16fe,0xa9300e8212981240,0xb03eeff5a8dc56aa,0x8ba225c8561c06ee,0xefdffbbef614defd,0x2877de74b4145f6e,0x1420e16000905441,0x910584cc14903c88,0x3aec2e6466a0146e,0x8a80010040004420,0x621a02813010c420,0xb71c0cfb2e30d5fe,0x97d29f58a4c456ee,0x404000000480,0x0,0x0,0x0,0x0,0x0,0x4100004000004,0x204000620001028,0x8000240000020,0x0,0x0,0x0,0x0,0x0,0x808440000401c0,0x0,0x0,0x2004000000100200,0x0,0x0,0x0,0x0,0x8100a00000101802,0x0,0x0,0x8010000000001080,0x2010000020000020,0xc00008001000,0x0,0x8090080001028,0x1000000204001000,0x9000000010000480,0x81a008161044140a,0x0,0x14020000020,0x0,0x8400000002000040,0x280000000000040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x7efe667da0f55ef2,0xc880c37b441d11ea,0x69d6fd9e961c1cca,0x7eb86ff226dd5cff,0x551eb6f896541eef,0x77b65d526634f7ee,0xc21ce330b41416eb,0xae038e10e01015f8,0xe8c5f3a3494143a,0x25b6d5f8a64cf4ea,0xc3280d00281c502a,0x4020462216104440,0x8d6af0ea763474fa,0xa7aa5fdaba14defc,0xfcfffefffedddeff,0x89fd37723415d6ef,0x560c10dc10151ca8,0x88436e040634d2d6,0x8861a0100045488,0x2c7e4cec001cd6c6,0x1088a0002000c020,0x8c212014001402,0x0,0x10020000040002,0x8010402000181400,0x20004000104000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80420000004040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xffa6ffffd9cfa77a,0xffffff3fd1efb3fa,0xffffffddf9efb3fa,0xe7ff7ffff9edb7fa,0xfdefffebf9edb3fa,0xfebf3ff3d9cfb7fa,0xffcfeb7ff1edb3da,0xfdffbfdbd1e4b3da,0xfffffffff9efb7fa,0xfffffffff9efb7fa,0xfffefffff9efb7fa,0xebfffffff9edb7fa,0xffdfffdbd9edb7f2,0xbffeeffff9efb7fa,0xddfffffff1ecb7fa,0x6ffe7f71f1cfb75a,0xfbfeff7ed9efb7da,0x8dc3bb9bf1efb3d8,0xaffefffbd9edb3ea,0xd4ff77fef1e493fa,0xffbeff71f94fb3fa,0xfab7f75ed1edb7fa,0xedcfbf1fd9eda3ca,0x6b9fdfdfe9edb77a,0xdcfffefff9cda3da,0xd9deef7af9cda2da,0xfbbfff7fd9efb7fa,0xd7d3fe2bf1e1a3e2,0x1d9c9d6f9eca3c8,0x34ffdfe9f9efb3ea,0xe96eff7bd86c23da,0xfff67f7fd1edb7da,0x6ba3f7bbd9eda6fa,0xfbf6cfebf9ce23fa,0x96b07b68d0c023e8,0xfdfff77bd1eca37a,0xd7fe3d64f0edb7c8,0xd9cc7736e8ed23e2,0xeeb6974b51c5a35a,0xa02cdcf158652368,0xd577d6fdf0eeb350,0xffbe6de299ed975a,0xfbd7ef36f9e5b3da,0xe253bbdfb1cda7c8,0x6fbecffbf1edb3fa,0x77f7ffeff8e7b7d0,0xd7bffff3f1edb3fa,0xcbb9f37ef0ed33fa,0x4fdffffbf9edb3fa,0xefdedfdbf1efb3ea,0xf4f7f7eff9edb7fa,0x59ceff73f1efb3fa,0xdbb8ff7df96db3fa,0x8dfe7f1ec9c5b2da,0x2ff6dfdbf1e5a2ea,0x3cb37fefd9edb342,0x35edae48e1658360,0xf59dbbf0d1eca6f8,0xc9ffff22d9edb3da,0x8ed6dfdfd9e5b7da,0xbf774de8796db2ca,0xd4ffbedad16923d2,0xd5cc7d7170ec26d0,0x8bdcf6fed1c5a3fa,0xffe796dbf1e9a3c8,0xb9bfcfeff1efb65a,0xfa74b6fa91c9a3f0,0x4d9c6fe2d94e32da,0xbbe4bf3fd1c1a65a,0x9e7f7e53f9c5a3da,0x33384fd1d1403688,0xe15f66e8f16d27d2,0x33fefd70d9ed93d0,0xffdfef6ed9e7b67a,0x88fefe03c1cd237a,0x229e5fdbd1cd365a,0xfcf72474d9cca548,0x984a7540c1e4a2c0,0xefbb7571d1efb372,0x4dcfb2afd0e9b6c2,0xef5e4e9ed1c5b16a,0xf6b37bfa9165a34a,0xf12eba6ad1c687ea,0xf696256cd1cd2358,0x8188f127602ca252,0x9d6d6456184a2fa,0xcfe24deac1cd32da,0x81092622104091c0,0xf28c0b5041a483c2,0xa9f0fae2d9edb3c2,0x5fee5fdff9ecb3fa,0xb7f94febf1c6b7fa,0xcf502671d1c01678,0x7bae7d7bf9eda34a,0xfbfffb3ff9efa7fa,0xccfbffdaf9efa6fa,0xbbbeefdff9efb2d2,0xf777f248f1c5b352,0xbf7afc42f965b3d2,0xdfbd7f7ff96db3da,0xc2c3daebf1efb35a,0xae7fdfded9cfb77a,0x14b77ee1f14fa3f2,0x8c40ac607867b640,0xbbb67fb6f9eeb6ca,0xcf57db0bf147a7da,0x8f5cc9c3d96da3d2,0x8cb3f7ead96da2fa,0x9ae5ff40c8cfb3d2,0xdf9f176451c5a6d2,0x8ff8f33cf1c4a3f2,0x69fa6ee469c722f2,0x829287bbd1c5234a,0xd2adbe78594923c2,0x8a06302671442358,0x8a83b2a25948834a,0x62889672d129a248,0x989194d0d0013040,0x9474ee09f96623ca,0x100020080800000,0x2944c13cd10ca2c0,0x8950935a802d224a,0xcf12cf70d0e5b3c8,0x7db6fefdf1af23e8,0x89149f30794d13c2,0xd914b33bc84da2c8,0x5fc33e18f8e523d0,0x2fdc8b9bd160b742,0x38d97169d960b2e2,0x1fb6fffaf1ecb3fa,0xf014b31c98c1a240,0x2edef0aad12432c2,0x48581f42c1a91348,0x2c1146e051cc129a,0x813c002091002148,0x56a28f06d140a3c2,0xafdfffaef1efb7fa,0x2b480940d940a242,0x262ace62f1c41348,0x154425c8d12c1002,0x8a8c25c9c14b2362,0xab4cd10191e50348,0x79731283f1a58242,0x15fa8f8e61eda2da,0xc4643ece51c4b362,0x8040000000140,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000400000400200,0x300000200040,0x0,0x0,0x80004000080200,0x0,0x0,0x0,0x0,0x0,0x200040040000,0x0,0x0,0x0,0x4000200010000080,0x100800000040008,0x0,0x0,0x41000040000080,0x0,0x8900e820510481c0,0x80000200000140,0x0,0x4000100050000000,0x0,0x0,0x0,0x0,0x0,0x0,0x1004000080000040,0x0,0x4000400040008,0x0,0x0,0xbba63de3f949a2c8,0xc9aff13bf1cfb3ca,0xebcf8ff9d1e5b2d8,0x8dbadfdaf1eda35a,0x480526b5940a3c8,0x6bda7d70f1cdb2ca,0xca2acb0ad9463352,0x8ac2db1ed96583c0,0x69ae4c6041618368,0x84c616a5c125b248,0x93820740884ca740,0xc99be1bbf94ca2d0,0xc9dfb39ff9edb34a,0xab7fffdbd9efa7fa,0xded7fefef9edb3fa,0x11342101d041a000,0xd2ae2f6ed9c9b37a,0x21050000000,0x28091880114c0202,0x4b857ced1052148,0x8040a710186003c0,0x40802840500022c0,0x0,0x108000040000080,0x160b5ec0d16903c2,0x510080082200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2100000000240,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80800000082000,0x0,0x120010080082000,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x37ece7619aacef24,0xffb7fffbdbbcfff6,0xffffff3fd9befdf6,0x6a7bb7cfdab8bdb4,0xa0041c0002800,0x0,0x8880620c0002004,0x800020008000002,0xefffffdfdbbeff66,0xffffffffdbbefff6,0xffffffffdbbefff6,0xfffe7fffdbbeff76,0xfbffffffdabefff6,0xeefffffbdbb8ffb6,0xffbfdfffdbbcfff2,0xf9eff7f3d9bcef76,0x7ffe3ff39bbefff6,0xbbfdff7e9abef5b2,0xeeebbbc89aa8eda6,0xcfbfcffbdabced36,0x7582fef4dab86d24,0xfd7fbfa09abefd34,0xdc84ef74d828e570,0xcdd3feb1d8b4fde4,0xabdfce4bdabcedf4,0xd4bdffeddabeedb6,0xf9f7ff73db2ccf62,0xffbeffffdbb8ff26,0x58d7db1adabebf60,0xabcff3db9a38ed64,0x4fbecf7258bebf74,0xd1fdaf52cab86f64,0xdbbe7f72dbaaff74,0xffffffff9bbeffe4,0xffffffc9dbb8bfd6,0xb4a7ffef9ab8fd66,0xfdf5bfeedbbeff24,0xefbeef72dbbcf572,0xeb84f23adab8e564,0xe9d6ffd2db8afd24,0x4bbecffadbacfd32,0x74fcf2f0da38ed22,0xff7e2df2dabcedf0,0xfbffff3bdabefd62,0x8eda3a9e1aace5e6,0xbfffdffbdabceff6,0xf5ff76fbdbb8e5d4,0xffbeff30d8beed76,0xdebeeff6dbbcef54,0xfffffebfdbbcff76,0xefffdfdfdbbcfff6,0xf7fffffbdabcfd76,0xcbeeff62dbb8ff76,0xdb96bf069ab87710,0xc7c2ee3bc220ef44,0xe6cd4f59db2cefd6,0xb4cf57ed9888efe2,0xddeebf49983cf574,0xfeb6bf6e9aacfff6,0xedfef73ad2b86f74,0xebefffcfdbbcfd66,0xbb337ffada884f36,0xf9edbf748ab8ed60,0xff4efe2ed8bc2770,0xa3b6f7afdabce7e0,0x7effffdbdabeed60,0x3fb7cffa9ab8ff14,0xb475fe7a9b88e596,0xd7be7df1dbbcef32,0xcbd5efe3d9bced36,0xf3c0165bd8b2a502,0xc6bf4fe85bb8ed54,0x17ebef6098a4fd22,0xffdefd709bbcffb2,0xebf3e757d8baed66,0x8ddffea7d83cef72,0x1b58cfdad8b47d42,0x94b77eeddab83562,0x8ce5e75399a89db2,0xf69d2e535b88f700,0xe8f1d31099a8cd30,0x4975525110b8e510,0xb5fb3fe39aaced44,0xd51991505aa8be20,0xf6b225609ba8ce90,0xef80f6a399386c24,0x8deb613912cad40,0xbf3b7f60d9b8bf74,0x325d967ad82c6420,0x2bba7d609aa8ef20,0xab9ffff212bcef54,0x7ed39ed298a0ef92,0x9a9c67e7d832cdf0,0xf4a7546dd8be6550,0xffbe7df9dabeeff4,0xebd6f30edabcb5f4,0xacefbaccd8bee574,0xa9becde39aa8ffa2,0xf9d777feda3af566,0xfaefec72dab8fd64,0xca8c636312a0e564,0xde5232a0d238c500,0xffdadddfda3ced96,0x35ff56e4d828cdb0,0xdfefff7b9ab6edf0,0xdfa17c77dab8dda2,0xefffffb7dbbcfff6,0x2ed7dfca9a38ff70,0xb4fbf6e7dabce7e6,0xd1fcffcbc8acefc4,0xfa9e3fee9ab86506,0xedfbffbedabcffe6,0x6eff3ecf9834ef66,0x14bbf7e0dbb84766,0x927bd7314ab86526,0xf7f27f7f9bbcffc6,0x8152e10a802c4c20,0x4dfc7e0eca282f46,0xfebd4fa898bafd14,0x1700f115d2886534,0x4dba2ff09ab8ae24,0xda9ee76f92b4e566,0x8e4abb5e92bc2f22,0xfbb2cff35bbaef22,0xdcffbef6dbbcfd74,0xfff8fdb2d8287526,0xfab4ff7f1a2efd44,0xaa47da0cc928f506,0x216c759da20a592,0x77e676e2c3ac6d62,0xdf66a77299a8edf0,0x5e8603cc90aa2502,0xa00fe0ce92bc8d00,0x4806000942002504,0xba84a100280760,0x814b85700024f560,0xba1a07709222c722,0x403dd72890362402,0x673ef450103a6544,0x266b47e09020fd62,0xd576f4dedab86746,0x9a8e496c90b02d14,0xcb27a21e932a2772,0x86f02a53c0286f26,0x88becf9f98a0d570,0xb673f6d4d0bafd60,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x880000c4000a100,0x0,0x0,0x404000000008100,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8104102408206100,0x950310680284c00,0x0,0x74d49e4803a82724,0x0,0x400402020000d100,0x142200000248800,0x480048000000,0x4554d62192ac2502,0x4c388d230838ed62,0x2988401e11888580,0x81e392101a247500,0xa409ddca9188ed22,0x161f72e113b84760,0x9a48e7721820e552,0x13a4676249280720,0xf9cbd3bb08acc560,0x3a46f81b18884702,0xb5ba55f3903c6520,0xd50087080a244504,0x72b6af7408ac6726,0xd180f79319b86542,0x800100240000,0x1c10ee6808a40540,0x1024206100882700,0xf6ac696cd8b8ad06,0xc1bbfe719ab82524,0x46e79670d2909d10,0x9bbac7da9020bd32,0x708c16c24a282724,0x638e7d7298ba6522,0xdb94cb22daa88f30,0xcc82be2999b885a0,0x62bec7c3daa06536,0x6665547110acad64,0x9d8e2c0399384500,0x80a12848006402,0xffeffabfdabaefb6,0xafdedfdad8aeefb6,0xdcf7feedd0aeff32,0xa76e277adbacc722,0x9aa1cb3edaaae500,0x0,0x284311011020b400,0x46410841aa06400,0x212058200a204400,0x0,0x6000000000002004,0x0,0x80400000008800,0x88302990240120,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x102054040202000,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x108000004001,0x0,0x0,0x8000000001020000,0x0,0x0,0x0,0x0,0x0,0xf0e4d44621064c62,0xf9c63773bd746d63,0xa3b27f6c7ef6ff62,0xefdfffbedfdeff73,0x0,0x10001000000140,0xa00000100100,0x800020000800200,0xefdffffffffeff73,0xfffffffffffeff73,0xbfffffffffdeff73,0xffffff7bfffeff73,0xfffefffffffeff73,0xebf7fffffffeff63,0xefffffdfdffe7f73,0xafffdfdf57d6ff73,0xd7fff6ff97f67f61,0xfffe7ff35fdeff63,0xfbf7fb3f5fdeff73,0xcffffb7bfffeff72,0x6fdf5ffeddfef771,0x9dfff6edd7deff60,0x8b9aef715f7e6562,0xf9af7f7f8d4efc63,0x4bdffe9b4fee6f62,0xab3fcfd11bce7572,0xb6f7dfff7dfef570,0xfbc7f7fb5dde7f63,0xfb3efffbbdfe7d63,0xcd47ff3e1fe67e73,0xaffddedfcffe7f73,0x9aff5fcb9bbe7562,0x99ed8fe84d967f63,0xfb9e3ff10ddeef61,0xc3dff3beffca7571,0x2c5efcc84d86fd61,0xbffffdfddddeff73,0xf5f6beef77e6f771,0xf7b67ff9dfd67d63,0x69cdf4380dceff70,0x5c133e8400ce6862,0x9f9b455b5fdef562,0xf5f777fa5bd66f72,0x6f7c6fdb3dc6eb63,0xdbc5cf3eafae7763,0xce5bbbffddf6ff60,0xffbffffffffeff73,0x95fd5ee9bff67d71,0xdb38af731df25760,0xf99eefb775de7f63,0xfffffffffffeff73,0xefffffdffffeff73,0xfefff7e5fdd67563,0xcd6f3ff79dd6ff62,0xdbb43a3b25d67f62,0xce07eef345da6770,0xcfceddb0d467530,0xfcf7757acfeef763,0x97fa2fdd8bc67d63,0xdf863fff95967763,0xdae7fdca5cc6ff61,0xab9efedbddd6f362,0xae786ffeddc67f61,0x96ecb57b11967d61,0xfb7e6f60cdd66561,0xcbf7b3ae9df67561,0x2f5f3ed3d9e67d61,0x3fbfedfc4fee7f70,0x76fff67d7fdeff63,0xfffefffbdffeff63,0xfbdaff7f9fce7f62,0xc243f75b91963f60,0x4bbe4f7b4fe66761,0xf777d6ea3fc67d60,0x9a3ced734d627c70,0xdeb2e7ff7dc67773,0xe5d6bfbb49d67f71,0x8e7fdf5bdfe6af60,0xf43edf3cd5c26d63,0xdbedfe7c59da7f73,0x4f143c6e81963960,0xcdd3f6201d4e4f60,0xc37c455dc31e7561,0x57f777efddde7773,0xb5d4ae60ddc66742,0x44b20fdf49d67140,0xcedbfe365e386561,0x2887ee15a52c7d61,0x343bffefd1b67f63,0x570cb6710da4e943,0xcffe75fcdd967f63,0xebe7fbfe77fe7d73,0xfcd97ec63dd67162,0x1ffa2d7b1dc65f63,0xdff7b44b0d365f72,0x7eac2d43dad66772,0xb3d4f216076e2560,0x9fcfba2e8d5e3561,0x8cbe4c0a1d967d62,0xfc5fb76a97f2f571,0xebfead771dd2ff60,0xcea4017ea5326560,0x8900928005820340,0xa61ec41a05527760,0xd6bed6f55dde6571,0x81006b6218c27441,0xce8c3f762bc05d71,0x8e5ffebe4f46fd63,0xbf5e7fd39792f553,0x879e77ebc7b66761,0xd96cf7f237f65f63,0xfaec2f7647b6ff63,0xeffe756747dc6f61,0x35c7ff5f45962f70,0x2caa4f2829c63e20,0x188100040802200,0xb2be7df091d2ef41,0xe9fdf6cf6fc6ff63,0xe61614860d864420,0x9a2a479a89563740,0xdcb0f375a9cca570,0x4d467df0238e4e60,0xf254e02209c27140,0xe8bb21f13861570,0xbdff6ff0ddd67d52,0xf57ffefedfdeef72,0x9e9eff720ffe7f71,0xb3806f4d2c44fd70,0xe5c6f68a4dd42f70,0x19bd9d387d2f762,0x94b226ed0fc27f73,0x9d73f7611ed65f73,0xd09e2d443dce4760,0x8067222803065c61,0x6a74174c91a6a660,0x10715d4009c24e60,0xd979f7741df65f60,0x10842e3605821740,0xf44960407024561,0xbf5cda4b09865660,0x87b842f39cd66b60,0xdce65d0210927060,0x6666256011e64f60,0x8101a32208c24a60,0x484a337121d27f60,0x8c4f2100c06b62,0xd434546000824a62,0x0,0x2204400000802200,0x50000001000800,0x120000008400000,0x0,0x0,0x0,0x0,0x0,0x20004000100100,0x22044000404040,0x0,0x80800000000140,0x0,0x0,0x0,0x0,0x0,0x210000002000400,0x0,0x0,0x8222256005024740,0x8840810008400140,0x848004001820100,0x90174d6005920340,0xfdf6d6f894963740,0x344342008164140,0xc9d593008a462c20,0x2c6bbb4010807540,0xa58756b824882761,0x315c5fe2d5c06763,0x94362570d59a6162,0xdb32cf7a878e7d62,0x987978f1dd66d63,0xb2de0dd21dd60160,0x8413fce281c67f43,0xe1642f72c5c67963,0x29626a001c4fd50,0x845802408d45100,0x631ccad2d9c46771,0x1db355e61dc66f60,0xc17df660cd426d70,0x96b41570092c6862,0xe5d67707d7a27760,0x80e1001960300,0x1001406002000550,0xb144942529961d20,0xf6b4a55321c07e21,0xeb90570422125571,0x44e9344428844560,0x8480000003000,0x887ef6e94dc47660,0xfb340f4845806942,0xcac0ca444b967420,0x8b5b93737fd27762,0x92aa052a05264672,0xddbc17f81de65f61,0x9a303c0248c63960,0xd31c5b5c8cd66761,0xdbfef6356dd67773,0xf9ff9ddeddce7f62,0xf6bffce77fde7762,0xdfffff7f3ffeff73,0x8a382f2099d22100,0x0,0x4000404000,0x1404100200864200,0x108000000002000,0x0,0x2004000000024000,0x4400000040020,0x80000800000140,0x8108200005260040,0x4800000000840000,0x0,0x0,0x0,0x200200021000,0x0,0x8000200002000400,0x10100001000800,0x0,0x0,0x0,0x0,0x400200000024000,0x0,0x0,0x0,0x0,0x0,0x0,0x420000000820,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xeb7f5fde77bc7ef9,0x0,0x2100000041000,0xdbb97f7ff23f7c6b,0xc1a8d90a0b3a2c00,0x1000020a000000,0x0,0x142200030200040,0xfffffffffffe7ffb,0xfffffffffffffffa,0xffffffffffff7ffb,0xfffffffffffffffb,0xfffdfffbffff7ffb,0xfffeffff7ffefffb,0xebffff3fffffbfeb,0xefffbbfbced43cf2,0xb5accdfb56bebe41,0xd7f7f7fadffd3f7a,0x7abe7ff3ffde2ce8,0xf3b7fffff7fe3fc8,0xcfff7ffffefe7ed0,0xaffecffb57ffbfc8,0x7eff7efeeebd6cf8,0xff7e7d73ffbc3ee9,0xeb9dff6eddbe2f99,0x8ff6fe0ffffebde1,0x81fadf5654b23fd8,0xdcb6d6f9fe9fbee1,0x9778ef77ff9d3e60,0xbbfe377f55bc7dc9,0xfff7ff3efffe7dfa,0xafdfdfdd769f3ed8,0x34f7d2e0ff943cd0,0xdffdbef777beffc9,0xc29fbf775ff7eee8,0x89d7fa08873c2460,0x1b9ffee376fe3f70,0xf6bbfffadffe2f68,0xf52fb7fbde9e7cf8,0xfffefd72fe9efcd2,0xe19bf31bd79b6fd0,0xe4e7163e729eae60,0x9a9d5fdbf29e3ce0,0xb1ded6e05ea83c40,0xdffe7df37ebffcf8,0xc392ef387fff3c2a,0x69c31fbe7ebe3ee8,0xfffffffffffffffb,0xf6ffffeb7ede7ff3,0xb5fe7f777ffffe40,0xfbfcfffffffc7dea,0xffffffbfffff7feb,0xffffdfdedffeffeb,0xf7777eedffbfefe8,0xf57f1f42d2b42c63,0x9382f76c7f3e6c81,0xcdfea61a7aae3de8,0x8b7ddf5756fc2fa0,0xf4177ce6dfb4bfd8,0xf7aaa7524e1e3ca1,0xffbe3ff5dfe5bd68,0xeb53f7bffebf3dca,0xbecdec0c70c2dd8,0x9f7acf33569cfd78,0x89fef470d43c3c40,0x76b62c44da391e42,0xebd7fcaaf79caee8,0xefff3fdbfff6efe9,0xddbe5de0ffdcac7a,0xddfef6f8febe3f68,0xffae7fe25ff63fe9,0xc376fb2e50fc7eca,0xfbf69fefffdeeef2,0xffbf4ff7ffb63ce8,0x58fe77fa7fbdbfe9,0x97ba3f41d39e3de0,0xfea5776f5efc2f6b,0x9bceda875f3ebecb,0xab5edf5a7e1c34cb,0x34b75eeb5fbe64c8,0x9bb4cf72f68e2c40,0xfbbc7f66f79c3fc8,0xccdbdb3e5fbe2ee1,0x1fffdf5b771e3eb3,0xfcff5fffdefeef69,0xd9f5aff8ef987de0,0x50a72d7356242cf8,0xc182f53bfbce1c60,0x8080a41ce003400,0x8eb3d6e0c7fab408,0xf5ec7feb7edafd81,0x7b167f75d7fe1ce0,0xc1e7d77fdbc2aec1,0xee52329bde883ed0,0x8bb2ff7b5e1e7f90,0xd57efc4e768e3c69,0xcaecf522f61c2e89,0xfbf7fb3a74af3ecb,0xde50135656242ea0,0xad3adf747f1efee0,0xd5f657ebc7b22d68,0xfbbebfd177bc76ca,0xfda2a3787ecc14d2,0x49d696201a2f3ec0,0xff1ecf1f5ebe7f81,0xf47676e8569cd6c8,0x5360ff71feb43f52,0xe3a83c23c7987460,0x93b4eb124f84acca,0x4bffdb5bfe9a3e40,0x85b67f6fdfb4be50,0xd3feb7745f9ffdea,0xf6b63ff6c79c3db0,0x8dfcf42656bc3de2,0x2fc6fb9f57f47df8,0x88bacf5bffb46cd2,0xd119f4796fba2c40,0x4394367efe022e00,0x8804e30442200400,0x49521040421c04a0,0x8e032f4972ec2448,0x9ed0787976900668,0xeeac65c3451c1da0,0xa98ccb5863960da0,0xcc7b12963a140f20,0x703a5fe1573e3469,0xfffeb7feffbf7ffa,0xffffff77fffe7ffa,0xdb9eaf5e7ffc2f68,0x8772b7024eb43c80,0xa9c8dcf7a8c2e89,0x84d656a8f314ec0a,0x987de77bce9a3dd9,0x5104690542aa24c0,0x48022202d2002e50,0xe937f2da54ff7cf0,0x342f63c754ec2ee0,0x6824a771cfb06c40,0xb41f40c6926000,0xc0821600542404c0,0xa9c4df4d52906fd2,0x6b0e4f61dadc6e40,0x94f02410560c2660,0x883e2dd3749e2ce0,0x81019f2000b86440,0x376632514e18ac00,0x19a44b2156304470,0x10682e8d4202e00,0x0,0x0,0x0,0x0,0x801408c4000800,0x84000080400,0x0,0x100800002080000,0x4000010000020800,0x0,0x0,0x0,0x0,0x2000008000c0000,0x0,0x1200000000200400,0x0,0x0,0x0,0x0,0x0,0x1a005d40540014c0,0x8180a3251b240620,0x20004082000000,0xbdbbfffa4bbcfd38,0xb7fffee9ffbfffe9,0x90e4c6346b420e0,0x9154ef4e5c3e2ec3,0xe8ca8689d2c82c50,0xaf8a4e5ac68ebcc8,0xd531fae07f983d01,0x94f8ed534e9c2d78,0xfb741b3a4e9eac6b,0xcfd1fe19fed51de0,0xafdfdfde7fbf3efa,0xf4355fe2deb62da3,0xdf8fbd735ebf3ce0,0xe3106e2e5e1416e2,0x8141ce085a962c48,0x29b8d9c01afe0c48,0x3dda15eb5e903e48,0xc1240c4a5e902240,0x448c0542110e0e20,0xe0f5f3b247122dd8,0x1400e00c0083c01,0x26800040080,0x90415040c6300480,0x642a7560439824c0,0xab84c37effb82c81,0xc44b3852529c2449,0x108016054040020,0xa86d624a4fde3c00,0x239a2e307e8028d0,0x80542c0012083401,0xaec2366f7fb4bee2,0xeb0afff8de1ffe48,0xfcfe16e8ff9a3e11,0xc138fc5012103401,0xccd2f6684f9d69c0,0xeffffebf7fbe3ff2,0x7ffedfdb56bc2fe2,0xf4fff2eecfb87d6a,0xfffff77afffe3ffb,0x769e2de00ea42cc0,0xa950e51ada1c2cea,0x2002000002002000,0x14547240120c3e20,0x4120180044a02040,0x0,0x4840c000408e2840,0x0,0x10100000104000,0x8099006212a02600,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x7fbeede3abdc2b9a,0x0,0x0,0x0,0x400002008400000,0x0,0x0,0x0,0x2f5f15d8d334672f,0x0,0x913ac7228bc12702,0xa0060090100100,0xb50fee220ac02230,0xbfff7f5f9bcc623e,0xf7ffffff8bdd2bf2,0xfffffffffbff6fff,0xffbffffffbff6fff,0xfffff7ff9bdc6bfe,0xfdfffefffbff27ff,0xffbbfffffbdf6fff,0xfffffffffbff6fbf,0xfffe7ffffbdc6fff,0xcbfeffffbbdf675f,0x4e7f7edf9bf463f7,0x40bf6fe28bd26736,0xf3fffefbdbfe2f77,0xff7e7df29b9f6f37,0xfbfdf77efbdf6ffe,0xffebfffffbfd6fbf,0xabbedfd28bbc63de,0xf5fffffe9bfe6fff,0xfbaebf759bf5630e,0xdfaeff7d9bbe6ff6,0xcfc6f623bbd4233e,0xbfdedfdbbbde675f,0x74f6f6eebade6bef,0xffffff7bfbfd2fff,0xffb56f7f9bd7277f,0xcff4df3a9add6796,0xebfedfd79afe2ff7,0xbcfedff79a99673e,0xf1edbfdbbb9c2f5f,0xfe973f42bb8023bf,0x8183f1a3aad82660,0xaf7bfbebba766f7f,0x87bed7f19bdd2b16,0xc97df719dad42b86,0xdefeff7abbef6b27,0xd95fff3edb1d275f,0x4eff1f5f9adc2fdf,0xcffefffd9bd227cf,0xb775d664bb0b26d6,0xefb26fe3dbcf2326,0x9b84a77fda9127ae,0xdfcf3fffdbff6f7f,0x7fffd7fbbbd66fbf,0xfc77fffffbff6fff,0xeffead73bbcc2bdf,0xefeef77fbbf86f7e,0xcedefe7bbbbd6bef,0xefffdfdb9bd66fbe,0xb6b7ffedfadd2f76,0x59fc7773bb9c667d,0xdb867f64bb8c2b56,0x8be5f635dbcd2312,0xefffdfdedbbc2fff,0xb6ff7febdaff6fff,0xddecb6d7fbdf2f27,0xfe962f739b9c6715,0xc021f3061acb2f26,0x3a8ebe5b9bdd0336,0x8e124df3f104222a,0xd54ce4f6f281653e,0x7df47d7bbbb5233e,0xc9f7fa2ffbcd6317,0xefdf5fdb9ad6633f,0x3dbddffbfa9c6f3f,0xc67677f79b9e678e,0x1ffefff0fbdd2796,0xfbfd772ebb9c636e,0xedefbf5dbbdc63fe,0x3f9bcff8aadc2f9f,0xdd7ffefbfbd66faf,0xdbd6ef7392962f76,0xe5a67f7f9af42f9f,0xcffef63bdb9627a7,0xa3dfdfd3abc9275f,0x34f7e6f09bcc2b0f,0xa4af2768db88210e,0xff912f7e9b0c63a7,0xcccff0169b8c270f,0xab5ddad39b8a2be0,0x94b57fec8ac0230a,0xf2dff51f9bd46b2e,0xffbe2f7f8bf46f4e,0xcfdfdf07f214672e,0xcc0f459ac00036,0x25bacff6a384437e,0xc52cbf9adb9c6f54,0x53363ff09aca6b5e,0xebdff30adb9c27f2,0x6e66349f93944312,0x82ba8c938bc223c6,0x79fcf67ecbd86b5b,0x499a76789bd00b22,0xb98df67cb382490e,0xdccb3703bb952726,0xffffdffbbb9c6b3f,0x7154e7e39394673e,0x3ffefff3fbfc6f1f,0xbb966f7e9b57675f,0xa3c7f81f9b8c632e,0xee74ed839b162b5e,0x56d076ed9bdc2f5c,0xd72a9f6abb402b76,0xffb26e6e9b802576,0xc3f6e624cbc06724,0x8f7cdc439b886b27,0xacfe7fe79b7c2b26,0xd9eead429ab62bbc,0xf8febfe29bc76bff,0xddfaf70eba1c6bad,0xff9e5f4ecb8e6722,0x8f8a6e6db294636e,0x70cfa7d9d914234d,0x4dc2437f9bd12346,0x8894f3e58b3f6302,0xed4e356793306964,0x799cde89bfc2b22,0x93f3de629a9c2335,0xea343dd08b10235f,0x8c8c00882002404,0x81c21af8ba080110,0x8b0c4b069b88223d,0xfffffffffbff6fff,0xfffffff7fbff6fff,0xfeafffffdbdd6fef,0xd947af13f2122705,0xef3e6fdfb3ae6bfe,0x25b0447493902106,0x8549a3b09acc0b82,0xfb8a4de49bdc6fb3,0x4ac3333892816b32,0x231d5cd892906e67,0x13cc77b92800b4e,0xc12c2b4199802322,0x88002578a2802100,0x86cb32449bd06716,0x2e9ddfce9bd96327,0x301a2eb3828422a6,0x9c66ac528b0403c4,0xb816143099804306,0x8174c52ad3002137,0x2c69168092900322,0x1581845013880322,0x81409e29cbc06037,0x0,0x0,0x0,0x0,0x0,0x8010010000004,0x0,0x0,0xc000008000200,0x0,0x0,0x0,0x0,0x0,0x0,0x801002000002,0x0,0x8800100280002100,0x2200040002000,0x200040000002001,0x204002000004,0xc0410091000020,0x8080891181080314,0x25002020004,0x200010082000000,0x404000080100000,0x4186c5803020106,0xb920f300b99423dc,0xabefb3869bcd2137,0xadbedd47ba582b9d,0x77f256ef93006bbf,0x82a2ff51abf02744,0x9aa6332b93906936,0x45499e09b0a02036,0xaededd5b9b9e6b25,0xb5bb66e09b5d6f1e,0x179677a98c46306,0x681436149b900322,0x8c6020aca802113,0x200897da9b702a16,0xb13fc2b08ac00316,0xd168ad44d3002d06,0x2040e2090084204,0x80431080802030,0x301084093800004,0x4005c4000a02204,0x8070ae50c2122714,0xb10592000124,0x8292d76e9a502b47,0x22500420a480202,0x8310021080180210,0x98122a0012010b06,0xd3ac0522931a2304,0xc1f0c75e9bdc2726,0xdccffe9abba02fbe,0xa29e5e12d094232f,0x75f5536a8ad02316,0x9a0d7de08b8f2736,0x4028e6068b880112,0xcbdab6b3d3922ea2,0x981a0fd2ba946326,0xdcffdffb9bdf6f17,0xffffef7ffbff6f7f,0xf680361282850106,0xc09a200880d2002,0x8040182040200,0x42446409200010c,0x8058a44080884316,0x80005200b000004,0x6000010850002004,0x0,0x420000800020,0x802a2f3013842206,0x1542d308a900254,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4100008000200,0x0,0x0,0x0,0x0,0x0 },
	{ 0x7ffe7ff3cbb6f7ff,0xbbd4ff7ed3b3f76c,0x4000040040008000,0x85eb4da07009d10a,0x94bb26ace095d32a,0x0,0x0,0x0,0x680a08034130c22a,0xfcfffffef3b9ffbe,0xfffffffff9bfdf7f,0xffbffffffbbdf7fe,0xeffffffffbbfffff,0xffffffdffbbdffff,0xfefffffffbb7f3ff,0xfffffffffbbfffff,0xffbffffffbbfffbf,0xeffffffbfbbbf7ff,0xfffffffffbbef7ff,0xbfff7ffff3b5f7ff,0xfffffffffbbff7ff,0xfffe7ffffbbff7fb,0xebd7ffbffb9cd3ef,0xeffbbefbfba5f3fe,0xb7ff5ffbf1a4dbae,0xfefffffdf9b7dfbf,0xfefefdfbf9b6ffee,0xfbdeff7ff9b6f7ff,0xbedb3e1bf3bbd329,0xebbecffeebb7d37f,0xf7ff7eebf9bbfbff,0xabfeedf3d1b5f7fe,0xefbef7fff9b9df7f,0xcfdff73ff9b0fd7e,0x1b1dc35ee999dfef,0x54b5fee2d1a9dbba,0xf16eff62d3bdd3eb,0xffbf7ffbdbb4d3fe,0xfb7fff7fe9a5d37f,0xefdedfdbe3bdd7fe,0xfcf7ffe9f9bfdf7a,0xfbfffffef9b1ffef,0x1e9eaf6cc9b1d3ef,0xe5efff2ce9b1d77e,0xfadfdf5bf3b5d3fc,0x9698dd7f8195d32c,0xd8efbf7ec184f3ff,0xfbfefffef1b7f33f,0xcbd5f1eee1b8db3f,0xcf4f7bdbf132c7a2,0xbfbf7ffbd9b3ff7f,0xff35e6e3e1b5c3ae,0xefbeed70e9b9d39e,0xf9d4db0ee1a4dbe9,0xeecbfa93c9b1fffe,0x8effcf6ff9b1ff2b,0x796dff7881b4f73f,0xdfbefff3f3b1dbaf,0xff94ef77cbb1ffaf,0xeffffb1ffbb4dbfe,0x8fdfdfdbf9b0d77f,0xd6fffeeff9b7dfff,0xfbefff7af98cfb7f,0xfa9eeffcebb5fbbf,0x3ddefebbfb91d16a,0x8bff5f5ffbb7df69,0xbf767eeff1b2d35f,0xd7adffedd9b6d7be,0xdabf2ff7f9abd73e,0xc94affbef1aad33f,0xa8acdffde124d3ee,0x9d3b4fc661b5d7e0,0x117cc7e461845732,0xffb67546f088dd7c,0xe9fff7ffebbaf7ef,0x3b5b5fd7e9b8f3ef,0xbdbedd7a7b8c53ab,0xd5cebf4cf9b0db67,0x5fcc7b42a1b1d32b,0xebeed73ff131f32a,0xeffbbfdff3b1dbff,0xffbfcffff1bfffea,0x14e5fd6ae1b1d32f,0x97beaf70f9aad3ee,0xffbcfbebe18dd3ef,0xefffff7fe995d7ff,0x6e7fdedacbb5d3af,0xb4f5dc616080d33f,0xf968f66bd192536a,0xdb0e2fffd8b0513f,0xeb53ffabfb3277ec,0xe3e4c15e190c7ae,0x9c9f7fe4e1a5c338,0x78c4f764c190532a,0xfeb67f7dcbb7d7f6,0xefd2d675ebb7f239,0xbefede43f9b0d93f,0xdffafffff0a9d3bf,0xf7fff7fbf3b2d39a,0xfffe7fffe19dd9ae,0xcbdefbbed934d3ee,0xeffe1bdffbb0fbef,0xf38bcd9ec9b0f3c8,0xfd75fe71f391d26d,0x4daa5652c3acd2ed,0xb302effd89bdfbbc,0xfcf7ffd3f1b4dbfb,0x390cdbf0d180cd38,0x7076fe6ba1b1d7ba,0x9f6eff63f9b0f3ae,0x9df4f77ee1a1d368,0xfd4ea2e61b1d35d,0xbbcacfdbdbb5f7e5,0x613058f44310e328,0x93423452c182d1be,0xf79f2975c195fa6e,0xedf3fe9af392f36e,0xcfffdfdffbb4ffef,0xbdffddbfe1b2d77e,0x65eca76ed0a8c37b,0xdea63ffed0b1ffbe,0xcbf7160ad910733e,0x3b7f1f5bd9b8d30d,0xbd3366fe89b1d36e,0xd37ebffdf991db3e,0x7f966774c930ff6c,0x8940c360c038c31c,0x5fdbbe1ee9b0fba4,0x8c1c7ca4c110f22c,0x418050220390c421,0x6386af71f90051e4,0x9100e22ce0204b06,0xca6beb4f8bb0d12f,0x8b185f716900c18a,0xfffffffbfbb7f7ff,0xfffffff7fbbfffff,0xebaef67fe9b2f37f,0x83c7a732d131d33a,0x48dccb5bc181d36f,0xd6377eeec9a45b6f,0x9b490f5a41b1d201,0xb794ff7ee1a1fb1e,0xc4ce6912c3b05308,0x16d84c49c090d108,0x300050002005200,0xdd101c62c109d000,0x998e0724c111d124,0x4f56ea234181c2e9,0x437b1bf969b3c378,0xdb09cf61c0a0d2a8,0xdbfdb7fb53b5f778,0x408672ea4b00c18a,0x82d4d432c915f226,0x4c9213505190c238,0x6e3153ae71b1c303,0xf1acf23efb31f322,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8400040001000,0x0,0x10001000008020,0x0,0x0,0x0,0x8010000000000108,0x0,0x0,0x0,0x0,0x0,0x8080014000004190,0xc00001100000,0x0,0x92cc51e1021924,0x82104481106040,0x42a214404001c008,0x72069b2451840339,0x87cb1759d3b05b79,0x6b3fcdd4d9345b2c,0xb57502905080d10a,0x17bafd71e3b8f328,0xda04737a7031f208,0x4142e816e890c020,0x8eb6cf5aebb2737e,0xf2a376c8c9b0f229,0x4826a3488930422a,0x59af2fe8e1b2937b,0x500000000204,0x29c53016101f280,0x14195cd0c090c113,0x9100416901900731,0xc024256888805728,0x8020000000800100,0xa1001c824810c200,0x16334250c0101308,0x504cf3088180813f,0x3880050cd0204206,0xc4c00209104240,0x64061000002443a0,0x0,0xca66181901209,0xc84a2460d8205200,0x8bc8e72ae9b4d338,0x81fbfa1b61b5e37f,0x83e6c46ea287388,0x10f51f4b81a463eb,0xceaeef02e1b4fb2d,0x290e72fc191d1fa,0x895ebe838190cbfe,0x2d7f5fd2f3b4dff9,0xfefffef3d3b7db7f,0xb3f9ff73f3b6ffef,0x1a323c200ba6c320,0x1846d6845190d178,0x10100040000000,0x154214081c00a,0x0,0x800b0041809020,0x0,0x0,0x0,0x12a2741e194c2a0,0x4004202000a00300,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x1a70250083040106,0x0,0xffdeffef8bd7f3fe,0xffffdffb8bd6f3fa,0x3041c36a03143122,0x0,0x0,0xdfde3f9f82d7fbfa,0xbfffdfdf8bd7f3fa,0xffbeffef8bd6fbfe,0xffffffff8bd7fbfe,0xffbeffff8bd7fbfa,0xffffffff8bd7fbfa,0xefffffdb8bd7f3fa,0xfe7fdfff8bd7fbfe,0xfdffffff8bd7fbfe,0xfebe2fff8bd7ebfe,0xcfffffff8bd7fbfe,0xbff7ffff8bd7fbe6,0xbfffeffa8bd7fbfa,0xffffbf7f8bd7fbfe,0xffbe3ffd8bd6ebfe,0xfbfbf7bb8bc7fbfe,0xffdfbefb81d7fb5e,0x9fbfdffb8bd7ebfe,0xfbfff7ef83d7f3fa,0xfffe7d7b8bd6f3fa,0xfbfdff7f8bd7f3fa,0xefff3ac78bd7f3aa,0xffbfdfff8bd6fbfe,0xf5ffffee8bd7f3f6,0xcfffaf738bd7e31e,0xf9a8f76d8b47f1fa,0xcbd7fe398b86fb22,0xaff9dfdf83d7b8f2,0xfefff7fb8bd6f3fa,0xdffeff7983c6f3fa,0xfbb63fef8b56fb32,0xffefff2f8bd7f3fe,0xcbffdd5b8bc7fbea,0xfdfffbeb8bd7fbfe,0xfdfdfe7f8bd6fbfe,0xfabe2ff68bd6fb72,0xc1f4eb7f8955d3ea,0x2a7e3ec88b96f24a,0x34ba65fc8bd4f32a,0xf7bca7fc8b46e39e,0x3dfe7f738bd6fbfa,0xabd5efe683d4b36a,0xeef77cde8bd4e35e,0xffbbf7f783c6e3f2,0xd53ef7ef8bd7f3fa,0x6f2e3df98b56fbf2,0xfbddf76682c7e96a,0x8de3bf5f8bc6ebee,0xc6184f228984eb16,0x7ef7f6f88257f3e2,0xef7e7d758bd7fbda,0xfb9efb7f83c7ebe2,0xffdffe3f8bd7fbe6,0xffffdfdb83d7fbfa,0xf6ffffff8bd7fbfe,0xffeffffb8bd7fbfe,0xffbe7ffe8bd7fbfe,0xcdcfdf1a8ad7b32a,0xeadfffcf8bd7f3fe,0xfddab7af89d7e3e2,0xf97fafd381d7faaa,0xdebe3ff68bd7f3ea,0xf78df33a8b96f3e2,0x6bfefeff89d6fb7a,0x3e294ffa8bd7fb3a,0xc57ff77b8356f9a2,0xd6be6f7f8bd7fa26,0xcb93db8489d6eaaa,0xdeda7e538bd6abea,0xbbb7cff78b86f3a2,0xbcf63ef58b55fb92,0x5fbcff728bd7d3e2,0xf8d4e6a48a46e38a,0xeccbbffb83d6a3fa,0xa7ff4fb383d7ebfa,0xf5f5bea08a56bbaa,0xfbb7fd728bd6fbca,0xfa36fb6f8bc7eaae,0xedffbe3983d4fbaa,0x6dfe5b8283d7f266,0x75bdd67a83c6e3f2,0x93bcae7a8ad4f21a,0xa38a2f6f8317b3e2,0xcfffff9e88d7f3aa,0xea9cdecb0bc6e3e2,0x8edf566083d5f342,0xe86d266289c5e14a,0xfedeafff83d7b3ee,0xeddfffff8bd7fbea,0x6fdedaaf8bc6e3ea,0xb779fff38b56f35a,0xf55dd7f98bd4a38e,0x87b67ff60bd4f3aa,0xab85caaf80c4e3e2,0x77ff7ffb89d7f3fa,0x9abe0fe383c7c36a,0xddffa6740a57b37a,0x40b6fe518856f31a,0x6ad4e37a01c4c33a,0x8de39eb98ad6a326,0x13bece638943fb22,0x30b7f4e88bc6c346,0xfbdeff738bd6f3fa,0xfdb0f77a83c6b8ea,0x8de7b93b83c6b332,0x1405c782c2a0c4,0x92937c908bd5c312,0xc06b37e80bd6231a,0xd3b02a7683c2e9a2,0xedcee61803c7a1f2,0xe368dbca83c7d162,0x3c9f44ea83c2e9aa,0xf5edbffc8bc7ebfa,0x5487ae5c8306e932,0xc3a2c73581c4e8a6,0xaff79fff0bd7e37a,0xd6ff4df98ad5b3de,0xf577b5f18ad7a3ba,0x48e67d7889c4f2a2,0x104e13380c6c056,0x415a36118384e102,0x8008c44000860200,0xd3519b588203fb6e,0xb2e3d028217b3e6,0x855a00388809102,0x8090e8102028122,0x9a1edff383c6fb6a,0x3d757feb8ad4f1ee,0xedfeff7083c7b3ea,0xfd96677d82d6a3fc,0xbc2be378a57b3aa,0x5a97d9428283e322,0x101256a808078102,0x17cef7a8bc7e26a,0xcfaa36078a869062,0x59cfe20280169356,0x1110820941c042,0xc8725fea00d7b3ea,0xf118667983d4f34a,0xd8910d520844a1ba,0xcc83f9098ac6a13a,0x18d99ec38346e3ba,0x28324b4a8086a136,0x8037144203d4b0a8,0x928437410044a02a,0xcae4e22780c6a3da,0x487f1bca035633fa,0x3fbecffb8243b3e2,0x9140644208149022,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1004042008400002,0x840000000020002,0x0,0x0,0x200400020200,0x0,0x0,0x0,0x0,0x0,0x0,0x802000040002,0x0,0x1000000200040002,0x420200080504004,0x64943c700256b122,0xb180f82e8346a362,0x8042360f8017f826,0xf9c65728856b126,0x74db2ff48b86c346,0xc220240400018122,0xfbd7ffff8bd7f3fa,0x9bd7df2f8bd6f3fa,0xe3ffde5a8357fbe2,0x503c56ea0906e3a2,0x8cbcaf728345f2e6,0xc1082a1009169302,0x4b523e000187e092,0x201211580a549040,0x280504200868222,0x0,0x480004020000a180,0x800081000000a020,0x24781c1b0357e322,0x980c00888809002,0x801207628311f20a,0x1b062f6183c5e252,0x800423840180a2a0,0xcc158000001a1a2,0x418c80000858020,0x40020000040002,0x20080080000004,0x898461028340e9ca,0xb2300304a020,0xb04cb5a83c6c22a,0xa0204801c40012,0xc98b2d500ac4a372,0x4290e1740294b122,0xcfcaf61f8ac7f3be,0x8091055681c4aa02,0x5c94fefb8b447362,0xdbffef6b8bd6bbf2,0xab102c0081045202,0xdcfae8a8b13f262,0x246001a0113d100,0x200500a002040002,0x480004028000a020,0xc00a080402004,0x0,0x50000000101000,0x40003508040410a,0x810478b80286a222,0x4008210001500080,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x190000406800000,0x3bdfb3dacfdbfe10,0xffffdff7cfcbfebd,0xf7b7fefa4fdafab4,0xadbe2f72a44722b8,0xffffff7fe7cffebd,0xffdfffbfefdffebd,0xffffdfdfefcffebd,0xfdf7dfffefdf7e3d,0xfffffffbefdffebd,0xffbfffffefdffeb8,0xdfdfffbbefdbfebd,0xefffdfd3edcf7eb9,0xfefbffffefdbfe35,0xffffff7befdf7ebd,0xfebe3fffefdf7ebd,0xefffffffcfdffeb4,0xffffffffcfcffebd,0xfffbfffeefdbfe39,0xffffbfffefdffebd,0xfffefff7efdf7e3d,0xdbdfffbfcedffabd,0x4ef77f3eefdbfe30,0x37ffcdf3cfcbfe84,0xf7bcb77fced3fe99,0xff9efdf26fdbfebc,0xfbf4ff7fefdb7eb9,0xefdfbf5fefc77e35,0xefbfeffbcfdffe3d,0x3dfff66eedc77eac,0x9dbc7fe30ed3f628,0xcbbef77fcfdbfe3c,0x4fdffe13cfcbfe8c,0xbbffdfdbefdffead,0xf5fb7cfaefdb7e28,0x9f7dff72eedffe31,0xdbb17f7fcdcffeac,0xffffff3fced6feb4,0xebffffdfcddbfe3d,0xbdffdfff4fdbfebd,0xf9febffecfcbfebd,0xfa3e3fe7edcafa1c,0xebe7fb7acfc62e3d,0x6fff9ef54e0d7625,0x3ebfcff3cfde7e35,0xd1ebf57b4d83ee35,0x93de796084d3f618,0x88d4fb3f4dc37eb4,0xaef79bffe793febd,0x3bbecfffefdffe3c,0xdc76bff94ed33eb8,0x7ff87750cccb7e21,0xd150772e0fcb3e10,0x84c78c100cc2b209,0xefb4cff0e5cbfa31,0xf5ff77eacc8a7eb9,0xe7fef5f2c7cf7e08,0xfb9cffffcfcbfebd,0xdfffffbfefcb7ebd,0xffffffdfefdffebd,0xffffffffefdffebd,0xfbeffffacfcffe39,0xffbeffffcfdbfebc,0xc954e72e0a185200,0x8b88ff5aeed95234,0xffbefffbeedbfe34,0xbf6dff36c6dbfea1,0x4ab63fe7c6965e95,0xcb74d7bcc4c3beb8,0x1573f58ccc2b620,0xb7ff7fefedcffab0,0x977f7febefdbfe81,0x9214b19084833600,0xeaf3f9aa4fca3e20,0xef7edfcbcfcf7eb4,0xfdbfeff2cfd7fe09,0x9676d4f94ec27a1c,0x48bc5ff3ec8b7e80,0xfbecff7ecdcf7a39,0xcefe32b3efcb3e19,0x7f3f5ff1cfdbfeb9,0xf5fef7696cdb7e24,0xf3bcff71eedb7e30,0xdbb6f76ee7cb3e94,0x8cdfbab68dc32e11,0xae96cfd3c6cbfabc,0x1007c6a8c6c35220,0xff6fff7be7cffeb9,0xffbe6f2acdc2fe30,0xefffffbeefdffe3d,0xbfffdfdfefcffe39,0xff9fd7f3c6c77ebd,0xff9faf70cf8afa3d,0xfbbe3ff7efcbfa3d,0x8bfef39b4edb3e08,0xeddfbd5bced37eb8,0x8fbacffaefc27ab5,0xdcfd379048c36810,0x7d26b7e4ac9de85,0x8257831a4bc11e24,0xadf71b40068ad288,0x378fdfd9c68bfea0,0xddf756fb6dcf7e21,0x7fbe7df3cfd7fea9,0xdbe4f72fcedf7e1c,0xffffbfffefcffebd,0xafbfdf7ecfcffebd,0xf5ffdeffcfcbfe34,0xffedff72cfdbfeb4,0x8396e33f8cc6d6a8,0x5ed67eb2ca437eb5,0x7b78bf99e4c73e3d,0xf4b8f763e58f7eb9,0x9b0b64720782de09,0xd2a6067066c23220,0x859dff18cccb1a1c,0xa62ff59c5c3fa90,0xbcd17722e582ae08,0x796af478eddbba98,0xfab42dd6cfcb5a3c,0xe7ceffeaefcb7e31,0xbf76affbcfd65eb1,0xb772dffacfd7bea8,0x606cbd474d895e24,0xfb8aeff865c72e01,0x88a2dbc24c463684,0x142220040121408,0x1498cd69cccb9201,0x9041f3230c037a09,0xc0b2258084805288,0xd3dff51e8dc37aa8,0xa040004a01814a00,0x2f3ccff0eec8a231,0xfdf5feabcfcb3e0c,0xd7af2d6347cbbe19,0xda2eff6c65cf7ea4,0xddf7ffbfcfcbfe34,0xe3ceddcececb9a3c,0xfc9d7feccddafe0c,0xa5437f63cecbfa31,0x3926704c58b7820,0xebffe72ecfcb7e35,0xab140cd745cb7a05,0x92964feac4c31a14,0x882cf60bc78a7e10,0xd0322dd147ca3a10,0xc102eb1a44483e00,0xa1000a0a45027c10,0xca06fc347cffa14,0xc874008889201,0xf6be3d67cf0a3a14,0xaa2df203854b1a90,0xc47e1a5d85821810,0xb65bc76260c33a30,0xe41323f00ec27a10,0x800040000400200,0x0,0x6000000040200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80000200400200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2010314040101090,0x8180807101c48a00,0x4201161080023280,0x809d45c00581be08,0x76b6d47a47c3faa0,0xa63424e048095ea4,0xf8d4e15744caf604,0xe4012880c880600,0x302846e146837224,0x30d4f64ae6cb5a00,0x828a2d4044436a89,0xf2a2bb5fcec69a19,0x892292a140cade30,0x8bef5c5a4ccb7a04,0xf0bf4df6c5d37e38,0x5b602c0a86d03218,0x42222c4041c40e24,0xd872a2020c3aa20,0xa180c250c48a9200,0x12220d4105c20a10,0xd004042480012a0,0x18822460a0c10200,0x480000000400200,0x981b120088885e20,0x24004084800,0x28234284022600,0xabd6756e4dc39a05,0xc842328b4cc35230,0x4482021004002601,0x200001000004800,0x8001000001000800,0x4002258009411800,0x202000040800000,0x742000000005680,0x48180b1184ca3200,0x1040220060000220,0x284d4044cb7a14,0xd700f16a05c62a08,0xdfdfbeb3efcafa19,0xb94a7d0045c21a21,0x54bf5ee6cdc37eb1,0x97fbff6acdd33ebd,0x9bbe6f3605c6ee38,0xe8614f2cc4927a05,0x2d2d240805621,0x800010000000600,0x50000800800,0x4206100240081230,0x0,0x4052000a05020818,0x0,0xa025042082840600,0x0,0x0,0x0,0x0,0x0,0x0,0x44000000020200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x1080500a00a00c6,0x0,0xebcfb37f2a4e80c8,0xffbffffffa7fbbef,0xf17d7658fa6610eb,0xfffefff7fa7fbbff,0xfffffffffb7fbbff,0xfffffebffb7ebbff,0xefffdfdffb7fbbdf,0xfffffff9eb7fbbef,0xfffefff7bb7fbbff,0xffffbf77fb7fbbdf,0xefffffbb7977bbcb,0xffffffdffb7fbbef,0xfffffffffb7fbbff,0xfffffffffb7fbbff,0xfebfff7ffa7fbbff,0xeffffffbbb7ebbff,0xffffffdffb7bbbef,0x9fbf6fff7a7f9bcf,0xffffffffea7fbbdb,0xffeefffffa7e9bfb,0xcbf5ff7f5a769be7,0x71c32f548a7219cb,0xfdbaede2ab7e98cb,0xfdfff67bfa7a93fb,0xfffc7ffa6b7e19cb,0xcbfdef3fea7e92ef,0xeffbbf7feb72bbcb,0xbfbeffffeb7fbbff,0x7dff76fa3a763bcf,0xb7a8fdf1fb7a1bcf,0xdea52f6f3b52bad3,0x4fcbbbdb2a769a87,0xabffdbdf6b7f3bef,0xdefff6ebea5f1aef,0xcadb7ff32a7f9bcf,0xefbe6c232b5213c1,0xcfffffbefa7b3aef,0xefdfdfdafa7fbbcb,0xbffefffbba7f93fb,0xdbfdf75ffa7bbbdf,0xdfb7ff7f2b7ebbcf,0xebf6ffff697a1afb,0x1add5e6ffa529aef,0x3709cfdb2a57baef,0x97fcf72a6856b98f,0x9216a32c28621389,0xebecffbbea76b8eb,0xeeda7f50aa731b7b,0x8c994ff3bb4a1847,0xfdf4f3f069761aef,0x7fec3f62ea74b2df,0xdbf7cb1e3b7a3bcf,0x47db3795aa539b9b,0x7bfdff32b761bcb,0xf591feebba7a93cf,0xf6bf7d733b521adb,0xfffefbfe6a721bef,0xdffffffffb7fbbef,0xffffffdffb7fbbff,0x3022c008010282,0x5b7e77e148321b93,0x200200000440000,0x0,0x0,0xf6ff7ffea97fbbcf,0xd9ffefea4a7239cb,0xeb23f5548733b57,0xcddbdb1ceb7533ef,0x4ffffedbeb7e11ef,0xfffb7ffbfb7fbbdf,0xb6bfff7c2b7e93df,0x7ffe3ffdea5f9bcf,0xcbfbfff7ba7a9bef,0xeeda1f593a52b9cf,0xf3aecd5a2b7713cf,0xf56cf77d6a7e9adb,0x5d3efdf27b7213eb,0xfbf6ff3f6a7f9bef,0xcf7bffd9ab7e19cf,0x23aac5130a6291e9,0x7c7fffec6a7ebaef,0xdfbfe7f07a721b8f,0xff9f7fed7b721acf,0x1bceba867a5292e7,0xac27d1dbaa749beb,0x94a55c680a4102ef,0xdd3cfffbfa7b1bcf,0xdffffffffb7e9bef,0xffffffbffb7ebbff,0xffffffdffb7fbbff,0xfffffffffb7bbbff,0xfffffffb7b7bbaef,0xffbe7ffdfb7ebaef,0xe522ef3b2a721a8b,0xab5e5ac62a5513cf,0xb2badf5bbb5a19cf,0x5072af5a0a7a13c3,0xf68ebf6cea7e12cf,0x897c72eb7278104b,0xede1e580a541bca,0xbe99effb485211cb,0xf4ffe7787976128f,0x4abefc736b7b33cb,0xebc7f72e6b7a19df,0xfffffffffb7fbbff,0xefbfdffffb7ebbff,0x79d45e4aa8679bc9,0xa13ead70287a9bcb,0x5b94bf656a7e194b,0xcd5a16012a7a3ad3,0x3b7edfdb0a7e1beb,0xfadb5ee6ea5a91eb,0xd9bdad6b6a7a99cf,0xc0142a6e88121082,0x58f8a2a48449883,0x27ea5cd8627a02c3,0xe494d4e1ab461b0f,0xf26caf2088521b4b,0xa2163d654a62936b,0xefeff73aeb571b8b,0xefffffddfa7ebbef,0x9ffbdffb6b771bdf,0xb15d27699b7a92cf,0xbab42565ea67b2cb,0xdbfa7bfa6a563ad9,0xa80a124148623a81,0x9882eac92a421bdb,0x552cf641aa4a10cb,0xa8c2de0287a114a,0x9c0c7300278024a,0xc7c0b7560b629a4b,0x26ba91c248421bcb,0x9cf41e6eab7693eb,0xdbbeff70ea7a9bdf,0x4f9c370d2a428aca,0x8be3de302a561a87,0x796cdd60a7a10de,0x92522644a862028b,0xc9273d600a6a1869,0x8bba16714b6a3047,0xe446e6380a631aaf,0xaa08fd1742581a43,0x426519fea1210c3,0xf89c82720a631acb,0xee9e2ff60a5713db,0xc91282382a4018db,0xa7129223285a91c8,0x92aa4e700a671a86,0x911dd22828430a8b,0x44d429d00868104b,0xe1b0c335aa6e80cb,0xb400e0248000949,0x8210000a0a440081,0xd5459016a8621083,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2000800000021000,0x0,0x0,0x80004002000001,0x0,0x0,0x4000010000400040,0x0,0xb1107704884b1bca,0xfb80f22ee8781aef,0x268290010810190a,0x80808800a000102,0x10280709800180a,0x100020000600000,0x8090e70c2b723247,0x88c0031120500148,0x20002000000000a,0x0,0xfa38e58039711a8e,0xc0a0a12e025012ea,0xcc4bfe89ba7a1afb,0xea744ed34a6303f9,0x96d75efa687211db,0x20c2003400040,0x9010824008029801,0x80c4901088650282,0x8a60104301509283,0x1004040002440002,0x403c2c40a2521048,0x0,0x902400000400282,0x456000820021848,0x10000200080040,0x0,0x0,0x28907202084418c8,0xcc827f183a5a30c7,0x8884d800a52120b,0x10400000600000,0x20804d6aa3721040,0x120480000006,0x1000020000000082,0x8810044808621000,0xc9e076408501387,0xbefeff726a54bbe7,0xecf6ef68ea5b9acb,0xcfdfbeaf796a93df,0xa25646c2ca76104b,0xf4f776ee3a7699cf,0xfbffff7aea779bef,0xfa982d4e28609bbe,0xc0032a0389501242,0xa611024008798343,0x4204d000a008041,0x0,0x8000206000500040,0x2840400008280040,0x10520002101040,0x23aa41ed205a13c3,0x12284442a8521002,0x0,0x0,0x0,0x0,0x8000002080000002,0x2004000420000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x108000101000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x4012001000180084,0xfbd7ff7f437d8ff6,0xfffbfffbe77fdff6,0x7b9ccfffe27fbff6,0xf7ffffffe77f9ff6,0xef7fff73e57eddf6,0xffbeff7fe77fdff6,0xdffffebfe77ffff6,0xafffdfdfe77fbff6,0xf4ffffefe77f9ff6,0xfffeff7b677ffff6,0xff96ffffe77e9df6,0xffdfffffe63ffff6,0xffffffdf677f5ff4,0xfdffffffe77ffff6,0xffffffffe77ffff6,0xfebe7f7fe73fffd6,0xefffffefe77ffff6,0xdfdfffff277fdfd6,0xbd3affffa33d1ff4,0xfffdffff633ffff6,0xd7b66dfb677f9ed6,0xfbf6fbbee77f5ff4,0xe5df1ffee73f5ff4,0xbebfdfebe37ffdd6,0xfdfff66f267e5fd4,0xfff67f73e77e3df4,0xfbecef2fe77f5df4,0xbffab36fe67f5ff4,0xcfbfcfff667fdff6,0xf5ffffefe67ffff6,0xdfbebf72e53fdd56,0xdbae77fd267c13d4,0x4d93d80c00767d40,0x8fb6df57617d9d72,0x312012c1603804f0,0xa54de76861059d56,0xcaae7fbe637fdff6,0xfdf7ffbfe77fbff6,0xff7effdfe37f5ff4,0xd73f7febe77eddf6,0xf9befff6673f9ff6,0xffbeaf7fe67ffff6,0xebd7fb1e277efef6,0xabfe1ccfa6759dd6,0x84eb6758a63cdeb6,0xbbbca371e53efdf2,0xfb96bde6e13e5fd4,0xf176ff7ee37f3df4,0x4ef7be58627f1ff4,0xbfbddf7be37d5ed4,0xdbdff67d673e9fe2,0x7ffeff78e37fbff6,0xfb96ff2f663c3fd4,0xec39adbe23d5ce4,0xaebf4f7aa37d1ed4,0xf4dffefce33e3bd0,0xb7bea570e27f9df2,0xdfbef77ee77f9ed6,0xcfffffbfe37fdff6,0xaf7993da663537d0,0x0,0x0,0x0,0x0,0x200031002140084,0xdfb2ff69673ffcf6,0x104702000111740,0xdabe2fefe37e8de6,0xedffdb5aa73fdbd6,0xefffffffe77fdff6,0xfffbfffbe77efff6,0xf7fff76f677f3fd4,0xf7fefdfd677f9ce6,0xc3f6fb3f63379dd6,0xdefe3fffe67f0fd4,0x7fbdcf42407e9ad2,0xfdfefefa227f9ff6,0xdbfe7fe1e67d3ff4,0xfbc2f76ea63cbbd6,0xcfffbebfe57d99e6,0xf79f4f7f627fdfd6,0xf5fff7fa627fdff6,0xfe3635f2423c9ed2,0xfbaeffaee33e5fd4,0xcd5fd68fe35efdd6,0xa90e4fdf633e7cf4,0xf6b476e8e33e5cf0,0xdffeff72677f5ef4,0xffbfffffe7775fd4,0xffffffffe77fbff6,0xffffffdfe77ffff6,0xffffffffe77ffff6,0xff7fffebe67fbfd6,0xfbbf3fffe77fdff6,0xc7d7bfbb633c9ef6,0xeffffb7f637dfef6,0x3a884fdb603c5fd0,0xd3bcb570203696c6,0x92c662ece0109bc2,0xa1a5fa2f627eb0b6,0x4e461247004f1cd4,0xbe1e7de1667d1dd4,0xf44cb66d663e34f0,0x68963d72a02c5ed0,0xfbfcffffe77e7fd4,0xefffbfbfe57fddf6,0xefffcfd0e677dfd2,0x71b4f6ff23778bd6,0xe3b8fd70613f5fd4,0xdab52b7ee7571fd0,0x49733a0b230e88e2,0xbb2bfb97e2359ed6,0xd62f5cb2822d9bd2,0x1048d610604c1174,0xfaad3f36261e17f4,0x27c6c80e623c14d0,0x27f6fbdb623d97d2,0xb2b06fe9673c1dd0,0x4128b778631d15d4,0xd994a7d1613e38e4,0x8ef01a0422361c84,0xefffffdfe77ffff6,0xfffffffbe77ffff6,0xcbf8ff7bc37c4ec4,0xff162bd462355dd4,0x8093eb6e26143654,0x6a40ca74265d05c4,0x40924c83601515c0,0xd555fc82a5171d50,0x33667f71647e14d0,0x7283eb2e623f98d6,0x194b1aca62141dc0,0xd9a89e3631c0ce4,0x50224eeae13cdd86,0x75befd60677e1df4,0xeaac7ed5623eb7c6,0xde6bfe1ae03cc8c6,0xb673db4ec376cc06,0x901381c060531444,0xede44bc0e2541d74,0x40843460e0101c84,0x8840ae0880553ae4,0x6a7e5190015c3bd4,0xb5f7577b607f3dd4,0xd9e4a460805fd986,0xf8022dd0641ec886,0x8840a062605c80a2,0x4f1c0e0b61185a74,0x31a2069e64764a54,0xd1aeb6ea427e9dc6,0xdf802f10623c0844,0x89dde32fc05b8876,0x49c01a15265e0624,0x36f99fe1247fda76,0x92c0144c65140e04,0x0,0x0,0x0,0x0,0x0,0x30000000001400,0x0,0x8000000000080080,0x2000010000001400,0x0,0x0,0x0,0x802001000800,0x0,0x8000002060000000,0x2004000000080080,0x0,0x0,0x0,0x0,0x240000080004,0x302c51200c14d4,0x0,0x8014ac8403c08a0,0x18886800001c00e0,0x1000060004a4804,0x49260782253c1f40,0x1391c53c646f10d0,0x6828a0000351444,0x3004a0000261c00,0x5193f068233e1694,0x93f4bf70233c1ca4,0xc8183122627c9812,0x4b52364a647e00f4,0xabaa0f1b603f97d6,0x8320040000188242,0x90e9272462361440,0xce80131e623c0010,0x45c6a310421ccc12,0x480000240000,0x15240b4181424800,0x12e0730002c00d0,0x4406000020001480,0x0,0x0,0x440000040100000,0x80002440000c04a0,0xd210772423351d50,0x8800230821101470,0x2540004882040940,0xc23e0562620412d0,0x940c2402001d11c0,0x80200c2222342040,0xc00000100004,0x840000000000240,0x29304f00a056a986,0x1102008000051400,0xc8295c2143760e40,0xfa26e77e627e9bc6,0xdeeffbbfe57fbdd2,0xed7bfbca823e9ff6,0xfcb776efe77c9ff6,0xfdff7f7be67f7fd4,0x93886ef6653e5dd0,0xa610431a223c5cc0,0xa0081182001812b0,0x0,0x0,0x8800280040300800,0xcd6120024648c02,0x206cba2062051080,0x343b47e4241435d4,0x6000003002440000,0x1200041000050044,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0xfffefffbff29ffdb,0xfbffff7fffa9ffdb,0xffefb6ff5f29ffdb,0xffffdffeefa9ffdb,0xfdffffff7fa9ffdb,0xffbefd71ffa9ffdb,0xffffffffffa9ffdb,0xffffffbfffa9ffdb,0xefffdfdfbfa9ffdb,0xf5fff7fd7ba9d7db,0xdf7ffffbff28dfdb,0xdfbe7fffefa9efdb,0xcddfffab5da9ff9b,0xffffffdeffa9ffdb,0xffffffffffa9ffdb,0xf9ffffffbfa9ffdb,0xffbefffdff89ffdb,0xefffff7e7fa9ffdb,0xaffffef9dfa9f7db,0xb7fedffb9b29ffdb,0xfffffffbffa9ffdb,0x7ff6ff7ddb29ef5b,0xa3f7ff7f7f20efdb,0x7a761f5f7b2967d2,0xf79effffbfa9f7db,0xb4fffe7f3fa9ffdb,0x6bfa7fe2df28dfdb,0xf1d2f72f5b29e79b,0xeddebe797a28af5b,0xff9fcfdb7b2897db,0x75ffd67e7ba0ffdb,0xfbbfaf73db29cfdb,0xdf9fe54b0b28d3db,0xefff3e17cb09ff5b,0xab5f5fc10b29ef59,0x947176e23a01e79b,0xfbee7f731f29ef5b,0x79d727673b08ff1b,0x8fdbff2f5ba9c7db,0xafffffdbefa8ffdb,0xbfffdfe3ff29ffdb,0xfdffffff7f28ffdb,0xfebe3ffc5b29efd3,0xcdfffe335fa8ff5b,0x2deab3e91b01eddb,0x8eb377ee1fa1cfdb,0x91fc976a0a01f79b,0xc7a4ff4423008393,0xebf7f7eeef88cfdb,0xeec7b6dabb20ffdb,0x3fbcefdb9b29f7d9,0x16f4f67a1aa9c753,0x5afefd71bb08e7d9,0xe8f4b72f4b09dfd3,0xc7db9e93dba0d7d3,0xa79bcfd2df21c71a,0x34d576fbffa8575b,0x9fbffe631fa9d79b,0xfbfafffbffa9ffdb,0xefdfffba1fa9ffdb,0x200001000000440,0x0,0x0,0x0,0x0,0x0,0x94b376ecfb29e7db,0x0,0x30000008010,0x6fdff7f6bfa9ffdb,0xffffffffffa9ffdb,0xfffbfffb7fa9ffdb,0xd4efffff4fa8efdb,0xd7fc7f6ecfa1df9b,0xcbd6fabf7f20eb9b,0xffdf7edb1b01d791,0x3fffcfdfbe89c5db,0x975cba7b5f21cf4b,0xffbe7d737fa8efdb,0xb3f7ff3a6b28e7db,0xdbd7f6ffcb28f7d3,0xefffdff32ba8ffd3,0xfdffdeff9f29c75b,0xbfffff77ff09dfda,0xffb7ffffbfa9ffdb,0xc3d6b61f1b28cf1b,0x9b5e5f5a1f08579a,0xfcfff6ed7b28efdb,0x994ee7697a09e711,0x0,0xdcc77eb79b21fcd9,0xffffffdfffa9ffdb,0xffffffffffa9ffdb,0xfffef7ffffa8ffdb,0xfa9e2ff3ffa9efdb,0xefdffffb7b09efdb,0xefdeff531f29efdb,0x9a3827ea4b008793,0xd538fc7018a9cd5b,0xa69a2565d9088753,0xe2e5f3fb3f28c7db,0xe6c7325a3b08cdcb,0xb99eebea3b01f753,0xbf7ea76f1b215f59,0xeb767d715b28ef5a,0xfbdcff0ebf00df5a,0xaed2fa8f7981cf5b,0xfffbd7f25f21efcb,0x4520b04809084343,0xbb6c6df28900dfdb,0xeabc755e0f01dd5a,0xdffe339fbf29ffd9,0xff7fdfdbdfa9ffdb,0xc4a4d4608f00c61b,0xa123ff381a08e7d8,0x571064703e004dd3,0xa4a6c8004b0021d0,0x652d9500921c55a,0x91bfcf603f00e3d1,0xffeeff63fb889fdb,0x5ebdaffd9f28f75b,0xc5dff726df01ef9b,0xaffffffe7f29ffdb,0xffef7fffbf29ffdb,0x83e12c52822885d1,0xc01635d0ac28450a,0xbd4c62f3a88df5b,0xfeee1f539388d75b,0x831ec9522a08a580,0xf07eb6ea0f08cf8b,0x309c34604928a180,0xf8a1db3e1a21e7cb,0x86a101d0900a0d0,0x2dbecd504b20d599,0x7565ede66a2881d1,0x5deeed624a08cd8b,0xce12a7060b087e59,0x4843fb041900cb41,0x379cdfda7f29c79b,0xb231256a0b080f4b,0x9342bb400820cd41,0x8310eb710b08c751,0xd5e27162b08c313,0x14c5d00a000851a,0x1462572809200a93,0xbd78e7a0a989f799,0x88860a624c88cb81,0x8bd67b204408cdd3,0x800808448002100,0x28a44164b08cf1a,0xd069f67c8b007f5a,0x84082d490708450b,0xe8e6e26c0b20fd5b,0x2fda92020921f212,0x133858301e205350,0x844010200a204900,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x404000001010,0x0,0x0,0x0,0x0,0x0,0xbac63fc46788d71b,0xc8f0c3270a00fe98,0x4200021002080101,0x0,0xd1c0a4126a004f59,0x4806202028080108,0xfb09dd639209dd1b,0x40021c101a000540,0x1cdf7a0829e34b,0x548800008013,0xedce7d70ba08ef19,0xa08c6f673f21d353,0x4942b2760900c71a,0x801e4d433e01ddd1,0xc00086e032001593,0x2e044048008350,0x8800020480008410,0x81d9ed0a86004f1b,0x201420008002000,0x0,0xc060ac002a00a119,0x140012000000,0x2004008000080,0x14800000a002010,0xa4466802008451,0xc001200008208010,0x20004001004000,0x100800000008409,0x110008200000,0x90ccb7008a82f19,0x904800000100,0x47c71020087c1,0x26800009090,0x480020201000418,0x294c4543a800140,0x585886e30b0003cb,0xb3361c700820271b,0xff9be75cdf29d719,0xc7e7ff9f3f28ddd3,0xafffdfdf3fa8dfd3,0xf5767efcef89dfdb,0xdfefff73df28ffdb,0xd5bc26677b28b79b,0xa84d4a1909084553,0x1004460e204010,0x469016700d00478b,0xf0c4b55c4909cb11,0xd21001424c284110,0x100e3240b088382,0x121092c2080885c9,0x95f2a5e00ba04948,0x9e75b7f85b884f5b,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x24020000100,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x204000014000,0x0,0x0,0x0,0x1000040000000300,0x0,0x0 },
	{ 0xfffefffbfdfeffee,0xfbdff3fe90ba4cee,0xffeffffffdfefde2,0xffbffffffdffffee,0xf5ffffffbdfeffee,0xdfbefd71fdbbedee,0xfffffffffdffffee,0xfffffffffdffffee,0xbb7edfdbfdfffde2,0xf6fdfffdbdfffdea,0xebffffffbdbbfde6,0xffbe7ffffdffdde2,0xfff7ff3efcfefd6e,0xffffffdffdffffee,0xfffffffffdffffee,0xf7ffff7ffdfffde6,0xffb6ff7ffdfbfde6,0xcfdeffffbdffeeea,0xbfff7ffff5fecdee,0xffbafffb35fb7dea,0xf57fbffffdffedee,0xffdebff7fdbf7d6e,0xf9ffffba7cbfff6e,0x6edf6edd7dbbff6a,0xbffbcffefc9edde2,0xd7df677facfafe66,0xdfea7d73ddfddca2,0xfbe7ff7f5dbafd62,0xdec2bfdb55bbec6a,0xfebefffb9dbeeeea,0xb4fc77e83cdadde2,0xb3beef7474f954e2,0xeeb7f6ff3dbb7dea,0xcde7de1e74bee562,0x2a165bcf34b06b60,0xb0dd6eea30fbec62,0x8b0e657295bc5f42,0xdba0f7fc5dbad5ea,0xefdfffbffdfffeea,0xebfedfdbf5bb7fea,0xbdfffff7bcbbfd6e,0xfb7e77723dffcd62,0xefbeffff75fe6fe2,0xefd2f3b975baff66,0xe9f7ddc6ddbb4c6e,0xffc36feb749efce2,0xfd7db7ea74df5d62,0xdbfeff6e0dbedde2,0xdbdef33f54beffee,0xee971e1f15ba4dea,0xdebfdff2b4be6c66,0x9e75fe6994baee66,0x7afe7d7135ea6d66,0x9bf5cf1715ba5d62,0x48621f2e25b0c340,0xfdba4f2bf5aacd6e,0xbcf577edddba7d66,0xff7eef7275bffdea,0xcffeffffdcbaffee,0x50000000080002,0x0,0x0,0x0,0x0,0x1020008004000,0x0,0x40000680082000a,0x0,0x2004000000820,0xc6c3366105beefe0,0xfffffffffdffefee,0xbb3bdffbbcbbfd6a,0xfc7dbd6a349e5ce2,0xfe9675fc34beff6e,0xc8d6fbaf5493fd62,0x4ff71ead75bafd66,0xbfdfdda87db9fd62,0xf77efffff5fa5fea,0xffbe77f2dcfacf66,0xbbd5f77e74ba4de2,0xefd3fe9bfc9f6d66,0x9eb8ccc924fa7c6a,0xd4dffbf73caaff6e,0xa335f7333cbbcfe2,0xfebf6f7efdfaed6a,0x6cc6d61554a8b462,0x4b9f5b5844d2ede4,0x94fddfeaf4fb6f6e,0x0,0x0,0x0,0xffffffdffdffffee,0xfffffffffdffffee,0xfffffffffdffffee,0xfefebffffdffffee,0xfbfffffffdfeffe6,0xfadfff6bbcbedd26,0xa7baddfb34a95de2,0x9dfcbef915bb7d22,0x7136f3e094125462,0x8bf0fb7a99fa4e66,0xbe67be4034fb1de2,0xfba9dffbbcfffd62,0x75befe46d4d84de2,0xbad425c0009a6ce2,0xdbffff3fbdf96de2,0xa1cabedb5dbefde2,0xbbbedcf034bc4fe2,0x7436ffd294ded5e2,0xbfba3d75acdedde0,0xcd97f726f5bb6f62,0xe95716b264b64ee4,0x2f3e97d655bf6fe6,0x92b57eee94badc46,0xe055af38bc92c546,0xda107b7c35ba6442,0x6fc5761ed1b34de0,0x8821051108c0042,0xbfbe4dbbd53b5fea,0x996cafee19fb4c4a,0xda90277f35ba4dee,0xc1e4b3261dbae462,0xefff5fddf5bb7de6,0xfdbfff7f75fefd64,0x423037733db05860,0x7b165d7715bb1c0a,0x5940f00245b2a16a,0x1ecd5b6038baec62,0x8093cc8310384ca0,0xf93c984a149a5c42,0x33f23d40149b4c2a,0x8b89c32414994c40,0x8edabea0349b5c42,0xb884f8345bc0da0,0x1443006491a88d22,0x94bc2d62648a4d62,0xdf96ffe53cba7d62,0xc8471e99f49a4564,0xa960f5a103a4c6a,0x9c35206a949b0ca6,0x2568874040926ce4,0xda802be0049a0c46,0x244021200024820,0x94421300a248e2,0x14bb6f694daa6262,0x4148a240149a0c24,0x4aa2a79604aa4162,0x8874218a048e0064,0x8d6b2408498ec60,0xe21065814214842,0xd1380400248ac942,0x2ebe7d6130d7ed62,0x83b1f37304ace160,0x80712460088e842,0x8046c1242a0142,0x2800900000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100800000200002,0x0,0x0,0x0,0x0,0x212205200040,0x1080000004020,0x0,0x0,0x5d4e7df23cbe5d42,0x5040aa0805a04120,0x41120c0010920042,0x808a20c904000,0x8420217014a85622,0xc80e7de08dba0ce0,0xfb1cf0041cda6542,0x5c2928281aa0c20,0x800210000024802,0x1021014010888802,0x19882d030daac8e2,0x8930034810194d40,0x48d312111cb90402,0x4a5148501c94e500,0xa84ca030026842,0x9182a36880ba644e,0xcb9342d814bb4de6,0xe000010814000142,0x2004a00001004a0,0x0,0x4030200004020102,0x24000c000000,0x44000801200200,0x208c00a009060a0,0x30000ce404105560,0x840826000080e040,0x2000400000000a,0x820000a08084000,0x200812d7148a2c42,0xc2aa8f7afc80756a,0x300300280040,0x102c0d00052049a0,0x8b5a400304846840,0x850368315a86042,0x231806d0b4084d60,0xd0421c68108e4820,0xa218d7a03090cd6a,0xd514ba742090cd6a,0x4ecfff3bbcaaec62,0xbafffdde3cba6d66,0xb6f7ffeffdbbffe6,0xfdef7773dc9f4cea,0xf6b33ebfb4aaddee,0x8481810808c24840,0x900001800084040,0x1000040000900000,0x412c20300082000e,0x22c0011200002,0x503c2001028000a,0x6f498d014982462,0x1cb15d7001e8c1e2,0xe875162004b644a2,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x480000000100400,0x0,0x0,0x102044010210002,0x8000010080000400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x1280d20e6337ee4,0x8010000000000280,0x2b5395dfe4f2f6ee,0xfffeffff76f3ffee,0xfcfffffff6f3f7ee,0xdffefff7e4b3feec,0xfbfef77ee6b3f7ec,0xfffffffff6f3ffee,0xefffdfd7e6f3ffee,0xfef7fefff6d3ffee,0xdfffef7ff6f3f6ec,0xffbefffbf6f3f7ee,0xeff7ff7ff6f3feee,0x7fffffdff6f3feee,0xfffffffff6f3feee,0xffefbffff4b3f7ee,0xfbbfbfffe6f3feee,0xeffff7ffb6f3f7ec,0xa7ffdff9d65377ee,0x15bb7ffab473feee,0xfdfffefff613ffee,0xfffefffff6f3ffee,0x89f6bf3674127eac,0xeed33ed37673fdec,0xbdbfedb3f4b13eee,0xf7d7fefae4f3feee,0x5faa7d60e67272ec,0xf9dcfeeef4b37fee,0xcedbbffbf4d3f7ee,0xbeff5ff3f6f3f6ee,0xfef7ffef769176e8,0xcfaafde2541336ec,0xff1cc37e867277ec,0xdd5bffd5b43356ee,0xfa3ecf572613f6ec,0xe6fef7e456937fec,0xfd637772a4b35eec,0xff1077f524b3f6ec,0xefffffbbb4d37eee,0xabdafdd3a6f276ec,0x95bf5fe3d433e7ec,0xd8a6a778b473f22c,0x7ef3af7a74b2f6ec,0xcbdafbfb7473f6ee,0x229ecbc36492776e,0x87aa5cb48450f2ec,0xfdfff77f64f2f7ec,0xffbe7ffcf673f7ec,0xd9fafb3be633fdec,0xafd71e57147277ec,0x7ba8affb6470dcec,0xc5f3f779749376ec,0x93fe7f7134f276ee,0xbb96cb1724537eee,0x35936a3625154e0,0x27f3cd007673f6e8,0xfdff7ffff4f3feae,0xdb2ffe50e0b376ee,0x4000240000000280,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xccc3ba34b4b270ac,0x7fffffdf76337fec,0xbfbaffdbe4f27aec,0xf5fdf66bb4f3f6ee,0xff9eff7c3493f6ec,0xc9ffffe7f4f1f6ee,0x4eff3e2fb6f356ec,0x37fcdfb2f6d2f7e4,0xcdffbebe74f27aee,0xca0c6d7034f176e0,0x79cce33ed69317e4,0xebfbfffbb4737eee,0xafefcf796632ffee,0xd5fff6f9f6b3ffe8,0xbf3eef73e61356ee,0xfbf7ff7ff633f6ec,0x7f5fffbef4b3f6ee,0xedffebd2b6b37fe6,0x267005a000102284,0x0,0x0,0x0,0xffbfffdff6f3ffee,0xfffffffff6f3ffee,0xfffffffff6f3ffee,0xfffe7ffff6f3ffee,0xefffff3fe6b3ffee,0x4ff6be8944b3f2ac,0x6b2c4c560114468,0xd2bf0d6804314262,0xa430b65d243354e8,0xcbdff7fe66f2f6ec,0xffdf3f7ff6f1fee8,0xfb9bdfffe4f37eee,0xfd7dee5ed6f376ee,0xf7be7c7be6d3f66c,0xfb97ef6e66b3ffec,0xaffa3eea7432f6ec,0xbfbad3f9f61231ec,0x55ffd6e836b2d2ec,0x93603540443214a4,0xca886d08003141ec,0xdcce3f1d8013f6ec,0xffdedbd0f6737ee8,0xb6f4fe6006b31ee4,0xbf6ef760845336e6,0xd314c6e8d4627ee8,0x4947431882101668,0x207044000905000,0x19245820010d460,0x486d25213432306c,0xffbc1e5a3452f4ec,0xc9d6d3f68493f8e4,0x5bfefade941276ec,0xb4addffbd47336e8,0xab5cb778665276ee,0xaeb63dfd5671f7e0,0x80743b6a8000d568,0x486e5bd8e010d3e8,0x809a0652a4323228,0xc012a25024306668,0x220290000101462,0xab5c431234915aa4,0xe42321b2012e4a4,0x63e5646e613f668,0x41feb1a8003150ec,0x4b2e7f82a47372a8,0x5a226462805052c4,0xf8e2e2a8612e4c4,0x6bfe9fd326d1f7ec,0xf0310464062276c4,0x41280d2006325044,0x38610450230e040,0x6c77e1a7432f060,0xfa44ded20492302c,0x85c017ab22923468,0x5a448420a0326260,0x48bc0570e082142c,0x8020000000000280,0x8c480408012c080,0xc100520000006c4,0x4071a40500026684,0xfa186d97147356ee,0x885832010002e6c0,0x4e80851148012e0,0x800004000801000,0x100201800010060,0x40100020200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x202000104000,0x0,0x60000010020,0x3000780c0102020,0x0,0x539213c0569254e8,0xc9a1832f2492556e,0x405b1b57a49076ec,0x498880140015020,0x90729700262272e4,0x120064000332000,0xcbeef36ec25254ec,0x480201c0008016c0,0x12e0c5240905460,0x84b6124036005460,0x931224402602106c,0xca0c20221411d0a4,0x480bfa30e03326e4,0x29332f5914b072e2,0x1080a640040002ac,0xc72da740209372ec,0xc380332680132684,0x41c3fc1060133668,0x14020000020,0x1000800340220020,0x8100300020304020,0x0,0x1800020000020,0x4080000001400,0x0,0xb40dac04a01368e8,0x8800450000114440,0x0,0x2585000103280,0x111165c2c2000360,0x1441000080221000,0x4800000440,0x52000002200000a4,0xee8212ca047372ac,0x1a2447e140323160,0x54f04dcca09052a8,0xd7136550e47076e8,0xfe84656c241277ec,0xc7f7fab9f4f377ec,0x33d2dfd124927eec,0xf4f2fefdd673f6ec,0x99aeb77366737ce4,0x8b80e6c24222aec,0x4cc81200001242e0,0xa000000000102000,0x80146006001444,0x30000000000440,0x12080e0000001408,0x4fcffe2ba493546c,0xf2d6fec9461157ec,0x80bb4c22964236ec,0x8400247000911680,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40200000801000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0xfb6a7f62592c95ea,0x8010000000000140,0x2b6b91dff9bcbedf,0xfffffffbf9bebffe,0xfdfffffef9bebfdf,0xfffefff7f9bebfdb,0xfbbfff7ef93e3fde,0xffffffffd9bebfff,0xefffdfdfd9bcbfdf,0xfcfffffff1beb7ff,0xfffffffbf9bebfff,0xffbe7ffff9bebfff,0xefffff3fd9bcb7da,0xafffffdff9bebfff,0xffffffef79bcb7df,0xfffffff7d1bebfdf,0xfefe7fffd9bebfdf,0xfdfff7fef99eb7fb,0x29dffeff711cbfd2,0xdefb7bf779bc9fda,0xffffbf7ff1bebffe,0xdbee7f6651be9fdb,0xe9f7ee6ef1bc96cb,0x5c931e57d094bfd0,0xbebbcff2519eafc9,0xf575feee513eaefa,0xf7be2d7971b41fdb,0xfbdff77fd93ebfee,0xaddbfebb71343f77,0xa2bdcffbd91cbedb,0xd9d3366b5936befa,0x892ea553d11c329a,0xdb06fb5f513ca7fb,0xcd6aba1ff12c91ca,0xebbe9ddbd13c9ed8,0x34f4f6aad936becb,0xc39a3770d936346a,0xfb9e6ff1d19cb65a,0xe8627b38d1343546,0xaaffffdbf1bcb7da,0x87beed43519c15ee,0xfb6fff6f59bcbeda,0xfeb63fe7f10cb7ce,0xef7dfa7f591497fe,0x9eeebe941b6b7ca,0xd3becdbb799cb7d8,0xb1dd57ec719e9dde,0xfffe7dfdd19cbfde,0xaba6fb79d114b5db,0x2ff71ed3790eb5cf,0xbbbeeffbf1bcb7db,0xd6bbfef8713eb7da,0xdffefff9713cb7db,0xcbd5fe4ed01c9dcd,0xafcbb2bf5134bdca,0x7fbe5fd851bebffa,0xb4bf3feed03c9edf,0x0,0x200000200008040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x280000001000002,0xefffffff79bebede,0xcfdf7fffd13cbfda,0xbfeefff3f1bcbfc7,0xf7ffffebd9b6bffb,0xfffeffeff93cbfdf,0xe9f6f6bed834bfd8,0xebd77ec3d1bcbe8f,0xefbceffdf1bcbffb,0xfdf4effee11ebddb,0xcefeed7a59b6b7de,0xf97ccb3e51149cfa,0xdffbfebbf1bcb6ee,0xfffdfff593e97fb,0xfdf5f6ffd11cb7da,0xdbbaaf75f13cbeca,0xdb8eefb7d936b6fa,0xdfeffe3b71b697ce,0xa9ffcfde799ebbfa,0x0,0x0,0x0,0x0,0xdfbdffdff9bebfff,0xfffffffff9bebfff,0xfffffffff9bebfff,0xffbe7ffff9bebfff,0xfffffffff9bebfff,0x2af6bf7f519ca6fa,0xfffe6f6af1aeb7ea,0xd3ddb67951be32ea,0xef5e7f67f1b4b5db,0xcbd87306d19ebccb,0xdcd6060fd1949dde,0xb9b3cfe15194be0e,0x307f77cd509c91da,0x7ebe7debd91c9fde,0x9314ee0c01142cca,0xc48200910108490,0xa11616d15114a078,0xf1c4bee9d82092d2,0x3d747f4351148fd3,0xcb29e02b5014a3d8,0xcefefebe71bca2da,0xefbfdfdb71beb7da,0x24744dc0510cb482,0x5ba6ac3c5030b7c1,0x7fb2567fd01896ce,0xc9e7b6aff194b7de,0x8255404a110c06e4,0xb0e75dc451163cd6,0xa1fc257a511c2af8,0xd096276fd10435cb,0x8007fb63d1149e87,0xaf363cdfd09c87c7,0xb6bbf77cf91cb7d2,0xd16036ec713c96c2,0xcbb237e4f11c11fa,0x900d33a400426ca,0x6e0188820048240,0x8f88fffa988c17ca,0xc153a27a5000a2c0,0x4552fe517034b7d0,0xcb83eb5b5910bcc5,0x5e4b925f71048178,0x3084000100150,0x7464f66840bc9dd2,0xe162fe0419c00da,0xc892742251a41148,0x6d33e2a51168cea,0x28106ac240941842,0xb4bfb0e75134b87a,0x91aea5205084b088,0x53367267d12435e2,0x8e83e30a911699c8,0xee465e98511ca81a,0x85a2022f910696c2,0xf504260851101e80,0x4400000240040000,0x81908001300480c0,0x260004858101200,0x20000428000c0240,0x2272c000e10c8,0x80044110000480,0xca88d32af81c30c2,0x8851840400c1088,0x9080b1051000082,0x8440144019009140,0x0,0x6011000000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x41000000480,0x200000210000400,0x0,0x200020000001002,0x21020640100005c4,0x5b96a761d00e17ca,0x830ed36e5194b0d2,0x4050123e5110ac56,0x880810800248048,0xd5d4f62cd13484da,0x682084001202342,0xd100b70c081c9a42,0x4ee39804d90434c2,0x622201425104844c,0x46c1744d114068a,0x4b963832d01492c0,0xcf14d3f851148ec2,0xc8cefa2170341fda,0x94e815ad984109a,0xba426060140358,0x8128282191163048,0x1220032040022240,0x29c0c32b580c96d8,0x200780180001142,0x300071250008490,0x90882548d1049048,0x8000020040000080,0x8402200c0001080,0x808005440003088,0x104c8000041082,0x4120853161048663,0x90102d4040248058,0x3808000100c0400,0x265d080040860140,0x1048c2100005c0,0x8b6866891949058,0x80440200100040c,0x0,0x494e93d041949252,0x1814ed40500430c2,0x44d1c2c471949748,0xfb28f562d0b6b7c6,0xc3b2420e511ea648,0x8c4bda19d136bcf2,0xabfe87da519cbfda,0xfcb61767f9bcb65a,0xff2ff76dd1beba5a,0xf694bfd6719eb6ce,0x4080c00040003080,0x141901050208050,0x1000010000000140,0x40210009040,0x9a0c2922591cb40a,0x45d2bc8a508636d2,0x6f5ef759513cbedb,0x171a55f1511c93fa,0x40140191040008,0x18000002040,0x0,0x0,0x82000000100800,0x0,0x0,0x80800000100800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200040000c0000,0x0,0x0 },
	{ 0xffef7f63d0e9bf6a,0x60000000400,0x408118100010140,0xffbfdffbf8ebfffc,0xf7fff6fff8ebfffe,0xfffe7f77d8fbfffc,0xffbdfffff8f9fffe,0xffffffbff8fbfffe,0xffffdfdfd8ebbffe,0xf4fffffff8f9fffe,0xffefff7ff8f9bffe,0xffbfff7fd8e9bffe,0xfdffffaff8f9bfe6,0xefffdfdfd8fbfff6,0xfcbf7ffff8fbfff6,0xdb7dfffaf8f9ff76,0xffbe7fffd8fbeff2,0xefffffbf98ebbf76,0xcfdf7f7fd8f9eff0,0xbff3dffff8f9bf26,0xf7ffbffab8fbbf72,0xbfb67f6ed8e1af56,0xfbf7ffbbd8e1ff76,0x7fd7ba39d8e9af6e,0x3cc95eda98412f74,0xf5f6f7f198eaaf70,0xfffe7de3d8fbaf7e,0xfbf7e36ff8e9bfd6,0xcc7bf7ab98e1bff6,0xefffcf7ff8ebaffa,0x74f77ffdd8e12f5c,0xef9aaf63f861b7e6,0xeba24b8e8849e9c6,0xdffabe13d8f1ef76,0xaffbdedbd8e8ef76,0x74a854c498e2af00,0xf5feff7198e1ef74,0xfbb6f766d8b0fd46,0xefcfff1bd8fbfff2,0x2e7e1fd5f8e0ef70,0xcc977fa998f8fb30,0xa51a5f6290e04720,0xfeb63dded8fbe79a,0xc7adf73ed8fbaff4,0xcfffbfcd8c9a704,0xbfff7ffff8e1e77c,0xf56fff6bd0c1bf0c,0xdffe4f73b852bff2,0x884dc36c98d22718,0x27ff3e5ab8d9bb42,0x8b9eefd3d8c9af7c,0xd1fef6edf0e9bf72,0xf2bc6d5398f0dd20,0xf9d4effb98e9ef70,0xeeee7bceb8e96f5a,0x1004000000208,0x0,0x0,0x12000002004,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xb8106f2290702760,0xdffeffbff8fbffda,0xefdfffdfd8f9af7a,0xd7bbffffb8fb6b7e,0xfdef777ad8abfff2,0x5fae6febb8f9b7f6,0xebf7ff2ff8f1ff76,0x69f2be1be8f1ef46,0x8fb9efdcf8d9eb72,0x9517b6f8f8e9ae70,0xcbbeff72f0693f74,0xfbf7ff7eb8e0f772,0xcec3ffcfd8e0e562,0xdfbeeffeb8fbbf7a,0xd5f73feed8f8b77e,0xdfbfffe3d8fbbf72,0xfbbcfffef8f9bf7e,0x9fd7ff1bd8e9af76,0xa25880dab0e96752,0x0,0x0,0x0,0x0,0x9fa9ffdbf8fbfffe,0xfffffffff8fbfffe,0xffffff7ff8f9fffe,0xffbffffff8fbfffe,0xf7f7ffbfd0fbaffe,0xe61ffafbd079af70,0xb7fe4ffbd0fbaffe,0x91ed756ab840be74,0x7bd657f4d8e1a772,0x699befead0e1bf78,0xb7577ffe90f3bf36,0x2c9eefe398e1af52,0xdbf3d67a90e2af28,0x41fe2dd898e12720,0xfa96cb2e98babd62,0x4dd3fe4c98cb2f70,0x8ab44c70d0e2bf00,0x1c7337e190baade6,0x93b1bd7090f1bd84,0xd384cb3cd8c1a500,0x43de921f90e16774,0xffffffdff8f9fffe,0xf7f67fe3d8a0af56,0xcd8a75439001d9e0,0xd3107e3cc029bf6a,0xa986e7b7d8402f66,0x67ffd7d8d8ebffd6,0xbff33ff3d8e1ff70,0xf5f48ed0b8c8ef60,0xc2843971c0802b70,0xa5e6d7129820ad70,0x4a3feb4b98a0bf78,0x969f5dba9809e350,0xf0f5befdf8f1f7dc,0xd192185098408350,0xa9d4f36e90a8677e,0x4864025010000510,0xa0b2042010226706,0xc041a8af88002d20,0x2063120c0006500,0x4bc4e10a3080af20,0x164a3218a8e90330,0x8182c81098009840,0x51a665eed089f752,0x95122c61d063af60,0x8dd0ef6a50a0ef98,0x85e2ca38b0c08e70,0x2f181ed010e1af70,0x7ab4bcd090e3bd50,0xd0acf76098eaefa0,0xcb9637d6b8fbab78,0xc966fba398a0bd40,0x55e065490f0bf00,0x563f672b90623760,0x10ac206290422214,0x808043cb0900a20,0x21c4803090410150,0x2c1e5141d8002214,0x420080000800,0x861850002414,0x5800080200202604,0x8300a10b90212050,0x608000010009100,0x8002000010800000,0x420200000082308,0x0,0x0,0x1104880010800,0x0,0x10100090000000,0x0,0x0,0x40001008400000,0x0,0x2020000000002400,0x8000004080000800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x48000400200,0xc088a00010882208,0x8a766df198e3a7d4,0x8284e2849049a3b8,0x48c0921098400600,0x12480a8821a040,0xdc30b26498202c30,0xc90ea770b880bf78,0xd2b5e90e8068bf68,0x8d4abfab9060ff22,0x220c74808602204,0x1040040000608200,0x4008af6090c22500,0xd1086d8690492c50,0x8122f28290b02630,0xa289155a1021f560,0x22500000018140,0x8000040000000400,0x4980203990488310,0x82907a03c0c02706,0x2100049300412644,0x11000000202000,0x8066074090a02c04,0x948409e4a8608760,0x808c22008002400,0x401004810002410,0x8400000208,0xd1a8727090a02600,0x1408084080080260,0x200200008080,0x60004c8020200340,0x8094c00230089290,0x400301080002002,0x440224489088a314,0x800656000412244,0xe96a120ff0418ff0,0x12014718904b2648,0xb050126498402c04,0xb35a744110e93762,0xd8822e6590c02c10,0xc6d3be19f8e19f50,0x2b3e87c7b0ea2b5a,0xb67474eaf8e2ab56,0x96ef7564b0eaab56,0xfbb02f7a9862b2e4,0x10c600290820010,0x0,0x4822054090c02202,0xa00000010040,0x482006aea0612a42,0x4d8a7737f843a314,0x1cafe3ab820aada,0x66607e2b0e92250,0x9000500090809800,0x0,0x0,0x0,0x401090000000,0x0,0x0,0x0,0x100080000000100,0x80800008000400,0x0,0x18000000210,0x0,0x0,0x0,0x0,0x0,0x200000200002400,0x100800000000100,0x108001140002120,0x4012006a8020a400,0x11200c0208202400,0x22d5000a04500,0xa250430608000520,0x24c4080086085a0,0x22000a0000a500,0x404000000000100,0x440004000000a480 },
	{ 0xffeef7f9fffefd63,0xbbd67b6dedee7c63,0x1b9b6b324beeb560,0xffffdffffffeff63,0xfdffffeffffeff73,0xfffffff3dffe7f77,0xfffffffffffe7f77,0xcffffffffffeff73,0xffffdfdffbfc7f77,0xfeffffffffeeff77,0xfbfeff7a7f7eff77,0xbbff7f7dcffc7f63,0xcffffffefbfeff63,0xefffffdffffeff77,0xfffedfebbfee7f67,0xfdf8ef7eff7cf761,0xfea62ff7df6cff66,0xefdbffbf9feeff26,0xeddbdfff8f7ef772,0x86aa5ed38b3cef65,0xfdfffffc7f7eff67,0xfebe7ff66d6a7f66,0xcb53db3e4ffcff72,0xeff7ffdbdb7af763,0x9f3edff3fb6e6d62,0xb374df7c0f7c7f62,0x4bbe5fe36f6e7763,0xdb56cb5ec3e8f743,0xeccaf6feedaefe23,0xef1eccd39b4e7f62,0xf4fffffefffeff73,0xfffefff2cfee6f73,0xff8cfe7ddb3e7d73,0xcffff93f5fec7f63,0xa3de5756db6c7742,0xbeffffeefffe7f73,0xf96a6f41cb48fd72,0xfebd7f66efae7f66,0xcfc7ff3acf3eef60,0xa6decedf4becff67,0xfcfaffe9cbeeed62,0xf96fff7bba68fd73,0xfe1e37ff5ffcf763,0xefdff737efeeef73,0xafdfffeb3bee6f77,0x36f26fead93c7746,0xfe7dbf7fdfee7766,0xef9e3decef6e7567,0xc9fdfbf64a0c7f67,0x8ed67ad94fbc7d60,0x9fdf4df3796c6f63,0xd7c2e67b8ffe7553,0xeffef7f04b6c7f63,0x8bd7cf7fdbecef62,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfffdeffedb6e7763,0xfffefffffffeff77,0xff7fffefff6cfd70,0xa70cd7791ffc6572,0xfffdfe7bbafcf563,0xffbe7d7ddffeff73,0xcbf6ffafdfeeff77,0xcfcf7f57ebecf573,0x9f5f7f8abffa6f22,0x9553fef1af3c6d61,0xefbe3ff3ef7e7f63,0xdb36fb63fbeeff77,0xaccb3eb5ef6e7561,0x3b9b4eea9fec7746,0xfcf7bfe9db7e6f67,0xffeed734bfc7f76,0xff94eff77fae7773,0xddfffebfef7eff77,0x8000000200202000,0x0,0x0,0x0,0x0,0x8389df5afffeff67,0xfffffffffffeff77,0xfbfffffffffeff77,0xffbf3fffdffeff77,0xeffffffffffe7f77,0x624efac789fefd60,0x174b7f7f4a7cff63,0xf18b776b4b6ef567,0xc2be6ded10acf475,0xcbcdf33edf78f763,0xcef737aecb4cf562,0xc49fd5f65b487562,0x9a6dd7f8eb88f720,0x7ebe7df2ff7cfd73,0xfa6cd92ae87c6c72,0xcf53be76cf48ff62,0xbf2def792ffe7d63,0xb0365eea0b4c7c62,0xa33eb5a00b666c61,0xdb85b63c883cf420,0x4dc7bb3aedee7e20,0xffffffdfdffeff67,0xb6efc7cc2d6efe63,0xe44410a30e400,0x461226380708e501,0xc8f3f87ac2087561,0x69d85f910f0c6405,0xb4be5fefcf6e7627,0xf9efb7dedffcf766,0x560c20e1833c7522,0xcddb9f1817746746,0xcfffefdfdffee766,0x1fa5dffa9ffce666,0xfdfcfffadfee7f67,0x8be645019ec4564,0xe8d5f37793bcf777,0xa9c39e7a082c4f60,0xd09b8d72cd4e7e61,0x96d476f96d4e7e73,0xc93e2de807bc2c71,0xfbddef6f573e7747,0x200301008084022,0x93b1fd3df60fc43,0x112480640d604c60,0x8eb2b5e393bc6664,0xd3066cf9aa2c7474,0x81a852110d642420,0x8d1c5f580b8e4c62,0x70b552e40b266442,0x492ecb601bfe4e64,0xc28064c020b84564,0xcefefffafbfef726,0xcefcd45348aced62,0xa6bef7a9cafced63,0xda7cb78643fe6162,0xca84041803220d60,0x2004000000000140,0x120000040020,0x0,0x5404208080005610,0x2050404008084062,0x0,0x8002114205001200,0x820000000000410,0x650410280a404400,0x0,0x0,0x0,0x0,0x0,0x4000200002000100,0x0,0x0,0x0,0x0,0x0,0x0,0x4100000000a000,0x0,0x0,0x0,0x0,0x0,0x0,0x22000000004001,0x0,0x553e3ef43bcc4e63,0x89853703206c6561,0xe4832a5009441436,0x2020014000007002,0x91f9227a4848ef65,0x28544478390c4c22,0xf3d3fb3e4fbe7f23,0x49e2eb130b226d60,0xb045554083c7440,0x6402000a400100,0x12020c4041042820,0xa10050002084422,0xcad39a293b6ced63,0x25d4f5803465d60,0x2080402000002060,0xa88cfb709b747c00,0x8102142000004140,0x212d001022400,0x50401000402060,0x2810000400,0x9000000000240000,0x400a044209000604,0xec92801744860,0x4044160002cc4420,0x0,0xc2751540438aa542,0x41800a0042006101,0x85128403022c20,0x400806791f402500,0x100400000084000,0x4200805002206440,0x186746081ba0d60,0x8000020402000400,0x45db100189886d60,0x1e78cf120fec7f62,0xf5d097ce9f4c7560,0x900020000a602540,0x0,0x400000808400000,0x82000002000100,0x3020400080404400,0x201008400000,0x0,0x0,0x0,0x4000010000080020,0x9000000002000400,0x10001000000410,0x0,0xc000000000400,0x8050008400820,0x0,0x0,0x0,0x0,0x0,0x0,0x1004000000084000,0x0,0x1800000080100,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4d00484202402600,0x210243000884060,0x84836e1087a8e503,0xa3dd58d00708ed70,0x4dc562bcf426407,0xbb6fdf604b7e7d62,0x709e1b304b6e7563,0xc185b1aa4d4a7f66,0x4a154b4057247460,0x801267f3906cfd20,0x30000000002000,0x1a0c072409286622 },
	{ 0x382421800221460,0x82061a400e003440,0x2deaede2ab922d48,0xffffeff3ffffff68,0xf7ffffffffffff68,0xdffefff7ffffff68,0xffffffffffffff68,0xefffffffdffffd68,0xffffdfdfffff7f68,0xfeffffffffff7f68,0xffeefffbdfbe7f68,0xffff7f7ff7feff40,0xffffffbfffffff60,0xefffffdfffffff68,0xf7ffffffdfffef68,0xfbfffffcfffffd68,0xfab63ff3ffff7d60,0xeffff69ff7f6ff68,0xbbff7efbdfbe7f60,0x3db14ffa4fb62d60,0xfffffffad7f6fd60,0xfbbe1d610bbe2d60,0xabdafb27d7ca4f48,0x7edfbffeffbf4f68,0x9df6fffbdeb7fe60,0x7ceef6f54fbeed60,0xfbbe7d62fff67d48,0xd936ff360fbf3d60,0xf6e79e4ac39a6f60,0x6faecb7bbfbecf60,0xfdfffffefff6ff68,0xffffff775ffeff68,0xdf9fffff7efe6f60,0x9fe6fab397be0460,0xbff2d558579e6560,0xfcf7fefeffff7f60,0xffffff7ffffeff68,0xdfffff679ff67f60,0xe447ee1a5fb67768,0xfbfdffd1ce9ffd68,0x86dbfeff5fbe7f60,0xf56fffff9fffff68,0xde9caf7e86b71f60,0xcfdaf7b95f96ed60,0x791effdf8ebe5d60,0xffffdffbffbefd60,0xdff9bd6576aa7c60,0xfbb63ffecfdfff68,0xebfffbff77be7f60,0x6e9f9f4c668b3c60,0xf38ec57287f69f60,0xded5f074d59e6f40,0xfffefff3ffffff68,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xffbebff6fffe7f68,0xeffffbffffbfff68,0xbfffbfc9ffbefd68,0xb799cffb6fb27f60,0xd1cdf678ff9af760,0x5deeff575f96e760,0xebb7ff7ebeeef760,0xefc3fedf37b62768,0x2798fdfa07d67d48,0x9ff3f6f077be5f40,0xff8ebe7217b6cd60,0xdbfdff2755b76d68,0xeffff207cfffbf68,0x2fbfcdd3c79e2f60,0xf4f5ef6ddefe3d60,0xdf9bef559f97ff60,0xff90f37fcfbf9f40,0xcfdfffbf9ffe7f68,0x0,0x0,0x0,0x0,0x0,0x0,0xffffffffffbfff68,0xffffffffffbfff68,0xdeffbffeffbfff68,0xfbf7f3fddfb6ff68,0xf5fffcbb5b76d60,0x36fbeff997bf6f60,0xd4ffbe7ae7b77d40,0xdd166f77d2b27f60,0xebbffffa0fbfbd60,0x67f63eb3daba8d60,0x30becdea5f9e7f40,0xb6b87efed69aed48,0x7bbe7dfa9ef76f60,0xfaf9fb2effb6bd40,0x6f4ba27947d8ed40,0xcd9ddf521f12bd40,0x54e5356fc39e7e00,0xdfba7d45d6b6ed60,0xc9f2f77757ff7d68,0xcde2e50e07b99d60,0xaf5ace5776967d60,0xf5b7feea1ebeaf40,0xd73a2f397f96a400,0x96b0b67842b62500,0x41890800922120,0x3c65b5905b61d60,0xa01a44a11fb28d68,0x5bec77731ff64f60,0xfc1c2d5917967e40,0x836445e13a902440,0x2bef99da17b67d60,0xc7ac52307b0e500,0xd77e2f60379e3d60,0x380060000200440,0x82b22b4b12923920,0x25511ffa728cbd40,0x9fdfedd35f9aff60,0x80227253140a1568,0x4286190080925b00,0xe204f1030a129540,0x4852161e43a96c40,0x8010191203a08640,0x9c77d1c67aba6d40,0xc49f2524f6986d40,0xefa4b6651fbf7d40,0x1156a6121f906f20,0xd91fcf4315967e60,0xa8741ee244126c20,0x1578846840d26000,0xca00140652d02500,0xc492cb00063e4d60,0x40909202d06000,0x300627e00b124000,0xfb2e77229bb7ed48,0x8a938368161e2d00,0x5800c51a14d71f00,0x8e60808018a5600,0x8010101800200,0xbc201014930820,0x820440008800600,0xc20001800000,0x10040000010800,0x100007000802420,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x842070004944140,0x0,0x0,0x0,0x0,0x0,0x40200000100000,0x4ad7d36a131a2d28,0x4fd23e00ce123d20,0x90100c6083160400,0x156ec6d08b852d00,0x4edc7d6b93b2ef60,0x920cc32613960d00,0x40c393de06b67d60,0x2ffedb5b4e36fd60,0x8050100002082400,0x2010040002000,0x8284016442827440,0x844a721102862d40,0xb7a8fd786126d60,0xb4f5f6e83fff2d60,0xc072444354942700,0x1080254000d23840,0x6e4126924e869500,0x2118c84a0ac61200,0x10206000000560,0x9010000000000900,0x8a080a0a8a020440,0x80a2a80001a24c40,0x23611c4002c40,0x480210000100900,0x8440870004063d00,0x528024041a900900,0xe000e08106b82540,0xe4230102c124540,0x8442401016100500,0x9188062a02021c28,0x40a2055011960900,0x8000a70010120540,0xeca120b1fb83540,0xbecf701e986760,0xb07796fcff9e6f20,0x928aaf11d68cad60,0x0,0x8000100010800000,0x0,0x140000120000,0x9210810000860640,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100001001000400,0x0,0x3000000000020800,0x0,0x0,0x0,0x21000200040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x81040402a000400,0x20423420120a0500,0xca206f3053961560,0xef86da0a53a22f20,0x8bbf9c0561e6d40,0x1c9f17e352b37d40,0xac1c266302be3f40,0x40261de052a68540,0xebf4f312179f3d40,0xe8031ec204b84f40,0xb4b3667add947f40,0xd33227191a925560,0x4d0e176286d22560 },
	{ 0x8ad43026d6006245,0x675e41b2042df4,0x1b3e4d1bc78c276c,0xfffffffed79e7fef,0xfffffffff7be7fff,0xfffffff3f7be7fff,0xffbffffff7be7fff,0xefffffffd6be7fff,0xfffedf5bf7be3dff,0xfefffefff7b67fff,0xfffffffbd63e6ffd,0xffbe7f7ad32e775e,0xefffeffed7be7bef,0xefffffdfd7be7fff,0x97ffdffbf63e7ff7,0xfd7fff7ed73e6ff7,0xffbf3ff7d7be7fff,0xe9dff71fd7bc6fff,0x2976fed1d62e72dd,0xf6f37f7ff7967ffd,0xf7f5ffffd61e77ef,0xfffeff7dd7be7ffc,0xebf7e73fd3b46ffd,0xe8d75fbed6a67aed,0xb6bd1ef2d68c23de,0xd5f6fe7fd604634f,0x7bbefd31d40c7afd,0xbb1ee279f4346147,0x5dd73f5fd2ac2ddf,0xefbe5fd0f6ae63de,0xfdf7feeed63c7bf7,0xdbfb7e60f716652e,0xdabcfe7bf6363ddf,0xdec3febf931661be,0xb3d4dfdbd79e71ff,0xc3fff7e9f7ae7df7,0xd3feffd8d60c2ded,0xffbebfe6d68e23ef,0xe9d7ff3ed69633d4,0x6ff6dddad6346bdf,0x3db3dff3c62e6ddd,0xdffcff53d61e77fd,0xdab2af6dd73e635e,0x89f3fe37d61e297f,0xb7ededed61665f4,0x37abef53d69439cf,0xb3efa76a938c3de5,0xfbe67fe5d28e2975,0x8bb6fdead49c2261,0xfcd7ae08d68c697e,0x3d812540d4202328,0xd5f3f77bd6be2fe6,0x8084c21440c0020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xffbf7ffff7be6fff,0xefeff7fff79e6fff,0xefffffdff73e6fff,0xfbbfeffbd73e2fff,0xc1cd3773d700276e,0xdef66d70d6262fce,0xebfffffff7be7fff,0xefdfffdaf6166fff,0x9ebd8debd62e29dd,0xc4f7bfebf6062fcc,0xdea6ef61d68669d7,0xfbd5ff2ed6042fdf,0xebdfbff3f6946bed,0xefbfcfdbf7b62f7f,0xbcfdfeff93be7fff,0x9b7effe3d6be6b66,0xfbffff7ff79e6b7e,0xa365f683f6a62de6,0x0,0x0,0x0,0x0,0x0,0x0,0xb6f7fffff7be6fff,0xfffffffff6be7fff,0xfefefffff7be7fff,0xfffffffef6be3ddf,0x3b7fffd9d6ac2e75,0x26eacf7ad6043b7e,0xf5fdff7bd79e6fff,0xfff6ff73d6be6f7f,0x89a2e737d60629df,0x4ad71ffbd7ae29f7,0xcbbadf7ad2ac2475,0x3cf67574d6a62bc6,0x67b63d50d62e65b4,0xcbf6e3eed62c6fdf,0xcd4a3e4993042c25,0x97a74d7ad71429f7,0xf51f16ead78c37fe,0xff3abc71d2a464e5,0xd6a22ffad6202f5b,0x8ffa9b1bd2843b7d,0xb578491292002a6c,0xb4b45ff6f2a42ff3,0xdf6efd42d60c7f97,0xcd083d64560c28ec,0x8fdfbb28f5ac6f50,0x25c5d08d40438c8,0xadb6ffbad6002df6,0xdb985d42d220297c,0x2c07f582bc01f2,0xa847f3e2820c2900,0x2c3f97d5d7242b5f,0x491178ad00c6943,0xd67dac78c60630ca,0x34021561460c0184,0x81a6c22692143023,0x44e24a1690042149,0x4dd2817256a42952,0x8424282976002816,0x3022d40c400209b,0xda80f966d6362972,0xc64b162192143901,0x92c86ad6002854,0x75779aced70629cd,0x8cede77252202be5,0xd21a6722c6a439cd,0x8583ae00c20c21c1,0x180d81440200d0,0xd0150a60d2022262,0xdafeef6092206759,0x4da41451c60c2892,0x83266202d2064386,0xab5b3217f7104be2,0xc000002080000100,0x910004a890206044,0x768ca651c60e2180,0x9018320d08020c4,0x8840085000043001,0xe024046091140200,0x41284470d20c2052,0xa40112000300,0x40200080100100,0x3540400c0000102,0x808201c0040080,0x0,0x0,0x0,0x0,0x0,0x2844000000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xcb065141d20020cd,0x6054020032002804,0x1d87f7bd7102f45,0x82814046002804,0x8052020482002002,0xd88ced42901c212d,0x8b02ce2cd21c216d,0x4c5b9a0ed624306e,0xa2390e5bd4843960,0xc0e49208c680005a,0x803c2420540c0020,0x8400cb4cd6040844,0x9d8b9a40d60c2545,0x981c048c5002401,0x94dbdeffd6a43bdf,0x2d680561d4046980,0x8080104800802244,0x2000c21094002160,0x68e7d040f28c20c9,0x92046a30041200,0x22144d0000118,0x4c92050292102642,0x443232054002045,0xa100085080042890,0x0,0x1002000c0000,0x522014080002840,0xa9aea90cc6a02544,0x40080000002400,0x8000002080004000,0x1102000c90842044,0x640a344892061334,0xc0809346d2041114,0x2e8a114ed6006985,0xd0864d21920c4374,0xb49475fad60c395f,0x2282c0030043820,0x200000440000080,0x840300042000010,0x0,0x0,0x801002000004,0x880020250000042,0x0,0x0,0x2000010002000100,0x0,0x0,0x10010000800,0x6000030000000,0x490010000020,0x41002b0002000,0x0,0x0,0x0,0x0,0x400002000002800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x48945521d6002904,0x42905b7cd61c330c,0xd0f082f7024000,0x3fc5f56d680217f,0x8922c1c0822c60c0,0xfd7f7f4fd7946dc4,0x40b00533c7002114,0x8005f264d4843069,0xaf1b0bc2d68442ee,0x326281c600281b,0xb0f09649938c3054,0x10002c00830c2000 },
	{ 0x8192412081941120,0xe20211c20b940322,0xfb9e4dff89d47afe,0xf7efdfac89dd4377,0xfffffffeabff7bff,0xfffffff3abff7fff,0xffffffffabfd7fff,0xcfffffbf8bff7fff,0xffffdfdfabfe5fff,0xf6ffffffabff7fff,0xfffff77fa9fe7fbf,0xff9fff7fabfd7bbf,0xe7cbef6f8bfe3339,0xffffffdfabff7fff,0xfebe7ff78bfd7b7f,0xfbffffff8bdc73fe,0xfbfebf7f8bf67fef,0xe9feffbb8bff53fb,0xf5ededa89dc132f,0xb6fbffffabfd7bbb,0xdffdbf7c8bff7f7e,0xb7b23fed83a46fb3,0xebfefbbe89bd53b7,0x68f697de8bba77e6,0x5e8ec5cf8ba81b7f,0xfdffbefb8bfd5f7f,0xdfeff70ab9d777b,0xf9fee3ae8bd47ba2,0xd6fbffdbabf67f7f,0x9f9f5f4689ba5398,0xf437eefa8bdc53ea,0x8dfc577089fa1ab2,0xcea7f66a8b941f3c,0xe986df098bfc63b6,0xbadecfdfab9a53ed,0x36be7ed88b9b132f,0x4d642fd28be2736a,0xcbb66fdea1b0132e,0xcffefef58bf54bfe,0xef7fdfdf8bab73b7,0xb5bf4fe389ef5af2,0xfbfebd3a8ba87be6,0xd6942676abd65372,0x6916d7620bdd6202,0x6c57fbc789b61b36,0xb73b5ff389b273ea,0xf0e4ffd8a89c53a7,0xc6beedc6839853e6,0xcad1739eabed6bf6,0x7cef3a998bf95b86,0x5990cfa38ae843b7,0x420000000000300,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6080000800,0xffffffffabff7fff,0xffffffffabfe7fff,0xefffffffabfc7fff,0xffffffffabff7fff,0xd72caf7fabf47fae,0xb77e7f7d8bf653ef,0xcbffffef8bfd7bfb,0xefdf3fffabde6ff3,0xbfffeff6abff4bd3,0xfdffbf7b8bfe537f,0xffbe7f7babfe7bff,0xfb97ff3eabf47fee,0xefffffffa9fe77ff,0xffffdfffabff7fff,0xfcdeffee8bb877bf,0xffbefd708bff5b7b,0xdb962b7f8b9c73a6,0x300400288400002,0x0,0x0,0x0,0x0,0x0,0x0,0x8675f6fd8bbd37ff,0xfbfffffeabff73ff,0xfef6efffabfc637b,0xffdfffbfabbd5bb7,0x2e7f5fea8bfc0f7f,0xb7ba7fff8bdd73ab,0xd3fdff7babff73ff,0xfffefff7abff7fff,0x89f5d772abb27fff,0xeff77eefabbe7f7e,0xffffcfff8bbe6fff,0xfd76f6dcabb77bee,0xb89c3d6a8b904ace,0xd190ef33ab9d7a2a,0x4e4f76b28bdf472e,0xb73e0d668b9043e6,0x75ac077a81a41a27,0x5bb22d0000140760,0xf7b67b78abbc0f60,0x48defb0588b05a8b,0x6b965f5aa9b46b3f,0x9070d5e88bb47a00,0xd9eebe4889b953af,0x7c2c6c70a0b51b12,0x87d2e91889bd6b42,0xa0b21958894c0b02,0x2c11010088a07281,0xe96c76e38ba4437f,0x9cce3e7581b04bbe,0xca06e37a89ae1f34,0xa9dfe1d38bbe5bbc,0x16b146caa9a04292,0xd23cbd7e8bf25379,0x794653e28ac82222,0x891af66989905278,0xf84a925b09d8037b,0xb63b2de283902b33,0xd47476c3abb0027e,0xddeefdd3abb45ff7,0xfaa4eb0a8bb05935,0x8e5a1e47893943a2,0x27978dc389fc4bb6,0x550c161d03b142ca,0xcb8c3d6009826531,0xc890e7ac8b10530e,0x8c5f263aa89473e2,0x809e055201221228,0x9cd454020ba24230,0x87b30d20a9d04b02,0x17be064309d0006b,0x804002000080020a,0x921655908190310a,0x84d710a289803223,0x902424c20b907a00,0x4090051181001022,0xd88040068b811122,0x0,0x0,0x6124052009026300,0x1002244000101320,0x40c5a30689841a34,0x4950021081041320,0x3000000000100200,0x8000100001000040,0x0,0x800010000100200,0x0,0x800080000040200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000200000100a00,0x800200600a000320,0x7575d0288984303,0x8db46dea8bdd53a6,0x85c02008100420c,0x904f24008a04282,0x4380e72189f10b32,0xa9539bdfabfd2afa,0x6f9dcbfb89fd7bef,0x9d97ddec8bdd73e7,0x8884821089401002,0xa00e100000c6280,0xc58a0008105b20,0x92260f480a8c43ec,0xbcf8ede789de0a67,0xc5aa8258a1807b22,0x808402000ac80200,0xe567ef1eab9957a2,0x22ced8d0a9f41b42,0x8480dc2082a01320,0x51a6014082a10326,0xc20462220b904208,0xc23e60289d82262,0x2010108008200200,0x422000200802010,0x120d50182815102,0x100161408400380,0x8000820408400000,0x40521668018503a3,0x4680a0800010,0x4035502088c02104,0x2424c881912282,0xfbc4d4268bac7388,0xa64afe40abb8333a,0x257247c0897572a6,0x7494dece89be5baa,0x4b28b4038b045a62,0x0,0x80100080000800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1800001100000,0x0,0x0,0x0,0x0,0x0,0x402000080000004,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8002000001000002,0x0,0x0,0x1483620a2204b08,0x12843b2e899853a8,0x2ca6e00098c6630,0xabdeded089b06387,0x289ac7ef89f473f2,0xd37ebf62aabc13b6,0xc2924d528bf04363,0x8072338a89d142a2,0xb48c94089101224,0xf130eff389ff72a6,0x31967c6a8b8c5bb3,0x1000094008800202 },
	{ 0xabd4f32acad5e12a,0xe2500252810c418a,0x80bfd6b2d354c02a,0x3c6e66f86185c0aa,0xfdfe96f87b4dc30a,0xfffffff7fbddfbfa,0xfffffffffbddfbfa,0xeffffffffbddfbfa,0xffffdfdffbddfbfa,0xfeffffeffbddfbfa,0xdfeeef7bfbc5f3ba,0xfffc7f3fdbd5f33a,0xedeffb3ee1ddd97a,0x7fffffdffbcdf37a,0xffffffebfbdcfb7a,0xd9feeffbf9cdf33a,0xfebfbf7ff3ddd33a,0xffedfffffbddd3ba,0xafffff7463d5c34a,0xb6fbeffbfbd5db7a,0xd5fef77ffbc5f31a,0xfbde7f65e1ddcb1a,0xe8d7fbafcb55e32a,0xcddbbe7fd3d5d3ba,0x8fbadff3f94dc37a,0xf3e7deeee9cdfbba,0xf5bcfd63ebdcd0da,0xfbb4ff9e71c5d31a,0xcedbb77de3c4c17a,0xaeffd7ffe3dde32a,0xf09fd6f7f3d4d96a,0x813eb7136908c33a,0xc79e37367b44d1aa,0x9dffbf1369d4c3ba,0x8bffdfdfd3d4cbea,0xf6f7ffeffbddebfa,0x9dee376ac9d4c32a,0xdbae777e5bd5c33a,0xf7d7ffdf7bcdd3aa,0xff5fffdbfbd5fb7a,0xd73f46eb5954933a,0xd17bab70e3d5d938,0xf4be3f52f9c5ebea,0xc6f87836e15dd26a,0xf8c085951c0411a,0xbdb7fffa63ddc2e2,0xf8ed767fc2cdf17a,0xb392fd69e1dce27a,0xdbf0f77eebc5cb3a,0x6ffffed3f154db3a,0x10020000008008,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfffffffffbddfbfa,0xffbffffffbddfbfa,0xeffffffffbddf3fa,0xbdf7fffffbddd3fa,0xa0baddeb4bd4d3ba,0xddfdaf7efbddf13a,0xfffeffffe3d4f37a,0xcbfdff2afbd4f3aa,0xe4bb3f8fcb84d82a,0x9f7bcfe3d9c4cb3a,0x97f5967ce2c5c32a,0xfffe7ffafbcdd1ba,0xfbdeef7ffbcdf1fa,0xefdfbbfbf8cdf17a,0xbe1ccff97bdcc32a,0xfdf7ffffdbd5fb7a,0xf7feff735bcdd33a,0xdae4ef3fc3c4e33a,0x0,0x0,0x0,0x0,0x0,0x0,0x20080040001000,0xe4f5f67ffbd5fbfa,0xfbfffffafbdccb7a,0xffbf7ffffbd5eb3a,0xefdfff7ffbc5eb7a,0xac5ffed94984c1ba,0xb7ba5ff2fbc5eb0a,0xfbfefffbfbddfb3a,0xffdeffe37bcdd3ba,0xdbc7e7fac3ddd37a,0xfcff3ec8f984d33a,0xf9adffafb4cf3ba,0xf57ff2f7f1dcc932,0xef3e3d78db55cbba,0xcbd6e9fedbc4699a,0x4bd27ebaf9ccdbea,0xd30e473b61c4d37a,0x44f6706ee180cb0a,0x338ad0083c4a042,0xafa6db77eb05d30a,0x44ff222e954cb42,0xa39a0d10da4ce37a,0xb6945763694cc33a,0x8546a570c0d0a12a,0xb2301f7cd104f91a,0x186c8004001d112,0x837e5e5679cdd32a,0x3cb85de283d443ca,0xa9505e2061ccc20a,0x4e9e2dffe3d4e35a,0x68de12324144911a,0x2f908e196004c328,0x4030ae18904411a,0xd33cb679ebc4eb0a,0xc88e7d72ebd4e30a,0xb14f31ec15ce28a,0x65c21e1641ccc2ba,0x1e9f1bc06b88c30a,0x845dee50c18de392,0x816a1784144c10a,0x2ac064020185c000,0xcdcbffaa6b5dc13a,0x281e4fc051c4c11a,0x8df45668d185e000,0x9406ad31cb89c10a,0xca80af36c3ccc32a,0x5eebec341c5d33a,0x2271d588344c12a,0x1000000400008008,0x984086108084e010,0xb08450804081c008,0x1d1ec0288848832,0x60164e8001d4424a,0x79318de743c4c39a,0xd80c27906041c02a,0x9a96a65160ddc06a,0x4850900840c00a,0xa76384229c4c002,0x480000008020,0xc860202004002,0x80020000008020,0x8126d14640c4c17a,0x6000000018102,0x25940066e0004102,0x10080000040002,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100100000008020,0x0,0x0,0x0,0x0,0x0,0x2000002000804000,0x80105008514400c2,0x40e200418004a022,0x5a8a4ff86bc5c13a,0x14269042e8804102,0x2a00d0080044202,0xc184fb02c0d0e12a,0xfb57dad3f35cc16a,0xe9becff1d3c4db3a,0x1431ca6000ccc10a,0xa520af728bc58b62,0x4296742489c4ca02,0x1c01ab14a848122,0x856414408d4c00a,0xa49c60600104c32a,0x8041400088040800,0xc80021000104410a,0x48988088808e00a,0xaf1206428345aa22,0x1d8057e75944c12a,0x4000a6686850c102,0x62102a028044a00a,0xedf5f5a2f3c5e34a,0x12d40a8080a000,0x842954db80cde82a,0xd1ca264c6180c30a,0x8810058041005008,0x81008a68814c484a,0x204020168008122,0xc00090001404040,0x204606c081406102,0x20b80d50f804890a,0xb3d67d6ce985d30a,0xba8a33c34b40d32a,0x9bbecf2b895ce36a,0x74be99eae154f36a,0xa021042080006108,0x52800000112,0x1800000000108,0x8040000008020,0x0,0x8400000050000000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4800000110,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6ae065724054811a,0xdacc3f46f140c13a,0x89c1e99ae3c4c30a,0x8b6e5f5bf081d12a,0x4607ae14bd4800a,0x700802fa4804c12a,0x3632320ac1c00a,0x8080000550000000,0x994e4b52e3c5a12a,0x9491dcf57955d19a,0x34a4a80041d4d118,0x4002000001004000 },
	{ 0xe9dcd260029af722,0x6082100104805600,0x82db8be14c1bb298,0xd010566506c22260,0x716fb7fa4e5afffc,0xfffffff3cedbfffe,0xffbfffffcedbfffe,0xefdfff7fcecbfffe,0xffffffdbcedbfff6,0xffff7eedcedaffe6,0xedffbffacedbffe6,0xffb9ff77ced3bffe,0xfdf7ef1bcecbb3a4,0xebdfffdfcedbfffc,0xd5ff7fefced3fff6,0xf92ef7facecbbfb6,0xfbffbf7fcedbfffe,0xcfdff77fcecbfbee,0xabd19ed142129276,0xbcbbffffcecbb366,0xf7f7f7daced3f3e2,0xfba67d604ecbfb70,0xe933fbfb8ec2b644,0xeff7bedfcedbfffe,0x9fbd4feacedab7e6,0xd5fdfed8cedbfb32,0x7ffefff3cedbb7f2,0xebd6f70e4ec3bbe4,0xafcbbf15ce4a32ac,0x2fffdfd9cedbbb7c,0x3ef77ff84ecbb7c4,0xcfba3531cccaba72,0xe9866f6ecec3af68,0xdfffffbfcedbfefc,0xad1ccfcececabaf2,0xf4f37ef8cedaf2e4,0xeb6e7feacec2f7e2,0x9a5a6f2ecec3bbe0,0xdeebffaacedbbadc,0x63fe575f4ecbb770,0x7c9e7ff3c4dbf1ee,0xa1a5aa588e8a162c,0xdebe2f7b8ecbb346,0x785ed74082403e94,0xce7e7f92ced3f6e4,0x95fbd6f3c6c3b724,0xf55fbeeac6dabae6,0x5ab6397ac6c2bba6,0xebd5ffaf4ed9fe7e,0x5cc22e1cc00bb266,0x200040040000004,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400400000101000,0xc94528a06822008,0xffffffffcedbfffe,0xfffeffffcedbfffe,0xefffffffcedbfbfe,0x7fdffe9fcecbf3e4,0xbfbeeffecedbfefa,0xfffdfffbcedbbffe,0xffbeffffcedbf2fa,0x6bc4fb3ac693b3ea,0xedff1ffeccc3fff4,0x4f9efdfbcedbfef2,0xf7ffe068cecbbef0,0xfcf6bde2cedbffae,0xfbb7f376cedbfffc,0xe273beffcedabe60,0xff9adfdb4ecbba70,0x74ff76fecedafb36,0xff7efff3cedbb7fe,0xdfff7b7fcadbf7e6,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe4f7f6efcedbfefe,0xdbfff77ecedbfff6,0xffbeffffcecbffee,0xc38ff72fce5bbba6,0xeffededd4e5bfee4,0x873f4ef24ac3f2b6,0xd4fdf7d5cedbff36,0xfd5cfd71cecbb7b4,0xc99ffbe3ced3ffb6,0xbddb3e9a88dab356,0x8bbec7f94cdbf6b6,0xb7b4766844c3f7a2,0xd7febf51cecbbbb6,0xb5eef764ed29e36,0xeedff6ffced3fbf2,0xaf9fcf5acedbbefe,0x532748eacec99b14,0x93bead70c6d19bd0,0xcb84eb3e4a43abb0,0xef5ef6a1c6c3bbd6,0x8c0818004022022,0x36e6dfec0edbfb62,0x934c760846c0a220,0xcb9377374edbfdae,0x7fdffcbacedbf7fe,0x80de7dd20e9b9260,0x4c022f800e42ae02,0xa9bef16bc09bb3f4,0xaebc2d8246d3f192,0xc0b4cd3286929320,0x214680c046828220,0xf4bb6f66c69bb36c,0xb1109578c083a660,0xfb9e3ff4cecbfb7c,0x8316d3d84a422f34,0x863e5c80868ab742,0xdbbfc9a1ced3bfc0,0x95a2fb594a5bfae4,0x71e55d0ced27a88,0x83a085000040ae20,0xe649b0164e53b512,0x61b41e100e02320a,0x4e617e1c483b6e4,0x0,0xfb89eb3ececbf2cc,0x2fdeffeec6dbbbea,0xd61574acc93b222,0xa6b736614453b4e2,0x842eadd14a4bbfa2,0x90086564024ab940,0x6e2204808040,0x0,0x2000200081000,0x1b6a24424c023620,0xfa882f55cc83a236,0xe852dbac0ecbb202,0x88d6e84dcecaa43c,0x3a200008040a2020,0x2c020080802020,0x81060020c6001020,0xc802dd20c8909244,0xe40000000402200,0x0,0x24080001000,0x0,0x0,0x0,0x0,0x4240800000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x420000000008040,0x0,0x800040000400200,0x80dcc00044002404,0x84400060c422200,0x6904c0802422222,0xf3c233f1cec3a670,0x1482f540ce823220,0x8bcce16ac842f7b2,0x9ad3721bc6dafa30,0x26cc0b8484829342,0x10ca884e021020,0x8528a560c282b004,0x48c0020680039700,0x41a60004123640,0x30842974880a220,0xf434a6e2c60bf164,0x2c298084822000,0xa390220d84c31840,0x81cac248c6821608,0xa0c07474e88d300,0x19402904818100,0x7004a20006818200,0x2010020040820002,0xc5c4e212868297a0,0xe7761fdbc693f7a0,0xe1011304688a004,0x91008152c4c3d920,0x220024408021040,0xa820e3124041b2b4,0x422a1004903080,0x147c082809600,0xf1f4ba524ecab6b0,0xdbe0573cccbfc32,0x91d4e72a46cbfb06,0x4a4613580ad2f642,0xa7decf71ced3bf38,0x309505e8c882d226,0x928a250000109200,0x0,0x0,0x0,0x0,0x2100000101000,0x0,0x0,0x0,0x0,0x0,0x18000808000,0x0,0x4880001000,0x9f9e7ffbc6cb862c,0x0,0x0,0x0,0x106000000,0x0,0x0,0x0,0xc00004800000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x15d265604459b824,0xc3ac6bbfc6d3b83c,0xa0cee8214cc89060,0xa08c8c244ca1824,0xa83224eac20a1624,0x913609044840db04,0x8210400084800000,0x201c08000c28250,0x68f6eac00ed2b320,0x34ab4e6280ca9b42,0xe154d6024e422260,0x4020ac10c000d104 },
	{ 0x8002d16880769201,0x447004808425b08,0x84989db888265b91,0x16ae579848f61a13,0xfffffffffff7ffbb,0xfffffff7fff7ffbb,0xfffffffffff7ffbb,0xffdfbffffff3ffbb,0xfbffdfdfcff7bfbb,0xfdfffffb7ff7ffbb,0xdfffff6afbf7ff9b,0xefb77fffbff7fabb,0xefcfffbfcff33b9b,0xebffff5f8ff77fbb,0xffbffffffff77fbb,0xfffffffe8ff33ba3,0xffff3f7fdff77fbb,0xedf7f7fb8ff77b9b,0xefff7af7dbd39e9b,0xbdbbdffbdf777ab9,0xf2f7feeb8fe77f9b,0xeadee535ced23f21,0x8393e01783f33a31,0xfe5f5ece8dd37fbb,0xcebb57fb4973ff13,0xdefee6fe3df37f8b,0xd7fefff99df3ff91,0xdb80ff3e69533e11,0xaffbbe5b9ef7fa1b,0x8dbdcff689535ebb,0xf5eff62f2d57fbb3,0xb38e3d610a523aa3,0xbba1ff7faad73bbb,0xeff7ff3fd9f3ffab,0xbfdfedcb0f72ba8b,0xfcf7f6f9db633f9b,0xd5eff77f5bf3bfab,0x59847f3528723291,0xd9d2debd9973ffb3,0x6ffddfdffdf6fb9b,0x22fbde710df37aa8,0xc502a7d60de33aa3,0xdeb61e73cde37eb3,0xeffbf6720cd77a8b,0xfdffe7bfcf7fab3,0x2f3addfb4ef77eb9,0x7363bffa2dc37f93,0x5fa63c62a9d35b2b,0xf9dff73eadd77ebb,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x81804d180072100a,0x989412120a509e0b,0xfffffffffff7ffbb,0xfffffffffff7ffbb,0xeffeffffeff7ffbb,0xfffffffb7ff37eab,0xffbbfffffff7fbbb,0xfdfcb7fb0f777e9b,0xd3befdf308765b81,0x81b0dfb148763e1b,0x6ed63e9b8de2bba1,0xbbbfeffb5bf3ffbb,0x55f5e47b13f25e33,0xf1b63d512f523a13,0xebfce77eedf33e9b,0xcfc3bf93e9773a11,0x65e9cbf119f25e0b,0xf5f5ffefdcf77abb,0xebbafd730df6be8b,0xdc162f602d537aa9,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8c7756ff0ef67e3b,0xfb7cfff7ddf3bfb9,0xfeff7ffeeff73fbb,0xc1d7f7f70ff33fb1,0xd573f591c777a29,0xb93b47f289531e03,0xd3feff33dcf23fbb,0xfd4effe26af7fbbb,0x8b86fb8e0ce37fa9,0xe9d3de9a08775b09,0xb3bbcefb6df37e03,0xf744f2680dc6ba9b,0xcfbe7ff24bf73e8b,0xdb93eb0eabf31e0b,0xf8f3b7038877122b,0xffffeffffff3fbbb,0xc8f7deeaa9773f03,0xdba79f700d733e99,0xc98b6b1bca5333bb,0xefdfbefcab777b9b,0x350311d200d23e02,0xdefcbfeb85c37e11,0xdd767fcbcb431fab,0xebbf7727edf77eb1,0xfffffffefff7ffbb,0x2b5c5cd84f532f01,0x10a2de6b09633aa1,0x18dc7f008e023f01,0x8aff17f6f8537f81,0xa396f2f70a72ba21,0xa0c18bcba84232b0,0x86b2f6e80dd37b0b,0x412cd51003507e01,0xffbefffffb773f9b,0xc9f5e36702f69a8b,0x7d2380b18f3123b,0x8697c7628b43ba1b,0x2594f2214041120b,0x2bfa6d510f82391b,0x8230f366087019a1,0xebf23a8e48171a1b,0x2b34435005c23a01,0x70ccdeeea9c35ea8,0x39045df288072292,0xeeafdb7e8ff77e83,0xfffffffffff7ffbb,0x34cc6c045503a03,0x54b0d7ec68711e23,0x54a835088021880,0xc686294008523911,0x90380244402021,0xa25215c608201e83,0x810000000004800,0xd9e9a7d2cf837a01,0x36b6005600631200,0x102401800e03e20,0x998c8c40d827f03,0xa210433a08415b03,0xb474ae4558521a1b,0x50823c400012120a,0x788850040c02368a,0xa46068305463a03,0x804810040020a,0x410420290152380a,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2080008000010,0x0,0x0,0x0,0x840000000600000,0x0,0x0,0x0,0x0,0x8100000000000280,0x154a00828025f10,0x28513510871060b,0x104430282121011,0xb526a26e19735b28,0xd56c3d1009c21b9b,0x8301c00208810810,0xf7cfb64b88621a1b,0xe38fd7726bf37e8b,0x94e3b6e52ad37f83,0xed007c220c03a68b,0xda52320638571a01,0x407388a0016920b,0x210011960801088a,0x24b5f6608e533a8b,0x41e2260008131a01,0x10c091400061213,0x8052ca4101c75a00,0x2ed7d95989d2fb2b,0x3454556009010a91,0xd040c20800063201,0x8e16312122621213,0xc386fd0601133e31,0x9a9eca0119423a9b,0xb7b6cff36fd37ba1,0x919c86020842172a,0x2030f45808711b81,0x826440020013160a,0x2201084008111201,0xb5878d9408d05a11,0x9540be7181531e02,0x3db66df2ca733bab,0x8a74e72009521a31,0xbda36419d673a00,0x889f6f7a89f27ab3,0x70e646ec204332b0,0x833829500150580b,0x0,0x4000040000000600,0x0,0x0,0x0,0x0,0x0,0x14008000800,0x0,0x0,0x0,0x4000001000600000,0xafbfeeddbe777bbb,0xfffffffbfbf77bbb,0xd5d57e6b485212a9,0x8024000040020a,0x8040042810601202,0x0,0x80040000021000,0x404100200402200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x82665422c502283,0xfbb4777239971f8b,0xc145de0148523601,0xdf1c4928c232a1,0x5000012000111800,0x0,0xc0820f4809601880,0x10100000010200,0x8b4dd408dd23f22,0x80a2079009600aa0,0x1474bc41cc432311,0x102004000440200 },
	{ 0x204200400260800,0x2796df086634a2c6,0x7bb2259a625200ce,0xffffdffb6f777fde,0xffffffffef7fbfde,0xdfbffff7ef7ffdde,0xffffffffef7fdfde,0xffefffbfef7f3fde,0xffffffdfef7fdfde,0xffffffffef7f7fde,0xffffffffef7fdfde,0xffffffffe777ffde,0xffffff9eef7f3fd6,0xffffdfdfef7779de,0xdffffffbef7f3dde,0xfbfcee7a6f771dde,0xfbbe3c732e7ffdde,0xcfd5f7bfe477b6de,0x6f57fe7ba3679fda,0xb7fbffff6a7f3dde,0xb3ed9753267e15da,0x7ffe5c75e177d9ce,0xebc7e37b637f34de,0x6f7e3ffc63745cde,0xbfadef73a77679de,0xb57ffeeea27799de,0x77be6f712737ddd6,0xf3f7fb7ee07f9fde,0x8fdbbe5be67f5dde,0x2f3fdb536276fbc8,0xdcf6f7ede27912de,0xdbafaff3e3773dde,0xff8f7ffeef7fbdde,0xbfdffe9fe677b5de,0xfe5ecfdee7361dde,0x71bdf7f2275f17de,0xdde8bf76a2779dde,0xfebe2f7e6a7433dc,0xc9dafe29677e0c4c,0xebff5f5feb7f15de,0xf91acf83627e795e,0xfb7ff77aeb7f15da,0x4a362d2b236a114e,0xa96dff332c5f344c,0x2f9eda4be67ed8d6,0xaa020fec00774856,0xfb779f7be7757dd6,0x5fae2f70a27e95de,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xbfffffdf6f7f3fde,0xfed7fffbe77f3dde,0xffffffffef7fffde,0xffbfffffef7fffde,0xffffffffef7fffde,0xefd7ffffed7f3fde,0xfffb7fffef7f59ce,0xdfffdefe6a777fce,0x76aa1de2605fa048,0xc3f6f37ea27a2bca,0xefffbeff6f7f3dde,0xeeb297d366765dda,0xc776f7fee77e5dce,0x5afe3df163763ace,0xfbffe577e3775fce,0xeffbffdfe2765dde,0xbffe4f63e37798ca,0xfcf5f6fb63779dca,0xdfea6fe1e77679d6,0xd68e216ca1769748,0x0,0x0,0x0,0x0,0x0,0x0,0x200020020000080,0xfc77f66def7fffde,0xdd6eb7f8627f17de,0xdf9fbfff2f7f5fd6,0xcfcafff6627f35de,0xeddffedf2576b7de,0x17a2475e623e81c6,0x83ffffeb2a7f35ce,0x884c756e637630ce,0xc1e7f22ee174585e,0x6afe3ed2e37618de,0xff9defff6637f9dc,0x975ffeefe775fdce,0xddae6c71ef7459c4,0xe1e4f36be67a34de,0xe8e67f776076bdc0,0xbfefcddfeb7f5cde,0xf4f3bffbe3771fca,0xe9e4af74e15600ce,0xc280c716621c1044,0x8dcaf80fa17e15ce,0xa1decd4e26777d5c,0xc425c2f4227415d6,0xe9f7bf70e55531c6,0x6e9a25786b5617c6,0xfec7fe3aa27e31ce,0x4d7a4a4f654e11ce,0xfc974eeb657617d0,0xdfdcbf4ae37c91de,0x82140130210600c2,0xc3c1ce0d621610cc,0x25100e5c60104844,0x84b36662e65616c4,0x112ea758206608de,0xfe3635546a6681ca,0xe87c600a233711c6,0x9773c8a265515d6,0x81088d9ba36a90de,0x142cce54626600c6,0x18253022375804,0xfa54630e64761896,0xe422e482a670040,0xe29add7627765dc2,0x91d2356ea64715ca,0x84082446423501d6,0xc080d1c226640148,0xab57ef8be21e11ce,0x69e5cbdbef6710c8,0xf6b944ec666719ca,0x98722042001115d6,0x8281000202640858,0x8080080000011082,0xba0e1d162163444,0x4a61003a4304806,0x98a6502231630d6,0xce9f646c6c6248ce,0x288a520824104804,0x8520200201010c0,0x823046f0a22400c2,0x952c257aa26608d2,0x4e42557062373940,0x6881812040761048,0x2f46388ae74012d2,0xa45975f12670204e,0x5761f6fbe23f84da,0x0,0x0,0x0,0x0,0x0,0x280000000000006,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100400000102000,0x0,0x0,0x0,0x3000000000102000,0x0,0x876131022661558,0x32280c81a06a1c40,0xa9561662631605ca,0x1d746dc22756105c,0xb800012000501140,0xc8febfdb627614ce,0x8ffe0fe3e676b9de,0xd4b477e92e779cce,0x70a27222a7e1c48,0x2806208a2040042,0x444a2201a03400c4,0x8b0eca5b2b76d9c8,0x410002000010050,0x8065076062240848,0x22900000a2040000,0x6000000001010,0x800ca4026000082,0x8030546a00725886,0x8000250022001100,0x1ac62962034513c2,0xa156e82420741dd4,0x8c291bdbe2561086,0x479ac3f7e6760cca,0xd1ce754c203cd9d0,0x7d8a2f202a7710ce,0xb34e33ba26611d2,0xe23d1dd2e05622c6,0x1dac4d28027c10da,0x98f720706257598a,0x6aba6572ab76f9ce,0x8bdac7ae627711de,0xc17b96c16615b14a,0x231e5f302a7419d6,0x90982688207425c6,0x8204500200a00ca,0x0,0x800020000080040,0x0,0x4002022008000,0x0,0x4100000440000,0x0,0x0,0x2000002002000040,0x0,0x0,0x0,0x8160600a84000c8,0x88088020460000,0x113a2700621800ce,0xff9c0ff4267c1ddc,0x8906837827361740,0xa1080000601148,0x480344000060048,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8204056162540408,0xdb9ebffeaf76ddde,0x89f5ce1a6170155c,0x12090046a0641100,0x419022a60300d40,0x800000200420000,0x2100021000000,0x940002062040000,0x45523630244610d2,0x828a07f3006615d2,0x8180208220000c8,0x100484060600000 },
	{ 0x8020000008200000,0x30684c89002800,0xc1c6cdc2091dced2,0xffffdffbdf7dffff,0xfcffffffcf3ddfff,0xffbefff3db3dffff,0xffbfffffdf7dffff,0xdfffffbfdf3dffff,0xfbffdfdfcf7dffff,0xfffff6fb9f3dffff,0xdfff7feb8f7ddfff,0xffbfffffdf7dfffe,0xefdfffffdf3cfeff,0xefdedfdfcf3dfffd,0xffffdfffdf3dffff,0xffedffff9f3dfeff,0xfefebfff5f7dfffd,0xc977f7af1f3cfef1,0x7bfeffdf9f3deeff,0xffffffffdf7dffff,0xf5aeff7fdf7cffdb,0x6bd665781e5cc6f4,0xcbd7ff3f8f3ddff2,0xfff73ff71b3dfdfd,0xbfaf4fb39a7cccfc,0xfdfdfe799b1ceffc,0x7dbe3dfa1b149efa,0xfbd6d13edf5ccffc,0x5ad71fdf4e3cf6d4,0xffbfcd878b1c4eec,0xddffffeb8b7deff7,0x9fdffdf2df7cdf99,0xffa67f7fcb5dfdff,0xebc7fa331f3cdefc,0xfe92dddfcf5ddfd3,0xdcb7deee9f3cfef4,0xedeeffca9f3dffdb,0xd7a667f10b7cf7d5,0xc9f7fb4e9b1dffbf,0xef1f9edf9f3cdebc,0x94f9dfdfcb19fedb,0xcff4f762855cdcfb,0xfba62d731a1cb6da,0xe9d3f6a68b3cdffa,0xaad6fbca1e3cf7db,0x9fbedffadf7deedd,0xddfe7fae1a29e7fe,0x520f643162dd596,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6fffffdfcb3ccfff,0xf7ffffefdf3dfedd,0xffffffffdf3dffff,0xfff7bfffdf7dffff,0xffffffffdf7dffff,0xafffffffdf7cffff,0xff6adffb1a7cefdc,0xff7ffffbdb5dcfed,0x6dfe7df91f0d6f5c,0x89227aa98901cc59,0xafffbfdfdf5cffff,0xffd8eff9497cfdd7,0xf765f7799b3ccffe,0xdfee6dc28b5deffe,0xdbf2fb3f8f3defff,0xcfdf721fcb3defba,0xffbfdfffcb7dfefe,0xf0fff6e88f2dce78,0xefbe6de3db3dcfde,0xf73c7dff9a38ffdf,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa4d5f6e8cb3c9efc,0xdbf7bf0e8a2cbede,0xde8f2dfa9f2dffdf,0xcff7f777db3cdedf,0xfbffdedf8f7dddfe,0x97a24f281a0ca5a2,0xd7bfeff69f3ddfdc,0xfd7efdfe5f3deff9,0x8b55bb2a0a3ca6bd,0x2fc72b588f2cdcf0,0xba9e4d7a4b3de3aa,0xf153f67d0f3dffdd,0x7ffeff6a9a3dded8,0xfb90e7760c1cefce,0xd8c2fb699b3cecb8,0xafbf8ffb8a7c9ffc,0xfaffffff9f3ddfdd,0x8bb60e700e4cf799,0xf9b0ff3b9f1cef9e,0x8a42fa7a8b2886db,0x2c5fdfc6535ccedb,0x52f6b6d6cf3cdef9,0x89fcba5a831c9efe,0xea4e0e36921c87d9,0xdfc8f72a8f2dcefa,0xc9eeb08f0e3ce5d4,0x636042a0308cc19,0x5028a4c0060ccb11,0xfe182f3f1f08df38,0x40c2220002080010,0x4a7e5b5e1f1dcfdc,0x3ca00d08025dc182,0xd1acb7de9f5dcfdc,0x98962efe0b4c7c10,0x810004710a000708,0xccd72f571b1cef92,0x841907b3161cce88,0x53a0faf05b1c8eb0,0x54aef56bcb3dcde1,0xbb907f6c0b2ddeff,0xc9df18500b0dcd79,0xe338ce534e1de6fe,0x805174800b200f10,0x40a4258000109291,0xf3a46e728a2c8af4,0xca461007882d45c0,0xa20885c10019cef8,0x56651e8ca146e98,0xb14fb668835dca7a,0x8b925a7cc93c0ea4,0x208000000c00,0x4000108212080020,0x2420081004400,0xa37238a28ccd2,0xb7124d0c8b1c4f34,0xe9e6e263830cdc1e,0x235c5b47020dcdde,0x14990671080ca6d0,0xc0012410080c2084,0x8212150101044890,0xc081000180080410,0x2d0201010c4c6020,0x14d007d10e20ce1a,0x24e606000e04c41c,0x0,0x20400040004,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x401000000008010,0x0,0x0,0x0,0x101000008002,0x0,0x0,0x0,0x2002000001010,0x0,0x0,0x0,0x8045d4106084c00,0x80c20000008442,0x2c1e35c505144ce4,0x1448000080440,0x0,0xa2a42d10081c8274,0x3310c3020b008949,0x99fb961b0f3ddbdb,0xfededf7b1f3cfff2,0xc25776ed1f18fad8,0xf376fc718b5dcfda,0x6600440292000c00,0x46c5d330083ca4c1,0x6934f40071dc796,0x98501c80060c4530,0x90cd0288908a810,0x9b0049625a10ec52,0x580a80002180490,0x2105c0ce0a044dc0,0x122010081000c00,0x112824440b28e110,0x223224020108c054,0xd1e69a03da308e58,0xff5edf994f3ddbb6,0x87befef60e18fff8,0xf5e73fe1495caef0,0xdafe6e41c91dcedc,0xc987e22fda1d5f98,0x4052135103180b90,0x138ccd910f44dc58,0x848aa2400c20dcd1,0x6b0e2d708810cec3,0x9ba4a72c8865ecd8,0xe2df32db5b1df71c,0x30d5a700b0d6f90,0x745f9676c93c8ed6,0x9a08251289215e18,0x0,0x4002000008000080,0x81180008006800,0x0,0x30000000100200,0x4000008000080080,0x10000008002,0x0,0x0,0x140060002080180,0x0,0x0,0x0,0x4c8080040880,0x0,0x4000218000180030,0xc9819ba28b30ce18,0xcdfb7e7b8f28deda,0x1204404002001458,0x0,0x0,0x80000200004800,0x0,0x0,0x44000000c00,0x0,0x0,0x0,0x0,0x0,0xc18c7d224439c4e1,0xfb32ef7a1f6dff96,0x80ece392cb3c0d50,0x804004da81088cd0,0x480230801184080,0x64046881094410,0x0,0xc88020200105cd0,0x203200d6010cc488,0x2049c00e086014,0x920082400a208619,0x9124644012008d42 },
	{ 0x41c0b109400001ab,0x6d14a0024880160,0x4400000000801,0xde94f07d640a01e2,0xfdf7fffffdba9fff,0xfffefdf3fdbadfff,0xfffffffffdbadfff,0xdfdfffbffdbadfff,0xefffdfdffdbadfff,0xfcfffffbddba5fef,0xff6eef73fdbacfff,0xffbe7ff7edba5dfb,0xefffffbffdbadff7,0xefffffdbe5bacf7f,0xfdbf7feb5dbacfeb,0xdbffff7f7dba9fe3,0xffbfbfef7dbacffb,0xefd7ff0ffdba07ea,0x7fffff5ffdba5ff3,0xfffbfffffdbadfff,0xfffff77bddbadff3,0xffb6ef7ffdbacfeb,0xe9f7d36b7dba9ffb,0xeff75fd675ba1f7f,0xbfdbe7d2cdba0562,0xf7bdf6fab99acfef,0x5ffe7ff28dbadfeb,0xd334af275d38cff3,0xebf39ffbfcba5fef,0x47ba4ff345aa9da3,0xf17f7f6ce4b807ea,0xfbbafd72cdb8cfbb,0xfbb5fff6edba4fc7,0xcfdefb9bdd9a8fef,0x6ffddfdbc5baddfb,0x3ebedee80d2ad5f7,0x5366fde1cdba57eb,0xdfbeef7f7d3adfe7,0xd7dfffca5dbac5e5,0x6f1f5ddb4dba1f7b,0x5fb717e69dba1f77,0xddefff52bdbacffb,0xb2b63f7255ba4dea,0xc6d1f9bf55badf63,0xefdade5b8dba55fb,0x95fafffeddba5ffa,0xd91dffeffcaa0fe3,0x8000200000800100,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x184206978d8a054e,0x6a5e5f580dba0fff,0xdcff7ffeb5ba17f3,0xffffff765dbacfff,0xfebeafff7dbacfef,0xeffffffffdbacfff,0xbffffedbfdba4fe3,0xba320ff86588136b,0xd3ee9ff97dba97e3,0x871678609510850b,0xabfffffffdba9ffb,0xfffffffffdbadfff,0xeffffffbfdbad7eb,0xd4fcff7c6c9a5f67,0xebbefd63ddba1f63,0xf3e5ff3edd9a8de3,0xff4bfebefd120dbf,0xffdf5f524dba4de6,0xf1d7f7ee7d3a4fe7,0xdffaef73ddba4fe7,0xd1842d50502a4fe7,0x0,0x0,0x0,0x0,0x0,0x0,0x101000020800,0xf637ffeffdbadfff,0xdb5effb7f49acde7,0xdebfae7765bacbeb,0xc7fc7537412a87fb,0xef7e3dc8cd3adfe3,0x73a09ff984aa0da3,0xf3ffbfff3d9a8fef,0x4a6e716261ba0fa3,0xc861aa82041a0c2a,0x48d51e5669ba87af,0xba300fba45aa04ee,0x5adc965b4ca8133a,0x57ae3f61813a0763,0xe1d5e746d0aa8f67,0x6ddfbe7eddba4f73,0x7fbaddfba5ba874b,0xedf7e6e3ddbadf23,0xfff7e773fdba9dfb,0xf9bee71f5dba0fe2,0xb56b7370cba0960,0xad56dc5ac53a1f2a,0x1490502411a00701,0x804502200084321,0xde3827342c100bbb,0xd6e77f3a9dba1fee,0x663d97c04daa0972,0x92917faa05b819a1,0xd4e4a6175daa0faf,0xb7166f7cddaa17e2,0x895cd2044cb81d2a,0xa771bc854ba5de1,0x973f5ff7d49a5fab,0x612cd53acc3a5562,0xfabaef6edd9a5d76,0xeaa5317b5d985d2a,0x41d70c5a513a4db2,0xf014e18d1384d20,0x1190c5798c084d22,0xcbac5c73c11205e3,0x800400000000012,0xeeeeb7ff7dbadffb,0x2a0c4f500d3805c1,0x90545421a4aa0542,0x85e42d616c8201ee,0xda60a32e70ba0726,0x46863e3561b8872b,0x89be87524cba0722,0x157b26e4149a0f43,0x91f9a771059a0deb,0x709c0002c5880cb2,0xddd2468c1c18c741,0x8120343d0100420,0x440080104020102,0x45ece71905b80fa3,0xc2a62e57350a8de3,0x2148592114081501,0xcb7f780194981723,0x92922a68049a80e5,0x168762714280320,0x4122f4610d289181,0xc280e003113a4504,0xae577f50d0920d5a,0xca1eeefadd9a4d66,0xf44fa2e97cb81702,0x0,0x0,0x0,0x0,0x0,0x10010001000100,0x0,0x100800001200000,0x0,0x0,0x0,0x0,0x800c0050000420,0x0,0x0,0x0,0x8000020004800000,0x0,0x0,0x0,0x0,0x3020040020000901,0x4182c2620cb285a3,0x60f3b2a24924d22,0x208020100184400,0x5144033945301702,0x8611c024aa4146,0x1d0c10c05a01b41,0xc2d236e3ddba4762,0x22904de48caa072b,0x75ae3eac158887bb,0xd3ab656135ba47aa,0xc890fd089d3805a2,0x49423214a53845e0,0xa4988cca602a456a,0x2000002004000100,0x2014000000213,0x3214060000b20442,0x1018720440081a1,0x2ddc65024aa556a,0x28064140120502,0x900826a40300623,0xc09229d41c8a0603,0x48cbf32ac5aa8783,0x4bdfd7d2fc8acf73,0xb6376ffb1dbacf43,0xddffbd7ffdaa8feb,0xfff6adf75dbacfeb,0xc9f3fb7eddbadfff,0xf3571fd6ccaa0f6b,0x479a659b343a0d42,0x683602a8052a1b04,0xcb8a3723059a016e,0x6204e20604820303,0xee7a515f05ba57a3,0xa58dded3ad0a1de2,0xf5be76fa7d3a97eb,0x8801ad1215ba4c21,0x0,0x0,0x10020000080002,0x24000020800,0x0,0x0,0x0,0x0,0x404000004000100,0x8010000000180420,0x0,0x0,0xc00000080100,0x0,0x0,0x0,0x1005a8000000301,0xefdf3bdffdbadffb,0xbf1fcdf2a5bacfeb,0x0,0x0,0x0,0x80000100000002,0x0,0x0,0x0,0x0,0x2004000080001,0x0,0x0,0x678e6f314c3a5723,0xccb27b5ee5aa1f3a,0xcde7ca22449a0f33,0x2654d4c82c120f03,0x122a92008a004a,0x208360040a014a,0x8220c680c801140,0xc49a24030c30570a,0xb4eee1cd4820d22,0x811eebf3edaa1da0,0xf0a93428c400157a,0x8204406014900400 },
	{ 0x902c161201242e4,0x48869a3124b26264,0xc886052f241063ec,0x6020000020,0xf5feffffbcf7ffee,0xfffffff3bcffffee,0xffd7ffffbcbbffe6,0xdffffebfbcffffee,0xbf7efddeb4bfffee,0xfefff6ef9cbbeeec,0xdffaff733cbbeeee,0xdfb66f5fbc9feeee,0xefe7febebcdeffee,0x6fff7f5bb4bfffee,0xfeff7ffab4db7fee,0xfbedf7fbbcdefae6,0xdabfbf7fbcbbfee6,0xefdffe3e3cd8ffe6,0xb95fbec3bc9be6ea,0xffbbdffbbcffffee,0xffffbfffbcffffee,0xd39c6f7634daf5ea,0xebf7fb3f349b7de6,0x7e7f97db3c92fee2,0xb7fcc9ff34032bee,0xdefefffeb4bb6fee,0xc7b2ed7a14f665e6,0xfb9bc73fb093ffee,0xf7d3feeeb497f6ee,0xe234c7ba3c95e862,0xd4f6fe583496f7e6,0xfbffff72bcbbfeee,0xffb2f77eb4b3feee,0xcedf7e7f3c9ffeee,0x6bdcd9dfb49afeee,0x54bf76dfbc32f4ee,0x83befd53bcb2f6ee,0x8a9c7f652492feec,0xfdcfdbddb4f6edee,0xefffdedbb4d7feee,0x7cbff7ffbcbffeee,0xd5aba7d924ff7aee,0xdfbf7fefb4b376ee,0xedf3ff9e24bf7ae4,0x5ffedfdfbc9ddce2,0xffffffffbcbeffee,0x8054586b14b6c2a4,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x462e13b4107c68,0xcb9adc9fbcbaf6ec,0xbcfd5ff9bcd77eee,0xdfe7ef7bbc9ffeee,0xffffbfffbcf7ffee,0xfff7ff3f9c9fffee,0xef7effdf3c337eee,0xfffbffffbcbfffee,0xf173ff7dbcbf7eee,0xdfb6777e3c3256e6,0xcbd7ffeb3cb37eee,0xefff5ffebcbbffee,0x9fbddffb1cfafeee,0xfdd7a7fd1cbecea6,0xd2f42df03096fee2,0xfa95ff5f34b6eee4,0xc7fbbeb134b2eee6,0xbfbfcff334b3fee4,0xdd77776b14bb77e4,0xdffefff3bcfbfeee,0xffac3ffe34ffffc6,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xb4f5dee3b4d3e6c6,0xdbf6f7ce349bfeec,0xf2be3efebcbbffec,0xcfe7f7a3b43376e6,0x29efd9593c9eeae6,0xb6fe4d7b34baf6e4,0xfb7fbf7eb4daebee,0x7a4e7cf3349be3e6,0x8a7cff2434de6fc6,0x95e0e058094e2e0,0xbabec5c21493e6e2,0x97e7f26994fe5eae,0x526c2df08092e3ec,0xd3dcff263cbafeee,0xef4bbfdb349afaee,0xa6ae4fea3497ebe6,0x1cfe77ea34b3ebc4,0xcaaddf1b4faffee,0xd480ef7ebc97faac,0xcf5bff2214b3fee6,0xf95ec7d23c93f1e6,0xdc76163684bc62e4,0xdaa1c71030d46462,0xc8bd2f3b941377ee,0xc182c6082cd84044,0x8b8a5f1114917d64,0x511607630906860,0xcb7ca74b10daeee0,0xca320bea2cfee164,0xccd810600936024,0x890d0a48943024c4,0x813047509c1064a6,0x8d60272934b0faa4,0x1618296004327284,0x8bceb98a18126cee,0x5ddb379b84333aee,0x4a8988a1000c86a,0x91a8b6b824da6944,0x240134000020,0x88f0c00408026cc0,0xfbf7bb4cbcbbf866,0x1e0ea7d3303276ec,0x94736fe634d6786a,0xa7792d55b43276ee,0xffadefe734337bee,0x3c4a6123010dca0,0xab6e8b4ab4b2e8e4,0x70b6d66834987562,0xd16f656a3c037fe6,0xaaac7c52b4fb6ea4,0x406a12a34b844a0,0xe9e2d8532410e066,0xb132032134c23024,0x59c5f50200d074e0,0xd61e3e7294977aec,0x4cc6ee0a94fafca4,0x4d3ce4b34127eec,0x86935c3a94137862,0x90001008000002c2,0x51c6f4c214b2fae6,0x89d5fa033093dbc4,0x140000000004040,0x20244ab12c80c2e4,0xfdbcb7e93cdef064,0x0,0x0,0x0,0x4010000080100000,0x0,0x0,0x600004001000,0x0,0x0,0x0,0x0,0x8004000000101000,0x0,0x0,0x0,0x851004300020,0x0,0x0,0x0,0x0,0x0,0x20004004200000,0x5000000000000480,0x88c98000088c820,0x82400224000044,0x4000c000200004c0,0x454e31a0340224c6,0x8080690030004060,0xe86bb2b6b8d1ecc2,0xa79dcf9230a2c8ee,0x948af7ea1cdb44e2,0x1eb225412092c6e4,0xefcffb6ab4bffce6,0x9d56f243b4bb7aec,0xac288ca0032c1c4,0x1018026684005828,0x60220380240016a0,0x328c040c0c204164,0xc956e40630b6c8ee,0x6c1f040092a0e0,0x8019546010106062,0xc948b660b01270e2,0x8256084122042,0x4e08d602308acec0,0x25ef1d014cd66e8,0x9fbbcfebbcb76aee,0xf23b26641cba67e4,0xfe5e6f7d349bfeee,0xcbbcf7efb4b3ffe4,0x6dcf1af5b4b3ebec,0x1c8ccd0a301046e2,0xd4a2f2f8049b5aac,0xc29abd623492ec66,0x91026b262092e6e0,0xadda169394927864,0x3709075210ba68e4,0x51bcf65ab0967ee2,0x9022a511943134e8,0x0,0x0,0x0,0x0,0x0,0x4020000000280,0x40200080100000,0x0,0x346218a800137484,0x10000244200000c4,0x0,0x0,0x8100000c00000,0x0,0x0,0x0,0x2200a80000028,0xfcc23efb3cb36fee,0x98a2474a10946fc2,0x6120ee404126aac,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8908e44310136aec,0xaa943f611c81ec62,0x8355e3379491eaee,0x89445b12301378a4,0xa42004c400102aa4,0x80282230802200a4,0xd6863f703432feee,0x4653221224b87060,0x16d4840348816ea,0x109acb72bc88e8e4,0xd1a4f6f3341334e6,0x80d600400d020 },
	{ 0x6884f20210385442,0x17711e4ff0341a70,0xe36acf7a5034d6db,0x4012023000202140,0xfffff7fff2fdffff,0xffbfff73fafcfeff,0xffbffffff2fcffff,0xfffefebff0fddffe,0xefffffdfd8f5deee,0xdffef6f2f2fc577a,0xfbffff7afafdfffb,0xffb77ffdf2b4fffb,0xffcbdf3ebaf4dffb,0xef7fffdffafdd7eb,0xb4dffeebf29474fd,0xf9ffff2bd0d4dfdb,0xfeb77f7ff03cffff,0xe7fef71e72d497eb,0xbbffdfcff23cf7fe,0xb5bbdffbf8f4f7eb,0xfffefffefaf4f7fe,0x7ffefd76f834bfeb,0x6b82ea1ad014bfea,0x6fffbffddabdffe8,0x1ebccffad0bc7cea,0xf5f7f6fef2f4d7fe,0x5966fdd2109cfedb,0xfbd5eb6670bcdeeb,0xecebbff1f2ddd7ea,0x8ecdcff9f095b5e8,0xf457dff69abddefe,0xfbf6ff63fad8f6ee,0xdbbeff7edabc56fe,0xdfcbbb3ffadcdff9,0xafffdfdbf0b1d7fe,0xf4f7ffefd875b7f9,0xbfffff7bf8fcfefa,0xdbb2ff7f7a54b6ff,0xdf4effbfdad1fff1,0xefffffdbb0fcf77b,0x7dfe5ff1f035dffa,0xfb763f63f2f49deb,0xfe763f7f3afcf6f8,0xc7f9f2bfda21d5e8,0x3f5f9efbf21ddff7,0xbffbfffffafcffff,0xfcd5fefbd07cffff,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8036e361020f648,0xabdfddd7b034d6e9,0xdcbedffcd27cfd6e,0xdb7cafb032d5dee6,0xbab8af79d25cdff2,0xefcffb3ff29cdffb,0xcf3fda7ff2f4feea,0xd7bbcffb9abdd6fa,0xf3d6b6f37afcf77b,0xd3f6fde232dddefa,0xebdfe72e5a7c5eed,0xafff7fd39875fffe,0xcebbeffbd0fc5fee,0xddf6b7fce8dc5efb,0x281c3d32300510aa,0xc3a3d33fb0b4def2,0xa7ffb7d9d0f4f7ea,0xffffdffffafdffff,0xf6fffefff8fcbeff,0xff3f2d329a74d6d2,0xf9044f3c301cd7e2,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe6d5566b3034aff5,0xfffeefeff0b4ffef,0xfeb6ff7df03cfffb,0xdbe9d7ba105dfffc,0xbfff3fdfd2bc7efa,0xc5b85e61b210dce0,0xf1ff96fab8fcdf7a,0x71dc77ee123cd7cd,0x985cb0a701844f4,0x8ed21356d0d536ea,0x33bfef5bd09c97ca,0xfdf7fff5d09d56ef,0x4fac7ef170d4de6a,0xfbd8e526d2d8db5a,0xece7ba0a122055c2,0x793eef73725cd7ea,0xf4ff1e722054deef,0x8ab60f71d2f4df6b,0xca94fb5f92dc5fc8,0x2147fbea70fd77ee,0x6b39c25cf000c6e0,0x949f7ee6da4dd6e8,0x24b617e3725d1fe8,0x8ad2332c82fc54cd,0xc1b3ffa8005c5652,0xc2de189752048663,0xb0374fa2da1c17ee,0x865a4004040142c,0x8fb2257c505494ce,0xc1060b08104074c0,0xaef6cec272d5f36a,0xb5aa1dd010745c92,0x934d06608041d5e0,0xcb86f47e1805f0ea,0x8153e634c0543ce8,0x66131e0050005640,0x8a1243819004f442,0xf334f4ed42c1b7ea,0x651a1d105014fcfc,0xcac4cb26d03c5ebd,0x21e21e1950a4f0e0,0xa73ccc4750f456fa,0xe8d416a1501c5492,0x2887d233090bce8,0xffb77d81507856e0,0xa1c4800078205420,0xe14641521035cc69,0x7a76f4e0f0b07cf8,0xc1670d60502564e9,0xc104629350141e82,0xcff7ffbbb0dddfff,0xaf56fdcad0c1d4aa,0xdd9077fb8254f5da,0xf1e4f7faf0345fea,0xb09635dd52f4d272,0xfc6201580b85642,0x966325c70141c60,0x1fb86fdb5834b6ee,0x98f5377bd004d6ea,0xfe9621e09a30bee8,0x89418072d0100640,0x8191224101406d4,0xa0b080415001d2e0,0x40c002000104010,0x0,0x0,0x0,0x100001010000080,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x600c2d20500486c0,0xc300404250240460,0x8626200010901482,0x9fb85f3232a05568,0x1404360800785c41,0x252090040240,0x0,0x1846224b0211cdec,0x2a100c8170009460,0x2099302208a45642,0x1084c008000100a,0xcb9ed77ffa81d7ee,0x4d56f228103c76c0,0x11d002d0001401,0x64202561400056ec,0x0,0x3946234320484e1,0x4fc297ba30b95ee0,0x1212954a20501471,0x46e104f22141062,0x80f80960b09411ea,0x4012020000002140,0xa0dcd010e0001660,0x6937ebc25098f368,0xf2bea7e1f04056e8,0xb1dd966b5a7455e8,0xfff67fe5927cd6da,0xebcdfbbfd0f4f7eb,0xffde7edff2b4feee,0x9274da110bcd759,0x144db67d327452ea,0x69361d32c03d56eb,0x528401260010fa58,0xba8a324190307c40,0x9b0b8bd3f2b49968,0x10e077ecdad4da40,0x9b50270050305e64,0x0,0x0,0x0,0x0,0x0,0x0,0x4852580040150868,0x0,0x484100202011062,0x0,0x0,0x0,0x0,0x0,0x0,0x204000c0000,0x4000032a50001000,0x0,0x6a10440d044102a,0x410626840401668,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8001000000040,0x0,0xd2a20741b084d4ea,0x820c62000014de48,0x86cccb02501c13ea,0x80cc4f4990345dd0,0x1a404100101410,0x98c125429024124a,0x2bc2c70108c1488,0xa4f4a028d010d008,0xb04e8e19701c7478,0x9efe6ff27854fffa,0x7145befb1250d6c5,0xc4e22d51d2385668 },
	{ 0x8b95f769d9cda25a,0xafdf9fd3d1ef22ca,0x239ccf54f9ed3348,0x81ac786ed1e92904,0xfffffffff9ffbbfe,0xffffff73f9febffe,0xffcaef7ff9ffbf5e,0xefdffebff9fdabde,0xaffecfcbd9ddaf76,0xdcf7fefff9eca77e,0xffbeff79d9e7bfee,0xfffffffff9efbffa,0xfff3ffbef9efbbfe,0xafffffdbf9fdbb7e,0xfeffffffe9ef2f5c,0xf36ef739f9ddbfd2,0xfeb6fffef1fdb3fa,0xcdc4f77ad9fd23de,0xebfedadbe8f723da,0x9fbffffff9ffa7fe,0xf57df766d9e5bb7e,0xfffefffff9efbffe,0xebfffb7fe9ef2ffe,0xbeef7fd1f1e7a7fe,0xefffc7ffd9f7aef6,0xd7fff65af9ed3ec6,0x6ffc2f7bd9fcaf4e,0xfb74f737f1ed326a,0xce6fb6cfd9ed3b56,0xef3edf7ff9fdabe2,0x7df3f6fbd1ffaa7a,0xfbfeff71f9fd3ffe,0xffbcff7ed9fdbbee,0xcffb3f3bf9edbbea,0xef7fdfdff9efbffe,0xf6f5fefef9ddbbf8,0x9f47bf7ef9eda77a,0xffbe2f7fd9e5bb7e,0xfffff6b6d9cd3efe,0xafffde59c9efb7fe,0xf2bf7defd9e1befe,0xf3acae72d97daf4e,0xdebe3b68f1dd3b7a,0x29d7e288b0ec2f64,0xeffeff7df0fdabdc,0xfffffffff9ffbffe,0xfdfdfefff9ffbffe,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xdecffebef9fdbbf6,0xbbffdf5ff9ffa7fa,0xba934748d14da368,0xf94c760bd1a6294a,0xfefa2f5ff1f7225a,0xffdfffeef9dda7fa,0xefffff5ff9ffbbfe,0xbfbbfffff9fdaffe,0xd7bfff7bd9fdabfa,0xfedefffbd9fc2bee,0xf184f60af0c5a3d2,0xeff7defff9edbf5e,0xefbdfff3f9edaedc,0xf467f2ecf96d2646,0x3feefff3f1f8afe0,0xfbfce72ef9ef236a,0xffff3ffff9edaffe,0xffbffffff9ffbffe,0xfdfffffff9edbfd6,0xcb7af7d0d9fcabc6,0xd2160304e14da244,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf4c5546bf9efbffe,0xfffffffff9ffaffe,0xfebf6ff7f9ec3bdc,0xefc6d3fef1ec23da,0x4a6f3fdfd8d7ab5a,0xeb2fffed1fdb3fa,0xd4bc746b99a5ab7e,0xdf5aef6ef1cdbad2,0xabf7fb3bd1edbbee,0x6dd21fd9f9ce2edc,0x5ebccf5ad16da3ee,0x7534a2e619412752,0xbfe7d70d8edae52,0xd130a26fd1ef22c8,0xa439f8ed1ec2e1e,0xdfbcdf3ad1fda368,0xd0d7dce0f141a246,0x86967d70c14fba50,0x4a826cccc9ec2376,0x8de6e227f1ec2a94,0x2cdeefcee9412f4c,0x90b77ee6f9ed2f52,0x59691252c9382354,0xbb8c2b74c8c53446,0x8091c30ab84c8640,0xeb52974fd9e9a7f2,0x24f5550ae84f1a40,0x19c0c65ae1a52a44,0xd622484eb0d40d40,0xc163f938e16d2a48,0x29611ac8408791e0,0xff934fa6d18f0342,0x24904280808250,0x7c446560d1c0935a,0x4024220640202284,0xcecf3359d9cda742,0x811ce7fbc1e70a62,0x9182d35ae9e5234a,0x41867570f1ed221a,0x30884b02d1aca04a,0xed7bbfd6f9f72b4e,0x8dbecf73f9feab5e,0x3444716298d12062,0x5327d72d0d72948,0xd026e10bc9ec264e,0xcb4c3613d15f3a68,0x2306c94ad149220a,0xb4b0c644202c2142,0x990ef532c16d224e,0xdf962b78d8f4b34a,0x88f291ded1c03a98,0x7ba04750d12b2bd0,0x34a20cf791c72040,0xd13074da90e72342,0x1a88094c91862046,0x81f7a3a6c96d2a88,0x2b401ad8514c135a,0x4f86d6741cda24c,0xd87cde78f1cc825a,0x6ff86d73c86aa7c6,0xe9c7f1baf918a34e,0x6de13fd8d0cdab76,0x3aa0ce8be0542b4a,0x8000200400000270,0x0,0x0,0x0,0x0,0x90434080062148,0x1000001040000080,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2800202000,0x1080000000010040,0x0,0x0,0x0,0x0,0x4000200000140,0x71de6079c97da576,0x100c70a99980200,0x204118d011480318,0x2e08e1d0c0cc0600,0xc13270524944209a,0x128409e8c1c0a600,0x8000001040001000,0x548000080008840,0x0,0x801401000c0042,0x922006408801a044,0xd0044a00018c2266,0xcc1101188602352,0x9012050098c00218,0x30b21600c060a140,0xa0815091020140,0x0,0x8612308ae064a600,0x400400000400100,0x92025ce3c056a142,0x3104a26a484c2340,0x200000280100000,0x4e80862080450208,0x20d19653c86db148,0x94926eec18ed2142,0x54a68d40d844a044,0x40b62744d1499240,0xebd8fb77f9e7bbec,0xaeff3edfd1e5afde,0xf9bdcffbd9ef32da,0x848eb6e8d8e1aa44,0x8b022d60c86da20a,0x88540420c0830050,0x885032d3e1242228,0x2dee4d69c8e127c2,0x44fe137ac10caa6a,0x9b382d42d1c4269e,0x0,0x0,0x0,0x0,0x0,0x0,0x1080041000000,0x4008100148401000,0x8010004080002042,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40a1d01240008278,0x5c0800880480340,0x1d83cf6ad9c6235a,0x81a6104851641208,0x1008016088002214,0x8000211080010210,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc144b630d8f4aa40,0x8a99736cd11622da,0x2acfc10ee149331a,0xcb7c9d03d9ecb314,0x0,0x1000800000002002,0x82b92e50d94c2376,0xa9c6d325c869a6e2,0x804494080c80600,0x95b7eff2c96fb3d0,0xd1adb774c1bd2ab8,0x8aa475f4d945a74e },
	{ 0x5858f1a018042572,0x6150173ed202ac22,0xa080110000520,0x9fff4f6b9feefd62,0xfffffffbdffeff72,0xffdefd77db6ce7b2,0xffbff7fedbaeff66,0xffdfffbfdbfafff4,0xfbdfffdf9afefff2,0xd5f7ffecdf6cfd72,0xf7efffebdabcf776,0xffbe6f7ddbfcf576,0xeff7fbfbdbfcfdc6,0x6ffefddb9fecedb2,0xb55edbf79a8ee7d4,0xbad876681aecb742,0xfeb43d7fdd7cfde2,0xcdfeffa7df6cf5b4,0xe9dffed3df6cfd62,0x973ffff7ddfcfdb2,0xf1fdf67fdfbcfd60,0xff767fefdff8ff24,0xebf6effedfeeffe2,0xfef69ef358fc7730,0xbabffffbdff8ffa2,0xfff7ff7bdffcff56,0xfffe7f72dfbcf7f2,0xfbffeb3fddeefff6,0xfafbb78b5fecfd36,0x6fbdc7bf9decf562,0xfff4f6fcda4cfd32,0xfffffff3dffcfff6,0xffffffffdffcfff6,0xcfffbfbbdffcfdf2,0xafdfdfdfdffcfd72,0xd4fbfee1986cbf22,0xfffefff1dffee7f2,0xfb907ff99feefda0,0xd979ffbcdbcaf5f2,0x8fffdfdfdf7effe4,0x3cffddebdb7efd00,0xfd7f75bfddf4f7d0,0xffbebfff1bbafff6,0xace3efabd924bfe0,0xabeeffffdf2efff2,0xfebefff7dffafff0,0xfdfdffffdffefff6,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4fcfbe349afefdf2,0x8fde575f9feeefb0,0x3ef77ce493b6ad74,0xf9fefffbdabcff46,0xfab77fb7dfbeffb6,0xffffff3edffefd76,0xefffffdfdffefdf6,0xffbfffffdffefff6,0xd51d777add7cb520,0xff9efffb9fbefff4,0xe87fd30edfe8ff20,0xfffebfffdffeff70,0x7fbfcffb9bfeff32,0xc576d77b9d26fc32,0xcf667de31a3ac530,0xeb86e122dbfcef30,0xbfd7fffbdfeeffa0,0xffffffffdffefff6,0xfdd2f7ee9df8ff72,0xa7de6db09fac6f36,0xfebebf7fcfe8fd66,0x41000010000020,0x0,0x0,0x0,0x0,0x0,0x1100000200400,0x1c145090d500a820,0xffffffffdffefff6,0xffbfafffdfbefff6,0xffffff3f9ffeff32,0x6fcffeefdfbcfd72,0x979b7ff7da24f5f0,0xa177bfca1de87440,0xfb9ea760dfbcfd46,0xfbe7ffefdffcff74,0xefdfffffdffaffb6,0xd69fc7fa997cf560,0xda7ad67d9a7cf540,0x29fe3d73dff8ed66,0xdbf4df16dbbef712,0xabe39bdecd68fda2,0xfb3edf3b58fea720,0x37e277684da8ffc2,0x9fbefff2dfecfde2,0xff9e777c5f3eef46,0xcda3fe8bda8cf930,0xabbbcd5a1b7c9f30,0x16bdd768c728ed00,0x8acf27729f40ff20,0xcaa0252e0d40fd00,0xd6c6cb1fddc8bd22,0xa798de909768bd20,0xb013002c95080520,0xa16b777a9d087520,0x5600057c15248570,0x8082270e1068e710,0xb560d4450aa6520,0x92004968008a520,0x915c030a900a8930,0xf38c7d319f2a6530,0x8b3efaa24d24f704,0x84659118228e580,0x99c1fb988ece120,0xd54fa6bdac86504,0xf3e2de2daa8f514,0x89c6e11fd8ac2574,0x58413be090b8a580,0x6b97cb02523ced10,0x40406001000a030,0x2f6aff639872fdf0,0x510651e1976c6400,0x88df3f3b5f28ff60,0xe03c5f409a2cc544,0x1170764858306412,0x814377ccd82ced20,0x8bc8295418e8cd40,0xcc77f3981f08cd62,0xe61a5713d0b8bd00,0x89d317699202bda4,0x97ea27a25a08ad44,0x88514210020a940,0x8a54c2145028ad00,0xa153ba590a68fd00,0x86300462910a2120,0x909406829ac83530,0x16907d514df86520,0xa9f5b1ee1228bd30,0x86880241102095a0,0x938a8d090820fd20,0x400002000000600,0x23c2000083100,0x0,0x80800008400000,0x10008084000,0x8010000000080100,0x40902c5019084402,0x4204e2220038a980,0x581e1098002600,0x0,0x128064048309800,0x8024054012002500,0x72a0244090000740,0x1449c30858042400,0x2fd319db1a702744,0x4b0440000008502,0x100106008000500,0x8080030440243000,0x181d3380ab84500,0x18ce4002303402,0x9e0085748d08b422,0x70f097020a306520,0x650a4c781d603404,0xa1b6215a1d08e510,0x308a9a4008046120,0xa8a8cc10d78c6a2,0x24000002000,0x240452098400400,0x1000139900085d0,0xcc6100c0d700600,0x1e8c4d81986ca460,0x115416699038fc00,0x8021350015202050,0x320a400402884502,0x96612104b880d02,0x114d1930720e180,0x41a0d46102208d40,0x1b600c60983c6522,0x8a0e050491080d04,0x40200000000400,0x0,0x14226f0011288d10,0xd0300e18d8003402,0x8012004001380000,0xcbd4b48fd16aff64,0x812149810009110,0x1abb7fff986cfd34,0xfc7526f118bcf520,0x5e020c42112aad40,0xe9fec36e9a4ae570,0x6ffb16e95abcfd04,0xb7ddcfdbdbecfdf2,0xf4ec76e2da6c7530,0xa2bc2e701af0a510,0x418083069a087520,0x146b1000a284740,0x214481008ac3540,0xd4537a831f6cff80,0xc81a2d22c8ec6522,0x20000202000400,0x0,0x0,0x0,0x0,0x8080000084000,0x800020000880000,0x0,0x1000226218208410,0x0,0x0,0x0,0x0,0x100000248000000,0x0,0xf29025544d402e00,0x8046d2120a784100,0xef5bbefbdfbce5e6,0xc1f4fab12222562,0x0,0x80000008800,0x0,0x0,0x0,0x800010000000a000,0x0,0x0,0x0,0x0,0x0,0x100320000000740,0xea3c6f4c9d722522,0xcaf1da108528e500,0x3d895d01026e500,0x801404a292888900,0x240440290020400,0xda26affe87c82720,0xe99e7e2615a8f562,0x17736401d7ef524,0x1e27e7731ab87720,0x90080761092ccd00,0x7ab67ff35af8ff76 },
	{ 0x114e18402262860,0x10064002031c00,0x890c70002818f40,0xffffffffbfff7f73,0xf7fdffffbfffff63,0xf7bffd719ffeff61,0xdfbefbff97db7d61,0xcfffbf6b9fff7f63,0xefdfdfdf9ffe7f63,0xf7b3feed1fdfff62,0x9beffff33fffff73,0xffbeff7fbffe7f63,0xefdfffbf8fdeff63,0xbfecf5a0bd76b60,0x5ffb5feb2fdfdf72,0xd1be777a9f86ed62,0xb2bf3f77b7ceff61,0xefccf73f37ff7f61,0x6f77ffd716bf7f61,0xffbb7ff3bfff5f73,0xf7ecbfeabbc67f63,0xfeb6edfea7ae7f72,0xabdffffe37feff63,0xea4f3fd59f82df61,0xb7dbc7fa0f97ef62,0xd75d3f7b8f9e2d60,0xfffe7df39fde7f61,0xfbffffffbfffff73,0xaaf3fbeeaf8e0f60,0xffffefff3fee7f73,0xe4f617ea37d77f60,0xffffff73bf5eff73,0xffd7ffffbffeff73,0xdfffffbf1ffeff63,0xfbdfdfdfbfde5f72,0x94fd7fff9fae7f73,0xdbfeff7ab5ce5f63,0xdfbf7f7f8fee7f63,0xfedfefffbfdf7f63,0xefffffdfbfffff73,0xffff7fe78fffef60,0x3f6ffffb9ade7f73,0x5e8c2f4e9b8e5d61,0xeff2f3bf3f867f72,0x29fbfad7bf9e6f61,0xb7befff19feeff73,0xffffffff1ffeff73,0xcd4efb629fffed71,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1c000e3f9b964742,0xc292df1893867960,0x943fdfe71b9fcf60,0xf97deffa97cecf62,0xfafebf7fbfff7f73,0xedfffb3ebffe5f71,0xafffffdfbffe7f73,0xffffffffbffeff73,0xf47dbf7b1bde7f63,0x7fbc37d43bc65c60,0xcafbb38f86d86f61,0xe9cf9e539fce4f73,0x7f9fcde39dc67f42,0xd4b2576d3ebe3f60,0xdfae7fe29a8f6d73,0xfbf4e77f1ffeff70,0xefefffff9fffff73,0xafbfdddabdbf5d63,0x7dfffeea8f9e7f63,0xbb9677700b8acf60,0xffbe7f7fbfdeff73,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1c14521a030e5f63,0xffffffffbfffff73,0xfffe7ff7bfae7f73,0xeffffe7f1ffe7f71,0x3fdfffff9fbf5f63,0x93b84ffaa22d6d60,0xf5ecb2f807ce7f73,0xda16ae6601167d71,0xfbfffffebffeff73,0xffcf7fffbfff7f63,0x376c95b332de6d61,0xf1ff366b97eeff63,0x79e626708f922f60,0xcbc6fbbf872b5f60,0x6cc3bfb3169e4b62,0x1b1ecf2b9e972d60,0xbadcd6cc9396ff72,0xa7baff708fdedd73,0x5f9ef77fbf8e5d73,0xaedfda030b033f40,0x21dfddd8ba8e0f71,0xc0d45ff817c67f60,0x493af5aa0b8e6d63,0xe212261c11e25651,0xd1f8fb2a11c25753,0xcb9dffdbab4a7f63,0xa118876317867e61,0xdf50f75c9b232d41,0xfab04f6e921b6d62,0xcddffade939eff43,0xc27fce4b3f97cd62,0x9ab327a61eb24f60,0x91b88d580b0a4753,0xd2366f6e26cefe63,0xc95eb2578346f463,0x6a428a440b020d40,0xbd99cfff1fd6df63,0xf5d4bfe99b972f70,0x7fb67d618b8fff62,0xe914f22f1bd25743,0xed4b1c1923867e62,0xafbc4d2c12e64f71,0xd0b4d6c620cc4962,0x85e46d6006d6d561,0xd796ffffbfcf4d73,0xadd79a0c93d36d40,0x2923d14804025d61,0xa7f0d7e813ce4f63,0xdbefb672ae9a4f61,0xfadc46261e804d70,0x4c4e790a0e124542,0x20d485412120540,0x3ed5257b83126d70,0x10b021500e024102,0x4c0a2d21051a5522,0xeb0053250b8c5571,0xb3d7fada0f7b4d62,0x492c76a209e4f61,0x31299f7122c46e30,0x97300d5403965553,0x8131c3341b924462,0xc4d11c718f424d42,0x88492804124141,0x7038e66035825160,0x1a84340121820e60,0x0,0x0,0x2e0d4004474920,0x1024080004000140,0x8360c6216c0c340,0xc800000200120140,0x20258021ac00000,0x2606000002004441,0x0,0x340988168a0900,0x8300000004060100,0x300250000104140,0x210105c92842820,0x1023000401080260,0x1084a41020000960,0x88242f4004c64460,0xa0d0f00293800640,0x1094866080045962,0xbaa04cae24c60460,0x9000040402802200,0x33a0296000826a43,0xa1a0770a0a1e4d62,0x846095093845120,0x200804800844001,0x9400d0400a060940,0x83280f4200cc4223,0x280c10000800660,0x8e46de4287c44962,0x232249580216ac30,0x4c480040820,0x1082080001001400,0x4400010202122000,0x6000000004001,0x62040a080a040c60,0x10400000840000,0x921ab61611a05541,0x82907a0cb1861a20,0xe0f6cf188a966f00,0xa08084008000d20,0x1004000000000820,0x0,0x7834b7c114834941,0xe5dcd11e31b64f61,0x81c5c801e100140,0x9100471102a28260,0xe07c56a00b980f61,0x9302020c0b425000,0xc397d31f1baa4d73,0x647e1f5e0fd3cf51,0x3f9dfffb9ef6fd72,0xa53277eeb3847f61,0x644a5d2213807f62,0x88848a1c11665160,0xe2411e5c21836f71,0x229a4de1241a0d61,0x243fe660318a5360,0x89086502a2025560,0x0,0x0,0x0,0x0,0x0,0x0,0x800100002008000,0x234481404ca6a60,0x480140000000c20,0x0,0x0,0x0,0x804004000040,0x0,0x0,0x763c3c5d871a4f63,0x8800201906810d40,0x0,0x0,0x0,0x0,0xa0004000040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8b3afdc0070e6f61,0xf71c7fed05824763,0x8bc7e31a1f965571,0xf3d694420392ff63,0x8c9b132131ce0560,0x10e876000060860,0x9ac096401e846540,0xf50c66002877c00,0x25fefa1387ce4f62,0x8888c3b21e027140,0xb04410a205426900,0x4294a77c03964563 },
	{ 0x8d0c12cd9b24d40,0x80908012002400,0xa6becfe0d79e7d43,0xffbffffff7be7ffb,0xfdfffffbf7be7ffb,0xffffff77f7be7fe9,0xdfb7ff7ff7be77f9,0xcffffbaef7be7fe9,0xffdfdfdff7bc7ff8,0xf4bfffecffbc3fea,0xdbeafff9dfbc7de8,0xfe9e6ffff7167fe8,0xdfffff7fffbe7fe9,0x6b7fdf49f3be3df8,0xfcd67fffddbe7f6b,0xff7efd76f69e7df2,0xffb72f7ed79c776b,0xcddffe3efebc3fe0,0xebdefbd0d7be7ce9,0xbdbeff7bf4be7d7b,0xfffd7ff95db47f68,0x5fe63665ff3e7600,0xebf4fb37f7be75ea,0xfedf3ebfd6ae6fe3,0x3dd84ff1c6b83561,0xd73eff7bd5b47ffa,0x67be7df1d6be3cea,0xebdef73fffbe3d6a,0xfed7fffaddbc6f78,0xffb7effaf6b23f52,0xf5fd3eefefbc35a9,0xffbeff73f7be7ffb,0xfb9effffffbe7dfb,0xcfeffafbdf9c7d7b,0xab61f7d6d59c7de8,0x76dcd6fbd7b43f60,0x9beb5f71d73c7fe2,0xdf9b3b5f77be7ee8,0xdfd7ffbeffbc7dea,0xffffffdfffbe7ffb,0xfffffffff7be6ffb,0x7bebfffdd7b667f2,0xfabe0f56d6947d7a,0xe9f7f6a7d69c7dfa,0xbfdfdefff73e6fe2,0xf7fbfffbf79e7ffa,0xf5fcffffff3c7ce9,0x48480102c10c1480,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x42000001000400,0xefdfdf49ffbc7f6a,0x17bbdfebd6b47fcb,0xfbffff38f7967f63,0xde94b762dc3c7d20,0xf7cbffeeff9c3df3,0x177fff5bf6bc7fc3,0xf79e4ff7dc9e6df0,0xddfdf6faff9e75ca,0xd286ef6d57bc26f8,0xcb80d22294b465e0,0xef7e1ebfd6bc7478,0xfefddff0de9c3dc2,0xd273f6e9d49e57c0,0x7bbc7d71dd9c7fca,0xfbfffb3fd6be7ffb,0xefffbfd7f7b47fea,0xff3fcf7fffbc7df0,0x75ff67fff6b47fd0,0x9f003542d73c24d2,0xffbeffffff9c7ffb,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8100000000c0000,0xfdfffffaf7be7ffa,0xffbe2ff7f7be7dfb,0xaffeffbeff9e7d69,0x2dfededfd71e6cf0,0xe7fefffbfdbc7ffa,0xb7fda644d4b477d2,0xd3467f7ff79e7fe9,0xcbf6ffeffebc6ef9,0x56933e7fd6bc05fa,0xb4fcdefdd3ae65c0,0xd1d7f678d5907df9,0xe97cfd7bf7b67ee3,0xcb76e5bedfb45f6a,0xfeebfedbfd1c7fea,0xafbfeffbdfbe77ea,0xbd3676fdfe8c3dca,0xfab23d71f7bc7fc0,0xdf9eef7edf1e7ee8,0x6fd7fa33f69e6440,0xaf57c90ab3ac2c48,0xc5f5d67156341ce2,0xf16f037897945568,0xedbc3f5bd4ba0d60,0xd8109600d3106d20,0x599641548c2440,0xfd0f032bd6306749,0x212cf60053b06620,0xffb7ffffffb87f69,0xdbf7f2bf47b83ca8,0xa21b1c0bd1040570,0x14ae0da0d2a02dd0,0x1000228d0004050,0x821e7a44b10c2d20,0x954c2285a003c20,0xc6e81e4ef31c2490,0xbdba4fdb569c26f0,0xf5fefffaf69c37f9,0x7fbe3d797fbc7de9,0xf891f32ef7bc34e0,0xc1436e39548621f0,0x2f76f7dbd69c64fb,0xf69ff67dde9e57fa,0xf7bc2f30b1bc07e0,0xebb2fffcf59c7df8,0xe2d5be11d41c1ef8,0x2475cda4e9c1450,0x72c9c4788e9c4460,0x59ac1712d49c1640,0x8aa82628521c2470,0x9ddadf1ef5184430,0x3f3e195877107c30,0x6b8e4d356386cc8,0x102c74051200530,0x180e0c20c6102900,0x6681972652802478,0x95412c67e180620,0x957a7fe3563c7649,0xe1cd96b8f68436f1,0x82200264d1004050,0x8320e132c59a4442,0x23d90c01c0086431,0x10060a90100400,0x10e0245050040490,0x200200000040080,0x48c04a0a870424c1,0x800210080806000,0x0,0x9050864413200468,0x819cad00f4902140,0x0,0x206e003c41c4000,0xa200044a10042980,0x92048460402c14a0,0x1000004040000080,0x0,0x4900b202441400c8,0x800200000408,0x16be06e2919c43c8,0x4100a000a0004400,0xc880294d748c2c82,0x146e020c5106800,0x6b9cd41897b42d20,0x88034946003c00,0x88790f69d5b43f68,0x58c44a44c7180408,0x80402b1c74902421,0x4ece9e38c7263420,0x222e0ec856902c40,0xd1d53278b62c3449,0x500000000100000,0xe850600286342c00,0xc263f6da50a43c61,0xe408441c5104000,0x8610166854062820,0x4800000080000100,0x4200cb2cd5840140,0x2e00042002002,0x10cad814902820,0x50802040400028c8,0x928f805ad0907cc0,0xd02414a2c2045480,0x88e48201f1200680,0x32084c80a1806420,0x86868ef3f01c3da0,0x6080000100,0xb8d10110049c1440,0xcfd4dc2754b43eea,0xc064e08c2800542,0xed08703d40c4ee0,0x5002825010986c68,0x82123d2450000a82,0x8030db8212182d48,0x4964e0bd2246cc0,0x5fbcfffff7bc7dfb,0x64f277e9f6b42d6a,0x412a3700d7344780,0xe9d4e2aaf59c3d40,0x84cba801d6142578,0x68a848a1dc0c25e0,0xfdbf76fff73c7f4a,0x9b3a9d2253143d92,0x0,0x0,0x0,0x0,0x0,0x0,0x81d0140872042170,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe414006813104420,0x8400000080400,0x0,0x0,0x0,0x0,0x0,0x0,0x10001000400,0x0,0x0,0x0,0x0,0x0,0x0,0x5fac4562961c2960,0xd2846880b1906c48,0xc7a8d80ad4982d70,0xaafdded6f61c7de8,0x94ca4168d2045422,0x4a283722d39405c0,0xf126bf609e346fc1,0x638ee731d41c7c48,0x655a9042c4123d08,0x8000061040200400,0x111007a12302460,0x40a2294014900408 },
	{ 0x8a0aeb6289fa2289,0x880168008a02210,0x80285eb18a040316,0xffffdffffbfe2fff,0xfdfffffbdbff2f7f,0xfbfeff73fbfb27ff,0xfadcffff9bfe2fdf,0x7fdfbfebfbfd2f7f,0xe7ffdfdfdbfe277f,0xf4fffefccbfe273f,0xf9ef27efabf62bdf,0xfbde7fffdbfb2fff,0xefeffffedbfb237f,0xbfff5fdfbbab222f,0xfcf7dffb9bb82f77,0xd3feff7e9bfb2fdf,0xfebeaff3abbf2ff6,0x8d5ff7ae8af1278f,0xf7fefffefbf7233f,0x9fbffffe9bbc2fff,0xf7efb7fb9bfc2f3e,0xdfd67ff0bbf8273f,0xc9daf3ae8be227cf,0xeff7feffdbfc2fbf,0xbf8eff7d8bd823f5,0xec7f2af0b989264e,0xffae3fe2cb9e27b7,0xf9a1e3ee83ac27ef,0xfffbffdffbdf2fdf,0x3ffedfab9bb023bf,0xfd77f6eacbbc277b,0xfffefff7bbde2fbf,0xffff7f7febb62fff,0xcfd6ff1f8bbe23ff,0x6b9ecfdaabde236e,0xd63736e89baa276e,0xb9feff7bfbbf2bff,0xffbe7fff9baa2fff,0xeede7f1edbda237f,0xffffffdffbdf2f3f,0xfffefff3fbdb27af,0xfbfffffb9bf2275f,0xfeb6bf7fabdf2fff,0xedffffbddbfa2f6f,0xafeffffbf39e27ef,0x5ddefffb8bbd2f7f,0xd1edfffa9b9f2b3f,0xc90ce9408b880316,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe7dfdfcfbbfe237f,0xadfe5fef8bda2b9e,0xfb3eff7ffbd02b1c,0xfeb7ff6e8bae279e,0xafdeff2b9bb3236f,0x6f57bbdfcbb02f7f,0x97bbffff8bd423ff,0xf7edb7cebbf827ff,0xe38e3d2c8b79237e,0xe3bce3238ba32b96,0xe0ce0a568b6422ea,0xaa187f70cb082396,0xdd75f2e58b942f78,0xef767f709b992f7f,0xfbd7ff77dbd8275f,0xaef3bfdacbe62777,0x6fbfcff1abce27ee,0x74a797ee8bfc277f,0xdbb2f5728bf02f2f,0xfffffffffbff2fff,0x0,0x0,0x0,0x0,0x200010002800000,0x0,0x0,0x0,0xffffb77b9bf52fdf,0xdfb63fefcbf8237e,0xedfeffb9939f23ef,0x2b5e574fdace2716,0xb0024ff3b3980347,0xfcfce5f08b93276b,0x867665bb82013e,0x63f7e76e8b9e2ff6,0x6e570e89138a2b1f,0xd3be1f7a0bb7227b,0xacf672a10b5c274e,0x59e2b8428380025e,0xfbf6ff7fabbf27bf,0xffffffffdbbe2f7f,0xfebfcffbdbf827ff,0x5cd1f27ad39e2b9f,0xdfbdbd708b9e27ef,0xeb903f74bb902224,0x99dffe2f1bf327bf,0x8bce8bd089f02639,0x5cf664e0ab9a215c,0xd1a6b4508bd1021e,0x91004f608b402a49,0xd8d387188b0c0b0a,0x8388061383942704,0x36be1fe6c3b2267e,0x40ec827083942314,0xdba0245eabfc031e,0xcddcf6a28bc22b3f,0xfefea5bbbb82f5f,0x1ef16d8c8be6270a,0xed3cbc72838d2b2f,0xdeb0f74682b42b1e,0xe012420c93802b0e,0x281006848bd82000,0x27fbee61aad12b9f,0x910eb05800842b01,0x739c7613bb802f15,0x9b18c30c03920b1d,0x6c4a000089802218,0x224e5188c02312,0x499612688a102305,0x903a2df083b22307,0xcf86676f1bbe2b1e,0x4ddbfe168baa270f,0x234ed3508b9a223b,0x3400766119d22305,0xf12e8be08bb42243,0x8b88050ec8c02305,0xca560b0e4a402b04,0x2ffccc59aad02315,0xa434406a8290030d,0x8049245012902704,0x6c91245009d00276,0x6042020000100a00,0x2d64200082802002,0x300a014088080a0e,0xa1018e9009880a52,0x4080260240202a00,0x1da76504a182207,0xccc23e11c2302205,0x200201124a002200,0x120025848402000,0x848414883002006,0x0,0x8000200240002001,0x0,0x4863d7698bf02f3a,0xc590a370a998223a,0x2c0001120000,0x9c90920088a80212,0x8000a4001800002,0x4050265001900030,0x4000800000208,0x0,0x808fc808c39c2217,0x2100000000000300,0x420002080800030,0x8000060082002402,0x95a0054002202b0c,0xc0c760020bc00365,0x6a7effc309952e63,0x80101a82802014,0x80e0fe42831a2303,0x70a8a400b280022c,0x89f4c33a9a8a0156,0xea441fdae19a2641,0x603c42939384022f,0x810ca25883102219,0x2c23540a200020e,0x8a10410082002100,0x8523212b3902105,0x22ccedad3902124,0x5020c448a1910013,0x20027470a2022401,0x81004b0480802031,0x8002910080002010,0xad76c21a8be0222b,0x548068a983b20202,0x805c213881a20b39,0x2195360283000b3c,0xc000511389300208,0xd2d4c100b900338,0x1490c6e81100026c,0xc02d225883100104,0x761e2d50ab342b43,0x8480100010000304,0x890a01000b000002,0x698155eab3a260d,0xd36d87685bd82776,0x816407011a0026c,0x8100846002010106,0x7ece3fd9dbf92376,0xbff9dff39bd527fe,0xd6e5f07b8bb82377,0xcb3215529bc82eb7,0xfac6ab0a8bce22a2,0x50e3b7109ba0270e,0xa82c9588a48230d,0xe8ffbec09bd82337,0x1088011000000300,0x0,0x0,0x0,0x0,0x0,0x82005200b080202,0x840020082000100,0x0,0x0,0x0,0x0,0x0,0x0,0x8180002000,0x48001100000,0x1000010080002000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc1423c4092822705,0x1e4c7d408880230e,0xe9d6ee800b99223f,0x2fdf5f488a05235f,0x9430d7658bc0233f,0xdde9a8c40bd02717,0xdabe6ffe938a2b5f,0x4120c7b0c3902a37,0x2b0808d2008c2290,0x80107a621a802205,0xf08174688b9c033b,0x9426094488402690 },
	{ 0x89c8d31c6034c328,0x584313220001c3a6,0xefffffffebadf7aa,0xffbfcf7efb3ddbbe,0xff7fffeff9bdf7be,0xdffffff7fbbcdffe,0xff90eb7fe3bdd3ae,0x89ff7e15fb30d3ae,0xfbffdfdbfbb5df3a,0xf4f57feef9b4db2e,0xffffff7bfb35dbbe,0xfb9677757315f3ba,0xfcffff1eebb5f7fe,0x7bceffd7ebaddbbe,0x3537eff7ebb5f7be,0xdbfbff7afb1dd3fe,0xfeb6ff7feba5f9be,0x61f6e33fc3bdd73a,0x6fdffffbfb34fffe,0xbebbfffffbb5dffe,0xfffdfffffbbdd7fe,0xfa9e7f6ee935d3be,0xcafefb7c5b05f5ba,0xfff79fd3f314fbbe,0xbaeeffbfbbdff6e,0x9dd2ff7beb25f3ba,0xff9e7f7beba4f96e,0xfb74fb6ffb21d7be,0x4fd23f7fd934df7a,0xeff7ddfeebbdd7be,0x77f7ffe8f980d16e,0xffbeff70db1cd76e,0xfab8df7ee0bcfbae,0xcfdbfefff335d9fa,0xfbdfdfdffbacfbfe,0x9638dfe4db21d7a8,0x84200400a004002,0x30041008304200,0x0,0x108141208304260,0xbdff7ffffbbcff2e,0xff6cef73fb20d7fa,0xfebe8f6e5b9ce39e,0xafe7ffabfba05f2e,0x2fdfffdeeba5d32e,0xffbaf57be8b1d37e,0x3144162888205320,0x200040001004000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2bfeffdafb24d1aa,0x75b7ffefeb14ebae,0xb77cef7efb3cd33a,0xfe3e377efbb5fbba,0x2dd6f796498dc3ae,0x2f9cdacbe38cf5aa,0xf7bfcffbebb1d76e,0xffbdbffaf185d1ba,0xf3766d71c3bcd7aa,0xca04a066c884d12a,0xaaf71ed5cb35f3b6,0x39a1ad82c115d10a,0xb01d9268a91453ae,0xde9c7df3c394f7e6,0xebe4d7eefbb5dfba,0xaafeb3dfdb35dbfa,0xf3bfcfffdbb4d77e,0xf5b7f7fbeab1d30e,0x4baefd63fb04d1fe,0xffb7eb7fe31df3ae,0x8947e8815838c382,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xb97def7ffb34f7fe,0xfeb22ffee9a5dbbe,0xe7fefb2ff3bdddba,0x48ff5f5feb2dd52e,0xfdba5feba9b9dbae,0xd377df72fb95d32e,0x9316b2954180c11a,0xf959d309f0a5d9ea,0xe7fc89cfe1a04b0e,0x9eb7efcbeb3473ee,0x95fd26e88b85c7be,0x69c6f77aeb00d9aa,0xfbc4ffffebb5f7fe,0xcffbffffebbddffe,0xafffdffbfbbdd7be,0xddf7feedfb15d376,0xb98cb5705391d30e,0xc3b7d358eb04d17a,0xdcbea284110c76a,0x9e03fa5b832c7f28,0xb6bf7feecb35fffe,0xdba2fdf1e321d36a,0xce922f70da00c12e,0x101e88050105204,0x8b50df598bb9db9a,0x16b7d6bf59305338,0xd3dc74416304d32a,0x93992f2ccb34f30e,0x142fe3073046122,0xe5ffd59f1a5d30a,0x8eaa5f996314433a,0x81c03620c110d204,0xfb967dc4cb21eb60,0x6801928052004528,0xce4a3a1fcb04d9b2,0x37aacfdffb25d1ae,0xc19c36b9c904f92e,0x77fe74f0fb95d39a,0x8990e702a914592a,0xcd832787104c12a,0xa6bf4f628b11d72e,0x70d15e43d33ccb3a,0xb1bdaf766b04d1ba,0xfdbcfb67db15cb2a,0xc07bff0fd934f72e,0x217fdfd28b14d33c,0xbcf9d7e86924d73a,0xad041f004814836a,0xd11066625810d728,0x4c8bc6082304453a,0x2f545d806924c7aa,0x46bfc0675985c06a,0xc118b800892042ea,0x86142a45c020c882,0x8ccbfa0ef190c37c,0x140014101104200,0x202840844080c008,0x80c4b76d638cc122,0x114b4304004c1a8,0x0,0x10918004000c10a,0x2010104242008102,0x4d312a0882052e2,0x22082040000108,0x0,0x8040101b6300c102,0x5fbedf51810457ee,0x949406f0d0305b0a,0x31a1f20d0105b22,0x8100212441845000,0x4002000000000102,0x480c0000000410a,0x100102000002,0x41280420c000f0e0,0x8000010000804000,0x8171c208c900c306,0x11108201040108,0x243213a48800932a,0x4120148003200260,0x2044000000102,0x45d3f630a39cd3ae,0xebff3d4af2a1f96e,0xb43ade73c895d7be,0x822419008000d822,0x400430004100c000,0x89e5d0e26884813a,0x2ac0808381105182,0x9845820220c31c,0xc560e478c325c908,0x43a0560e3045526,0x3314c30cc1a1d11a,0xc0080850008000,0x21a1f5ae1044922,0x410468e0100b08,0x240080001000,0x5a56472e8115d3e2,0x244820a4000d00a,0x2800cc5a0b80c112,0x748052e0c080d4a4,0x51ed3760e134c3ea,0x8010220760118208,0x1443220882107312,0x2bd99fc20304c5aa,0x105b1b6bda10521e,0x8c4041e0005122,0xc888294c7810c30a,0xc058f80a80241122,0x1d5002004624,0x109349088114402a,0xc004080000300200,0xb6a47448d320d1ea,0x8928904b49045302,0x2e1117c9f390531a,0xbbbf4fc6f9a5db86,0x5b662a9db0055a8,0x13007d008220472a,0xa84830e8124d322,0xe08b3149c8344326,0x833ecd70f8384396,0xd0b076e80b34d7e2,0x804000a104202,0x8200010440008012,0x0,0x0,0x14040001000,0x0,0xe410040080040126,0x0,0x0,0x0,0x0,0x0,0x0,0x400000000a,0x0,0x0,0x40080200000,0x0,0x0,0x1000020000c000,0x0,0x20004000004002,0x0,0x0,0x0,0x1000000480200000,0x0,0x0,0x8100000040020,0x0,0x0,0x6cce7de0e904596a,0x5ff673d7cb14db2e,0xf7f4f21cc194d32e,0x2977d290cb24cb0e,0xfc73556aebb1d6aa,0x9127a44ad130d30e,0xc7952530db38c204,0xa85c5628c325c19e,0x6113de80c104c92e,0x29042625b104202,0x24c103004102,0x720020124000d00a },
	{ 0x4133f32382c6bbe2,0x4786ec0002caf762,0xffffffff8bd7effe,0x7fbf5ff28ad7f3fe,0xd4fdfe6a8adff6a2,0x9b7ef771cbcff7a4,0xef9bf77f89dff774,0xcfd7feab0bdeb76a,0xab9affdb8bc6e732,0xf6e3fdecca56ffee,0xffffff7fcbdfffee,0xfedf7fff8bdff3fa,0xffdfeffe8bdef3fe,0x6f5edb5f8adff772,0x3dfadbfb8ad7f7e2,0xdb5ffffb8bc7f3ea,0xfaf77f7f8bdffbea,0xa9b7d3a58ac6e368,0x33dfffff8bdff2f2,0xbfbbdffbcbdffbee,0xdd7ef77b8ac6fbee,0xffde7f778adef378,0xcbd6fabfcbdff3f6,0xfcefbedd8ad7ffea,0xffffeffb83dff3fa,0xf113f6e5c3dcfb2e,0x49bebdfb8ad6b362,0xbbd7e32dcadefa6e,0xeed7fffb8bdef3fa,0xd7bfddff8ad6f7f6,0xf5f4eefe8bdef3b6,0xffbefdf309dfb3fa,0xdb8efe6e0bd7f748,0xede7ffbd8bdfbbea,0xe75cdeda8bcdfbb2,0x0,0x0,0x0,0x0,0x0,0x0,0x960662201c40248,0xeaa02d5800d4801a,0x8750a00400c40218,0xb2f7b00540202,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xae1fdfdacbdeb6ee,0xfeff77ffcbdfffee,0xdf75fe6e89ded362,0xfcbc3f7b8a56f2a2,0xeff7ff3e8bc7fafa,0xeb5ef75f83dffb7a,0x7ffdff7cacff3a4,0xffffffffcbdffffe,0xfbfef7f1825bf3e8,0xc187fbe88053f2ea,0xeaf7de39025ff7e2,0xbfbdcffb8acfffe6,0xb4f5f26b8acef7ea,0x5f7cfd72cbd6f3e6,0xeba4eba7cbd6fa8c,0xc5f7fb9bcbd7f7ce,0xe6b6df7182c7fb28,0xbcb67fef8adff7ee,0xfbbc7f738bdef7fe,0xebbeff7f8adeffe2,0xabddf72f88defbee,0x0,0x0,0x0,0x0,0x0,0x0,0x1474509a828af184,0xde7fa77ac3d7fb6e,0xfffe3fff8a5ff7e2,0xe5e7d7698a56bbe2,0xabde7e5f8ad7bbea,0xbdaa7ff78b5afb6a,0xfdbc9ffa88d7bfee,0xcffc75618acebb2a,0x88c7fbe0c2daf166,0xefc633db8a57f3e2,0xdf9edffb8b5fffe2,0xf0ffb6694adfb3e4,0xfb9a0d7a8ac7e270,0xfbffff7f8bdfffbe,0xeffffffb8bdffffa,0xffbccf72cacff7ee,0xfcfef7eb0bd7a7ea,0xbbf6fd638bcfbb26,0xfab7bfe6c9d6f7ce,0xcb521e014a45f2c4,0x6756dfc208c6f3e2,0x56f0f4f280c2a3e0,0xddeabfe28ad7bbb2,0xfabc2f7e08dfd32a,0x4cb424888c45280,0xa359df5a88d6f27a,0x12936c60c8c4ea24,0x7bad356988c6f220,0xda862fee0ac6a2e2,0x81c3fb2b8ad4f0a6,0xcffe7a9f8ad2f3ae,0x9cba4c6a8ad7eaa2,0x204307788084fa34,0x4de0755108c4b362,0x8134190c801afb60,0xcd5f7a8883d6b3c8,0x412c8ff7c0c3a2e6,0xb676f779cad7b3f6,0xffbefdf3cbdffbfe,0xeb40cb3e8bc3b3e8,0xebf69e33c2d67be6,0xecbfdff5c2d7f3f6,0x18d2e2e08b83b6c0,0xc4a4a861c85c320e,0xd986fffd4adef2a6,0x876afe1180d4fb42,0x2126cb81cb866374,0xb0108240018e1300,0xf2b046428115f394,0xebb80f2a8a4eb1ea,0x50ad10888425268,0x2b94cc020044a260,0x64b666e100dbb682,0x962824528205a360,0xcb2e0d0e8a44a240,0x8306efec804ef7a0,0x8362d408802f720,0x1f20de5c830aff00,0x80518fe888846200,0x1a882c02c0559a14,0x0,0x52162080099840,0x280002000,0xc14d66988c48926,0x1024004000022180,0xb184e02008d48208,0xa020800001481240,0x916052a0047a060,0x9252ec0850b3e0,0x9116a4700bd4f210,0xd102030080008824,0xc00002002000,0xa080180d00004,0x440000000008200,0x0,0xcb1632050ace1702,0x4144800000508100,0x85084810880b060,0x82500080401008,0x40000008100,0xdc920a3480067622,0xc9c2fdb182dfba22,0x207a3c84a979204,0x489241608200aa60,0xd5cf36698ac4fac2,0xc3023c100882f220,0xe1b1f2e382c7e3ac,0xaede3618cac6b344,0x2008652081a9220,0x9602827802527020,0x492e29798b55a668,0xb0002b020052d720,0x205212f800878062,0x18a084880849220,0x1503108a0ac08208,0x80913510024c9620,0x9a9def6c0356b36a,0x10475a000807b222,0x6dbfcac08adef2b4,0xd436fbe0c2cef236,0x1b2845b3c244ab06,0x50609c008447220,0x580900008c02220,0x1200928602c5a200,0x141b21224048702c,0x4000001000008100,0x4802270002529208,0x8382d300884e720a,0x208082a140828704,0x102481c0501004,0x4008f4080a442228,0x0,0x83f2236ec844f126,0x40b21d400096b092,0xbedfddfa8acffb6e,0xf772b2740ad6b22a,0xc38631a302ceb102,0xe1f483368b5cf226,0x9c778008845f60a,0xf6eedee088cd9aa2,0xc0f276e2c2d1b244,0x8108e41280c4e102,0xc100801280025820,0x0,0x0,0x2000002000020200,0x2100000008020,0x1004000000008100,0x0,0x0,0x800010000400002,0x0,0x0,0x0,0x0,0x0,0x40020000020200,0x0,0x0,0x0,0x1001000000040200,0x0,0x0,0x2000400000005000,0x0,0x0,0x0,0x880080002000,0x0,0x140000002000200,0x2000800000008020,0x0,0x49a833020887f380,0x39be5fe4c8debbae,0x61c1a212820c9222,0x83df74983d8b32c,0xaf974dea81def2e8,0xd191af5a8a54d666,0x884052000d21240,0xe082014a00a224,0x671d7051025a32a2,0x140bcc43081eb222,0xa24004e202d29260,0xc2a024908202f280 },
	{ 0xcbf6edaa4cc7be30,0xabd4fa02cfd3feb6,0xfeebffffedffbfb3,0xff9ecfffcddb7fbc,0xf49e7feecdd33eb3,0xffbe7f62cddfbebf,0xff9effffcfdfffba,0xcfc2fe3fcdd3fe93,0xffffdfdfeddbff93,0xfcfffef8cfe3fbb2,0xcdb7ff59cdc7febe,0xffb07fffeee7af93,0xffe7ff7fcfe3eeb6,0xabffffdfefcbbf3c,0xfedfdff7cdf7fe31,0xffeeef7bcdd77eb8,0xfe967feecdeb3ebc,0xc5cecf3fcee6ffbc,0x2feedbfdcdf73eb7,0xb7baffffedeb6f9a,0xf97fb77becdbfe31,0x77be3f764cc3de39,0xcbfdf3ffcffa7bb2,0xafffdfd9c7d7be35,0x5ffccf3bcfd37e92,0xf4ffd6f4cdc3df9a,0xd7eeedf1cdd23e9a,0xfbb7d33fedeb7ebd,0x4cb3899244a2609,0x2fbfcff76fc77ba5,0xf1cc9fefcecbbe34,0xbeba3f57edd77b92,0xddb8fb7fefcbbb3b,0xdfffffb7cfcfffbb,0xe9411086cec18c10,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x9000000008000100,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xab4b10da4caa3eb4,0xb6f6d7efcca7beb4,0x9323bf38eecfae9d,0x7ebf2ffdcdeb2bb9,0xefe7ffb6efdfdfbb,0x6f777e584dc3ca93,0xec8edf7bc6c7fe1a,0xffffbfffefcfffbe,0xf7f6fc6d4febbf9c,0x89f4d17eecde3eb1,0xcc935e03e1931a14,0xb49ac7eecd8b9e90,0x95b6d06c4e036e3a,0xfb7e2dfaccc31eb2,0xf397f33dcee71ab0,0xc6fb969a48531abf,0xf3f4ffecdd35fb1,0x7fffdffbcfdf7fba,0xfb7fbd61cfe37bbe,0xfca25f6cc4d3be34,0x47d29cb204db3a13,0x0,0x0,0x0,0x0,0x0,0x0,0x4010002008000100,0x91a9a752ccc35eb8,0xffbe2fffcff3bfb4,0xdf56d39a0fe3ff98,0xeffecedfcef7ffbb,0x9f7af7ffcfdbff9a,0xb9ddfefbedfbbe9b,0xde9265d1cc476e13,0xabdffffecde3eb99,0xedff16d3ccdabe1f,0xf7ffcffbcfc7febd,0xfcb792654dc3deb1,0x6c3ae5f2cea38f34,0xebfef77ecff7feb8,0xfefbffdbeffbfbbb,0xbfbfcfffcff7ffb3,0x74f376d84df3ba91,0xb70a3d628fe72f92,0xfb95e7f7cfcf3bb3,0x845a72094dc26b16,0xf37f5d0aece73b1d,0xc4ead5d087c33a18,0x8b4dcf614dc22f1c,0xfbab3f7acff3bbb2,0x4106f60249430f01,0x81c5c184cc22eb0,0x209267ed4c583e21,0xb179f769ccdabe9a,0x9a9aa7248ce3ab10,0xd96f3fecddfbab3,0x8f5b9b7f8df3bf33,0x846dd01820e10,0x5954a5420c82ae92,0xa0102551c8901930,0x8b40203bc883a390,0x668abc502c062b82,0x86d7cff5edc72f1b,0x54f776fa4ecf3a96,0x7ffe7d73ede77fb7,0xf390e73a4ec73a00,0xe64b1f5bcec33e96,0xc3c4d23ecda9b98,0x9c9c46fcc4c31a3c,0x251e8e208dc70b88,0xd895ec22c6c1a624,0xd9e2ba786fa63b83,0x2b7f5ec7ccf25bb9,0xf4b876638fe73ab1,0x9163074088620211,0x8800010048000500,0x948f12881c11a10,0x1a6cc9584cc30eb1,0x41a54e60cd2b888,0x530c821801c10a00,0xe0b31d0001426a08,0xf984da01c0423a20,0x17e768465c2ba28,0x961a4d4084c62e00,0x40a50444804800,0xa065440c6826800,0x81a0c21444412e20,0xa44e90c20c8b2e08,0x210620800c29200,0x4010801001020808,0xa24047047820800,0x82816a064c823aa0,0x43720068020282,0x1cacddab24c2a239,0x404002e04c22200,0x828017620dc30a08,0xf00609000d111908,0x1800830044810228,0x200008a44802000,0x1482120048000820,0x80022140c0921000,0x1202030448800302,0x406201000025808,0xa04845524c9034a0,0x0,0x8004056080525a10,0x36800d700ee10288,0x8d80b310c8c10a82,0x95df6e1a4dd03e80,0x80884ea34e002a11,0x11348c3c8e907900,0xd3be27744c426e30,0x2e02a08400290,0x204a6a1b09471b10,0x79ccb204c80123d,0x4c5dc3084eb21998,0x4c206d50cd820910,0xcb038b6ea8430ab9,0xe026a4000e0202b0,0xa000c0044820408,0x3524166684c30601,0x438afc1068801211,0x42106708c852130c,0x44cb200040c08a20,0x1580a8224c03a80,0x32b0666a8cc3ae99,0x804a274808858f01,0x8084340400454a10,0xc0004002000,0x62ac1a09c4c30014,0xc8ba4de0c4c33290,0x9014060c28420311,0x9d1c037c8eca2a00,0xc354c134ccc3eaa8,0xc52021b24c0a220,0xc81200a80c823880,0x94141e48e8020331,0x800380a0020200,0x6854000048800114,0x2d5ef9ca6cc62e90,0xbb3a6d736fd33ab2,0x1465e044c8c20a90,0x6f3e04f08d833b92,0xb05027248c420a83,0xc1422a1045830e08,0x9ff7ddf24dd23eb8,0x3d7cdac049f3ffb3,0x92202d614c83a704,0x800808000080,0x60001000800,0x0,0x0,0x93408dd00c834f1a,0x80004000808000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80000240800000,0x0,0x0,0x400200000022000,0x0,0x0,0x0,0x0,0x0,0x0,0x30004020000,0x800c80000020212,0x6000000020010,0x0,0x2100040000020,0xfff72e5a4dc3bf91,0xc5c6e73bedd03f1c,0xaf7ef4db49c37b9a,0x346f56e7cd8aad80,0xdb3ce75bccc36f90,0xde1e1f7e4dc2fab4,0x44a21a8c5020e08,0xab58b24ae8c24b15,0x10186d4bc8c11f04,0x93a0206325c18a10,0x528e1f644cd31b3a },
	{ 0xcadcf39eb25214cf,0xfbffff7fff7f97df,0xdffbff79f77ab7eb,0xbffeffff3756b7fb,0xdff77feb2a7e11df,0xffbfe777eb7f17eb,0xffffffffff7e9fef,0x6feffeffbf7fbfff,0xabff7ddbaf7b92d7,0xf6fdffeffe72b56f,0x9d7ff7ffef7fb5ff,0xff9d7f7ffe7eb3df,0xefffffbfef7eb5df,0xe35ffedabf6611cb,0xbdf67ffbef7a97eb,0xdb7ef778af7b17cb,0xdbb63ffeeb7a93fb,0xe7f7fbeb2b5fbfcf,0xeffffeff6f7e37df,0x5ebf5febef7f1aef,0xd57d3fe5ef6a18cf,0xfafe677c2a739787,0xe8f5f73a6f7fbacf,0xced35f57ae6f30eb,0xa8bfec6b666e1749,0xd57ffe7ebe77b1db,0x6f3eff62e37731d3,0xd8d4eb3e6f76908b,0xe6c3be022e7a804a,0x75b146f3606215c1,0xf4b7dee8337e12da,0xffbfef722f779fdb,0xdf8aff7e6e7010cb,0x8904a20008601801,0x0,0x0,0x0,0x0,0x400200000000082,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2bb9ded8fa00941b,0xffffffeb6f7218ef,0xf577bf73ea7a177f,0xfebe2fefef7e15c7,0xdfd7fb3f2b5e1beb,0xaede3ecf7e721beb,0x26b9e7e1b24016cb,0xbffff7ffff7b15cb,0xffbebf737b7713ef,0xabc5ea336f7311db,0xe5f71f506f4215db,0xbcbbc5ef777216c9,0x93fabe7e227e13e8,0x6f9e7d796a72bfc3,0xe9b47faf6c671f9f,0x6cdb9b9b677811fb,0x6f7fdfd1ab7ab6cb,0xf4fffffcbe7f94cf,0xdbbd7ff22e7b97e3,0xfbadfbfe6f5ebdcf,0x79d7bf9fef76bbcb,0x0,0x0,0x0,0x0,0x0,0x0,0xb491b2eea06315d8,0xd1f975506e5a16cb,0xdbd6377c7a723beb,0xcfc7ffb33a70bcef,0xabffbfc86c769cdb,0xdfbe7ffefa76949f,0xfbffbffcef77b7df,0x54b633702a70118f,0xe2f5d7e6a37a12ce,0x37c7be79eb5e9bcf,0x28e456aa6421b46,0xd7fdf77aee7697df,0xdffe7d72ff5a91cf,0xfbb7ff6efa77b22f,0xfff7b7fbea739cef,0x6fbfeff9fb779fdf,0xfddebe6eee7e91cf,0xebddbfe2ea7fb3cb,0xfbf6773fea7ebfcf,0xe90feab0a6218ab,0xa1a8ddff464304f,0x5a9af77cb7473a8b,0x5bae356aa15a90db,0xdabe253ce75214da,0x1006760602121941,0x4b6eded1687292c9,0x4877d4a01771740,0xe129ff72aa7a15e7,0xdebe3bffea6394cf,0xedf7f7996b5617df,0xe87fffe3fb6318cb,0x4da26f3b225211db,0x9a31a776207211ce,0x4c166571f44039eb,0xd056c3142646116f,0xe9d37fe6ba46328f,0x37d8ced062571367,0x4075a202c850120b,0xd6b40df83a58934b,0x99009111c072100f,0xcdcbf851ea521fdf,0x7b0e6bf3a2461c8e,0x340408a18a021087,0x19bc2d46ae708201,0xbea003621a421283,0x1747de0a7c72137b,0xdf9fcfdffa7abfef,0xf7ff7fef7a7fbfef,0x11682f70ab42174b,0x3090225410623248,0x8691ca0a7240924f,0xa12ccca21529449,0x490026a70101402,0x13f18f48684200cf,0xa920550006200a6,0x600000000006,0xae4e20002270914b,0x65244800010114e,0x4cbc7031a8461186,0x226046000521106,0x8a02432c2a6e0241,0x468600302c421208,0x1c06a280480046,0x202610a2805c0042,0x8200a0008000004a,0x108462564421401,0xec42011961501a0f,0x26fac72ae06619cf,0x118a27d8a6569003,0xcac22d00a246204a,0xd68e0e8e287f10c9,0x8c683132686490c1,0x1a06cd4802411549,0x210040220121000,0x81028500a0703000,0x800040000021000,0x808a1a20080a1041,0xa640041a4401602,0x404404000621000,0x4000a00100611001,0x480411f0225092c1,0xf7effa6aeb6a12cb,0xa95e7e8b2776b687,0x12134ec022621891,0x94e53e016c4312c9,0xf502510a060204a,0x8a00632c486000c3,0x40a4a6c000001,0x800c4fa0225a10e2,0x1522009048400207,0x20121d0003581041,0xe000080021400008,0x9422001a0001492,0x48641004040110f,0x34306ae02552108a,0x92b4a500244611ca,0xc79c409e636890e9,0xc3f29913717893cf,0x2719d95008621301,0xb03056cc2666108e,0xd12c07882a54184b,0x10143530a460100b,0x8c2e28c234810c1,0xa2340c872472208a,0x8434d221e242820b,0x1944044c8070000a,0xc6922f466e111483,0xc1d7f6302a6b91eb,0xc74a0024121082,0x632422822460287,0x35cc35422a531081,0x4f021405104214eb,0x8192800b24661200,0x7d70b4b2e7010ab,0x9fff4ff3ee76314b,0x9e9612d1bb6215c1,0x4efe7df2e67236ab,0x9b9aed2ee56e93db,0x68200180460002,0x23be0c222f42226f,0x71f5feeb3276915f,0xd22c6521b9701031,0x0,0x50000000402000,0x0,0x40040020000080,0x80982100a2600001,0x280000000021000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000100000600000,0x0,0x44000500000,0x0,0x0,0x0,0x0,0x0,0x0,0x201000021000,0x200400000001400,0x40080000400040,0x2b7001caef63159b,0x2000002000000082,0x4884054120609590,0xc790a3172b7a16c3,0x65defe287b76150f,0x2bfa59d13b6a1049,0x34eadd4c216216cb,0xe9fea75aea7eb7cb,0xf29e3fb02a52b5df,0xc5e7c3266663914f,0x6edb1e9b6e72312f,0x120000000400040,0x9dbc350222471180,0x7fb0bd7f3e7e18df },
	{ 0xdbe8b568483dccf0,0xfbffff7fea7fbfff,0xfefffeda887ebfff,0xcbbedf77ca7ebff7,0xfd77feee8b3edff7,0xfff6ff77c97efef7,0xffbff7ffeb7fffff,0xefffffffcb3ffff7,0xabffffdb4b3feafd,0xf6ffffffeb3fbeff,0xcbeff7f3c97eafdd,0xdbf4ffff693cbfdf,0xffffffbbcb7fbfdd,0xffffffdfc97fbffd,0xb4f3ffe3e93fbfff,0xf7ffbf7feb7efefd,0xdebfe7ffc93ffff7,0xefdfff7fcb3efde7,0xeffefeffcb7ebeff,0xb637ffeb087faefe,0xd7d6ff4e693caeff,0xffde3f77893feef7,0xe9fef3bb6a7eb7fe,0xea73df5fca7cfef6,0xaebd4eefea7d3eec,0x97fff6ff8b3ebcdf,0xefe86c63483cb0dc,0x39dc6e2e883c2af4,0x84c3a201681c81c9,0x2185b11683c0dc0,0x54f776e86a7caee2,0x9fbfff77883eabf7,0xdd2c2c48083585d2,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1e4e10491e8bc2,0x948bc6eb403e8ec6,0xff3baf76c83c8fdd,0xfebe3fefea3ebeed,0xc7fbf775ca3eeef5,0x5f1e5ae74b1faaed,0x923bc7f6eb3c5ef4,0xfdffffffeb3fffff,0xfffeff7fcb7effff,0xcbfffbffcb7dfeff,0xe8434b40801d4f74,0x3fffdfefe23fffd7,0xb6b667fc683edbe7,0xfbbe7d716a3eb6ce,0xfbbdef6feb3fbfd6,0xc5cfbf304834bec0,0x843ecf62083ea615,0x15b736ee823c8cf7,0x9fbeff53e87f8df7,0xf7b3ef7e693eafdd,0x2fdfa2becb3dfbd5,0x10008208208010,0x0,0x0,0x0,0x0,0x0,0xb6931fee483cb2d5,0xfbe4f768017e9fd5,0xffff3ffec97fbdd4,0xedf6dab74b75bc5e,0xbbde93df4b3c8ff6,0xfdfeffeb493d9ffe,0xf1edb7788a3f8ff8,0x8b9ef93200379dd4,0xa9fef1afe93eacf9,0x8b620f52083eb9a2,0xbfbbffebea3efffe,0xd507b0e94934bdd0,0x7ef63d734b3eebd6,0xabdde26ee87dafdc,0x88c3f7c24a3e82c5,0xb3bfcff3c97eebf5,0xdd767f7f4a7dbefc,0x95baaf50cb3cbef0,0xcf92bf2f883fafff,0x58567abac23cccf6,0x2a00f5900934ac83,0xb6b954ee697caac0,0x5d668572e9579a42,0xfa3c67666876a5d7,0x83e5a70a411c4e14,0x148918008143004,0x1c307103422c8980,0xb77aa771c93eabe5,0xf6b6a75e497ea8ed,0xefd7ef9fcb7faeff,0x2003ba99417c8c81,0xb60607f50a2c81d9,0xe16d346a097c8cfd,0x94963de1887caffe,0x8bc4e332411e9cf1,0xafcfbee8c83eeeff,0x8582191b4a34a3c6,0x97b2fb51493c9ae4,0x14a82d70483dabd6,0x4144011a01080810,0x80009610411c2890,0xef9f5fd44b7d4ff4,0x65f6b4ea405e86f2,0xa0343c30003690c1,0xd504ef6c483ca3fc,0x1d61604e0115c40,0x969dfd3c83e9dfc,0x502e4660c11689c4,0xf1ac677ae91daed7,0x7b9574b00a762954,0x8081080240460000,0x4c9e100143048c90,0x94143ec00826acc6,0x9a042608011c0844,0x88520040092c0800,0xc9c2f620403d0674,0x28bd003001e82c3,0x2421410080108a10,0x3055017080158ec0,0x282c356c410e8941,0x4a006102ca048841,0x1c01e0008248cc0,0x4a1e7780856b005,0x2434263088300854,0x46902070881489c0,0x8300cb08087c0054,0x34716000854820e,0x2028400008300260,0x11801463017c08d0,0x803629424817a0c0,0xd2362b648a749ad2,0x8dcec2134a1e2ec0,0x2000901000110040,0x0,0x8000040040160000,0x8190021601248c01,0x6004600600303040,0x80411060200040,0x4048169082c8a94,0xa144a20280108ed0,0x415580a14040c,0x69c2a00a082da072,0xc4e5b5a483eead0,0x8092cef9003cacd1,0xdc707761c82e0bd8,0xca946c71083c1e34,0x0,0x66000000088012,0x8030040240140050,0x14be4028880c8e9a,0x86001200840,0x80040360401080d0,0x88522e03c82ca84d,0x0,0xf08410214a340a08,0x931d152001748844,0xf6029b68ca3ca814,0x2966bf21411e4a74,0x299bdb51083c97d2,0xc49673f9e97eaedf,0x42e7f760683c82d3,0x200010008000040,0xc8478321001d81d2,0x800c1417086cc810,0x2304f4080108c45,0xf9d487b8483c82d9,0xff162f340b3daed3,0xcc0a26030811115c,0x2ce61289685c8799,0x0,0x20020000200040,0xe0105d44411e8c45,0xeb66a92b421c9696,0x484b3711083c82d4,0x38ac8141882c8c80,0xd8fef676083ead54,0x1eb20570093e8b93,0xc38ec00cc03cbbf4,0xedc2bb9ec83cadf0,0x21344b60003c0084,0xf574fee9893ebef5,0x2a4056241360c40,0x4008000080080,0x0,0x0,0x0,0x44020802141000,0x0,0x0,0x0,0x0,0x0,0x81000001000800,0x10200080040,0x0,0x0,0x8010000000080020,0x0,0x4000400000101000,0x0,0x0,0x0,0x0,0x0,0x21001000800,0x0,0x0,0x200040000101000,0xbb266be3c33c0ce0,0x343920a0917ba41,0x36508340e8100a44,0x14a9010003140a40,0x7b8e566a483c1ef4,0xfe6eae72483e8ff8,0x49c6fc02081c92dd,0xaf5ccd5b093cbad1,0xbcbbeee90b3cafd3,0xf3b8bf7f483e82d7,0xfeb22fdfe97eaff2,0xc9eee2aa0b3caa92,0xc3ced9474b3ea8d1,0xb7e2477b493cafe1,0x2071b779693c8cd4,0x694afe7e493cbef5 },
	{ 0xffbfffffff9bdf5b,0xfbffffffff9bdfdf,0xcfefffffef9bdfdf,0xbfbfdfff3f9bcfdb,0x74d7feeb3a9bdf1b,0xbfdeed70b59bcf5b,0xfbbcfffe5b19df5b,0xffffffbb7f9bdfdf,0xbf7fffdb3f9adf1b,0xf5b7feec7b99d71b,0x6dee77ea5f0b5fdb,0xfbf7fffeff9bdfdf,0xeff7ff9f7b9bcf5b,0xaf5ede9f9213df5b,0xb6bfffea9f8adf5b,0xfbffff6a7b9adf1b,0xffbf2f7f7f1bdfdb,0xc3f6ff3bdb03cf5b,0xbf5fffff7f9bcf53,0x27be7fffdb9a9d1f,0xdffdfff8139adfdf,0xfffc7d6feb9acf5f,0xe3fef37eaf8adf9b,0xfed61f5a8f8bcfdf,0x9bbc4ff33b0acb12,0xffbdf67d7f9bdfdb,0x4f3eff733f1bc7df,0xcbd4e6ae060a4f5b,0xae67080a080bdd93,0x605400083440a,0x4412aee80700471a,0xdf865f60578ac71a,0xf88e0d6a0b8bdfdb,0x0,0x0,0x0,0x0,0x200100000081,0x0,0x4000404004000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8b3c4c065f18470b,0xb7b637eaeb82dfdb,0xf97efffa7b9a0f5e,0xfefe2fffb39adf5a,0xcdf3beb5579acd5f,0x21c61a9241084b1b,0x400102801004010,0x79baaff2778acf5b,0xc7ba7d737f0bdf5b,0xe9f7f33fdf8bdf5b,0x6bc64e114d02cd0b,0x78aef77a231acf17,0xdf68aff1238bdddb,0xe7be2d71ce0a4fcb,0xebfeff76cb9acfd3,0x9cffff7b7f9adfdb,0xa39ecdc93f1b4d5a,0xfefff7e9df9bdf5f,0xb3b617704f03c15f,0xdb9fcf79df03d757,0x4fdeffb6bf8bcfdf,0x0,0x0,0x0,0x0,0x0,0x10020000020002,0xbdbb7fff1f9a5f5f,0xfbe5a7788e0b5f5b,0xffbf2f67d79acfdb,0xedd7ff663b1a4fdb,0xa17ff6ffbf9a5f5b,0xb6bfeffb939a5f5a,0xd3f09f52451adf53,0xdbbe7d7d079bcf53,0xabf1c02b0b9a0750,0x66943cf853005f1a,0x331e0dc20701d310,0xf6f6776b2e9bcf5f,0xeaba2df03e0b5d58,0xcbb0e53f3b0b5b5b,0xe6f2b69b4f0acfdb,0xffbfefffdf9bcf5f,0x35febfed7e9bdf5f,0xff1eed77469bdfdb,0xfbb7feeebf9b5d13,0x44c62e1c0e924f19,0xebbdddca960847db,0x7411f4209308450a,0xbffeb773dd8b5f5b,0x48b9073e5b18df13,0xcdc3f009a788d71b,0x223114c649020343,0x3cae7b6a130a475a,0xfffbbf7f5f99df1b,0x588e3f7c7e0ad797,0xeddff72edb03df5f,0x29c75fde730bdb16,0x3411288a0702c51c,0x616d3588161ac497,0x9a86150c12014d0a,0xbc6a22a770acd43,0xeeff1e79178bcf5b,0x573ac7d317825f1b,0xc40c301a63114581,0x53cc3c60810a9f5b,0xc881c53603920446,0xd832a1130089d11,0x19a48b2b13084b03,0x10a0044402008301,0x2010050001000203,0xda32474e1f03491b,0x4cc2df110b8a010a,0x11844110004d01,0xb4afd06c179a4f11,0x1002e1080100500,0x540008580100870b,0x8cf12a8132128f0b,0x2a911e521a025d50,0xb4a75020160a4d4b,0xdb7e7f73db9a8f0b,0x8870760400a071b,0xc020100000020b05,0x4432a484802094a,0x1b84d0310125c50,0x1015420018a0042,0x812e2c4000020a4b,0x9028d280f004302,0x4d5a2e65079a441a,0x908564f86b089113,0x945596180103cd1b,0x28262c601208cb13,0x8980e34f0e08d71b,0xc0022000002000a,0x8284cb600a120750,0x181306f06800c311,0xab0434320602464a,0x8a02277802085f1a,0x180480000004c40,0x83814421482c311,0x80a8526900020bcb,0x813c0000200a0140,0x6010052604824609,0x40831300100010a,0x2001108000014408,0x344010a422094189,0x5b68254213014310,0xb090474603000542,0xcb8ecb2ebf0ac5db,0x68f602cb0102dd1a,0xb5b8d7731393dd52,0x21ac90720a024f50,0xc800ed33118ac597,0x88c00100001c19,0x842200000020801,0x5400442014110,0x185426490103c915,0x5206055017804743,0x8900e2256110c512,0xc16312210308070b,0xad01e021098ac101,0x844454000110451a,0x8d071500a180701,0xe1a09e27208a5d18,0xcdefbe3f238b4fd9,0xbf7fcfd76d8bd7df,0xd497f6f217025d1e,0x8164745f1b0bcf1f,0xcc381710890a4f53,0xc05dc20c2111c513,0x2b563c5c0a02c751,0x1107d30e084312,0x990d745c600b4344,0xb3349774928bcf19,0xc948f6aa18880f0b,0x64544e4d25185753,0xb49042e103094a91,0x409610008284c3,0xf48e5d613f92575b,0x184c006200ac941,0x2cf7ff4f1f8b5fdb,0x2b1a6d997d934d42,0x817be778330a8f5f,0x4f367d71fb8357d3,0xb914d30ebb824f13,0x895212860282855b,0xa57edffb6f9a4f5b,0xbcf3b7ef7f9a4f5f,0x9bcab7004382cb53,0x200200010000100,0x0,0x0,0x1080000000004800,0x8030040002020810,0x8200000000000002,0x0,0x0,0x0,0x0,0x0,0x80100000100400,0x0,0x0,0x0,0x0,0x802000000440,0x42001000004a10,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000400a02080401,0x95a90800a004719,0x210c8830a080302,0x4420049012024840,0x4bceff627b8a8f4f,0xf1302f727a1a0f1b,0xc5c6ef3b1b1acfdb,0x6fdf58d1be8b5f5a,0x96bf7feb5f8bc7db,0xf7f9bf6b9f9bdfdb,0xdea4156e669bcf5a,0xc9d0f737c892d75b,0xefa1c5b0693cf52,0x1a8a07d352180753,0x407c0642130b4f57,0xf6963dd00e1bdf58 },
	{ 0xfffefffafdf9ffe6,0xfffdff7ffdb9ffe6,0xeffffffbfdb97fe6,0xffffcffefdf97fe6,0xdfff77ef3df97fe2,0xffbeff7295b9efe6,0xdf9ff77f7cd9ffe2,0xdfdbff9ffdf9ffe6,0xbbfedfdffdbdffe6,0xd67e3fe854b94ee0,0x9f2ff76a34a8efe6,0xdff47f7ffdf96fe6,0xefe7ff9efcf4ffe6,0xa6fd7ed798e8e4e6,0xbfdedfebfcfdf7e6,0xf73ebf5a7cf97fe4,0xdffebff7f5bdffe2,0xeffefffff5bdffe6,0xbfdffedb3dfdffe6,0xffbb6ffef1fd6ee2,0xb17cff7b95fdfde6,0xdffeff7cbdb97fe6,0xcbf1f3bb5cbc6de6,0x3f570fd94db8c6e6,0x98bf6fef55bdf6e6,0xfdfff7fffcf9ffe2,0x5b96ad7bb5b86ce6,0xfbdef33e15f8fee6,0xffc7bffbb5bd4fe6,0xafbf5f7a75b866e2,0x347e06e13db05ee0,0xdffeff72bdf86ee2,0xc104212544306ee0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x69449042098c9e4,0x53f47f21cb965e0,0xfffeff3275b9efe2,0xde9737e6b5b8eee2,0xdfdbd768d4bcffe6,0x4c00e28508086a4,0x400400000280000,0xff6fb77ac4ddefe2,0xdca63d61d49c67e2,0xe9fff36b35b9ffe6,0x63771e5354b97fe0,0x1bbacff21cb9fcc6,0xfd72ff6234d8efe2,0xffbe2df07dbd77e2,0xe324f737dcf96ee2,0xcad3ffde74fdefe6,0xe73fce931cf8eca2,0xf5f776e7ddfcefe6,0xb79c3df025f84ae2,0xb28b673965f05ce2,0xcd67be1234bdffe2,0x0,0x0,0x0,0x0,0x0,0x0,0xffff5ffffdfdffe2,0xffffef73bdb97fe6,0xba3e39604df867e0,0xcfffffbf34bdefe2,0xdfcfffdbb4bd7de2,0xd73b5ff3a4b97ee6,0xd3fcbf78f5f925e6,0x5b36bd225db967e2,0xe3dff33a15b9d7c6,0x6fff7e9ab4b8f7e4,0x4a9cd49e34904be2,0x94d6f2e815b04fe0,0x6bbefd71f49cefe6,0xbbc4fa377cb86fe2,0x98ce53dbbdb9dce2,0x2d7fce5195f9c7a0,0xf5d47ff91dbdefe0,0xfbbdef72bcb8e7e2,0xfb97ff7ffdf9ffe6,0xc3feff13349d7de6,0xeb7f9ddffdbcfde6,0x57fc76c8fdb96ce4,0x99647c4354886de4,0xd786efdc95b96f62,0xc8dee20a68956ae2,0x21a134211394822,0x149d549e11b928e2,0x8061a44215b06cc6,0xd33e3f6755b9dfe6,0xc9e6d38481b9eca4,0x2be79e58dcb96fe2,0xb03e0d7341282be2,0xd1e5d6683cbdc6e0,0xfb2e2ce95db9ffe2,0x88c1a0021c905520,0x4ea77e9094b976a0,0x13aec57a34b87ee6,0x8536a4d014117462,0xd02a7dc051b96c62,0xfbd0e33a7c114ee6,0x1080321100106084,0x2b184d18041144c4,0xd52772aa11a8e6e6,0x212ab56084206de0,0x3004012430910440,0x948160011310042,0x800ecc431090c684,0x22a6846014317442,0x126e074051204ea0,0x8280000200282,0x100800000100400,0x8565d0054918520,0x5c045100052032c0,0x1baccc720499c5a0,0x282006011310002,0x8070020c1c084500,0x2159c80a40b14422,0x89407a0149804c2,0x9ce50520453104a2,0x7522084e209867e0,0xca2afb6a3c394fe2,0x6e4115081314c00,0x5010ccc1adf10560,0x1080500c10840942,0x482c0c2211284c22,0x100802000110280,0x26ebfc1f9f046a0,0x69ecbdbbdb987a2,0x9cde7fea85fd7ec2,0x812511010882560,0xd2107848300064c4,0x555280000386840,0x2010485c04988820,0x54400028009020e0,0x880c047214a06660,0x100082401880c002,0xc0000000280,0x10a1442800046c0,0xd9379a870116de0,0x94486021480cae0,0x8004022010004560,0x51ed1bcb4b974e2,0xaffefe9f3cf9dce2,0x9e2a0ff255b9fce6,0xf5fef736f491fde0,0x43e06e510df86940,0xe111c10135a441c6,0xc4601440c006d60,0x32004e0010290120,0x10500100280280,0xc29235102d3046e0,0x4901930200a8e280,0x88603e8111b9e400,0xa801f1a34b94162,0x18506690412800a4,0x1086ad0214b146a2,0x4b82ea33348c7282,0xc2a2122100d90020,0xaaae4c503095e560,0xedf9fe6025b163e6,0x9ceb267635b92ee0,0x4004007000006680,0x2cdedae0ccb925e4,0x41881b1194a80542,0xac74c0a31c996ce4,0x4040a60084101540,0x4a962cf200996aa4,0x8980340210840442,0x860001000042a0,0xa242601485a040,0x949cb053149854e2,0x4814501c54204262,0x802cf00615f061a2,0x2a968a4120ac0960,0x2d3bddb3aca0d7e0,0xfd7ef6faf5b92fe2,0xafbc073034a957e2,0xd144f77835b96f42,0x8ce32a80289964e0,0xeffbddea3cbde7e2,0xf6b76eefb5b9ede6,0x1868ef701c883162,0x2100204000042c0,0x0,0x80010000000042,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100800000880000,0x0,0x0,0x0,0x300000102000,0x0,0x0,0x20001894c0000a0,0x20200001200000,0x0,0x0,0x0,0x0,0x0,0x14000880000,0x20810000400,0x0,0x0,0x0,0x83486f40501867a2,0xfd1627bb1c9c4d62,0xf453ca2971b02de6,0x2dfed7d1fcb827a6,0x9fdb976b35bc6fe2,0xfbedef6a75fd7fe6,0x1abe1f7575b8fee2,0x42c3d33811fc0462,0x22579ebb35e82be4,0x6d375df234d9cfe4,0xf5fdffe6b1bdefc2,0xfac47d9615b96fe2 },
	{ 0xfffffffff6fff7ee,0xfbffffffeedff7ee,0xdfffffbf76f7f7ee,0xfffeff7bfef674ee,0xfffff7f7b4fff7ee,0xbfbeff77f697f6ee,0xffbff77ff4f6f7ee,0xdfffee5fc6f677ee,0xeafedddbec7e766c,0xd4f77efbae5fd6ee,0xf9feff6ab6d674ee,0xfbbffffffef6f6ec,0xefffffbda4f6f4ee,0xbfdcfbdb34b776ee,0x34b7fdefd6b6f3ee,0xfffefff6fe36d4ee,0xfffe3f7ff6d6f1ee,0xeffeffbffe7ff5ee,0xefffffdffefff6ee,0xbabbeffffef7f6ee,0xe57ebf6aeef676ee,0xfffe7f71e6f6f6ee,0xcb8efbfe7c5e55ee,0xfff6fe5bbe76f4ee,0xdbbeeffbfebe76ee,0xf5fff7f8feb2f06c,0xdfbe357344f676ea,0xc9c4531fa43253e0,0x4a667e4f6416d6ee,0xd1f5f17bcd45368,0x7596beea6ad6546e,0xd7feaf63b4f6f6ec,0xd9362fb3e452f466,0x8400201000c1080,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa0ded192632f768,0xd7e7efba296f6ea,0x9b7eed6ae43750a6,0xee9bbf7fb6bff2ee,0xeff6f7bf7636f6ec,0x0,0x0,0xfbdca77ee6d317ec,0xcb8e2f6c1e96736e,0xe9f5e10ef496756e,0x6a5e7e7be4bd50ee,0xbdbfddebb6bc766e,0xdcfdd2edb6f6d6ea,0xeebe3df2e4d577ea,0xf985ffbfcedf55ee,0xdeef3efb66a770ee,0x73b6c6501607f06c,0x7df777ffa6d6f6ee,0xd7b9e553be9434ea,0xd392fb3aec9ef7ea,0xc04bf208267c50e8,0x0,0x0,0x0,0x0,0x0,0x0,0x8cfb7571e47f566c,0xfffcfffafe7ff6ec,0xde9647ebb452766c,0xe8426a0e843d54e0,0xafffffc8e436f6ee,0xbf32cefed6b6d7ec,0xfffdff69fe7ff6ee,0xdfe65de2c637f46e,0x89f2c77aa43e32ac,0x265279cc341653e8,0xa9895d214227642,0x10de57e986bed06e,0x5a1c2db3243ef3ee,0xb9d6db2e9ebe14e6,0x4f6bfe73f47776ee,0x8fafdd183c1ed666,0x55fd5279a292d76a,0xf3befd70ee36d2ee,0xdfffff7ffebff7ee,0xcddf9649961456ea,0x8436ce0224127068,0x44fc7be8943250ec,0xc9025602ac127062,0x7c1e37eb3630d760,0x8ce7e6bf60e714ea,0xe05c9c4282375064,0x4044640801214e0,0x4340bf008040148a,0x429106162c05d4ec,0x81a4d3653444106e,0xb5d10c6a0b65268,0x34b2e7ea247d566e,0xc8341c019c045468,0x9a94426294b4146a,0x881cfe22545c5460,0x69261eb41453f4ec,0x25fc4cebb462516a,0x0,0x9fae7fe0acdc55ee,0xf880cba436b611e6,0x1fdb3a58163677ee,0x7b1ed515ac3e52ee,0x24322064040e3168,0xb2e2d40169c53c8,0x80002900001510a0,0xc0a3824360411428,0x987cb5a345ef026,0x546626400414d00e,0x10ea962ac1614e4,0x9314222400d05268,0x8200668200105068,0x2a1150d1241e7160,0x84233428029410ea,0xa570466a80145022,0x2002000104000,0x8000700080001060,0x2358134000525208,0x16905d1a841e7060,0x81a9166094741002,0x710a5a78200e306a,0x490cff2a147650ee,0x48671a10141470e4,0x88055240134024,0x100820000020,0x100040002000040,0x2154420000160204,0x204eb34ef4226028,0xbb3e4f5326b672ee,0x7cd43efc3476366e,0xc1322f72be9450e8,0x8a0020340144022,0x85830818034d002,0x9ccd850765410e0,0x167056a0484212ee,0x4124ac72001454e8,0x2a080c0084001260,0x40032a32840cd464,0x3cc4d0ae010142a,0x41a03461402026c,0xb044062204905444,0xba20256094345262,0xe48bf0d0d414d6e8,0xeedf7f5d847775ea,0xa3b24ffba87256ac,0xd231a619bc37f666,0xf676b7d4feb3376c,0x81f2080026302a,0x2e5a13c82014166a,0x8008402080027000,0x1130064000022244,0x41803a4060126020,0x9b64033040d4502a,0x48120080120004,0xd04480a1410446c,0x510c32840222f26a,0x8109f1e014001268,0xc2042ba20414446e,0x4955a62f801672ca,0x2610c100140412c8,0xf4b5dee1e43454aa,0x914eb3381436d06a,0xc3b23706c254d0e8,0xa177ca18801650ee,0x288a975ae811566a,0x8cde59a3e251d460,0xc13480602084116a,0x5096181024041422,0x4840000000401080,0x825a1104325004,0x2214120010d008,0xd54ca66880d0506a,0x5bc43d2014227268,0x28f5bbeec4bc56e8,0xa63e12db8876744a,0x37d181001622706a,0xd48770718410526a,0x21763960a8e6f0aa,0xcbd1af2a249c706a,0x2bc7de1794df74e4,0xdefffffb8cfe75ee,0x7fd7ffeaec36d6ae,0x9442c06004410e8,0x1000004000001002,0x0,0x2000810100040042,0x1088000214020004,0x0,0x800000100000c000,0x8020500000041008,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20000200000040,0x0,0x0,0x0,0x0,0x0,0x2200000140000,0x480000040008,0x0,0xd1aaad50a01e52e8,0x5b820a7e94325064,0x9976e60e843f6266,0x6ffff9a867650ea,0x7c3657e8b45676ee,0x9bbcd572a6f6f6e6,0xfab63df32ed4f5e8,0xc5cafbbeecb6d66c,0x6bdf4dd08294d642,0xb6fa5ff7a436d66c,0xa1bdff7afc16726e,0xfb36777de4b7f7ee },
	{ 0xfdfffff7f17dbfff,0xfffffffff1fdbfff,0xfffffffff1fdbfff,0xfffffffff1ffbfdf,0xf5feffed71f7b7fa,0xfffefdf2d1febfda,0xffbfff7f71ff9ffb,0xeddfff3fd17dbbfa,0xafffdfdff09cbfda,0xdef7f6efd1dea7da,0xcf6fa76551fcbec8,0xfdb67fffd1ddbefe,0xffdfffb971ffbffd,0xfe7efed9f1feb7fa,0xdd3fff6bf1fdb7fc,0xfbfffff3517d8ffe,0xfebe3dff717daff9,0xee3fff76517f97f6,0xefffffdf51ffb7da,0xffbbfffb51ffbff7,0xdad335477150b3f0,0x9a9e7d7dd171bfd2,0xabd77bdff17ca7f1,0xeef7dfde517d83d2,0x5296cff351debf5e,0xb455ef7a51f43fda,0x3ba87de351ffa7fe,0xebc4f32c713ca67e,0xeeebb77bf151bfc3,0xffdee7b517d37dd,0xf5f676fe515fa7eb,0xfefe6e707171a755,0xd006a5ac50948ec0,0x8c02b10a5030a240,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x808807105010b6d1,0xadfe57f3717fa3ea,0xc07cefc05075b378,0x5eb3277c515daf7c,0xc9f7c33d715bbfda,0x0,0x0,0xf42ca5d0d13ea4d8,0xd282adc850549fd9,0x83e7f77f51dcb7fa,0xbe57bfdb707fbfd8,0x2dbbc95bf17dbe4e,0xd37fe7fe51fe97db,0x7f3c6f7bd1bd96fa,0xcbcfff3ff1bcbff3,0xcff83a1f71fd3ed8,0x7d9a57d25038bee1,0xfcf7feefd07ebfdd,0xc5a9af7251dd9ed3,0xff97f77d5171affb,0xcfcfbeda71ed275e,0x0,0x0,0x0,0x0,0x0,0x0,0xb5fe75e671bdb6da,0xdbcfef7371fda7ff,0xccba2760205f8ed8,0x8580ca20510010b0,0xfafededf715eb7fa,0xb19acffbd13e9ef8,0xf27cb62d511b27f0,0x8f965d7940b7bdc4,0xeb3eff70513ca7fd,0xbae7b65a51fd8ed1,0x1202070910510ef0,0xbe32eb7051752f47,0x412efce211508e59,0xeb10d33e71dd27fa,0xbee3bf4e51dda7d8,0xefbfec7e5175a7db,0xf776b6fef1ff9ffc,0xfff66f7371d7bfda,0xfbdfefef71bcbffd,0xe3d6d64a514da7fa,0x8b5759523149875a,0x5015d5e051100af2,0xe9486f63011d0e64,0x880c070c411806f2,0x52cbe73a71bf27dc,0x251d1a10517c0a42,0x42265663101c868c,0x914a2700d0240240,0xc2863f7d713f87d0,0xe1d077bb517ca7d0,0x584e0787d1118dec,0x36385fd151262748,0xd8e13678516526eb,0x5f8e2d20d0141ed9,0xcbc034b7519c86fa,0x6838865270d486c0,0x9a9b2fd9507cab50,0xf566d67a71570b7a,0xcfb63f22501caedb,0x84000600200340,0x4043043c90140e8c,0x31c8ef150fc2f5a,0x71db07ea5135b6d0,0x2ba2800005d8250,0x98b4830a51108601,0x2ede0209d10086ce,0x1a091c0a410406c9,0xa0c04010080640,0x81bcaf68d11426d6,0x202020400250040,0x40271870040484,0x802180010100010,0x50201a2411081d0,0x31406572418c2660,0x2263c40d0008484,0x4152800050080600,0x2c050eda603482d0,0x2421000000000644,0x50c0301050148c43,0xc6162573503d8650,0x8220c42250648388,0xfd60e00005186d5,0xa0040800418040,0x60d9d669512496de,0x1cc7d21502136e0,0x4391431841ec0348,0x60121623410c048a,0xb9eef3b71778e5b,0x1062e6445035a240,0x32ab700f17c0f44,0xc020c200109c0200,0xc58ef80041dc1f46,0xa842894e415c1202,0x189156e9704582c0,0x9d20044211110643,0x8a080b024106204a,0x782864010100dc0,0x8a2c0b4251248250,0xb4020fe060942b58,0x10a0052851102a10,0xa0042720710886c0,0xcfbbfb225075bed8,0xaffffff3f1ffaffe,0xb6fb5feaf1acb77e,0x91cf34f851d5bfe9,0x2632b564517d87c8,0xcb80eb7151548fd8,0x4412260840062042,0x8010040040000,0x0,0x6812f122514c8260,0xc050490851040358,0x6642861001248048,0xa8a2c948510cb140,0x85f114a271680282,0x2030580040000554,0x7d8df277512087c3,0x952f388514587c8,0x298949cb51501372,0xf5bcfed271fca7da,0x986fad52d174aff8,0x92906b2441120dc4,0xc563cb3c7114a743,0xaa1657cb313d9e69,0x703241e6501483f0,0x996cad88514da3c0,0xc9082c1011100c87,0x410a0004004a,0x24160401409403c0,0x10904c8410100e80,0x4301f75251248684,0x4822af64513d05ea,0xc1d4d32d51dc17ca,0x696610011134065a,0x69f446e507c8f90,0x9089b350115006e2,0x4514396251b53242,0x9144416c51ec82c0,0x2621258701c3240,0x2fbcdfd1515cbafa,0x18fff67451bc87cb,0xd1ac847091500752,0x0,0x0,0xa04080010100c00,0x0,0x0,0x4801000008,0x0,0x0,0x0,0x0,0x40100002040,0x0,0x480000008040,0x0,0x0,0x0,0x1800040000010,0x204000010000400,0x0,0x0,0x18003001020042,0x0,0x0,0x800010000180000,0x0,0x0,0x4000000200840000,0x21000002040,0x98eccc9511811d0,0x0,0xd28aa77051552e4d,0xcea05376519c85ce,0x4f44212e11c783ca,0x2b34c71951b40dc2,0x342a7f49103caac8,0xf9c5afeb717f8ff8,0xb2127d80511186d8,0xc0a8f42851148eda,0x2dd03c92513c8a52,0xd6b6f7ebd17fb7ea,0x9b7fff6071fdbfd3,0xdf766f77517d8ffa },
	{ 0xfffeffefdaffbfe6,0xfbf6f77edaf7bd66,0xcffbffffdafeff74,0xffffff7bdaffffb6,0xf5f7feefdaf4ff46,0xfbf73ff3daf5bdd6,0xfb97ef3f9af0ff44,0xffffffbfd8fdafe6,0xafffdfdedaf7efe6,0xd4f6f6efd87eaf64,0xd9bfbd7d9af86574,0xffbf7ff89af2bfd6,0xcfe7ef1adaf3bf74,0xdbfe5f5b92f4bf36,0xdf3f5ffbd2fbbff6,0xf9acf7ea9af5af74,0xfebe7fefd2f3ff66,0xa7f7ffffdaf0af26,0x7ffffedada7eedf6,0xffbbcfffdafcf514,0xf3fd177e9af22740,0xfbba7fff9afced46,0xe9dffb7d9afcedd6,0xebf61fde926caf42,0xf7dadffb9871ade4,0xf77ef7efdaf8af76,0x9b7e7d739af4adf0,0xfbd6ff6e9af0ad70,0xed6a7f899af8ade0,0xcfbfdffadab5fdf6,0x7ffffeffdafdfff6,0xbbbefdf3da7cff06,0xff9e7ec7dab0b546,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x15b84ce0c2e5ed40,0xc97cfa2998e4af50,0xdca62d4ad8682466,0x840000000300000,0x80080000000400,0xc3857ba92308534,0xffc9f7eb98b0e576,0xc484190010243522,0xcbfcf60692f0bd64,0xfdd37edfdaf1bd36,0xdf9b7dfbd2b7ad66,0xd4ebe6f49a68ad60,0xd1be0d72dac03556,0xd1b9672a9a642740,0xceebbb9fdae1ed36,0x2abad3489ae02d14,0x8860c6e89864ad10,0x5aa6ade092f06f64,0xf59cf33fd8f0bf60,0xcfcfbebbdaf5bf56,0x0,0x401000000004100,0x0,0x0,0x0,0x0,0xb4ffdfe992b4fde4,0xf9ffd76a9af7bcd6,0xfabd37f3d2b0f566,0xf1506a1080b0ad02,0xef77bfdfd2bffdf4,0xfeeb6f739af6bd76,0xb773f77fdae9ad16,0x286a52c903dad44,0xcbc7f72bd2b6a544,0x4bc40e539ab02d40,0x379ad4b290a28460,0x4437fa6ed8e8bf64,0x2144ec609224ac44,0xebd5cb02daf6a504,0xeedf3effdafdbf76,0x7eadcffbdaf4f566,0xf5e3d6eb98e8ed36,0xd33e7d73da34ade2,0xdbee7f36daacad16,0xfffebffb9afcad74,0xf7ecd4a9afd2c42,0xd4befcfadab0acf2,0x454e7961d8e82416,0xcfb0377d98f4bf70,0xdfdffeba9afefd74,0xcadfbb439a64bd64,0x2471742892244500,0x100455e992254514,0xfee72f6fd877c560,0xc752d73892a5dd10,0x4b9fde51dae02d60,0x369af5f1d0a4cd70,0x394454a1d0a43420,0x574e7d47d8a1b5a0,0x8106022480200440,0xe591b4b92402f64,0x573dcef098e6ad70,0x100280d490a40050,0x39de41609002a9f4,0xe9f497099ac0ad44,0x86233c0c860a440,0x26ff5fe0d0b0b5f4,0x7161777a1842edb2,0x88b607729860e534,0xf293fb7e9a76bda0,0x40462a32c020bd40,0xc13c34090a02470,0x1413ce6490d09b00,0xd5af7d63daf1f564,0x968423e09200a5b0,0x8950a2008aa4c500,0x81d835000020a564,0x1003031298a16500,0xf04c470290246550,0x40b33572082c2540,0x848e904c92800560,0x61cfdbf99801a5b2,0x75ae5f2298c0bd40,0x8009b74090f01a40,0x33be6f659ae1c550,0xebc4c1d0d228a4d2,0x814100080082004,0x0,0x100800002004,0x200100200040,0x8000200000200400,0xd4a120088182400,0xcf9fdff39af92574,0x9c545fe01a682514,0x8009a41090110c10,0x4800240000,0x83c28c03c0702202,0x8010014288802122,0xb04c489000a814,0xc1e9adf0d0b0a580,0x8b2eee429000bd74,0x89511f8892a02504,0x2b5505429ae40400,0x1814000080002040,0x91780d0818408510,0x4ab22413d2882d40,0x8fc7f317dae9ad30,0xf777ce19d8e42e40,0xfc327ffb9a7c6774,0xd4ff9f479a66f546,0x76ae7d72d8f0bfc0,0x9de56c89ae0b504,0xa047f81198a42140,0x4080050090000400,0x2008000300000,0x29400d6080612606,0x5884602a00348540,0x842125090202440,0x211648c000202546,0x4907c8890310c40,0x81002f228a002542,0x488771909020ed24,0x49521e101082fd14,0xb220e9392a6ac04,0x34b60de29808f5a6,0xd960ae61d8f46510,0xfabc6974d0a0ed74,0xe443eb028ab0ad46,0xb05a0190102400,0x1000480012000500,0x91f1af1990f42724,0x8834047080092406,0xc054182200000560,0x207c86c310302414,0x90826ca08000ac40,0xf1bc24430220b922,0x44466c20d840a910,0xbd2b13890a1acc4,0x8e820a0092242524,0x16914989d201b400,0x198702892020110,0xc600304acaa02400,0x4a444b26d2a0a514,0xe492b4aa9af4a510,0xcabaef49d8f4b5a0,0x5554b6f2d0403f20,0xc121a60090b081f0,0x8400000004100,0x882120042002500,0x80180000040,0x1000800000000400,0x0,0x0,0x0,0x2100000010000020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4100000010000400,0x4020040080002400,0x840000090000000,0x0,0x0,0x0,0x20200080000040,0x4000400000008080,0x0,0x0,0xc010008000400400,0x9d7bafffd234fd56,0xdadfaf1f9ab2bd16,0x8573c7089232add2,0xa3e8c50092101d04,0xb9bad76b92f43560,0xc9eebfa298e12776,0xfbaebf719af4bd74,0xcdd1b616d2b10d60,0x2a5e99d8daa4ad52,0x9cb34febdab3bd22,0xf1fdefefd2b5a5a6,0xffbe7df4d2bdadc4 },
	{ 0xffffffff7fdeff73,0xfffffffffffe7f73,0xffdffffffdfe7f73,0xffffdfffdfdeff73,0x7cffffe7effe7f71,0xbbbefff34dfeff73,0xffbdff6fcf7eff63,0xcf56fef3ddfaff73,0xef9ddfdbdfdee773,0xfcf9dfe47dde7f63,0xdfffff738f7e7f63,0xfb9c7f55cfdaff63,0xdfceff2d4d4e6963,0x6f7fffd1cd5e7f63,0xf63b5deb2f4e6d63,0xbbcff2794cda6962,0xdeb27f76cfea7d63,0xffdff3afdffe7f73,0xfffefe1fcfdeff63,0x9fbb7ffacfdeef63,0xf5ec77edcdde6f73,0x7f667d7dcb4ef571,0xe9f6f76ebf8e7d73,0x6af796db4dde7740,0xbefbeddbdf7e7561,0xd7fefe79cd7aed73,0xef1e77e24feee563,0xeaf6e36605587563,0xdedabe1b895c7762,0x67fb4fd24dc87d62,0xb5f7f6e84fce6771,0xbf7a6d722f6e5773,0xfbbdeffdff4e7f62,0x1548204b8106560,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x403302549880000,0xdbf8b7d8c944f603,0x0,0x4000040000000600,0x221086c06824a460,0x37bb5ffbcdecfd62,0xe745a7fbcdd86f61,0x9000372429024d62,0xd017e3a24a6a7560,0x2cf67fd24e5a6f62,0x9fdfddba4dcaff62,0xf42eb4f504d82740,0xbf3e3d300f5e6760,0xebb7f736ed1eff61,0xeefabf561d7a4f63,0xfa9cdf7fcdac6d63,0x8ed33fe0294e6d62,0xb79aef31efce7762,0xd70cfe6cee4c7761,0xcddaf6930fe8ff62,0x0,0x0,0x0,0x0,0x0,0x9f295f1a9d3e6f63,0xf47f5ffb7f5c7d73,0xd9bdbee65fc87d73,0xfaa03717198e4763,0xc004db248a407472,0xeffebfdffffaff73,0xefbdd7ebadfaef70,0xd301b7a1d1007141,0x96b2f56cc97c7f63,0xebddf6e779deff60,0xa9490e516dce7920,0x84130a6009624140,0xddd25ee0450af742,0x6dfe65b20dcc4f62,0xa944f72c050caf61,0x4fdbba9f64daef62,0x20c440808420541,0x55c53e690752ed70,0xdaecbd424efe6d71,0xf7b6fff7cb7e7f72,0xcdde9f3bef4aed62,0xfffeefdfefeeff63,0xd63517681f264760,0xc9afad66fdcce561,0x98892d3219501f20,0xdfffffbeefeeff73,0xebeff7dbcadc6732,0x37fb57e949c8f762,0x910ad670194a6553,0xdafaff7f5f5eff73,0xc8e4d3ee0c6e7962,0x7adf3bcd0fce5d73,0x240a2df207986d62,0x4251773a2b5a6d61,0x569e7b579eec6560,0xab7ff22bec4efc63,0x6ed39fceaf04e462,0x56655568ed422960,0x4244565204002441,0x406e74484040a460,0xf004693e0f4e6522,0xebcb9e7688467563,0x3fbe4d1b1c7a4d63,0x749652ed0a1e7762,0xa001853001022541,0xf400754a01080d60,0x48430e00050a4062,0x6b7f03d0cd4afd61,0x61e494cc4d286460,0x88498202080a7442,0x4012044002040,0x81c0a103809ca120,0x46100001084000,0x1489002003020540,0x2124b622ca0c6540,0xaab0a7495d006d62,0xc990939f49427442,0x28d75b9d5b087d70,0xbd204932842a7761,0x9271a240498e6102,0x7284b22c2d0c6461,0x8e581620b4e2560,0xebeb9e1f27caf562,0x4004c8b10c14a140,0x144800000140,0x2a02004000140,0x260408484000,0x828a010488a502,0x21be4f422f0c6763,0x301406620018a760,0x823a6d7107067761,0xd8b4255209446223,0x94642090848e020,0x2b6e4b58cb0e7560,0x520002001020440,0x4ca6496087007762,0x12200764210a0360,0x4c19d1808484c22,0xa7c115205405442,0x812c30004203441,0x142400000080700,0xfebe7ffeedde7f73,0xe9d7f7ffefce7f73,0xe9f7fede6f5e4573,0x37bf5fed074a7773,0xd9c9f778c50a7143,0xa3220e4018004f70,0x8ac0622ecf407640,0xe6f239466ca86162,0xcf9de971c95e6762,0x4c1070200f027500,0x340640084002140,0xdbbeaf3e8ade6560,0xc9daf219ce2e6761,0xab3e185b880c6661,0xb011014200885622,0x8410014000042421,0x5bafe6230708e761,0x809080b08040540,0xab1edfc387986d61,0x14907ce207285c71,0x882d272110404740,0xcbb067748d746f60,0xc5f2faa949de7f63,0x80c7b4105884532,0x880a3b8107084542,0x804a8b4205026960,0x8b941f1ccd026462,0x10902000020140,0x6908032002085d20,0x42820d030c244140,0x9400814030000960,0xef8616605d2ead60,0x8b2bb8c2cd467541,0x10530642050ca561,0xc98a0510c104741,0x3a226800884560,0x2ab40da019487941,0xcb14e12242886740,0xadc21a100b404e70,0x631e4d5b350c3f60,0x80100ee80d486122,0xdfa8ed324f406743,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2000002000000440,0x500000000140,0x0,0x0,0x0,0x4084040000040820,0x9ceca5738bd86f42,0xab12753407cc4761,0xc8d3be9e019aef63,0x8b5f8b5a08d85f63,0x1ce765a06d5e5570,0xf5ccb5618aee6563,0xcd8e35590f1e5760,0xcdcefe77ed1e7d70,0xffffbff6df6c7f73,0x37bbedeb1fce6f62,0xfdf7b4fdcb9e7d60,0x47306c534f88ed40 },
	{ 0xffffffffdffef6e8,0xfbffff7fdffeffe9,0xffffffffffff7fa8,0xffffffffffbffda8,0xfdffffefdf9ffde9,0xff2eff73daffbca8,0xddeee776df9e7ca9,0x9ff7febfd7beffa9,0xafdecfdbf7be7f29,0xf7f6fff0febe7de9,0xdfffff73f7bfbe68,0xffbe7ff5dbfe6fa9,0xdfffff7fffbefca1,0xbf7eff5adfdefea1,0x9ebf7767dfbe3f68,0xbbfea772b7de7fa0,0xeebebfeffaecbfa9,0x6bfef7bed7bf7fa1,0xbbdefedf7fb76ce8,0xbfbffffbfede7fa9,0xfffdbffefbfebf68,0x176e3d6c7eda4d28,0xc0b4d16a629e2ce0,0x2aff9bebfe9f7cc0,0xbfbbdffbffbfffa8,0xf4befefddf9f7ca1,0x6dee7d625c9f3ee0,0xf3d4ef5cffbf3ee0,0xdef3bfd11bbefd20,0xf124da35e9d2c20,0xd1c7bfe3d39e7ec0,0x291ef7304b8d3c28,0xfdb4ff7bffbeed28,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x21edd8c24616ada0,0x33125b97139e3c60,0x93352e6852862ca0,0xea96bf7a5abefea0,0x83fbf3625fbe3f80,0xabe95adb838fbf68,0x32bc47690a962b00,0xc4d4be4b55b63620,0x82a609e0458416c0,0xa399c70703b6ee21,0xc2625f65ce9f1da0,0x2284e40c4823e00,0x5170afeeeb0e3ca9,0xbfbbaf5657dfbfa0,0xf3a0667e599e54a0,0x907d20232c48d40,0x0,0x0,0x8000040000080400,0x0,0x6000000060000,0xef7fdfdaffbefda1,0x3cbef3eddf9fbda0,0xf9f9f777ffbefda0,0xdab62771cefead20,0xcfca6f31d61c3480,0xfede3f9e7b9f7fe0,0x87aa5ffb56de3ea0,0xd454ac2c1ab6ad28,0xffb67f7d769c7ea0,0xc8cdfb7b56bf7da0,0x4ddb1e9e8adf3c01,0x145d4fc0420c3488,0xd8f082400aec1600,0x572e75f0179e7d20,0xfb878b36579c2e80,0xeecbb6b31abbbf20,0x255e5ed1c5be3c21,0xd17f36ee1b9e2ea0,0x9ffcae31c03e3da0,0xefbbff7ec7be7fe0,0xc4d3b3335bae7ca0,0xbb66cfdbd3d63e69,0xf6bf5fecbede3e21,0x412eb7095ca63ca8,0xdefe3eab53ae2d29,0x88c3ee2a16127c29,0xab3f5d53d38e7c81,0x93ac4ea16923d20,0x5084d0c414907c41,0xdabcaf7ed79efca0,0xcbcdf3ba9bd6af60,0x4657df5642d67e20,0xff8ec5b6f3fcafc0,0xffffbf7b7f9e3da9,0xdfdeed7657de3ce1,0x8904e33a52a23d20,0xbfdabe5d87b7fda0,0xa632d69303b73d00,0x5050a7f411971860,0x90272182261400,0x1a81cb030293bb00,0x8040081200185400,0x8eafdf7bbabfcf60,0x5477b6e87b962ea0,0xce38356347953ee0,0x220840020202a820,0xc823000000e0400,0x84c4082261400,0x7422f0e80236bd01,0xd4a6856012843c20,0xa332250486906200,0x8023aa2110852c00,0x0,0x918c13b141c40a0,0x8078e7004a1e44a0,0x260640420c0480,0x414ef318508c0da0,0x222e50805a840080,0x8284b6a9b15a801,0x32402628522c2400,0x8804f00c07dc2400,0x0,0x1124002800600,0x20280c8b82162c20,0xb18c0c5a55106c08,0xc220251000b12d00,0x180a01202061420,0xe072b010c7be2c80,0x85ad4c404a225400,0x4032926144b60600,0x8010240022104400,0x8000250003000400,0x1080001800000,0x21000120000,0x4a04240860280,0x114633208a842800,0x91100a14038e0000,0x8543e02807902200,0x200542528c1080,0x120424050062020,0xe944426301f82c28,0x8e14030402926c20,0xc351fba61b9e1e40,0xeecbbfeb7eb67c20,0xb7bbeffb5a9e3fc0,0xf0bc9f7ddb9f2c60,0xa720246003823400,0xc395d204028e3c00,0x88576e86629eedc0,0xc5ee48215d1e3c00,0x934d41a0b803e40,0xc1a0d91428434a0,0x49b1e73056962220,0xea4e025558fc6480,0xa23e4e805a963ce0,0x11cc96cb0fd46dc0,0xa033ad11c3b24ca8,0x7910361852de0ca0,0x8e20480002a42680,0x23d38fc15eb72d20,0x14f48ee84e063d80,0x808a60224112c00,0x7a942f07029e3ca0,0xc6d5a816161e7c08,0x2936de56ea962ca0,0x800203ed468c2800,0x210e020000a22420,0x0,0x5154726201883d08,0x15190c810a62c00,0x182490000980440,0x184c968288962c20,0xe042377c1292bc20,0xc182f13a129f1d20,0x8a69800108e2420,0x702441604a080d00,0xe1bcb261ce963f00,0xcaa025c0e38684a0,0x88904364620ab480,0x8d01b01833c2c80,0xa79bcfdf3e9e3c40,0xd1d7d67e43dffda9,0xd801af401796e800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x140000041000,0x0,0x300000120000,0x0,0x0,0x0,0x0,0x0,0x0,0x4000800040000400,0xd73f237846d6eac0,0x988c2b64743e0420,0xc9c59a98dfb65c08,0x77fadfdacaafbd20,0x949e83e39bbb4d68,0xe8285780538a6420,0x46bf34f082823e20,0xe9f7ff7edfdeff20,0xeededfd39bb67f28,0x66af67bbc3bf2da0,0xb7feff79da8fbc28,0xcf34bf36e5de7ea0 },
	{ 0xdffffffff7be6f57,0xffdffffff7bc2f7f,0xfffffffff7be6fff,0xffff7ffff7be6b5e,0xfcfffefff7966fff,0x9ffebf73f7ae6fdf,0xffbaffeff79c2fc7,0xffdffe3fd7b46f77,0xbbdf5f5ff6b66b5e,0xfff5fff8f73e2fff,0xdb5f3fe8f7b66be6,0xfe8f3fffd6ae63cf,0xcffffe1ed7942b67,0xeffedfdbf6bc6f77,0xb5b7feeff2bc67f6,0xf5acbe78d6902ff7,0xfe9e2f77f79c6776,0xe7fff7ffd7bc67ff,0x495fffd5d286677f,0xf7bb7ff2f7b66ffe,0xd1bdf67bd7242f57,0x9f8c4d6fd7a4237e,0x88b1d38893a06362,0x4fdf3fdbd3866f7f,0x5f5dfffab2ac235e,0xf7f7f764d696275e,0x82a67141d6140b16,0xd38e7b6fd794637e,0x80c1b970808003c2,0x8efe4f6b97942976,0x35d4e7ebf30443a6,0xbebf7d75b3b46bdf,0xb0846568d48c2b4a,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400200001000002,0x2a1fd9c8d3946363,0x14bb7ee391b4214f,0x997ebf7bf6bc6f7e,0xcf1ae75ff614231e,0xc3e7b73ef7ac4b4e,0xaef23efbf7b42b5e,0x928a051aa280254c,0x9765f26cd304272e,0x5ba83d51d6042b3f,0xfbc6f72cb3a2274e,0xeefbbffff79c27ff,0x2df84dc3b2042602,0x11db36eed6b62346,0x89492dc1b3046116,0xf905ff7ff7be6fdf,0x4c004b1000100,0x0,0x0,0x0,0x0,0x0,0xef5cfed7f7902ade,0xb4fa0ffff7b62f37,0x997ee60393002f47,0xc092bf70d6202b66,0xefdffefff7be6bdf,0x2ddfee7bf7b4277e,0x7793ef73d6142f71,0x8378256ed2342246,0xfa86256ed684296f,0xa300fbaa77002216,0x5ee65e5bb2a4670f,0xb2191a72d7102306,0xc0f4d612b6166740,0xe2bcecdbf6906b26,0xbb16fb3eb7b8634f,0xcc4eb6c1d206636f,0x8abefbfd9684234c,0x51f796faf7806f5e,0xba722d62d2b02acf,0xdc022a6a92164304,0xdce38f98b1242f43,0x797ebddcd7966357,0xf6f4f6f7f79e0b1e,0xd1ccf600c0102157,0x64b12161802c234c,0xe6d6ea1af610236d,0xbc9c9c893146725,0x54e9712a92162541,0x810c500ad4040804,0xca8aa670929623cf,0xe2c3df88d5242626,0xedff9fdfd792674f,0xcfbbffbfd3966f7e,0xd97cff7bd6b0277e,0xffa67564d7946376,0xc9c0f17bf6962223,0x4052361ae6006304,0x19bc4f92f2842706,0x9e50ae53b2942a7b,0xaae4d629216631c,0x9bdd8725f2242717,0x18911e9292002127,0x8ffcdf0db6b0265f,0xb4355ae290142f1f,0xbfae3f20c71623dc,0xc00010000200,0x89c2e8008210225d,0xa002010041000002,0xb0ba52401620014c,0x8d76ef10f6b46342,0x7294028e9014021c,0xa87e6b169210271d,0x844484012020104,0x88820f2156102604,0x4040a160d4902644,0x72142d61d2842552,0x44001854000802,0x200c0040000214,0x86030e28d5040202,0x88348640d1000316,0x1800a4840802442,0x8166aa0ac3022b12,0xe4200008180005a,0x91000c531602206d,0x4184006890206070,0x60b82c00c680410e,0x210020010000104,0x9402b202d600610c,0x536536092144324,0x8450c67992902264,0x2010000000002400,0xa890402080900250,0x0,0x40000810000800,0x202004c010800860,0x9406dc3d40c0a94,0x38082d2dc620025c,0x146a6229210264e,0x1e5d150b0042a83,0x34d04c28f4106a00,0xa100852a80102704,0x601e2420c2142304,0x8946fa01d194230f,0x60f63a7992002f07,0xb382cfa89304036c,0xd76ef54bb31c630f,0x92a0e2183202147,0xd3007602c2a02016,0xe7fb1fcbb78063d7,0x451cec29130262a,0x2024a27952002744,0x400844a2d4100850,0xf9706952f7900146,0xaa071a8392182b4c,0xaf9fce5d97b62bf7,0x8cd05660b41002c4,0x843ae531d6b00346,0x2100252010042245,0x4140061092100a4c,0xbf9fdfdfd6b02ffc,0x74bd54afb2844bde,0x37e650203b40346,0x9ba4af6cd6040b72,0x48070f82d0002a68,0x48177cc3d6942706,0x34bb7e689190093e,0x41048648c0802102,0x820002000004,0x402212008200240c,0xa8e0943b1104104,0x2a610292004b00,0x4228562292106302,0x1528074102122a02,0x108010000100,0x43de134b91a46143,0x864a8e5292142371,0x8002026040802412,0x930a2d4054042e04,0x900c220c4840302,0x49cb12d093844346,0x8002000000000a00,0x2190566ab0800b42,0xd62e2502c23427cc,0x0,0x0,0x0,0x0,0x0,0x0,0x4000000000800040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10002000000,0x0,0x100010044000000,0x0,0x0,0x200000480000040,0x0,0x0,0x0,0x0,0x102600010000a04,0x0,0x0,0x0,0x97ada523d70a2257,0xcb0832b0d7142a0d,0x4fd7bf88f7120356,0x2eff54cb93962762,0x203160609180404a,0x59c05402b2042320,0x8a0a0f34c7000306,0xa3deef2ad78c27f7,0xcb961b0fd0802b07,0xa6b3ddfaf7866e57,0xfb77f7fff7b66fde,0xfffefd77d7bc2756 },
	{ 0xffaefffbfbfad7fc,0xfffffffffbfe7fff,0xffdfffffebfff7be,0xfffffffdfbfc733d,0xfdffffebfbfe7737,0xddfebd67ebdef7be,0xfffefffffbfff7ff,0xebfffe3bdbdc493b,0xebcedddffba87b5f,0xf6fdffe8ebf45f3f,0xfdefff73fbfaf7fe,0xdfbf7ffffbbff7eb,0xffc7fffffbb4f3bf,0xefdf5fd7fbebf7bf,0x35b5d7e9fbbcd377,0xd957ecf2d9b157bf,0xfebebff7ebb5d37a,0xebfeffffcbfdf77f,0xabdeffdffbe25b7f,0xfda7ffffebdcffff,0xfbfdbe7ffbb87fff,0xf7fe7ff5fbfed3be,0xcbfdff3ffbf5ff7f,0x8eff3e61dbb5d337,0x9b9adffbcbb17f3f,0xf5fffffeebb5d777,0x4d1e7f73ebd7473e,0xcb84ff2ee9e4f1e9,0xaed7fa08dbb0c17e,0xd3becdebeabc47fc,0xf5d3fefbf194ff2b,0xfffeff73fbfe73ff,0x240000000208,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x81000000800010,0x4cc216008bc46a20,0x7fdebe9febbe537d,0x9ebb5fdeebf0f39b,0xf1fd27305b00d368,0xdfde7f76fbb67b7b,0xcb97fb77cbd0577f,0xc1d25b10d9a05241,0x3ffecfe9dbbc77ed,0x50e156685b804628,0xc39c3560c3b05309,0xcbd4ff6ce0b41338,0xca66ffffcbd957ab,0x839e4e23cb3052e8,0x50c557fbc93b732a,0xcdbeef72dbb8d37e,0xffae67fbdbf1737b,0x0,0x0,0x0,0x0,0x0,0x0,0xa3fddddafbf45b2d,0xf7b357eae8f5fbeb,0xf0e2a77e8bec7325,0x7eff3f7ffbf053db,0xefffffffcbff73ff,0xb5f1fcbebd64f34,0x808a4fffc3b01b5d,0xf9fd9ffbfbbe7b7f,0xfebe6ffffbb8d37f,0xe3b1f31aebdc533a,0x4ef8bc51d9c4532f,0x303c4f91d1805338,0xccff7fdedbf4333b,0x4fbe1dd2f10051b8,0xeb54eb1e7b9163be,0xa97fb0db89b47bac,0x3fafccdafbb87be8,0xfefdf6fffb907ffe,0x9f9cdfd1dab073fc,0xefbe777fdbbcfff6,0xcfeeffbedbd5433f,0xa17b5ddb891c5362,0xfcfcd6fdfbf9333b,0xef2efff3dbb5fbba,0xcfa43770fb9c6de3,0xc8a3fb39c9f8372c,0xef48de5e8bb2f3fc,0x342b7fc58b94d93c,0xdba4e5c279c3033b,0xcad8176a83b44be2,0xa9f3dff7f9b5d3ef,0x2b7dfe73c1927b5e,0xb4ba6f6ff93873ee,0xfdffff7cfbf4f37e,0xf7aefe4efbb257e3,0xebbdfbeffbb4df3e,0xcfdffe79fbf8dffe,0xba9e8dfecbf477fc,0xdcd2c66889b64236,0x1aa47d61e8205328,0x9312834c8848031a,0x8cc210c18924436a,0x3b08050be8180902,0x86903427100015a,0x92b82f5009d2037d,0x5d36f3260b887366,0x2146000888014202,0x4b0100a03200,0xf49f7feedb305dfe,0x116d41426080413a,0x5a9c6a3ccaa46b38,0x80438869c880d094,0xa84f3cc8ea802176,0x2090438d89a00250,0xc000706168400260,0xf20029448bb10b6a,0x100001000000110,0xa5089c808a0c310,0x9291056888801332,0xc81e27e0ea891723,0xaa0246269404120,0x80800080000004,0x4204060048001308,0x845000a0c0201120,0xb41e36e889bc066d,0xb8c77e2f0b057b8,0xc100410889104a00,0x25a325b090463fa,0x8804be009600308,0x102c08820020030e,0x402a296000800523,0x400004000004080,0x4100900000000328,0x89084713d1a072b0,0x4a074408130530c,0x88071678828453a5,0x5cae0e74d9e05b69,0xd5c80682304328,0x87380189c01b22,0x10048001000060,0x2000220169800001,0x4886016240041728,0x741d810d8944129,0xb4a1bc240a052a8,0xb6b0dfc8790df3aa,0xd1fc76fb6bf973eb,0xcf9e3d72e3b45b27,0xa9a4763febb27739,0x2e7eb650d9f01369,0x3f9a0d1ae9685b38,0xa467722dc99143af,0x289e0430cb3352a2,0x8300ef25dbd05320,0xdfdabec78bb55377,0xe48804328940526a,0x181108e2e1901010,0x2b0a6540988c370,0xc914073861045720,0x904dd0088805303,0x4192df92c0a01b1c,0xf4b4f771f9701916,0xfbfeff7bd9937f7e,0xc2165bf449b27309,0x81d6e768d190130b,0x95d5c528bf9332f,0x2127248094b0162,0xd141b7780980437b,0x202008006000,0x4080100040001208,0x246a24162900214,0xa82001a1e0001200,0x124200000d200,0x6406747cd1a01234,0xf006da46e134c364,0xa4e2263089887062,0x20a000a80300118,0xd4053080cb207208,0x15e3410538640a0,0xd285873e612043ec,0x4957b3618ba453e1,0x243c4890d0800334,0x109456d78914576c,0x0,0x0,0x0,0x6000000000108,0x0,0x0,0x8000100000800001,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x402a0003820000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1202106001800121,0xcf92277fcbc57127,0xe5c4c128e9d0772c,0x6a90d88aca885211,0x24a0556a8091133a,0xe967af38c9d6316e,0xdabe2ffee9be7bef,0xc24ffb6ec93c7b2e,0xfa7e9fdbd9b07fee,0xf7fb5feb8bfedffa,0xd4fdbf7cfbbc7ffe,0xfab67feaeb98d16c },
	{ 0xeffcff7fabdfe3fa,0xfffffffeebd7fbfa,0xbffffffbebdfe3fa,0xdbfffffbebcfcbee,0xb5fd5febab5ff37a,0xbfbefff3cbdfdbf6,0xffbefff7cbdff3de,0xcfc2be1dc184c33a,0xeefecbc0abc7e27a,0x50bef7ebc3d5e39a,0xdbfcf6638bcff11a,0xfa9d3f7eebd7e96e,0xfdffff3fe3dfcbfa,0xeeffffdfcbd7837e,0xbfff5f62ebd7e37a,0x9177b77c83d4f33a,0xfe973f77cbd7f366,0xa7f7b73fc3d5e3e2,0x6fffffdbcbd7e3ee,0xb6fb6feba3c5d33e,0xfefdbf5fcbcde36a,0xebbe3f702b45c32a,0xcbdfffff8bdfdbfa,0xe8767e7fc1c7e37a,0xcf9fd5fbcac5a33a,0xd477f666abc5d38a,0xef3eff72c9cff37e,0x8b912b0ec3d5c016,0xefc7be68e185cb3a,0x2bb6694a81d4c31a,0xfffdfeffebddc37a,0x132087004081c008,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x454ce00000c0c02a,0x689fbfd98ac4d32e,0x87bbdd9e88c7e36a,0xf1ffffffe3c7f36a,0x6bb675f0e245e372,0x8bd3f36a8b4ce32a,0x3edd7fd183cdd32a,0x49a5fbfe1c5ca2a,0x36d252e861c5c068,0x44422840010cc122,0xdba2ff3ee3c5d22e,0x6a5632e201c7826a,0x8de94d7903c6c35a,0x14b6c6308354c162,0xdbbcff32c9dce2be,0xff162f4ec1d4e1da,0x0,0x0,0x0,0x0,0x0,0x0,0xeffefed68345ea1e,0xbdb3dff38b8fd25a,0x99fafe58e3d6c222,0xf2962d77a1c5d12e,0xc7d4e336a8c6eb6a,0x95e5efd88cdc07a,0xffff7fe18bcffa5e,0xd32fbd638354f33a,0xfffe7ffbebd5c39a,0xe984f70fab55db12,0x4f5615198186c306,0x959f686aa3d6db0a,0xf6f7fe6febd5fb7e,0xd2a43060814de062,0xebbc7742cbccc38e,0xae02b8038184c166,0x8fbeebd3c3c5eb3a,0xb0c50d600147c382,0xcb9eef7383c5cb7e,0xfb9cdbbd83d7db7a,0xcbdff786a3d5cb7e,0x87e7de5283d4db2a,0xf4f6d7eacbcdf132,0xfbfde7fb8bd7fbee,0xd38623504844c33a,0x8ca20b092304c102,0x489c5dc1e246c256,0x943a7ecb6944c3aa,0x140d7c20104c32a,0xca32876e09d4c31a,0xcf579f0dc9d5ca9a,0xab1fdbd2cbc5c28e,0xb7bf4ffbcad5f96e,0x89fd7e7fcbcff33a,0x7b9e3ff2c3dce32a,0xcbd7f6e3a3cec31a,0xa573e58c9c4c09e,0x8ffa069ec144e322,0x7c7e76e8a345f162,0x29563f3102858122,0xcac043268bc7d2ca,0xa26aa04103c4c312,0x8001006088004002,0x34ccb6c8c8d5c80a,0x89eca72009d1c3a2,0xf11c537883c7f21a,0x200000201000080,0xa40841418140c10e,0xb01100208884a000,0x2040424000c0800a,0x969c3f6c48d58362,0xf044ef270354c352,0x2a43d948810ceb80,0x27802cab4bccc09a,0x84ac70600945c028,0xdb146f7dcb5de3ca,0x480802000408202,0x840032c0145c200,0x21a2be68185d20a,0xd21c80218004a20a,0x1a8000000102400a,0x14cd0204108c0a0,0x60981ca06201c202,0x208000000000102,0x104a40048002,0x4d407601a283d02a,0xfbdce33eabc5ebfa,0x8013942e2b47c0c2,0xe560d2ac80ec012,0x142108201018208,0x5022056020d4c350,0x0,0x454200a0046820a,0x1420802a0444000,0x11026040468040,0x9000006080022040,0xf4906840498dc000,0x480822a88c4c100,0x4df211128354a23a,0x4c416630306c152,0x8c040880539060,0xa84062089044302,0x8d012000a040202,0x29dfb6cbc987e34a,0xf7ffdffbabd7ebfe,0xf7ffffffcbdff17a,0xffdefd79abd5eb7a,0x88c4e3038b03f126,0xaac71fc88bc4e382,0x30814f718186d150,0x76e772f183c4c23e,0x4886756a08c4835a,0x200ccb6c01c5c022,0xa051a02280049162,0x8292ca000147c102,0x91736eca09c1c23a,0x1934e4308381c32a,0xec886421c1d79060,0x1c5821548068052,0xaf36cfdb82d4d33a,0x4b27eaf83dff20a,0xc8aff560898ff07a,0xfe1e5f34e387e27a,0xc7eed99b80d7d156,0xa91a5a030154c2b2,0x4818038044400c,0x8f0c070a8156c060,0xce02231f820ed026,0xc912dc7a83c4da3e,0x6a5ec64882c3ea22,0x2004f00c288c80a,0x120c004209060002,0x7404416a4ac4802a,0xfb5deb3be384c3b8,0x2c4600000305c202,0x1a8cbd1a0dcd10a,0x752c7233c15491fa,0x21107c708087d142,0x88f0af22c180e32a,0x2f42a3408146d30a,0x889a042b01448342,0x945636a20b04419a,0x100301008000102,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1002000440000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8ba3f77e8bd7f17a,0xcf6af71c83c4d34a,0xabeecfdbcbd5d37a,0xbebf5cf38345ebea,0xd563a26a4b55c00a,0xfe962f7fcbcde3fa,0xedf6eb178bc7a2ca,0xaf1d5ebf83c7e1da,0x92a7c7faa3d4d33a,0xf3eadffb8bcdfb3e,0xe31a3d78cbc4e2a2 },
	{ 0xfffefffbcfdbff3e,0xffdebfffcfdbfea6,0xddfbfffbcfdb7efe,0xef7effffcfdbfefe,0xf5fffff6cfc3fbf2,0xfede7f62c7cb7e36,0xffbffffec7db7fe2,0xcfdfbf334cd2fea4,0xeb9fdfda4fdbfea2,0x9cfed7e74fc37ef2,0xdfffff7bcfdafee6,0xfbb73fdfcfcaffb8,0xeff7ffbacfdb7bf4,0x8b9adf5a4ddb7e60,0xbdbe7fff87cbfeca,0xf96cf77ecfdb3fe0,0xff9e7fffcfd3feb2,0x41c7c575cdc27e60,0xbbffffdfcfdbffb6,0x97f9fffacfcb3aa8,0xf5eef7fa4fcafec8,0x7df67d7bcec2faae,0xcbcff3ff4edbfeb2,0xa4c0b21e048af860,0xb7bfcdf94fcbff22,0xb19c3269cfc27a90,0x7f7efdf1cedb7e76,0x7100eb314bc26a20,0xa64bffab4bd8fe32,0xfe9fcffbcfd3ff6a,0xf6b37feecfdbfeb2,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x79d6e72acedbfef2,0x2212778046c22e20,0x8732dfb10ec03822,0xf9fdbfeb4fd3ffb4,0xd3b67c404cdaf634,0xea00330888c22ea2,0x26fe1f434dc93e82,0x239fcf72078a7a20,0x9419fe204dda3a84,0x1aa42d6045c3aa14,0xa300c1644ec32300,0xc5cbffcb4ec27ff0,0x8f3adfd34fca7af0,0x945707ea47c37a48,0xab12ad524e8a36a0,0xc3b427c447d23a30,0x0,0x0,0x0,0x0,0x0,0x20104002000,0xebde5fdf4ec2fe20,0xd8bb5f7b4ecbbfa0,0xfb7fff6bcfcbfab6,0xceae2de7cecb2aa2,0xc4e6d3364eda7e10,0x41cd59d98dcbba4a,0x37195ef347cabf20,0xd1bdff764fca7ebc,0xfe3e77774ec32e34,0xcbf5f34f4f9bffbc,0xab5a32dacfcafe0c,0x393bcde8cfc39a64,0x96b7d4f24fd37a82,0x8e2c2d604dd3d80c,0x207e32203c03e80,0x4c02054806881600,0xbb3ecd738bdb7ae0,0xfcf276ffcfd3fe70,0x5f9eff7387c2efa0,0xf316f3f44eca7f22,0xcedbfe3a4fd37b1a,0x2bffdfdf0fcbffb8,0xfc3e7cb64fc37f22,0xc5a6e7f34fdbfaa2,0xef9ff77fcecbfef6,0xc7d66f1ac5cbbe90,0x28089e5b0ed27ac0,0xf4984abb4e1a76a0,0xc86c4d400382be22,0xfabc7d7d4ecbee28,0xc9c3de73c7d83ea4,0x6b7ff2dbcedb7e3a,0x77bb67fb4fc3fa10,0xd1f6f5fec5d2fe2c,0xbfe677ed87db7ba8,0x8885932146c30620,0x461236910c83aa20,0x9339cd4a0fc2ff82,0xf73ed6ef4ed8fea2,0x66125450404223a0,0xdaf0c32645131ff4,0x8a6a336a4c527eaa,0x8489cb208fc18a08,0x74b7ffe7cdc3be70,0x861080802020,0xdbbc7baecfcb3ab0,0x952b32306c3a720,0x9de095b4d0a7a30,0x3130d64847032c20,0x9b18a5504ec0a604,0xc7927554cf817a28,0x80a4ae3746c05e80,0x16945846807204,0x4b0ce2306014a08,0x3000b42283822a00,0xf0843e4003400e00,0x20800002020,0x26540e200a432620,0x4199458104c03e00,0xb08d920805c12a20,0x2044004020000,0x100040008001000,0x2206101802422240,0x80c8000002280,0x8410300000103080,0x513a094001830a20,0x150040002005200,0x96f3b25046cbbe24,0x8810c95206037280,0x0,0x80416044921000,0x0,0x144000000020300,0xa802ee9205593a02,0x10e1044004121600,0x3460737387824e08,0xca827fa4460a02a0,0x81462820c4832210,0x48ba1e4ec4c1beb0,0x12400240800220,0x380c00260498b082,0x10000a2000085800,0x85c3ae0a0ec97a80,0x5bda1a71471b3ae2,0x1c7b1fb10cd1faa0,0x993d2f3b4ed2fb86,0x7f367f52c7927ea8,0xc174fb3644db3fb0,0x67ff2ffe0fdbfe90,0xb0a1cde14ed17a04,0x3414c3b847c22eb2,0x24104000020620,0xf145e16e4cd9b2a8,0xc4eafef34dda3a2c,0x26050000004038a0,0x147036cd4cc232b4,0x22d8ec0046c87600,0xbb80df7e4793bea2,0x8051fa82024227a0,0xa7d797500fd37b88,0xfcbffeffcecb6af6,0xd9c075404c8ab0a0,0xfe3c2fb6cfdafa78,0x4682eb2e44835f2c,0x505a0202812a80,0x1826010a09c09800,0x4021a44205820820,0xcaac3da2c6c8f030,0xcf8ada364dd33a36,0x1944a03444a7910,0x8090046006822640,0x110035018c807200,0x4802250400c30a40,0x8981b2bb0c432f80,0xfcdbfef7458b7828,0x7bf5e9b478bbaac,0xd47054e00692ff20,0x421e0db04882eaa4,0x8024636449015a24,0xcb43fdca0edabfb2,0xaf5a499bcfc37e20,0x76f1ffe40fdb7e2a,0x2100510044a0800,0x0,0x840000000400010,0x0,0x0,0x0,0x0,0x0,0x100002080,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa0004800000,0x0,0x120d4101400800,0x0,0x0,0xcbea657f4dcb3eb4,0xddd7ff3fcfdbfafc,0x8f5fffd74edbfeb4,0xaef71feb4ecbfe34,0x480446524c003006,0xffa63f73cfdbfea8,0xc1f5c22ccfcabb30,0x4b1d4f1b4cc9fe34,0x3faa6ffbcfcbbe6a,0xf37fff5e8edb7bea,0xfefe7d7546dbfea6 },
	{ 0xfbecff73f872bbb3,0xfffeff7e2a72bbbf,0xfffffffbea76bbbf,0xfffffffbfa73bbbb,0x75ec7efbaa77ba9b,0xd3beff703a161bbf,0xfedcf77fe8723a9b,0xdcd6be232a52bab3,0xfbdf97d6e8773b9b,0xf4f2f7edfa529ba7,0xbbffff7bfa76bb9f,0xf7a57ffeba533b9f,0xffcfffbefa739bbf,0xaff6ded3ea529bbf,0x3ebfffedea523abb,0xf9fffffa6a729b9b,0xfffeffffea77bbbf,0x87daf20bea723a2b,0x2ffeffffaa761bbf,0x9fbacffaea763bbf,0xf5bfaff5ba523bab,0xfbfe7fa2ba739abb,0xebfdf5be7a763bab,0xeeff9f7bba723abf,0xffbbcfefaa733b8f,0xfdfefeeb6a729bb1,0xf3a66f7b6a139a9b,0xbbac7737a852bb3f,0x4c5276faaa721aa3,0xaffecfc1ea723b8b,0x6e0084002a0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xebe6a97ae872128f,0x1010420020420001,0x97ba67e7ba563a9f,0xd0351c7988421319,0xb6be3f74fa52bb07,0x89b270432012128a,0x6cf79fd0e87613bb,0xa6bbcfe92a5633ab,0x9708326028111a8b,0x333c3ca008723283,0xebc0d51e087232ab,0xbed3924b68570aa3,0x158d575338523a93,0xa5cd176a8872328b,0xdbaeaf76a86232b1,0xf9842755a843323b,0x0,0x0,0x0,0x0,0x0,0x8060004a8020108,0x675c5edaaa72b3ab,0x866f3ff7aa5213bf,0xffffff7ffa77bbbb,0xf9bf5fff6a72babf,0x8ac2ebbaba701b93,0x225a92da2a521a8b,0x8c3f7ef98856ba9f,0xd1f586feaa561b8b,0x5bb63f5e2856bb9b,0x8b84ef3eba703a8f,0xefffb7dbfa77b2bb,0x8b3c7fdfa87232a7,0xd5d2fbdeea5712bb,0xe9ee65f27a5232af,0xc30220300a500083,0xa425a8a28701882,0x2d1fcdc2a27212b3,0x15d636aca870923f,0xc7bdfd70aa27329e,0xeb826d7aaa123b9a,0x4edf9cb7a05212bb,0xb254d753087212b3,0xd0fff6e0e822831b,0xcd8fe7622873bbbf,0xfb9e7f7efa521b8f,0x69cfdea8e872b28b,0xabfef75be063128f,0x8cf25afb486238bf,0xf2f6b7ec5a739aa1,0xda162665a852321f,0x6100c6242a6001a3,0x2d7ffdd8ba529a8f,0xbcbb4f09a812139a,0x5564a4fc88721b21,0xdf5c2d65a012908e,0x695e536eea72b18f,0x9d73a95b8523a9f,0x8d9a797ab872130b,0xbc57b4fa3a72ba8f,0xdffe2fe2aa52b28e,0xc9e453488852123a,0xe8c3f23f7a563a9b,0xa5a94c50aa521283,0x5454460a4a001883,0x9102a730a242800f,0xc900ef3e88721b9b,0xce8ffab0aa531ba3,0x61200421200,0xd01400200870010a,0x203861e028722202,0xca9e6f1c38523317,0xcd6ea6822a720293,0x434f1ad0e862220b,0xd484b8328721202,0x484c310450401283,0x10a41e020a121901,0x80400008000100,0x940000400021201,0x24000020010,0x9346f2f15840ba81,0xc7b024508a428313,0x8820500080121000,0x8d715d22a42b28b,0x62b445e048721a01,0x92f332404a601a01,0x204c020800000b,0x2004010208000082,0x409bb619b8421a8f,0x86080420028228,0x4000800000600000,0x118a010022121001,0x880400a48101801,0x4600701228321800,0x2c7a050308521a36,0xa0f656cce852123f,0x830aad710a523b83,0x60a22f4180523219,0x892ac10aa8121b0a,0x2080800128508080,0x43c26c0e8020091,0x8101301060400209,0x1ab4273000663a00,0xc8c07202a8423092,0x206300608413a01,0x818acb52e8423a23,0x1c757fabe8179bb3,0x7bb62f74f85291ab,0xb5af312a85213ae,0x6ecb96d8e8569299,0xb29bccf388771ab0,0x341266e4ea501291,0x80a7d5008001a1f,0x8bd3030a42021a07,0x48c9122f1876b281,0x9084c1a0052120b,0x61503463a862b890,0x82a0ad6088021a89,0xd3a6fb662a52930b,0xc9d2de102a023ba3,0x92c3df5aba729aa3,0xf6fa56ee6a7713bb,0x348c322a8522207,0xf995ff6fba533b9f,0xe5fbee28e842b23f,0x1fde5f81b8721a8a,0x34ffd6636a521b0f,0xe9ec747a2a761aab,0x8608256068038291,0xd1661d7418563331,0x3808968002220b,0x108428020100206,0x122056023a523083,0x7fccfd7268731baf,0xa885e30a2072ba27,0xa222110521210,0x44048000240000b,0xc144176868621301,0x4bc2570a8621321,0xb9b8fbacba52120b,0xc63a3d112a5232ab,0x800014000501200,0x7cf7feeffa72baaf,0x8000200000061200,0x0,0x0,0x0,0x0,0x0,0x0,0x40020000000a,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x3000000008000080,0x0,0x0,0x0,0x0,0x0,0x1000002010001000,0x0,0x0,0x0,0xa50008600007,0x0,0x40080020020000,0x0,0x0,0x0,0xfbb4777f2877339b,0xefcfffaeaa76babf,0x8bfedf5daa523a9f,0x74fe7fe328773bbb,0xfb59fff228763b9b,0xff8e3f6e6a739b9f,0xcff3f7bd2876b38f,0x8b56000a0a62031f,0x800843b1e052101f,0x70bcbae83252b201,0x4bbc2f432852b38f },
	{ 0xffeefffbef7d1ffc,0xeddfef7f6f7e1f5c,0xefffffdf6f7e17dc,0xbffedfffef757ffc,0xf5f5f66a6b775dd0,0xdfea7c74417659d8,0xefb47fffe77f1ddc,0xdbfbfebba37f2cfc,0xfadfdfdf6f7e1fdc,0x56fff6ee67775d48,0xf8ffff66637f1fd4,0xfab47fefe77f5efc,0xcff7ffbfe7767ddc,0xefdfffdbef7f1dfc,0xcdf757ffef7f5ee8,0xb17eeff903745fd8,0xf2942fefe37e3fdc,0xc5defbff63775fec,0x93ff7f4b637d5cd8,0x93bb4ffaab7f5d5c,0xd1ecfef8e7761a4c,0xfe9ef7f2677679c8,0xeb97dbbae7553f5c,0xcfd7bf156a751f48,0x65bacdc9667e15c8,0x956e766eeb7f1d48,0x5abe7d536676345c,0xe3c5f33ee7551dd0,0xec6bf79d61773fec,0x8bbdefd86f7f1fd0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe8d7df3861751c54,0xa9ce8e886f760070,0xadbaaff3e97f1bdc,0x85809640024c1150,0xc15c7e60217610cc,0x4b82e93781561cd0,0x606220821511c18,0xbfb96feea53d3fe4,0xddf676e9c3362e64,0x541e7d7163751360,0xf39cef6baf7f1fdc,0xfb573ffbed7d1f54,0x2e9f0be163741f4c,0xcc36afefe53f1df8,0xc3beffe2e3741fe4,0xff94ef2d687f5bc4,0x0,0x0,0x0,0x0,0x0,0xd6c32619e0761acc,0x6f5fdfdee3761b4c,0xbfd5dda3027f1374,0xddf8f770ae7c3d68,0x76b31fdb2f751ed4,0xc4783600424617c0,0xab0f439140765c44,0x1f9b4fe366764ef4,0xfb6cff7fe3775be8,0xdaa47ded6f7417dc,0xe1d7f20e615713d4,0x5c8f9e49a71f1dd4,0x28ecfa262541450,0x708f86e1a6340a84,0xc4ec75e2a4157ed4,0xb8fce34e4742ae8,0xfed236ff627e5e4c,0x8bbf1f5a6f7f15dc,0x10ef86a821645dc4,0xd2be4d70e2374ac4,0xdbc6f73467741cd8,0xcdeffb3d677e7cd8,0xafdfff53676f5ddc,0x77b6fffa677e5f4c,0xdfaff773e77f5dec,0xfb743fdc627e1edc,0xdfd3ffbae7745dfc,0xafffffdbef7f5ffc,0x40606e1024e0240,0x5b4c276ae20658cc,0xc2822f592b7619c0,0xe184f3bee3741dc8,0xecf27c9dee7d19cc,0x56996f7b6c753fec,0x9184b07f42567bd8,0xfe9ee77de37d79fc,0x88a8d5aa63741404,0xea1a90c227460048,0x8f084cc265741040,0xd594a271657c195c,0x7baefdc163761ff4,0xf222e30e63773348,0x19d0336363520ec8,0xbbbcdf33c0371bec,0x456246823040cc8,0x9f807550087519c8,0xe12cfe71e33629c0,0xfd2ff2a407e08c0,0xa851cdcfa2741dc8,0x9490e5c026741cc8,0x504016824001050,0xdd22296e21741fc4,0xcc7dd06417631d0,0x244020c80040884,0x404000000100200,0x1302e00005214d0,0x9034873003761644,0x4480612060544a00,0x6000060000000,0x9451140200344a40,0x100812002040050,0x5eec604074027c,0x2106200322010050,0x800020240100050,0x8102088102741000,0x1000a64000140084,0x200040000100420,0x8900e020263c0060,0xc983ac100651258,0x80354e20265610d4,0x14042880251808c0,0x80280c0080440940,0xc100713180144a84,0x8cc0301c62361414,0xa01a1fd165661100,0x8030506a21445bc8,0xf1aa2470623403e4,0x50282470230419d4,0x88c31600e0101e00,0x2a42821024521218,0x3080048840170040,0xc9425468414f18c0,0xb012bf1d295e10d4,0xc7d2febae1340e44,0x2952ba04214d1cd8,0x632672023781e48,0xd5bc97026e741440,0x71a23c02a65518d4,0xc37df2bfa37c1b5c,0x843325926080158,0x16bfddf7e3763dd0,0xf577137c66777cc8,0x8050026040140,0x80d4e264663c48c4,0xf9d2165b83746ac4,0xe3bb5e72a6741bc4,0x8512d6c800361e40,0x98040560a2700860,0x470dc022253509c0,0x6bd6ff1b6b6719d8,0xa75edf1766791d48,0xd4b67eeda6775fdc,0xedec377a41761bf8,0xd7967ffdef775dd4,0xd1a0fba2023c0ed0,0x6a3efd5a62775fc8,0x9c3ac773027e035c,0x0,0x880d4001360a50,0x108000100080,0x2900726640000,0x1048a223141000,0x8159a63527631940,0x6802502002081148,0x8344bba663560dcc,0x49504e1060661408,0x23b28e10243138d0,0x84e3706d224c5cd8,0x108a544002141304,0xc316e32a017c036c,0xc1a1a836025e48ec,0x198ccd4827141c74,0x947776ec6f6f1dc8,0x200c00001000c8,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40800200800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8080000000040004,0x101000000050,0x0,0x0,0xb3983c506d7518ac,0xdfc7fe22667419e4,0xeffedf53667e3dc8,0x7cfaf36361760bc0,0xf96c6742e5344a6c,0xfabe2fffe77f3fe0,0xefdeff7e67755f5c,0xe0dedbd265371cc4,0x34da5753e5763944,0xf5fcf6fee7777fcc,0x561466f56d751fc0 },
	{ 0xeff7ffd31fedefff,0xf797fffb3f6cdffd,0x4eebbfd91f6ce3f3,0x6defdf789e6defdb,0x356d7ee81b0dd57f,0xdeb63d50374cc591,0xff94476f3f0ccff9,0xc6dff73b8e6cf7d7,0x2fd7ef5f9aaddddb,0x3cf656cc0b0de7db,0xfbf6777b1f29fdf1,0xffb27ffd2f4cd7fb,0xc3def7781e6cf7ff,0xefdedfd93f2dc7fb,0x9ebeffff2becff9f,0xdffffffabfadef9d,0xf6b67ffebf2dcdd7,0xeffbff7e3e6dcfdf,0xefffdffdbf8cffdb,0x77b8c5740e4c67c4,0x98525558290ce575,0xe79e3d7cbb29f5d9,0xebbef77e3e2cffdb,0xffdebf5b3facf7fd,0xbba4fff3278dc79d,0xb5ffff7f1fedffda,0x290c0da0032886d3,0xfbd44f3e028dd7b8,0xdee7b7ff3facf7dd,0x2200000000080040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xcfd2ffef9a2ccfd8,0x365c000088855c,0x368aa5f50828a11b,0x83e7442082ca5c2,0x5ebe6d731e28ff99,0xcba3f103030d5d10,0x65f91e198e2cc6d8,0xdfdf4fbf1f6dcff9,0xc4322670a84c85c2,0x6bec7fe08b6cedf3,0xc98477170f2cc49b,0x4ffbbb5bafacffdb,0xffbf57fb3eaddffb,0xfdf676f6bfecf7bd,0xd7fe6d71ae4dfdf9,0xdeb4bf7b3aad47f6,0x0,0x0,0x0,0x0,0x0,0x8dc7b7b3ab28c79b,0xcbf7fd581bac87f9,0x60344de38708d516,0x5ae0fd62040dc6b5,0xaa32af641b2ccddb,0xc5c1b6101b00c789,0x211ed7583a0ca75d,0x37ddfff71bcdedf2,0xd7dd2ffc0f6dd2d5,0x7290234c1a2ca092,0xbf1f30f234cc598,0x2ec5beca2a2de79b,0x980dcf333a05e750,0xc47df471ac20c7d3,0xeffe7dfb3f4dffdb,0xf307ce1a2ae5dd9b,0xe9213ca1a24cd02,0x863ecc650009cc85,0x4b06644064cc418,0xb3bf3f20a829dfd9,0xcd86f37f3aec9793,0x8a83fea71f2cc7da,0x883f5e5a1a28ef5b,0x4e617691a89cfeb,0xffffeffabf4ddfdb,0xdebe67dd1e28cfe3,0x6ed3f20e1b0ced91,0x2beedfd30a6dcfd4,0xe4bbd7e92a28efd1,0xa10a252013084990,0xfeffafff1f2deffb,0xcffaf7393b0deff1,0xefffee7f3f0defd9,0x99a237fe1eadd55b,0x1ddcf7f03e6ccddb,0xfe9467641f2cc4d9,0xc083b90b0b28c7dd,0xc4da36499a2df713,0x3a89218412000110,0x80cea3381f0cc49b,0xf9feec711e2deff9,0xdb08f76f0e0de759,0x4fdbbdf70e0df7db,0x4eb94d402a2887d0,0x10f0ccc20b28c9b0,0x80d840200a2c8401,0xc2807401020885c8,0xc94bef110b2ccdd9,0xa94cc4912014510,0x963524601209c711,0x49642408020c8109,0x4f00634e0a65cf88,0x40820d0204084502,0x308da92080c851a,0x4c205020800e881,0x4000a80002095010,0x890005011a048581,0x5244608002c44d00,0x8004002000100,0x3000465220008391,0x90909040282081d1,0xc2aaad422909f190,0x834460462008cbb0,0x4d01a003208071e,0x9409042228000381,0x1000004000008010,0x50002000100,0xc044452c1a21c00a,0x240021002008302,0x2180291080881d0,0x109004620788cd91,0x52b22c501a28a470,0x92918d64210cc549,0x50d18812a08110,0x6211d45a0a09cdd1,0x1020624008080180,0x102a800200c0510,0x9a8177241a4de55b,0x9c2e1140300ccd8,0x248014822880301,0x940600830804ce43,0x5900002000048110,0x91020000208010,0xc5fbf3181e4ce7dc,0x2c5f2fdb1e08db9a,0x279a67da0fa9cff9,0x71e435299709d6d1,0x7a220d701a08859b,0xe9f7f31f3f2ccffb,0xcfdf3fd33f2cd5da,0x2e124d581a28a199,0x9130a646052cc5c4,0x448c0560002ccd81,0xc015ca28014dc491,0x8cc136200acc871a,0x2dd6c7101a2c845a,0x94f27e84b888a753,0x9b8bbf719e4dc7db,0x41006479170cc409,0x9fcdeb238309d59b,0x1261bd11a0ca3d1,0xfcb4fee91f2cd5d3,0xdb65e77a2eccefdb,0xc0bf2e52090ccd53,0x4542a2bc0b2c87d1,0xf1c1384060ccd17,0x991257622a08d379,0x20060000080101,0x89340c7008080703,0xe144d5ac0a08e4d2,0x150160000008180,0x84a9000808181,0x9b38277c3f2cc359,0x280000004004000,0x284c0220600c518,0x86c00a020a214281,0xbf99d9fb1705cf9b,0x1560036a8644c40b,0x252225200b24d490,0xb98423790c0d7ad1,0x674a7eb80a2c86db,0xc7fe4d693a08a799,0x31dc166b0d2ce6f7,0x14008000040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x840000000008001,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80100000080400,0x804020000100,0x0,0x80c00200000c012,0x0,0x0,0x0,0x0,0x108c01a0880c0458,0x44802754020d4729,0xc6dfff9e3f2dd5dc,0xaffd725b3e0cffdb,0x94e715280a8cef1b,0x797efe7fbeacef5b,0xcbfe3fff9f8cffdf,0xcff6f7bfbf2ceddf,0x2f9f7edaaa6ca7df,0xdffeffffaf2de7f9,0xdbf4753d3f4dd7d9,0xc2b075d00b4cc759 },
	{ 0xef7ffff7cdbafffb,0xebf4ffffddbadff3,0xaffbffdfddbafffb,0xdffefffaddba7f2a,0xbd76b5ef05bacef3,0x5bf2ff72ddba5de2,0xdb9cfb7e5dbccf73,0xcdffbf941cbe4fab,0xbfdedccfddfa77ba,0xd0f4f6ec159a4f23,0xf5ff7ff29dfe4f3b,0xffa677fe5dfedfe7,0xcdf6fc0a5dfa5f33,0xefdfdf538dfa5f53,0xfcfb5fe9d5ba4fea,0xf9e6f75f4dbaef6a,0xcaae2feed5fedf67,0xeffbffffddfeffeb,0xfe7fff6f5dbe4f6b,0xa6f27f620d385d6b,0xd57d367dcd3a5f6a,0xff967ff7ddba6ffb,0xabfce7be99ba4ff3,0x206f93e1418e616a,0xfed95f3a4db84be1,0xd7e6c7740dfe4ff3,0x5b2c3d621daeefbe,0xdbd6cd3fddba47e2,0x8ee27ab00dbe4f6b,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x280004000100,0xe1cfe93645ba1522,0xab5eb6120df84f63,0x14b29ea99488073a,0xbffdbe64c09ae7e7,0xdadabdf155be0d62,0x8152f3614d3a0d2a,0xd286501c820540,0x1e894d49509a0da3,0x21ef076c849a1c0a,0xfbbe7dfbddba6fe2,0xf394e72a4dba6f62,0xea5ffbdbcdfaef2f,0xffbfffffddfeffff,0xf5dfffebddbeff6f,0xc98eff71dd2a5d63,0x208040418000500,0x0,0x0,0x0,0x0,0x0,0xa932ffaac8becfea,0xea77d7d60dbeed63,0x8010a01220002,0x943025600818472a,0xfabe3cc0dc00654a,0x48c5ba84541a0da0,0xf5e19c80d9edf21,0x867a535b8dba4d2a,0x30c476c04092cfea,0x55027b2094180d60,0x8957aba2c492cd62,0x675dbe504dca01fa,0xb398cfd215aa17f2,0xf534f6e2cdba07f2,0x6c060fe2049a0641,0xcb84ed7005be7de6,0xf949bab58dbe4f22,0xba3f6f6089beee32,0x4c503e69599a4d22,0xfef72d779dbaf763,0xd31eefefdd9adfa0,0xcfcfff1b4dfaef62,0xbbbddfdf9dbeefea,0xfcfd7ce45dba476b,0xdb5ef36acc824f73,0x8e942e6554b865a2,0xd3c7ff9a91facf47,0xc27e9cc345ba61e2,0xfcf7ffefddba5ffe,0xd978ee7a55baceab,0xffbfbfffddfeffef,0xffffffffddfafffa,0xcddefe9f5dfe7f62,0xb5c8bff9449add6a,0xf1fff57a9dfedd6f,0x6e9e37701cfecc62,0xcbc4fea2949acd7a,0xd8f39e5e589a4f6b,0x1f126f5a55204ee2,0xd7b0b7de0df8f86a,0x95bc6d701ce8dd6a,0xa350e7000c9c4142,0xe8b96494db8cf6b,0xb2e09534dba0d23,0x9450a5ec4cb84901,0xb263da2548a4d2a,0x89921271499a470a,0xc382b22b892a4da2,0x1ba9075215e8d540,0x94c0926001b00962,0x207602108a4442,0xc3000794488a4104,0x90b08f0a4590174b,0x25b1152c1b4cc22,0x12000000a0,0x94c206801a08820,0xd300052204205b00,0x8020600001200002,0xad25050ae0e20,0x320a650201b84022,0x9524017800a8c462,0x2010041008280100,0x100800000080002,0x212114090c420,0x880005a004084120,0x4027040088a4120,0x592514109020b03,0x5000d80044a80100,0xb4e0b0f4ca80320,0xb3a4dfbdcba4d01,0x54d056ea0c9a5d24,0x5e3af512d4a84ca2,0x8904113489c48502,0x9018a01048a080a,0x80448120810050a,0x3420540000186c03,0x91ad8b700c9a5d20,0x8880672d4caa1961,0x846220000226842,0x4b02101900286862,0x1004048010284600,0x5080000000c80020,0x0,0xcfa7f6175db8c77a,0xaf6f1a599d9a4da6,0x16230fb1d8024d12,0x8101169189884500,0x1a9a3e500db807e2,0xa964e3395de21d3e,0xa84b1351c9a24d6a,0xa5134d928c92e40e,0x54b4b8240cfa4cf3,0xc1965fd085ea074a,0x59d4c12208bacf00,0x28ca304d48920352,0x2d0dc98049c80b23,0x78123472158ad426,0x53aec450442259f2,0xfd60cd5b5d385ff3,0x7063c0a4c98cd01,0xb287cd1accb20f20,0x47154e008ba4f62,0xe5e7ff6b5c9afd3a,0x2bb8ef3509bacfae,0xc673ce1015baccea,0x22879355c4a8642a,0x2401497305b8040a,0x886c2662c48a0540,0xec9c3cd004e20347,0x861201080c004100,0x241a10d05c9804a0,0x3ea9c7521cfa876e,0xaddc140801ba6a03,0x82822a0001220352,0x61842005048a4840,0x69e633f75d2a5f63,0x90060004880002,0x80101408c0820500,0x1e262542dda04721,0xe891420c44a88062,0xbcc7ff5d9d9acd6f,0xfb9e8febddba4f6e,0x44010200000000a,0x3a00243044c0016a,0x4000200000100400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10004001000,0x21000000820,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8400008004000,0x0,0x0,0x0,0x880820080004c00,0x0,0x0,0xdd04b7600caad1e2,0xdebe3f761dbec56a,0xd1237a310daa4b00,0x4a40120904905801,0x4535f7ef0dba4ff7,0xa967224289ba8deb,0xffbe3f78ddb87fab,0xe7bdf3245ca84f73,0xbfdfffd9ddba4f6a,0xbcb7dfeeddfadf6a,0x552d07790db4f56a,0x5d9c6fb20dbadf7a },
	{ 0xdbdffffbf6336eee,0xffdfffd7b613feec,0xdcfffafab633f6ee,0xfffffffff633feee,0x5ce7f6f9f41276ec,0xfbbafdf0b0136fe6,0xfb8073d63013e6ec,0xc2d6b23f16336ae6,0x295fdfd9b43276e6,0x7cb5f6fdb6337fee,0xf7fff77d761366ee,0xdffe7f71b43373ee,0xc4cf9b39b433f6e4,0x6b70dfcbb63356e4,0x36f777eb34336fe4,0x70eff7f93433f5ea,0xee3e3fffb433f7ee,0xebd7ffbf3633efee,0xfb5a17f3161336ec,0x29a4c60841078e0,0x93a9b64cb61162c2,0xcf7eed62361256e2,0xebbeeffff63367e4,0xcbde5eddf433faee,0x37bdcfdb1233f6ee,0x9c734c709412e5ee,0xba1c7d327431f0e6,0x9bfecb773433feec,0x8020000020000040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd48adf2aa413e2e0,0xffdb3a5bb631fbe2,0xb48826e80413ece0,0xf189af3e3432fbec,0x46527d60f413e4e4,0xc944cb0b2432626e,0xa257ef46943362ea,0xbe734748b0104dc4,0xd732e67e9633f7e6,0x26b4bd72141337ec,0xdbf0c376161375ec,0xcecb7e33b6326de6,0x3feacfdb7013eee6,0xfdf776fbf61373ee,0xfd48ff31343376ee,0x0,0x0,0x0,0x0,0x0,0x8a1020c20022460,0xeff7e73f3632e6ee,0xea5695d83413f7ee,0x0,0xd548a070943226e4,0x14900052203056c0,0xc487fcbe343266e6,0x4f5ebfddf410ffee,0xbcb2edfa36323fe0,0x51a47452101366e2,0xfacc6da2a43076ea,0xeb52ef2b7432e7ee,0xe0c218d032104ce0,0x8f9fcffbb613feee,0x1624f4e5063336e4,0xf5be5fe2b61176e6,0xc1e45f303613f4e4,0xaec132923002e3e4,0xbffdc64214336eee,0x14f5f66a3202e7ee,0xab3bbd6034117bc4,0xe3a5f36b3622e6ec,0xdfdfffbfb633feee,0x8f775fcb161267ee,0xc4f477e2e432c7e4,0xbbdbaff03431d7ee,0xfa287c7e3633ee62,0xdec2bf9bb43372e6,0x4e9fce4f54114762,0x9db447ea363370e4,0xfbffff7bb633f7ee,0xfebffffff633ffee,0xfffffffff633ffee,0xfffffffff633ffe6,0xb3bf4dfff633cfe6,0xf1edf7f8343377e6,0xfb7a6750f6334ee4,0xc336f3fbd4126de4,0x6ffb1f43861372e6,0x8822864104315260,0x9023223020114260,0x1ae0de0361066c6,0xdb96a37e30133fe6,0xeb5e9f4fb433ffe6,0xe90f414080134064,0x90577ef7b4036fe6,0x12be45725612eae0,0x3830ab2a2433306c,0x469b9402b43100ae,0xcd249021433f6c4,0x90464064104060,0x0,0x18202140011c060,0x1402a33ab01352e0,0x2410008000104480,0x182040204206040,0x80240a0810000442,0xad1005c0b410c8e4,0x435543240610f5ec,0x4004400100104084,0x920007010101042,0x1200832080304280,0x800e354004106144,0x8982c022002246c4,0x556108a141108ec,0x1202000040,0x540084e45410c2a0,0x146472000004140,0x8200010404204000,0x7527211801162e6,0xe040102342004c4,0x50001245140007c0,0x4000460001040ac,0xe880522050026aa0,0x14e120030110060,0x82dcd40160224c6,0x1c946a88102226c2,0x801e564014022ac4,0x40102a1000104084,0x100c0081000e042,0x848445420611c1e4,0x1112871b0012e268,0x0,0x240000110000,0x812081100430c000,0x8540e6232100aee,0xbc234de0c433c2e4,0x24150a203240c2,0x7c48646b361146c4,0xc904a222102242e6,0x6cd7b61d341236ec,0x1111059612113440,0x74a68662240227e4,0x4a602549101246c6,0xa184c70634127ee8,0xa2421250200223ec,0x4412010270000022,0x1081142610100462,0x63829c30821106e8,0xcf00e743903226e2,0x80c30010006842,0x3c6df98141376ee,0x30fb7ef5b4335ee4,0xdffd2f7bb6337eee,0xffb0677672324ee8,0xeedf6eaa543277e4,0x14e3b00241279e4,0x847164e01411666a,0xd3b2ac04101310e6,0x4a88044c2420ccc0,0x8142f2291633e544,0xc80a3206101162e0,0xb6410d4034004cc4,0x80702410300064c0,0xde00744e94326de4,0xc9cfebeb363077e4,0x6fde3fd0941373e0,0xfbbfc45ba01366e0,0x2002026030030044,0xc6940030b6204464,0x6124420256114164,0x404338103400e2c6,0xc39e4f6b7432f6ee,0x4cd506e0401276e4,0x2010000000000480,0x4200600010110400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2020000000060,0x0,0x108000840000e040,0x1be8f70f032c7ea,0xfb0e79fc341277ee,0xedebfd18163277ec,0x2a7eff8314304ec6,0x8432fcec303367e6,0xffbebf7b1633fae6,0xfe1e7fb7b412ffee,0xfeffffcef433fee4,0x9ed7fefb76136fe6,0x9bb347ee723376ee,0xf168b7772633c6e6,0xed12016210336462 },
	{ 0xdef67ff9f8dc9ffa,0x9fde5ff750dcb7db,0xdffffffff0bcfffa,0xf7fffe57709cbef2,0xfc76ffebf09cb6fa,0xfdfaff61f09cb68a,0xdaacff7ef89c96fe,0x5dd7beb5f094fffe,0xa9e6854ab0dc36fa,0xb4b8cbe4d83c3242,0x9df6b572f0fcbeee,0xffb27f7ff8dc76fb,0x4adbf6e8509cfec2,0xaffed9db709c94ea,0x96733ee7f894b7e2,0xf869e7d770dcbee8,0xdaa3af7670dcb4f2,0xed9eff3fd0dcfffb,0xaa5fee5a70543cf8,0xbeb95fd278d4d7dc,0x916db67a6054f8f8,0xfa9e75f570bcb7e2,0xcb4ffffff0949eea,0x65ff15ca5814b66e,0xbdfa4913709cbee0,0xfd5efeef70bcb7d8,0xbfbe3d69d0bcbff5,0xb2068b0228405428,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6d4edb96d0b4bdd9,0x8956df52f01cb66a,0xa205c2b09c10a0,0x600d97cad014b4c8,0xfd9e7dfdc89496db,0xe1fffffff89c9ffe,0x577ffaf0dc3ebb,0x3ffdcfc7f814bfef,0x4504e008504cb060,0x7f3e7ffbd0bcb7fb,0x3b2e53e509c3493,0xeb7bbffff09c9ef9,0x7b3e0d50d08c3092,0xbdd77ffff0dcb7fa,0x91baef72f01cb4ea,0x0,0x0,0x108000020000020,0x0,0x0,0xbb126375d09cb4da,0xcfcffe36f0dcfe6c,0x200a108140801422,0x8ed57463501c9e4f,0xd578fdd0f0541ead,0xba3e2f75e8fcf6fa,0xc9d5f73f509cbfce,0x6fff7fd7503cffdf,0xbee557ea709c947a,0xd1699742f0141688,0x6b0c36245804140a,0xc8f1fbbac01c9edf,0xeafe7fd1f0b4bff0,0x5fbcff4ef0b4beff,0xa95da64050849480,0xe4527de3d004b260,0x8b8ba91b5014be79,0xd6c38b52701434e2,0x61646d7e09c3ae2,0x2c369361503c90d2,0x83b8041170140092,0xd21cdb6ed0b4b6f3,0xcfdbff2bf09cb6bd,0x2b1dff537094b6ea,0xfcb5fefff8dcfeff,0x53efaf7b78dcdeba,0xd9a2eed050c4b64a,0xa1c6ed0a704c92e8,0xcfff7fdff8dc7679,0xb0927d7778d49fe8,0xfffffff7f8fcffff,0xffffbffff8fcfffe,0xfffffffff8fcffff,0xefffffdff8fcffff,0xb77bdffff0dcfffb,0x9171ff9378dc9efa,0xfacabd6e505cbfd9,0xe9f7ff6fd09cbfdf,0xfc1a3779408cf6d2,0x508c903501c1a80,0x17661648509c1c02,0x45826552d09c9480,0x8810c3000050180a,0x215b3e03f0b4bae8,0xcabfdfdbf01cb6e8,0x142642e0401816f4,0x8030201010102240,0x40242394600452e2,0x144c62018109c42,0x6f188f1650149692,0x5498fca9504c34ae,0x54a657f1f804b7d9,0xd38cee56d80cb5c8,0x34c7ce08701485f6,0x1904a0000140682,0xe3320522f0140422,0x60062070b004d0c0,0x8200052050100490,0xc154f57790b412d2,0x8ee3201d01c9cde,0x42271002000a560,0x803712c00c9080,0xb00c3d45901094f8,0x9d0602b101490f8,0x8860c00b01018a0,0x81264d6008349042,0x0,0xc30c4211701c0030,0xc150c502107414e8,0xa200321e0050160c,0x938c0120008094c8,0x20160000101410,0x802100220001022,0xc1854012308414a2,0x8000200040040000,0x22c4801903854c0,0x8c9017620090141c,0x7620fc060c4101a,0xcb1c6662708496cc,0x2c947312501496c8,0x82167e566004b4a8,0x332a4febd89cfecb,0x8840060030849040,0x8000250040002480,0x2c00800100c80,0x0,0x8490142050109888,0x8520240400014a6,0x40b62ff0f09434d8,0xc9d0ea22d094b6ae,0x2d06214060049068,0xb7a94f895814b7fc,0xb7f454ea98d49c85,0x24b560508ca0f8,0x1d07202f0149010,0x2b7a3c84401ca340,0x2228c1850108444,0x90db26c8d09496c0,0x813c6d327094f358,0x88002c21d000b408,0x4842301070043082,0x6039d250101c1743,0xc0995669581c74c2,0x9e862560d030a6d0,0xdf9e777d70dcffea,0xaccaef12f094b578,0x398dfc0781c3e80,0x26200c61501496d0,0xd37dfb68f0f4b6fc,0xde942fec085c1e9c,0x81c1d000600c0038,0x25187e94901c164b,0xb000441018001080,0x0,0xf0844d6a701c31e0,0x81208140700004a0,0xfff6ae93f0a4f6e0,0x9aa4cb25050b890,0x4202060001c0c80,0x47a82f40f084f36a,0xfa44e262280094ee,0xae03da406014ace0,0x903804e0503494c0,0xc30d6424010374e,0x200000008040,0x4840000010,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20010000080,0x0,0x0,0x0,0x400004000801000,0x0,0x0,0x0,0x0,0x6006000002644,0x0,0x0,0x0,0x14443443600c54a0,0xdbfff75bf0bcf4fe,0x4cb42f7c501c9e98,0xedd7f7ba50d4beda,0xea0fda51d034fdea,0xc365642700414ac,0xdbfeff7a707cb7fb,0xdebf2f7ff074b7da,0x8ac3fb66701436f8,0x2b76db7cd0b47ece,0xbf99cffad09cf4d2,0xd5daf722409416ba,0xe8b42e7470189efc },
	{ 0xfbfeffffd9efbf7e,0xfbf6be77d9efabce,0xafffffded9ecbbde,0xffffffff99ffbbde,0xf4a677fcd9edba7a,0xdbbebf41f1e7a3ca,0xcf8cdeb799e3af7a,0x8fcbf6bbd9efbbfa,0x297e5fdfd9e6ba6e,0xf477b7f5d97fb952,0xd9cbed66f955aa7e,0xfb9cbf7cf1edaf7a,0xc163e87690c3abd4,0xab98cdda99c2ba9e,0x14bdffebf963bfde,0xf9acbe7ed1fbab7a,0xdfbf7ffbd9ffbffe,0xefffffb7f9f9bbf6,0x6b5fdbda99d9a77a,0x3a2e05f891808b58,0x801a8452d1830042,0xded6fd66d9cf267e,0x8b6ee769916fbf5e,0xeddb1a51f98f335e,0x5795fffe99fbaad6,0xfffdfefef9ffaffe,0x49083700112c020a,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x851120010460618,0xa34e5e5091e02766,0x98468431082246,0x50eeb77091840a1a,0x4eba7c51b1c1bb34,0xcbefbb2fd9e7364e,0x8eff7fc89155bb7a,0x6fbfcdffd9e9b7f6,0x9cf3e92d5961231c,0x61742f709962abdc,0xda64fb2ed9ffabda,0x98c63e0198ed2250,0x9b3ccd7fd938296e,0x51f6a6f6d94b3372,0x8202004000409200,0x0,0x0,0x0,0x0,0x0,0xfa14f3fcb1108bfa,0x8190850ab008a300,0xa0c448a90802804,0x541611e9d1e82340,0x884745688048ab52,0xf298271799e98e38,0xe0c4d71e91c2b36a,0xaff9e099948a3fc,0x8dbafebb9964bf4a,0x3168b66391900378,0xda260f4ef1ac2278,0x83e7e3683967225c,0xe9fe5eddf9fda37e,0xff9effdff9ffbf7e,0x9473f6fe9965b15e,0x751605c0d184b01a,0xfaa2cf6af1ed2258,0xf5db92b1b9ec2618,0x291ece93b0e92250,0x7fffbfff99e1277a,0xbb9e7c53b16da3d4,0xd91ee3bfb1e7bffc,0x8febfebbf9efa7fa,0x60e484099c00e10,0xbcb646ecd14abf5c,0xb3fcb7e0d945b37c,0xdb965bf491e6267a,0xe1c6cf1bd1413b76,0x2f5c56d9f14eab58,0xf4fe57ebf9f9affe,0xfffffffff9ffbffe,0xfffffffff9ffbffe,0xfffffffff9ffbffe,0xfffffffff9ffbffe,0xbff36ffbf9fbbf76,0xd5eff5fbf9efbb7e,0xef46fbe5d9f8a552,0xebd5f3eff9efa3dc,0xaec71fd7d9c9aefa,0xaa3a4bfbd9c30d40,0xd494d6f8f861ab7e,0x5db20dc8d1212366,0xa8dd660ec9c10542,0xa9ca26c1f0c13252,0xbbbadf7bf9ed277a,0x100256c441492342,0x5b6afd535104b23a,0xc382230d9000a9e6,0x1cd33a08d1458768,0x2026c80290000268,0x94110460c1440278,0x2204201008802,0x54a0e2c3c8d12e50,0x4342e233d8d2264a,0x68117ac099c02212,0x10084a8311248308,0x0,0x800014c01419002,0x81b6d20c9065a908,0x7bc69f55d0c4abc8,0x3000000000010250,0x8001500010810040,0x90442200024a00a,0x80100a004024a80a,0x4c0081000440308,0x90020d60110c8308,0x42024290238000,0x24041000012040,0x80042502c0002060,0x690100000202142,0x400100000080200,0x0,0x0,0xca02a24391e0bb40,0x0,0x385519c211403306,0xc91d06e9963aa7a,0x4b243440c968ab00,0x8020112290401900,0x52d80800412352,0x209815899c10202,0x141d4c82112089e0,0x1141000001200202,0x3692055c9965a14a,0x0,0x108401c0000040,0x14d14e64d0613240,0x4140103110022140,0x22a62d60d189aa62,0x9d5c20011a5225e,0xcdc36d61d9c03354,0x801043719840a112,0x1410ca0091028860,0x102274800490748,0xcb90eb2ed9c9ab42,0x9401d40c9c20140,0x65686b19030224a,0x55dfa4ea91e5ae7a,0x208050040002004,0x110036c91a02218,0xca52e71790402652,0x286c01091242044,0x54f1dcc9f168a366,0x8020a64351203210,0xdddf7d3c99cea3de,0xe757e116f9ce2a4e,0xeb5e7d57d9f9bb58,0x6c3505ba81e5a962,0x840a700c8050048,0xd6a70d63d9eaa23a,0x8a20720a1040b394,0x405002042004220e,0x100205001080220,0x0,0xb5160c519948230e,0x8020100141008140,0x7fff7fd3d9d3bb76,0xad3cffe3d961bb58,0x98b45378d1cbab22,0x110054200602204,0xd1206a229101a8c4,0xcdd35eba91cb0372,0xff1edffbf9ebaf7e,0xf034d06891512672,0x800240210001010,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x41000040002000,0x0,0x400002000082000,0x0,0x0,0x0,0x0,0x0,0x804010001000,0x0,0x0,0x1000020000002002,0x4400001000008,0x0,0x0,0x20400008100,0x0,0x0,0x3050d67dd1212b66,0xa069777851672332,0xfdb63f76d179abf6,0xe94ff3baf139baf6,0x2bfefe5ed865ab5c,0x5c551120c8412062,0xdd6def7a91f7ab7e,0xca9eaffbd9ffab5a,0xeb55d494d9c58a5a,0x25fcd3c9d1718aca,0xadb35dfad9efab7e,0xf9fdf66fd9ecbe72,0xb090257d9943bc26 },
	{ 0xdbf7fd7ecbbe7d67,0xffffffffff7ef767,0xbffbfffb7beef563,0xf6fedf735bec7762,0x90bf5eab09c67562,0xbab84d620d2c7566,0xddf67b63ebe67764,0xd7dffe5bcb7cf560,0x820e494000202520,0x10020a8000000522,0x959605f80a206500,0xc084270000002400,0x1010b1000080420,0xe2400b4800080540,0x8090060000002500,0x41eca06800246500,0xf290af5020206520,0x8000100000000120,0x80000100200400,0x1001041a10200400,0x0,0x5080055000086060,0x100800000000400,0xa100100a80400,0x888c8000002400,0x10506000206500,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8380240108646002,0xcfdbfe36ef4aff63,0xe7f7fecafdaef763,0xafbb6e3a0d4a7563,0xd5aeb34c0d327463,0x66bc3e708bc86560,0xcbe4f36cca646523,0x6886f9c08b425720,0xffdfdffb9bfce767,0x2c5ff6651d786663,0xfbf67ff3edfe7d26,0xfbdceb7ecf767746,0x14000000a000,0x67bfdf76df7ee567,0x6000000120,0xc2858d6051026725,0x0,0x0,0x0,0x0,0x6240101024840,0xbbf6fe197d4c7762,0x440bc10802884700,0xe9e4b1c39487d63,0x3cfa57f6cffc7f61,0x99cf6d02b1487d21,0xffb7277f9fee7f63,0xc976f77b2ffe6627,0x6cda9a5b88606726,0xbb2453a596e7445,0xd0ce370a0a687721,0xef767772abc47d63,0x6987fb2302ac7760,0x42c24e5a4b3a6f42,0x3c97ffdb1b7e4724,0xa7efb2fda8ec6722,0xb8d67de0d5f86f42,0xe3d9d73c89226762,0xc08a0e110b620560,0x2b88cc6cb486360,0x801177e008647720,0x9ada7c4208647764,0xdbd5fb2ed9667f64,0xc5c3fe3adff86741,0xabd697d24ae86562,0x1c1518292b42f120,0x89eca730afec7521,0xffaa7fffa97ef762,0x944f72a9aace503,0xebdcd55d3bf8e763,0xffff7fefdffeff67,0xfffffffedffeff67,0xfefffffffffeff67,0xfffffffffffef767,0x3fdfffddcffe7763,0xbebaffebe9ecf567,0xf37efffb5dec7763,0xfbf67fefafee7f63,0xcbceffa32bfc6d67,0xfddffe7b0fee7723,0x959785eacfea6d63,0xfd56f2f9e76a7762,0x49a6744208007522,0x8a9422360234a500,0xcc018d8802c6001,0x279b04128a2ce520,0x90f4f6fc592e6d25,0x2002857200283460,0x7ea0ab780bb64760,0x6816320209267500,0x234a51d0cf086121,0x1425526b5938a720,0x100a4320a404500,0x99046224484c6120,0xc6c6a33b6a6c6560,0x21a105088006000,0x8280043a02084443,0x0,0x8001800008a020,0x383080a040540,0x24100c000200a600,0x6002000400,0x71a0026a4a202500,0x5a8e380003882602,0x8884130a0a244500,0xc60a0e118b426506,0x80a004e040006701,0xcc44030802004544,0x8146001082160,0x8b82be661f707562,0x8882720918084503,0x1000490008044002,0x4080049801085002,0x4a00e3018a602400,0x9e84277c0ae46503,0xcfcb86110ff46760,0x2f3ec4c088647560,0x462108408402400,0xd93b3e7015347e42,0x4aa832420a246500,0x8c2837a02386722,0xa3c7a0312ac4646,0xd014a380a82742,0x49d34201426062,0x4206200002282500,0x80900001001120,0x14000300000,0x8002000000240000,0x8050180021084820,0x4400000000000400,0x1800000000400,0x0,0x401000000400,0x0,0x804000000000400,0x200400a00080500,0x0,0x0,0x0,0x80304000007001,0x110000000000400,0x800020000000400,0x0,0x400002000000100,0x0,0x0,0x10200000400,0x4400000000400,0x0,0x0,0x246802300500,0x0,0x0,0x10200000080100,0x0,0x10408200000a400,0x40200001020000,0x84c2120202000500,0x12004e0000000540,0x11b2c24408a00560,0x0,0x0,0x0,0x1000018100000740,0x80040000005000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400000408000002,0x0,0x0,0x0,0x100080000200400,0x804000404000,0x0,0x0,0x0,0x0,0x5491f66cf9ee6762,0x71afee730f7e6762,0xffba6fd74f6e7f61,0xcccffff6dffaed63,0xeb7ef7578f36e767,0x3c331fd3313a4765,0xfd6ef5aeefc8e522,0x5a8aaf661b7af760,0x87338b2113667500,0x82c792a36a48a524,0x412777273b83524,0xdcbcb6fbbfac6762,0xf22a7de41bfe7502 },
	{ 0xfbeeffff9fff7f69,0xfff6ff778bbe7f71,0xfbdfff5f9bbeef68,0x7ebefff20fb73d70,0xbcd936e68e326d70,0x90e3bfd08e9e2d68,0xd7b4777f0bf2cf78,0x21d7c32a123e6d61,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80400002000400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc5dfff388e9e0f60,0x2fcebec996bead61,0x948a5e1b0a18a560,0x90ef2edf17ff8f68,0xd5a279728e264d61,0x8b54e5b612af6840,0x282336210882c20,0x1f934ffa9edecd60,0xb604823004900d68,0xfbfebdf89eba6f69,0xdbf7fbbe8f9e7d68,0x946080003100508,0x879ece5482a7ef68,0x0,0x962c45401aee1520,0x0,0x0,0x0,0x0,0x9eaf5c011920559,0xeee4af6f0ebe2d70,0x10200020100,0xefbfffdf9ffeff79,0xb4b17ff99e942f60,0xbdcdbfc00b2ebd60,0x9eb62d7f93ff2f60,0xced27f1096921d60,0x4b5edfc116b24d60,0x963a7fda8e937f61,0xd9fef7ff9fdebd78,0xfbbebffd0fbf6f69,0x89d6e32a9ab44d28,0x2fdfbe53869e3d60,0x7bbe0d1b1e8a6d60,0xb1f4d75212ce4d61,0xcec475c30eda2d60,0xfbedf72a133e6d28,0x94c9bb4b9e144d60,0xaba8871206c63b20,0x75566907364c60,0x910c0520038e2660,0xdb93eb0d96b6ef68,0xdecffe939ffebd78,0xaffed9d917b67d41,0xd4f356a80f904568,0xf9f8cf6016b66d60,0xdf326e5517f2ed68,0xcb67fb3202fafd60,0xef7edcdf9efe6f68,0xb7367ceb169e6f79,0xdfffffef8fbeef61,0xffffff7f9fffff79,0xffffff7f9fffff71,0xbfffbfdf8ebecd70,0xb7bfef738f9e3f61,0xd1fd77fa8ebe7d68,0xfffc7d7c8ffe9d61,0x8bf7fbfe0f1f4d79,0x6fffba5b9fffef69,0x7fdefff39fb27d50,0x716eb4780e292d60,0xeffeff739ffeff78,0x8010202208120440,0x4dcaba5186002d41,0x9db01d528e9a1d60,0x7090f4dc16b60d60,0x9380020007008540,0xc99ce63f0b2a4d68,0xa28ca20b8e12c160,0x204004008040420,0x50b3065216902140,0x1b8c006a97a34960,0xd38021798f960500,0x90c3ae210a950f20,0x295ac2518a848f60,0x742a1fe30aea6640,0x8041040000260400,0x240001000400,0x8088c00010820c10,0x0,0x401040804100820,0x4020114a00270c40,0x3a84304882840400,0x8a20200a02900500,0x0,0x2200004802080420,0x8001014003860400,0x4ab06c5007b04540,0xab2443221bb22540,0x8800a0002000440,0x84e8c23302bd2860,0x10545ee096160d60,0x10200c2010960000,0x7000292a04361560,0xc1c29a3c040e1960,0x10808d4301ca4960,0x8000046000020060,0x858a215003a80740,0x4300412610840400,0x8848e10c84220940,0x614400010920420,0x346017e904902960,0x8132084001820620,0x98062f80122c0520,0xc51ad3390f863740,0x105002000a002400,0x1010400200020940,0x8074020000122d60,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2010000000020040,0x0,0x80100000040020,0x0,0x0,0x0,0x0,0x2020000060000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4400000000120000,0x0,0x0,0x80010000002800,0x100100902002400,0x0,0x0,0x108000000400,0x0,0x0,0x0,0x0,0x400040a000000,0x800080001000400,0x0,0x0,0x1200000000000900,0x80080000c00000,0x0,0x401000000000440,0x0,0x0,0x0,0x0,0x0,0x0,0x2000002000200400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100002002002000,0x0,0x0,0x0,0x0,0x90b83cea9abe2d60,0x15e6276b168e5d60,0xfe983f6e96debd61,0xcffeff0e9fbe6f71,0xa9efddd30edf7d60,0x149ad56a87b40d60,0xd87f2d7f9ef2cd60,0x9b970f7481be4d60,0xe4a3f24d1e927568,0x22b5ca4606162c28,0x1530d67292862f60,0xc5ad65fa0ae1ad38,0xcab6757d1e934d68 },
	{ 0xfb8ef7f3dfbc3ff6,0xcffcb7fff6847d65,0xaf7fbefb560e3f0f,0xdedefe7cf68c3f6d,0x54d63e80568e3c9c,0xbff03f30d68c3ee5,0x7e8e6f6ad69c67c6,0x182250ac28c2482,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x420000002000004,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe6c7a212fe0c1600,0xa35f1fcaf2167586,0xb22e436256842a6e,0xd06134d556842782,0xdf164734d6b42bf0,0x8352ff78d70c1e42,0xb6cd1b52568c03e4,0x8eaa6ff4d38e370c,0xf5fffffaff9e6e8f,0x7ffe7ff2df9e77ee,0xfb84ff3ff6be7de7,0xa0631086d4147c2c,0x2198ce80100c1468,0x0,0x0,0x0,0x0,0x0,0x0,0xdaaef5f0d69e475c,0xda9d2c3d5f9c2ec1,0x2144c402d4300b40,0x77567dcbf3141536,0xdc171fe1df963784,0x892c5e6bf6806146,0xffb63ddfffbe3ce2,0xc7d6b778f29c2762,0xaaf3ca41c6ac2546,0xb4bbcd76f6be3ce6,0x55b4b7eaf7be3fc0,0xa2123c7cd6bc3542,0x80805202a2020502,0xebe5daf0da2c2ff0,0xbffedffbd6a637f6,0x5424406ad69e20e0,0x457665b0d31c6502,0xf38ceb32d3b07d42,0xdddfbebbf7b477c4,0xef3fcfd3d79c36ef,0x841016c052802d44,0xd7ba1f50d3146f85,0xf9b7f7fef79e7fee,0xefdef69fd79e77ff,0x1a5fddcbf69463c7,0x10563c6ac684014c,0x913fff7af7b40b4c,0xd3280a2ab2946061,0xc122e52a568401e8,0xaa935edbf3382fc6,0xacbb5feadf3e3ce4,0xff5dfef0d7943db5,0xdff3bf7ff79e7fee,0xfdffffffffbe7ff7,0xe1ffdfdff7be3fed,0x9fbffffef79c7fee,0xf7fdf7faffbe7ff6,0xfeb6f5fef79e3fec,0xc9f7ebbed79c3f66,0xedeebe4ed21e77ee,0x99bbcf93d79e3595,0xf5e7b67ad61677ed,0x7bee2ff2ff9e4f54,0xd910ad38b3a00b44,0x800b9820120c2105,0x1cb945ca92943885,0x1080748850000116,0x4536ad71d19c3de4,0x600ce4a8c2182d01,0xa15e8a09d6002d04,0xb5c480292002b06,0x414080005400,0x8417041b1002041,0x204846002000,0x4588d10cc50c404c,0x621ad810d69420a2,0x44a040042080,0xd001200410001460,0x480005040000504,0xc107f90e52022d06,0xe8020fc272082505,0x6604640d4902004,0x76854040080,0xb404ec6086102948,0x85806487041464,0x486080092000b00,0x16101396440c1402,0x9194c002c4120402,0x6212fd22fe182f48,0x8b246b3a56822a64,0x40260042000110,0xc44000090100004,0x0,0x22145f20568003c0,0x4844a32050802a84,0x480216a974842840,0xba705fc8f69c2504,0x20ac5068540c2a21,0xe9028a40d11c0570,0x80824450802006,0x84828200049c2420,0xab4f5f52df9e2d06,0x75a4056a92846b26,0x4042a540b680024c,0x4220348ec2045410,0x4c0c600e2000490,0x6000000040080,0x820a0601c6024100,0x0,0x0,0x800020000080400,0x0,0x8080000002800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100244000000,0x0,0x0,0x0,0x0,0x0,0x0,0x2040002000004,0x0,0x0,0x0,0x100800044000000,0x840000000001400,0x84020410800840,0x0,0x200001010000100,0x2008010000200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x500000030000000,0x0,0x0,0x0,0x8104100000,0x0,0x0,0x0,0x800010002000004,0x0,0x44040208040480,0x0,0x2020000800040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100800044000000,0x0,0x108044000000,0x80120442d59c0460,0x14543ce9d6162ba6,0x9b2e2ff3d69c25e0,0xbfa5f7ffd79e7f44,0xdffdfb9ed78e6bee,0x6ebd1e02d02e1d46,0xe4e15d6dd71c5704,0xc5248460e6140404,0x9ab8a70592bc76c1,0xe7cef732fe0c3dae,0x6fdefd5fdf987fec,0x920acfc71f143e61,0x80522650901c0400,0xd2322b74f20e2d8c },
	{ 0xffdefdf1e9fa73fb,0xdbd7bf6681b82b7a,0xfffb7fdfebba7fdb,0xbffedffd83a777fb,0x34d46ee08bcc4377,0xdbdfe5f08be077fb,0x8300e584e0102282,0x44000080000010,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc4d0f21009c46202,0x6d0f2e5bc39063e2,0x129aa5eb8b907359,0xf77bf67189f673fb,0xd28e7d43ab7c2383,0x8193ee038bd03748,0xaff71f5983b262ef,0x8fce4c08a1902163,0x51a6a0f20bcb236a,0xdb6e6e7989f87779,0xf3f4ef670bfe62e7,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4004000000001200,0xfb6fff718bdf62e2,0x10001002000000,0xc8c20c0003036002,0x17c5f00e9286220,0x179e17ebcb8062a8,0xf11cf7528bd06778,0xfffe2f7fabd96efb,0xf5e4ee078b956bbb,0xfbffffdfcbbd6b7a,0x8fba5fdaabd16bff,0x8514b46a09186323,0x5c2efdec8bb04bfb,0xcae0f303a9c8271b,0xbbdfdf48abda62f1,0xfa9fa3cf893067fe,0x9520f24183b06203,0x528438c089c423b2,0xf9e4d31e8bb163ba,0xc3ca9efa89b26bab,0xc8140c1200b03210,0xf276efabb063a7,0xb2162c7049846eeb,0x9bb4ff7f8bd477fb,0xcfffffbfebfc77ff,0x6bff9fdbabb26afa,0xdcb7527e8bbb620b,0x91ffb761e93c6e9b,0x8bb41e7d8bbc3211,0xee7eef39abe563ff,0x6ffb7fcf0bb967bf,0x863efaebcbb56332,0xff7ff77b8bfc73ff,0xf2b02f5f8bf06bee,0xedd7f7a98bf17777,0xbfd7fedbabbc6373,0x1e9a6feec1307be8,0xdf95ffea8bfe63df,0x7fae7fefabf866b3,0xc9f3eb79ab94669b,0x4fd73e16ab310a02,0xd79a97b58bfc2772,0xfdf7f7faabca77fb,0xdda06d60c1b47372,0x91d7cb0409344200,0x80c39bbfcb186602,0x22e4850a3100242,0x6444b4708b802403,0x49023401a1902002,0xf914ff0488d86299,0x47d72a12c94065f0,0x3d2c08800a03701,0x70700c728b82220b,0x824450088604202,0x9000220303800028,0x8090222280803209,0x984801a2804010,0x200f201220204,0xd0e034000ba00330,0x8802e148080210a,0x8780a01908f82201,0x864c00040002301,0x16984128012060a3,0x8454140089806020,0x92240d0c81102351,0x4020000002001,0x21c0082002011,0x288086e8a8d02b02,0x9424702803186261,0x4202259001884220,0x8990822d83b02212,0x108000100200,0x3846240b800218,0x749d18280a806331,0x9200000000006001,0x8b2234070b8c2271,0x44d22e2101940b23,0x35f5d0943982a21,0x804000800001,0x2002032000a03220,0xc20046008002261,0x4e0224440bd20260,0xa943180a8191421e,0x9080552a8b004122,0x5040228888882202,0x8e102d1ce9c06388,0x41809010000043a0,0x2a1ab0484b902201,0x1491492888506280,0x10c0110000802121,0x0,0x0,0x0,0x0,0x0,0x0,0x800000808400000,0x0,0x0,0x0,0x0,0x0,0x8100000100200,0x201000002,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x204000100200,0x0,0x0,0x140022000800218,0x804001000020,0x2000002000080200,0x2010000080002000,0x0,0x0,0x0,0x8400800001,0x0,0x80802101040220,0x2020000000208,0x0,0x0,0x240000200200,0x0,0x0,0x804080002000,0x0,0x0,0x0,0x0,0x0,0x0,0x140000010002,0x1000010000000200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x910a4b918b884b32,0xd7357ff0abc663e3,0x51e71f694bd06b93,0xdfb3772e8b997373,0xedd7ffababda7b33,0xab7fffdacbba77fd,0x43356a308b052e6,0xcbcda56a8b0d6b2b,0x5604084402106320,0xcec3d82689996b13,0xef5e1fce8bb463e9,0x15105802418060f8,0x354d52108980422e,0xc682032482802618 },
	{ 0x4be77f68fb6dcf6e,0xfbdeffeffbfcd77e,0xe4d8fbfae174d73a,0xbe3effffe3f5d7fe,0x100102c201044128,0x90100d408284912a,0x0,0x0,0x0,0x1000008000000108,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x44800010d0000004,0xb5df3efbf3e4db26,0x82806f4a41c4832a,0x807abf50c054592e,0xc1c26f2463c4c20a,0x8b84f2f4f3f4d36a,0xc8065e864844c11a,0xc69c75c240d4c33a,0xfddfbe7f7bd5c3ba,0x44ac7d785145c35a,0xfbefcf3ff1e5d77e,0x0,0x0,0x0,0x0,0x0,0x0,0x400400040000000,0xd444f27141c5c10a,0x712e0a6a83d5d16a,0x0,0xca43a644c854c11a,0xa3d5dbc9e174d132,0xb43f56ffeb04cfa6,0x90b8e676d1e493be,0xfb962da5c3e4c32e,0xcfc85631e875910a,0xefdfffdbfbd4db3a,0x96ae47f6ab45d93a,0xd567f7f6e3c4d1aa,0xfbbeffecfb7ddbfe,0x8bf0aa33fb55c91a,0x68cf1fe76984c32a,0xbf8fcffbeac5d93e,0x88c5d67a83a44336,0x4cd62678c01dc31a,0xfb40fd6afbf5d32a,0xc4c08030c0c19000,0xa9c884258540202,0xd84025e5d354412a,0x933ee376e3c5c36e,0xd9bcef7fe3f4c33e,0xcfd7bf3fd9c5d33e,0xbfdfffd9fb55db3e,0x3e7476c4dbd5d3aa,0xc57eb77acb78531a,0xfb963fffe1f4d33e,0xedf6ff3edbfdd1fe,0x8bdf175bdbe5d3ba,0xbff9fff9db6cc7fa,0xd3d8afdaf975d17e,0xfbbe3fc7ebc5d9fa,0xe9d79b1ffbf4d3aa,0xefffffdafbfddf2e,0xf7ff6ffef3fdd73e,0xd771ae7ee365d33a,0xd3fe7774f324d13e,0xab567fa3dbf0d11e,0xa206bac3e3d4da7a,0xbbbbdf62e9e5c96e,0xf5b5f279e3f4d726,0xeffe75e2f975d252,0xa3d6e3377b4dd30a,0xe84200b80240452a,0xae9fce67d8f4db3a,0x414560e050d806,0xc3eced31ebc5c19e,0xc0b5f33183544d36,0xfdefda39f3dddf3e,0x218441a210c4202,0x94905274490dc202,0x4949076289044816,0xc80639386105811a,0xc8d0a38a6104c088,0x940825bc0f8d108,0x2808010000c00a,0xc020a5108105d00a,0x2a0045ae104c10e,0xc004b21e8b005322,0x811129461041312,0x140868000150c020,0x0,0x240001000080,0x12000000120,0x120040000000,0x18014c4220504042,0x801000000108,0x8c2a2063008080,0x82857002410441ba,0xe0618e810144837a,0x8d9121705984c11a,0x100001c000048030,0x840007c04344010a,0xc80420028101d000,0x8ce9580003044428,0x30a0d404250c10a,0x23116915801c102,0x4c8020468060012a,0x40223e480b44c42a,0x45c7a7aaebf4d11a,0x4184d508070811a,0x22821c1005000,0xc13a2468e905d92a,0xdaae2d784bccc10a,0x4d5e6210ea84d12a,0x8564bd08844911a,0xa77c4632d961c17a,0x90d138356844052a,0x0,0x0,0x0,0x0,0x400020004000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2004000001000080,0x0,0x0,0x840000000000102,0x0,0x0,0x104002880100102,0x0,0x0,0x0,0x0,0x0,0x4002000000c00000,0x0,0x0,0x0,0x0,0x0,0x0,0x20840008000,0x100000240008000,0x0,0x0,0x200000200004008,0x0,0x40201000080,0x1a0100061000200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100020000040200,0x0,0x0,0x0,0x10001000040002,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20080001000200,0x0,0x0,0x0,0x100800080001000,0x902d9d2e1c48b08,0xf4d434abf3d5d5be,0x971a9768c3085d2a,0xff74373ee3e5d33a,0x6dce720fc1f5d18e,0xaa7effdaeb74db3a,0xb47e7febdb7dd77a,0xf9567e63f30dd1be,0xfebf7f75fbedd77e,0x814e70388b204b02,0x43dd9ecbd154d33e,0x16b66ee2d0e45d3e,0x711cdd7bc2f1d52a,0xeb065d644b05d23a },
	{ 0xeb90b746c040aff0,0xdbf5e72ec0d7ba64,0xdcbfffd7c2dff3e6,0x75b4e4ed88c7b320,0x30904bc28082e220,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xb6baff2e4adef3b2,0x206d12680a12b260,0x4b062d758242f3e0,0x69a7b236cac3bb90,0xaeff5ee9cadeb336,0xfabfefafcad7bfe6,0xdfffb67fcacfff26,0x8000050040008220,0x0,0x0,0x0,0x140420040a408200,0x0,0x0,0x0,0x0,0x50957ee1c097ff62,0x112c006a0045f102,0x0,0xcdcec52acac3af64,0x2bf6ce978296f3e6,0xad977fe7c8daf626,0xfb69bfe2ca9af340,0xf7963fe4c04bfb62,0xcbcfd77bcac3fbf4,0x29dafbcf4ac7bb76,0xd7be5ffbca43b7c4,0xf7edf759c2d7bbe2,0xff4e7d79cacfff66,0xeb8673e64ad3fe32,0xefefbf7bcadefae6,0x3b97cd6a48c7af64,0x2041002000408030,0x4fe275db4ad2f2e4,0xd98f7becc2cbbf40,0x81021e99c0d4b8a0,0x71cdf5242c3ab40,0xf0009460c8020e24,0x800aec60484297a0,0xfab1632ec2c2b2f0,0xc9deb2abca42faa4,0x8a4f47cf8287b7e6,0xfffffe6bcadbffe2,0xf9eeb7f8cadbf6e6,0xfe9e7f7dcadbffc4,0xfdd7ffeacacefb36,0xafddffdbc2c7ffe6,0xf4ff7fedc287fb64,0xdffe7e58cad7bff2,0xde94bff7cacbf776,0xcdfdf66f8253fee0,0x215c7f51c847ab36,0x9bbb5f7acac7f7a2,0x95fdffffc8d3f7d2,0xafe6fff0ca5af3b6,0xc889fa23cacafe60,0xefcf3fdbcadbfef4,0xc3ba4f52cad6b236,0x423452680a12b280,0xc7bc3df2c2dffb66,0xd976aa7ec2cdff36,0x8ad33ec3c0c79b72,0x41215168085a022,0x905c54100a40a280,0xe2d648842a3c0,0xfe80cf688ac2fa24,0x88c6ff6f8adef322,0x1bd0db508a92bae2,0x5c9156e68842ee20,0xe220060040868202,0x88945bf840439fb0,0x38cf202880273a0,0x804000880402800,0x4806432c8803220,0x402482c00ac0b002,0x1210213340818620,0x460041b800438202,0x60c22a298a5d5202,0x9405042040c00a00,0x8088240240410220,0x2000040000008220,0x4000800002000200,0x8000000280000004,0x208d01048c00004,0x420400440020204,0x108082000481800,0xb06430c4042a236,0x2002aa00c2005600,0x1004000101000,0xd0501ec088425a10,0x8148fc704a03a222,0x212840c00820,0x4c4f3292ca93f764,0x2b64640a48099504,0x3480142e804ad940,0x14a07408a10b220,0xf685b6404ac0aa14,0x81cc284802a232,0x48009048000220,0x20e24a24002a602,0x503028c040c08010,0xdfb66fffcac3f7e6,0x8022280302485210,0x800080000020040,0x303a0ca34206a0a2,0x53be2dad8216fb22,0x0,0x0,0x0,0x2080000020020,0x20020002000200,0x0,0x0,0x0,0x4040080001000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8400000020100,0x60000008200,0x0,0x0,0x0,0x0,0x806800430040,0x20040000402000,0x0,0x2000008000008200,0x0,0x1400000008400000,0x0,0x0,0x0,0x0,0x0,0x2010002000200,0x0,0x0,0x0,0x80040000002080,0x0,0x4100000081000,0x0,0x0,0x0,0x0,0x0,0x120000020020,0x0,0x0,0x0,0x800400000081000,0x0,0x0,0x0,0x0,0x0,0x0,0x2200000008100,0x120000000808000,0x0,0x0,0x4002000008100,0x0,0x9ebd4fd1cad3db64,0xfef6fcf7cacfff76,0xb9670d61c893b362,0x8bba77fdca1af6a4,0xbaabfe12cacbf7d4,0x8bdefd5e8ad3f7a6,0xd1b4b33c242fb24,0xb9de75eac2c2fbb0,0xfaff3ff6cadbfff4,0xc94293284054bab2,0xaef9ab1ca5afa04,0x91324c738240abe4,0xd5b13e58c24bfb66,0x53a007c142c3a220 },
	{ 0xcbde6f674fd3fba1,0xe3f6d273cccb3a0b,0xced3feda8ecab3bb,0xeeb9fffa4da1bf13,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x102c70628cca2280,0xc2b61f7448823b13,0xd9c0d16b8ec33a11,0xaffabd5bced6bb01,0x4dbeefdf8fdffa83,0xf4ffabffeddb7fab,0x10c27608d023e11,0x0,0x0,0x0,0x80040000000220,0x0,0x0,0x0,0xa005308601201,0xa0056828020392,0x0,0x8000112000022280,0xd8c3b6188e833a11,0xea56de95ce963609,0xfebb1fe7ced3bb89,0xf124f760cfe19f81,0xfabe2c708ef33383,0xc9aac32d0ec3fb91,0x297eecdbcec3ba89,0xb6bb7feb4fdb7f9b,0xd47ebdff2cc3de9a,0xf6744ffc6cd2be93,0xebf4f73bcdc09f81,0x6e765f5f8ccfbb99,0xff7fdf7eaffbfebb,0xd40574690e921899,0x4aa24d618d92b910,0xb1a4db868dea9a09,0xa4213590c0b3b00,0xb1047830e533a01,0x9054bee30e92b891,0xdebffff3afdf7fa9,0xcd92e313adc33b12,0xc58ad2280e8b8399,0x881acf5b0eda3e99,0xfffc5fe0eff3bb93,0xd7b67f63cdd2ba89,0xfe9e3f6c8e8bfb9b,0xcec3fe3b8dcbba99,0x2acffd538cc3fe19,0x56975d6aeec622b1,0x97f9ff5feedafe91,0xdebe3f6a0eebbb23,0xcec0fff20df27f31,0x6f5ef95b4ec3ef11,0xdcab4dfb2efbfa99,0x716e366b0f533aa9,0xefb63c7a4443fa29,0xe899fbb88783de01,0x4bfa1e1f8f923881,0x83ceb20c0a2313,0x1444a6620e024a11,0x970a2d418ea2b383,0x3a76631e4ec31799,0x4ce93e832c0033a1,0x288c40986422201,0x1091bec608632381,0x5762b30c4ccba09,0x2a50c06c08416b80,0xcce6d2b50dcafa9b,0x200a18520a4002a1,0xc818e06026c23281,0xc14027080c503688,0xbb4170088021e10,0x86114a1009421a91,0xb56fa528e433782,0x4224a050e4b0281,0x0,0x92256886812221,0x4d81f40007820810,0x9cc76f7985cffe19,0x1000080040000001,0xa10050088808080,0x0,0x88400004809008,0x204000008000001,0xa00ce488d809880,0x80320e2188c00b00,0x4808641200521603,0xc9f0635bcd83ee81,0xc862ea200c92bb13,0x2832012080020600,0x4022804208001811,0x2a340008109101,0x401022020a000701,0xd407a2200d01221,0x800080008000010,0xb010188204c00202,0x98602c0009406801,0xa130054786820001,0x7c6ea68ab434e93,0x190ca100e819819,0x90542886820010,0xd240a7180c949a09,0xf5140c7a2e877a93,0xc120820146103001,0x8ccdc028493da09,0x908805894e823081,0xd124a762adc23e38,0x0,0x0,0x0,0x0,0x40800004800,0x0,0x802000800800,0x80080000000600,0x2000200000201,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8002000008000001,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100800000000202,0x0,0x0,0x1080002000001,0x0,0x0,0x0,0x0,0x1800000000280,0x0,0x0,0x0,0x0,0x0,0x800000800004800,0x0,0x0,0x0,0x0,0x6000000020000200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10020000000220,0x10100004002000,0x0,0x0,0x0,0x0,0x0,0x0,0x1000000480020000,0x0,0x0,0x60080020000,0x224e51daa5c61a09,0x74d7ffed0cdabe89,0x8071245281810e00,0x8084004000001801,0xc34afeb109833e32,0x494cddda8fc3ff09,0x14162b034e422219,0x936bf772cfd33b9b,0xfeb7bf7d8dcbffab,0xcb41dfb56cd3baab,0xa9f6dcc38cc3fb10,0x2cb267cc0dd23a09,0xc02492190583a800,0x142626600f82a811 },
	{ 0x8106f761a25a16ce,0xcb76cf362676174e,0x4af3ffebe37e17ca,0xe38dcf7aea7e90fb,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4eb82c50206b15d3,0xcb47f2673a5617eb,0xcdef5f5faa7597fa,0xbabb6ff2a67f15de,0x700032e922521512,0x54ae3de0227310c9,0x802000400001,0x242301322101444,0x24024026000000,0x0,0x0,0x0,0x0,0x8284010085010c0,0x20004000600000,0x0,0x281820200101482,0x88c5faa20275144a,0xa4fc5c56226414da,0xa1bae76a677314de,0x556eaf40ee7313ee,0xdaab3df3767713da,0xcb20f77030529558,0x6d3e3fc820239150,0x1fbbcf79257713fa,0x71efa77c315f17de,0xd0ae316165711459,0xa1e6f387605211de,0xef5fff766b7b91fe,0xb9ba8d58ae6a10c6,0x15cea3e6207a12e2,0xcbfe1d00226c00c2,0x93165b34e27a80ca,0xaec23c9a6e52154e,0x433b4f43227210da,0x4c0680020a1140,0x118a25607270114a,0xd3822f36e26100cb,0xc6dbda30b65f15da,0x287e9f5a2a6b115a,0xf4b656e1a35416ea,0xff7efffab37f15da,0xffae6f7faf7b11ce,0xddfefe3a3e7f97ca,0xffdf7fd4737e91fb,0x4c75c69a252114b,0xddfe7e53777417ff,0xdfbe7ffe6f5f95de,0xed86e72c7a7b15cb,0x29f79edbb27694ce,0xa4ba7eeee36214da,0x4100056be2521482,0xffbe7fffef7f97ff,0xd9f7ff7b6f7f15cf,0xeff6bfcb377695ca,0x687c568325610d0,0x44033790a70000e,0x4a4a2460241210d2,0x3a4213e267106ca,0xc7cbb33fa37a15ca,0xd4e4000450076,0x79fdfffe657117da,0xabbee77326771548,0xdb9d47e6235215ea,0xcec9363b387f17ca,0x331addd3225a11ca,0x34bd76e5337c154e,0xa9c45e68317611da,0xca002a4fe008114e,0x3909000200240,0x106d14000700140,0x102020200211440,0xd020200200701048,0x9a940f14007210b7,0x4180801800671040,0x87484002121042,0x411000224001100,0x91896400202211c8,0x0,0x100020000000006,0x52023223011048,0x80004e1a62020000,0xd2109c08a142000a,0xc70634202150145a,0x81e6f3222a52114a,0xcce29f15675294ce,0x20014060101000,0x836520020521482,0x80082020200000da,0xc08053042052000e,0x84222e0a0250111a,0x304001000521000,0xa4320264207a1048,0x46a41801481148,0x2800a46001a105a,0x7018e28206010c9,0x40875900a2621442,0x10220c80a0521080,0xa8c64020700048,0xfeb66f7ee75615da,0xa402d20224501080,0xe814f3c2622a1740,0x2000e5b226a0148,0xd544af53237184c9,0x0,0x0,0x8200420000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20860000000,0x4000400021000,0x0,0x100200080040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000046022500000,0x0,0x0,0x0,0x0,0x0,0x4200100040,0x0,0x0,0x0,0x0,0x1000080000021000,0x10001000000050,0x240000021000,0x600000000001100,0x1000010000021000,0x0,0x0,0x0,0x200200101000,0x0,0x8100021000,0x20040008400000,0x0,0x0,0x0,0x0,0x9000000000001100,0x0,0x0,0x0,0x0,0x0,0x0,0x8000024000080048,0x0,0x0,0x0,0x0,0x81000000000048,0x9763f0b6e5994ea,0x62fad7d326729042,0x34b95568677a11fa,0xc120840020500009,0xb8424106010110a,0x4800070e607210c0,0xaa4231db632115ca,0x14f657e3227b11d5,0xf9ffbff66b7f155e,0xfff6af7c207e95df,0xd994e73da05715d6,0x265fce5a024d155e,0x44a6ef0275211c8,0x9410880024211140,0x0 },
	{ 0xc9b27fcd063ecf96,0x9bfbdf7a9b3debbd,0xc5c3fbf3da3eede1,0x20ca1f8002568682,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf100011400308a82,0xa9c8f0c21a2d484c,0xdfd77f3d1b3dcff7,0x189feffe5b3e69dc,0x4556328a922f4950,0xc8386660480c80bf,0x0,0xe14212594c176048,0x144a50883de894,0x0,0x0,0x0,0x0,0x3908c98189344ac0,0x0,0x8000200004004000,0xf29401841a0ca0c3,0xcdc3fe8a4d356af4,0xa855d052086cccd1,0xf4f277ea1f3ce7c6,0xf95caf72c93ceef4,0x4e9f3f659f17cbdd,0xe174f6888d3cecec,0x816fece1a6e2d44,0xf3220f035a384aa8,0xd1789e421a25c8d0,0xc80674000c314804,0xebfffffe5f7feff7,0x6fffbfffdf7feff7,0x3dbcdff94f3ecefc,0x3022564c403c8092,0x200680204020c,0x8b00f5aa1f3c4a58,0xaeca935b4b3ceb15,0xc4ab8d6b520dc59d,0x51f2dcba0e1ee5e2,0x8c02256089246c80,0xfe902f464e3d4f90,0x894385038a648c42,0xfbf8dd5bce7defaa,0xb4fffee5cb7ccef6,0x93eef7790b3fcf86,0xdf127b7c037e4f90,0xfcdc770c973dcfcd,0xff7f5d5fdf3fcff8,0x21ba430a0a56a18a,0xc964a4021715e302,0xff96af74d73ecd94,0xef60f6e35a7c66d8,0x1d6fed2ca36ac82,0xc925ff28a3d4a64,0xd220a0d01f2d4a38,0xf67e7f75433eaef4,0xcad3e32b5e3fe4b4,0xcff7ff7b8b7ecffd,0xbe3b4d52cb5cee80,0xd592f22d5b6ee890,0x5d6e7d72d33e0f84,0xd8c0e72e423fcad3,0xc6fbfeabdf3feedd,0xef31c5684b35eac2,0x80be7eec4a7c81f9,0xca8865430b3e8b49,0x828463e0130c4840,0x9020681083680e9,0x1c7655460a36c8e9,0xc41502fcca3c0b64,0xd94ce6731e1d49cc,0x80101378045c4cd0,0xa106db1a493c88c5,0x91e1340130dcc9c,0x4d3849a04a34ae88,0xc95287401214e590,0x2284002040100,0x811099100e04490c,0x7a1202000900c0c0,0x100002634014c0d0,0x100100100080,0x0,0x800820240104800,0x2001088008080088,0x16095791c33f4ac0,0x4c1ca8a8011c4ad0,0x402c48e240348c81,0x8bc496650a1c6f58,0xe94233d19b2c0adc,0x101841011216000c,0x26140a0000c803,0x60000014000,0x200c00801044004,0x4d2202041144600,0x8c0900180230,0x10034001104200,0x81063180040cc041,0x170028610b64cc80,0x8c80ea0243144c2c,0x484458484a1c00d0,0x502a401000a0,0xc10ab4720f24c70b,0xba922a790b3dc0de,0x2004611412348181,0x87643a580a3ccb81,0x122b47424015c8c6,0x164242a1a35c0b6,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100800000080080,0x0,0x0,0x60000002040000,0x0,0x0,0x0,0x800008000440000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x11000040000800a0,0x0,0x800000800040004,0x0,0x0,0x801001004000,0x0,0x0,0x40000808200000,0x0,0x10800000000c0000,0x820404340184840,0x8000020004004000,0x100008080,0x0,0x0,0x0,0x30000040100,0x0,0x0,0x0,0x0,0x0,0x0,0x60008002000,0x0,0x0,0x0,0xc00000080040,0x120012000000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xcc63e11ca7fcfda,0xad9eccc29f7deed1,0x86fb76600b3febde,0x8246802040d4410,0x91901356163dc216,0x40038150083608c0,0x375eddc18e34cbc2,0x948b5de34e3d4a6c,0xb96f2f68d63eefe2,0x3a963fb5cf3dea92,0x8c5a63048b046948,0x4055504c0838a2a1,0xb633fff31f6dcfd8,0x8bad1600e24c194,0x0 },
	{ 0xc3aaec60503a0560,0xeafaf7beda1a4ff0,0x30134a8343088551,0x1012222100188513,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000f0a1413a0c20,0x4e02376c9a0a4d93,0xb6badf7828ba8fff,0xdcdca7de8128c73a,0x404800000101a,0x0,0x1b6a13df61b28feb,0x1a044c0850024c64,0x20400000000a01,0x0,0x0,0x2501012001804a2,0x5f187bd2711a9dfb,0x4040de4089a2c501,0xeffefdf3d9ba0fff,0xd1263d66c902dff9,0xeed277baf9bac75f,0xffeff57f9889733,0x1cff5debcb8adfbb,0x11aa26e820105ffa,0x8c8a3862f20a057a,0x4dd3faa24b184d32,0xae4e3fd953ba5752,0x482511e098a1552,0xb1bca213490a5f70,0x7a960d4cc11816bb,0xabfefbfd53ba5f7e,0xecdfffffb9aa9ffb,0x1f9dcffbf93adf73,0x9798aa70932a07df,0xb00040000084802,0xcbf4e32ed91a4da6,0xeff3be797a9acfbb,0x63ae9d121b0a1f33,0x54d356e488088733,0x3b03d50a118853b,0xff34776dcb9a8f73,0x5edbbf9f68ba8fe7,0xeffeffcfe9badffb,0xf5b7fee1d8ba8deb,0xdb7f6d6af33adffb,0x9b986ef5500a157a,0x7fd7fe3a9a3a9f77,0xe09f95def93a8ff2,0x26be5ee971b81773,0xd5e9ee5702188ffb,0xc6862f77d39a1d7e,0xcdd6f3d7db9acffb,0x4870da61220a8553,0xae1247e1908a8f01,0x61401619a2101550,0x73d62d17f19a1733,0x49d0f29140000fab,0x6fdfff70b99a1f77,0xf3ba9dffdb3007fb,0xb57f67fcdbba9df2,0x4fb62d704a1807f3,0x9bb7732f3092c77b,0xdf4f7ec9783a4feb,0x2b9ec7df3bb21ffe,0x9041daeac21a0f7b,0x8288e5b330ba07eb,0xc0a5f6e4313a8f43,0x4f03f60400ba4e4b,0x8b975d1e112a4d4b,0x52233ee80a084b13,0xe81003a882080751,0x3ab06e7b430acd7a,0xcfc3ffb359bacf23,0xa850929200029b5b,0x88a030000180702,0x2000006200028810,0x480024a080820700,0xa004882080080520,0x485790c30b824d12,0x2020000020000100,0x1143804a189a0d02,0x9010001440020124,0x88904024800a0436,0x0,0x30084c1220108591,0x8102a2c0032202f1,0x85007d40c00a8572,0xabe0f72e6812cfbb,0x2100000000201,0x800205001100500,0x9013000480001750,0x0,0x48080722880085b0,0xc658b0a240204f49,0x108480a02080413,0xc000168540080323,0x40230901180402,0x9a00130c13020cf2,0xa15488000018040b,0x4a5805599008870b,0x80b644e0910845a3,0xd9001602020a041a,0x5400004801200702,0xa16404201ba20591,0xe97cfaced39287fb,0x18220a8001000302,0x50205004c0080528,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1200000000a,0x0,0x8100001010,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80000801000010,0x0,0x0,0x0,0x400010000000a0,0x0,0x1002000000100,0x0,0x0,0x8000000100020040,0x0,0x0,0x9200240200028251,0x0,0x0,0x0,0x44000000120,0x0,0x20000200000081,0x6000080000400,0x0,0x0,0x801000000120,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x21230026002054a,0x1217c1134b008772,0x72bee6eae28a4ffb,0x9d77ef731a9a8ff3,0xfb9e277db19a0f3e,0xcdcefe8ec81a8d6b,0x61855b588128c7eb,0xb0f25e6a99828747,0xd440b40c6008850b,0xfeba3f7158ba9d7b,0x4fdff37e988a8fdb,0xaec4ec498310ddf0,0xb4bb6eee9308df1a,0x8806002008010,0x78b61bc8011a0d3a },
	{ 0xeff7fe713cb97eee,0xcbc6d53e0c9a73e6,0x8008880004020104,0x4a22500c220264,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2957be0010a84a42,0x4ff216551d3b5fa4,0xb6fa2de035bae760,0x1422a2c40c280240,0x128490204086840,0x2054c524849018a4,0x5eef9e1335bbd7c6,0x38190000aa0004,0x0,0x0,0x0,0xcb47b7333c8b4ec6,0xa34f59cb3cda78e4,0x94767ee03cb94dc2,0xbdd66f633cba5fe6,0x430e2d2104194d80,0x986864004a81280,0x8111581211b066a2,0x36f676f0343a6ee6,0xfffeff77b5ffffee,0x9aae3fe19cbb77e6,0xe9d4f32c34bbdce6,0xa97a7a0b14b84ae6,0x588189934982d80,0xc5380e00049a63e2,0xdb8effe73c977ee6,0xeb1af77b15bbefee,0xecf7bf5fb5bbfdee,0xbfbedf7bb4bbfeea,0x1d5efec034026aee,0x82a86c43353817e2,0x43ccf01e153067ea,0x6ecbba3514ba7ee2,0xdf56d7fbbcbaffea,0x148c024000866044,0x93a6273014133ae8,0xee94654035f8fec2,0xcdc7be2d35ff6fa4,0xb877dfdf1dbbfbee,0x76fcfeea9cbee5c6,0xd9c1766e1493dbc6,0xd3346fee35bbfbe2,0xd4fedff53cfbffe6,0xcb17df5a35fbfde6,0xbaa45f2814c25fe6,0xdbeda57b153370ee,0x98a6237000ba4aa4,0xe1b282041482f262,0xccd85114a85a40,0x428887a51caac046,0x9240164604127060,0x97a62d451db86fe2,0x1ce7392913b64e6,0x66953bd0941b7a86,0xb6bfcf8c1df36ae6,0xf634c66e3cbbeae6,0x5ab2b65014d337e8,0xb9464e2624f007e4,0xafdfbbd134b3f2e6,0x9f9acf723cb36ee6,0x71264e7c34d3f4c6,0xf2beadf034bb7bee,0xc3207f39349850e6,0xdd5adefe8cb370e6,0xad5cd18b1495dc62,0x94b474a8053137e2,0xaf64566a9c196fe6,0xd2a2037e10abeee0,0xf5c7bbbc0cbbfdee,0x9c8a14134911068,0x53c04e4141028a6,0xe0758622b53359c6,0xc6106a42202246e6,0xe0142402208800c4,0x205e9400002801c6,0x200040010000002,0x8000002800101140,0x41000000c000000,0xe020a266009a7066,0x848822481c103400,0x29cc1710cbdc482,0x2000220100120024,0x90c040008394000,0x69a1d12a20325fc4,0x662300604938064,0x4103004001100,0x1022048400022684,0x80e2d701888c560,0x0,0xc4870013c105000,0x200001a00900000,0x50daf6ec3484dbc2,0x41ba2d0010d06042,0xec102c0024884940,0x6056c20000024866,0x880804831200002,0x3c150160101140c6,0x8110420a14909002,0x2082c0200200062,0x8582f62804127ce6,0x7b7713c614d2ffa0,0xa224f491411796a,0x84a4b60224f961e4,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4080000280000,0x0,0x0,0x0,0x0,0x8802388000984040,0x44030000000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x42000008004000,0x0,0x0,0x2900010000040c4,0x10800000280,0x0,0x208000000002040,0x0,0x24000000020200,0x0,0x0,0x0,0x2008000100002,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x802000010000002,0x0,0x0,0x0,0x80040000c000000,0x0,0x3d7ef5d215bbefe2,0xe4fff47834bbdbe6,0xdfe7f77934bb5fea,0xff9e3f8435faebee,0x88d3ea2a34b666c2,0x289819800093e064,0x94275e663cbbe9e6,0xd81cea121c83e1e4,0xdfa42ff534b979c2,0xaac7fb923c3276e8,0x2ed29fd91dbac1c6,0x1034042101120c4,0x4020208000084480,0xf612273639b16cc2 },
	{ 0xcbb6fffffe34f6fe,0xfbd7b2261cb45edc,0x3babefdbf479d77e,0xf1f6da6c7e44d6e8,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8b64c4044c13024,0xdfffcff7f8d5d7fa,0x2416622008901a02,0x0,0x0,0x140380084106020,0x14000204000,0x0,0x0,0x80c00a0001166a,0xcbe3b293b6341fe8,0xa61ecadb1e9dd760,0x309164ea10745d5a,0xd9bebd4a26bc57f2,0x42802e7c34d01228,0xc54f680284345c62,0x2101403048014c4,0xc9124601c143680,0xfd6dff63f63cf2ea,0xfeb62f7f2e9c7fe7,0xefeedb12343cffec,0x259bfb7b887c56d9,0x9f2055dcb494d422,0xd48d343b14a414ce,0xfff67dfbc4bcdcfd,0x89deef3fb6355ceb,0xeff77bdf74f4fcee,0xdfffdffffe9cdefe,0xf66ff7fe7e3c5fef,0x2023c61101cdd41,0xcbc4d31e167c5c52,0xeee3329be49cdefe,0x8b144e7af61df6e8,0x1464074010110c25,0xa12e2d6034157ced,0xc984c7e6201cc6fa,0x9ecbfe8b16ddd7ef,0x4e5eeb52341cc7f8,0x54fbfee4f4b4f6ec,0xb56bfe78a63c1cee,0xe3a7f7dabc71dca8,0xefd6fb9896bc54fc,0xefbedfdcf6fd5dfc,0xbc5f5b7ab090d678,0xfdce75563494f3ee,0xda802d43044516e6,0x438bf76f54b4dc7d,0x8008000000c000,0x2000602c0001028,0xe0cfb2792020d6e1,0x429001a030101228,0x2b64662224b0db6c,0x22a63ad0303c5ee8,0xefffdf7ff6317efe,0xdffcf6fda494fee6,0x6adc7df3c87c5fea,0xabf5fb3fb0dc1efb,0xeddbfbdfb4f457ff,0xeb3c477a2c9c7efc,0xf1f7feed343c5ebe,0xb3aafd72b414d6aa,0x9b325d7414d45dd8,0x89f7a70826b455e9,0x2ffe9dcb545476b8,0xf09ef66cd010d6ea,0x15670368604514a0,0x12186e2c4041d424,0xc1526a02b4407420,0xef6d9d034141e66,0x81aa03600414226a,0x815046080804d082,0xa42650801c00a2,0xc063e6288004d4aa,0x694412050435ca60,0x800030020000028,0x0,0xe50c71210010165a,0x80807012042015c0,0x21600001c03c0,0x61e4002201c50a0,0x810000004001000,0x2004e4600200c240,0x4144830e401c442d,0x84cab2084010546c,0x0,0x62040014040080,0x4180158040001468,0x8640806001040,0x8041061000100e00,0x0,0x100280000020,0x4082010002001042,0xc210224044204164,0xc20040040028,0x118440280101498,0x242442402040102a,0x95895f402210c668,0xdba80d46b414f26a,0xadf4f2721e35d66e,0xe9dede5dbc155efa,0x8382c7582418dcae,0xc1521662c4354668,0x0,0x0,0x4000100000100010,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1920014001400,0x0,0x0,0x100200c00401000,0x0,0x0,0x0,0x0,0x8001020824002020,0x0,0x0,0x0,0x4010000000104000,0x0,0x20040080000020,0x0,0x0,0x8000000200100010,0x14000000480,0x0,0xa000000000c0040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4008000000240,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4800000401,0x0,0x0,0x0,0x0,0x0,0x208c314000028,0xd4f4fee234dcfee5,0xdfede76bd6bdffee,0xea362f71661469ea,0x6846c23ca05954b0,0xaa18ced3805c5ce9,0x3d3a798b7c1c7660,0xb9dcff68fc3c56ee,0x82b63d7034705e6c,0x106cb3ce404f02e,0x184006890400146c,0x1a020808400110a0,0x108a72c101c1c0a,0x7434277c84e054a8 },
	{ 0xb996fffff96eb2fa,0xeabf3effd18eb2ea,0x831c4f30586082d0,0x40009840102010c0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6f7f5bdf51ed236a,0xd5fae7ffc9c9276a,0xf1fefe7988c9274a,0x8504416080002002,0xebfcf36a89c92762,0x4c6232b0a0c52242,0xb6bf4c72f06ca360,0x100476c6f0ec1280,0x0,0x0,0x0,0xeac153b80482442,0x4000967aa06c2200,0x82902d4088c12240,0xffd4eb7bf9c9a758,0xc5cf9630d048a1c2,0x2b4e9fcad1a5b3fa,0xf0141ca8d100a2ca,0x91ed677879ecb2da,0x12832724602c23d2,0x50060040002140,0x629a1549c1410382,0x973fceebf1ee2342,0xf18e3262d0a533d8,0xf2962ff7f9cd83f2,0xceceb2aaf9c9a3ca,0xee1ffe4fd145b1ea,0x1fb8ffa9d9658758,0x9d5de7f171e4a2da,0xd7b67d7ff1e5b3f2,0xfbf7fbbbf9edb3fa,0x7ededebff1eca7fa,0xbfbfcffbd9edb7da,0xf4fdfbe0d1afb2d2,0xafbe7df3f1eca3ea,0xfbd4ffeff9efb3ea,0xfec3be17f1aea3fa,0xf3d494869c5a14a,0x65ed161bd1699372,0xe2fead73d9eda3ea,0xffbffefff9ccb7da,0xcffffe3ff9cda3fa,0x797ddfdaf1eda2e8,0xd4b2dee4d1c9877a,0xdfcfbf7bd9efa3da,0xdffefff7d9efb3da,0xfdcfff1ed14ca37a,0x6dfdbd7d1ccb3ea,0x3a32dfc1600d234a,0x6864f270d005a34a,0x68082c82d108a040,0x8a267f24d085a2d2,0x21730001008a042,0x200040000000210,0x1000006800400208,0x86ac2f45d1c19280,0xabfcfbe3c9efa778,0xa2d73ad0d1452372,0xbb9cdf7bf1afa1fa,0xf4577679d94da3da,0xcbde7f63d14db6da,0xf9f5e72ad964b3d8,0xeff7bff9f1efa2fa,0xefdfdf5ef9eda7fa,0x7cfe8fefd16cb7fa,0xb79c3d10e165a25a,0x8094f6ecc108a340,0x87e4beabd84da3d0,0x65650c1d0a422d0,0x805ea8710422d0,0x856406007004a040,0x880012228440000,0x8810188120050240,0x229ccb86d06d87c8,0x8404002040440400,0x1220660e0042040,0x601204a2c148a258,0x89c3c72498cc3242,0xa1001c0842002,0x81045c260011340,0x510c802050448148,0x4024077450082262,0x820000400200,0x400340000040148,0x2000200008040,0xd601d0a048440298,0x6420b440408501c0,0x1a0c2751100b1c8,0x60c1fa13400da360,0x6000000240000,0x484040050048000,0x8000004040000010,0x3288400e20a40042,0x180a10361a40038,0x88888c3502610c0,0x441000040008010,0x808305805004a080,0x5884262141013162,0x0,0x1500c00400081c0,0x4054102051048082,0xd300830851200250,0x10010000410000,0x1820732700482d0,0xc022608108482c0,0x2000108100100c0,0x1000800040002000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8001001200060,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1200000000400100,0x2000400000400200,0x10040000200200,0x10011000000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x804000000000210,0x80040000000140,0x8400000000000240,0x0,0x800021400240140,0x0,0x0,0x0,0x0,0x0,0x108014000400340,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100080040008000,0x0,0x0,0x0,0x0,0x8000200040000080,0x0,0x0,0x0,0x0,0x282100000040008,0x0,0x0,0x2000800000010040,0x14a454c8e8c52240,0xc129ff21d9cc23f2,0xc29a3850d1469352,0x95e6fe28d16da3da,0x4e9e5fdbf96592d2,0xb9115fe258cca3da,0x804205e0d1612240,0x720023c6516413d2,0xedf7ffe5f9edb2da,0x142a0402004a042,0x82000000000210,0x140c600008610240,0x92327f71e1acb3c2 },
	{ 0xf90c776318acfd74,0xdb4a860cd83aa560,0x480000890000800,0x3022000200160,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x48d63f4211b8ef60,0xffffffff9abcff76,0xdff5f5fe983eff24,0xff9e3d75dbbc3d24,0xe9dff73fdbb8fd34,0xcfcb4e7cd8bcef30,0x653a0d4310acb534,0x94bce02f10b0b530,0x4bc074409a300406,0x8010000008000002,0x0,0xfdccd0599b8ef36,0xf46696ecd8a4cdd0,0x51fe7d71d8b8bde2,0xdf56e7fc10baad74,0x2fc4ee0f5aa8cf20,0xbe77cf5cda2eed24,0xf49d56e0d8ac2524,0x9857e470d824ed22,0x80822e4492202414,0x1041020000084000,0x249444308a82102,0x8e127de890a8ad66,0x47cc350812269dc0,0xde975e6d982cf512,0xebd4f7bfd1aaaf36,0x7f58134bdaa4e562,0x9fb3cff39abae736,0x932dfe7cdbacf586,0xdfb60d759abeef56,0xfb9bffbbdab4ed36,0xfedfbfd7dabced72,0x8ffecffbdabcfd22,0xdf75f67f9bbcff76,0xfffe7d72dabcff76,0xb3d5cf26d8369f32,0xdde29281d2b2bd44,0x218e492282888920,0x9150b3ec90b2b614,0x54be3d705aa8afe6,0xdfb0ef7fdabcff36,0xccd7ff2bdbbc6566,0xe5ccdd2da88ed10,0x94b4b4e890a8b594,0xb9ee75f7dabc6f34,0xdebe7ff6dbbaef64,0xe7d6d7bcdab8ffb6,0x4f5a7b498a3c6706,0x44e050a89a248762,0x71a723e202b4a960,0xc8202448012c0140,0xe4d094709824f514,0x61810c09000ac80,0x30006100028e404,0x8c10068888304500,0x9d967ff9da387f34,0xe9d4f72edabaed26,0xcac61ed99aac6d30,0x4faceda3193c6524,0xe0d152e9d820a926,0x58824652802cec80,0xfa45ea66d824fd14,0xcf4277dedaa0ad16,0xb7bf4f689a28ff60,0xeddd1ee899a8fd54,0xb539fd719a8afda2,0xc8d0655818acf524,0x83d4be1f9a2ced32,0xa67406001aac6104,0x2800084000,0xdaa828438808e7a2,0x10042a2448206100,0xcd61a82010b0ad24,0x939cca123a0560,0x2016a1022a100,0x24000000400,0x60102100800089a0,0x50000000004100,0x2d208019200a580,0x122401008300540,0xc1c550002080520,0x46442180184740,0x8180450212084410,0x2e80810a102cd100,0x8904c20d2002400,0xc800c8004400,0x40010000400,0x4000200008100,0x0,0x0,0x0,0x2010000001080000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x28000000100,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100800000300000,0x0,0x8028400000008900,0x0,0x0,0x4000100010800000,0x3000240080000800,0x0,0x0,0x0,0x0,0x0,0x100000188000000,0x0,0x0,0x0,0x0,0x80000202000400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa010140088208140,0x400101280000400,0x0,0x0,0x12000000100,0x0,0x0,0x0,0x3000200102000500,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x24000000200400,0x0,0x802000080020,0x0,0x0,0x5004849c9080ac04,0xdd3cac5a98b8ed34,0xf29e77c6daac6f04,0xfff5af3adaa8ff70,0xefffdd5cdbaef776,0xb7f9ff669aaa7f32,0xf9fffe74d9bcff36,0x14b60b401280ac12,0xcfdfffbedab87756,0xfedf9edfdbaaad24,0x3288443888a86400,0x906b8f508a302506,0x4202000084402 },
	{ 0x49b445701dc6f141,0xc984010751406960,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8200000080002000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa18974e17027540,0xffba7ff3dd76fd73,0x96fcbf791d567562,0x1fd6557025486561,0x895233ba12a0e742,0x19c61a30159a4160,0x43184b480d527140,0x972642e915744161,0x502434c285c67201,0x9bc8fb6e05c67f70,0xcccfb79805b64f70,0x2b9ece0b8d947561,0xc4b4b6ac15b64771,0xe98fed62acde6760,0xdbb6677618d60740,0x4adbbb3f5dde6f62,0x635ed7c79fd86560,0xf4207de4df0e2560,0x102054006840500,0x58042d301080f041,0x180e6b290842461,0x2189e4000060570,0x15a2de401c464572,0xc148e6020d46bd20,0xf6b50f70254e7f70,0xebe5ff2b3dde5f73,0x2fdecfdb1ffe7f71,0x1dba57f37df63f70,0xd12c3f7b5bda7f63,0xdeae2f6f1dc6f763,0xcbf7ffaf5ffeff63,0xeff7fef68f8e7f73,0xa3aecfff0df67763,0xfc74b3f10786ff63,0x428e3f33e5d6fb60,0x2287270407407740,0x4810309401c06c43,0x2eac4f121c067163,0xc14021801820260,0xfebeff70dffe6771,0xbb0c6e7dc5fc6d71,0xcdc2f7afcbde7762,0xa202040201800940,0xd6c634ec0dd64d73,0x95ceff71d7c27f72,0xebb6bf5aafc4ed61,0x8deaff6a45e67571,0x6b3ad3512dc26733,0x94f65643d0863760,0x194c602104c44560,0x6cc0257a47c46641,0x8000000000004002,0x8901f8911862b60,0x8b025f1a070a5662,0x38347188946162,0xda006dcd05de6360,0xc9b5d7231fc66f62,0x6e9a0ecf8aee7761,0xc5fcdda29dd67761,0xd490d84553827671,0x2a0754081a66c61,0xc394e724dbc66561,0x244a389881d6e262,0xd39acf6129ae6461,0xfd7fb7fa2db66763,0xfff26fd30966f563,0xe494eb93450efd61,0xa52b1280c065571,0x20561689d18a6121,0x9030d8c00d406840,0x254e13420d404341,0xc82406721dcc7321,0xc1044c0208cc0060,0x200a189000024062,0x20d2b1692d4a7520,0x8000064244802240,0x2010241000164142,0x10b932009406611,0xd40388042006402,0x200010008400000,0x2065000060402020,0x400004101000a100,0x6180f311c8842f60,0x8426e0081c02410,0x90040200940100,0x4000208101003800,0x0,0x20400000140,0x0,0x80c200000000540,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1000080005000000,0x2000026010820020,0x84000000440,0x0,0x0,0x0,0x0,0x8200000000000440,0x10020001000200,0x40001000800040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2008002000400,0x0,0x0,0x0,0x0,0x14004000100,0x8000000002000400,0x0,0x0,0x0,0xc00000000060,0x0,0x0,0x0,0x8000100002000400,0x0,0x0,0x0,0x10001000840000,0x0,0x0,0x0,0x0,0x0,0xc00000020100,0x0,0x0,0x0,0x4800000201024040,0x0,0x0,0x0,0x0,0x0,0x81000000000820,0x0,0x0,0x40520080842000,0x0,0x102000000840000,0x2104800008402000,0x0,0x0,0xc8efa74099fe2360,0xcbac73f70dd67561,0xfff6ffbb9bfeff63,0xe31ddf9b9dd67762,0xbdb77ffa1ddeff61,0xd1aab25801a27262,0x1ae0253201487361,0xe5ffffbfed467b63,0xab123c818d042340,0x608575005426a42,0x8120a64a0dc65652,0x303c09400d464960 },
	{ 0x89c6f03313083e08,0x0,0x1441088010000248,0x84b83668d6200e40,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200040000000900,0xee322dd0c70d8c90,0x795ecf285f9c2cc8,0x84363c00c39a0c48,0x10116840562a0c40,0xd63c165206bc2d60,0xcea2ad62542e3dc0,0xdbf1ff73509d7d41,0xf0ea72b1fa903e79,0x9681d71072ee3c42,0x55ad84fe779d3df0,0x13822734d6b02620,0xb1ed724c6167caa,0xdec7bf3bfebf3dc9,0xaaffdddf7ebe3ee1,0x268446ef7fbc7d6b,0xa448277a56acbed0,0x2182f0144106080,0x8559a1046192880,0x1443602460420c0,0xe4200460061d6840,0x1180f2047e200d60,0xe2b63f7a729d3df8,0x65d3f732d3dd2e80,0x68cf1e5c5bb6ecc8,0xcfadf636fdabd02,0xd19cfd009a843f48,0xfffe7ffd5f1fbfb2,0xebf7f77ffefe6eeb,0x6fdd9ed37fbe6de8,0xfef9cdfb7fbe2cea,0xde36b5fc5bfe3d70,0xfbeefff3febe7fe3,0xd900e30c02ba2c60,0xcccb3091cf082d00,0xe63c0e447a002cc9,0xe9f5be7ad324bce9,0xdfbc2e76d79b2de0,0xf09e7f07cfb8ec62,0x75ef208c7d02c00,0x3e934d5b4e933c00,0x949776a1fe023f6a,0x696ea7625efa6cc8,0x8aa627c41a2c3ec8,0x4004062088010c00,0xcbfc5d40be128c48,0xb8c173e5da9f2d68,0xd70c75718eae2cc9,0x40228f7a479a0c20,0xec18000421014a0,0x47d6be3a4f9cbde0,0xa4ff4760471f2882,0x18d476fb1e0e2de8,0xa7367f7d4ebe1ce0,0xcb57d72357fe2cf3,0xf7ce37dd7fbe2fa3,0x1f9feffbd7ac3eeb,0xd16526b946ab2d00,0xefbe78687a8c3681,0xdbd4fb6e673c3dd0,0xceeb9e5dd6263d59,0x2fbecf5b7aae3ca9,0xffdfd7ead6bc2de1,0xdf3aa5355eb72f61,0xdc143f91c7dcbd82,0x8046ea00440214b2,0x1080e1020001d00,0x441910a092202c00,0x1042364220202c00,0x8200e20802012c00,0x8856061883053d80,0x810008924e008000,0x408082e0460c9d00,0xd328a2d04f26bc00,0x88023d44e7023c80,0x4100066002282d00,0xa900080270040002,0x2a000000041420,0x11883400120c2c40,0x802040002012800,0x102480041100400,0xc5421c0010a02e08,0x901c048102020c48,0x8021400044061800,0x2100000010800,0x8000004000800040,0x0,0x0,0x0,0x0,0x808000040000010,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x3804040040082,0x0,0x800010000002001,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400840000080,0x10020000010800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x500000104000,0x400040a000000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8100000020800,0x10040000c0000,0x0,0x0,0x0,0x0,0x0,0x10400080000800,0x8202110010c00,0x88000000000c0000,0x0,0x0,0x0,0x0,0x8010000000002800,0x0,0x0,0x0,0x0,0x0,0x1100800444000408,0x0,0x0,0x0,0x0,0x0,0x1800000002800,0x0,0x6002000400,0xd9fc2f61761ebcf0,0x933673e2579e3e40,0xcffbf7beffde3eeb,0xaa7ededffefffeca,0x85f75de7569d6eeb,0x1840254658d40400,0xa02240010e00460,0x5000c000020c0000,0x0,0x0,0xd3efef70de993c61,0x9bf65f7c709e1ffa },
	{ 0xcae43beefb8f2b2e,0x840060000042280,0x83a85ef28a1d670e,0xf25ff748bbcd2b26,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x72b62df98bcb6256,0xc3f46622b2104104,0x4dd55e6899ac277b,0xb110c6f39a1c2344,0x3126062492812926,0xf5a477a2d300211c,0x950812010002700,0xca5aba9392302b3e,0xa6b7c77b9a8e6b04,0xf1f77eedbb9c277f,0x1b8f5e182900322,0x1a22090002024,0x4eea3fb7dab66767,0xaffffedf9bde6ffe,0xb0b2dae8cac42396,0x93ac3f339ad80b12,0xc50000010000100,0x3a57beb3bbec2716,0x9dc7f538ac86734,0x308bd56182906726,0xed6ef73b9b8c2336,0xe4b6a7699b952f36,0xedd6dbae9bb4633f,0x6d9c7b499a80277c,0x16814fcf8af12713,0xc12d8460e3902302,0xffde7ff48b9c6f2f,0xcbfffb7f9bdd2fbf,0xfffffffddbd76ff7,0xbfbffffb9bcd27ff,0xd5fdfffb9bdf6f76,0x9b6e6d739ad9636e,0xb70db2e0ac82323,0x8852301a8b480b00,0xab9ccf989b9a2e32,0x10b596e99a94223d,0x8ff677f3bbfe670e,0xcbb67f3e9bcc6fb6,0x234682228b062301,0xbb297d189b886723,0xb495feeedbc82d37,0x806c266ada45230f,0x82802f4082a02306,0x80020e2090000234,0x44000a0000000,0x3cba5bb799186716,0xfb6eb658d3106be6,0x4884047083100304,0x8152e208ba1c2703,0x12b2819a182324,0x34b272e09b502b1f,0x844c5e8299006a53,0xde3025d49ad04b14,0xe888f3b2b3b02356,0xfef7dfdfdb9f633e,0xf57eeefb9a946bb5,0x510c24699390253e,0x18e5f5283812032,0xd2f4ca62a394031e,0xaad992829a9c232c,0x3f38cf429bc9637e,0xb5d776cb9b5c6b6e,0xdbfefd67fbfe22bf,0xc49035129b002525,0xc8c632969a8c2716,0x203f17c58adc2222,0x441500200a800210,0x202025729a900322,0x88916810a2042500,0x44cbce089ad90332,0x8a04860912802524,0x414400081040002,0xd100a064ca002282,0x8202300718002704,0x848242839ac00106,0x82100100802300,0x488040081002000,0x2000020000202000,0x840000c00102200,0x802061a288882210,0xe4421e1a81942323,0x20004000040200,0x104506108080020e,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x600000080002000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x24080040000,0x0,0x0,0x0,0x820000080000008,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x500009000000,0x14080000100,0x0,0x0,0x0,0x4000001008400000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8008080004000,0x0,0x4080000000002001,0x0,0x0,0x0,0x22000002020000,0x0,0x0,0x0,0x0,0x0,0x0,0x1004000000100200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x108000082000000,0x8000020000000280,0x1080000080200,0x210024090000030,0x0,0x8008200090040004,0xc2947ff81b840326,0xcfdffefd9b9e2fe7,0xefdfffdffa8d6fae,0xe5bb47f99ad42b1f,0x5108a9008a232212,0x10241000800030,0x0,0x2400000002000100,0x64000010002,0x9350a55093902b07,0xfffe6ff7fbd66f7f },
	{ 0xcbd7d7b7ebb7d7ec,0x1a537e030190c3b8,0x10240e304000110a,0x1240aed8d100c56e,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xde962d6cd9a05319,0xc0dc520c6820c31e,0x2e4ab2d981b842f7,0xc0904fb9e0ac4f33,0xd18c921cd081d132,0xbb62ff2c9a8517a,0x3a8a800080218200,0xce5bb4dbe325d52e,0xafbbcd7ff1b3f37b,0x70403f6089b44338,0x92c73d505194d33b,0x3004080000200314,0x4741a03280809a18,0x9b9ddfdbcbb2dbfc,0x58fcdee0e284f32f,0x1342f500e0a5c028,0xb3067a5e1b0db88,0xfdffffb7ebb1d76f,0x334efcd8c906d3fc,0xf4fa5ffbd1b7d1ec,0xc96cff7ae1b6f3fe,0xdcf62fffe194e3ef,0x658bf088e9a4d362,0x29413ed001b1c209,0x9eb06ff9e181d73e,0xf176ae7af194d7ee,0xdb96f9f2d1bddfbf,0xabfd5bbff9b0fbef,0xb2a6fbd8e1b4d37f,0xffbaeffbf9b1dbff,0xdf37a6ffdbb4f36c,0x2a2c3d60d192d303,0xd9204316019042a5,0x5297bb56e1807379,0x6fbdcfc0e9b6f368,0x7ef57beae3b4e33b,0xc1ee6f70e9b0d7f6,0xfa3c7f6af9b5fb6d,0xdccff63ff1b0d3ee,0xfe752dc37102d3aa,0x16eade40c919f122,0xb00002302000c120,0x200406000004288,0x0,0x295ca500101d20c,0x1cb7d77bf034d3aa,0x20249001200320,0xce30674041a48336,0x8000000100800001,0x115a120b4181c1a8,0x94335a60e331d302,0xf17df66ac9b3d33e,0x9f943f21e0a1d369,0xcbaffbb2e1bdd76f,0xdd6e7abfc990d33e,0x1bbecfff73bbf31a,0x844106616808c108,0x4504e56ae19048f2,0x359e012e188534d,0xecea9a50e1a0d9ae,0x273edffbd9b8f3fa,0x657cdefaf8b0dfbe,0xffeeff73fbb67ffd,0xebb2e778c9b0537c,0xc8c38e00e0200a68,0x6dff1d5881b26379,0x4408c0a402a04302,0xd14860008024c02a,0x2818090088204208,0x480060000000,0x8100a0061020000,0x400100180a00010,0x60e64060884001,0x30a1050280a00360,0x8af002048a204160,0x5002004000,0x200008c40018008,0x2025326189225328,0x81062c4060a0c202,0x3800800000208,0x0,0x8310050001100160,0x14117426e1001844,0x0,0x0,0x0,0x0,0x0,0x0,0x20020000018000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4008000000120,0x0,0x0,0x10040000004008,0x1004000040000000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400000840001000,0x0,0x0,0x0,0x0,0x0,0x800100000200,0x0,0x0,0x8000100001000020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x48000000120,0x801000008008,0x0,0x100000100800001,0x40004000000420,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20002122001c000,0x0,0x0,0x0,0x1000080000008008,0x0,0x0,0x0,0x2000002000200200,0x0,0x10000000108,0x0,0x0,0x0,0xa0250041008128,0x0,0x0,0x44050000000,0x0,0x40008004000,0x0,0x0,0x0,0x0,0x0,0x0,0x2100000018000,0x0,0x0,0x11e6b540e198d32a,0xa11e4eb0e0a0c38d,0xe1d3ef1acb84d7ee,0xab7a5fdaf1b0d32c,0x416dca00001c62e,0xf97c2623d114ff2e,0x549e357089b2431d,0x8180c30e801051ec,0x42581068904200,0x10884c8001048168,0x8050020001004020,0x108420000004422a },
	{ 0x40f430248084a8a0,0x481e5e3381c6aa62,0x80068000048202,0x10418cc00a40a042,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc6b86d1083d3883a,0xc9a1de2683d4f122,0xa750bfd803c6e3e2,0xc9a4d41838622a6,0xc18c78e183c4c19a,0xef3a2df88b4431ea,0xc0b0e20680c63142,0xe6fa325f0bc62192,0x122e4d450354602a,0x84d41ab60bc4c22a,0x225528046c140,0xc0000a140204802a,0xc8624e0003460102,0x5b9fcedf83d7f3fa,0x56f6eee18b57f3e2,0x99ef647a8b47a1c2,0xcbdaae378bd6e836,0xa784c21a8bc69262,0x28505d5c89466bda,0x34ca55ca82c1f01a,0xf77def5e0bc7a3c2,0x80000d0002028122,0xd5c27b1e88c6c160,0x1a177b548946e2f2,0xf7be6ff28bd7fbe2,0xfff5bffd8bd4fb7a,0xdfec7ffb83d7f3be,0xc980e82e015671a2,0xafc53e4a83d6b38a,0x13bcc77b8bd5f1b6,0x4000910000040012,0x5a80e4083c6c120,0xd8d463008253e342,0xc0c33ab383c7f072,0x4d7cceea80d6b37a,0x746276f18356e23a,0x97a635600347e268,0xfbbbff7f8bd7fb7a,0x7ed7b6a68bc6a3ee,0x6ff69bcb8bd6f372,0xb6f7edea8996fb7a,0x80010000008040,0x323060408944a308,0xc08206480206a1c2,0x23f09fd78bc4c1f2,0xb7bb7de68bc6eb6a,0x9008a01080041a00,0x9804000a00060230,0xa10689090046a222,0x6c90bb4b824623ca,0xbcfb6ffb8bd7e37a,0xdeffdffa8bd7b3be,0xffb47fef8bd6e3fe,0xabf7df3c8bd7bb3a,0x6ed53e3881c6eb6a,0xf6bf5ffc8bd6f3fe,0x8487106b8904e122,0xe29214200354c112,0x7ac3410a82c23162,0xe8cb9ed583c6e162,0x2b3c4eef83d6f36a,0xfdf7767e83d6eafa,0x9fbaaf708bc6eba2,0xde34bb7e8b44f3b2,0xc070a42283049392,0x294d98038a54a220,0xb40100208201a000,0x94c8a30082029322,0x1200000002008000,0x10000402000,0x80005c100102c220,0x48271000154c282,0x20004000008020,0xa02016000460048,0x810192048340208a,0x40001000020002,0x840263220355e200,0x1241000000448100,0x402405900844a0a0,0x200000008200,0x60e4081100c68362,0x218080080028244,0x404008000008200,0x0,0x0,0x0,0x10000480004000,0x204000400008,0x0,0x0,0x0,0x0,0x680004000060300,0x8001020800848000,0x0,0x6000000008440000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x81000001000080,0x110002002000,0x0,0x0,0x0,0x4000000c80002000,0x0,0x0,0x9150200001a080,0x0,0x0,0x0,0x0,0x0,0x0,0x400004000500000,0x0,0x0,0x0,0x0,0x8800000000000102,0x0,0x0,0x0,0x0,0x0,0x0,0x10020080004000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6020002080040204,0x0,0x0,0x0,0x2028400080005000,0x0,0x0,0x20401040000,0x8100001004000,0x0,0x0,0x280000020002,0x200203000000,0x0,0x1000010000020020,0x0,0x0,0x0,0x820000808000,0x0,0x2004000808000,0x0,0x0,0x0,0x800000800008020,0x901110010240a208,0xdb4c7c238ac7a1d2,0x8f4078450352f14a,0x8ccdf6a488c6b272,0xaa52044a0144224a,0x8000026080003002,0x40420800008032,0x202000040002,0x4040e1200104a220,0x20080000500000,0x800031080501000,0x10a4141000c08060,0x0 },
	{ 0x88c1200a49491a04,0x8c0080080800201,0x2860c72cfc88a00,0xb7c7966bccc1de30,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x52222c464d823804,0xebfefff7ed4f7e3d,0xcebe3ed94cc9de38,0x909d8ffa64ca263c,0x54e6b35bc7436e30,0x822ad024cc012a0,0x4b14c60ecfc2aa30,0x414614144580d820,0xe6bfcff3cfd7fe90,0xf4f336f885cbfe39,0x852637f28187bea0,0xff2877df44ceb63c,0x302f80285c22a00,0x1b3eef5fecdb3631,0xf426f6d0c4cabe94,0x884ae672c793fe94,0xfbbe2b7ec5c20e25,0x41d6d31fedc77e1d,0x8ffddfd3edcf7e3d,0x9f3e47ebecc67a10,0xf34bf772e7cbfe01,0x9a020a4006422a88,0x580c1200d805a80,0x221220034d402a14,0xb5f229ef4fc2fe88,0xe7cc973bccd2f69d,0x7b1ef472ccc32ab5,0xf284eb8ac692b004,0x267311da4dca2a30,0xb89a0d80cec02e90,0x9147bb7a87c37e11,0xfffebdf9e7cbfea4,0xbb102b2fc6cb1608,0x86ea728904ca7e11,0x42a4d15044929810,0x4a6046048c95000,0x8462d22c1c00e00,0xcfb77fff6dcffebd,0xd062602a58a3aa9,0x9b76cf5847cffe19,0x94251469a44b1a20,0xa97f2f7504dbba31,0x1614130404033290,0xcc8eb20781cbbe10,0xa910265806c21a08,0xddbfffeee5cbde95,0x508244888c01210,0x94024400420210,0xd821080408102a0,0x6edff6dfecdbfebc,0xb7ff6ffdcdcbfead,0xfffffffee6cbfea9,0x3386b461c79b7e39,0x8bf7e36fcfca7eb1,0x6fd2be3fcf4a3eb5,0xb2290ca041cb9e90,0x941606480e4a7220,0xaa122d5049c11e20,0xca16674e67c25e30,0xe253f9532fd35a98,0x3ffacffbcfcbbab0,0xc9f5d7fee7ce7eb4,0xf3f3af70cfc3fa95,0x57a77fe84dcbbe3c,0x80e00e1000404e80,0x8a5e8d43cec21e30,0x40420c005002800,0x184c651841c2aa14,0xc00a4ec04e401a8c,0x2000421000400a01,0x404c14da46490e14,0x8002000000400200,0x862014080d01800,0x8290000002420208,0xa084c10c04523200,0x0,0x80400000081000,0x1040042544402204,0x52282d8048c23808,0x8812180044805400,0x1020000400800,0x8090040000020010,0x84400c2088c00200,0x0,0x102001000022010,0x0,0x0,0x0,0x0,0x200104800000,0x0,0x60008204005008,0x0,0x0,0x0,0x0,0x42004808081000,0x0,0x0,0x280000006000000,0x10800020008,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2000020004800000,0x0,0x0,0x100800040000020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x810000410000,0x8000000200081000,0x0,0x0,0x0,0x0,0x0,0x0,0x24080800000,0x40400000081000,0x0,0x4000000404002000,0x0,0x0,0x0,0x0,0x4000000106000000,0x800080004800000,0x0,0x0,0x0,0x0,0x0,0x0,0x820000101000,0x0,0x0,0x0,0x0,0x8040000400200,0x4000200000800200,0x4000480800000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10100020000200,0x0,0x800010000020010,0x200200080020000,0x0,0x0,0x18c0212204107000,0x880411244280de20,0xc190d82066c00610,0xabf64f4a04cbde00,0x8002000000800200,0x840000000020010,0x4000000204020000,0x0,0x4000402000200,0x10002a40400820,0x4020821044020004,0x1014092008010200 },
	{ 0x888012ce0001849,0x0,0x804410008000900,0x5040929200521202,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe21005260270102b,0xebdcef0a4a311b93,0x0,0x0,0x40021000121000,0x9bda26500a58134f,0x89207b34ba5210c9,0x2003398000681041,0xacbd556a4852920f,0xf577d76b287813ce,0xbfbefd737b72bbeb,0xfaafff7abb5eb3c3,0xffd7e6187a56bbef,0x2d08c819424d824b,0x10002000001001,0x8832242208621282,0x200108000801,0xdfd7fffdfa7f9bef,0xebffdfdbf87a9bdb,0x8c865d05787a1363,0xdf3dfe7b687f3afb,0xf6be3decba7b1bcb,0x2646220408702082,0x1b4efacea373120e,0x3dfb6ffbbb7a9bf7,0xd52ca756aa669a8b,0x5a8e447e5876130f,0xa926d3686a5a12df,0x6016160008400980,0x20b86cc0a86e0242,0xb52fbb566a7219db,0xaa8eed68e8723b8b,0xfb77ef29295219db,0xe6dab6162b668189,0xe554fa3aa5533cb,0x70a0a67c2a561089,0x104196020660000,0xeba6ff37fa633bcf,0x8fdffe76fa7eb9ef,0xe9375edb68721b8b,0xa694542348600917,0x412081e000021000,0x8b867622295a9acb,0xadd7880a6862901b,0x6bfedfd2e8723bcf,0x920085410a001107,0x18f0a65008501a8e,0x82c0f00214a904a,0x81f653180a5810ce,0xa7fefa710a57bb4b,0xfff94ffbfa729bbf,0xf6feafeb2a7a9acb,0xe3fe7ff4ea7eb9cb,0xfbd7e76e6b7bb3db,0x2e4a30cfaa6313cf,0x9ae4bf7187a12d3,0x500a324208101906,0x41180e9008701217,0xb8c4f6288a50a21b,0x8b720e23ea0a184b,0x8dfecef7aa5e11d9,0xfdfff7fe2b76badf,0x9b24ffd22a7f93cf,0xee2e5d4ea262934b,0x43c2360bba6010cb,0x28a4c900320812ca,0x9231042a28608181,0x2000042210401040,0x12120a000870000a,0xa081d6020a4a12c1,0xa10000a221001,0x420504088021003,0xd125160898601b03,0x490004000423000,0x8448222280001c6,0x21478a402248024e,0x2004008000100,0x10f0310ba06011c1,0x2261500004000ea,0x88e22008501849,0x4480021801400148,0x88c0080b410301,0x0,0x4002100000440202,0x0,0x2002000040000001,0x1000010008400000,0x0,0x0,0x0,0x0,0x0,0x0,0x8000200008000100,0x0,0x0,0x0,0x80d00040000083,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40000008001,0x0,0x42000008001,0x0,0x0,0x0,0x0,0x0,0x0,0x88000020008000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x800008000021000,0x0,0x0,0x0,0x0,0x2280008000848,0x800080001400000,0x0,0x8400000000008001,0x0,0x0,0x0,0x0,0x0,0x8080210008500001,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc000420000010282,0x80000100021000,0x0,0x0,0x3024d420044002e,0x83902b0120429041,0x8580a60228129081,0x22a4840820029081,0x22482110501206,0x1140e4174040184b,0x42882dace20800c9,0xc080702140529001,0x61215c220721082,0x0,0x825220042004e,0x4652484800410202 },
	{ 0x8004204002c0000,0x622080003140804,0x8080042001000840,0xa4841020002c00d0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xcaac45f0e13e79c4,0xab82f73902349776,0x64a320ae22c0874,0x8f764f7a81373dd4,0xa054167302353170,0x5b887447401cb5d6,0xc330331f64761454,0x10224000000,0x0,0x0,0x0,0x0,0x9feffeab677f5dd4,0xdd6da9f623d3f64,0x319a5329227c9e92,0xf92a8f7a261f16f0,0xcca66774c23e9ff6,0xe3f4e626e23c1ed4,0x833137ae71c1ab4,0xb792fefbe66e9df2,0x9acc5648c14f1dc0,0xfeb41560223f9f06,0x880031a220380bf0,0x6d53de5b663c5dd4,0x1b985ef0237e3dd4,0x316d5428c03c9c96,0xdceb671661eb5d2,0x6900d730623648c0,0xc8439c0765211440,0x6e2aedc3003cb656,0xbdf55efae67f3fd4,0x100e306240141484,0xe38afd2e65341bc4,0x5fc3da11677417c0,0x6a5aca53e43c1ed4,0x4401c8427240a40,0xc902556b643d10d0,0x20900a0021000240,0xd4ece31b603e82f2,0x2afceccae5181e70,0x810426a000d1484,0xd94476e2006c95b2,0x7cae3fe4e55d5cd0,0x8936d72240346f64,0xb95e5ecc627d96d6,0xb4014062251548c4,0x19fcbf6a665f9e32,0xe6907ff5653d9df6,0xc9f2f97b6337ddd6,0xa5f75edc637d3dc0,0x14c10522340340,0xd4943e6541765940,0x12a64562410c0c34,0x8000093000060080,0xd1d337db623f9dd6,0x6faecffbe33e7df4,0xd477feef635e5bf4,0x2bbc7c72e23f39d4,0xfe84776c83359cf2,0x44807201601821c4,0x32e589222240940,0xa4725160241015d4,0x9188150200240e44,0x80220e00000c1340,0x0,0x44165e00231c09e0,0x426000260040,0x160a640213404c0,0x81002d0100061500,0x2000520080001c00,0x208864640141084,0x80e0c140340800,0xc1401402021c00c4,0x42222520001c02d0,0xa0001000800,0x2940908100380060,0x4010000000080080,0x4100201000041404,0x0,0x0,0x0,0x0,0x1000200000200040,0x0,0x11000060000,0x0,0x0,0x0,0x0,0x0,0x0,0x200000800040040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100800102000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000200000080060,0x0,0x0,0x0,0x8010024000140104,0x49260200c0350860,0x0,0x0,0x6000000000050,0x80040000101000,0x201000010040,0xc60000000,0x0,0x0,0x0,0x0,0x400004000040400,0x0,0x24000000080040,0x0,0x0,0x0,0x0,0x2000000c0000,0x0,0x48000100200,0x0,0x0,0x0,0x8060002000c1010,0x0,0x1100040100081180,0x991e5860660515d0,0x514cc332663c08c0,0xac1c06ca413c18d0,0x3c975eac025e0fc4,0xd2dc45a0601927e0,0x82022567e6175850,0x950902162101050,0x211218c800309e52,0x2000200080014,0x108a0524215311d0,0x5a63d50631c1150 },
	{ 0x6000000004004000,0x40030c0006004310,0x809ea20028008150,0xe075e151e2008513,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x211827220808969b,0xe154c1061700cb13,0x8dc1b6030c20e2d1,0x74ecf720f21d118,0x7551b26c1f20de18,0x973ab5450f017a10,0x1004004808204010,0x0,0x0,0x0,0x0,0x0,0x0,0xef4f9cdf5d08df53,0xde955fe35f29cf93,0xe92eef73df89cf5b,0xda3a263f1ba8cdc9,0xc8c6d31b4f01475a,0xc94e1759882095da,0xd49c47aafd21c7c9,0xd265bc5b3b89c311,0x7f5e7d7083089753,0x83c7efe20b016d4a,0x2e425b684a8173d8,0xcecec415b00ff5b,0x5536b37b5b88e39b,0x598cbd221f89e599,0x0,0x40080000208000,0x845842210a214309,0x10a5f6eddb01d751,0x9300bc002200054a,0xce202f061f00d81a,0xc82fa885a208309,0xa16150a22800512,0x16787cc4280885c8,0xbffefff6ff89dfdb,0xfff67f7fbfa9ff5b,0xddffff3b4fa9efda,0x4b8ef81913084ddb,0x351f5bad8f01ff19,0xf9dce73a3f00ed1b,0x9a80a32c12088b0b,0x65d2c468580885d3,0x9573e56db08d5d3,0xb6b2c7e84b21df5b,0xddfdb47aeba9d7db,0xffbefffeffa9ff9b,0xe9e2fb327a01cf91,0xdc26a101b21ebd1,0x479dce402f20c75a,0x9154b03e1e08db11,0x80b0444001a0cb11,0xb924820e1a00971a,0x46d33c600a0095d9,0xbf7ecf7b3b21c793,0xb495f6ffdf21dfdb,0x8abe7df11f21f7d1,0xca94b43f3b21eb99,0xcdd8f0a35b29e3d8,0xa842d95007084913,0xf43526e45f28e112,0x4847731e084a43,0x49960f0a1b000b81,0x8000100001000002,0x220000c002008403,0x1423220113000211,0x790c221a8400d518,0x8010257002288110,0x101c00001000450,0x27a155a0400cfc3,0xe28052000008689,0x1a1c001012208b01,0x2004000000100,0x8844026096004700,0x60000001004000,0x8000040800000408,0x1000801000800540,0x44020000100,0x41a000009010,0x0,0x0,0x0,0x0,0x0,0x40001000000300,0x0,0x0,0x0,0x0,0x800000404004000,0x40208000100,0x0,0x0,0x0,0x0,0x0,0x0,0x840000008000100,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40004008000200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x101002000100,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000508040080100,0x84820000c004100,0x0,0x0,0x0,0x0,0x0,0x820000014000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2100000000402,0x560c22001200011a,0xcd42beb01ba0af1b,0xeffedddbcb88cf1b,0x8cba59401b01db5b,0xd301a5608929e553,0xfb1eb7e4be01cf9b,0x100800000001010,0xe88a31451a208290,0x1140000a000140,0x3c5556220e01e318,0x480865420a20cadb },
	{ 0x81928204220a40,0x0,0xc10066310826440,0x901028d084184420,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x804182000c8c8000,0x2e98744005b84d62,0x4a90600d0c284700,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200000020800,0x2004c5254880562,0x5945112090184442,0x9963285414ba0722,0x4018272014880542,0xcfdbff30d5fe5d6a,0xfd57fef73daa5f62,0xb5966ded04aac9e4,0x59d40ce604b86d6a,0x5eb66df3f5bed562,0xc92aa3781d2a6f42,0x7836178015ba2de2,0xbabafdda5cb8fd6a,0xd4f5167a15badc66,0x0,0x0,0x0,0x3422140230280142,0xf415dd7a949b6d42,0xc080040040200060,0xe900601014809002,0x80f2bc111daa7966,0x12969f1254ba0466,0x0,0xedf675f33dfbfdee,0xff9f3ffffdfaffee,0xcfcafa9bddba4f6a,0xbd48b1a15b0cc62,0x14077961009acc4e,0xc07cae5810836142,0x5a1aa74cf0c94c6a,0xc252cbbe34987562,0x4e7e768c85ba0de2,0x2db3effd141b6d62,0xdffffff955bb7f6e,0xbbbcfd70f5bb6ce2,0xc167faea148a8c66,0xb7c8ff433493cd64,0x20ccdc14c124d24,0x3545f76d349b8d62,0x4068757015a86c42,0x8800610210d84402,0x7cb9a40649a40e2,0x7f768fd375ff4d6e,0xfdf5fffbd5ff7dee,0xd1bf79f17dfaf4e2,0xfa44b7fc74cae9e2,0xddab12bbc8a4d60,0x8364d08349e5442,0x91b3666e1c8aec4a,0x8246200020826,0x20045000080c800,0x4d02ce0c05b2584a,0x2100000808000,0x300055c404c0c9a0,0x91702340141a0442,0x60900004000a000a,0x180400100884800,0x86e0050008840e2,0x78066c210b14642,0x1060b01000086462,0x610a92010194c42,0xaa54403630016840,0x18800e9035326440,0x4a94d2210b84e22,0x40064020880040,0x0,0x0,0x0,0x8a08210008a4800,0xfe830a8e040,0x0,0x0,0x4000100000020800,0x101001200000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x48001200000,0x8010000080000400,0x0,0x0,0x0,0x0,0x0,0x84000010000400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1000040000100400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x108000020004,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x62800284800,0x40020000c000000,0x0,0x0,0x0,0x0,0x0,0x0,0x80108004000,0x0,0x6008005201000,0x200280000400,0x0,0x0,0x0,0x0,0x200400000900000,0x0,0x0,0x0,0x0,0x10400004040,0x0,0x0,0x808000000000a0,0x113a9750901264e0,0xf40625d2948a4c62,0xc40233a14b84402,0x4501404181220ca0,0x480230880888400,0x7804060050894060,0xdca02743449a5ca2,0xc090d68804aa4540,0xac5008000090406a,0x620078004b80400,0x90a40140608a0062,0x8008480010000402 },
	{ 0x80800000000028,0x2101142c4103060,0x4120042220805020,0x7015d27d020262e0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa080608002202c4,0xdcbb5fe9f65276ec,0x7967767b6091d668,0xff264f7b9490f6ec,0x88c6eb6e247374ae,0x825f5ab40013f68c,0xb45056f1a03272a8,0x35c4d03ac222d6ec,0x48000402000204c0,0xe845b36a648174e8,0x76c7df4b0431f268,0xf7b10572803336a4,0x2000002000000084,0x0,0x0,0x1100000000040,0x8a3c6f60c4b23aec,0xfc327a711433f6ec,0xc6a21f40207176a8,0xc7875040567252e8,0x208000000880,0x4080000104000,0x0,0xfffffff3b6f3ffee,0xffb26f54e212f6ee,0xe5ce6e0a66113ce8,0x2d179007808054e0,0x109665d68490502c,0x806000124200,0xfc72645ed473526c,0x8001f80000123484,0x8000114000000,0xa4b047e3c2a353ec,0xb7ffbf70a471f7ec,0xc6963d31341346ec,0xe0f7e33b26f3d5ee,0xfffe1e7bb632f7ec,0x855ec06680d15468,0x8c549ee86492d4a6,0x70eefd7324b3dee4,0xc365fb3e7453f6ee,0xb967968cb4b0d464,0xeb8e4fabd652d2ec,0xfef776fff673f6ee,0xdfbeff73b6f3f6ec,0xfb96ffa3e6f3d6ee,0x220623022050d020,0x8a1207486011d6a0,0xbfbbffece6f3f5ec,0x22100000004e0,0x3a9402242490d062,0x8c0000204105000,0x0,0x4323028a4102060,0x822210400148c,0x2a802441000114e8,0x210020000001540,0x8020094200000068,0x8d40492214105060,0x100cb40300314028,0x560a5920040014ac,0x8864c22014131640,0x484706011410c084,0x1aacc021610d06c,0x81408230240014ac,0x10001040000020,0x110000002000040,0x2380004b01020,0x14ed5768201246ec,0x34aa6fee743272ec,0x4002020000020,0x0,0x800220082100040,0x0,0x8400002420022020,0x0,0x100400000000240,0x0,0x51500000823028,0x24010100c0100020,0x20080000020200,0x280000000000028,0x0,0x0,0x48004200000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x804000604801004,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10001014000000,0x0,0x0,0x0,0x0,0x8000040080000020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6004000004,0x0,0x0,0x0,0x0,0x8200040080001420,0xc000000000040,0x0,0x600000000000028,0x0,0x0,0x0,0x0,0x0,0x0,0x10000200004000,0x0,0x0,0x60000080100000,0x20004480200c240,0x0,0x140280440801460,0xa2843f1e867050e4,0xa00068220082188c,0x90800200004440,0x0,0x42a7403611d468,0x688235420092fc88,0xc1320200a0105028,0xa400000800104280,0x402208004022004,0x100840024002020,0xa22a2d68947052ec },
	{ 0x8b82f2bb51bcb7da,0x60f110c0401c1088,0x430e4332d114b15a,0x93a0a210511431c2,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8854526a901c1c8b,0xf0be1768903c364a,0xc3a4eed6d02c36c6,0x2904900080001002,0x695e76f0599c3fd0,0x1680a7fb519c94d2,0x9837dee950169dcc,0x90204e0cd1041408,0x8834cf3c9804948b,0x0,0x0,0x0,0x0,0x0,0x2842a0004010b050,0x292ec548401e3a5a,0x657d4268513226e2,0x755180141812,0x4800100200,0x0,0x0,0x547476e0b19c969e,0x262c4080141882,0xd6806f62d0ac34d0,0x8566f22bf01496c2,0x8f61f4158849148,0x41e1a09581c1018,0xd18cb6d8710cb148,0xa102f2a1034b7c0,0x2944032b4104254b,0xaa4e08d05110b64b,0x91b20ff2519c3692,0x367cbe505994bafa,0x6f963f74712ca448,0xc9bff72bd93eb7de,0x6edd5f7d593ebeea,0x27becda9d91c33da,0xb465f7d471bc9fca,0xcb7e7f6271aeb3da,0xcb5ee20a711c2942,0xc8d3f393d114bfdb,0xbfbfdfdfd19ebfea,0xb4fddee8511eb6ce,0x37befdf3501cbdcf,0xda8eeb9fd09cb6d2,0x131b12d114160a,0x671c1bdfd81c109a,0xf5bfffead91c9fdf,0x865004a51101c90,0xd200230c510092c0,0x404a600040149040,0x2202948040048048,0x901504a00104016a,0xbea545418ea078,0x83946571699630c8,0x8002c06111101682,0x28075908680412ca,0xeb2a0fd3513ca7d2,0x51007603511c1cc6,0x822848411014148a,0x8152083050101406,0xfc01a0841143242,0x53a8c463511cb4d8,0x10601440401210d0,0x0,0x0,0x0,0x40800000401,0x0,0x0,0x0,0x0,0x0,0x0,0x8482420102004c0,0x0,0x1000c30000342240,0x0,0x1000080000140000,0x0,0x8100000010100000,0x0,0x840000000000140,0x0,0x8110040041000012,0x2000040000020040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8040000001002,0x0,0x0,0x0,0x0,0x0,0x1010000210101800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1001000000000240,0x0,0x0,0x0,0x0,0x200040000000140,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x82050008001480,0x44040210241048,0x0,0x2020000008040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x50000000000480,0x80100020040,0x400200010000080,0x44e460c01c3350,0x2122a005104804a,0x942e22a511c175f,0x805000104000a,0x8a2434051020440,0xf1a4a260111485c2,0x12000dc040148058,0x4002000840000,0x0,0x1242a0101030c0,0x0,0x9d0c3d687984b1da },
	{ 0xe9d2673698d02d66,0x6000015090808800,0x208000008400000,0x1600b24210e12260,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8440260280690700,0x10001080402100,0xcccbb0809820ad14,0x24ed8c088680720,0xb5ba1d8a9061f7f0,0xa043270088809d00,0x288445e0d0803850,0x0,0x0,0x0,0x0,0x0,0x0,0x8039400008a100,0x2a809088840ac00,0xa17c1ef09070be50,0x8008600120012242,0x0,0x0,0x8201449300800100,0x36b347e8b8689f38,0x90825500d00a2e00,0x5bbc6df4d861bb76,0x8f8ae72a88d0a744,0x875edb5a90f0bd14,0xc321100d0202c00,0x110a03a098500260,0x5883822400602520,0x8000d000c0009100,0x669e7fddd0e9bff4,0xc806fe088492752,0x91da376bd8e2a302,0xf2de7d7d98f9a7f4,0xfbf7febe98c9bf7e,0x6dfe7ed9f8e1b77a,0xdddb7fead8eab784,0xd57dff7bf0e9af56,0x9b167dd0b869be50,0xb3c4ef4eb0f0ef72,0x8a72fe5298c2bff0,0xeebfed5b986baf72,0xf4fb367ed0f2e736,0x812a0d208008a690,0xfabc777ed06befd4,0x856f28ac8602c40,0xba415d02d0412650,0x50a00ec0d022ac10,0x800c000808412042,0xb100012080010120,0x8402820890a02500,0x10100010000010,0x1000080000010040,0x4000720890004120,0x4204043398c00320,0x100800080000000,0x844161190602330,0x14824b81a0410770,0x10e01c0098804d00,0x3906504800408160,0x400280000000,0x8ee312539068af20,0x8a00cff99843a1e4,0x2000002008400000,0x0,0x0,0x190000000,0x480000400100,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1000004080000000,0x80800080100000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80040000208000,0x0,0x0,0x4010620202200,0x0,0x0,0x0,0x0,0x20090000000,0x0,0x1800040890000000,0x404000000008800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x401008400000,0x20004000080200,0x100040010000020,0x2004000002400,0x60000010040,0x0,0x0,0x4000800000000120,0x100020080000800,0x0,0x21090000000,0x10002008000400,0x0,0x0,0x0,0x8002000010000010,0x4400008400000,0x0,0x0,0x0,0x0,0x0,0x4000008000400200,0x100100010001000,0x26101e70986085e0,0x1818a0a08400520,0x4110580000418480,0xc92eaa3b880ae10,0xe30c06a290e0b618,0x801401c000489200,0x480000000000100,0x6511c5090e82a04,0x100010000400200,0x8061020080002442,0x4a006102c0c12000 },
	{ 0x6084c310003ca420,0x0,0x128002001024001,0x9224160808084f62,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xdd6200a2b0c4561,0x29feeb71bb7e6f66,0x1fa04f50cb3c6720,0xb10cb04828f84120,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x3820021080020,0x483815002c84102,0x0,0x0,0x0,0x0,0x1f7fdf5adb3c7762,0x8c95c4e000347660,0x80a25742050a4463,0xd38405a28b042750,0x44a0c0002080120,0x808000000a000,0xb8800de12148ec70,0xc070060ae0006421,0x3a92332009007c61,0x8480c00125405020,0x26111eca2a486520,0x346a7280032c4442,0x73aeb5d2956cff43,0xfef66f649b3ced62,0xc9eef3afdffcf507,0xfef73fd973ce7f61,0xaf3ace8bcb7e6722,0xf4e58279eb7e7721,0x6afea5facbaa6763,0x9bc5cb264a7cfd21,0x37e114605c86e20,0xe6dedeabfb6ce763,0x707373fa2d6c5e61,0x10a8ed0048882570,0x4a08e10212084430,0x68d4110120c8e021,0x200e419a18606f61,0x14115ca29b2c6561,0x108045600524f000,0xcc44240189362000,0x8020000000240000,0x600084000,0x1081042040406001,0x206046022884420,0x4606106402883960,0x0,0x45301a484a402560,0x1600a09109243400,0x9816000082120,0xda8822240a48e544,0x8900420242882400,0x9f11c060a6c6470,0x1108000208084000,0x500000004001,0x0,0x408010000400,0x0,0x2020000002000400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80800000084000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1400000000000404,0x0,0x0,0x510201482000,0x1080000402000,0x0,0x0,0x0,0x0,0x0,0x0,0x500020000020,0x0,0x0,0x0,0x0,0x0,0x0,0x10001000000400,0x4000000001020000,0x0,0x0,0x0,0x0,0x100020010000400,0x0,0x20400000000060,0x0,0x500000404000,0x880101008045000,0x6000000005000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100002008400000,0x0,0x0,0x0,0x4844202000880c20,0x4002066808080422,0x80000202000100,0x2001508a02040d60,0x489026800240740,0xd100a4900a8c4c33,0x3840000848042420,0x0,0x4000160000200860,0x0,0x50e012100f000403,0x2001c009004200 },
	{ 0x8bbffb6fbeffbd60,0x4c1000002022800,0x420c402000c40408,0xf13efb6a96bf7d60,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4220100104408,0xc5ca37bc5ed62d68,0xff7fdfdfdffeef60,0xb3b85d29079d3d40,0x44000861000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x7c33e0902981440,0x280691081c900540,0xb40ce6641ab23c00,0x1b2c1778539e7d40,0x5186204c17902940,0xe5d7f8a814be5560,0x2a744c1af2b84440,0xb000016044822100,0x602c240046102440,0x82912e65c2ba2d60,0xcf9c13254a1a3d28,0xa6d0bec7069a7d00,0x2eaa6f52fbf22e40,0xf1df57edc7f77f68,0x4aba2502c68e2c60,0xc1047a8514962528,0xefd79f575ebf7d60,0x2498dd5147902f00,0xf575c6624e9a7548,0xdfac2d635fb2b768,0xfbdeef6f6fb67760,0xa7fef9115b330d48,0x6efcdfb00fba6d40,0xd5bfdeefdfde5f60,0xe8183d415d922d20,0x889020000220440,0x2830203101400,0xa142090183302940,0x80909468063a1428,0x100a20400a104400,0x8202010c04125140,0x846aaa0016942140,0x10130020001400,0xa2800005000540,0x100661144120500,0x24024040840400,0x4000209003100400,0xbc9180005200d40,0x289045f902962d20,0x8c01000060800,0x2aa2d400e128940,0x88c00002002900,0x767d510bd61f00,0x1614c548828a2d40,0x1a40020840842420,0x4000008020000400,0x8000100000001400,0x40001000001400,0x0,0x100006200000440,0x0,0x2000042000940000,0x0,0x0,0x0,0x0,0x0,0x0,0x800004000a000000,0x0,0x0,0x0,0x0,0x0,0x0,0x1002000060000,0xa00004100000,0x0,0x0,0x130040204100940,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80000040a000000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x21010800000,0x0,0x0,0x0,0x0,0x8000000000020040,0x0,0x0,0x0,0x0,0x1000040000000440,0x0,0x1000010004100000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10800020020,0x0,0x0,0x0,0xa00000840000,0x0,0x0,0x0,0x2000002000021000,0x0,0x0,0x0,0x0,0x1000200010000400,0x0,0x0,0x2c20010840400,0x4040000021000,0x4080a000000,0x4824780000220440,0x9810aa780ef80400,0x80a0010a0a826400,0x281096c810840540,0xc8b8456a05964760,0x9300de0016963548,0x461024900e220c60,0x884493048a002900,0x0,0x92200002001400,0x500002000400,0x4e60040002082540 },
	{ 0x6050010882046080,0x800008080002000,0x120000000040080,0x2acbe71d224267d,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xca10211010843041,0xcfc3fff1f23e6bdf,0x4fdffec4f71c7e9c,0x401080004000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6000000080002000,0x88e7f2a2d6b20bdc,0xa00ac042d02400e0,0xfd7c7e6ef7346b5c,0x99ac7441d0862240,0x9a40c30692246304,0xc9d7eb98f60e7f44,0x4006100002002804,0x142303c0c2026804,0x990e7402c6122a54,0xa302e7292002746,0xa9d6fb02d6046cff,0xefffffdfd7b6675f,0xb4fa47e2f28625e4,0xd0d07c14d6160160,0x292b44044042c82,0xc920eb1ad2846147,0xa571e50c2802858,0x25ebc8cbd60c6eb4,0xf51dbabfd79c3ae6,0xffbe3ffbd28e6bff,0xf3d6af3ff6be6fed,0xac16bfb9d2363add,0xbfaeff69d72e2fe8,0x9532ff68d60c635f,0x291091002501,0xc82cf7ded60c25fa,0xedeba00d6142d97,0x8a4e2056d514186c,0x96f20660540620b1,0x22c042092040014,0x4202600630a00040,0x2200000002800,0x2400000000002001,0x80520000802040,0x1061328c4022501,0xc1902535d6200344,0x4650002050002880,0x29ed8810c2063205,0xb0100ce0d6042bc4,0x22d951d20e2003,0xc002664c82006204,0x900012a9004204c,0x8080080200002400,0x20040000041000,0xc804010002020,0x0,0x0,0x800020000002400,0x0,0x880000c0000,0x0,0x80000280002000,0x0,0x0,0x0,0x0,0x0,0x44000010000040,0x4000100000002001,0x0,0x0,0x0,0x8000100080000008,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x11c82102000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x722000402cc,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8100000002000,0x0,0x0,0x0,0x0,0x0,0x8000180000100,0x0,0x41000041000,0x300000c0000,0x800840000010,0x0,0x0,0x40830000000,0x0,0x0,0x2200044000000,0x0,0x0,0x4002000800040,0x0,0x4000040000000204,0x0,0x60000040020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10000404002c0040,0x8400662c4002244,0x2090000200002400,0x844321282002115,0xcabddf93562027cc,0x1c0c018354043860,0x10d0a904c2222148,0x5480000040022001,0x0,0x0,0x0,0x80150010202004,0x54020946d2002d80 },
	{ 0x4000c11201921201,0xa1d2900b81200b24,0x40802800000,0xa401022801204b02,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x45006102881041e0,0x68002620091041a0,0xcfcffbffabf553b7,0x211a1050890a4302,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xeb9c773fabba73fc,0xa15c760309d04722,0x90cc9d2a8614266,0x9694756b01b44227,0x8340967981bc6ea2,0x83be6fc381b673ab,0x8fd7b38289f44b6a,0xca4e395381161324,0x42054e003804062,0x886c5f61a9b20339,0x5eb72df289b45f63,0xcf5ff7be8b9d53ef,0xbfdeffdfabdc7fbe,0x1743437289b57a12,0x9069a7768a841233,0x53b27c548af83398,0xc900600180080624,0xeed797008be07f2c,0x165d479e8bd4472e,0xb6f526500b9045a2,0x4feead6089f44327,0xcbb5ff26a9fc4a67,0xfc43bf4dabda63b0,0x9feecddb8bf8426b,0x35246ef8891c1b22,0x428825e689300a68,0xdb8d4fb9abfa2326,0xc9eb7e138bbc57ea,0x67164bc88bb8732a,0x34a24ce40ad062b0,0x40a0404088080304,0x0,0x5d4608a81a0133c,0x1004080001000,0x90040880200100,0x5064a46009025320,0xa010004880010122,0x250020001900001,0x18101000001320,0x16324d44a9d34432,0x601c2888c00302,0x1a32045080811126,0x8021810008c00101,0x0,0x401000008004000,0x1434042480200a08,0x0,0x0,0x0,0x0,0x88008400000,0x2000012080200004,0x0,0x8000100000000100,0x4000402004000,0x200200c008800201,0x10442090a004000,0x9010002080000004,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6001000002,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x800002001000002,0x600000000800100,0x8000040000100200,0x0,0x0,0x100800000100200,0x6000008080000,0x30008080000,0x0,0x0,0x400209000000,0x0,0x202090008006000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000001001020000,0x0,0x0,0x1022080002004208,0x100000802800000,0x400000c880800810,0x8000100000000100,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000100000800001,0x0,0x0,0x0,0x2042028180000124,0x0,0x41001100000,0x4b822420a10046a0,0x68564010088803ab,0x5c01c68810426c,0x2726a09844323,0xd4acf57088585a86,0x41000040200,0x4020400040120,0x0,0x0,0x404000000040200,0x10412c8090440a2 },
	{ 0x2004020000008028,0xeb4e94caf104e91a,0x810ea407000c008,0x424200003000102,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x816a2e768984e102,0x2a60c6089442288,0xc9d4dfa34bc5d13a,0x80080000008008,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xdeb6fb7e7bcdc3da,0x64000a40809012,0x392ce1080c58820,0xc6f2f6ecc1c5c33a,0xc18e622b514cc01a,0x8b200107c0806102,0xfe5fca62f1c4c01a,0x6f555fc829ccc3a2,0x10400000000102,0x162a91080c4e01a,0xfd36af75c3d5eb7a,0x8fe9f777f345f37a,0x2e7ebadfd985f33a,0xb6674fffebc5eb3a,0xf87d066af350f10a,0xe6a0356d0b50c10a,0xc85aea6ae3c45b0a,0xeefe5f56f2c4f13a,0x4e9e2fe3cbc4e8a2,0xdc55bf7c4adcd16a,0xc04e4c34b0cc110,0xc8c5c770e144e13a,0x7433ac343c4431a,0x192cb97e1c4f02a,0x94d13561c2d5ca22,0x10ae2d704044c30a,0xbb953734e354e202,0x48c230000154c300,0xa11b4650c304e10a,0xc43272e80104839a,0x102020083002000,0x10042a042104c010,0x824430000340404a,0x44060000000,0x20802c0007000,0x11ac00607804c000,0x188004606100c100,0x450020078100e022,0x8206100683046012,0xb40044120100c102,0x10016014000c022,0xe18025424304c032,0x8000024401004002,0x60000080002000,0x410038480408962,0x100800000018000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1200480061000200,0x0,0x1400000c000,0x0,0x81c00001000092,0x24404028004002,0x1201004000,0x0,0x0,0x0,0x0,0x0,0x0,0x100100000000102,0x0,0x0,0x0,0x0,0x0,0x104000001000080,0x0,0x0,0x0,0x0,0x0,0x20180004000,0x0,0x2004400000c40200,0x0,0x0,0x0,0x10040000004002,0x0,0x0,0x4000200050000000,0x0,0x0,0x0,0x40100000040200,0x0,0x0,0x0,0x0,0x1004000000000108,0x2004002000002,0x0,0x0,0x0,0x204000808000,0x8080040004000,0x0,0x0,0x10100001040000,0x80040000500000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2000008001040000,0x2020000000102,0x0,0x4002000008008,0x0,0x0,0x48080004000,0x0,0xc000000804000,0x0,0x0,0x0,0x40020000004008,0x8010404040004108,0x840000000000012,0x480000040008000,0x100801000100400a,0x800010000000c000,0xc840240000048032,0x88241fa00a04812a,0x5000070881042200,0x2046d51269440182,0xd4b2bbeb23ccc1ea,0x1000a258c9042200,0x169401508344400a,0x0,0x40e190405081c020,0x98885130288c122,0x4004000000008020,0x0 },
	{ 0x4848410240422190,0x18663291c0cab624,0x669bc5f044c18214,0xd4be50480ed07a00,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc18cd40001030a0,0x2cbec6fa8ecbb2fe,0xd9fe7f724cdaf628,0xffbc2fffcedbb67c,0x384c4f48213b224,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa910358286c02828,0xc306f77d46c0f870,0x88f23e26484aba62,0xa2200b4306c28a50,0x56a61e804403a230,0x20005014400000c,0x8a16273ec6d3f620,0xe856af3fc6cab8dc,0xe3d6f7dbce8bf6ec,0x1d9665ee8ed3b3ca,0xf17fbfea4edbf7f6,0xa1e87340203f2a0,0xf5d2f232ced3f2f4,0xefffffffcecbff6e,0xcbab4fdbc6dbf37e,0x10e8cd104ed01014,0x12c6bd40c8d23f84,0x837e738906829300,0x69f77fc24e9ab38e,0x99ac55b846c3b2c4,0xf67df671cec3bb94,0x31884d7144c22032,0x535066228a80a2c0,0x8e080040c802220,0x8e9f4f7d8edb3a32,0xa05316e08ed9ba80,0xc426fd404ec20234,0x10886c040e002200,0x8a43c22ac4ca1034,0x24084a904ec0a100,0x345558e4c682b100,0x1207a4100109180,0xda807ac64ec2a300,0x800210080c480000,0x208108140105004,0x2091042008402880,0x100a41242005204,0x428224000cc03200,0x4000400040800000,0x140020004800000,0xa5a24fe08443a724,0x83c16a184c2ab00,0xaa9025504cc38a00,0x800182644482a420,0x6900070804828600,0x200060948008210,0x20340060c4802020,0x0,0x0,0x0,0x0,0x1004044008020240,0x0,0x1000004004020000,0x0,0x0,0x0,0x0,0x8000001040800000,0x0,0x4010000000020040,0x80040002002000,0x0,0x200000200018000,0x200400000400800,0x0,0x0,0x0,0x0,0x100800000020100,0x0,0x0,0x40040000402000,0x0,0x0,0x0,0x80080002000200,0x0,0x0,0x0,0x800020040000020,0x0,0x0,0x200040000002020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000240100121040,0x0,0x0,0x0,0x0,0x0,0x8000000004800000,0x0,0x0,0x0,0x0,0x202000081000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200000280000004,0x808100100028240,0x0,0x0,0x820004002000,0x0,0x0,0x200040004000008,0xd4600e802208,0x416040921000,0x82a43b4280c39a02,0xf44c200404222b0,0xe35e7fcfcedaffec,0xb510012082c0f220,0x4061081088c00210,0x10000a2040020120,0x140043886001000,0x30cc20844838000,0x14d35b224693b7a0,0x40123800089200,0xb2c044200600f2c0 },
	{ 0x364623618723981,0x5c4f0a194472b211,0x86914ea222135a03,0x4010834008700282,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8cd0d26205d12a00,0x7d49afe905c16e20,0x4002001000800,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x600600500060200b,0xd3923d925dd27a1b,0xc28c864c2d02bb09,0x50420004002001b,0x8edc0c4a88327f01,0xd062f6e5c9321e83,0x748a7c00292b291,0xfb34eff75b73ba1b,0xefffff8ac9d77b8b,0xef5bf7df5ff6be9b,0x1cf75ffb6a267bbb,0xd37a2fd2ea773a81,0xfa9e2ff338521f33,0x8fdef338ce73fb9b,0xeffededf7ff77e9b,0xdfb6493a8cd37f9a,0x91bc5f610a529b0b,0xa73e7de518579a01,0xe8d072040843ba23,0x6bffba1b09f3be81,0x3abae6ca0b337a01,0xddfff6fecff73e9b,0xc5087d006d527a81,0xa4c2f2e88771b0b,0x8763ba797c53bab3,0x97b9c9705b729a31,0x54b27ee919f23831,0x1008ae504870018b,0x9b822b5f1ed33a0b,0x4302260204513221,0x6910c11248702001,0x9437344e08527a8b,0x900c40600003380b,0x102c1008210300,0xf4e4e10e0a235b91,0x2240415840521201,0x430c4ac4c12380b,0xc800246000021010,0xc10204000c022100,0x4301f20208721e01,0x8208000001,0x0,0x100c724802521601,0x284340014013200,0x1028b2012423201,0x204012180c103100,0x4105413208d25a01,0x9050326908701001,0xa80020001010a00,0x0,0x80800000000201,0xa00021201,0x20004000001001,0x0,0x8000244001402800,0x2004e02008021090,0x0,0x0,0x0,0x820000101000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x120008000001,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x918a410700411221,0x0,0xa102000080420200,0x8010000008000080,0x0,0x0,0x8000800000000202,0x0,0x2008000001800,0x0,0x0,0x0,0x4000000140000001,0x800010000001001,0x0,0x0,0x0,0x0,0x0,0x48000000600,0x0,0x0,0x0,0x80800008000100,0x88001000800,0x8025000008400281,0x828001000800011a,0x0,0x42024004012800,0x20060008000810,0x94e682000501a01,0xf4902b6618523691,0x6602d8ea49621b03,0x6755d65a8a761aab,0x9585450044d02601,0x932882200812118a,0xc000200208600200,0x400200000004800,0xc54108000761608,0xca2554008901900,0x1080100800004a02,0x80004000800200 },
	{ 0xabfefa2b6a767ddc,0x6431a0024521184,0x2020490000106a40,0x1022001024020000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xb5ddd66be27e1d4e,0xaa147d63a276154a,0x9924132a02340846,0xc9c18a28624610ca,0xaa84c93622e0858,0x44840028001010c6,0x8320260002664858,0xb5322f7c6377184c,0x96658a0420561484,0x81c514e22661550,0xa8b20448002c0040,0xfffd877227761158,0x4c0207682352905e,0xeed3ab3e6676dcde,0xefffffcfe67f5fde,0x959beff9a27e6dde,0x9fdc1738626ddb48,0xeebe3fd02677b5de,0xc8d4eb6c626494c6,0x4eb65b5f6434194e,0xd2baefe2667e32d0,0xe09676fc207d1dc4,0x82a60d6060761444,0x8858110003542082,0x84ce274862694946,0xefbeef7b676f1fda,0xd176eaa2345cd6,0xe9b4e7f2a6761bca,0xbba46292e76811d2,0x4945a06002411ce,0x9b6c050000660040,0x421cae002143146,0x20256060061100,0x4a021a0002162240,0x800000400001080,0x908080924521050,0x8a240022100144,0x100824000260048,0x12a0051062140048,0x8f80d60222741480,0x10c000460000,0x809a05ada208805e,0x910cf40e21705c86,0x1840b6422040086,0xa0542366016e0940,0x4084022040008,0x481220000148,0xc0502230606249c0,0x0,0x800002000001100,0x0,0x0,0x10001000100002,0x108000000420000,0x200010000000006,0x0,0x0,0x0,0x0,0x200400000001400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100100000000006,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100002040040000,0x0,0x0,0x1000000260000000,0x0,0x0,0x0,0x2400000000400008,0x8010000000600000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x12014000501040,0x0,0x0,0x200040000000050,0x0,0x0,0x100020040040000,0x0,0x0,0x0,0xa00000000000006,0x800404820440080,0x400a1020000082,0x0,0x0,0x40241000201900,0x5290af66273c7cc6,0xe9d5e636e27e304a,0xcb7f5f9aae7fb1de,0x18b2c7c1246b1540,0x408740226200c2,0x2020044000241100,0x880900008400140,0x20d21a5a23741ac4,0x400244000101000,0x12000000006,0x4000241000200040 },
	{ 0xe331e10bc91c8fb2,0x4090162885104c60,0x21d441008006c58,0x9003024000540300,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x140000000000440,0x8f6c05500a19c6da,0x408c16e9890cdc9e,0xfabe7df39f1cefff,0xf3bef7efcb3cefdd,0x4dd7feae8f15ffd9,0xe1d5df5e8b2dcfde,0x202080610000cc50,0x85c6a7108f0ce911,0x1a141e6c1b017618,0xcc6fe228c2ccaca,0xbfecf5c4b3dc6d7,0x64ce5febcf3cddde,0xb9eaff731e5ddedc,0x48c015409146480,0xcfc7f2be0e3ce7d7,0x7ffedecbcb3cffff,0x77bbdff91f0dffdf,0xb8efbf711e3dfdd7,0x92907d53971c4ef4,0x8282ffba1a3cb691,0xeef6b6cb4b5ccf97,0x31fcd7e1b0ccdda,0x9cf796e89f1c6fd4,0xcbaa3de21a3cbf94,0xc1821b145a3ce4da,0xea67be3b9238cfd9,0xbbbedfdadb5cefdd,0xbdd71eed1f7cfd7d,0xd3fffd739f3dcf9c,0xea86b78e1b2c89f4,0x875382004a38cc50,0x2ff29dd3170ccd19,0x109e7ef2521ccd98,0x9083f942431c4fc0,0x8002a0044104c00,0x44880332120c0010,0x518d3d08109cc51,0x242008a01220c010,0x120825600301c810,0xc48101c001014c40,0x650c800003008e90,0xa6e0001010c8010,0x80100a31022c8518,0x24000008000200,0x428aa5628a0c8cc0,0x1038b20020c9490,0x5080861000108488,0x8a07c48b1a0dc694,0x13682480a34a780,0x20004000000408,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80080004004000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20080000100080,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80100008002000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x24000100080,0x0,0x0,0x8800000000008002,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8200000012000000,0x8000020000100020,0x0,0xa00008010,0x0,0x2050040100c00,0x0,0x0,0x0,0x0,0x404000000c00,0x0,0x0,0x0,0x0,0x0,0x20408000008,0x8000004000001010,0x10000200100200,0x0,0x0,0x0,0x0,0x2020002000100,0x0,0x0,0x0,0x0,0x0,0x8040000008010,0x0,0x418615701a28b418,0xffbfff7fdb7dfffd,0xefdfffffdf7dffff,0x2a3ed6dc033cce2f,0x15051b814510ccb1,0x0,0x2a34be109a20a641,0x8000300082000900,0xab6e1808031447d8,0x20220d811200d810,0x410000390a000640,0x2141088000cc0 },
	{ 0x4bc1f5e84128c7a3,0x687796401182052a,0x402004100000,0x260040010a0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc0edc12e0db85240,0xe8e3b470ec9a0d6a,0xfbccdf33d384ff3,0xad7d33ec6daac3ef,0xdffeff61c5badfb3,0xfb8ff2190d9acd3b,0xc5e2be31fdba9f23,0xbfdfdfda5dba1f63,0x7c9e5ee9d5ba45f2,0xfbea2f7bdd321ff3,0xfaa5660fc11a4f6a,0x6146460811220542,0x6944c9404cba0d44,0x247e113101084702,0x880c742029000502,0xd032257005181f81,0x29c6e92acd2847f3,0x2bd3f9c74caa476b,0xbfb3df6999aa0ffb,0xd3dfb6dbfd329fe3,0xea8c3560549acd33,0x98c001210380402,0xb0da9fdfd0b84773,0xf638d1237db29d2b,0xd5e6b1ebdcaa0feb,0x92120d50f5200701,0xfbd6f562dd3a5d72,0xddca1e0d75a80fa2,0x56bdc7fb293a4ffb,0xf4d4feea6cb2cdc3,0xe1382de0158a8f2b,0xc3936206048a1ba1,0x802ca2001220345,0xedf4cf4914aa574b,0x44f5f76a1528150a,0x19045c0000120e01,0x20032c08080303,0x6c0f102041886c3,0x4256135095a21412,0x81005800120001a,0x9149ac1410288933,0x489025100100030b,0x29448628092881a1,0x401e5000c40015a0,0x400a01632482030f,0x314105a0112a454a,0x32823e4411284701,0x100c00904081303,0x24613514120094b,0x2ab06b5205285303,0xe415c02094980f01,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000002000000100,0x0,0x0,0x800020840080020,0x0,0x0,0x0,0x0,0x12040000020,0xc00000000002,0x0,0x0,0x4000601141020902,0x0,0x10040001200000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10000800000000a0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x11000280000,0x0,0x8000100004400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20080008000100,0x52400080101,0x0,0x0,0x200480800020241,0x0,0x0,0x0,0x800080080000400,0x0,0x0,0x0,0x0,0x0,0x0,0x9032f84e212a81c3,0x414e7403880a0d97,0xfb567b79489a8fe1,0xfffffffefdbacfff,0x2d02d0814c8085b3,0x857a5020702007e0,0x4008b464c1220572,0x2090001000100402,0xa1040214448a0025,0x90023211043809e0,0x11143e40c0a0126,0x3870726849220362,0x804240005000102 },
	{ 0x104010280120004,0x4c2104004126020,0x930846121482aaa4,0x97b0a25434d3c6ae,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa220044024106020,0x0,0x0,0x0,0x0,0xec4ca470341278e4,0xc944f52c3c98f6e2,0xcc52b218341176e6,0x11bc4b7534975ee4,0xb0fe57eeb492efe6,0xdf9ffd70b4ba6eee,0xdfb3e37eacd37eec,0xc5ffb8bf34d1f2e6,0x28edf8dc8c90f8e6,0xc4b43ee23082ebe6,0xb9ce7ff334fae6ee,0xefd0eff7bcb3f6ee,0xd95e6f1a34b3fea6,0xa641c1a081042ec,0x80050ce004903140,0xd1ae04203411d060,0xca902e323491f06a,0xc11bf54914bbfcea,0x6e588a48308a46c2,0x200022800004940,0xc4809070088a4ce4,0xdcbe77f3bcb7ece6,0x88991221303222e0,0x80131a4220127044,0x88410022c000,0xd571fc73809aea62,0x5ff675723433ffee,0xdad4f32fb4955ce0,0x1843a6a104d1d0ec,0x694c358109cc042,0x90d1dc01141324e4,0x80062d60a09262e4,0xca91260e14905ae2,0x84c28f3290908ce6,0xceb75fde38dde5e6,0x9cde7ee31492c7ee,0xc1a6116a34b6e166,0x8040010000002,0x2020000104000,0x6000010000400,0x490002000c040c4,0x84c4022804904aa0,0x40a420f008d26824,0xc842e22804f460ee,0x8550000200920260,0x6181742100004ea,0x1004201614904000,0x3a60c4c00914062,0x8983c36b14136002,0x41914a15149653c4,0x30290101340264c4,0xf307a4e834b8c4e2,0x1200000020000040,0x0,0x40001000100020,0x200040000220000,0x4008000004800,0x1000c1014200400,0x300400104900020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x800000800110000,0x4080100000020080,0x0,0x820000000000280,0x21000000280,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x60000000110000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2105000944000,0x0,0x24020800882040,0x4002000004040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000080000004040,0x0,0x0,0x0,0x8002000900000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000032000022280,0x0,0x200000800022000,0x0,0x0,0x0,0x200001c004022004,0x0,0x0,0x0,0x0,0x0,0x0,0x100820200002a0,0x25a617f014bacae6,0x81101c65a0954062,0xa5dfff2a3c96ffe6,0xea7e955214926bc6,0x513512a38926062,0xe5c6a660340044e6,0x40020134041024a0,0x6956677a34126ee4,0x2161800808802e2,0x8010227028026060,0x41003031a400486c,0x4c341a4490110462 },
	{ 0xa122850008460,0x89e4080152105cfa,0x8edc4e03003433c2,0xfc7fb672faf0d66c,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd9eaff741054b6cb,0xd3be3564329c7cfa,0xc181a243a00454a7,0x0,0x481000810008240,0xe50df060029d74fa,0x45264dc2f80411e2,0xc386e7fff864f568,0xefefbf5fd05cd7ce,0xfabb4d7a30b89fe8,0xf5e7ffec7a74dfee,0xbba6fff2f23cf7ef,0xf7b5ee3370bdd5fb,0xefcffe7bd294f7ea,0xa4ffcf4a70d896e9,0xfcf5defb5034bfe9,0x797e7561701536ee,0xfff6ff3672d4d7fe,0xcd7ff73ed8f4deff,0xa05bdfda70a0b7e8,0x14a758a8020036c0,0x141264180501218,0x81292080001248,0xebece3be2afc777a,0xabde8acb10cc94ca,0x6200040000404c0,0x8000200010000400,0xe20225854200d462,0x89a1d33a9804548b,0x62534b40522410c8,0x35a8cf01301d16e1,0xdb6ca756c8bcd66a,0x7ff67df3e05c53f2,0x92800a1c001456c8,0x85017500004d08a,0x611467a504d146a,0x2011c071100056c0,0x192c6f7002b4f2c0,0xcbb78bbf109c9ede,0x494630a28054d6c9,0x4830480020000560,0xb0b4d661125c56c0,0xd94cffb29224b7c2,0x284100000401404,0x50420000000420c0,0x208000000000140,0x48000100200,0x1004444a00801541,0x44082822200c5460,0x44c1e00800941541,0xa8c60980128054c1,0xb41557e3b081b660,0x80092210805c5080,0xa802a00000c4240,0x81224108100406c0,0x10e5010c01080,0x51a81c3804034c2,0x64272024089c1240,0x0,0x8110820218109010,0x0,0x401000004000,0x0,0x0,0x200000200002040,0x808180080801000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x800100000240,0x0,0xc000000001008,0x4000200000000404,0x0,0x0,0x0,0x0,0x0,0x0,0xa00040008,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8020050000000,0x20000100010,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200000240,0x20040000100200,0x4000019002000040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1000020000000240,0x8010000001008,0x0,0x0,0x8010000000004000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20000801000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x500000004000,0x0,0x0,0x48847520d8101ec0,0xd88c280e38101463,0x6dc5ef9210141fff,0x479cd71b50149ff6,0x9cfc5771b2b456ea,0x937e264018941f69,0xc88e2cea30bc14fe,0xc1c4d93e720436ea,0xac15388e101c34d8,0x9004b6100416c2,0x9488240450181062,0xb8e6fc2727498e2 },
	{ 0xa0c641698081aa68,0x834e0811d1693142,0xa084d93412d236c,0xe9e6b678d1eaa3ea,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x14b24ee6f9b42356,0xdb7f1568c16f33ca,0xe056354c4102a0e2,0x0,0x21e308c8c8602a44,0xa7bedd69d8f9265a,0x14317644d9490340,0xc7863d62d9e42544,0xfb6cf72ad9d5beda,0xfeffbafff9fdabf6,0xbffccf5ff9fdab7e,0xf7f7f7ffd9efabde,0xd3feff73d1ff27fe,0xffded7f4d9f5b3fe,0xdbdfbb1bd9edbbde,0xe3ffdfd7f9ee234a,0xfdbdff6af1ffb7de,0xfbe7fffff9efa36e,0xefbeffe2d9eea25e,0xadf1ff1ae0ecab7a,0xaf7cfd9d1ad22fa,0x2008a80082800,0x4024804080412208,0x82122400c0402480,0x81220000c0100010,0x4010000040040000,0x4415ce1f1613202,0x20020040001000,0xde1225dcc8e53244,0x180c0004000a040,0x8ed0061241450278,0x9696c738d0ac2084,0x0,0x12000202000,0x110002088c00000,0x0,0x2128489149442204,0xb4100274f1100210,0xeefa3400d100a864,0x80823040040218,0x214db212d1652278,0x11dd881d1802006,0x84103060c1040a50,0xddfe7f62f1efa3fa,0x5b1805b0c1152234,0x99c6b20ec14ca362,0x4014525001008252,0x62b0576af1c40282,0x1100a2400144220a,0xa000458c08000c2,0x458484200040a354,0x62469203e0892340,0x20100162c0400310,0xc030f501d1908020,0x88903520c0e42b10,0x2042080040808,0xca5e51c8608270,0x5b0c098c1008860,0xe0558260c8448270,0x0,0x0,0x40020000002002,0x0,0x0,0x0,0x0,0x0,0x480080000800,0x2200010800000,0x0,0x2002c0000480200,0x22c0088002004,0x0,0x0,0x40046000240060,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80100000000140,0x0,0x0,0x100080080010000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10100040000080,0x801000400100,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000100000008040,0x0,0x10000000210,0x0,0x81000080000000,0x0,0x0,0x0,0x0,0x0,0x800010000400100,0x804000000840000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200004011000000,0x0,0x244840002004,0xc00000000140,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x18020000200,0x0,0x0,0x2002880090000210,0x1080140080042008,0x40823500608480d8,0x9bb44f1cd14ca490,0xfc7ff3ed8fd3f5a,0xb5a1f5bc9edba42,0xb0b205e0c0f6a348,0xc409710251802142,0xc7b6136cc84d2656,0xa85cca0de1c3b646,0xf617db436144a2ea,0xad1275bfe1af2a52,0xdff6bf1fd9cba37e,0x18040c0260849010 },
	{ 0x8bf5ef6b0b6ce522,0x8c1f3a6b5fa42732,0x3fe8cf93d2e0bdd0,0x99d28e681ae8e530,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xac4c504d084000,0x5fbacff7dfecef62,0x5040365290acc530,0x0,0x0,0x42d9594f1822b520,0x1b280c009000ad20,0xc23cac6815a88500,0x2fbc7dcadfeafd20,0xdaf5ff7edffe7f64,0xfffbfeaddffce7f6,0xe7bfcf6bdf7cfdb2,0xfdffffffdffeff74,0xfffeff73dbfeefe6,0xff7eff7fdffaffe6,0xcdcefabbdffcefb6,0xbcbeffdf9fbced56,0x7cfafef7d9acff12,0xdf2e667ada66e772,0xcd3a0a705ab87542,0xc2c1ba0215ea6940,0x2280814808284520,0x1c3775a4d5ea7f20,0x500c060048240400,0x206142490000030,0xc00000003000,0x10040000000400,0x13008210800b400,0x200100008100,0x5c090000120,0x4000000000002400,0x4202004010084020,0x308028480006500,0x60263090003c4100,0x0,0x50000c00004700,0xcac32ad89a30d500,0x494c1db48a8ad00,0x151270fc8908e724,0x44920d400128a500,0x530065040888c960,0xc00c960c15242412,0xa28004210009010,0x10645dee8024a540,0x2124052a1824f404,0xb20271698305122,0x40c2060a0a802500,0x60000008400000,0x3c395f404008aea0,0xc179050008208580,0x5680024218a06500,0xc482a02090ac8500,0x150a12080968b402,0x840210000000e100,0x50c03770900ac100,0x40802d80520aa500,0x8120210b9024a920,0x60000a8000b400,0xa904a42190acc14,0x2404326d00889500,0x0,0x44000000008100,0x0,0x0,0x48010000020,0x0,0x0,0x0,0x0,0x0,0x8040044010000400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10020000000120,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400200000400,0x0,0x41021000202400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1004000000402,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x440000000000600,0x0,0x100000190000000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x242010241000,0x0,0x0,0x21000084000,0x0,0x0,0x0,0x0,0x0,0x6040002000,0x0,0x0,0x820000009000,0x18040180002120,0x0,0x9010416090268100,0xcf307774dd76bc10,0x88ede186970a3520,0x685e5dd0cd60a480,0x4376e61412c3f40,0x11a005280a20c900,0x4e9222d885286504,0xcf46e32c50ec7560,0xa008108008086420,0xaa294f019040ad22,0x916827f09a486524,0x101005d00030a440 },
	{ 0xdadff7682fd24d61,0xcc31b118a8c4f61,0xbb9aeff317f76d60,0x88505461118e5160,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2fffffdb9fdeff71,0x1ebbcffb9fde5d73,0x0,0x0,0x0,0x0,0x0,0x0,0x43362df0159a5770,0xd940f73205065562,0xed6bb74f17165d73,0x6fbedf62b3de5d62,0x77fd7efcb7ce5f73,0xdfffff731fdeff63,0xdf06f76f0b525563,0xc182c23001061020,0x69a4d5d211961560,0x4cbc5eee8e8e7f61,0x99cad37214964962,0x82387f1607966f70,0xc549e92a9fce4470,0xc2d48c43069a6732,0x14b55e6429864f62,0x90b4a258a7406b62,0x22625c007961640,0x2016002884240,0x0,0x4006a1002060440,0x8050054006020460,0x0,0x10100000020800,0x86c00400a860040,0x8408101020900,0x801002000400,0x800000202002000,0x840000001020000,0x0,0x820a0502a6006461,0x40c634e901a61d61,0x2432741207968d60,0x578e79fa9fd7fb60,0x100800002028020,0x210014c002a1060,0x10c0d8cc324c6121,0x95a03f200a4e7020,0x7600276801424d41,0x481090806080530,0x88040100021400,0x106200010c40,0x0,0x8804043810825160,0x8000000400020800,0x800118002000440,0x84e41108c0420,0x9181746bb28e6460,0xc022054809860740,0x4193800620404023,0x4e0000800064021,0x4180160000a4020,0x140920a06000511,0x200240020020060,0x0,0x140000100100,0x210000000020040,0x4100000802006002,0x582004000500,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1004004800404900,0x0,0x0,0x4400010800000,0x0,0x0,0x0,0x0,0x0,0x0,0x2004000020040,0x0,0x0,0x0,0x0,0x100000800040020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000000000000140,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000160000021140,0x162600640ec6c940,0x8800000000080020,0x8100000000060,0x8020040000000470,0x0,0x0,0x0,0x0,0x0,0x500000000820,0x804000000000401,0x0,0x0,0x20000080a000000,0x0,0x944102202404402,0xcb3065528206fc01,0xee56f2f89e9f2d60,0xa9c5f4213865d62,0xcb9e46501c65160,0x7080a660168a0a40,0xdb80277abac64d72,0xadc5e21186d87b00,0x21c25a9104064963,0x2000cdd107cacf21,0x532c2f28862a6d53,0xb2e344e98861d40 },
	{ 0x18ed370c7006408,0x826612c2d42c6f49,0x10c69244082c02,0xb4b62aa072b03480,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4ecbbe30568e38f3,0xefdfffff5fbc7ca8,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x7d7fbfebfebc67f9,0xcdfefd73dfbc7ce8,0x0,0x0,0x8002091342282c00,0xf4f7f7ecdfbe74e8,0x905f2d71d73e6de0,0xdbd0ef76d51c4768,0x4082de3076106cd0,0x255036555e1a6c80,0x80294420d6102e00,0x40200000000408,0x0,0x0,0x60121c1093104460,0x81900cea91a80660,0x0,0x4a00001040820040,0x8000400008000400,0x42081761c9943480,0x1011000291000800,0x0,0x0,0x0,0x8080000190000120,0x8000040800200400,0x845204ec56902420,0x8025042d4000708,0x800620c10000448,0x144040800201400,0x801290000300,0x50bceee8f3102322,0x4964963895b427c8,0xa962d1092146da9,0x2041c010b1100100,0x88901102001400,0x40017e0d4000142,0x1840a30e64100422,0xca802c2244001442,0x100800010000040,0x42000000104000,0x200046012002200,0x120404c040001420,0x71062c6244302449,0x8ed420254200548,0x4a200d50720c242a,0x8090800002040480,0x60000002001,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2020080000100,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80181010042001,0x0,0x0,0x8100000040000400,0x0,0x80000104100000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20c080081200140,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4002004100000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x14040020000,0x0,0x0,0x10040044000000,0x0,0x0,0x0,0x0,0x0,0x0,0x2004000000100000,0x10001010000040,0x8800012490080500,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc000000000420,0x0,0xa000000080040000,0x400904044002001,0x9f4e7762f48c35e2,0xeb387d72d71c6500,0x4906919c121c2e20,0x2aa5f548d23c3dc8,0xd4fb7fe9f71c7df8,0x1000504044000002,0xcacc05485d041460,0xc223621a802760,0x15338005e8035d8,0x38800183951c2900,0x122ae578f6106521,0x99c60f60d3942fb0 },
	{ 0x8134d778ab902fb5,0x4ecf7e71ebdc273f,0x67f8cd62986d229e,0x8c60020442802b2c,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x280000008000200,0x10280000008,0x0,0x1004000000000c,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd4f4726882902357,0x8001200082000204,0x0,0x0,0x0,0x804080000010,0xc17bf7f29b9c2b0e,0xdfbc6ffeabfe233f,0x4cc7ee188ba822f9,0x398de1b91900ace,0x2d414600280220f,0x4000305283000024,0x0,0x0,0x802200002002005,0x49404a009080208,0xd010060202022604,0x2100002000004,0x0,0x200004000800010,0x0,0x440022001002020,0x0,0x3a0830282a0220f,0x40431a1180800238,0x86cb44009a182204,0x2048482000110,0x50a300082000102,0xe30016028a800b06,0x40232a0a2802012,0x837c0a5aabc22032,0x709075e48b08205e,0x5126a67293a0023e,0xa00362c2002010,0xc4309008a382604,0x4444a82800010,0x9013048282000216,0x990c206acac02407,0x1000000400000208,0x18584001200240c,0x40024802002005,0x3011402080000386,0xd008a4009a40030e,0x4690574443802006,0x8303ea269b6c260e,0x4007101008602600,0xb0144ba08a40021e,0x3000064008c10012,0x0,0x0,0x0,0x0,0x202c008c00010,0x200201081000012,0xeb0477f88adc2a4f,0x800870001900010,0x940000009400000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100e00201000248,0x800004000002400,0x40200080200,0x0,0x0,0x0,0x0,0x0,0x200102201000020c,0x0,0x0,0x0,0x0,0x0,0x0,0x5000000002000004,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8600000001902000,0x0,0x0,0x0,0x12000000208,0x0,0x0,0x44001000020,0x0,0x0,0x8000400001100000,0x300000000208,0x0,0x804001000008,0x0,0x8400002a0100200,0x2020002000000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200000800002,0x20800000300,0x0,0x0,0x0,0x498c7d628ac22205,0xf314667d8a9a2379,0x4054b28689182308,0x6a5cdadc09f0226b,0xfd3ac7ffdb9b22ee,0x858d06681bd00114,0xdab02f7c0baa2a1b,0xd0e0e522991022ac,0x0,0x88818149802110,0xd18826da8b84230b,0x2890054282a02320 },
	{ 0xc96ae90a790cd11a,0xd8763e26cb2453a6,0x7b3c7f728ba4d3ae,0x80438ee0c184d128,0x40000000120,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4080a7c28a005122,0xdffebf7fd9b5f3be,0x46432e188980538c,0xbd81f108018572a,0xc8960120c321d200,0x8140264003005286,0x88012040008000,0x80800008004000,0x80e0080004006,0x0,0x0,0x800040080200000,0x0,0x8004000000204,0x6000100200,0x0,0x4402004001004120,0x18180104000d080,0x807000c280205182,0x4614c10680041162,0x51ee2648c090d10e,0x895ead32d834932a,0xf9b8c06a4b04c12a,0x8448228c110c0e0,0x212040e83240886,0xccb230b84080c02a,0x9021542c120c0e2,0x10050063000000,0x855c000012052e0,0x40100048e0200000,0x810090180205068,0xd1001e48c120c284,0x18001004000,0x100a020000041a0,0x10100000000120,0x101342a0e020c00c,0x42c040060001120,0x48bc55d0f2b1d32a,0x2002000100200,0x40022400c0004004,0x9299cb23c305cb2e,0x1140b724812c5080,0x40001000080,0x0,0x0,0xc00000004002,0x0,0x0,0x11000008020,0x482020000804060,0x1044280005008,0x813040200800410a,0x0,0x8000020000084000,0x10800000000a,0x2100000000110,0x80000800000120,0x0,0x200160004000,0x4002000000004080,0x0,0x0,0x0,0x0,0x0,0x80800050000000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x240380104080,0x0,0x0,0x20080000008008,0x4000800008000002,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2800200200,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20020000000120,0x4002001001040204,0xa02060000120,0x80100080100000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x202100180000908,0x0,0x0,0x94bcd7e2ea14d122,0xda947effa995c1fa,0xd1b4b10af810c386,0x1ffdfdbfb2dd76a,0xbdb7cff2db14d3be,0xf9d8af03c915d3fe,0xde920d72cb34d32e,0xc1c4e4210390c3ea,0x480000004002,0x123006408804c822,0x100002000100200,0x8000250041004080 },
	{ 0xc90875da83599a60,0xefdf7ed3cbd7f3ae,0x169d4fb381d7fbaa,0x2f62dc282d4f220,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40a061200868260,0x7f56dfdb4ad6b78e,0x3cb377a2c89ad224,0x93022748ca025224,0x202004c0a003000,0x800020080000004,0x265091ca8242a3e0,0x140005600200a100,0x0,0x8000400002008000,0x0,0x0,0x20a0800002402000,0x22020000022040,0x1200000008040000,0x0,0x802000000020020,0x6d040000244a020,0x787707f808cfb38a,0xa394af008accbad0,0xa210420408123300,0x84716a1080c4e280,0x2a0d9d428854f244,0x400020008000a200,0x108242008442000,0x923a2a128280a706,0xd8c234a80a14b2a2,0x2100c01a8000a200,0x90130de40042a340,0x10e05280a049210,0x3008042108061240,0x84c422000202a220,0x400000800005000,0x0,0x814c12288280a200,0x80c08002220,0x2002080002004,0x4200160280083024,0x10288080003080,0x852ca0418a062a22,0x100080000101000,0x0,0x2200000400002,0x0,0x0,0x8108200000828020,0x80811008081000,0x100800000002002,0x240000004,0x0,0x0,0x0,0x0,0x0,0x0,0x840000000400008,0x0,0x0,0x0,0x400000002020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000000208001000,0x0,0x8000000200020100,0x0,0x50000020002,0x0,0x0,0x480000840000,0x24000040200,0x4000001000040200,0x4020000008040,0x0,0x0,0x0,0x40800400040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x120000005000,0x0,0x100040000500000,0x0,0x80530208060300,0x0,0x0,0x0,0x0,0x800800c00000,0xa4000c200818020,0x80400000020100,0x0,0x0,0x0,0x0,0xa004400000540200,0x1000142102d02008,0x1060820828c3800,0xdf835f768207bb24,0xc8f2f3298a83bf20,0xaa7cffdb82dff3ea,0x25c516e382daf320,0xd12a7f1e8ad6e6e2,0xdaae27eb89c6f6e8,0xa2c2fea60847f362,0xa4a1044020693a0,0x802406e080d4a220,0x4c508008068320,0x4b80030401447220 },
	{ 0x8810100080840200,0xeedf3eabcdeb7b95,0xd095c560cc923c14,0x9284b22a4dc3aab0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40001000800,0x0,0x0,0x47d8571348e01fa0,0xbefedf7bcfcf6faa,0xdfeeffffedfbffbe,0x1a88042000462200,0x100c03228020a10,0x108040000004,0x8008482204880680,0x0,0x82244020014a00,0x10800010800,0x0,0x658c40948c63a00,0x100c2800020610,0x240020000200,0x8800000000808000,0x802000640820010,0x280005c028031280,0x21504678ac0a2a01,0x1200000000400800,0x6b00e6084c622890,0x54c63e1481823e03,0x800e4c0044423824,0x849514620ee8d800,0x12405200f000a82,0x9a205f4c4cca0285,0x6c6000008421200,0x24004801400a00,0x4060156148355b20,0x112026c248111720,0xce102143c9660a20,0x8140901144402220,0x80000100021000,0x0,0x2000102186802000,0x9020014800031201,0x8820320108231900,0x800000101000800,0x204026084020008,0x1128820020400a02,0x0,0x8000010004020000,0x0,0x0,0x0,0x0,0x4000200000000600,0x0,0x100030240400220,0x5000000000000201,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1004000004080000,0x0,0x1800000800800,0x4002000040022020,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000000004800000,0x0,0x0,0x0,0x400008000800,0x0,0x0,0x0,0x0,0x0,0x8000000000004800,0x202000000400200,0x0,0x0,0x0,0x40040000020008,0x1000a0040000014,0x8104000004822000,0x6000080020000,0x0,0x0,0x0,0x802000400200,0x0,0x0,0xc14000040420200,0x0,0x0,0x0,0x0,0x0,0x2f1c084422208,0x1ba66f6549c26f81,0x89f4932006533208,0x8ece4fd268414fa1,0x47bb55a60ccbae83,0xc1da6c4905c0aa01,0xfb96afde6de2af98,0xc9d4ff6e6dc60e12,0x94a68010c822890,0x65b2cfe24dd73a95,0xa0ac58420cc38680,0x9bc47d6cc4821e1b },
	{ 0xc901c072a2421648,0xebffdfdf7a7a17cf,0xbdfbefe3a67e38cb,0xdfdfb6c9617b17cf,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa0d481000029001,0xa5fe5fe2aa7691eb,0xd12de7722c7631cb,0x20102d21204a10c0,0x4682da80274011d0,0x115000500001,0x89b247fa42508209,0x0,0xa00000000600000,0x54020000600048,0x8010044208501041,0x1481c40102029049,0x8000000000101000,0x43c2020021180,0x20010a2000800b,0x800020022000000,0x60060482a4000048,0x2c0200a2001182,0x80424112000140e,0x13060002404a1049,0x4dfab2116552155e,0x355005466a000149,0x4b1d9c2b2409703,0x43b670925011cb,0xd426264408461687,0x8208ca002b101083,0x51c464020423080,0x2080002880500002,0x36f0668b070904a,0xcc82338ca852104f,0x8010000000001400,0x0,0x820000800401483,0x2000020020000008,0x4800400200,0x44000000080040,0x62020c0000501148,0x80401020400028,0xd0348464220814c1,0x0,0x0,0x0,0x1010140060021000,0x0,0x880000600000,0x210000040000001,0x0,0x0,0x8010000000001001,0x0,0x0,0x0,0xc00000400008,0x0,0x0,0x240000420000,0x0,0x4e0060000082,0x200040022000000,0xc000800008600082,0x0,0x8000000000021000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x120060000000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80000800000082,0x0,0x200010022000000,0x0,0x0,0x80050202600046,0x0,0x200000440000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc000020008000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x115020420008,0x0,0x0,0x0,0x800002024000000,0x0,0x0,0x0,0x0,0x0,0x0,0x80010002000040,0x2004024000420001,0x84625e02602184a,0x111a2f007a021049,0x68c7c044215a10ca,0x6f9cdbd8ee5411af,0xb5f637e0eb72929f,0xd8f0bedeab5293df,0xda366ef1e24417cb,0x8384d695e7121081,0x8a1adeb3704813cf,0xabba477a6e7b12e7,0x9b18246648721141,0x6d125c2164501693 },
	{ 0x6902fb22403558d0,0xcc05e8b0a3c8dd0,0x1ba9dff369361ed4,0xd67ff2f80b3dece7,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd88da57641375e54,0xfebf7fffcb37fbf6,0xcb5bfffec3375dd4,0x0,0x1420424000040c04,0x41409200000c00a0,0x5e04552008048184,0x840000008200000,0x200004000000c00,0x0,0x8020000100400,0x44b00458e030a810,0x80008310082c0100,0xe0c29206603400a4,0xf9fcd8168348236,0x4000206008000c40,0x630a05002348040,0x888801220a340200,0x41038c1c086caa90,0x38160d84611809d4,0xf4351ef4097cabd7,0x85cb5560482e06d0,0x41a325140936980c,0x4482700800548094,0x31a5a1041140c54,0x8c923481400c0c50,0xb84000001c00c4,0x8880004400008150,0x44000000000140,0x2a00010e08362080,0x10002000080010,0x0,0x20590001c00c0,0x81508000000c05a0,0x16020000280820,0xba0b820829c840,0x10400000000040,0x0,0x6000240000,0x0,0x4400002040000,0x2860000000,0x1080000000100020,0x0,0x8020000000040004,0x0,0x0,0x0,0x0,0x0,0x8000040008200000,0x0,0x1100000200101140,0x0,0x40080000000140,0x0,0x0,0x2004000100200,0x0,0x8000020000200800,0x0,0xc000000000050,0x0,0x0,0x0,0x0,0x0,0x0,0x80020008000008,0x0,0x0,0x0,0x0,0x0,0x0,0x800020000080004,0x0,0x0,0x0,0x0,0x0,0x0,0x20004000200040,0x20080000100080,0x0,0x0,0x0,0x0,0x0,0x0,0x120000100400,0x2010e4000040144,0x10000100400,0x1001000008000040,0x0,0x0,0x0,0x0,0x1000040000200800,0x0,0x0,0x0,0x0,0x88008000040,0x0,0x1280000008000340,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10800240000,0x0,0x0,0x0,0x80e75694b3cacf5,0x8f914bf2082c8cd0,0x50d6f63288398edb,0xab5b5e934b4e8dd2,0xb5f65fefc97e8ed4,0xd5fbbfe8eb7edef4,0xda921ffa493d79d4,0xe3d7d79f4a3eaff4,0x900e8a8f023c0874,0xda0f2561483ca7d4,0x501c06c2681c13c0,0x729e2be24b340878 },
	{ 0xc022d222138a048b,0x4eca1e791b02cf1f,0xb6bdcde3361a4f4b,0xf5eab762171bdf5b,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x290072180209c589,0x6b177f506998075b,0xc134482101024902,0x1002001004000,0x80800220000044a,0x8010000000020040,0x8022100000000102,0x6b2e4c431b99c703,0x3028224081100402,0x8242105004803,0xd5a0215b6180871b,0x8010d0011008050a,0x215e4f8c168a4142,0xe435447233084d91,0x202c65e212080510,0x448534000812c550,0x8010480000080102,0x2002d01080008510,0x1455932920889512,0xa21800101412,0x2010019013004010,0x0,0x0,0x18e38005024b02,0x24000001004000,0x1210000410080508,0x4000230004104400,0x408000000000300,0x8402000200,0x560402ec43080759,0x4000008000008010,0x0,0x6000000080400,0x480000000402,0x4000008000000012,0x8200000008001,0x800000200080400,0x0,0x804000000020800,0x40040002000200,0x200053002020112,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100104004000,0x44000020040,0x0,0xc0104000080101,0x0,0x14000100400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x44c00003000500,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x802010000400,0x0,0x80080000000408,0x0,0x488c6771511b45d9,0xd2943e525410df11,0xa943fea22b1acd1b,0x4b585dc9081a8715,0x35aff16b530acfdb,0x7d7ebedc5719df13,0x4aa6a6661b92df55,0x8c7e130c93029e1b,0x4dca5a18cc034741,0xe7beed943935f57,0x95b97601783c352,0x22803b6532028703 },
	{ 0xd1c1a22911b86fe2,0xeaf71e723cb96ce6,0xc792dfb3f4b85fe6,0xfe943ee3dc9c66e6,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000200000280000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1b064d106480e162,0x2065f34235bca022,0x408a15200c116100,0x320a70654900de0,0x4c131004a04a82,0xa0ae0c2410c0e942,0x10336880252004c2,0xc1a225001895e8c4,0x808010420c085042,0x80ce10052049c0,0xe48df4194185ce4,0x4a256e821b94ae2,0x61a3ab7b3cf86ee2,0x909149001c114d20,0x801280230084880,0x6014488410000486,0xa1b27b633091cdc6,0xe4650c0894286fc2,0x90013014084500,0x1c4c60014909440,0x652c360000a86842,0x0,0x1000044020000042,0x1000000401200000,0x8000a0010014440,0x4026000004104800,0x20a0040004980000,0x5605804000a8cc00,0x0,0x0,0x0,0x0,0x400002004001000,0x4800000000200040,0x840000000880000,0x8100000006000,0x0,0x1002000000480,0x0,0xa0000100002,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x44000000004040,0x0,0x0,0x200200001200000,0x0,0x10108030000800,0x0,0x10200000004040,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100800000280,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80000000000000a0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x201000280000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x820000000004400,0x0,0x0,0x0,0x200004000000084,0x0,0x10201104005040,0x0,0x0,0x0,0x0,0x0,0x0,0x4000000201200000,0x0,0x10000200280000,0x0,0x18ceb610098c2e0,0xc110333430bc40e6,0x5ed7bab0f499c7e6,0x8730ce1a35bc66c4,0x1046928110282662,0x935cae585d3064e2,0xce982f5254806de0,0x86d637fe4ca873e0,0x202dde8040d4e5e0,0x87ba6e72d4f96fe2,0xd103a67c20b8cec0,0xd38a677c30386ce4 },
	{ 0xeafefb6a8454f36e,0xa7d33acb0c3e766a,0x579dc7fbbeb7f2ee,0xa6f5f671b41652ea,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xaa4d43841556e8,0x8c545010020450ca,0x1002054084103000,0x829b4204a082502c,0x8048119ba4103028,0x86fe5ff1c23355e8,0x34240080020440e8,0x8301e581000236c2,0x4000201002000068,0x812086010023102a,0x108284200400e0,0x4511a6ab480146a,0x8102a758a2911060,0x6d934400040610ec,0x82870f08201e10ac,0x2400c000000011c0,0x400038000305000,0x28c24004146080,0xa20004800104020,0x100002000104000,0xa000020800000040,0x2020000000010020,0x1110005240401420,0xc004000294000020,0x200200104000,0x84c00e00000c1428,0x95e05deaa21f56ee,0xa171de68843612ee,0x1102040000021600,0x0,0x44c1001016000440,0x8000002000040008,0x0,0x0,0x12000401000,0x0,0x0,0x0,0x0,0x120000004000,0xc000000c0000,0x8004000000040,0x28000000240,0x0,0x0,0x44000000140000,0x100080000000060,0x0,0x0,0xa00000000001008,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40000808001000,0x0,0x0,0x0,0x0,0x0,0x200004000000240,0x0,0x0,0x0,0x810000104000,0x0,0x480000001400,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40080000004000,0x0,0xc000080001000,0x0,0x0,0x480000020004,0x80010014000000,0xa000000028,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4840040000041082,0x0,0x0,0x0,0x0,0x80003080000c0028,0x0,0x0,0x0,0x8040000000240,0x0,0x0,0x0,0x0,0x0,0x0,0x80000100022000,0x0,0x0,0xa8675420037d6a8,0xff3c7ffe5e37f7ee,0xe9f7f9a6f4fe74ea,0xea1dd649ac92776c,0x2e72577ae21650ea,0x9ce187d82092702c,0xfe862f7ec07776e8,0xa916f2b2a01553e8,0xa5d0e50041674ac,0x363e8cb6dc127024,0xd5fe3659a8f7d46e,0x2c903724228311e8 },
	{ 0x8bda730b5116a7ce,0x59d217bd511fb7f6,0x375c4dc9515d85d8,0xb5b6f6f371feafce,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1400402000c2248,0x2084050100800,0xf194cd3ad130afd1,0xc03090001104824a,0x4054450140401,0x70449040400ca340,0x8b08e04000412340,0x91802c4500298240,0xc9c6322751a48680,0x232a495151100ae8,0x2210446850648500,0x8022004240048248,0x3a1e6f2051518e56,0xc983ff5ad006a5c7,0x42121208105102c0,0x401048040140010,0x250cd368d11da048,0x2210061401000258,0x100001000102050,0x44000020040000,0x8040000000240,0x80a0080000400240,0x0,0x201050e0401902c0,0x800e0010100090,0x90944ced6106827a,0xf6defaff717597d8,0x0,0x0,0x0,0x2108010001100802,0x0,0x108000000000240,0x0,0x0,0x0,0x0,0x0,0x0,0x8000800010000080,0x0,0x404010008000,0x52010100400,0x0,0x110408040100280,0x8002000000400100,0x90042000108050,0x0,0x0,0x0,0x0,0x10002000100010,0x0,0x4000200050000000,0x0,0xc0000180000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2100000008040,0x0,0x0,0x0,0x2800000140,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x24000000010040,0x24040110180400,0x20440002000,0x0,0x0,0x0,0x0,0x200110000400,0x0,0x0,0x0,0x4886053051102362,0xa9947f74517d2fd2,0xcd2ce73ef0fdabf8,0xe3715d58707daad8,0xb7ff5fef71fdb7f3,0xb1cd150c503ca644,0xd3bf2d7551f40fcc,0x2946e3a6711027f5,0xaafed6de515da7d3,0x34b26eda6040a7c2,0x536554a850e427e0,0x688e3c3151152178 },
	{ 0x35a018c08500,0xcee736139090b956,0xb39acdc3d840a794,0x54e496e8d0f32260,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x201000202000,0x1007c0298c02000,0x600000008800,0x2024810022004,0x8800082090100010,0x40126810a03000,0x85080d8098408850,0x400400a080800040,0xa08486029010a904,0x8008840c8400000,0x14312022d8401800,0x9802824290002042,0xca1aaf06d8e8b042,0xc22a8b8040aa80,0xa20800c2d8100010,0x2401400d0001000,0x0,0x121010d0800000,0x100001000082000,0x0,0x8002c44080400c40,0x4140040090000000,0x200040080000040,0x104000000008080,0x0,0x8012000280000040,0xde7c74e898e9be22,0x0,0x0,0x0,0x0,0x0,0x0,0xa0820000402402,0x0,0x8000108400000,0x0,0x0,0x0,0x0,0x200280000020a800,0x4002000008080,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4400000008080,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2004010800000,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x21890001000,0x0,0x0,0x804016000202002,0x9b308a6e9850b180,0xcbd7c70ed021b9d6,0xab12165ad8c03212,0x5325f689082ac90,0x812c687290802816,0xa692254d98e0bc44,0x89505228c8c00400,0x2dc21a589863b304,0x428400000202000,0x90ccc7609082ad02,0xe3922d70d078ad14 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 }
};



